// Copyright 2023-2024 DreamWorks Animation LLC
// SPDX-License-Identifier: Apache-2.0

///
/// @file TransmissionAlbedo.h
/// $Id$
///

/*
 * This file encodes the average directional albedo and
 * the normalization factor required for the Cook-Torrance Microfacet BSDF(BRDF+BTDF).
 * This class provides accessors to encapsulate the
 * indexing into the albedo table. Values can be looked up as:
 * E(etaT/etaI, cosThetaO, roughness)
 * oneMinusEavg(etaT/etaI, roughness)
 * These tables have been calculated using moonshine_devtools/compute_albedo
 * Ref:
 * A Microfacet Based Coupled Specular-Matte BRDF Model, Kelemen'01
 *
 * We also provide a separate (only (Fresnel*BRDF) table which is used to sample
 * the reflection component which results in far less fireflies overall for
 * transmissive surfaces).
 */

#include <scene_rdl2/common/math/MathUtil.h>

#pragma once

namespace moonray {
namespace shading {

class TransmissionAlbedo
{
public:
    // Transmission Albedos
    static float E(float eta,
                   float cosTheta,
                   float roughness)
    {
        float w;
        size_t iLow, iHigh;
        float e = 0.0f;
        if (eta < 1.0f) {
            index(1.0f/eta, cosTheta, roughness,
                  w, iLow, iHigh);
            e = scene_rdl2::math::lerp(mEExit[iLow],
                     mEExit[iHigh],
                     w);
        } else {
            index(eta, cosTheta, roughness,
                  w, iLow, iHigh);
            e = scene_rdl2::math::lerp(mE[iLow],
                     mE[iHigh],
                     w);
        }
        return e;
    }

    static float oneMinusAvg(float eta,
                             float roughness)
    {
        float w;
        size_t iLow, iHigh;
        float e = 0.0f;
        if (eta < 1.0f) {
            iLow = rIndex(1.0f/eta, roughness, w);
            iHigh = scene_rdl2::math::min(iLow+1, COMP*COMP-1);
            e = scene_rdl2::math::lerp(mOneMinusEavgExit[iLow],
                     mOneMinusEavgExit[iHigh],
                     w);
        } else {
            iLow = rIndex(eta, roughness, w);
            iHigh = scene_rdl2::math::min(iLow+1, COMP*COMP-1);
            e = scene_rdl2::math::lerp(mOneMinusEavg[iLow],
                     mOneMinusEavg[iHigh],
                     w);
        }
        return e;
    }

    static float netaRange(float eta)
    {
        // IOR has been calculated for the range [1.05, 3.0]
        return (eta-1.05f)/(3.05f-1.05f);
    }
private:
    static void index(float eta,
                      float cosTheta,
                      float roughness,
                      float& w,
                      size_t& iLow,
                      size_t& iHigh)
    {
        const size_t lIndex = rIndex(eta, roughness, w);
        const size_t hIndex = scene_rdl2::math::min(lIndex+1, COMP*COMP-1);
        const float if1 = lIndex * COMP +
                  cosTheta * (COMP - 1);
        const float if2 = hIndex * COMP +
                  cosTheta * (COMP - 1);
        // Take cosTheta's ceiling for the index
        iLow = static_cast<size_t>(ceil(if1));
        iHigh = static_cast<size_t>(ceil(if2));
    }

    static size_t rIndex(float eta,
                         float roughness,
                         float &weight)
    {
        const float i = nIndex(eta) * COMP +
                        roughness * (COMP - 1);
        const float indexI = floor(i);
        weight = (i-indexI);
        return static_cast<size_t>(indexI);
    }

    static size_t nIndex(float eta)
    {
        float nZeroToOne = netaRange(eta);
        nZeroToOne = scene_rdl2::math::clamp(nZeroToOne, 0.0f, 1.0f);
        const size_t i = static_cast<size_t>((nZeroToOne*(COMP-1)));
        return i;
    }

    static const size_t COMP = 33;
    static const float mE[COMP*COMP*COMP];
    static const float mEExit[COMP*COMP*COMP];
    static const float mOneMinusEavg[COMP*COMP];
    static const float mOneMinusEavgExit[COMP*COMP];
};

// 33x33x33 table calculated using moonshine_devtools/compute_albedo for the following:
// F*Reflection + (1-F)*Transmission
// Ranges:
// EtaT/EtaI [1, 3]
// Roughness [0,1]
// CosTheta [0, 1]
const float
TransmissionAlbedo::mE[COMP*COMP*COMP] =
{
 0.00000f,0.98313f,0.95548f,0.96441f,0.94742f,0.93089f,0.92386f,0.95647f,0.94750f,0.94352f,0.93320f,0.92587f,0.95081f,0.94876f,0.94555f,0.95058f,0.94374f,0.93866f,0.96342f,0.96366f,0.96461f,0.96524f,0.96422f,0.96938f,0.96724f,0.97009f,0.98262f,0.98356f,0.98430f,0.99014f,0.99251f,0.99525f,1.00000f,0.00000f,0.98940f,0.97963f,0.97736f,0.97365f,0.97147f,0.96596f,0.97149f,0.96987f,0.96808f,0.96546f,0.96524f,0.96986f,0.96953f,0.96978f,0.97045f,0.97003f,0.96961f,0.97568f,0.97559f,0.97660f,0.97694f,0.97859f,0.98024f,0.98187f,0.98332f,0.98694f,0.98855f,0.99030f,0.99266f,0.99497f,0.99741f,1.00000f,0.00000f,0.98914f,0.98618f,0.98231f,0.97952f,0.97773f,0.97491f,0.97444f,0.97441f,0.97387f,0.97271f,0.97261f,0.97345f,0.97284f,0.97461f,0.97468f,0.97464f,0.97584f,0.97672f,0.97790f,0.97881f,0.97983f,0.98111f,0.98271f,0.98382f,0.98552f,0.98765f,0.98943f,0.99134f,0.99328f,0.99542f,0.99775f,1.00000f,0.00000f,0.98495f,0.98550f,0.98112f,0.98049f,0.97851f,0.97560f,0.97555f,0.97424f,0.97341f,0.97382f,0.97265f,0.97254f,0.97413f,0.97437f,0.97388f,0.97519f,0.97530f,0.97675f,0.97810f,0.97889f,0.98040f,0.98187f,0.98295f,0.98436f,0.98577f,0.98782f,0.98937f,0.99134f,0.99327f,0.99547f,0.99775f,1.00000f,0.00000f,0.94788f,0.98333f,0.98261f,0.97962f,0.97705f,0.97650f,0.97368f,0.97372f,0.97318f,0.97464f,0.97339f,0.97432f,0.97321f,0.97457f,0.97358f,0.97555f,0.97590f,0.97704f,0.97749f,0.97835f,0.98050f,0.98115f,0.98266f,0.98428f,0.98570f,0.98737f,0.98944f,0.99140f,0.99331f,0.99549f,0.99775f,1.00000f,0.00000f,0.92180f,0.96325f,0.97714f,0.97723f,0.97626f,0.97542f,0.97433f,0.97400f,0.97409f,0.97326f,0.97304f,0.97311f,0.97246f,0.97306f,0.97423f,0.97449f,0.97543f,0.97648f,0.97752f,0.97851f,0.97974f,0.98142f,0.98281f,0.98442f,0.98550f,0.98762f,0.98936f,0.99123f,0.99329f,0.99541f,0.99772f,1.00000f,0.00000f,0.91617f,0.94323f,0.96630f,0.97468f,0.97624f,0.97570f,0.97419f,0.97356f,0.97289f,0.97303f,0.97182f,0.97186f,0.97298f,0.97354f,0.97386f,0.97447f,0.97515f,0.97697f,0.97705f,0.97865f,0.97998f,0.98114f,0.98276f,0.98400f,0.98589f,0.98725f,0.98922f,0.99114f,0.99337f,0.99545f,0.99755f,1.00000f,0.00000f,0.91568f,0.92907f,0.94880f,0.96545f,0.97132f,0.97269f,0.97257f,0.97393f,0.97304f,0.97277f,0.97161f,0.97361f,0.97277f,0.97347f,0.97311f,0.97439f,0.97522f,0.97617f,0.97767f,0.97832f,0.97953f,0.98104f,0.98235f,0.98356f,0.98598f,0.98776f,0.98940f,0.99121f,0.99315f,0.99518f,0.99749f,1.00000f,0.00000f,0.92505f,0.92568f,0.94000f,0.95143f,0.96304f,0.96759f,0.97195f,0.97124f,0.97256f,0.97339f,0.97108f,0.97268f,0.97295f,0.97271f,0.97369f,0.97382f,0.97497f,0.97646f,0.97695f,0.97828f,0.97901f,0.98093f,0.98259f,0.98377f,0.98563f,0.98740f,0.98914f,0.99113f,0.99324f,0.99535f,0.99760f,1.00000f,0.00000f,0.93755f,0.92788f,0.93476f,0.94047f,0.95580f,0.96115f,0.96769f,0.96884f,0.97092f,0.97213f,0.97127f,0.97170f,0.97230f,0.97195f,0.97281f,0.97339f,0.97457f,0.97629f,0.97632f,0.97835f,0.97920f,0.98035f,0.98221f,0.98377f,0.98548f,0.98718f,0.98888f,0.99097f,0.99312f,0.99520f,0.99737f,1.00000f,0.00000f,0.94139f,0.92560f,0.93047f,0.94213f,0.94557f,0.95347f,0.95924f,0.96388f,0.96707f,0.96907f,0.97008f,0.97062f,0.97256f,0.97096f,0.97327f,0.97297f,0.97393f,0.97611f,0.97660f,0.97843f,0.97896f,0.98057f,0.98193f,0.98314f,0.98554f,0.98708f,0.98903f,0.99058f,0.99321f,0.99523f,0.99739f,1.00000f,0.00000f,0.95281f,0.93413f,0.93094f,0.93302f,0.94358f,0.94914f,0.95107f,0.96082f,0.96274f,0.96717f,0.96747f,0.97061f,0.97038f,0.97200f,0.97377f,0.97343f,0.97447f,0.97554f,0.97593f,0.97809f,0.97807f,0.98024f,0.98128f,0.98387f,0.98488f,0.98660f,0.98913f,0.99089f,0.99292f,0.99499f,0.99750f,1.00000f,0.00000f,0.96201f,0.94088f,0.93541f,0.93815f,0.93980f,0.94663f,0.95379f,0.95776f,0.95725f,0.96258f,0.96439f,0.96729f,0.96706f,0.97124f,0.97229f,0.97214f,0.97414f,0.97497f,0.97504f,0.97697f,0.97805f,0.97920f,0.98138f,0.98305f,0.98473f,0.98665f,0.98842f,0.99063f,0.99293f,0.99472f,0.99736f,1.00000f,0.00000f,0.94770f,0.94199f,0.94098f,0.93592f,0.94011f,0.94414f,0.95059f,0.95400f,0.95652f,0.95924f,0.96131f,0.96455f,0.96642f,0.96892f,0.97122f,0.96989f,0.97357f,0.97280f,0.97489f,0.97712f,0.97854f,0.97958f,0.98178f,0.98272f,0.98465f,0.98640f,0.98903f,0.99118f,0.99241f,0.99491f,0.99739f,1.00000f,0.00000f,0.94441f,0.94361f,0.93646f,0.94254f,0.93732f,0.94198f,0.94859f,0.95193f,0.95845f,0.95855f,0.95948f,0.96093f,0.96566f,0.96714f,0.96843f,0.96867f,0.97152f,0.97404f,0.97448f,0.97564f,0.97741f,0.97890f,0.98130f,0.98268f,0.98400f,0.98606f,0.98802f,0.99078f,0.99262f,0.99458f,0.99726f,1.00000f,0.00000f,0.92852f,0.92940f,0.93426f,0.94186f,0.94255f,0.94472f,0.94988f,0.95230f,0.95398f,0.95617f,0.95726f,0.96036f,0.96007f,0.96506f,0.96552f,0.96826f,0.97016f,0.97073f,0.97362f,0.97555f,0.97718f,0.97958f,0.98047f,0.98366f,0.98433f,0.98621f,0.98792f,0.99030f,0.99226f,0.99476f,0.99721f,1.00000f,0.00000f,0.90269f,0.91486f,0.92102f,0.92703f,0.93415f,0.94382f,0.94641f,0.95241f,0.95249f,0.95607f,0.96095f,0.95863f,0.95954f,0.96190f,0.96536f,0.96630f,0.97019f,0.97088f,0.97257f,0.97481f,0.97588f,0.97881f,0.98060f,0.98210f,0.98383f,0.98613f,0.98784f,0.99045f,0.99221f,0.99433f,0.99689f,1.00000f,0.00000f,0.86963f,0.89242f,0.90494f,0.91841f,0.92968f,0.93276f,0.93935f,0.94599f,0.95245f,0.95751f,0.95773f,0.95956f,0.96118f,0.96250f,0.96131f,0.96338f,0.96667f,0.96943f,0.97331f,0.97465f,0.97413f,0.97752f,0.98026f,0.98058f,0.98255f,0.98541f,0.98760f,0.99005f,0.99211f,0.99438f,0.99702f,1.00000f,0.00000f,0.82278f,0.85415f,0.87856f,0.90183f,0.91140f,0.92576f,0.93618f,0.94012f,0.94756f,0.95296f,0.95692f,0.95966f,0.95931f,0.96334f,0.96393f,0.96276f,0.96701f,0.96763f,0.97075f,0.97373f,0.97408f,0.97597f,0.98015f,0.98100f,0.98361f,0.98494f,0.98746f,0.98977f,0.99135f,0.99472f,0.99679f,1.00000f,0.00000f,0.77799f,0.81713f,0.85341f,0.87573f,0.90359f,0.91122f,0.92403f,0.93803f,0.94370f,0.94955f,0.95207f,0.95597f,0.96018f,0.96063f,0.96171f,0.96573f,0.96466f,0.96879f,0.96991f,0.97080f,0.97387f,0.97647f,0.97748f,0.98098f,0.98396f,0.98466f,0.98762f,0.98846f,0.99176f,0.99395f,0.99655f,1.00000f,0.00000f,0.72952f,0.78271f,0.81941f,0.85079f,0.87607f,0.89389f,0.91085f,0.92237f,0.93088f,0.94207f,0.94938f,0.95415f,0.95761f,0.96477f,0.96482f,0.96805f,0.96416f,0.96441f,0.96859f,0.96694f,0.97200f,0.97397f,0.97537f,0.97924f,0.98157f,0.98367f,0.98704f,0.98904f,0.99158f,0.99349f,0.99624f,1.00000f,0.00000f,0.68083f,0.74000f,0.78617f,0.82336f,0.85305f,0.87628f,0.89511f,0.91220f,0.92306f,0.93275f,0.94370f,0.95081f,0.95359f,0.95783f,0.96048f,0.96792f,0.96777f,0.96699f,0.96713f,0.96913f,0.97208f,0.97389f,0.97562f,0.97715f,0.98032f,0.98330f,0.98623f,0.98891f,0.99170f,0.99332f,0.99680f,1.00000f,0.00000f,0.63416f,0.69196f,0.74629f,0.78637f,0.82486f,0.84911f,0.87686f,0.89620f,0.90787f,0.92219f,0.93422f,0.94647f,0.94913f,0.95336f,0.95911f,0.96398f,0.96497f,0.96709f,0.97099f,0.96866f,0.97212f,0.97076f,0.97455f,0.97900f,0.97976f,0.98190f,0.98590f,0.98805f,0.98980f,0.99340f,0.99614f,0.99997f,0.00000f,0.58499f,0.65082f,0.70875f,0.75189f,0.79422f,0.82953f,0.85415f,0.87477f,0.89429f,0.90874f,0.92062f,0.93586f,0.94952f,0.95222f,0.95462f,0.96242f,0.96386f,0.96910f,0.97138f,0.96890f,0.97124f,0.97348f,0.97598f,0.97834f,0.97940f,0.98316f,0.98414f,0.98853f,0.99118f,0.99396f,0.99689f,0.99995f,0.00000f,0.54204f,0.61313f,0.67052f,0.72395f,0.76159f,0.80136f,0.82744f,0.85784f,0.87369f,0.89329f,0.90797f,0.92098f,0.93337f,0.93991f,0.94654f,0.95920f,0.96245f,0.96555f,0.96894f,0.97177f,0.97180f,0.97408f,0.97364f,0.97669f,0.97638f,0.98224f,0.98297f,0.98721f,0.98938f,0.99216f,0.99534f,0.99985f,0.00000f,0.49850f,0.57075f,0.63403f,0.68515f,0.73385f,0.76813f,0.79766f,0.82989f,0.85747f,0.87622f,0.89467f,0.91196f,0.92161f,0.93356f,0.94149f,0.95118f,0.96065f,0.96414f,0.96846f,0.97263f,0.97240f,0.97498f,0.97387f,0.97482f,0.97788f,0.98015f,0.98188f,0.98481f,0.98882f,0.99097f,0.99627f,0.99982f,0.00000f,0.45729f,0.53201f,0.59700f,0.65134f,0.70083f,0.74005f,0.78054f,0.80617f,0.83593f,0.85649f,0.87846f,0.89754f,0.91055f,0.92459f,0.93120f,0.94367f,0.95035f,0.95974f,0.96348f,0.96866f,0.97280f,0.97560f,0.97321f,0.97661f,0.97493f,0.98009f,0.97970f,0.98660f,0.98761f,0.99145f,0.99517f,0.99975f,0.00000f,0.42157f,0.49660f,0.56013f,0.61698f,0.66943f,0.71028f,0.74651f,0.78452f,0.80928f,0.83713f,0.86167f,0.87892f,0.89296f,0.90812f,0.92597f,0.93246f,0.94818f,0.95444f,0.95980f,0.96743f,0.97030f,0.97362f,0.97330f,0.97652f,0.97600f,0.97781f,0.98232f,0.98322f,0.98868f,0.99135f,0.99434f,0.99968f,0.00000f,0.38584f,0.46115f,0.53124f,0.58782f,0.63785f,0.68309f,0.72515f,0.75451f,0.78994f,0.81648f,0.84408f,0.86086f,0.87915f,0.90224f,0.91186f,0.92417f,0.93868f,0.94527f,0.95246f,0.95932f,0.96776f,0.97128f,0.97603f,0.97623f,0.97542f,0.97788f,0.97958f,0.98348f,0.98722f,0.98996f,0.99439f,0.99960f,0.00000f,0.35484f,0.42945f,0.49447f,0.55287f,0.61173f,0.65514f,0.69459f,0.73597f,0.76838f,0.79315f,0.81741f,0.84568f,0.86394f,0.88217f,0.89917f,0.91376f,0.92715f,0.93989f,0.94546f,0.95513f,0.95950f,0.97035f,0.97708f,0.97672f,0.97784f,0.97817f,0.97800f,0.98142f,0.98557f,0.99053f,0.99407f,0.99943f,0.00000f,0.32938f,0.40222f,0.46379f,0.52613f,0.57747f,0.62584f,0.66769f,0.70504f,0.74259f,0.77251f,0.80146f,0.82801f,0.84385f,0.86952f,0.88673f,0.89727f,0.91566f,0.93070f,0.94193f,0.94918f,0.95526f,0.96322f,0.97108f,0.97619f,0.97870f,0.97681f,0.97702f,0.97980f,0.98716f,0.98803f,0.99304f,0.99937f,0.00000f,0.30158f,0.37559f,0.43859f,0.49673f,0.55138f,0.59804f,0.64398f,0.68607f,0.71640f,0.75092f,0.77803f,0.80771f,0.83392f,0.84848f,0.86992f,0.88823f,0.90356f,0.92191f,0.93142f,0.94707f,0.95412f,0.96193f,0.96903f,0.97313f,0.97502f,0.98313f,0.97953f,0.97771f,0.98414f,0.98909f,0.99149f,0.99912f,0.00000f,0.27962f,0.34894f,0.41227f,0.46979f,0.52414f,0.57137f,0.61700f,0.65964f,0.69365f,0.72601f,0.76033f,0.78774f,0.81683f,0.83858f,0.85641f,0.87443f,0.89436f,0.90982f,0.92151f,0.93283f,0.94500f,0.95916f,0.96584f,0.97174f,0.97676f,0.97950f,0.97831f,0.97916f,0.98332f,0.98742f,0.99163f,0.99908f,
 0.00000f,0.99068f,0.97367f,0.95929f,0.94525f,0.92872f,0.95982f,0.95701f,0.94871f,0.94065f,0.93409f,0.93176f,0.94480f,0.95024f,0.94531f,0.94884f,0.95245f,0.94819f,0.94773f,0.96697f,0.96793f,0.96754f,0.97148f,0.96922f,0.97283f,0.97314f,0.97814f,0.97855f,0.98081f,0.98807f,0.99290f,0.99718f,1.00000f,0.00000f,0.99019f,0.98198f,0.97647f,0.97254f,0.96908f,0.97209f,0.97334f,0.97005f,0.96851f,0.96609f,0.96504f,0.96972f,0.96985f,0.96924f,0.97012f,0.97066f,0.97061f,0.97032f,0.97592f,0.97673f,0.97802f,0.97968f,0.98056f,0.98226f,0.98385f,0.98581f,0.98779f,0.98978f,0.99245f,0.99497f,0.99761f,1.00000f,0.00000f,0.99264f,0.98563f,0.98166f,0.97898f,0.97493f,0.97667f,0.97495f,0.97385f,0.97297f,0.97352f,0.97261f,0.97349f,0.97317f,0.97458f,0.97469f,0.97458f,0.97582f,0.97665f,0.97790f,0.97848f,0.97994f,0.98107f,0.98289f,0.98417f,0.98587f,0.98748f,0.98947f,0.99134f,0.99338f,0.99544f,0.99778f,1.00000f,0.00000f,0.98305f,0.98689f,0.98230f,0.97933f,0.97867f,0.97646f,0.97533f,0.97385f,0.97409f,0.97328f,0.97403f,0.97299f,0.97336f,0.97349f,0.97393f,0.97489f,0.97576f,0.97685f,0.97755f,0.97874f,0.97989f,0.98158f,0.98238f,0.98453f,0.98603f,0.98774f,0.98951f,0.99128f,0.99343f,0.99550f,0.99775f,1.00000f,0.00000f,0.94246f,0.98240f,0.98266f,0.98027f,0.97641f,0.97597f,0.97455f,0.97382f,0.97375f,0.97303f,0.97359f,0.97395f,0.97334f,0.97423f,0.97493f,0.97469f,0.97538f,0.97635f,0.97781f,0.97883f,0.98026f,0.98071f,0.98274f,0.98420f,0.98586f,0.98754f,0.98914f,0.99133f,0.99337f,0.99552f,0.99776f,1.00000f,0.00000f,0.91959f,0.96607f,0.98046f,0.97795f,0.97772f,0.97552f,0.97512f,0.97388f,0.97433f,0.97382f,0.97310f,0.97301f,0.97208f,0.97357f,0.97434f,0.97530f,0.97573f,0.97645f,0.97685f,0.97828f,0.97940f,0.98137f,0.98281f,0.98390f,0.98577f,0.98760f,0.98937f,0.99118f,0.99327f,0.99540f,0.99776f,1.00000f,0.00000f,0.90669f,0.93650f,0.96280f,0.97377f,0.97670f,0.97536f,0.97450f,0.97346f,0.97153f,0.97289f,0.97183f,0.97314f,0.97279f,0.97366f,0.97320f,0.97397f,0.97547f,0.97607f,0.97719f,0.97876f,0.97966f,0.98134f,0.98237f,0.98395f,0.98551f,0.98732f,0.98937f,0.99112f,0.99348f,0.99537f,0.99762f,1.00000f,0.00000f,0.91850f,0.92222f,0.94458f,0.96110f,0.97169f,0.97168f,0.97221f,0.97189f,0.97272f,0.97269f,0.97258f,0.97291f,0.97271f,0.97312f,0.97345f,0.97443f,0.97523f,0.97594f,0.97695f,0.97796f,0.97940f,0.98089f,0.98230f,0.98383f,0.98559f,0.98732f,0.98927f,0.99109f,0.99303f,0.99535f,0.99756f,1.00000f,0.00000f,0.91964f,0.91392f,0.92854f,0.94684f,0.95859f,0.96620f,0.96992f,0.97247f,0.97116f,0.97180f,0.97220f,0.97247f,0.97234f,0.97251f,0.97325f,0.97435f,0.97477f,0.97580f,0.97660f,0.97792f,0.97942f,0.98076f,0.98210f,0.98353f,0.98546f,0.98713f,0.98930f,0.99107f,0.99304f,0.99530f,0.99758f,1.00000f,0.00000f,0.92650f,0.91910f,0.92384f,0.93340f,0.94468f,0.95539f,0.96464f,0.96798f,0.96944f,0.97033f,0.96997f,0.97032f,0.97173f,0.97147f,0.97227f,0.97416f,0.97409f,0.97588f,0.97651f,0.97793f,0.97901f,0.98090f,0.98214f,0.98368f,0.98531f,0.98723f,0.98859f,0.99095f,0.99298f,0.99531f,0.99759f,1.00000f,0.00000f,0.93673f,0.92374f,0.91988f,0.92801f,0.93820f,0.94530f,0.95372f,0.96013f,0.96615f,0.96622f,0.96906f,0.97069f,0.97209f,0.97153f,0.97290f,0.97285f,0.97396f,0.97570f,0.97592f,0.97738f,0.97892f,0.98037f,0.98165f,0.98335f,0.98495f,0.98691f,0.98872f,0.99082f,0.99306f,0.99531f,0.99745f,1.00000f,0.00000f,0.93927f,0.92752f,0.92398f,0.92640f,0.93413f,0.94025f,0.94574f,0.95290f,0.95784f,0.96371f,0.96615f,0.96852f,0.97162f,0.97157f,0.97182f,0.97333f,0.97543f,0.97657f,0.97610f,0.97747f,0.97899f,0.97925f,0.98178f,0.98339f,0.98512f,0.98703f,0.98881f,0.99063f,0.99286f,0.99499f,0.99730f,1.00000f,0.00000f,0.94726f,0.93255f,0.92961f,0.92799f,0.92862f,0.93728f,0.94108f,0.94759f,0.95265f,0.95751f,0.96028f,0.96346f,0.96614f,0.96913f,0.97128f,0.97230f,0.97359f,0.97456f,0.97683f,0.97625f,0.97853f,0.98022f,0.98164f,0.98312f,0.98448f,0.98666f,0.98906f,0.99046f,0.99252f,0.99514f,0.99745f,1.00000f,0.00000f,0.95589f,0.93403f,0.93340f,0.93028f,0.93389f,0.93514f,0.94151f,0.94508f,0.94781f,0.95173f,0.95728f,0.96126f,0.96356f,0.96533f,0.96996f,0.97082f,0.97192f,0.97379f,0.97467f,0.97598f,0.97800f,0.97945f,0.98130f,0.98205f,0.98519f,0.98705f,0.98836f,0.99070f,0.99258f,0.99500f,0.99723f,1.00000f,0.00000f,0.95623f,0.94401f,0.94154f,0.93303f,0.93363f,0.93269f,0.93741f,0.94304f,0.94652f,0.95128f,0.95126f,0.95591f,0.96088f,0.96353f,0.96493f,0.96759f,0.97091f,0.97245f,0.97436f,0.97597f,0.97680f,0.98007f,0.98077f,0.98307f,0.98480f,0.98690f,0.98839f,0.99030f,0.99237f,0.99449f,0.99728f,1.00000f,0.00000f,0.95836f,0.94740f,0.94196f,0.93588f,0.93154f,0.93985f,0.93581f,0.94187f,0.94503f,0.94705f,0.94967f,0.95175f,0.95858f,0.95837f,0.96329f,0.96637f,0.96835f,0.97055f,0.97328f,0.97539f,0.97698f,0.97779f,0.98015f,0.98113f,0.98386f,0.98545f,0.98787f,0.99039f,0.99206f,0.99479f,0.99693f,1.00000f,0.00000f,0.95290f,0.95410f,0.94404f,0.94088f,0.93461f,0.94263f,0.94029f,0.94208f,0.94499f,0.94733f,0.95119f,0.95279f,0.95301f,0.95565f,0.95999f,0.96534f,0.96698f,0.96831f,0.97186f,0.97308f,0.97595f,0.97722f,0.97937f,0.98091f,0.98323f,0.98467f,0.98762f,0.99028f,0.99250f,0.99447f,0.99679f,1.00000f,0.00000f,0.94274f,0.93721f,0.93883f,0.93657f,0.94007f,0.93803f,0.93932f,0.94039f,0.94483f,0.94297f,0.95169f,0.95040f,0.95427f,0.95484f,0.95727f,0.96066f,0.96481f,0.96589f,0.96776f,0.97300f,0.97404f,0.97594f,0.97916f,0.98208f,0.98277f,0.98530f,0.98713f,0.98937f,0.99185f,0.99411f,0.99717f,1.00000f,0.00000f,0.92274f,0.93272f,0.93562f,0.93796f,0.93635f,0.93609f,0.94176f,0.94370f,0.94257f,0.94869f,0.94807f,0.94839f,0.95515f,0.95384f,0.95427f,0.95906f,0.96041f,0.96232f,0.96545f,0.96796f,0.97393f,0.97419f,0.97748f,0.97822f,0.98142f,0.98438f,0.98739f,0.99073f,0.99173f,0.99466f,0.99686f,1.00000f,0.00000f,0.89679f,0.90984f,0.92433f,0.92765f,0.93368f,0.93021f,0.93085f,0.93916f,0.94006f,0.94646f,0.94863f,0.94716f,0.95372f,0.95381f,0.95389f,0.95928f,0.96044f,0.96174f,0.96445f,0.96650f,0.96988f,0.97192f,0.97560f,0.97816f,0.98009f,0.98424f,0.98597f,0.98825f,0.99141f,0.99411f,0.99708f,1.00000f,0.00000f,0.87103f,0.88644f,0.90482f,0.90961f,0.92150f,0.92464f,0.92703f,0.93959f,0.93930f,0.94386f,0.94585f,0.94969f,0.94973f,0.95490f,0.95404f,0.95678f,0.95848f,0.96039f,0.96261f,0.96300f,0.96813f,0.97029f,0.97481f,0.97588f,0.97983f,0.98276f,0.98773f,0.98785f,0.99044f,0.99407f,0.99609f,0.99998f,0.00000f,0.83870f,0.85827f,0.87738f,0.89286f,0.90532f,0.91320f,0.92704f,0.92790f,0.93484f,0.93866f,0.94448f,0.94274f,0.95212f,0.95565f,0.95271f,0.96109f,0.95970f,0.96056f,0.96081f,0.96052f,0.96670f,0.96803f,0.97328f,0.97635f,0.97789f,0.98176f,0.98622f,0.98800f,0.99076f,0.99306f,0.99621f,0.99996f,0.00000f,0.79854f,0.82775f,0.84839f,0.87485f,0.88711f,0.89507f,0.90220f,0.92097f,0.93027f,0.93060f,0.94061f,0.94705f,0.95135f,0.95258f,0.95216f,0.95490f,0.95941f,0.95746f,0.96075f,0.96264f,0.96612f,0.96493f,0.96931f,0.97356f,0.97648f,0.98052f,0.98375f,0.98667f,0.98999f,0.99251f,0.99586f,0.99990f,0.00000f,0.75302f,0.78754f,0.82272f,0.84221f,0.86039f,0.88272f,0.89390f,0.90333f,0.91378f,0.92176f,0.93013f,0.93697f,0.94526f,0.95156f,0.95269f,0.95563f,0.96037f,0.96187f,0.96259f,0.96430f,0.96421f,0.96636f,0.96855f,0.97234f,0.97601f,0.97821f,0.98068f,0.98505f,0.98869f,0.99276f,0.99527f,0.99969f,0.00000f,0.71591f,0.75337f,0.78114f,0.80952f,0.83418f,0.85705f,0.87464f,0.89246f,0.89966f,0.91448f,0.92555f,0.92908f,0.93813f,0.94073f,0.94692f,0.95071f,0.95679f,0.96024f,0.96387f,0.96459f,0.96665f,0.96346f,0.96687f,0.97204f,0.97469f,0.97894f,0.98128f,0.98518f,0.98747f,0.99156f,0.99422f,0.99955f,0.00000f,0.66628f,0.71541f,0.75216f,0.78139f,0.80879f,0.83140f,0.85624f,0.87329f,0.88878f,0.90109f,0.91152f,0.91706f,0.92902f,0.94054f,0.94579f,0.94942f,0.95359f,0.95941f,0.96305f,0.96047f,0.96507f,0.96591f,0.96856f,0.97066f,0.97467f,0.97586f,0.97810f,0.98226f,0.98641f,0.99037f,0.99313f,0.99931f,0.00000f,0.62399f,0.67408f,0.71219f,0.74988f,0.78339f,0.80685f,0.82897f,0.84955f,0.86947f,0.87581f,0.89908f,0.90964f,0.92749f,0.93076f,0.93737f,0.94446f,0.95244f,0.96121f,0.96190f,0.96128f,0.96570f,0.96527f,0.96882f,0.96773f,0.97437f,0.97435f,0.97749f,0.98111f,0.98495f,0.98898f,0.99218f,0.99908f,0.00000f,0.58805f,0.63466f,0.67917f,0.71224f,0.74801f,0.77557f,0.80118f,0.82672f,0.84949f,0.86631f,0.88076f,0.89770f,0.90582f,0.91716f,0.92746f,0.93687f,0.94122f,0.95179f,0.95480f,0.96171f,0.96152f,0.96352f,0.96708f,0.97163f,0.96829f,0.96961f,0.97744f,0.97951f,0.98316f,0.98813f,0.99172f,0.99851f,0.00000f,0.54696f,0.59608f,0.64395f,0.68404f,0.71882f,0.74983f,0.77691f,0.80395f,0.82747f,0.84585f,0.86248f,0.88026f,0.89550f,0.91163f,0.92084f,0.92761f,0.94032f,0.94594f,0.95237f,0.95809f,0.96116f,0.96528f,0.96529f,0.96983f,0.97133f,0.97293f,0.97451f,0.97847f,0.98009f,0.98616f,0.99111f,0.99811f,0.00000f,0.50229f,0.56083f,0.60814f,0.65050f,0.68281f,0.71695f,0.75434f,0.77452f,0.79877f,0.82363f,0.84797f,0.85957f,0.87952f,0.89595f,0.90206f,0.92061f,0.92870f,0.93690f,0.94714f,0.95193f,0.95861f,0.96313f,0.96699f,0.96966f,0.96979f,0.97294f,0.97206f,0.97669f,0.97950f,0.98478f,0.99012f,0.99761f,0.00000f,0.47192f,0.52145f,0.57701f,0.61710f,0.65559f,0.69311f,0.72408f,0.75686f,0.78079f,0.81119f,0.82930f,0.84352f,0.86000f,0.87692f,0.88989f,0.90454f,0.92018f,0.93089f,0.93809f,0.94706f,0.95233f,0.95756f,0.96516f,0.96908f,0.96986f,0.97054f,0.97384f,0.97475f,0.98106f,0.98314f,0.98991f,0.99720f,0.00000f,0.43884f,0.49242f,0.54209f,0.58536f,0.62360f,0.66517f,0.69452f,0.72837f,0.75213f,0.77883f,0.80680f,0.82787f,0.84696f,0.86291f,0.88100f,0.89211f,0.90738f,0.91711f,0.93148f,0.93875f,0.95077f,0.95824f,0.96067f,0.96756f,0.97066f,0.97292f,0.97528f,0.97269f,0.97706f,0.98038f,0.98621f,0.99674f,0.00000f,0.40549f,0.46069f,0.50933f,0.55407f,0.59782f,0.63297f,0.67091f,0.70076f,0.73516f,0.76022f,0.78095f,0.80805f,0.83128f,0.84613f,0.86857f,0.88249f,0.89581f,0.91059f,0.92040f,0.92991f,0.93846f,0.95371f,0.95445f,0.96170f,0.96877f,0.96860f,0.97426f,0.97253f,0.97367f,0.97793f,0.98570f,0.99608f,
 0.00000f,0.99362f,0.97342f,0.95990f,0.94291f,0.94723f,0.95616f,0.95138f,0.94312f,0.94418f,0.94085f,0.93363f,0.95691f,0.95602f,0.95533f,0.96199f,0.95280f,0.95586f,0.95765f,0.95389f,0.95226f,0.95696f,0.97332f,0.97402f,0.97134f,0.97664f,0.97856f,0.98070f,0.97498f,0.98936f,0.98540f,0.99589f,1.00000f,0.00000f,0.98917f,0.98324f,0.97744f,0.97219f,0.97393f,0.97183f,0.97056f,0.96916f,0.96906f,0.96698f,0.96489f,0.96975f,0.97074f,0.97024f,0.97048f,0.97176f,0.97189f,0.97239f,0.97318f,0.97353f,0.97721f,0.97936f,0.98049f,0.98248f,0.98397f,0.98597f,0.98775f,0.98960f,0.99269f,0.99454f,0.99749f,1.00000f,0.00000f,0.99003f,0.98527f,0.98427f,0.97999f,0.97731f,0.97569f,0.97481f,0.97421f,0.97397f,0.97208f,0.97275f,0.97327f,0.97411f,0.97385f,0.97399f,0.97550f,0.97602f,0.97688f,0.97745f,0.97861f,0.97996f,0.98127f,0.98261f,0.98422f,0.98559f,0.98770f,0.98940f,0.99127f,0.99318f,0.99528f,0.99774f,1.00000f,0.00000f,0.98300f,0.98511f,0.98315f,0.97920f,0.97888f,0.97667f,0.97419f,0.97408f,0.97506f,0.97393f,0.97308f,0.97343f,0.97359f,0.97355f,0.97426f,0.97481f,0.97605f,0.97646f,0.97764f,0.97895f,0.98032f,0.98161f,0.98255f,0.98407f,0.98594f,0.98738f,0.98958f,0.99137f,0.99343f,0.99549f,0.99775f,1.00000f,0.00000f,0.94512f,0.98353f,0.98117f,0.98035f,0.97806f,0.97664f,0.97471f,0.97415f,0.97372f,0.97228f,0.97396f,0.97378f,0.97394f,0.97393f,0.97487f,0.97458f,0.97568f,0.97731f,0.97798f,0.97875f,0.97990f,0.98110f,0.98267f,0.98441f,0.98599f,0.98736f,0.98931f,0.99133f,0.99336f,0.99546f,0.99777f,1.00000f,0.00000f,0.91241f,0.96281f,0.97876f,0.97773f,0.97748f,0.97640f,0.97473f,0.97407f,0.97408f,0.97277f,0.97309f,0.97248f,0.97374f,0.97363f,0.97324f,0.97451f,0.97539f,0.97611f,0.97750f,0.97904f,0.98006f,0.98145f,0.98250f,0.98391f,0.98570f,0.98770f,0.98937f,0.99117f,0.99321f,0.99540f,0.99775f,1.00000f,0.00000f,0.90591f,0.93341f,0.96447f,0.97238f,0.97475f,0.97565f,0.97456f,0.97397f,0.97370f,0.97308f,0.97259f,0.97224f,0.97304f,0.97247f,0.97404f,0.97440f,0.97507f,0.97658f,0.97758f,0.97874f,0.98010f,0.98123f,0.98247f,0.98394f,0.98557f,0.98743f,0.98911f,0.99123f,0.99324f,0.99545f,0.99758f,1.00000f,0.00000f,0.91338f,0.92090f,0.94228f,0.95937f,0.97054f,0.97233f,0.97252f,0.97265f,0.97244f,0.97166f,0.97223f,0.97189f,0.97344f,0.97296f,0.97354f,0.97392f,0.97531f,0.97596f,0.97732f,0.97831f,0.98006f,0.98063f,0.98283f,0.98355f,0.98553f,0.98721f,0.98945f,0.99130f,0.99321f,0.99531f,0.99751f,1.00000f,0.00000f,0.91982f,0.91241f,0.92716f,0.94278f,0.95550f,0.96395f,0.97043f,0.97072f,0.97145f,0.97137f,0.97278f,0.97200f,0.97183f,0.97298f,0.97432f,0.97502f,0.97514f,0.97630f,0.97682f,0.97775f,0.97978f,0.98014f,0.98247f,0.98416f,0.98567f,0.98713f,0.98949f,0.99119f,0.99316f,0.99534f,0.99761f,1.00000f,0.00000f,0.93078f,0.91577f,0.91908f,0.93124f,0.94164f,0.95491f,0.96254f,0.96541f,0.96997f,0.97109f,0.96995f,0.97154f,0.97198f,0.97204f,0.97298f,0.97416f,0.97460f,0.97540f,0.97708f,0.97775f,0.97899f,0.98030f,0.98236f,0.98346f,0.98546f,0.98707f,0.98901f,0.99107f,0.99295f,0.99513f,0.99758f,1.00000f,0.00000f,0.93819f,0.92236f,0.91584f,0.92524f,0.93314f,0.94123f,0.94914f,0.95772f,0.96283f,0.96654f,0.96874f,0.96898f,0.97090f,0.97242f,0.97282f,0.97372f,0.97461f,0.97568f,0.97660f,0.97774f,0.97922f,0.98019f,0.98201f,0.98352f,0.98508f,0.98728f,0.98878f,0.99105f,0.99303f,0.99525f,0.99751f,1.00000f,0.00000f,0.94178f,0.92190f,0.92202f,0.92032f,0.92841f,0.93639f,0.94411f,0.95030f,0.95719f,0.96164f,0.96438f,0.96780f,0.96965f,0.96874f,0.97174f,0.97330f,0.97345f,0.97599f,0.97625f,0.97676f,0.97829f,0.98111f,0.98119f,0.98342f,0.98490f,0.98701f,0.98885f,0.99101f,0.99279f,0.99493f,0.99755f,1.00000f,0.00000f,0.94994f,0.92976f,0.92581f,0.92192f,0.92548f,0.93499f,0.94123f,0.94204f,0.94703f,0.95308f,0.95930f,0.96297f,0.96607f,0.96910f,0.97111f,0.97117f,0.97345f,0.97540f,0.97570f,0.97662f,0.97823f,0.98015f,0.98211f,0.98346f,0.98548f,0.98695f,0.98892f,0.99047f,0.99262f,0.99512f,0.99741f,1.00000f,0.00000f,0.95153f,0.93377f,0.92365f,0.92688f,0.92740f,0.93030f,0.93339f,0.93779f,0.94446f,0.94959f,0.95269f,0.95831f,0.96206f,0.96621f,0.96674f,0.97099f,0.97157f,0.97354f,0.97481f,0.97558f,0.97733f,0.97986f,0.98113f,0.98319f,0.98444f,0.98640f,0.98862f,0.99053f,0.99245f,0.99501f,0.99740f,1.00000f,0.00000f,0.95913f,0.94142f,0.92745f,0.93019f,0.92533f,0.93110f,0.93099f,0.93490f,0.94179f,0.94178f,0.94863f,0.95256f,0.95550f,0.96051f,0.96433f,0.96706f,0.97055f,0.97310f,0.97395f,0.97565f,0.97986f,0.97897f,0.98068f,0.98263f,0.98404f,0.98648f,0.98816f,0.98981f,0.99238f,0.99475f,0.99721f,1.00000f,0.00000f,0.96322f,0.95203f,0.93509f,0.93232f,0.92716f,0.92895f,0.93394f,0.93670f,0.93785f,0.94222f,0.94507f,0.94801f,0.95496f,0.95772f,0.96072f,0.96413f,0.96537f,0.96875f,0.97155f,0.97304f,0.97665f,0.97951f,0.97960f,0.98274f,0.98371f,0.98631f,0.98809f,0.99008f,0.99270f,0.99492f,0.99728f,1.00000f,0.00000f,0.96206f,0.95318f,0.93781f,0.93363f,0.93578f,0.93298f,0.93377f,0.93615f,0.93804f,0.94282f,0.94218f,0.94493f,0.94875f,0.95226f,0.95841f,0.96011f,0.96264f,0.96618f,0.96932f,0.97112f,0.97504f,0.97731f,0.97920f,0.98124f,0.98346f,0.98586f,0.98754f,0.98985f,0.99253f,0.99446f,0.99711f,1.00000f,0.00000f,0.95556f,0.94862f,0.94668f,0.94172f,0.94118f,0.93583f,0.93503f,0.93844f,0.94060f,0.94230f,0.94349f,0.94382f,0.94975f,0.95151f,0.95060f,0.95455f,0.96005f,0.96341f,0.96543f,0.97021f,0.97156f,0.97481f,0.97666f,0.97980f,0.98230f,0.98588f,0.98691f,0.98961f,0.99194f,0.99405f,0.99665f,1.00000f,0.00000f,0.95540f,0.95336f,0.94790f,0.93998f,0.94061f,0.93578f,0.93332f,0.94023f,0.93738f,0.94328f,0.94370f,0.94546f,0.94675f,0.94896f,0.95064f,0.95339f,0.95666f,0.95962f,0.96444f,0.96720f,0.97118f,0.97440f,0.97428f,0.97941f,0.98126f,0.98419f,0.98568f,0.98896f,0.99107f,0.99409f,0.99694f,1.00000f,0.00000f,0.93810f,0.94292f,0.93761f,0.94282f,0.93732f,0.93551f,0.93726f,0.93658f,0.94225f,0.93980f,0.94877f,0.94672f,0.94367f,0.95174f,0.95176f,0.94895f,0.95542f,0.95732f,0.95869f,0.96359f,0.96714f,0.97130f,0.97402f,0.97688f,0.98051f,0.98202f,0.98631f,0.98864f,0.99147f,0.99369f,0.99619f,1.00000f,0.00000f,0.92278f,0.93006f,0.93094f,0.93789f,0.93196f,0.93628f,0.93631f,0.93751f,0.94055f,0.93672f,0.94443f,0.94471f,0.94572f,0.94871f,0.95278f,0.95103f,0.95589f,0.95622f,0.95965f,0.96372f,0.96497f,0.96917f,0.97106f,0.97450f,0.97932f,0.98196f,0.98282f,0.98743f,0.99088f,0.99259f,0.99567f,0.99998f,0.00000f,0.88873f,0.90331f,0.91292f,0.92527f,0.92689f,0.92909f,0.92414f,0.93123f,0.93520f,0.93975f,0.94016f,0.94616f,0.94495f,0.94891f,0.95050f,0.95168f,0.95394f,0.95593f,0.95897f,0.96034f,0.96020f,0.96603f,0.96756f,0.97355f,0.97794f,0.97949f,0.98306f,0.98596f,0.98933f,0.99178f,0.99561f,0.99990f,0.00000f,0.86287f,0.88599f,0.89366f,0.90756f,0.91168f,0.92106f,0.92236f,0.92246f,0.93054f,0.94121f,0.94100f,0.94533f,0.94967f,0.95065f,0.95156f,0.95265f,0.95266f,0.95509f,0.95594f,0.95908f,0.96018f,0.96501f,0.97002f,0.97175f,0.97454f,0.97641f,0.98132f,0.98347f,0.98836f,0.99116f,0.99531f,0.99973f,0.00000f,0.83432f,0.86113f,0.87030f,0.87999f,0.89477f,0.90707f,0.90948f,0.91749f,0.92047f,0.93176f,0.93413f,0.93929f,0.94528f,0.94809f,0.95175f,0.95106f,0.94747f,0.95492f,0.95726f,0.95879f,0.95932f,0.96222f,0.96444f,0.97141f,0.97264f,0.97562f,0.98019f,0.98277f,0.98641f,0.99086f,0.99479f,0.99945f,0.00000f,0.79590f,0.82297f,0.84341f,0.85738f,0.87506f,0.88571f,0.89615f,0.90593f,0.92125f,0.91712f,0.92890f,0.93746f,0.93883f,0.94185f,0.94699f,0.94935f,0.94980f,0.95444f,0.95795f,0.95621f,0.95922f,0.96338f,0.96213f,0.96554f,0.96898f,0.97266f,0.97759f,0.98255f,0.98708f,0.98891f,0.99341f,0.99897f,0.00000f,0.76100f,0.78914f,0.81604f,0.83060f,0.85266f,0.86674f,0.88359f,0.88785f,0.90087f,0.91096f,0.91659f,0.92887f,0.93078f,0.93808f,0.94095f,0.94585f,0.95115f,0.95539f,0.95609f,0.95663f,0.96072f,0.96087f,0.96525f,0.96489f,0.96724f,0.97078f,0.97443f,0.97973f,0.98252f,0.98726f,0.99234f,0.99852f,0.00000f,0.72039f,0.74911f,0.77841f,0.79980f,0.82031f,0.84512f,0.85819f,0.87197f,0.88110f,0.90192f,0.91253f,0.91350f,0.91919f,0.93146f,0.93400f,0.93991f,0.94573f,0.95429f,0.95696f,0.95632f,0.96029f,0.96155f,0.96061f,0.96325f,0.96362f,0.97026f,0.97313f,0.97844f,0.98236f,0.98480f,0.99061f,0.99788f,0.00000f,0.67854f,0.71509f,0.74519f,0.77581f,0.79465f,0.81971f,0.84248f,0.85527f,0.86422f,0.88004f,0.89200f,0.90493f,0.91409f,0.92292f,0.92863f,0.94043f,0.93963f,0.94791f,0.95413f,0.95860f,0.95948f,0.95779f,0.96339f,0.96412f,0.96393f,0.96717f,0.97079f,0.97374f,0.98005f,0.98520f,0.98855f,0.99706f,0.00000f,0.63734f,0.68013f,0.71419f,0.74247f,0.76407f,0.78849f,0.81416f,0.83035f,0.85224f,0.86246f,0.88096f,0.89040f,0.90733f,0.91475f,0.91869f,0.92745f,0.93770f,0.93942f,0.94685f,0.95122f,0.95689f,0.95864f,0.96425f,0.96413f,0.96299f,0.96783f,0.96903f,0.97210f,0.97736f,0.98066f,0.98764f,0.99646f,0.00000f,0.59823f,0.63907f,0.67974f,0.70880f,0.73227f,0.76256f,0.78524f,0.80520f,0.82245f,0.84663f,0.86363f,0.87693f,0.89224f,0.90453f,0.90945f,0.92358f,0.93013f,0.93752f,0.94211f,0.95271f,0.95092f,0.95889f,0.95923f,0.96162f,0.96504f,0.96551f,0.96526f,0.96959f,0.97557f,0.97967f,0.98614f,0.99512f,0.00000f,0.56195f,0.60392f,0.64041f,0.67922f,0.70746f,0.73547f,0.76162f,0.78370f,0.80429f,0.82725f,0.84327f,0.85849f,0.87182f,0.88996f,0.89782f,0.90865f,0.91829f,0.92890f,0.93508f,0.94287f,0.95259f,0.95575f,0.95757f,0.96155f,0.96476f,0.96393f,0.96554f,0.96733f,0.97333f,0.97741f,0.98305f,0.99464f,0.00000f,0.52765f,0.57050f,0.60716f,0.64125f,0.67629f,0.70620f,0.73423f,0.76153f,0.78821f,0.80534f,0.82488f,0.84038f,0.86012f,0.86827f,0.88602f,0.90017f,0.91402f,0.91977f,0.93110f,0.93749f,0.94743f,0.94895f,0.95598f,0.95825f,0.96064f,0.96193f,0.96654f,0.96502f,0.97130f,0.97317f,0.98165f,0.99341f,0.00000f,0.49413f,0.53691f,0.57699f,0.61725f,0.64766f,0.67663f,0.70377f,0.73470f,0.76171f,0.77787f,0.80202f,0.82252f,0.83664f,0.85492f,0.87108f,0.88359f,0.89714f,0.90688f,0.92174f,0.92678f,0.93960f,0.94534f,0.95157f,0.95798f,0.95749f,0.96293f,0.96242f,0.96422f,0.96768f,0.97353f,0.97793f,0.99172f,
 0.00000f,0.99659f,0.98461f,0.97781f,0.97035f,0.95028f,0.95236f,0.94554f,0.93222f,0.93568f,0.95891f,0.95175f,0.94791f,0.94509f,0.94513f,0.95259f,0.94170f,0.94787f,0.94693f,0.96554f,0.96740f,0.96772f,0.97156f,0.97430f,0.97254f,0.97413f,0.97763f,0.98170f,0.98568f,0.98893f,0.99632f,0.99578f,1.00000f,0.00000f,0.99013f,0.98461f,0.97925f,0.97603f,0.97307f,0.97180f,0.96927f,0.96690f,0.96637f,0.96955f,0.96916f,0.96858f,0.96959f,0.96960f,0.97043f,0.96971f,0.97064f,0.97123f,0.97572f,0.97720f,0.97811f,0.97968f,0.98093f,0.98240f,0.98437f,0.98591f,0.98821f,0.99013f,0.99225f,0.99494f,0.99741f,1.00000f,0.00000f,0.99171f,0.98715f,0.98256f,0.97993f,0.97770f,0.97678f,0.97449f,0.97486f,0.97453f,0.97357f,0.97306f,0.97308f,0.97233f,0.97340f,0.97469f,0.97470f,0.97502f,0.97640f,0.97764f,0.97856f,0.97972f,0.98122f,0.98289f,0.98442f,0.98593f,0.98749f,0.98926f,0.99128f,0.99325f,0.99533f,0.99775f,1.00000f,0.00000f,0.98285f,0.98341f,0.98232f,0.98049f,0.97735f,0.97621f,0.97475f,0.97321f,0.97374f,0.97352f,0.97335f,0.97268f,0.97366f,0.97423f,0.97415f,0.97450f,0.97632f,0.97721f,0.97790f,0.97861f,0.98009f,0.98141f,0.98240f,0.98439f,0.98622f,0.98744f,0.98938f,0.99145f,0.99327f,0.99546f,0.99779f,1.00000f,0.00000f,0.94105f,0.98263f,0.98025f,0.97940f,0.97709f,0.97598f,0.97730f,0.97392f,0.97435f,0.97356f,0.97267f,0.97298f,0.97388f,0.97406f,0.97444f,0.97531f,0.97526f,0.97674f,0.97702f,0.97880f,0.97989f,0.98115f,0.98244f,0.98392f,0.98558f,0.98745f,0.98941f,0.99141f,0.99331f,0.99553f,0.99772f,1.00000f,0.00000f,0.91859f,0.96016f,0.97617f,0.97823f,0.97601f,0.97569f,0.97378f,0.97335f,0.97242f,0.97253f,0.97270f,0.97301f,0.97368f,0.97325f,0.97415f,0.97509f,0.97589f,0.97582f,0.97712f,0.97878f,0.97982f,0.98119f,0.98263f,0.98409f,0.98563f,0.98774f,0.98945f,0.99126f,0.99326f,0.99552f,0.99773f,1.00000f,0.00000f,0.90959f,0.93269f,0.96165f,0.97292f,0.97622f,0.97463f,0.97430f,0.97359f,0.97307f,0.97263f,0.97191f,0.97215f,0.97276f,0.97355f,0.97393f,0.97504f,0.97552f,0.97601f,0.97807f,0.97835f,0.97976f,0.98131f,0.98245f,0.98405f,0.98564f,0.98743f,0.98938f,0.99129f,0.99347f,0.99557f,0.99762f,1.00000f,0.00000f,0.91074f,0.92164f,0.94075f,0.95962f,0.96873f,0.97320f,0.97360f,0.97201f,0.97243f,0.97185f,0.97316f,0.97237f,0.97222f,0.97338f,0.97468f,0.97470f,0.97504f,0.97630f,0.97767f,0.97846f,0.97990f,0.98099f,0.98255f,0.98394f,0.98505f,0.98731f,0.98931f,0.99138f,0.99319f,0.99538f,0.99758f,1.00000f,0.00000f,0.91529f,0.90893f,0.92589f,0.94411f,0.95487f,0.96620f,0.96779f,0.97087f,0.97057f,0.97373f,0.97138f,0.97280f,0.97328f,0.97216f,0.97335f,0.97416f,0.97521f,0.97635f,0.97663f,0.97788f,0.97929f,0.98092f,0.98229f,0.98387f,0.98551f,0.98722f,0.98909f,0.99122f,0.99308f,0.99526f,0.99762f,1.00000f,0.00000f,0.93094f,0.91207f,0.91911f,0.92862f,0.94357f,0.95651f,0.96114f,0.96613f,0.96855f,0.96898f,0.97033f,0.97220f,0.97108f,0.97292f,0.97371f,0.97418f,0.97474f,0.97575f,0.97688f,0.97828f,0.97890f,0.98079f,0.98236f,0.98356f,0.98528f,0.98727f,0.98896f,0.99094f,0.99283f,0.99524f,0.99747f,1.00000f,0.00000f,0.93500f,0.91577f,0.91463f,0.92281f,0.93365f,0.93974f,0.94986f,0.95877f,0.96309f,0.96583f,0.96801f,0.96956f,0.96979f,0.97257f,0.97322f,0.97321f,0.97504f,0.97630f,0.97670f,0.97805f,0.97885f,0.98053f,0.98144f,0.98378f,0.98559f,0.98695f,0.98880f,0.99083f,0.99282f,0.99515f,0.99750f,1.00000f,0.00000f,0.94380f,0.92086f,0.91235f,0.91849f,0.92914f,0.93474f,0.93996f,0.94782f,0.95477f,0.96133f,0.96450f,0.96730f,0.96923f,0.96938f,0.97208f,0.97304f,0.97393f,0.97557f,0.97620f,0.97786f,0.97933f,0.98042f,0.98134f,0.98331f,0.98499f,0.98681f,0.98876f,0.99097f,0.99283f,0.99501f,0.99748f,1.00000f,0.00000f,0.94592f,0.92788f,0.91832f,0.91867f,0.92244f,0.93168f,0.93449f,0.94121f,0.94525f,0.95161f,0.95823f,0.96261f,0.96540f,0.96915f,0.96893f,0.97206f,0.97237f,0.97507f,0.97667f,0.97643f,0.97874f,0.97951f,0.98107f,0.98349f,0.98525f,0.98688f,0.98908f,0.99048f,0.99269f,0.99484f,0.99730f,1.00000f,0.00000f,0.95286f,0.92979f,0.92612f,0.92542f,0.92407f,0.92488f,0.93358f,0.93932f,0.94055f,0.94653f,0.95197f,0.95614f,0.95993f,0.96330f,0.96593f,0.96972f,0.97189f,0.97320f,0.97519f,0.97561f,0.97631f,0.97957f,0.98106f,0.98415f,0.98427f,0.98607f,0.98866f,0.99099f,0.99228f,0.99475f,0.99704f,1.00000f,0.00000f,0.95691f,0.93762f,0.92849f,0.92695f,0.92015f,0.92915f,0.93138f,0.93300f,0.93977f,0.94174f,0.94578f,0.95148f,0.95489f,0.95927f,0.96373f,0.96574f,0.96963f,0.97080f,0.97219f,0.97559f,0.97721f,0.97865f,0.98169f,0.98329f,0.98466f,0.98682f,0.98806f,0.99069f,0.99232f,0.99488f,0.99707f,1.00000f,0.00000f,0.95468f,0.95015f,0.94170f,0.92650f,0.92751f,0.92305f,0.93148f,0.93110f,0.93249f,0.93937f,0.94256f,0.94544f,0.94853f,0.95693f,0.96017f,0.96202f,0.96416f,0.96749f,0.97178f,0.97280f,0.97539f,0.97763f,0.97952f,0.98163f,0.98377f,0.98542f,0.98739f,0.99033f,0.99232f,0.99443f,0.99716f,1.00000f,0.00000f,0.95681f,0.95653f,0.93760f,0.93319f,0.93215f,0.92318f,0.93014f,0.93098f,0.93413f,0.93896f,0.93915f,0.94183f,0.94558f,0.95032f,0.95394f,0.95744f,0.96122f,0.96584f,0.96745f,0.97101f,0.97343f,0.97652f,0.97711f,0.98041f,0.98293f,0.98538f,0.98721f,0.99002f,0.99261f,0.99457f,0.99710f,1.00000f,0.00000f,0.96273f,0.95166f,0.94660f,0.94140f,0.93596f,0.93000f,0.93410f,0.93600f,0.93581f,0.93599f,0.93819f,0.93984f,0.94644f,0.94528f,0.95069f,0.95382f,0.95940f,0.96003f,0.96466f,0.96847f,0.97231f,0.97387f,0.97806f,0.97834f,0.98247f,0.98437f,0.98652f,0.98936f,0.99200f,0.99478f,0.99716f,1.00000f,0.00000f,0.95998f,0.95464f,0.94742f,0.94505f,0.93992f,0.93392f,0.93625f,0.93436f,0.93431f,0.93695f,0.93981f,0.94004f,0.94502f,0.94630f,0.95089f,0.95327f,0.95468f,0.95757f,0.96086f,0.96665f,0.96800f,0.97283f,0.97446f,0.97797f,0.98129f,0.98251f,0.98548f,0.98845f,0.99093f,0.99446f,0.99638f,1.00000f,0.00000f,0.94442f,0.94451f,0.95182f,0.93976f,0.93803f,0.93933f,0.94247f,0.93369f,0.93430f,0.93844f,0.93779f,0.94347f,0.94201f,0.94523f,0.94784f,0.94874f,0.95056f,0.95507f,0.95732f,0.96284f,0.96530f,0.96779f,0.97232f,0.97545f,0.97965f,0.98129f,0.98500f,0.98738f,0.99089f,0.99361f,0.99650f,1.00000f,0.00000f,0.94308f,0.94329f,0.94531f,0.93759f,0.93832f,0.93789f,0.93860f,0.93776f,0.93754f,0.93305f,0.94261f,0.94040f,0.94146f,0.94644f,0.94790f,0.94855f,0.95168f,0.95309f,0.95560f,0.95781f,0.96399f,0.96468f,0.97037f,0.97308f,0.97659f,0.97993f,0.98505f,0.98649f,0.99077f,0.99299f,0.99622f,0.99996f,0.00000f,0.92445f,0.92861f,0.92764f,0.92974f,0.93211f,0.93728f,0.93493f,0.93472f,0.93697f,0.94359f,0.93967f,0.94551f,0.94304f,0.94707f,0.94682f,0.95000f,0.95272f,0.95178f,0.95522f,0.95653f,0.96072f,0.96428f,0.96682f,0.97065f,0.97379f,0.97834f,0.98172f,0.98521f,0.98854f,0.99159f,0.99529f,0.99986f,0.00000f,0.89858f,0.90736f,0.91982f,0.91846f,0.92386f,0.92786f,0.92856f,0.93662f,0.93461f,0.93657f,0.93837f,0.93897f,0.93949f,0.94385f,0.94949f,0.94723f,0.95028f,0.95132f,0.95327f,0.95646f,0.95610f,0.96108f,0.96427f,0.96759f,0.97132f,0.97662f,0.97828f,0.98344f,0.98714f,0.99101f,0.99432f,0.99960f,0.00000f,0.86900f,0.89075f,0.90087f,0.90901f,0.91754f,0.91588f,0.92723f,0.92496f,0.93089f,0.93850f,0.93428f,0.93785f,0.94535f,0.94447f,0.95088f,0.94727f,0.94972f,0.95101f,0.95495f,0.95247f,0.95705f,0.96004f,0.96086f,0.96477f,0.97027f,0.97324f,0.97705f,0.98170f,0.98485f,0.98900f,0.99242f,0.99911f,0.00000f,0.84751f,0.86096f,0.87496f,0.88687f,0.90293f,0.90884f,0.91485f,0.91657f,0.92630f,0.93103f,0.93453f,0.93303f,0.94015f,0.94150f,0.94547f,0.94742f,0.94927f,0.94894f,0.95338f,0.95407f,0.95437f,0.95821f,0.96132f,0.96285f,0.96626f,0.97076f,0.97386f,0.97944f,0.98427f,0.98790f,0.99064f,0.99850f,0.00000f,0.80924f,0.83710f,0.84937f,0.86980f,0.87854f,0.88849f,0.90055f,0.90392f,0.91154f,0.91827f,0.92199f,0.92727f,0.93379f,0.93426f,0.94397f,0.94843f,0.94919f,0.94862f,0.95027f,0.95193f,0.95547f,0.95833f,0.95915f,0.96033f,0.96512f,0.97001f,0.97233f,0.97610f,0.98089f,0.98645f,0.98928f,0.99752f,0.00000f,0.77361f,0.80220f,0.81811f,0.83993f,0.85533f,0.87261f,0.87737f,0.88908f,0.90129f,0.90942f,0.91611f,0.92090f,0.92309f,0.92896f,0.93401f,0.94658f,0.94950f,0.94679f,0.95407f,0.95456f,0.95440f,0.95665f,0.95606f,0.95947f,0.96371f,0.96509f,0.96865f,0.97320f,0.97939f,0.98285f,0.98861f,0.99655f,0.00000f,0.74707f,0.76895f,0.78814f,0.80913f,0.83670f,0.84599f,0.85421f,0.87032f,0.88766f,0.89647f,0.90844f,0.91301f,0.91986f,0.91889f,0.93414f,0.93558f,0.93819f,0.94181f,0.94806f,0.95065f,0.95600f,0.95846f,0.95596f,0.95970f,0.96019f,0.96235f,0.96489f,0.97084f,0.97691f,0.98135f,0.98541f,0.99531f,0.00000f,0.70193f,0.73592f,0.75846f,0.78423f,0.80393f,0.81756f,0.83967f,0.85300f,0.86794f,0.88048f,0.89063f,0.89845f,0.90374f,0.91628f,0.92835f,0.92850f,0.93477f,0.93897f,0.94262f,0.95308f,0.95271f,0.95419f,0.95691f,0.95884f,0.95944f,0.96140f,0.96460f,0.96727f,0.97148f,0.97758f,0.98344f,0.99405f,0.00000f,0.67152f,0.69778f,0.72455f,0.75250f,0.77464f,0.79552f,0.81794f,0.83322f,0.84911f,0.85543f,0.87255f,0.88575f,0.89348f,0.90494f,0.91277f,0.92188f,0.92842f,0.93573f,0.94075f,0.94539f,0.94605f,0.95167f,0.95562f,0.95577f,0.96000f,0.96144f,0.96190f,0.96582f,0.97194f,0.97595f,0.98159f,0.99276f,0.00000f,0.62986f,0.65801f,0.69077f,0.71870f,0.74167f,0.76547f,0.79041f,0.80484f,0.82803f,0.83954f,0.85411f,0.86614f,0.87663f,0.89123f,0.90266f,0.91054f,0.92106f,0.92863f,0.93446f,0.94126f,0.94890f,0.95164f,0.95416f,0.95656f,0.95776f,0.96020f,0.96168f,0.96044f,0.96472f,0.97040f,0.97884f,0.99046f,0.00000f,0.59110f,0.62495f,0.65822f,0.68746f,0.71140f,0.73845f,0.75905f,0.78076f,0.80128f,0.81996f,0.83994f,0.85044f,0.86584f,0.88082f,0.88979f,0.90361f,0.90521f,0.91812f,0.92289f,0.93602f,0.93994f,0.94849f,0.95224f,0.95391f,0.95512f,0.95847f,0.95933f,0.95899f,0.96376f,0.96746f,0.97608f,0.98903f,0.00000f,0.55692f,0.59153f,0.62526f,0.65386f,0.67812f,0.71006f,0.73390f,0.75432f,0.77895f,0.80168f,0.81526f,0.83173f,0.84749f,0.86276f,0.87891f,0.88811f,0.89885f,0.90676f,0.91763f,0.92682f,0.93290f,0.94135f,0.94553f,0.95024f,0.95392f,0.95578f,0.95736f,0.96393f,0.96158f,0.96513f,0.97290f,0.98755f,
 0.00000f,0.99003f,0.97467f,0.96431f,0.94616f,0.96186f,0.96342f,0.95962f,0.95817f,0.95035f,0.94643f,0.94471f,0.94513f,0.94362f,0.93877f,0.96326f,0.95306f,0.95914f,0.95648f,0.96208f,0.95836f,0.95987f,0.96624f,0.96898f,0.96136f,0.96643f,0.98323f,0.98269f,0.98646f,0.99013f,0.99328f,0.99600f,1.00000f,0.00000f,0.99013f,0.98010f,0.97865f,0.97412f,0.97502f,0.97305f,0.97184f,0.97044f,0.96910f,0.96834f,0.96812f,0.96785f,0.96694f,0.96700f,0.97216f,0.97189f,0.97311f,0.97396f,0.97462f,0.97610f,0.97691f,0.97861f,0.97957f,0.98149f,0.98305f,0.98672f,0.98888f,0.99083f,0.99295f,0.99511f,0.99762f,1.00000f,0.00000f,0.99061f,0.98653f,0.98301f,0.97867f,0.97785f,0.97709f,0.97526f,0.97470f,0.97390f,0.97383f,0.97310f,0.97293f,0.97386f,0.97346f,0.97357f,0.97470f,0.97566f,0.97613f,0.97811f,0.97846f,0.98001f,0.98149f,0.98182f,0.98412f,0.98610f,0.98754f,0.98943f,0.99157f,0.99331f,0.99547f,0.99777f,1.00000f,0.00000f,0.98177f,0.98527f,0.98244f,0.97883f,0.97784f,0.97682f,0.97530f,0.97439f,0.97464f,0.97342f,0.97382f,0.97304f,0.97295f,0.97412f,0.97465f,0.97457f,0.97569f,0.97650f,0.97784f,0.97935f,0.97991f,0.98110f,0.98277f,0.98409f,0.98587f,0.98774f,0.98953f,0.99131f,0.99326f,0.99548f,0.99777f,1.00000f,0.00000f,0.94274f,0.98280f,0.98234f,0.98015f,0.97688f,0.97535f,0.97415f,0.97500f,0.97324f,0.97337f,0.97374f,0.97317f,0.97331f,0.97409f,0.97398f,0.97495f,0.97480f,0.97673f,0.97781f,0.97879f,0.97986f,0.98122f,0.98249f,0.98422f,0.98574f,0.98785f,0.98916f,0.99131f,0.99343f,0.99536f,0.99780f,1.00000f,0.00000f,0.91519f,0.95924f,0.97723f,0.97784f,0.97663f,0.97488f,0.97513f,0.97414f,0.97306f,0.97306f,0.97349f,0.97155f,0.97196f,0.97404f,0.97512f,0.97475f,0.97561f,0.97607f,0.97729f,0.97872f,0.98003f,0.98044f,0.98253f,0.98425f,0.98586f,0.98776f,0.98926f,0.99123f,0.99331f,0.99549f,0.99771f,1.00000f,0.00000f,0.90713f,0.92995f,0.95752f,0.97216f,0.97607f,0.97456f,0.97412f,0.97393f,0.97224f,0.97258f,0.97196f,0.97317f,0.97227f,0.97365f,0.97412f,0.97489f,0.97481f,0.97686f,0.97695f,0.97854f,0.97932f,0.98081f,0.98283f,0.98438f,0.98573f,0.98731f,0.98920f,0.99118f,0.99328f,0.99540f,0.99756f,1.00000f,0.00000f,0.91286f,0.91722f,0.93625f,0.95786f,0.96822f,0.97258f,0.97398f,0.97215f,0.97317f,0.97196f,0.97245f,0.97273f,0.97330f,0.97369f,0.97335f,0.97487f,0.97564f,0.97649f,0.97719f,0.97840f,0.97944f,0.98081f,0.98276f,0.98417f,0.98566f,0.98727f,0.98918f,0.99117f,0.99318f,0.99532f,0.99754f,1.00000f,0.00000f,0.91890f,0.91382f,0.92528f,0.93922f,0.95603f,0.96344f,0.96647f,0.97047f,0.97224f,0.97178f,0.97213f,0.97290f,0.97181f,0.97306f,0.97375f,0.97380f,0.97582f,0.97607f,0.97688f,0.97793f,0.97991f,0.98086f,0.98241f,0.98362f,0.98527f,0.98714f,0.98918f,0.99112f,0.99325f,0.99537f,0.99758f,1.00000f,0.00000f,0.92224f,0.91117f,0.91735f,0.93039f,0.94086f,0.95290f,0.95945f,0.96558f,0.96868f,0.96980f,0.97226f,0.97236f,0.97285f,0.97293f,0.97290f,0.97459f,0.97492f,0.97539f,0.97639f,0.97746f,0.97945f,0.98087f,0.98218f,0.98390f,0.98536f,0.98697f,0.98901f,0.99107f,0.99308f,0.99511f,0.99755f,1.00000f,0.00000f,0.93087f,0.91781f,0.91310f,0.92285f,0.92960f,0.93851f,0.95002f,0.95725f,0.96173f,0.96572f,0.96825f,0.96991f,0.97101f,0.97170f,0.97283f,0.97298f,0.97554f,0.97479f,0.97673f,0.97695f,0.97895f,0.97959f,0.98178f,0.98338f,0.98576f,0.98679f,0.98898f,0.99064f,0.99264f,0.99507f,0.99746f,1.00000f,0.00000f,0.93751f,0.92016f,0.91073f,0.91882f,0.92628f,0.93433f,0.94066f,0.94693f,0.95377f,0.95957f,0.96487f,0.96641f,0.96919f,0.97142f,0.97159f,0.97374f,0.97384f,0.97468f,0.97670f,0.97844f,0.97880f,0.98044f,0.98195f,0.98321f,0.98528f,0.98688f,0.98878f,0.99048f,0.99286f,0.99514f,0.99742f,1.00000f,0.00000f,0.95208f,0.92827f,0.91766f,0.91797f,0.91858f,0.92899f,0.93165f,0.94106f,0.94794f,0.95229f,0.95624f,0.96140f,0.96495f,0.96643f,0.97142f,0.97060f,0.97334f,0.97429f,0.97562f,0.97717f,0.97778f,0.97977f,0.98214f,0.98284f,0.98538f,0.98652f,0.98867f,0.99072f,0.99275f,0.99514f,0.99721f,1.00000f,0.00000f,0.94857f,0.93657f,0.92904f,0.91681f,0.92036f,0.92830f,0.92750f,0.93577f,0.93748f,0.94452f,0.94744f,0.95413f,0.96083f,0.96133f,0.96568f,0.96865f,0.97092f,0.97255f,0.97512f,0.97702f,0.97909f,0.97896f,0.98123f,0.98279f,0.98482f,0.98697f,0.98790f,0.99028f,0.99224f,0.99482f,0.99729f,1.00000f,0.00000f,0.95236f,0.93978f,0.93178f,0.92790f,0.92616f,0.92158f,0.92704f,0.93414f,0.93264f,0.93902f,0.94299f,0.94962f,0.95411f,0.95786f,0.96048f,0.96698f,0.96860f,0.97114f,0.97315f,0.97553f,0.97758f,0.97808f,0.98103f,0.98154f,0.98412f,0.98583f,0.98837f,0.99100f,0.99266f,0.99480f,0.99719f,1.00000f,0.00000f,0.95856f,0.94779f,0.93417f,0.93129f,0.92589f,0.92243f,0.92902f,0.92850f,0.93440f,0.93486f,0.94007f,0.94374f,0.94812f,0.95223f,0.95689f,0.95964f,0.96644f,0.96703f,0.97076f,0.97231f,0.97455f,0.97810f,0.97891f,0.98149f,0.98295f,0.98591f,0.98801f,0.99065f,0.99281f,0.99488f,0.99718f,1.00000f,0.00000f,0.96221f,0.94598f,0.94003f,0.93486f,0.92973f,0.92196f,0.92778f,0.93015f,0.93250f,0.93167f,0.93844f,0.93868f,0.94405f,0.94910f,0.95042f,0.95753f,0.96064f,0.96317f,0.96851f,0.97023f,0.97241f,0.97511f,0.97754f,0.98064f,0.98312f,0.98562f,0.98810f,0.98969f,0.99195f,0.99506f,0.99684f,1.00000f,0.00000f,0.96326f,0.95430f,0.94083f,0.92837f,0.93381f,0.92909f,0.93367f,0.93179f,0.93256f,0.93248f,0.93696f,0.94006f,0.94407f,0.94366f,0.94882f,0.95082f,0.95528f,0.95691f,0.96557f,0.96653f,0.97001f,0.97353f,0.97644f,0.97938f,0.98146f,0.98494f,0.98709f,0.98933f,0.99192f,0.99369f,0.99677f,1.00000f,0.00000f,0.96073f,0.95384f,0.94960f,0.94241f,0.93647f,0.93398f,0.93199f,0.93297f,0.93331f,0.93605f,0.93594f,0.93821f,0.94161f,0.94129f,0.94624f,0.94784f,0.95242f,0.95548f,0.95879f,0.96282f,0.96578f,0.97134f,0.97384f,0.97685f,0.98077f,0.98214f,0.98519f,0.98838f,0.99081f,0.99371f,0.99639f,1.00000f,0.00000f,0.95211f,0.95744f,0.95289f,0.94586f,0.93879f,0.93324f,0.93686f,0.93633f,0.92938f,0.93374f,0.93410f,0.93562f,0.94115f,0.94044f,0.94415f,0.94749f,0.94974f,0.95116f,0.95611f,0.95864f,0.96347f,0.96767f,0.97132f,0.97449f,0.97746f,0.98148f,0.98403f,0.98757f,0.98965f,0.99325f,0.99607f,0.99999f,0.00000f,0.95065f,0.94738f,0.94594f,0.93927f,0.93776f,0.93772f,0.93712f,0.93520f,0.93040f,0.93133f,0.93597f,0.93690f,0.94444f,0.94062f,0.94391f,0.94501f,0.94784f,0.95128f,0.95152f,0.95597f,0.96028f,0.96554f,0.96709f,0.97079f,0.97608f,0.97930f,0.98340f,0.98599f,0.98949f,0.99222f,0.99543f,0.99994f,0.00000f,0.93920f,0.93821f,0.94361f,0.93419f,0.93702f,0.93733f,0.93829f,0.93940f,0.93658f,0.93316f,0.93547f,0.93716f,0.93877f,0.94286f,0.94468f,0.94507f,0.94544f,0.94997f,0.95015f,0.95450f,0.95694f,0.95964f,0.96583f,0.96923f,0.97354f,0.97485f,0.97933f,0.98321f,0.98797f,0.99087f,0.99458f,0.99979f,0.00000f,0.92705f,0.92724f,0.93044f,0.93595f,0.93324f,0.93407f,0.93076f,0.93620f,0.93782f,0.93333f,0.93698f,0.94295f,0.93977f,0.94156f,0.94244f,0.94647f,0.94454f,0.94970f,0.95307f,0.95366f,0.95731f,0.95747f,0.96178f,0.96606f,0.96978f,0.97310f,0.97817f,0.98150f,0.98603f,0.99014f,0.99292f,0.99933f,0.00000f,0.89789f,0.91862f,0.91668f,0.92213f,0.93070f,0.92740f,0.93231f,0.93140f,0.93208f,0.93653f,0.94102f,0.93802f,0.93662f,0.94185f,0.94622f,0.94528f,0.94639f,0.95154f,0.95006f,0.95348f,0.95510f,0.95629f,0.96114f,0.96363f,0.96723f,0.97139f,0.97363f,0.97961f,0.98402f,0.98626f,0.99213f,0.99856f,0.00000f,0.87202f,0.88440f,0.89625f,0.91157f,0.91356f,0.91758f,0.92435f,0.93019f,0.93061f,0.92794f,0.92853f,0.93580f,0.93890f,0.93838f,0.94643f,0.94672f,0.94487f,0.94816f,0.94870f,0.95399f,0.95632f,0.95261f,0.95672f,0.96061f,0.96302f,0.96691f,0.97161f,0.97747f,0.98212f,0.98541f,0.99018f,0.99761f,0.00000f,0.84999f,0.86518f,0.87827f,0.88351f,0.89208f,0.89923f,0.90620f,0.91581f,0.91650f,0.92478f,0.92635f,0.93197f,0.93439f,0.93463f,0.93959f,0.94292f,0.94791f,0.94898f,0.95158f,0.94954f,0.95508f,0.95569f,0.95718f,0.95899f,0.96169f,0.96630f,0.97005f,0.97318f,0.97866f,0.98335f,0.98806f,0.99647f,0.00000f,0.81745f,0.83354f,0.85434f,0.86448f,0.87313f,0.88218f,0.89304f,0.89952f,0.90828f,0.91527f,0.92002f,0.92358f,0.92923f,0.93227f,0.93677f,0.94224f,0.94153f,0.94502f,0.94653f,0.95119f,0.95148f,0.95319f,0.95420f,0.95799f,0.95847f,0.96244f,0.96424f,0.97225f,0.97595f,0.98056f,0.98513f,0.99488f,0.00000f,0.78080f,0.80547f,0.82203f,0.83832f,0.85629f,0.86863f,0.87529f,0.88228f,0.89497f,0.90344f,0.90581f,0.91377f,0.92096f,0.93027f,0.93380f,0.93493f,0.94062f,0.94399f,0.94710f,0.94720f,0.95081f,0.95340f,0.95511f,0.95664f,0.95735f,0.95984f,0.96217f,0.96771f,0.97215f,0.97636f,0.98318f,0.99373f,0.00000f,0.74076f,0.76615f,0.79687f,0.81003f,0.82819f,0.83839f,0.85652f,0.86436f,0.87873f,0.89201f,0.89630f,0.90403f,0.91157f,0.91414f,0.92325f,0.92878f,0.93287f,0.94156f,0.94595f,0.94863f,0.94681f,0.94670f,0.95327f,0.95465f,0.95471f,0.95860f,0.96000f,0.96394f,0.96711f,0.97345f,0.98003f,0.99154f,0.00000f,0.70987f,0.73635f,0.75930f,0.77765f,0.79849f,0.81613f,0.82862f,0.85074f,0.85970f,0.86766f,0.88304f,0.89005f,0.90151f,0.90683f,0.91455f,0.92356f,0.92944f,0.93524f,0.93643f,0.94405f,0.94442f,0.94839f,0.95090f,0.95396f,0.95322f,0.95512f,0.95748f,0.95787f,0.96569f,0.96989f,0.97618f,0.98930f,0.00000f,0.67742f,0.69747f,0.72769f,0.74573f,0.77416f,0.78756f,0.80402f,0.82601f,0.84176f,0.85482f,0.86457f,0.87666f,0.88418f,0.89841f,0.90306f,0.91459f,0.92186f,0.92666f,0.93434f,0.93585f,0.94091f,0.95061f,0.95112f,0.94921f,0.95203f,0.95474f,0.95167f,0.95837f,0.96167f,0.96500f,0.97321f,0.98728f,0.00000f,0.63605f,0.66260f,0.69936f,0.71162f,0.74551f,0.76499f,0.78528f,0.79876f,0.81887f,0.83624f,0.84856f,0.85943f,0.86863f,0.88435f,0.89443f,0.90404f,0.91068f,0.91632f,0.92726f,0.93474f,0.93456f,0.93978f,0.94745f,0.95148f,0.95058f,0.95366f,0.95380f,0.95647f,0.95727f,0.96206f,0.96977f,0.98502f,0.00000f,0.60105f,0.62909f,0.66602f,0.68402f,0.70876f,0.73250f,0.75324f,0.77312f,0.79382f,0.81058f,0.82960f,0.83920f,0.85676f,0.86309f,0.87948f,0.88869f,0.89819f,0.91208f,0.91589f,0.92360f,0.93182f,0.93810f,0.93922f,0.94211f,0.94838f,0.94928f,0.95018f,0.94956f,0.95484f,0.95727f,0.96537f,0.98330f,
 0.00000f,0.99506f,0.98016f,0.97299f,0.95863f,0.93426f,0.94289f,0.94477f,0.92857f,0.95345f,0.95584f,0.95540f,0.96018f,0.94879f,0.95141f,0.95243f,0.95297f,0.95207f,0.94417f,0.94998f,0.96922f,0.95784f,0.96331f,0.96645f,0.97309f,0.97625f,0.97628f,0.98095f,0.98614f,0.98987f,0.98878f,0.99625f,1.00000f,0.00000f,0.98960f,0.98192f,0.97935f,0.97479f,0.97056f,0.97178f,0.96793f,0.96703f,0.97069f,0.97081f,0.96970f,0.97046f,0.96986f,0.96975f,0.97032f,0.97116f,0.97102f,0.97175f,0.97317f,0.97673f,0.97751f,0.97899f,0.98063f,0.98250f,0.98441f,0.98639f,0.98814f,0.99036f,0.99267f,0.99456f,0.99732f,1.00000f,0.00000f,0.99064f,0.98531f,0.98140f,0.98012f,0.97720f,0.97625f,0.97519f,0.97408f,0.97379f,0.97379f,0.97373f,0.97276f,0.97357f,0.97371f,0.97464f,0.97495f,0.97495f,0.97612f,0.97737f,0.97910f,0.98013f,0.98117f,0.98286f,0.98444f,0.98607f,0.98756f,0.98942f,0.99118f,0.99336f,0.99549f,0.99772f,1.00000f,0.00000f,0.98447f,0.98676f,0.98149f,0.97975f,0.97790f,0.97668f,0.97534f,0.97420f,0.97442f,0.97284f,0.97295f,0.97326f,0.97360f,0.97438f,0.97444f,0.97510f,0.97559f,0.97641f,0.97739f,0.97902f,0.98030f,0.98100f,0.98261f,0.98419f,0.98598f,0.98754f,0.98941f,0.99120f,0.99345f,0.99552f,0.99767f,1.00000f,0.00000f,0.94279f,0.98242f,0.98226f,0.97907f,0.97851f,0.97677f,0.97465f,0.97258f,0.97298f,0.97353f,0.97325f,0.97384f,0.97373f,0.97379f,0.97475f,0.97468f,0.97557f,0.97612f,0.97796f,0.97893f,0.98027f,0.98131f,0.98281f,0.98400f,0.98610f,0.98748f,0.98941f,0.99153f,0.99339f,0.99541f,0.99776f,1.00000f,0.00000f,0.91419f,0.96051f,0.97814f,0.97920f,0.97706f,0.97541f,0.97392f,0.97410f,0.97388f,0.97323f,0.97267f,0.97341f,0.97364f,0.97361f,0.97440f,0.97563f,0.97570f,0.97737f,0.97740f,0.97846f,0.97971f,0.98098f,0.98278f,0.98427f,0.98573f,0.98751f,0.98938f,0.99117f,0.99329f,0.99542f,0.99777f,1.00000f,0.00000f,0.90845f,0.93154f,0.96032f,0.97297f,0.97684f,0.97424f,0.97342f,0.97355f,0.97227f,0.97257f,0.97231f,0.97262f,0.97304f,0.97354f,0.97381f,0.97500f,0.97554f,0.97625f,0.97754f,0.97843f,0.97959f,0.98102f,0.98247f,0.98443f,0.98584f,0.98731f,0.98952f,0.99108f,0.99325f,0.99533f,0.99762f,1.00000f,0.00000f,0.91112f,0.91912f,0.93732f,0.95591f,0.96769f,0.97377f,0.97397f,0.97362f,0.97321f,0.97260f,0.97310f,0.97266f,0.97301f,0.97393f,0.97385f,0.97502f,0.97570f,0.97620f,0.97723f,0.97789f,0.97999f,0.98093f,0.98216f,0.98391f,0.98562f,0.98727f,0.98898f,0.99115f,0.99312f,0.99533f,0.99747f,1.00000f,0.00000f,0.91762f,0.90876f,0.92701f,0.94152f,0.95374f,0.96484f,0.96993f,0.97065f,0.97255f,0.97151f,0.97290f,0.97202f,0.97129f,0.97290f,0.97324f,0.97413f,0.97491f,0.97605f,0.97678f,0.97787f,0.97959f,0.98042f,0.98238f,0.98396f,0.98575f,0.98721f,0.98937f,0.99110f,0.99319f,0.99535f,0.99755f,1.00000f,0.00000f,0.92208f,0.91047f,0.91206f,0.92847f,0.93951f,0.95131f,0.95908f,0.96495f,0.96784f,0.97113f,0.97096f,0.97111f,0.97229f,0.97265f,0.97333f,0.97386f,0.97560f,0.97566f,0.97715f,0.97840f,0.97880f,0.98031f,0.98207f,0.98370f,0.98502f,0.98736f,0.98859f,0.99104f,0.99282f,0.99495f,0.99738f,1.00000f,0.00000f,0.93879f,0.91015f,0.91248f,0.91762f,0.92794f,0.93940f,0.94980f,0.95471f,0.96156f,0.96681f,0.96951f,0.97221f,0.97031f,0.97275f,0.97241f,0.97404f,0.97480f,0.97501f,0.97637f,0.97687f,0.97871f,0.98036f,0.98213f,0.98307f,0.98507f,0.98694f,0.98896f,0.99081f,0.99306f,0.99527f,0.99748f,1.00000f,0.00000f,0.94061f,0.92406f,0.91278f,0.91442f,0.92353f,0.93073f,0.93843f,0.94677f,0.95492f,0.95922f,0.96297f,0.96578f,0.96825f,0.97040f,0.97141f,0.97387f,0.97411f,0.97466f,0.97587f,0.97767f,0.97884f,0.98036f,0.98192f,0.98362f,0.98491f,0.98711f,0.98880f,0.99052f,0.99287f,0.99527f,0.99734f,1.00000f,0.00000f,0.94095f,0.93246f,0.92179f,0.91622f,0.92056f,0.92683f,0.92979f,0.93691f,0.94560f,0.95149f,0.95447f,0.96226f,0.96556f,0.96752f,0.96767f,0.97145f,0.97234f,0.97455f,0.97744f,0.97638f,0.97835f,0.98000f,0.98131f,0.98262f,0.98467f,0.98673f,0.98854f,0.99038f,0.99262f,0.99493f,0.99726f,1.00000f,0.00000f,0.95101f,0.93155f,0.92015f,0.92167f,0.92037f,0.92427f,0.92814f,0.93138f,0.93880f,0.94443f,0.94739f,0.95366f,0.95664f,0.96296f,0.96418f,0.96754f,0.97052f,0.97182f,0.97419f,0.97646f,0.97776f,0.97995f,0.98052f,0.98316f,0.98417f,0.98681f,0.98866f,0.98974f,0.99258f,0.99472f,0.99739f,1.00000f,0.00000f,0.95645f,0.93734f,0.93197f,0.92199f,0.92043f,0.92181f,0.92575f,0.92981f,0.93459f,0.93984f,0.94045f,0.94863f,0.95288f,0.95459f,0.96055f,0.96372f,0.96765f,0.97126f,0.97236f,0.97541f,0.97783f,0.97951f,0.98139f,0.98268f,0.98392f,0.98595f,0.98830f,0.99038f,0.99261f,0.99517f,0.99718f,1.00000f,0.00000f,0.95532f,0.94464f,0.93474f,0.92780f,0.91700f,0.92633f,0.92299f,0.93005f,0.93457f,0.93636f,0.93969f,0.94298f,0.94538f,0.95268f,0.95663f,0.95844f,0.96337f,0.96725f,0.96977f,0.97175f,0.97548f,0.97796f,0.97993f,0.98141f,0.98372f,0.98533f,0.98748f,0.98993f,0.99285f,0.99471f,0.99729f,1.00000f,0.00000f,0.95615f,0.95092f,0.93577f,0.92775f,0.92817f,0.92353f,0.92561f,0.92816f,0.93003f,0.93079f,0.93639f,0.93908f,0.94285f,0.94579f,0.95047f,0.95544f,0.95732f,0.96301f,0.96588f,0.97029f,0.97272f,0.97457f,0.97682f,0.98114f,0.98343f,0.98512f,0.98721f,0.99002f,0.99199f,0.99425f,0.99724f,1.00000f,0.00000f,0.95881f,0.94929f,0.94264f,0.93104f,0.93193f,0.92599f,0.92764f,0.92547f,0.93142f,0.93299f,0.93149f,0.93793f,0.94245f,0.94481f,0.94664f,0.95046f,0.95500f,0.95921f,0.96159f,0.96635f,0.96993f,0.97329f,0.97579f,0.97896f,0.98099f,0.98456f,0.98624f,0.98985f,0.99104f,0.99407f,0.99713f,1.00000f,0.00000f,0.95860f,0.94864f,0.95376f,0.93794f,0.94061f,0.92949f,0.93213f,0.92923f,0.93064f,0.93723f,0.93252f,0.93507f,0.93998f,0.94466f,0.94144f,0.94909f,0.95056f,0.95300f,0.96003f,0.96335f,0.96564f,0.96948f,0.97238f,0.97569f,0.98023f,0.98303f,0.98649f,0.98805f,0.99071f,0.99315f,0.99674f,1.00000f,0.00000f,0.96229f,0.95500f,0.95559f,0.94004f,0.94098f,0.93347f,0.93640f,0.93020f,0.93042f,0.93512f,0.93248f,0.93880f,0.93911f,0.93903f,0.94578f,0.94846f,0.94836f,0.95166f,0.95258f,0.95794f,0.96158f,0.96478f,0.96954f,0.97416f,0.97690f,0.98033f,0.98338f,0.98656f,0.99061f,0.99362f,0.99601f,0.99999f,0.00000f,0.95874f,0.95616f,0.94668f,0.94603f,0.94449f,0.93703f,0.93628f,0.93617f,0.93290f,0.93397f,0.93850f,0.93751f,0.93734f,0.93806f,0.94354f,0.94191f,0.94615f,0.94973f,0.95026f,0.95534f,0.96069f,0.96393f,0.96570f,0.97052f,0.97603f,0.97762f,0.98258f,0.98567f,0.98843f,0.99179f,0.99516f,0.99991f,0.00000f,0.95142f,0.94444f,0.94203f,0.94930f,0.94091f,0.93974f,0.94056f,0.93592f,0.93637f,0.93794f,0.93655f,0.93546f,0.93968f,0.93930f,0.94021f,0.94373f,0.94568f,0.94861f,0.94882f,0.95127f,0.95794f,0.96124f,0.96321f,0.96593f,0.97127f,0.97346f,0.97895f,0.98305f,0.98730f,0.99121f,0.99356f,0.99969f,0.00000f,0.93588f,0.93656f,0.94049f,0.94289f,0.94029f,0.93645f,0.93708f,0.93866f,0.94081f,0.93682f,0.93463f,0.93702f,0.93831f,0.94028f,0.94229f,0.94262f,0.94270f,0.94447f,0.95068f,0.95155f,0.95393f,0.95517f,0.96056f,0.96519f,0.96783f,0.97040f,0.97748f,0.97947f,0.98439f,0.98806f,0.99244f,0.99920f,0.00000f,0.92335f,0.91744f,0.92177f,0.92980f,0.93248f,0.93473f,0.93099f,0.93423f,0.93684f,0.93487f,0.93443f,0.93567f,0.94215f,0.94203f,0.94290f,0.94722f,0.94351f,0.94259f,0.94930f,0.94816f,0.95341f,0.95340f,0.95409f,0.96007f,0.96464f,0.96754f,0.97269f,0.97692f,0.98181f,0.98680f,0.99077f,0.99818f,0.00000f,0.89364f,0.90946f,0.92674f,0.92264f,0.91869f,0.92663f,0.92366f,0.93013f,0.93330f,0.93026f,0.93393f,0.93576f,0.93578f,0.94015f,0.94309f,0.94302f,0.94544f,0.94357f,0.94414f,0.94764f,0.94958f,0.95284f,0.95405f,0.95678f,0.96060f,0.96492f,0.96890f,0.97548f,0.97967f,0.98441f,0.98785f,0.99671f,0.00000f,0.86790f,0.87858f,0.89519f,0.90970f,0.90524f,0.91616f,0.91280f,0.92028f,0.92450f,0.92918f,0.93021f,0.93617f,0.93573f,0.93463f,0.93989f,0.94295f,0.94664f,0.94277f,0.94519f,0.94626f,0.95077f,0.95116f,0.95327f,0.95425f,0.95751f,0.95986f,0.96505f,0.97195f,0.97599f,0.98114f,0.98581f,0.99524f,0.00000f,0.84484f,0.86114f,0.86892f,0.88172f,0.88638f,0.90287f,0.89987f,0.91191f,0.91384f,0.91642f,0.92513f,0.92478f,0.93271f,0.93898f,0.93117f,0.93817f,0.93708f,0.94070f,0.94715f,0.94871f,0.94853f,0.95023f,0.95283f,0.95292f,0.95683f,0.95879f,0.96378f,0.96753f,0.97364f,0.97635f,0.98234f,0.99372f,0.00000f,0.80922f,0.82774f,0.84802f,0.85678f,0.86744f,0.87287f,0.88550f,0.89816f,0.90353f,0.90739f,0.91376f,0.91662f,0.91805f,0.92543f,0.92852f,0.93787f,0.94070f,0.94189f,0.94647f,0.94523f,0.94505f,0.94827f,0.95357f,0.95164f,0.95505f,0.95519f,0.95947f,0.96114f,0.96961f,0.97444f,0.97937f,0.99143f,0.00000f,0.77896f,0.80051f,0.82037f,0.83894f,0.83917f,0.86040f,0.87301f,0.87606f,0.88381f,0.89701f,0.89979f,0.90366f,0.91540f,0.91845f,0.92501f,0.92931f,0.93660f,0.93641f,0.94246f,0.94302f,0.94177f,0.94698f,0.94779f,0.95013f,0.94945f,0.95573f,0.95548f,0.95840f,0.96469f,0.97047f,0.97416f,0.98910f,0.00000f,0.74391f,0.76668f,0.78522f,0.80621f,0.82061f,0.83520f,0.84634f,0.85697f,0.86854f,0.88118f,0.88781f,0.89605f,0.90248f,0.90667f,0.91575f,0.92354f,0.92995f,0.93210f,0.93860f,0.93889f,0.94366f,0.94449f,0.94540f,0.94791f,0.94716f,0.95232f,0.95340f,0.95403f,0.96132f,0.96680f,0.97189f,0.98658f,0.00000f,0.71097f,0.73961f,0.75422f,0.77715f,0.79111f,0.81121f,0.82338f,0.83637f,0.84257f,0.86295f,0.87527f,0.87889f,0.88794f,0.89849f,0.90275f,0.91139f,0.91767f,0.92044f,0.93225f,0.93714f,0.93961f,0.93728f,0.94300f,0.94503f,0.94658f,0.94991f,0.94691f,0.95332f,0.95561f,0.96126f,0.96717f,0.98422f,0.00000f,0.67450f,0.70371f,0.72100f,0.74471f,0.75955f,0.78327f,0.80242f,0.81650f,0.82734f,0.84281f,0.85225f,0.86209f,0.87288f,0.88732f,0.89358f,0.90257f,0.91382f,0.91559f,0.92557f,0.92784f,0.93587f,0.93803f,0.93507f,0.94185f,0.94779f,0.94740f,0.94833f,0.95064f,0.95393f,0.95864f,0.96310f,0.98152f,0.00000f,0.63690f,0.66635f,0.68903f,0.71482f,0.73737f,0.75822f,0.77332f,0.78682f,0.80426f,0.81883f,0.83469f,0.84257f,0.86060f,0.87252f,0.88095f,0.89171f,0.90330f,0.90580f,0.91245f,0.92169f,0.92371f,0.92959f,0.93476f,0.93537f,0.94365f,0.94106f,0.94712f,0.94647f,0.94980f,0.95509f,0.95909f,0.97826f,
 0.00000f,0.99972f,0.97894f,0.97936f,0.96554f,0.95067f,0.95730f,0.95138f,0.95180f,0.94177f,0.93435f,0.94083f,0.93515f,0.95974f,0.96048f,0.95916f,0.96160f,0.96308f,0.96140f,0.96215f,0.95861f,0.96446f,0.96688f,0.96973f,0.96755f,0.97031f,0.98291f,0.98431f,0.98666f,0.98606f,0.99431f,0.99595f,1.00000f,0.00000f,0.99138f,0.98615f,0.97911f,0.97593f,0.97272f,0.97221f,0.97149f,0.96956f,0.96728f,0.96513f,0.96691f,0.96780f,0.97066f,0.97169f,0.97115f,0.97321f,0.97270f,0.97363f,0.97502f,0.97566f,0.97683f,0.97833f,0.97963f,0.98142f,0.98343f,0.98694f,0.98857f,0.99059f,0.99273f,0.99519f,0.99749f,1.00000f,0.00000f,0.99106f,0.98453f,0.98225f,0.97975f,0.97836f,0.97780f,0.97617f,0.97442f,0.97359f,0.97234f,0.97286f,0.97278f,0.97366f,0.97417f,0.97387f,0.97485f,0.97532f,0.97622f,0.97778f,0.97850f,0.97927f,0.98114f,0.98253f,0.98423f,0.98579f,0.98748f,0.98957f,0.99129f,0.99333f,0.99546f,0.99771f,1.00000f,0.00000f,0.98549f,0.98387f,0.98280f,0.97929f,0.97836f,0.97657f,0.97412f,0.97351f,0.97414f,0.97369f,0.97279f,0.97378f,0.97329f,0.97468f,0.97473f,0.97468f,0.97587f,0.97699f,0.97719f,0.97850f,0.98002f,0.98116f,0.98267f,0.98440f,0.98601f,0.98723f,0.98949f,0.99127f,0.99331f,0.99547f,0.99771f,1.00000f,0.00000f,0.94128f,0.98280f,0.98168f,0.97892f,0.97683f,0.97497f,0.97432f,0.97447f,0.97363f,0.97264f,0.97361f,0.97327f,0.97415f,0.97369f,0.97506f,0.97455f,0.97586f,0.97680f,0.97776f,0.97864f,0.97959f,0.98154f,0.98257f,0.98437f,0.98579f,0.98745f,0.98942f,0.99145f,0.99331f,0.99555f,0.99777f,1.00000f,0.00000f,0.91188f,0.96023f,0.97596f,0.97940f,0.97705f,0.97554f,0.97395f,0.97459f,0.97422f,0.97255f,0.97340f,0.97234f,0.97317f,0.97326f,0.97442f,0.97415f,0.97578f,0.97643f,0.97723f,0.97814f,0.98019f,0.98107f,0.98271f,0.98397f,0.98564f,0.98732f,0.98921f,0.99129f,0.99324f,0.99544f,0.99771f,1.00000f,0.00000f,0.90544f,0.93251f,0.96212f,0.97269f,0.97583f,0.97487f,0.97422f,0.97391f,0.97277f,0.97364f,0.97284f,0.97292f,0.97217f,0.97355f,0.97395f,0.97504f,0.97548f,0.97669f,0.97748f,0.97856f,0.97981f,0.98107f,0.98243f,0.98390f,0.98560f,0.98783f,0.98934f,0.99121f,0.99332f,0.99543f,0.99762f,1.00000f,0.00000f,0.90840f,0.91582f,0.93883f,0.95702f,0.96796f,0.97161f,0.97383f,0.97179f,0.97139f,0.97080f,0.97289f,0.97182f,0.97277f,0.97317f,0.97324f,0.97462f,0.97489f,0.97567f,0.97701f,0.97847f,0.97944f,0.98083f,0.98203f,0.98444f,0.98586f,0.98740f,0.98932f,0.99118f,0.99313f,0.99529f,0.99756f,1.00000f,0.00000f,0.92019f,0.91028f,0.92637f,0.94049f,0.95583f,0.96399f,0.96830f,0.97039f,0.97218f,0.97134f,0.97151f,0.97127f,0.97259f,0.97383f,0.97319f,0.97435f,0.97576f,0.97588f,0.97789f,0.97801f,0.97952f,0.98056f,0.98228f,0.98409f,0.98523f,0.98726f,0.98929f,0.99111f,0.99309f,0.99518f,0.99749f,1.00000f,0.00000f,0.92576f,0.90752f,0.91881f,0.92560f,0.93934f,0.95205f,0.95974f,0.96601f,0.96837f,0.96853f,0.97143f,0.97141f,0.97086f,0.97292f,0.97321f,0.97379f,0.97451f,0.97661f,0.97693f,0.97754f,0.97928f,0.98121f,0.98187f,0.98367f,0.98539f,0.98724f,0.98901f,0.99101f,0.99301f,0.99519f,0.99754f,1.00000f,0.00000f,0.93634f,0.91311f,0.90971f,0.92291f,0.92601f,0.94066f,0.94693f,0.95619f,0.96124f,0.96565f,0.96774f,0.96995f,0.97045f,0.97393f,0.97333f,0.97379f,0.97402f,0.97552f,0.97709f,0.97786f,0.97973f,0.98000f,0.98179f,0.98316f,0.98505f,0.98704f,0.98883f,0.99066f,0.99256f,0.99519f,0.99755f,1.00000f,0.00000f,0.93848f,0.91948f,0.90889f,0.91958f,0.92305f,0.92891f,0.93687f,0.94400f,0.95335f,0.95889f,0.96270f,0.96657f,0.96756f,0.96933f,0.97104f,0.97289f,0.97298f,0.97543f,0.97651f,0.97780f,0.97841f,0.98027f,0.98098f,0.98301f,0.98501f,0.98675f,0.98899f,0.99092f,0.99293f,0.99494f,0.99737f,1.00000f,0.00000f,0.94572f,0.92555f,0.91546f,0.91633f,0.91779f,0.92289f,0.92707f,0.93390f,0.94491f,0.95084f,0.95651f,0.96262f,0.96383f,0.96721f,0.96824f,0.97035f,0.97373f,0.97382f,0.97658f,0.97726f,0.97804f,0.98078f,0.98206f,0.98297f,0.98470f,0.98669f,0.98916f,0.99050f,0.99244f,0.99503f,0.99731f,1.00000f,0.00000f,0.95606f,0.93019f,0.92263f,0.91742f,0.91691f,0.92121f,0.92581f,0.93633f,0.93848f,0.94287f,0.94797f,0.95381f,0.95726f,0.95987f,0.96486f,0.96782f,0.96891f,0.97194f,0.97465f,0.97610f,0.97766f,0.97832f,0.98187f,0.98313f,0.98434f,0.98664f,0.98906f,0.99068f,0.99283f,0.99491f,0.99694f,1.00000f,0.00000f,0.95453f,0.93983f,0.92258f,0.92163f,0.92356f,0.92167f,0.92631f,0.92850f,0.93142f,0.93685f,0.94054f,0.94709f,0.95084f,0.95761f,0.96131f,0.96479f,0.96816f,0.96942f,0.97346f,0.97432f,0.97641f,0.97856f,0.98087f,0.98254f,0.98379f,0.98679f,0.98857f,0.99029f,0.99216f,0.99481f,0.99714f,1.00000f,0.00000f,0.95672f,0.94834f,0.93196f,0.92808f,0.92132f,0.92347f,0.92199f,0.92934f,0.92831f,0.93354f,0.93763f,0.94250f,0.94583f,0.95172f,0.95514f,0.96030f,0.96263f,0.96586f,0.96893f,0.97221f,0.97505f,0.97614f,0.97949f,0.98178f,0.98416f,0.98535f,0.98859f,0.99084f,0.99215f,0.99482f,0.99717f,1.00000f,0.00000f,0.95582f,0.94466f,0.94016f,0.92945f,0.92731f,0.92303f,0.92733f,0.92902f,0.92972f,0.93448f,0.93495f,0.93927f,0.94642f,0.94681f,0.95016f,0.95378f,0.96048f,0.96154f,0.96506f,0.97028f,0.97304f,0.97490f,0.97716f,0.97942f,0.98329f,0.98476f,0.98677f,0.99013f,0.99229f,0.99455f,0.99712f,1.00000f,0.00000f,0.95989f,0.95625f,0.94160f,0.93615f,0.92969f,0.92936f,0.92776f,0.92587f,0.93240f,0.93183f,0.93373f,0.93915f,0.94120f,0.94207f,0.94557f,0.94920f,0.95309f,0.95754f,0.96141f,0.96463f,0.96899f,0.96989f,0.97463f,0.97714f,0.98115f,0.98316f,0.98636f,0.98924f,0.99196f,0.99413f,0.99696f,1.00000f,0.00000f,0.95734f,0.95632f,0.94077f,0.93548f,0.93352f,0.93499f,0.92858f,0.93094f,0.93091f,0.92963f,0.93378f,0.93330f,0.93706f,0.94073f,0.94235f,0.94653f,0.94813f,0.95432f,0.95713f,0.96261f,0.96569f,0.97092f,0.97367f,0.97552f,0.97865f,0.98189f,0.98446f,0.98798f,0.99073f,0.99390f,0.99640f,1.00000f,0.00000f,0.95861f,0.96414f,0.95026f,0.94281f,0.94066f,0.93159f,0.92607f,0.93417f,0.93420f,0.92919f,0.93092f,0.93352f,0.93352f,0.93798f,0.94078f,0.94515f,0.94578f,0.94938f,0.95216f,0.96119f,0.96012f,0.96407f,0.96885f,0.97305f,0.97710f,0.98022f,0.98282f,0.98512f,0.99005f,0.99232f,0.99611f,0.99999f,0.00000f,0.95835f,0.96084f,0.95675f,0.94526f,0.94193f,0.93702f,0.93560f,0.93370f,0.93359f,0.93529f,0.93049f,0.93642f,0.93415f,0.93795f,0.94419f,0.94258f,0.94537f,0.94480f,0.94897f,0.95591f,0.95789f,0.96173f,0.96419f,0.97036f,0.97306f,0.97672f,0.98132f,0.98434f,0.98752f,0.99193f,0.99509f,0.99990f,0.00000f,0.95615f,0.95811f,0.95230f,0.94401f,0.94115f,0.93629f,0.93616f,0.93667f,0.93481f,0.93627f,0.93531f,0.93570f,0.93689f,0.93779f,0.94036f,0.94015f,0.94208f,0.94642f,0.94575f,0.95010f,0.95261f,0.95736f,0.96247f,0.96364f,0.96971f,0.97295f,0.97815f,0.98229f,0.98620f,0.98987f,0.99363f,0.99963f,0.00000f,0.94847f,0.94258f,0.94697f,0.94436f,0.93831f,0.93909f,0.94033f,0.93688f,0.94004f,0.93545f,0.93753f,0.93338f,0.93620f,0.94020f,0.93918f,0.94433f,0.94347f,0.94500f,0.95076f,0.94887f,0.94931f,0.95541f,0.95994f,0.96262f,0.96629f,0.97063f,0.97548f,0.97884f,0.98317f,0.98648f,0.99155f,0.99896f,0.00000f,0.93121f,0.93292f,0.92662f,0.93348f,0.93403f,0.94399f,0.93815f,0.93197f,0.93629f,0.93791f,0.93601f,0.93240f,0.93344f,0.93573f,0.94221f,0.94011f,0.94330f,0.94258f,0.94656f,0.94400f,0.95054f,0.95362f,0.95476f,0.95691f,0.96284f,0.96802f,0.97178f,0.97615f,0.98017f,0.98474f,0.98939f,0.99778f,0.00000f,0.90572f,0.92163f,0.92005f,0.92522f,0.93243f,0.93414f,0.93307f,0.93639f,0.93036f,0.93545f,0.93271f,0.93257f,0.93802f,0.93397f,0.94477f,0.93843f,0.94428f,0.94388f,0.94602f,0.94876f,0.94713f,0.94825f,0.95047f,0.95514f,0.95939f,0.96261f,0.96723f,0.97172f,0.97764f,0.98173f,0.98668f,0.99632f,0.00000f,0.89376f,0.90225f,0.90898f,0.91209f,0.91461f,0.91847f,0.92250f,0.92051f,0.92069f,0.92604f,0.93181f,0.92924f,0.93296f,0.93344f,0.93870f,0.93547f,0.94377f,0.94188f,0.94628f,0.94626f,0.94362f,0.94912f,0.94888f,0.95258f,0.95510f,0.95879f,0.96340f,0.96790f,0.97375f,0.97787f,0.98318f,0.99446f,0.00000f,0.86741f,0.87498f,0.88433f,0.89419f,0.89352f,0.90574f,0.91139f,0.91511f,0.91857f,0.92029f,0.92915f,0.92847f,0.93061f,0.93453f,0.93401f,0.93973f,0.93976f,0.94376f,0.94631f,0.94450f,0.94308f,0.94960f,0.95347f,0.95221f,0.95303f,0.95523f,0.95856f,0.96393f,0.96956f,0.97446f,0.97952f,0.99195f,0.00000f,0.83596f,0.85194f,0.86648f,0.87545f,0.87910f,0.88824f,0.89593f,0.90205f,0.90413f,0.91452f,0.91743f,0.92178f,0.92465f,0.93014f,0.93607f,0.93703f,0.93690f,0.94012f,0.94475f,0.94418f,0.94502f,0.94781f,0.94781f,0.94727f,0.95009f,0.95412f,0.95526f,0.96070f,0.96550f,0.96923f,0.97606f,0.98880f,0.00000f,0.80183f,0.82632f,0.83808f,0.85417f,0.86053f,0.87241f,0.88244f,0.88662f,0.89166f,0.90175f,0.90239f,0.91648f,0.91067f,0.91833f,0.92635f,0.92538f,0.92878f,0.93853f,0.93571f,0.93928f,0.94411f,0.94773f,0.94629f,0.94777f,0.94789f,0.94937f,0.95341f,0.95271f,0.96142f,0.96665f,0.97157f,0.98665f,0.00000f,0.77016f,0.79326f,0.80843f,0.82217f,0.83377f,0.84936f,0.86134f,0.86675f,0.87727f,0.88300f,0.89177f,0.90272f,0.90316f,0.91263f,0.91648f,0.92429f,0.92680f,0.93302f,0.93497f,0.93432f,0.93838f,0.94123f,0.94447f,0.94529f,0.94260f,0.94701f,0.94994f,0.95058f,0.95494f,0.96070f,0.96807f,0.98305f,0.00000f,0.73778f,0.75846f,0.78150f,0.79026f,0.81027f,0.82000f,0.84163f,0.84630f,0.86098f,0.86968f,0.87799f,0.89062f,0.89441f,0.90524f,0.90736f,0.91427f,0.91735f,0.92163f,0.92679f,0.93138f,0.93512f,0.93883f,0.93922f,0.94172f,0.94582f,0.94394f,0.94653f,0.94667f,0.95369f,0.95690f,0.96387f,0.98038f,0.00000f,0.70448f,0.72756f,0.74653f,0.76925f,0.78287f,0.79519f,0.81301f,0.82134f,0.83458f,0.84819f,0.86053f,0.87253f,0.87872f,0.88681f,0.89706f,0.89957f,0.90456f,0.91460f,0.91825f,0.92332f,0.92979f,0.92830f,0.93907f,0.93739f,0.94000f,0.94368f,0.94097f,0.94176f,0.94765f,0.95085f,0.95669f,0.97685f,0.00000f,0.66765f,0.69528f,0.70937f,0.73501f,0.75333f,0.77227f,0.78706f,0.79931f,0.81383f,0.82774f,0.84116f,0.85403f,0.86014f,0.87713f,0.87765f,0.88648f,0.89947f,0.90589f,0.90950f,0.91985f,0.92064f,0.92978f,0.92897f,0.93798f,0.94220f,0.93737f,0.93702f,0.94255f,0.94184f,0.95015f,0.95296f,0.97309f,
 0.00000f,0.99070f,0.97495f,0.96869f,0.97015f,0.96094f,0.95663f,0.95578f,0.95146f,0.94143f,0.94710f,0.95091f,0.94487f,0.93738f,0.93079f,0.94605f,0.94793f,0.95836f,0.96223f,0.96436f,0.96718f,0.96400f,0.96741f,0.97467f,0.97640f,0.97978f,0.97808f,0.97974f,0.98399f,0.98469f,0.98734f,0.99580f,1.00000f,0.00000f,0.99203f,0.98274f,0.97878f,0.97844f,0.97556f,0.97293f,0.97019f,0.97061f,0.96914f,0.96777f,0.96886f,0.96742f,0.96801f,0.96765f,0.96883f,0.96887f,0.97274f,0.97453f,0.97587f,0.97727f,0.97731f,0.97929f,0.98115f,0.98252f,0.98433f,0.98605f,0.98782f,0.99030f,0.99233f,0.99460f,0.99743f,1.00000f,0.00000f,0.99139f,0.98568f,0.98288f,0.98037f,0.97890f,0.97548f,0.97548f,0.97403f,0.97525f,0.97285f,0.97420f,0.97322f,0.97407f,0.97404f,0.97390f,0.97500f,0.97571f,0.97617f,0.97725f,0.97877f,0.98020f,0.98162f,0.98324f,0.98402f,0.98604f,0.98757f,0.98918f,0.99134f,0.99328f,0.99542f,0.99774f,1.00000f,0.00000f,0.98375f,0.98487f,0.98099f,0.97912f,0.97734f,0.97594f,0.97581f,0.97438f,0.97419f,0.97302f,0.97279f,0.97280f,0.97284f,0.97359f,0.97393f,0.97481f,0.97599f,0.97673f,0.97808f,0.97878f,0.97985f,0.98123f,0.98307f,0.98397f,0.98603f,0.98755f,0.98954f,0.99148f,0.99324f,0.99541f,0.99767f,1.00000f,0.00000f,0.94297f,0.98116f,0.98035f,0.97922f,0.97754f,0.97607f,0.97486f,0.97466f,0.97344f,0.97349f,0.97331f,0.97398f,0.97317f,0.97401f,0.97489f,0.97542f,0.97588f,0.97684f,0.97746f,0.97916f,0.97998f,0.98157f,0.98267f,0.98427f,0.98584f,0.98744f,0.98937f,0.99132f,0.99336f,0.99537f,0.99775f,1.00000f,0.00000f,0.91158f,0.96099f,0.97899f,0.97768f,0.97825f,0.97557f,0.97472f,0.97349f,0.97203f,0.97289f,0.97272f,0.97256f,0.97331f,0.97344f,0.97408f,0.97516f,0.97578f,0.97667f,0.97760f,0.97872f,0.98022f,0.98182f,0.98246f,0.98444f,0.98582f,0.98756f,0.98938f,0.99125f,0.99334f,0.99553f,0.99775f,1.00000f,0.00000f,0.90527f,0.93106f,0.95894f,0.97381f,0.97535f,0.97600f,0.97319f,0.97297f,0.97361f,0.97195f,0.97348f,0.97250f,0.97322f,0.97312f,0.97355f,0.97480f,0.97469f,0.97673f,0.97728f,0.97817f,0.97915f,0.98092f,0.98274f,0.98399f,0.98592f,0.98761f,0.98927f,0.99119f,0.99325f,0.99530f,0.99759f,1.00000f,0.00000f,0.91056f,0.91645f,0.94131f,0.95766f,0.96954f,0.97327f,0.97328f,0.97286f,0.97303f,0.97210f,0.97186f,0.97210f,0.97314f,0.97384f,0.97414f,0.97418f,0.97481f,0.97634f,0.97666f,0.97829f,0.97924f,0.98095f,0.98212f,0.98376f,0.98581f,0.98758f,0.98907f,0.99115f,0.99330f,0.99518f,0.99758f,1.00000f,0.00000f,0.91167f,0.91230f,0.92294f,0.93877f,0.95576f,0.96362f,0.96915f,0.97188f,0.97301f,0.97143f,0.97190f,0.97153f,0.97221f,0.97255f,0.97364f,0.97461f,0.97476f,0.97636f,0.97697f,0.97778f,0.97975f,0.98079f,0.98237f,0.98416f,0.98571f,0.98713f,0.98903f,0.99113f,0.99331f,0.99533f,0.99760f,1.00000f,0.00000f,0.92184f,0.90824f,0.91634f,0.92787f,0.93829f,0.95107f,0.96061f,0.96461f,0.96834f,0.97120f,0.97205f,0.97210f,0.97283f,0.97304f,0.97354f,0.97470f,0.97492f,0.97595f,0.97685f,0.97838f,0.97989f,0.98049f,0.98230f,0.98368f,0.98537f,0.98758f,0.98919f,0.99116f,0.99301f,0.99521f,0.99752f,1.00000f,0.00000f,0.93653f,0.91312f,0.91549f,0.91922f,0.92878f,0.93690f,0.94879f,0.95538f,0.96127f,0.96537f,0.96648f,0.96945f,0.97062f,0.97300f,0.97207f,0.97340f,0.97440f,0.97557f,0.97640f,0.97777f,0.97953f,0.98037f,0.98224f,0.98336f,0.98529f,0.98680f,0.98873f,0.99081f,0.99304f,0.99520f,0.99748f,1.00000f,0.00000f,0.93979f,0.91844f,0.91527f,0.91054f,0.92048f,0.93022f,0.93694f,0.94553f,0.95323f,0.95777f,0.96207f,0.96556f,0.96727f,0.97098f,0.97267f,0.97283f,0.97507f,0.97490f,0.97691f,0.97748f,0.97854f,0.97971f,0.98205f,0.98315f,0.98475f,0.98681f,0.98823f,0.99089f,0.99277f,0.99486f,0.99743f,1.00000f,0.00000f,0.94415f,0.92661f,0.91496f,0.91577f,0.92083f,0.92550f,0.92835f,0.93600f,0.94499f,0.94910f,0.95660f,0.95943f,0.96294f,0.96614f,0.96878f,0.97118f,0.97347f,0.97395f,0.97571f,0.97729f,0.97868f,0.98013f,0.98134f,0.98275f,0.98529f,0.98622f,0.98857f,0.99047f,0.99270f,0.99489f,0.99739f,1.00000f,0.00000f,0.94401f,0.93249f,0.91800f,0.91460f,0.91703f,0.92107f,0.92787f,0.93214f,0.93311f,0.94280f,0.94817f,0.95233f,0.95765f,0.96168f,0.96608f,0.96816f,0.97097f,0.97224f,0.97426f,0.97602f,0.97796f,0.97938f,0.98116f,0.98324f,0.98481f,0.98633f,0.98877f,0.99034f,0.99238f,0.99493f,0.99721f,1.00000f,0.00000f,0.95242f,0.94222f,0.92561f,0.92305f,0.91844f,0.92123f,0.92231f,0.93067f,0.93015f,0.93728f,0.94117f,0.94917f,0.94999f,0.95598f,0.95950f,0.96233f,0.96694f,0.96924f,0.97245f,0.97480f,0.97629f,0.97830f,0.98052f,0.98222f,0.98466f,0.98682f,0.98813f,0.99030f,0.99289f,0.99478f,0.99696f,1.00000f,0.00000f,0.95438f,0.94635f,0.93375f,0.92848f,0.92295f,0.92430f,0.92433f,0.92333f,0.93026f,0.93300f,0.93512f,0.94073f,0.94435f,0.94995f,0.95357f,0.95828f,0.96168f,0.96495f,0.96979f,0.97279f,0.97507f,0.97699f,0.97897f,0.98093f,0.98366f,0.98577f,0.98817f,0.99031f,0.99238f,0.99483f,0.99716f,1.00000f,0.00000f,0.95753f,0.94957f,0.93456f,0.92896f,0.92311f,0.92407f,0.92306f,0.92703f,0.93040f,0.92994f,0.93369f,0.93818f,0.93971f,0.94427f,0.95009f,0.95445f,0.95744f,0.96051f,0.96435f,0.96765f,0.97158f,0.97493f,0.97723f,0.97874f,0.98148f,0.98591f,0.98681f,0.98984f,0.99219f,0.99492f,0.99691f,1.00000f,0.00000f,0.95716f,0.95284f,0.94031f,0.93614f,0.93422f,0.92609f,0.92827f,0.92224f,0.92970f,0.93360f,0.93169f,0.93307f,0.93777f,0.94022f,0.94682f,0.94946f,0.95390f,0.95739f,0.96223f,0.96582f,0.96712f,0.97135f,0.97608f,0.97881f,0.98035f,0.98267f,0.98732f,0.98850f,0.99163f,0.99371f,0.99709f,1.00000f,0.00000f,0.96252f,0.95076f,0.94498f,0.93654f,0.93939f,0.92886f,0.92521f,0.93003f,0.93160f,0.93114f,0.93773f,0.93468f,0.93665f,0.94115f,0.94341f,0.94296f,0.94795f,0.95145f,0.95729f,0.96057f,0.96611f,0.96816f,0.97290f,0.97452f,0.97847f,0.98126f,0.98551f,0.98737f,0.99021f,0.99352f,0.99670f,1.00000f,0.00000f,0.96185f,0.95743f,0.95364f,0.94437f,0.94067f,0.93528f,0.93145f,0.92769f,0.93050f,0.93483f,0.93087f,0.93321f,0.93717f,0.93648f,0.94182f,0.94021f,0.94464f,0.94939f,0.95287f,0.95672f,0.95921f,0.96312f,0.96923f,0.97276f,0.97430f,0.97983f,0.98342f,0.98587f,0.98905f,0.99207f,0.99586f,0.99998f,0.00000f,0.95672f,0.95332f,0.95047f,0.94135f,0.94282f,0.94187f,0.93789f,0.92736f,0.93505f,0.92674f,0.92638f,0.93523f,0.93292f,0.93564f,0.94071f,0.94058f,0.94309f,0.94191f,0.95063f,0.95298f,0.95728f,0.95998f,0.96416f,0.96860f,0.97186f,0.97556f,0.98009f,0.98398f,0.98736f,0.99140f,0.99498f,0.99987f,0.00000f,0.94951f,0.95462f,0.95119f,0.94873f,0.94534f,0.93694f,0.93862f,0.93674f,0.93377f,0.92982f,0.93388f,0.93575f,0.93496f,0.94008f,0.93888f,0.93977f,0.94173f,0.94302f,0.94559f,0.94890f,0.95138f,0.95590f,0.96147f,0.96496f,0.97101f,0.97191f,0.97824f,0.98133f,0.98515f,0.98975f,0.99244f,0.99953f,0.00000f,0.95119f,0.94698f,0.94952f,0.94541f,0.94664f,0.93978f,0.94042f,0.93838f,0.93247f,0.93454f,0.92993f,0.93688f,0.93621f,0.93546f,0.94058f,0.93794f,0.94065f,0.94645f,0.94354f,0.94545f,0.94937f,0.95487f,0.95742f,0.96083f,0.96328f,0.96820f,0.97354f,0.97829f,0.98225f,0.98604f,0.99042f,0.99868f,0.00000f,0.94419f,0.94677f,0.94328f,0.94281f,0.93741f,0.93999f,0.93644f,0.93559f,0.93861f,0.93313f,0.93071f,0.93630f,0.93723f,0.93365f,0.93420f,0.94077f,0.94028f,0.94190f,0.94437f,0.94612f,0.94746f,0.94795f,0.95490f,0.95474f,0.95891f,0.96567f,0.97098f,0.97358f,0.97909f,0.98383f,0.98873f,0.99711f,0.00000f,0.92494f,0.93217f,0.93483f,0.93346f,0.93528f,0.92905f,0.92987f,0.93121f,0.93758f,0.93956f,0.93361f,0.93328f,0.93690f,0.93414f,0.93681f,0.93899f,0.94038f,0.94143f,0.94309f,0.94644f,0.94472f,0.94913f,0.94825f,0.95431f,0.95712f,0.95946f,0.96525f,0.96940f,0.97447f,0.97971f,0.98429f,0.99544f,0.00000f,0.90591f,0.91241f,0.91641f,0.92545f,0.91923f,0.92099f,0.92059f,0.92443f,0.93027f,0.93165f,0.93519f,0.93686f,0.93439f,0.93516f,0.93672f,0.93781f,0.93696f,0.93982f,0.94441f,0.94448f,0.94254f,0.94497f,0.94782f,0.95059f,0.95057f,0.95771f,0.96093f,0.96616f,0.97115f,0.97561f,0.98195f,0.99294f,0.00000f,0.88872f,0.89165f,0.89997f,0.89921f,0.90624f,0.91213f,0.91522f,0.92385f,0.92446f,0.92154f,0.92223f,0.93100f,0.92848f,0.93250f,0.93493f,0.93797f,0.93649f,0.93955f,0.93442f,0.94373f,0.94131f,0.94763f,0.94701f,0.94700f,0.94778f,0.95170f,0.95731f,0.96304f,0.96684f,0.97131f,0.97611f,0.98977f,0.00000f,0.85792f,0.86462f,0.87573f,0.88417f,0.88890f,0.90196f,0.90194f,0.91304f,0.91247f,0.91876f,0.91667f,0.91875f,0.92273f,0.92953f,0.93085f,0.93391f,0.93553f,0.93845f,0.93952f,0.94032f,0.94155f,0.94554f,0.94396f,0.94782f,0.94683f,0.94820f,0.95243f,0.95733f,0.96216f,0.96697f,0.97196f,0.98702f,0.00000f,0.82451f,0.83539f,0.84344f,0.86717f,0.87573f,0.88357f,0.88451f,0.88840f,0.89421f,0.90172f,0.90429f,0.90912f,0.91221f,0.92024f,0.92954f,0.92754f,0.93085f,0.93593f,0.93597f,0.93749f,0.94004f,0.94204f,0.94437f,0.93985f,0.94397f,0.94314f,0.94903f,0.95157f,0.95596f,0.96143f,0.96777f,0.98370f,0.00000f,0.78936f,0.81307f,0.82287f,0.83941f,0.84697f,0.86190f,0.86702f,0.87696f,0.88376f,0.88936f,0.89541f,0.89950f,0.90625f,0.90927f,0.91750f,0.92227f,0.92381f,0.92854f,0.92896f,0.93342f,0.93503f,0.93973f,0.93979f,0.93907f,0.94128f,0.94258f,0.94502f,0.94792f,0.95049f,0.95636f,0.96288f,0.97975f,0.00000f,0.75920f,0.78334f,0.79449f,0.81060f,0.82208f,0.83361f,0.84956f,0.85739f,0.86822f,0.87794f,0.88513f,0.89101f,0.89194f,0.91046f,0.91358f,0.91629f,0.91402f,0.92070f,0.92415f,0.92582f,0.93137f,0.93527f,0.93780f,0.94059f,0.93911f,0.93793f,0.93875f,0.94359f,0.94526f,0.95232f,0.95706f,0.97676f,0.00000f,0.72898f,0.74720f,0.76365f,0.77812f,0.79772f,0.81012f,0.82513f,0.83020f,0.84331f,0.85081f,0.86834f,0.87247f,0.88188f,0.89100f,0.89328f,0.90683f,0.90748f,0.91036f,0.92080f,0.92247f,0.92601f,0.92530f,0.93194f,0.93586f,0.93540f,0.93605f,0.93786f,0.94043f,0.94135f,0.94679f,0.95270f,0.97306f,0.00000f,0.69428f,0.71833f,0.73628f,0.75338f,0.76866f,0.78462f,0.79887f,0.81147f,0.82526f,0.83341f,0.84601f,0.86020f,0.86352f,0.87547f,0.87969f,0.88723f,0.90302f,0.90274f,0.90903f,0.91533f,0.91652f,0.92690f,0.92795f,0.92884f,0.93288f,0.93462f,0.93442f,0.93329f,0.93564f,0.94323f,0.94681f,0.96780f,
 0.00000f,0.99550f,0.98505f,0.97335f,0.95053f,0.94526f,0.94570f,0.96727f,0.94316f,0.95831f,0.94851f,0.95661f,0.95664f,0.95121f,0.94954f,0.95529f,0.95390f,0.95682f,0.95791f,0.94512f,0.94386f,0.97129f,0.97460f,0.97524f,0.97969f,0.97688f,0.98283f,0.98274f,0.98761f,0.99074f,0.99350f,0.99631f,1.00000f,0.00000f,0.99108f,0.98456f,0.97891f,0.97611f,0.97135f,0.97020f,0.97253f,0.97081f,0.96914f,0.96967f,0.96912f,0.97012f,0.96966f,0.96891f,0.97006f,0.97116f,0.97219f,0.97205f,0.97225f,0.97402f,0.97847f,0.98001f,0.98091f,0.98302f,0.98451f,0.98690f,0.98850f,0.99062f,0.99280f,0.99502f,0.99749f,1.00000f,0.00000f,0.98998f,0.98441f,0.98289f,0.98085f,0.97703f,0.97597f,0.97593f,0.97487f,0.97359f,0.97310f,0.97312f,0.97304f,0.97415f,0.97412f,0.97360f,0.97440f,0.97595f,0.97674f,0.97701f,0.97863f,0.98002f,0.98122f,0.98287f,0.98446f,0.98602f,0.98738f,0.98929f,0.99120f,0.99333f,0.99541f,0.99770f,1.00000f,0.00000f,0.98428f,0.98641f,0.98097f,0.97927f,0.97741f,0.97683f,0.97447f,0.97457f,0.97416f,0.97404f,0.97373f,0.97287f,0.97416f,0.97384f,0.97428f,0.97512f,0.97617f,0.97721f,0.97732f,0.97885f,0.98031f,0.98122f,0.98260f,0.98404f,0.98587f,0.98753f,0.98936f,0.99119f,0.99332f,0.99553f,0.99776f,1.00000f,0.00000f,0.94485f,0.98218f,0.98088f,0.97937f,0.97762f,0.97485f,0.97485f,0.97425f,0.97304f,0.97364f,0.97325f,0.97296f,0.97284f,0.97384f,0.97404f,0.97522f,0.97557f,0.97713f,0.97785f,0.97804f,0.97984f,0.98089f,0.98238f,0.98458f,0.98582f,0.98761f,0.98953f,0.99139f,0.99339f,0.99551f,0.99773f,1.00000f,0.00000f,0.91616f,0.96117f,0.97748f,0.97870f,0.97704f,0.97379f,0.97466f,0.97360f,0.97281f,0.97311f,0.97330f,0.97315f,0.97454f,0.97384f,0.97412f,0.97456f,0.97602f,0.97649f,0.97727f,0.97857f,0.97982f,0.98052f,0.98273f,0.98439f,0.98563f,0.98727f,0.98944f,0.99140f,0.99330f,0.99550f,0.99771f,1.00000f,0.00000f,0.90503f,0.93372f,0.95889f,0.97457f,0.97365f,0.97580f,0.97428f,0.97398f,0.97286f,0.97223f,0.97261f,0.97273f,0.97314f,0.97394f,0.97443f,0.97384f,0.97525f,0.97698f,0.97644f,0.97841f,0.97951f,0.98131f,0.98265f,0.98398f,0.98570f,0.98749f,0.98915f,0.99136f,0.99340f,0.99541f,0.99760f,1.00000f,0.00000f,0.90899f,0.91881f,0.93702f,0.95820f,0.96880f,0.97199f,0.97239f,0.97279f,0.97254f,0.97269f,0.97230f,0.97197f,0.97310f,0.97381f,0.97444f,0.97392f,0.97504f,0.97587f,0.97719f,0.97849f,0.97967f,0.98073f,0.98211f,0.98445f,0.98583f,0.98747f,0.98921f,0.99115f,0.99310f,0.99537f,0.99751f,1.00000f,0.00000f,0.91915f,0.91211f,0.92215f,0.93753f,0.95333f,0.96451f,0.96828f,0.97117f,0.97188f,0.97258f,0.97260f,0.97199f,0.97177f,0.97257f,0.97210f,0.97418f,0.97525f,0.97591f,0.97630f,0.97775f,0.97938f,0.98099f,0.98207f,0.98358f,0.98529f,0.98720f,0.98935f,0.99103f,0.99312f,0.99526f,0.99758f,1.00000f,0.00000f,0.92845f,0.90709f,0.91514f,0.92827f,0.93801f,0.95067f,0.95862f,0.96441f,0.96767f,0.97137f,0.97179f,0.97166f,0.97223f,0.97230f,0.97374f,0.97412f,0.97503f,0.97578f,0.97691f,0.97810f,0.97926f,0.98073f,0.98240f,0.98337f,0.98556f,0.98723f,0.98911f,0.99111f,0.99297f,0.99521f,0.99757f,1.00000f,0.00000f,0.93170f,0.91090f,0.91237f,0.91897f,0.92692f,0.93636f,0.94801f,0.95623f,0.96146f,0.96491f,0.96777f,0.96916f,0.96972f,0.97089f,0.97325f,0.97398f,0.97461f,0.97588f,0.97679f,0.97743f,0.97858f,0.98100f,0.98197f,0.98374f,0.98528f,0.98711f,0.98904f,0.99081f,0.99301f,0.99522f,0.99738f,1.00000f,0.00000f,0.93911f,0.91915f,0.91271f,0.91629f,0.92376f,0.93172f,0.93937f,0.94375f,0.95082f,0.95698f,0.96439f,0.96490f,0.96914f,0.96967f,0.97092f,0.97341f,0.97413f,0.97503f,0.97624f,0.97737f,0.97873f,0.98087f,0.98219f,0.98300f,0.98509f,0.98685f,0.98885f,0.99046f,0.99287f,0.99500f,0.99741f,1.00000f,0.00000f,0.94083f,0.92858f,0.92053f,0.91354f,0.91958f,0.92479f,0.92747f,0.93571f,0.94343f,0.95136f,0.95761f,0.95874f,0.96393f,0.96582f,0.97042f,0.97253f,0.97216f,0.97479f,0.97513f,0.97760f,0.97847f,0.98016f,0.98221f,0.98293f,0.98487f,0.98667f,0.98870f,0.99016f,0.99277f,0.99488f,0.99747f,1.00000f,0.00000f,0.94604f,0.93023f,0.91632f,0.91252f,0.92040f,0.92136f,0.92801f,0.93111f,0.93782f,0.94088f,0.94596f,0.95262f,0.95783f,0.96066f,0.96478f,0.96677f,0.97128f,0.97203f,0.97351f,0.97467f,0.97869f,0.97904f,0.98061f,0.98340f,0.98432f,0.98652f,0.98806f,0.99055f,0.99282f,0.99484f,0.99720f,1.00000f,0.00000f,0.95366f,0.93697f,0.92769f,0.92025f,0.92030f,0.92153f,0.92430f,0.92652f,0.93090f,0.93644f,0.94123f,0.94817f,0.95046f,0.95647f,0.95922f,0.96464f,0.96630f,0.96855f,0.97245f,0.97339f,0.97713f,0.97872f,0.98059f,0.98326f,0.98448f,0.98564f,0.98792f,0.99015f,0.99250f,0.99473f,0.99683f,1.00000f,0.00000f,0.95374f,0.94129f,0.92643f,0.92391f,0.92106f,0.91627f,0.92216f,0.92542f,0.92719f,0.93148f,0.93768f,0.93799f,0.94406f,0.95007f,0.95189f,0.95882f,0.96128f,0.96570f,0.96808f,0.97074f,0.97489f,0.97691f,0.98007f,0.98158f,0.98386f,0.98583f,0.98764f,0.98997f,0.99221f,0.99456f,0.99702f,1.00000f,0.00000f,0.96235f,0.94566f,0.93882f,0.92959f,0.92576f,0.92644f,0.92826f,0.92731f,0.92826f,0.93027f,0.93078f,0.93529f,0.93850f,0.94338f,0.94751f,0.95417f,0.95558f,0.95907f,0.96574f,0.96695f,0.97235f,0.97460f,0.97669f,0.98088f,0.98184f,0.98611f,0.98723f,0.99020f,0.99191f,0.99450f,0.99676f,1.00000f,0.00000f,0.95914f,0.94780f,0.94250f,0.93284f,0.93180f,0.92036f,0.91967f,0.92936f,0.92665f,0.93083f,0.93207f,0.93458f,0.93634f,0.94058f,0.94283f,0.94857f,0.95084f,0.95645f,0.96030f,0.96421f,0.96905f,0.97082f,0.97577f,0.97809f,0.98150f,0.98237f,0.98664f,0.98860f,0.99144f,0.99455f,0.99674f,1.00000f,0.00000f,0.96048f,0.95284f,0.94521f,0.94009f,0.93349f,0.93145f,0.92997f,0.92470f,0.92672f,0.93183f,0.92838f,0.93523f,0.93740f,0.93867f,0.94308f,0.94215f,0.94609f,0.94910f,0.95890f,0.95655f,0.96346f,0.96711f,0.97110f,0.97525f,0.97912f,0.98044f,0.98364f,0.98754f,0.99054f,0.99428f,0.99707f,1.00000f,0.00000f,0.96096f,0.96202f,0.95093f,0.94251f,0.94097f,0.93741f,0.93060f,0.93079f,0.92844f,0.92988f,0.92935f,0.93039f,0.93146f,0.93726f,0.94087f,0.94503f,0.94127f,0.94769f,0.95041f,0.95542f,0.95907f,0.96479f,0.96715f,0.97196f,0.97604f,0.97933f,0.98297f,0.98617f,0.98902f,0.99185f,0.99550f,0.99998f,0.00000f,0.95558f,0.95777f,0.95004f,0.94477f,0.94399f,0.93477f,0.93242f,0.92680f,0.93098f,0.93242f,0.93137f,0.93125f,0.93595f,0.93346f,0.93882f,0.94019f,0.94150f,0.94209f,0.94707f,0.95061f,0.95506f,0.95836f,0.96383f,0.96774f,0.97170f,0.97756f,0.97892f,0.98309f,0.98768f,0.99032f,0.99392f,0.99984f,0.00000f,0.95576f,0.95593f,0.95646f,0.94715f,0.95094f,0.94378f,0.93476f,0.93203f,0.93378f,0.93131f,0.93166f,0.93214f,0.93336f,0.93446f,0.93725f,0.93622f,0.94267f,0.94549f,0.94524f,0.94758f,0.95035f,0.95444f,0.95755f,0.96325f,0.96944f,0.97051f,0.97542f,0.98083f,0.98549f,0.98784f,0.99208f,0.99938f,0.00000f,0.96329f,0.94398f,0.95150f,0.94971f,0.94649f,0.93688f,0.93717f,0.93865f,0.93622f,0.93030f,0.93271f,0.93184f,0.93352f,0.93825f,0.93671f,0.93930f,0.93963f,0.94195f,0.94427f,0.94526f,0.94680f,0.95004f,0.95366f,0.96007f,0.96534f,0.96757f,0.97149f,0.97554f,0.98181f,0.98467f,0.99015f,0.99864f,0.00000f,0.94413f,0.94510f,0.94802f,0.94153f,0.93666f,0.93864f,0.93476f,0.94051f,0.93786f,0.93112f,0.93349f,0.93337f,0.93479f,0.93621f,0.93775f,0.93917f,0.94001f,0.94118f,0.94425f,0.94259f,0.94642f,0.94784f,0.95065f,0.95390f,0.95988f,0.96384f,0.96938f,0.97235f,0.97676f,0.98317f,0.98634f,0.99658f,0.00000f,0.92537f,0.93703f,0.94301f,0.93979f,0.94221f,0.93843f,0.93669f,0.93395f,0.93185f,0.93334f,0.93290f,0.93826f,0.93413f,0.93376f,0.93550f,0.94016f,0.93986f,0.93879f,0.94361f,0.93843f,0.94510f,0.94309f,0.94999f,0.95226f,0.95550f,0.95823f,0.96279f,0.96804f,0.97248f,0.97804f,0.98230f,0.99432f,0.00000f,0.91541f,0.91281f,0.92590f,0.92815f,0.93218f,0.92674f,0.92825f,0.92660f,0.92662f,0.93086f,0.93653f,0.93254f,0.93588f,0.93404f,0.93768f,0.93305f,0.94219f,0.93788f,0.93749f,0.94266f,0.94230f,0.94314f,0.94665f,0.94659f,0.95144f,0.95354f,0.95856f,0.96349f,0.96812f,0.97328f,0.97921f,0.99149f,0.00000f,0.89818f,0.90559f,0.91335f,0.91153f,0.90971f,0.92599f,0.92177f,0.92117f,0.92110f,0.92851f,0.92925f,0.92540f,0.93362f,0.93261f,0.93493f,0.93677f,0.93499f,0.93851f,0.93517f,0.94026f,0.94121f,0.94221f,0.94298f,0.94587f,0.94844f,0.95166f,0.95168f,0.95938f,0.96321f,0.96947f,0.97405f,0.98798f,0.00000f,0.86631f,0.87757f,0.88653f,0.89790f,0.90253f,0.90528f,0.90577f,0.91752f,0.91601f,0.92106f,0.91683f,0.92846f,0.93040f,0.92661f,0.93196f,0.92982f,0.92888f,0.93677f,0.93692f,0.93886f,0.94017f,0.94459f,0.94269f,0.94356f,0.94726f,0.94610f,0.94819f,0.95462f,0.95944f,0.96217f,0.96913f,0.98422f,0.00000f,0.84548f,0.85452f,0.86533f,0.87168f,0.87790f,0.88429f,0.89797f,0.90014f,0.90786f,0.90455f,0.90930f,0.91485f,0.91947f,0.91939f,0.92517f,0.92868f,0.93027f,0.93398f,0.93103f,0.93616f,0.93489f,0.93961f,0.93970f,0.94066f,0.94256f,0.94165f,0.94562f,0.94934f,0.95508f,0.95828f,0.96417f,0.98091f,0.00000f,0.81068f,0.82440f,0.83626f,0.84133f,0.85304f,0.87216f,0.87383f,0.88547f,0.88804f,0.88938f,0.89929f,0.90855f,0.90714f,0.91483f,0.92114f,0.92373f,0.92410f,0.92598f,0.92896f,0.93385f,0.93481f,0.93344f,0.93737f,0.93790f,0.94090f,0.94176f,0.94132f,0.94395f,0.94882f,0.95410f,0.95994f,0.97652f,0.00000f,0.77917f,0.79477f,0.81350f,0.82875f,0.83434f,0.84478f,0.85998f,0.86278f,0.87467f,0.87890f,0.88392f,0.89195f,0.89495f,0.89865f,0.91098f,0.91473f,0.91418f,0.91908f,0.92842f,0.92612f,0.92542f,0.93534f,0.93281f,0.93368f,0.93950f,0.93575f,0.93521f,0.93885f,0.94192f,0.94717f,0.95287f,0.97211f,0.00000f,0.75249f,0.76141f,0.78661f,0.80188f,0.80729f,0.81793f,0.82949f,0.83927f,0.84956f,0.86396f,0.87129f,0.87923f,0.88261f,0.89035f,0.89762f,0.90386f,0.90722f,0.90973f,0.91774f,0.92151f,0.92404f,0.92605f,0.93242f,0.93236f,0.93096f,0.93181f,0.93346f,0.93504f,0.93677f,0.94042f,0.94555f,0.96875f,0.00000f,0.71826f,0.73825f,0.75066f,0.76531f,0.78000f,0.79571f,0.80812f,0.81723f,0.83483f,0.83923f,0.85177f,0.85978f,0.86842f,0.87711f,0.88524f,0.88803f,0.89452f,0.89860f,0.90458f,0.91114f,0.91773f,0.91928f,0.92102f,0.92490f,0.92568f,0.92784f,0.92689f,0.92942f,0.93223f,0.93445f,0.94302f,0.96371f,
 0.00000f,0.99759f,0.98364f,0.98218f,0.96850f,0.94779f,0.95333f,0.95571f,0.94696f,0.94559f,0.96224f,0.95512f,0.95630f,0.95307f,0.95149f,0.95498f,0.95526f,0.95137f,0.95684f,0.95921f,0.95785f,0.96316f,0.96884f,0.95760f,0.96946f,0.98144f,0.98190f,0.98581f,0.98692f,0.99180f,0.99243f,0.99621f,1.00000f,0.00000f,0.99126f,0.98568f,0.98097f,0.97645f,0.97317f,0.97116f,0.96954f,0.96845f,0.96733f,0.97253f,0.96853f,0.97016f,0.97040f,0.97062f,0.97135f,0.97081f,0.97241f,0.97312f,0.97452f,0.97568f,0.97648f,0.97804f,0.97967f,0.98085f,0.98497f,0.98685f,0.98849f,0.99095f,0.99300f,0.99502f,0.99754f,1.00000f,0.00000f,0.98951f,0.98774f,0.98175f,0.97887f,0.97795f,0.97650f,0.97589f,0.97437f,0.97327f,0.97276f,0.97314f,0.97295f,0.97316f,0.97408f,0.97416f,0.97552f,0.97559f,0.97644f,0.97762f,0.97864f,0.97980f,0.98074f,0.98251f,0.98425f,0.98578f,0.98751f,0.98944f,0.99124f,0.99336f,0.99552f,0.99774f,1.00000f,0.00000f,0.98294f,0.98608f,0.98229f,0.98050f,0.97818f,0.97683f,0.97449f,0.97419f,0.97313f,0.97360f,0.97277f,0.97307f,0.97277f,0.97456f,0.97390f,0.97480f,0.97652f,0.97632f,0.97771f,0.97814f,0.98031f,0.98142f,0.98270f,0.98411f,0.98593f,0.98763f,0.98929f,0.99134f,0.99319f,0.99537f,0.99778f,1.00000f,0.00000f,0.94123f,0.98119f,0.98210f,0.97912f,0.97878f,0.97619f,0.97369f,0.97507f,0.97378f,0.97328f,0.97272f,0.97249f,0.97312f,0.97343f,0.97405f,0.97552f,0.97590f,0.97632f,0.97746f,0.97819f,0.98023f,0.98120f,0.98271f,0.98428f,0.98603f,0.98760f,0.98914f,0.99138f,0.99332f,0.99543f,0.99778f,1.00000f,0.00000f,0.91458f,0.96090f,0.97640f,0.97850f,0.97735f,0.97682f,0.97474f,0.97479f,0.97334f,0.97362f,0.97352f,0.97262f,0.97314f,0.97422f,0.97433f,0.97455f,0.97522f,0.97609f,0.97770f,0.97886f,0.97999f,0.98141f,0.98265f,0.98430f,0.98586f,0.98755f,0.98952f,0.99153f,0.99322f,0.99540f,0.99776f,1.00000f,0.00000f,0.90709f,0.93150f,0.95868f,0.97314f,0.97642f,0.97597f,0.97306f,0.97300f,0.97336f,0.97200f,0.97271f,0.97299f,0.97337f,0.97423f,0.97388f,0.97460f,0.97571f,0.97627f,0.97731f,0.97845f,0.98001f,0.98085f,0.98263f,0.98389f,0.98584f,0.98760f,0.98910f,0.99111f,0.99327f,0.99549f,0.99764f,1.00000f,0.00000f,0.91480f,0.91749f,0.93888f,0.95473f,0.96764f,0.97171f,0.97385f,0.97303f,0.97383f,0.97213f,0.97141f,0.97271f,0.97258f,0.97279f,0.97389f,0.97496f,0.97504f,0.97564f,0.97754f,0.97770f,0.97977f,0.98144f,0.98272f,0.98415f,0.98562f,0.98744f,0.98919f,0.99084f,0.99328f,0.99529f,0.99751f,1.00000f,0.00000f,0.92098f,0.90761f,0.92395f,0.93781f,0.95238f,0.96303f,0.96958f,0.97131f,0.97261f,0.97274f,0.97263f,0.97280f,0.97172f,0.97239f,0.97288f,0.97457f,0.97568f,0.97604f,0.97715f,0.97861f,0.98036f,0.98079f,0.98238f,0.98389f,0.98543f,0.98726f,0.98912f,0.99108f,0.99320f,0.99532f,0.99755f,1.00000f,0.00000f,0.92024f,0.90726f,0.91349f,0.92823f,0.94012f,0.94871f,0.96082f,0.96502f,0.96772f,0.96991f,0.97080f,0.97229f,0.97188f,0.97166f,0.97292f,0.97334f,0.97446f,0.97535f,0.97707f,0.97798f,0.97977f,0.98102f,0.98186f,0.98414f,0.98550f,0.98689f,0.98898f,0.99127f,0.99316f,0.99514f,0.99753f,1.00000f,0.00000f,0.92723f,0.91454f,0.91033f,0.91763f,0.92659f,0.94031f,0.94662f,0.95473f,0.96131f,0.96524f,0.96756f,0.96909f,0.97113f,0.97160f,0.97196f,0.97360f,0.97356f,0.97569f,0.97722f,0.97746f,0.97911f,0.97965f,0.98175f,0.98345f,0.98547f,0.98725f,0.98863f,0.99128f,0.99283f,0.99511f,0.99749f,1.00000f,0.00000f,0.93666f,0.91981f,0.91227f,0.91744f,0.92044f,0.92809f,0.93592f,0.94124f,0.95125f,0.95783f,0.96154f,0.96527f,0.96782f,0.96993f,0.97149f,0.97279f,0.97422f,0.97529f,0.97599f,0.97776f,0.97846f,0.97984f,0.98173f,0.98315f,0.98494f,0.98671f,0.98859f,0.99058f,0.99296f,0.99512f,0.99735f,1.00000f,0.00000f,0.94308f,0.91813f,0.91928f,0.91270f,0.91486f,0.92347f,0.92823f,0.93806f,0.94020f,0.94913f,0.95388f,0.95854f,0.96235f,0.96585f,0.96886f,0.97142f,0.97262f,0.97446f,0.97588f,0.97632f,0.97833f,0.97933f,0.98175f,0.98346f,0.98500f,0.98688f,0.98856f,0.99049f,0.99251f,0.99496f,0.99734f,1.00000f,0.00000f,0.94752f,0.92983f,0.92056f,0.92019f,0.92236f,0.91764f,0.92276f,0.93258f,0.93457f,0.94212f,0.94641f,0.95134f,0.95740f,0.96153f,0.96492f,0.96818f,0.96996f,0.97266f,0.97374f,0.97614f,0.97799f,0.97967f,0.98128f,0.98263f,0.98438f,0.98685f,0.98879f,0.99002f,0.99269f,0.99498f,0.99713f,1.00000f,0.00000f,0.95235f,0.93137f,0.92682f,0.91887f,0.91814f,0.92160f,0.92009f,0.92923f,0.92985f,0.93549f,0.94165f,0.94348f,0.94913f,0.95615f,0.95808f,0.96210f,0.96597f,0.96965f,0.97164f,0.97429f,0.97728f,0.97796f,0.98084f,0.98236f,0.98458f,0.98585f,0.98762f,0.99055f,0.99250f,0.99465f,0.99692f,1.00000f,0.00000f,0.95007f,0.94077f,0.92919f,0.92096f,0.91933f,0.91483f,0.92302f,0.92691f,0.93242f,0.93213f,0.93840f,0.93835f,0.94538f,0.94999f,0.95512f,0.95716f,0.96236f,0.96491f,0.96823f,0.97100f,0.97285f,0.97678f,0.97950f,0.98088f,0.98335f,0.98565f,0.98760f,0.99036f,0.99250f,0.99453f,0.99700f,1.00000f,0.00000f,0.95992f,0.94836f,0.93464f,0.93027f,0.92776f,0.91996f,0.92309f,0.92437f,0.92962f,0.92781f,0.93317f,0.93622f,0.93797f,0.94107f,0.94622f,0.95121f,0.95736f,0.96075f,0.96491f,0.96704f,0.97095f,0.97561f,0.97623f,0.97933f,0.98213f,0.98498f,0.98760f,0.98919f,0.99205f,0.99486f,0.99717f,1.00000f,0.00000f,0.95474f,0.95054f,0.94121f,0.93600f,0.92910f,0.93127f,0.92834f,0.92627f,0.92481f,0.93066f,0.93028f,0.93279f,0.93823f,0.94033f,0.94342f,0.94876f,0.95163f,0.95844f,0.95863f,0.96181f,0.96652f,0.97073f,0.97235f,0.97840f,0.98022f,0.98303f,0.98587f,0.98823f,0.99105f,0.99400f,0.99711f,1.00000f,0.00000f,0.96854f,0.95769f,0.94532f,0.93599f,0.93097f,0.93372f,0.92643f,0.92596f,0.92789f,0.92717f,0.92848f,0.93128f,0.93692f,0.93554f,0.93891f,0.94447f,0.94584f,0.95081f,0.95562f,0.95806f,0.96328f,0.96584f,0.96980f,0.97464f,0.97693f,0.98070f,0.98450f,0.98706f,0.99029f,0.99334f,0.99652f,1.00000f,0.00000f,0.96377f,0.95232f,0.95443f,0.94963f,0.93736f,0.93321f,0.93490f,0.93041f,0.93018f,0.92854f,0.92772f,0.93324f,0.93147f,0.93487f,0.94201f,0.93816f,0.94377f,0.94485f,0.95051f,0.95275f,0.96108f,0.96330f,0.96626f,0.97042f,0.97544f,0.97755f,0.98163f,0.98532f,0.98929f,0.99259f,0.99535f,0.99997f,0.00000f,0.96182f,0.95881f,0.95804f,0.94815f,0.94043f,0.93202f,0.92461f,0.93386f,0.93093f,0.92845f,0.92747f,0.93206f,0.93299f,0.93452f,0.93805f,0.93843f,0.93978f,0.94069f,0.94551f,0.95044f,0.95261f,0.95697f,0.95996f,0.96685f,0.97140f,0.97479f,0.97916f,0.98296f,0.98684f,0.99070f,0.99408f,0.99981f,0.00000f,0.96364f,0.95886f,0.95427f,0.94864f,0.94124f,0.93965f,0.93719f,0.93383f,0.93579f,0.92818f,0.93288f,0.93178f,0.93246f,0.93297f,0.93602f,0.93856f,0.93980f,0.94143f,0.94227f,0.94532f,0.94940f,0.95229f,0.95534f,0.96390f,0.96774f,0.97103f,0.97552f,0.97911f,0.98429f,0.98720f,0.99155f,0.99933f,0.00000f,0.96125f,0.95835f,0.95062f,0.95037f,0.94934f,0.94216f,0.93739f,0.93695f,0.93847f,0.93318f,0.93192f,0.93398f,0.93253f,0.93199f,0.93501f,0.93970f,0.93706f,0.93984f,0.94204f,0.94299f,0.94558f,0.94961f,0.95513f,0.95886f,0.96260f,0.96664f,0.97032f,0.97609f,0.98014f,0.98373f,0.98909f,0.99805f,0.00000f,0.94963f,0.94470f,0.94695f,0.94371f,0.94784f,0.94312f,0.93698f,0.93869f,0.94176f,0.93573f,0.93519f,0.93490f,0.92805f,0.93499f,0.93171f,0.93846f,0.93908f,0.93725f,0.93968f,0.94287f,0.94442f,0.94395f,0.94905f,0.95333f,0.95850f,0.96215f,0.96736f,0.96969f,0.97649f,0.98101f,0.98567f,0.99624f,0.00000f,0.93957f,0.94138f,0.93943f,0.93808f,0.94128f,0.93669f,0.93321f,0.93548f,0.94265f,0.93454f,0.93234f,0.93450f,0.93195f,0.93585f,0.93556f,0.93798f,0.93887f,0.94163f,0.94081f,0.94229f,0.94331f,0.94292f,0.94708f,0.94944f,0.95443f,0.95767f,0.96264f,0.96599f,0.97323f,0.97676f,0.98053f,0.99345f,0.00000f,0.92254f,0.92995f,0.92780f,0.92610f,0.93271f,0.93235f,0.93775f,0.92777f,0.93548f,0.93564f,0.93755f,0.93580f,0.93261f,0.93695f,0.92982f,0.93357f,0.93656f,0.93870f,0.93762f,0.93919f,0.93940f,0.94438f,0.94308f,0.94737f,0.94877f,0.95480f,0.95768f,0.96273f,0.96632f,0.97048f,0.97594f,0.99029f,0.00000f,0.90439f,0.91753f,0.91826f,0.91917f,0.91829f,0.92008f,0.92277f,0.92510f,0.92529f,0.92710f,0.93214f,0.93327f,0.93313f,0.92927f,0.93521f,0.93562f,0.94166f,0.93511f,0.93726f,0.93711f,0.94052f,0.93952f,0.93943f,0.94490f,0.94746f,0.94877f,0.95242f,0.95677f,0.96039f,0.96533f,0.97088f,0.98669f,0.00000f,0.88477f,0.88625f,0.89627f,0.89640f,0.90330f,0.90815f,0.91480f,0.91320f,0.91847f,0.92255f,0.92345f,0.92283f,0.92446f,0.92584f,0.93369f,0.92896f,0.93346f,0.93103f,0.93582f,0.93408f,0.93891f,0.94008f,0.94108f,0.94113f,0.94216f,0.94351f,0.94600f,0.95215f,0.95560f,0.96110f,0.96588f,0.98229f,0.00000f,0.85464f,0.86848f,0.87916f,0.89055f,0.89364f,0.89648f,0.89557f,0.90103f,0.90645f,0.90824f,0.91268f,0.91726f,0.92159f,0.91948f,0.92568f,0.92781f,0.92656f,0.93138f,0.93174f,0.93525f,0.93262f,0.93451f,0.93782f,0.93911f,0.93679f,0.93826f,0.94155f,0.94392f,0.94745f,0.95440f,0.95988f,0.97808f,0.00000f,0.82478f,0.84457f,0.85118f,0.85362f,0.86811f,0.88005f,0.87862f,0.89224f,0.88821f,0.89439f,0.90275f,0.90225f,0.91482f,0.91277f,0.92239f,0.92023f,0.92616f,0.92660f,0.92800f,0.92935f,0.93091f,0.93343f,0.93450f,0.93757f,0.93712f,0.93303f,0.93634f,0.93943f,0.94425f,0.94842f,0.95448f,0.97296f,0.00000f,0.79586f,0.81475f,0.82832f,0.83589f,0.84068f,0.85666f,0.86141f,0.87247f,0.87651f,0.88228f,0.88694f,0.89491f,0.89648f,0.90462f,0.91274f,0.91349f,0.92014f,0.91978f,0.92115f,0.92396f,0.92351f,0.92743f,0.92907f,0.93057f,0.93433f,0.93199f,0.93086f,0.93529f,0.93927f,0.94192f,0.94776f,0.96814f,0.00000f,0.76292f,0.77817f,0.79857f,0.80660f,0.82163f,0.83061f,0.83530f,0.85078f,0.85953f,0.86446f,0.87554f,0.88162f,0.88575f,0.89220f,0.89662f,0.90742f,0.90458f,0.90998f,0.91364f,0.91257f,0.91921f,0.92555f,0.92776f,0.93007f,0.92626f,0.92888f,0.92467f,0.92926f,0.93205f,0.93607f,0.94148f,0.96338f,0.00000f,0.73998f,0.75225f,0.76558f,0.78180f,0.79361f,0.80583f,0.81644f,0.82724f,0.83685f,0.84568f,0.85177f,0.86295f,0.86984f,0.88122f,0.88624f,0.89019f,0.89418f,0.89573f,0.90243f,0.90773f,0.91493f,0.91713f,0.91462f,0.91998f,0.92536f,0.92276f,0.92200f,0.91994f,0.92332f,0.92852f,0.93631f,0.95789f,
 0.00000f,0.99643f,0.98632f,0.98102f,0.97353f,0.94877f,0.96338f,0.95599f,0.94797f,0.94266f,0.92833f,0.94272f,0.93821f,0.94313f,0.96353f,0.95915f,0.96137f,0.95842f,0.95820f,0.96434f,0.96687f,0.96826f,0.96783f,0.96771f,0.97259f,0.97457f,0.97711f,0.98319f,0.97310f,0.98906f,0.99288f,0.99624f,1.00000f,0.00000f,0.99143f,0.98523f,0.98017f,0.97716f,0.97455f,0.97213f,0.97139f,0.96978f,0.96736f,0.96749f,0.96790f,0.96759f,0.96779f,0.97076f,0.97140f,0.97258f,0.97311f,0.97304f,0.97468f,0.97574f,0.97704f,0.97886f,0.98005f,0.98230f,0.98369f,0.98589f,0.98813f,0.98960f,0.99238f,0.99506f,0.99755f,1.00000f,0.00000f,0.99001f,0.98544f,0.98278f,0.97949f,0.97694f,0.97644f,0.97556f,0.97412f,0.97248f,0.97398f,0.97254f,0.97402f,0.97351f,0.97375f,0.97440f,0.97467f,0.97521f,0.97672f,0.97775f,0.97858f,0.97939f,0.98083f,0.98305f,0.98384f,0.98602f,0.98753f,0.98925f,0.99126f,0.99336f,0.99542f,0.99774f,1.00000f,0.00000f,0.98611f,0.98483f,0.98277f,0.97910f,0.97751f,0.97561f,0.97449f,0.97398f,0.97381f,0.97442f,0.97273f,0.97356f,0.97288f,0.97399f,0.97383f,0.97557f,0.97578f,0.97683f,0.97779f,0.97935f,0.97956f,0.98134f,0.98267f,0.98419f,0.98611f,0.98766f,0.98940f,0.99149f,0.99340f,0.99551f,0.99777f,1.00000f,0.00000f,0.94322f,0.98175f,0.98128f,0.97870f,0.97782f,0.97576f,0.97366f,0.97435f,0.97329f,0.97450f,0.97252f,0.97340f,0.97402f,0.97391f,0.97423f,0.97458f,0.97578f,0.97688f,0.97760f,0.97868f,0.97970f,0.98131f,0.98240f,0.98419f,0.98586f,0.98753f,0.98946f,0.99136f,0.99332f,0.99543f,0.99772f,1.00000f,0.00000f,0.91202f,0.96056f,0.97819f,0.97843f,0.97665f,0.97422f,0.97512f,0.97379f,0.97336f,0.97307f,0.97325f,0.97321f,0.97213f,0.97351f,0.97425f,0.97424f,0.97562f,0.97655f,0.97743f,0.97895f,0.97991f,0.98130f,0.98246f,0.98442f,0.98554f,0.98768f,0.98928f,0.99144f,0.99331f,0.99552f,0.99776f,1.00000f,0.00000f,0.90519f,0.92890f,0.96167f,0.97366f,0.97487f,0.97488f,0.97325f,0.97284f,0.97337f,0.97247f,0.97238f,0.97290f,0.97299f,0.97323f,0.97397f,0.97476f,0.97524f,0.97686f,0.97714f,0.97862f,0.97983f,0.98079f,0.98225f,0.98433f,0.98593f,0.98754f,0.98930f,0.99133f,0.99343f,0.99542f,0.99764f,1.00000f,0.00000f,0.90610f,0.91521f,0.93727f,0.95620f,0.96761f,0.97099f,0.97298f,0.97377f,0.97334f,0.97230f,0.97205f,0.97302f,0.97242f,0.97381f,0.97405f,0.97412f,0.97512f,0.97709f,0.97703f,0.97803f,0.97951f,0.98103f,0.98219f,0.98373f,0.98597f,0.98719f,0.98926f,0.99113f,0.99331f,0.99528f,0.99752f,1.00000f,0.00000f,0.91523f,0.90658f,0.92240f,0.93663f,0.95436f,0.96433f,0.96848f,0.97008f,0.97205f,0.97134f,0.97086f,0.97283f,0.97293f,0.97308f,0.97321f,0.97426f,0.97500f,0.97552f,0.97696f,0.97763f,0.97935f,0.98085f,0.98243f,0.98369f,0.98556f,0.98713f,0.98922f,0.99107f,0.99332f,0.99523f,0.99756f,1.00000f,0.00000f,0.92484f,0.91061f,0.91546f,0.92746f,0.93828f,0.94907f,0.96090f,0.96377f,0.96783f,0.97124f,0.97057f,0.97281f,0.97206f,0.97299f,0.97297f,0.97373f,0.97494f,0.97578f,0.97682f,0.97844f,0.97837f,0.98085f,0.98258f,0.98364f,0.98545f,0.98702f,0.98905f,0.99080f,0.99296f,0.99524f,0.99760f,1.00000f,0.00000f,0.92700f,0.91696f,0.91802f,0.92034f,0.92868f,0.93893f,0.94771f,0.95494f,0.96073f,0.96514f,0.96741f,0.97054f,0.97082f,0.97204f,0.97196f,0.97293f,0.97427f,0.97489f,0.97624f,0.97766f,0.97888f,0.98090f,0.98170f,0.98332f,0.98531f,0.98720f,0.98889f,0.99066f,0.99292f,0.99506f,0.99737f,1.00000f,0.00000f,0.93685f,0.92260f,0.91085f,0.91637f,0.92164f,0.92658f,0.93525f,0.94401f,0.95069f,0.95658f,0.96078f,0.96552f,0.97016f,0.97007f,0.97123f,0.97157f,0.97318f,0.97497f,0.97600f,0.97704f,0.97874f,0.98043f,0.98244f,0.98304f,0.98531f,0.98640f,0.98870f,0.99076f,0.99264f,0.99502f,0.99725f,1.00000f,0.00000f,0.94556f,0.92274f,0.92112f,0.91233f,0.91620f,0.92597f,0.92944f,0.93432f,0.94368f,0.94658f,0.95341f,0.95764f,0.96346f,0.96706f,0.96840f,0.97232f,0.97243f,0.97412f,0.97625f,0.97600f,0.97836f,0.97977f,0.98147f,0.98291f,0.98490f,0.98682f,0.98882f,0.99060f,0.99275f,0.99499f,0.99733f,1.00000f,0.00000f,0.94515f,0.93354f,0.92490f,0.91146f,0.91214f,0.92344f,0.92656f,0.93237f,0.93564f,0.94102f,0.94435f,0.95284f,0.95769f,0.96001f,0.96263f,0.96782f,0.96846f,0.97247f,0.97529f,0.97556f,0.97722f,0.97939f,0.98089f,0.98292f,0.98393f,0.98709f,0.98834f,0.99042f,0.99231f,0.99493f,0.99716f,1.00000f,0.00000f,0.95568f,0.93736f,0.92471f,0.91801f,0.91998f,0.92452f,0.92136f,0.92923f,0.92991f,0.93289f,0.93851f,0.94694f,0.94884f,0.95457f,0.95841f,0.96262f,0.96701f,0.96924f,0.97110f,0.97378f,0.97699f,0.97813f,0.98094f,0.98230f,0.98408f,0.98599f,0.98789f,0.99025f,0.99251f,0.99488f,0.99740f,1.00000f,0.00000f,0.95819f,0.94333f,0.93253f,0.92660f,0.92476f,0.92301f,0.91904f,0.92494f,0.92764f,0.93053f,0.93324f,0.93743f,0.94382f,0.94853f,0.95284f,0.95598f,0.96160f,0.96381f,0.96738f,0.97082f,0.97324f,0.97669f,0.97956f,0.98132f,0.98358f,0.98625f,0.98842f,0.99036f,0.99225f,0.99442f,0.99711f,1.00000f,0.00000f,0.95172f,0.94972f,0.93335f,0.93125f,0.92306f,0.92318f,0.92460f,0.92306f,0.92387f,0.92359f,0.93127f,0.93767f,0.93671f,0.94495f,0.94721f,0.95097f,0.95640f,0.96044f,0.96454f,0.96778f,0.97014f,0.97433f,0.97689f,0.98020f,0.98142f,0.98380f,0.98662f,0.98945f,0.99177f,0.99475f,0.99689f,1.00000f,0.00000f,0.95869f,0.94872f,0.94399f,0.92845f,0.92937f,0.92600f,0.92538f,0.92621f,0.92576f,0.92693f,0.93170f,0.93619f,0.93690f,0.93732f,0.94054f,0.94552f,0.95174f,0.95413f,0.95927f,0.96402f,0.96697f,0.96978f,0.97375f,0.97697f,0.97955f,0.98267f,0.98514f,0.98869f,0.99141f,0.99375f,0.99689f,1.00000f,0.00000f,0.96055f,0.94767f,0.94456f,0.94178f,0.92996f,0.92921f,0.92729f,0.92961f,0.92421f,0.92726f,0.93079f,0.93232f,0.93581f,0.93561f,0.93858f,0.94099f,0.94670f,0.95081f,0.95341f,0.95930f,0.96117f,0.96533f,0.96992f,0.97400f,0.97863f,0.98182f,0.98393f,0.98734f,0.98951f,0.99360f,0.99692f,1.00000f,0.00000f,0.96581f,0.96078f,0.95170f,0.94718f,0.93992f,0.93332f,0.92739f,0.93116f,0.93109f,0.93242f,0.93190f,0.92923f,0.93142f,0.93826f,0.93580f,0.94134f,0.94160f,0.94561f,0.95151f,0.95435f,0.95838f,0.96144f,0.96428f,0.97000f,0.97330f,0.97706f,0.98055f,0.98430f,0.98830f,0.99161f,0.99557f,0.99997f,0.00000f,0.96418f,0.95418f,0.95527f,0.94149f,0.94529f,0.93641f,0.93258f,0.92918f,0.92925f,0.92641f,0.92988f,0.93082f,0.93287f,0.93532f,0.93681f,0.93803f,0.94121f,0.94143f,0.94438f,0.94849f,0.95280f,0.95856f,0.96124f,0.96677f,0.97040f,0.97385f,0.97799f,0.98198f,0.98590f,0.98996f,0.99390f,0.99978f,0.00000f,0.96249f,0.95918f,0.95270f,0.94368f,0.94412f,0.93835f,0.93279f,0.93434f,0.92733f,0.93182f,0.92965f,0.92813f,0.93174f,0.93330f,0.93394f,0.93883f,0.93956f,0.94091f,0.94336f,0.94534f,0.94833f,0.95235f,0.95588f,0.96233f,0.96735f,0.97090f,0.97496f,0.97868f,0.98231f,0.98665f,0.99144f,0.99924f,0.00000f,0.95539f,0.95790f,0.95679f,0.94881f,0.94943f,0.94385f,0.93661f,0.93402f,0.93449f,0.92684f,0.93413f,0.93418f,0.93207f,0.93320f,0.93249f,0.93650f,0.93885f,0.93662f,0.94029f,0.94105f,0.94381f,0.95013f,0.95388f,0.95580f,0.96161f,0.96476f,0.97093f,0.97554f,0.97813f,0.98458f,0.98761f,0.99778f,0.00000f,0.95357f,0.95121f,0.94923f,0.94906f,0.94601f,0.94465f,0.94131f,0.93480f,0.93911f,0.93630f,0.93144f,0.93326f,0.93430f,0.93101f,0.93295f,0.93733f,0.93778f,0.93958f,0.94005f,0.93892f,0.94376f,0.94516f,0.94842f,0.95229f,0.95563f,0.96196f,0.96534f,0.97034f,0.97490f,0.98025f,0.98440f,0.99550f,0.00000f,0.94454f,0.94298f,0.94308f,0.94355f,0.94317f,0.94118f,0.93984f,0.93845f,0.93305f,0.93250f,0.93674f,0.93789f,0.93062f,0.93247f,0.92945f,0.93516f,0.93729f,0.93822f,0.93789f,0.94001f,0.94245f,0.94400f,0.94425f,0.94754f,0.95090f,0.95501f,0.96032f,0.96471f,0.96977f,0.97372f,0.97916f,0.99288f,0.00000f,0.92671f,0.93744f,0.93480f,0.93857f,0.93704f,0.93134f,0.93195f,0.93443f,0.93265f,0.93700f,0.93275f,0.93543f,0.93786f,0.93353f,0.93167f,0.93580f,0.93738f,0.93408f,0.93833f,0.93801f,0.94129f,0.94072f,0.94275f,0.94518f,0.94696f,0.95276f,0.95601f,0.95753f,0.96384f,0.96989f,0.97399f,0.98912f,0.00000f,0.90428f,0.91216f,0.91698f,0.92504f,0.92274f,0.92074f,0.92578f,0.92913f,0.92725f,0.92921f,0.93137f,0.93021f,0.93081f,0.93414f,0.93676f,0.93309f,0.93482f,0.93321f,0.93587f,0.94232f,0.93839f,0.94065f,0.93633f,0.94228f,0.94404f,0.94609f,0.94837f,0.95237f,0.95861f,0.96376f,0.96887f,0.98488f,0.00000f,0.89275f,0.89887f,0.89875f,0.90541f,0.91548f,0.92266f,0.91479f,0.91681f,0.92709f,0.91817f,0.92287f,0.92361f,0.93041f,0.93038f,0.92977f,0.93037f,0.92864f,0.93391f,0.93162f,0.93990f,0.93523f,0.93673f,0.93841f,0.94193f,0.93965f,0.94501f,0.94542f,0.94866f,0.95290f,0.95622f,0.96204f,0.97955f,0.00000f,0.86427f,0.87610f,0.88698f,0.88842f,0.89657f,0.89814f,0.90347f,0.90796f,0.91722f,0.91319f,0.91473f,0.91839f,0.91807f,0.92511f,0.92359f,0.92564f,0.92778f,0.92965f,0.93288f,0.93437f,0.93208f,0.93282f,0.93522f,0.93684f,0.93789f,0.93675f,0.93771f,0.94134f,0.94445f,0.94977f,0.95654f,0.97490f,0.00000f,0.83859f,0.85174f,0.85659f,0.87038f,0.87582f,0.89037f,0.88387f,0.89537f,0.89939f,0.89848f,0.90518f,0.90524f,0.91290f,0.91614f,0.91429f,0.91798f,0.91944f,0.92183f,0.92066f,0.92770f,0.93472f,0.93165f,0.93033f,0.93260f,0.93132f,0.92838f,0.93178f,0.93759f,0.93859f,0.94510f,0.94915f,0.96996f,0.00000f,0.80897f,0.82528f,0.84010f,0.84896f,0.85255f,0.86160f,0.86761f,0.87751f,0.88192f,0.88928f,0.89388f,0.89631f,0.90477f,0.90661f,0.90702f,0.91033f,0.91458f,0.91535f,0.92323f,0.92290f,0.92069f,0.92579f,0.92739f,0.92845f,0.92687f,0.92505f,0.92943f,0.92980f,0.93405f,0.93726f,0.94289f,0.96462f,0.00000f,0.78093f,0.79053f,0.80451f,0.82218f,0.83042f,0.84173f,0.85132f,0.85310f,0.86405f,0.87057f,0.87379f,0.88749f,0.88315f,0.89593f,0.89913f,0.90395f,0.90234f,0.90896f,0.91405f,0.91732f,0.91637f,0.92175f,0.92367f,0.92329f,0.92200f,0.92134f,0.92459f,0.92534f,0.92711f,0.93153f,0.93559f,0.95966f,0.00000f,0.74796f,0.76070f,0.78071f,0.79383f,0.80675f,0.81004f,0.82343f,0.83013f,0.84173f,0.84885f,0.85795f,0.86149f,0.87023f,0.87976f,0.88026f,0.89211f,0.89875f,0.90193f,0.90449f,0.90102f,0.91423f,0.91013f,0.91522f,0.91399f,0.91647f,0.91472f,0.91706f,0.92031f,0.91998f,0.92230f,0.93001f,0.95306f,
 0.00000f,0.99922f,0.98866f,0.98095f,0.97273f,0.95878f,0.96494f,0.95659f,0.95929f,0.94532f,0.95418f,0.94884f,0.94919f,0.95170f,0.95228f,0.93927f,0.95528f,0.96325f,0.96860f,0.96986f,0.96601f,0.97262f,0.96414f,0.97534f,0.96739f,0.98062f,0.98024f,0.98115f,0.98481f,0.98539f,0.98468f,0.99585f,1.00000f,0.00000f,0.99173f,0.98583f,0.98036f,0.97811f,0.97571f,0.97245f,0.97186f,0.97082f,0.96969f,0.96922f,0.96913f,0.96875f,0.96835f,0.96920f,0.96950f,0.96990f,0.97327f,0.97465f,0.97569f,0.97675f,0.97793f,0.97938f,0.98074f,0.98202f,0.98432f,0.98612f,0.98811f,0.99076f,0.99243f,0.99433f,0.99747f,1.00000f,0.00000f,0.99149f,0.98574f,0.98218f,0.98023f,0.97712f,0.97664f,0.97428f,0.97478f,0.97276f,0.97332f,0.97299f,0.97215f,0.97345f,0.97332f,0.97400f,0.97535f,0.97536f,0.97675f,0.97817f,0.97851f,0.98007f,0.98129f,0.98296f,0.98415f,0.98577f,0.98752f,0.98961f,0.99133f,0.99318f,0.99547f,0.99778f,1.00000f,0.00000f,0.98196f,0.98479f,0.98211f,0.97989f,0.97866f,0.97622f,0.97505f,0.97446f,0.97380f,0.97317f,0.97289f,0.97328f,0.97362f,0.97410f,0.97404f,0.97521f,0.97567f,0.97644f,0.97785f,0.97834f,0.97973f,0.98155f,0.98267f,0.98440f,0.98580f,0.98771f,0.98929f,0.99123f,0.99345f,0.99545f,0.99778f,1.00000f,0.00000f,0.94150f,0.98192f,0.98143f,0.97989f,0.97761f,0.97576f,0.97551f,0.97410f,0.97444f,0.97329f,0.97216f,0.97296f,0.97312f,0.97349f,0.97405f,0.97489f,0.97584f,0.97685f,0.97751f,0.97891f,0.98017f,0.98160f,0.98276f,0.98448f,0.98599f,0.98766f,0.98930f,0.99121f,0.99331f,0.99552f,0.99775f,1.00000f,0.00000f,0.91383f,0.96057f,0.97725f,0.97761f,0.97708f,0.97530f,0.97405f,0.97368f,0.97294f,0.97297f,0.97274f,0.97323f,0.97317f,0.97357f,0.97396f,0.97476f,0.97556f,0.97594f,0.97732f,0.97919f,0.97958f,0.98122f,0.98299f,0.98411f,0.98560f,0.98734f,0.98930f,0.99119f,0.99330f,0.99543f,0.99778f,1.00000f,0.00000f,0.89890f,0.93033f,0.96059f,0.97277f,0.97635f,0.97416f,0.97272f,0.97280f,0.97185f,0.97355f,0.97290f,0.97247f,0.97262f,0.97314f,0.97425f,0.97458f,0.97559f,0.97647f,0.97774f,0.97880f,0.97974f,0.98115f,0.98276f,0.98404f,0.98587f,0.98715f,0.98919f,0.99115f,0.99334f,0.99544f,0.99761f,1.00000f,0.00000f,0.90604f,0.91415f,0.93818f,0.95763f,0.96839f,0.97214f,0.97375f,0.97090f,0.97342f,0.97178f,0.97139f,0.97291f,0.97264f,0.97217f,0.97373f,0.97439f,0.97569f,0.97630f,0.97689f,0.97873f,0.97984f,0.98107f,0.98256f,0.98381f,0.98527f,0.98737f,0.98922f,0.99110f,0.99312f,0.99525f,0.99756f,1.00000f,0.00000f,0.91900f,0.90868f,0.92086f,0.93702f,0.95305f,0.96111f,0.96818f,0.97021f,0.97262f,0.97101f,0.97227f,0.97269f,0.97320f,0.97237f,0.97370f,0.97416f,0.97522f,0.97651f,0.97610f,0.97832f,0.97923f,0.98087f,0.98239f,0.98363f,0.98530f,0.98728f,0.98929f,0.99075f,0.99313f,0.99525f,0.99754f,1.00000f,0.00000f,0.92572f,0.91107f,0.91575f,0.92728f,0.93642f,0.95011f,0.96029f,0.96592f,0.96726f,0.96925f,0.97120f,0.97183f,0.97275f,0.97244f,0.97387f,0.97420f,0.97460f,0.97636f,0.97718f,0.97847f,0.97981f,0.98094f,0.98232f,0.98339f,0.98553f,0.98720f,0.98870f,0.99101f,0.99294f,0.99529f,0.99743f,1.00000f,0.00000f,0.93717f,0.91383f,0.91357f,0.92091f,0.92869f,0.93762f,0.94924f,0.95258f,0.96196f,0.96578f,0.96753f,0.97059f,0.97118f,0.97133f,0.97228f,0.97307f,0.97488f,0.97588f,0.97656f,0.97769f,0.97916f,0.98011f,0.98152f,0.98348f,0.98501f,0.98701f,0.98905f,0.99084f,0.99311f,0.99517f,0.99751f,1.00000f,0.00000f,0.93818f,0.91767f,0.90934f,0.91310f,0.92623f,0.92594f,0.93624f,0.94502f,0.95122f,0.95745f,0.96257f,0.96490f,0.96783f,0.97032f,0.97137f,0.97368f,0.97440f,0.97566f,0.97583f,0.97803f,0.97903f,0.98033f,0.98206f,0.98374f,0.98508f,0.98691f,0.98908f,0.99065f,0.99301f,0.99509f,0.99746f,1.00000f,0.00000f,0.94645f,0.91916f,0.91458f,0.91575f,0.91571f,0.92454f,0.92885f,0.93572f,0.94107f,0.94723f,0.95460f,0.95825f,0.96444f,0.96648f,0.96881f,0.97012f,0.97219f,0.97423f,0.97634f,0.97694f,0.97839f,0.98049f,0.98135f,0.98273f,0.98527f,0.98653f,0.98895f,0.99027f,0.99286f,0.99492f,0.99730f,1.00000f,0.00000f,0.94991f,0.93322f,0.92519f,0.91711f,0.91781f,0.91938f,0.92587f,0.92982f,0.93458f,0.94033f,0.94583f,0.95264f,0.95590f,0.96088f,0.96524f,0.96712f,0.96974f,0.97213f,0.97444f,0.97596f,0.97852f,0.97962f,0.98124f,0.98300f,0.98460f,0.98675f,0.98819f,0.99072f,0.99261f,0.99482f,0.99701f,1.00000f,0.00000f,0.95438f,0.93728f,0.92466f,0.91999f,0.91636f,0.92005f,0.92451f,0.92760f,0.92954f,0.93378f,0.93646f,0.94315f,0.94966f,0.95216f,0.95884f,0.96422f,0.96641f,0.96882f,0.97007f,0.97338f,0.97675f,0.97813f,0.98081f,0.98284f,0.98433f,0.98593f,0.98814f,0.99006f,0.99280f,0.99502f,0.99708f,1.00000f,0.00000f,0.94896f,0.94020f,0.93290f,0.92048f,0.91899f,0.92257f,0.92031f,0.92574f,0.92833f,0.92945f,0.93513f,0.93777f,0.94295f,0.94856f,0.95254f,0.95651f,0.96133f,0.96392f,0.96703f,0.97058f,0.97392f,0.97600f,0.97902f,0.98096f,0.98366f,0.98595f,0.98795f,0.98964f,0.99246f,0.99469f,0.99736f,1.00000f,0.00000f,0.95357f,0.94783f,0.93481f,0.92814f,0.92443f,0.92252f,0.92729f,0.92287f,0.92533f,0.92927f,0.93151f,0.93488f,0.93965f,0.94465f,0.94480f,0.95236f,0.95601f,0.95816f,0.96459f,0.96882f,0.96971f,0.97375f,0.97710f,0.97875f,0.98212f,0.98490f,0.98724f,0.98920f,0.99226f,0.99438f,0.99692f,1.00000f,0.00000f,0.95539f,0.95149f,0.94557f,0.93685f,0.92725f,0.92686f,0.92629f,0.92652f,0.92811f,0.92575f,0.93012f,0.93467f,0.93683f,0.93882f,0.94220f,0.94695f,0.95257f,0.95260f,0.95686f,0.96361f,0.96711f,0.96943f,0.97464f,0.97727f,0.98034f,0.98308f,0.98527f,0.98863f,0.99102f,0.99395f,0.99666f,1.00000f,0.00000f,0.95517f,0.95378f,0.94071f,0.93860f,0.93060f,0.92776f,0.92638f,0.92901f,0.92495f,0.92814f,0.92958f,0.93035f,0.93394f,0.93621f,0.94110f,0.94266f,0.94470f,0.94883f,0.95538f,0.95813f,0.95969f,0.96552f,0.96962f,0.97334f,0.97659f,0.98072f,0.98323f,0.98658f,0.99013f,0.99348f,0.99602f,1.00000f,0.00000f,0.96030f,0.95485f,0.94872f,0.94054f,0.94129f,0.93085f,0.92781f,0.92611f,0.93202f,0.92822f,0.92736f,0.92727f,0.93292f,0.93685f,0.93422f,0.93807f,0.93935f,0.94473f,0.94889f,0.95122f,0.95892f,0.96260f,0.96520f,0.97125f,0.97305f,0.97697f,0.98145f,0.98328f,0.98845f,0.99196f,0.99546f,0.99997f,0.00000f,0.95720f,0.95429f,0.95371f,0.94960f,0.94394f,0.94020f,0.93622f,0.92929f,0.93213f,0.92783f,0.93060f,0.92790f,0.93129f,0.93658f,0.93509f,0.93773f,0.93906f,0.94290f,0.94403f,0.94961f,0.95329f,0.95835f,0.96089f,0.96466f,0.96935f,0.97419f,0.97829f,0.98081f,0.98440f,0.99032f,0.99348f,0.99975f,0.00000f,0.96283f,0.95679f,0.95646f,0.94929f,0.94625f,0.93896f,0.93486f,0.92936f,0.93192f,0.92934f,0.93165f,0.92777f,0.93272f,0.93155f,0.93370f,0.93358f,0.93911f,0.94443f,0.94084f,0.94610f,0.94908f,0.95201f,0.95518f,0.96047f,0.96279f,0.96980f,0.97392f,0.97772f,0.98259f,0.98719f,0.99139f,0.99913f,0.00000f,0.95303f,0.96171f,0.95553f,0.95013f,0.95130f,0.94082f,0.93859f,0.93724f,0.93704f,0.92752f,0.93111f,0.92937f,0.93413f,0.92948f,0.93584f,0.93522f,0.93580f,0.93644f,0.93657f,0.94359f,0.94280f,0.94897f,0.95103f,0.95613f,0.95935f,0.96300f,0.96986f,0.97282f,0.97789f,0.98242f,0.98744f,0.99754f,0.00000f,0.94680f,0.95225f,0.95467f,0.94448f,0.94223f,0.94474f,0.93824f,0.93924f,0.93515f,0.93554f,0.93226f,0.93501f,0.93339f,0.93469f,0.93626f,0.93555f,0.93679f,0.93910f,0.93961f,0.94141f,0.94345f,0.94617f,0.94625f,0.95166f,0.95568f,0.95906f,0.96458f,0.97005f,0.97417f,0.97743f,0.98372f,0.99525f,0.00000f,0.94425f,0.94894f,0.94744f,0.94567f,0.94470f,0.94036f,0.94044f,0.93138f,0.93407f,0.93532f,0.93376f,0.93551f,0.92910f,0.93208f,0.93599f,0.93371f,0.93397f,0.93375f,0.93525f,0.93936f,0.93803f,0.93936f,0.94183f,0.94582f,0.95121f,0.95440f,0.95782f,0.96291f,0.96723f,0.97153f,0.97793f,0.99215f,0.00000f,0.92851f,0.93535f,0.93858f,0.93881f,0.94181f,0.93484f,0.93819f,0.93551f,0.93399f,0.93311f,0.93110f,0.93350f,0.93189f,0.93395f,0.93267f,0.93700f,0.93173f,0.93517f,0.93706f,0.93500f,0.94216f,0.94101f,0.94251f,0.94492f,0.94528f,0.94740f,0.95183f,0.95721f,0.96216f,0.96778f,0.97218f,0.98809f,0.00000f,0.91662f,0.91519f,0.93036f,0.92916f,0.92930f,0.93240f,0.92749f,0.93824f,0.92886f,0.93173f,0.93236f,0.93156f,0.92952f,0.93238f,0.93573f,0.93284f,0.93032f,0.93441f,0.93410f,0.93508f,0.93468f,0.93599f,0.93925f,0.93736f,0.94165f,0.94376f,0.94831f,0.95112f,0.95502f,0.95991f,0.96513f,0.98335f,0.00000f,0.89894f,0.90526f,0.91623f,0.91820f,0.91504f,0.92340f,0.92492f,0.92564f,0.92182f,0.92303f,0.91928f,0.92744f,0.92453f,0.92757f,0.92837f,0.92996f,0.92908f,0.93131f,0.93210f,0.93426f,0.93507f,0.93898f,0.93525f,0.93500f,0.93954f,0.93918f,0.94238f,0.94533f,0.94831f,0.95530f,0.95896f,0.97793f,0.00000f,0.88293f,0.87421f,0.89444f,0.89100f,0.90759f,0.90438f,0.91409f,0.90700f,0.90810f,0.91455f,0.91496f,0.92525f,0.92575f,0.92286f,0.92488f,0.92556f,0.92947f,0.92907f,0.92721f,0.93268f,0.93390f,0.93241f,0.93251f,0.93143f,0.93291f,0.93408f,0.93286f,0.93530f,0.94314f,0.94690f,0.95333f,0.97206f,0.00000f,0.84677f,0.85691f,0.86926f,0.88077f,0.88070f,0.89040f,0.89224f,0.90422f,0.90807f,0.90335f,0.90803f,0.90689f,0.91032f,0.91649f,0.91805f,0.91984f,0.91975f,0.92194f,0.92436f,0.92928f,0.92570f,0.92614f,0.92606f,0.93055f,0.92843f,0.92918f,0.93148f,0.93159f,0.93517f,0.93989f,0.94506f,0.96663f,0.00000f,0.82217f,0.83304f,0.84554f,0.85195f,0.85971f,0.87296f,0.87380f,0.88635f,0.88465f,0.88971f,0.89966f,0.89290f,0.90025f,0.90663f,0.90802f,0.90943f,0.91142f,0.91869f,0.91680f,0.92281f,0.92490f,0.91969f,0.92427f,0.92555f,0.92583f,0.92415f,0.92560f,0.92756f,0.92887f,0.93247f,0.93766f,0.96008f,0.00000f,0.79337f,0.80653f,0.81821f,0.82316f,0.83432f,0.84321f,0.85846f,0.85496f,0.86670f,0.87294f,0.88273f,0.88077f,0.88568f,0.89140f,0.89615f,0.90814f,0.90107f,0.90703f,0.90814f,0.91231f,0.91081f,0.91569f,0.91848f,0.91999f,0.91640f,0.91757f,0.91791f,0.91778f,0.92143f,0.92563f,0.93067f,0.95462f,0.00000f,0.76335f,0.77913f,0.78695f,0.79857f,0.80886f,0.82270f,0.83557f,0.83962f,0.84587f,0.85096f,0.86209f,0.86337f,0.87212f,0.87923f,0.88235f,0.88837f,0.89640f,0.89662f,0.89905f,0.89876f,0.90863f,0.90551f,0.91170f,0.91282f,0.90990f,0.91234f,0.91233f,0.91363f,0.91420f,0.92014f,0.92494f,0.94862f,
 0.00000f,0.99577f,0.98209f,0.98211f,0.97313f,0.95844f,0.96469f,0.96605f,0.95472f,0.95532f,0.95808f,0.95159f,0.95170f,0.95382f,0.95301f,0.95277f,0.95364f,0.95579f,0.94524f,0.95008f,0.95895f,0.95326f,0.97579f,0.96786f,0.97953f,0.98084f,0.98276f,0.98147f,0.98143f,0.98899f,0.99454f,0.99667f,1.00000f,0.00000f,0.98949f,0.98183f,0.98200f,0.97760f,0.97468f,0.97518f,0.97235f,0.97179f,0.97234f,0.96929f,0.97019f,0.96967f,0.96941f,0.96995f,0.97032f,0.97130f,0.97117f,0.97228f,0.97262f,0.97517f,0.97614f,0.98027f,0.98057f,0.98309f,0.98458f,0.98672f,0.98854f,0.99025f,0.99264f,0.99519f,0.99751f,1.00000f,0.00000f,0.98963f,0.98473f,0.98279f,0.98038f,0.97757f,0.97607f,0.97555f,0.97405f,0.97417f,0.97367f,0.97332f,0.97339f,0.97392f,0.97386f,0.97368f,0.97492f,0.97633f,0.97627f,0.97769f,0.97836f,0.98077f,0.98121f,0.98309f,0.98431f,0.98583f,0.98762f,0.98950f,0.99144f,0.99349f,0.99546f,0.99775f,1.00000f,0.00000f,0.98220f,0.98585f,0.98227f,0.98009f,0.97760f,0.97586f,0.97405f,0.97347f,0.97292f,0.97372f,0.97373f,0.97311f,0.97334f,0.97405f,0.97499f,0.97514f,0.97577f,0.97691f,0.97811f,0.97904f,0.98002f,0.98145f,0.98260f,0.98470f,0.98605f,0.98778f,0.98931f,0.99114f,0.99329f,0.99543f,0.99776f,1.00000f,0.00000f,0.94405f,0.98263f,0.98053f,0.97920f,0.97838f,0.97460f,0.97519f,0.97449f,0.97298f,0.97378f,0.97340f,0.97330f,0.97324f,0.97331f,0.97369f,0.97536f,0.97647f,0.97639f,0.97749f,0.97913f,0.97933f,0.98116f,0.98307f,0.98461f,0.98601f,0.98784f,0.98943f,0.99129f,0.99327f,0.99554f,0.99776f,1.00000f,0.00000f,0.91346f,0.95987f,0.97754f,0.97779f,0.97747f,0.97581f,0.97517f,0.97331f,0.97523f,0.97257f,0.97294f,0.97287f,0.97427f,0.97466f,0.97496f,0.97437f,0.97561f,0.97645f,0.97755f,0.97867f,0.98010f,0.98142f,0.98263f,0.98445f,0.98556f,0.98748f,0.98950f,0.99117f,0.99331f,0.99550f,0.99778f,1.00000f,0.00000f,0.90602f,0.93039f,0.95937f,0.97216f,0.97573f,0.97407f,0.97384f,0.97456f,0.97261f,0.97325f,0.97280f,0.97322f,0.97322f,0.97287f,0.97445f,0.97453f,0.97590f,0.97652f,0.97830f,0.97782f,0.97973f,0.98156f,0.98243f,0.98418f,0.98589f,0.98739f,0.98947f,0.99094f,0.99327f,0.99540f,0.99763f,1.00000f,0.00000f,0.90971f,0.91843f,0.93465f,0.95919f,0.96838f,0.97217f,0.97381f,0.97306f,0.97278f,0.97304f,0.97270f,0.97240f,0.97286f,0.97302f,0.97322f,0.97423f,0.97454f,0.97679f,0.97690f,0.97796f,0.97983f,0.98087f,0.98227f,0.98354f,0.98572f,0.98755f,0.98918f,0.99125f,0.99304f,0.99539f,0.99751f,1.00000f,0.00000f,0.91843f,0.91186f,0.91885f,0.93856f,0.95288f,0.96402f,0.96993f,0.97115f,0.97126f,0.97115f,0.97157f,0.97315f,0.97302f,0.97271f,0.97344f,0.97420f,0.97617f,0.97593f,0.97665f,0.97791f,0.97969f,0.98092f,0.98219f,0.98413f,0.98549f,0.98700f,0.98885f,0.99091f,0.99316f,0.99534f,0.99756f,1.00000f,0.00000f,0.91933f,0.90934f,0.91657f,0.92491f,0.93758f,0.95032f,0.95955f,0.96435f,0.96681f,0.97079f,0.97117f,0.97205f,0.97267f,0.97215f,0.97312f,0.97387f,0.97422f,0.97651f,0.97694f,0.97730f,0.97886f,0.98098f,0.98210f,0.98362f,0.98514f,0.98720f,0.98893f,0.99118f,0.99318f,0.99533f,0.99744f,1.00000f,0.00000f,0.94058f,0.91219f,0.91566f,0.91607f,0.92697f,0.93864f,0.94651f,0.95452f,0.96160f,0.96353f,0.96884f,0.96991f,0.97096f,0.97224f,0.97258f,0.97295f,0.97354f,0.97526f,0.97627f,0.97759f,0.97895f,0.98036f,0.98166f,0.98306f,0.98530f,0.98748f,0.98883f,0.99079f,0.99291f,0.99497f,0.99744f,1.00000f,0.00000f,0.94636f,0.91954f,0.91206f,0.91487f,0.92007f,0.92832f,0.93605f,0.94474f,0.95333f,0.95750f,0.96160f,0.96612f,0.96833f,0.97140f,0.97059f,0.97312f,0.97322f,0.97585f,0.97636f,0.97750f,0.97865f,0.97980f,0.98110f,0.98311f,0.98510f,0.98724f,0.98866f,0.99121f,0.99272f,0.99512f,0.99739f,1.00000f,0.00000f,0.94646f,0.92089f,0.91917f,0.91341f,0.91979f,0.92315f,0.92821f,0.93632f,0.94213f,0.94947f,0.95562f,0.95873f,0.96108f,0.96534f,0.96864f,0.97132f,0.97251f,0.97455f,0.97512f,0.97682f,0.97859f,0.97999f,0.98182f,0.98349f,0.98469f,0.98725f,0.98861f,0.99064f,0.99248f,0.99511f,0.99723f,1.00000f,0.00000f,0.95462f,0.92873f,0.91519f,0.91307f,0.92056f,0.91906f,0.92585f,0.93007f,0.93486f,0.93728f,0.94697f,0.95117f,0.95740f,0.96128f,0.96466f,0.96725f,0.97020f,0.97117f,0.97319f,0.97684f,0.97756f,0.97972f,0.98118f,0.98286f,0.98413f,0.98651f,0.98872f,0.99051f,0.99230f,0.99472f,0.99707f,1.00000f,0.00000f,0.95883f,0.93583f,0.92270f,0.92506f,0.91535f,0.91761f,0.92331f,0.92566f,0.92922f,0.93311f,0.93917f,0.94370f,0.94954f,0.95328f,0.96034f,0.96219f,0.96530f,0.96815f,0.97102f,0.97320f,0.97612f,0.97719f,0.98031f,0.98249f,0.98467f,0.98561f,0.98888f,0.98993f,0.99270f,0.99474f,0.99718f,1.00000f,0.00000f,0.95786f,0.94514f,0.93266f,0.92088f,0.92042f,0.91781f,0.92445f,0.92548f,0.92623f,0.92815f,0.93321f,0.93666f,0.94436f,0.94789f,0.95275f,0.95768f,0.96108f,0.96345f,0.96899f,0.97032f,0.97287f,0.97614f,0.97871f,0.98202f,0.98285f,0.98542f,0.98845f,0.98996f,0.99232f,0.99489f,0.99710f,1.00000f,0.00000f,0.96232f,0.94382f,0.93918f,0.92611f,0.92573f,0.91951f,0.91830f,0.92465f,0.92712f,0.92623f,0.93110f,0.93393f,0.93598f,0.94013f,0.94491f,0.95173f,0.95602f,0.95976f,0.96162f,0.96630f,0.97022f,0.97259f,0.97684f,0.97809f,0.98220f,0.98431f,0.98687f,0.98936f,0.99216f,0.99445f,0.99680f,1.00000f,0.00000f,0.95661f,0.94736f,0.94167f,0.93571f,0.93007f,0.92555f,0.92262f,0.92076f,0.92531f,0.92567f,0.92950f,0.93326f,0.93550f,0.93875f,0.94112f,0.94754f,0.94940f,0.95439f,0.96023f,0.95980f,0.96511f,0.96964f,0.97310f,0.97670f,0.98031f,0.98273f,0.98666f,0.98826f,0.99121f,0.99431f,0.99681f,1.00000f,0.00000f,0.96240f,0.95530f,0.95178f,0.93883f,0.93199f,0.92440f,0.92658f,0.92415f,0.92168f,0.92711f,0.92815f,0.93123f,0.93517f,0.93841f,0.93853f,0.93963f,0.94488f,0.94925f,0.95149f,0.95492f,0.95922f,0.96569f,0.96760f,0.97282f,0.97788f,0.98037f,0.98439f,0.98655f,0.98894f,0.99288f,0.99653f,1.00000f,0.00000f,0.96233f,0.95106f,0.95071f,0.94725f,0.93678f,0.92943f,0.93094f,0.92583f,0.92942f,0.92675f,0.93481f,0.93328f,0.93354f,0.93548f,0.93708f,0.93635f,0.94109f,0.94505f,0.94670f,0.95080f,0.95879f,0.96024f,0.96374f,0.97056f,0.97291f,0.97746f,0.98008f,0.98445f,0.98701f,0.99046f,0.99538f,0.99995f,0.00000f,0.96292f,0.95556f,0.95317f,0.94503f,0.93660f,0.93808f,0.93314f,0.93327f,0.92953f,0.92951f,0.92878f,0.92807f,0.93075f,0.93610f,0.93556f,0.93600f,0.94003f,0.93807f,0.94590f,0.94431f,0.95305f,0.95592f,0.95939f,0.96568f,0.96706f,0.97473f,0.97711f,0.98164f,0.98503f,0.98890f,0.99296f,0.99972f,0.00000f,0.95598f,0.96591f,0.95571f,0.94962f,0.94192f,0.94008f,0.93562f,0.93029f,0.92900f,0.93252f,0.93015f,0.93058f,0.93246f,0.93123f,0.93321f,0.93312f,0.93350f,0.93875f,0.93775f,0.94193f,0.94708f,0.95164f,0.95730f,0.95939f,0.96277f,0.96755f,0.97268f,0.97624f,0.98053f,0.98690f,0.99020f,0.99898f,0.00000f,0.96292f,0.95292f,0.95644f,0.94331f,0.94750f,0.94619f,0.93387f,0.93567f,0.93343f,0.93172f,0.92805f,0.93637f,0.93239f,0.92802f,0.93590f,0.93330f,0.93631f,0.93715f,0.93937f,0.94299f,0.94317f,0.94402f,0.94951f,0.95406f,0.95812f,0.96437f,0.96608f,0.97258f,0.97705f,0.98192f,0.98606f,0.99757f,0.00000f,0.95434f,0.96141f,0.95200f,0.94725f,0.95027f,0.94550f,0.93444f,0.93932f,0.93709f,0.93600f,0.93088f,0.93455f,0.93161f,0.92967f,0.93891f,0.93296f,0.93494f,0.93686f,0.93645f,0.93965f,0.94078f,0.94080f,0.94700f,0.94933f,0.95442f,0.96139f,0.96380f,0.96662f,0.97365f,0.97550f,0.98152f,0.99442f,0.00000f,0.94579f,0.95344f,0.95095f,0.94197f,0.94766f,0.94760f,0.94165f,0.93691f,0.93117f,0.93023f,0.93201f,0.93194f,0.92781f,0.93136f,0.93621f,0.93168f,0.93324f,0.93378f,0.93498f,0.93662f,0.93545f,0.93912f,0.94398f,0.94463f,0.94767f,0.95385f,0.95770f,0.96131f,0.96624f,0.97056f,0.97589f,0.99045f,0.00000f,0.93704f,0.94098f,0.93373f,0.94033f,0.94445f,0.93776f,0.93686f,0.93641f,0.93035f,0.92985f,0.93515f,0.92681f,0.93123f,0.92969f,0.93240f,0.93358f,0.93315f,0.93456f,0.93706f,0.93648f,0.93675f,0.93869f,0.94031f,0.94412f,0.94321f,0.94623f,0.95237f,0.95486f,0.95933f,0.96593f,0.96894f,0.98672f,0.00000f,0.92347f,0.92217f,0.92736f,0.93973f,0.92684f,0.93103f,0.92779f,0.93196f,0.92959f,0.93033f,0.93164f,0.93027f,0.92942f,0.92900f,0.93041f,0.92965f,0.93076f,0.93100f,0.93426f,0.93336f,0.93974f,0.93796f,0.93748f,0.93788f,0.93940f,0.94256f,0.94418f,0.94886f,0.95469f,0.95903f,0.96338f,0.98154f,0.00000f,0.90468f,0.90616f,0.91925f,0.91747f,0.91627f,0.92919f,0.92220f,0.92358f,0.92027f,0.92334f,0.92809f,0.92391f,0.92799f,0.92830f,0.92818f,0.93108f,0.93200f,0.93052f,0.92772f,0.93038f,0.93373f,0.93278f,0.93274f,0.93427f,0.93702f,0.93297f,0.93785f,0.94166f,0.94412f,0.95116f,0.95523f,0.97554f,0.00000f,0.88376f,0.89380f,0.89304f,0.89976f,0.90591f,0.90950f,0.90668f,0.91411f,0.91428f,0.91363f,0.92065f,0.92070f,0.91935f,0.92750f,0.92676f,0.92894f,0.92480f,0.92737f,0.93198f,0.92909f,0.92860f,0.93164f,0.93383f,0.93112f,0.92936f,0.93077f,0.93273f,0.93350f,0.93942f,0.94389f,0.94703f,0.96811f,0.00000f,0.86095f,0.86357f,0.87926f,0.88118f,0.89104f,0.89338f,0.89897f,0.90133f,0.90232f,0.90562f,0.91319f,0.91435f,0.91369f,0.90974f,0.91172f,0.92189f,0.92160f,0.92365f,0.92691f,0.92470f,0.92183f,0.91945f,0.92921f,0.92765f,0.92428f,0.92339f,0.92704f,0.93092f,0.93126f,0.93649f,0.94109f,0.96365f,0.00000f,0.83636f,0.84728f,0.85266f,0.85838f,0.85951f,0.87314f,0.88071f,0.88505f,0.89255f,0.89423f,0.89591f,0.89814f,0.90523f,0.90304f,0.90886f,0.91676f,0.91277f,0.91459f,0.91535f,0.91600f,0.92032f,0.91826f,0.91982f,0.92309f,0.92151f,0.92194f,0.92074f,0.92236f,0.92481f,0.92851f,0.93329f,0.95663f,0.00000f,0.80340f,0.81232f,0.82371f,0.83313f,0.83782f,0.85133f,0.86025f,0.86700f,0.87177f,0.87612f,0.88631f,0.88554f,0.88396f,0.89325f,0.89730f,0.90313f,0.90434f,0.90250f,0.91069f,0.91220f,0.91181f,0.91207f,0.91269f,0.91835f,0.91624f,0.91329f,0.91486f,0.91523f,0.91454f,0.92029f,0.92687f,0.95048f,0.00000f,0.77671f,0.78030f,0.79431f,0.80715f,0.82330f,0.83310f,0.83739f,0.84540f,0.85035f,0.85759f,0.86271f,0.86510f,0.87245f,0.87718f,0.88102f,0.89054f,0.89070f,0.89388f,0.89626f,0.90287f,0.90631f,0.90404f,0.90833f,0.91110f,0.90917f,0.91104f,0.91032f,0.90391f,0.90810f,0.91244f,0.91809f,0.94424f,
 0.00000f,0.99500f,0.98471f,0.97673f,0.94816f,0.94723f,0.97119f,0.96955f,0.96355f,0.95606f,0.95256f,0.96415f,0.95606f,0.96032f,0.95828f,0.96230f,0.95709f,0.95663f,0.95922f,0.96080f,0.96212f,0.96197f,0.96050f,0.96621f,0.98086f,0.98164f,0.97632f,0.98472f,0.98457f,0.98550f,0.99414f,0.99702f,1.00000f,0.00000f,0.99023f,0.98304f,0.97755f,0.97290f,0.97231f,0.97368f,0.97216f,0.97158f,0.97123f,0.97096f,0.97088f,0.96999f,0.97024f,0.97039f,0.97107f,0.97212f,0.97249f,0.97333f,0.97448f,0.97466f,0.97694f,0.97816f,0.97988f,0.98362f,0.98538f,0.98593f,0.98877f,0.99030f,0.99279f,0.99498f,0.99762f,1.00000f,0.00000f,0.99120f,0.98508f,0.98191f,0.97998f,0.97673f,0.97632f,0.97478f,0.97419f,0.97305f,0.97426f,0.97433f,0.97357f,0.97332f,0.97338f,0.97461f,0.97436f,0.97607f,0.97621f,0.97763f,0.97913f,0.97954f,0.98140f,0.98264f,0.98433f,0.98588f,0.98757f,0.98928f,0.99134f,0.99332f,0.99539f,0.99772f,1.00000f,0.00000f,0.98274f,0.98510f,0.98233f,0.98069f,0.97797f,0.97649f,0.97451f,0.97373f,0.97403f,0.97354f,0.97360f,0.97363f,0.97316f,0.97392f,0.97467f,0.97508f,0.97604f,0.97674f,0.97720f,0.97896f,0.97981f,0.98120f,0.98252f,0.98440f,0.98548f,0.98771f,0.98943f,0.99118f,0.99325f,0.99550f,0.99776f,1.00000f,0.00000f,0.94038f,0.98104f,0.98191f,0.97813f,0.97772f,0.97607f,0.97505f,0.97522f,0.97341f,0.97328f,0.97394f,0.97323f,0.97243f,0.97411f,0.97442f,0.97525f,0.97549f,0.97633f,0.97771f,0.97860f,0.98020f,0.98160f,0.98282f,0.98406f,0.98579f,0.98757f,0.98925f,0.99139f,0.99322f,0.99545f,0.99776f,1.00000f,0.00000f,0.91425f,0.95707f,0.97781f,0.97630f,0.97647f,0.97691f,0.97466f,0.97362f,0.97272f,0.97227f,0.97280f,0.97265f,0.97372f,0.97417f,0.97413f,0.97439f,0.97658f,0.97675f,0.97729f,0.97818f,0.97985f,0.98083f,0.98269f,0.98410f,0.98564f,0.98738f,0.98928f,0.99124f,0.99325f,0.99549f,0.99773f,1.00000f,0.00000f,0.90492f,0.93026f,0.96095f,0.97185f,0.97512f,0.97594f,0.97533f,0.97388f,0.97368f,0.97351f,0.97316f,0.97402f,0.97360f,0.97394f,0.97347f,0.97402f,0.97577f,0.97645f,0.97809f,0.97876f,0.97973f,0.98112f,0.98205f,0.98416f,0.98575f,0.98739f,0.98916f,0.99125f,0.99317f,0.99546f,0.99761f,1.00000f,0.00000f,0.90670f,0.91236f,0.93563f,0.95813f,0.96690f,0.97245f,0.97280f,0.97304f,0.97364f,0.97287f,0.97213f,0.97299f,0.97315f,0.97273f,0.97373f,0.97435f,0.97533f,0.97546f,0.97651f,0.97833f,0.97970f,0.98074f,0.98247f,0.98356f,0.98587f,0.98740f,0.98927f,0.99108f,0.99328f,0.99525f,0.99756f,1.00000f,0.00000f,0.91710f,0.90669f,0.92024f,0.93964f,0.95375f,0.96376f,0.96784f,0.97178f,0.97151f,0.97127f,0.97203f,0.97206f,0.97274f,0.97389f,0.97358f,0.97520f,0.97598f,0.97597f,0.97702f,0.97775f,0.97975f,0.98070f,0.98218f,0.98402f,0.98565f,0.98710f,0.98887f,0.99094f,0.99316f,0.99541f,0.99762f,1.00000f,0.00000f,0.91904f,0.90850f,0.91567f,0.92699f,0.93847f,0.94952f,0.96091f,0.96522f,0.96677f,0.96985f,0.97108f,0.97159f,0.97214f,0.97280f,0.97279f,0.97415f,0.97438f,0.97594f,0.97693f,0.97777f,0.97899f,0.97995f,0.98259f,0.98419f,0.98500f,0.98724f,0.98885f,0.99073f,0.99312f,0.99509f,0.99742f,1.00000f,0.00000f,0.92913f,0.91542f,0.91643f,0.91949f,0.92788f,0.93778f,0.94700f,0.95397f,0.95963f,0.96453f,0.96676f,0.96982f,0.97142f,0.97148f,0.97292f,0.97346f,0.97519f,0.97551f,0.97649f,0.97739f,0.97917f,0.98031f,0.98185f,0.98348f,0.98487f,0.98691f,0.98889f,0.99085f,0.99318f,0.99531f,0.99751f,1.00000f,0.00000f,0.93918f,0.91741f,0.91270f,0.91403f,0.91873f,0.92467f,0.93305f,0.94385f,0.95080f,0.95693f,0.96101f,0.96474f,0.96776f,0.96968f,0.97223f,0.97320f,0.97412f,0.97456f,0.97613f,0.97761f,0.97825f,0.97986f,0.98165f,0.98337f,0.98508f,0.98663f,0.98850f,0.99041f,0.99300f,0.99502f,0.99752f,1.00000f,0.00000f,0.94003f,0.92109f,0.91356f,0.91642f,0.91593f,0.92335f,0.93012f,0.93635f,0.94255f,0.94791f,0.95392f,0.95923f,0.96301f,0.96710f,0.96871f,0.97112f,0.97266f,0.97338f,0.97592f,0.97828f,0.97860f,0.98056f,0.98143f,0.98261f,0.98436f,0.98693f,0.98872f,0.99083f,0.99262f,0.99499f,0.99748f,1.00000f,0.00000f,0.94447f,0.93092f,0.92221f,0.91539f,0.91864f,0.91812f,0.92377f,0.93127f,0.93422f,0.93979f,0.94563f,0.95260f,0.95725f,0.96111f,0.96367f,0.96737f,0.97009f,0.97249f,0.97436f,0.97620f,0.97777f,0.97940f,0.98143f,0.98324f,0.98489f,0.98654f,0.98865f,0.99059f,0.99258f,0.99509f,0.99727f,1.00000f,0.00000f,0.95806f,0.93835f,0.92806f,0.92209f,0.91891f,0.91858f,0.92274f,0.92513f,0.93092f,0.93063f,0.93623f,0.94514f,0.94947f,0.95271f,0.95913f,0.96139f,0.96535f,0.96985f,0.97193f,0.97375f,0.97607f,0.97765f,0.98133f,0.98205f,0.98366f,0.98646f,0.98820f,0.99036f,0.99264f,0.99479f,0.99705f,1.00000f,0.00000f,0.95479f,0.94079f,0.93034f,0.92462f,0.92217f,0.91883f,0.91919f,0.92410f,0.92902f,0.92912f,0.93338f,0.93704f,0.94494f,0.94792f,0.95244f,0.95662f,0.96059f,0.96281f,0.96745f,0.97193f,0.97369f,0.97537f,0.97873f,0.98195f,0.98345f,0.98498f,0.98750f,0.99012f,0.99296f,0.99474f,0.99719f,1.00000f,0.00000f,0.95545f,0.94344f,0.93583f,0.93283f,0.92263f,0.92181f,0.92454f,0.92444f,0.92802f,0.93108f,0.93280f,0.93489f,0.93733f,0.94138f,0.94642f,0.94898f,0.95406f,0.95928f,0.96331f,0.96658f,0.96954f,0.97468f,0.97624f,0.97887f,0.98230f,0.98473f,0.98769f,0.98933f,0.99291f,0.99449f,0.99684f,1.00000f,0.00000f,0.96173f,0.95020f,0.94480f,0.93080f,0.92634f,0.92112f,0.92472f,0.92227f,0.92565f,0.92814f,0.92774f,0.93480f,0.93430f,0.93451f,0.94074f,0.94186f,0.94826f,0.95362f,0.95619f,0.96226f,0.96480f,0.96906f,0.97428f,0.97548f,0.97782f,0.98296f,0.98593f,0.98728f,0.99097f,0.99404f,0.99707f,1.00000f,0.00000f,0.95717f,0.95228f,0.94889f,0.93860f,0.92968f,0.92769f,0.92367f,0.92620f,0.92629f,0.92666f,0.92783f,0.93025f,0.93042f,0.93531f,0.93698f,0.94259f,0.94440f,0.94798f,0.95419f,0.95610f,0.96095f,0.96783f,0.97007f,0.97106f,0.97686f,0.98054f,0.98373f,0.98632f,0.99041f,0.99262f,0.99603f,1.00000f,0.00000f,0.95917f,0.95728f,0.94886f,0.94862f,0.93724f,0.93212f,0.93001f,0.92664f,0.92830f,0.92352f,0.92729f,0.93042f,0.93173f,0.93118f,0.93466f,0.93623f,0.94102f,0.94484f,0.94739f,0.95055f,0.95540f,0.96092f,0.96514f,0.96844f,0.97237f,0.97558f,0.97961f,0.98448f,0.98802f,0.99091f,0.99524f,0.99996f,0.00000f,0.95957f,0.96097f,0.94873f,0.95127f,0.93946f,0.93801f,0.93055f,0.92954f,0.92917f,0.92490f,0.92569f,0.92968f,0.92643f,0.93184f,0.93584f,0.93549f,0.93999f,0.94128f,0.94531f,0.94747f,0.95072f,0.95488f,0.95938f,0.96382f,0.96816f,0.97194f,0.97649f,0.98141f,0.98438f,0.98873f,0.99247f,0.99972f,0.00000f,0.95909f,0.95995f,0.95663f,0.95206f,0.94768f,0.94113f,0.93470f,0.92929f,0.92986f,0.92991f,0.93181f,0.92922f,0.93137f,0.93080f,0.93132f,0.93678f,0.93731f,0.93942f,0.94018f,0.94323f,0.94678f,0.95108f,0.95431f,0.95856f,0.96169f,0.96733f,0.97126f,0.97750f,0.98128f,0.98509f,0.98996f,0.99888f,0.00000f,0.96032f,0.95287f,0.94944f,0.95900f,0.94452f,0.93831f,0.94043f,0.93830f,0.93673f,0.92703f,0.92932f,0.92874f,0.93115f,0.93439f,0.93595f,0.93117f,0.93612f,0.93744f,0.93467f,0.93656f,0.94430f,0.94488f,0.94973f,0.95473f,0.95958f,0.96081f,0.96627f,0.97008f,0.97549f,0.98068f,0.98589f,0.99700f,0.00000f,0.95517f,0.95504f,0.94944f,0.94694f,0.95008f,0.94277f,0.93618f,0.93861f,0.93609f,0.93461f,0.92872f,0.93277f,0.92840f,0.93539f,0.93232f,0.93326f,0.93480f,0.93572f,0.93640f,0.94101f,0.94071f,0.94146f,0.94460f,0.94793f,0.95301f,0.95757f,0.96216f,0.96624f,0.97019f,0.97491f,0.98097f,0.99399f,0.00000f,0.94481f,0.95400f,0.95419f,0.95093f,0.94686f,0.93824f,0.93759f,0.93635f,0.93519f,0.93073f,0.93672f,0.93382f,0.93418f,0.93233f,0.93406f,0.93379f,0.93417f,0.93235f,0.93822f,0.93916f,0.94006f,0.93860f,0.94094f,0.94301f,0.94635f,0.95166f,0.95634f,0.96059f,0.96594f,0.96998f,0.97415f,0.98973f,0.00000f,0.94098f,0.93917f,0.93997f,0.93895f,0.94063f,0.94367f,0.93800f,0.93893f,0.93568f,0.93653f,0.93416f,0.93094f,0.92755f,0.92876f,0.93020f,0.93249f,0.93217f,0.93116f,0.93386f,0.93520f,0.93570f,0.93776f,0.93920f,0.93936f,0.93946f,0.94538f,0.94891f,0.95232f,0.95803f,0.96248f,0.96733f,0.98480f,0.00000f,0.92472f,0.93660f,0.93010f,0.93604f,0.93665f,0.93559f,0.93215f,0.92874f,0.92853f,0.93078f,0.92842f,0.93343f,0.92793f,0.93060f,0.92939f,0.93358f,0.93163f,0.93167f,0.93002f,0.93044f,0.93489f,0.93736f,0.93698f,0.93601f,0.93743f,0.94083f,0.94174f,0.94731f,0.94983f,0.95441f,0.95853f,0.97955f,0.00000f,0.90898f,0.91992f,0.91842f,0.92089f,0.92958f,0.92831f,0.92442f,0.92309f,0.92212f,0.92499f,0.92460f,0.92600f,0.92710f,0.92674f,0.92987f,0.92663f,0.92513f,0.93321f,0.93115f,0.92989f,0.93093f,0.93363f,0.93247f,0.92983f,0.93236f,0.93468f,0.93639f,0.94204f,0.94338f,0.94912f,0.95203f,0.97318f,0.00000f,0.89511f,0.89668f,0.90209f,0.90696f,0.90962f,0.91370f,0.92186f,0.92375f,0.91785f,0.92098f,0.91631f,0.92155f,0.92539f,0.92191f,0.92307f,0.92814f,0.92762f,0.92763f,0.93200f,0.92619f,0.93056f,0.92639f,0.93157f,0.92895f,0.93309f,0.92866f,0.92904f,0.93280f,0.93784f,0.94036f,0.94328f,0.96655f,0.00000f,0.86789f,0.87629f,0.88896f,0.88571f,0.88747f,0.89777f,0.89888f,0.89947f,0.90034f,0.90470f,0.91032f,0.91157f,0.91433f,0.91544f,0.91935f,0.91820f,0.92086f,0.92009f,0.92126f,0.92214f,0.92273f,0.92405f,0.92434f,0.92300f,0.92297f,0.92151f,0.92196f,0.92601f,0.92845f,0.93148f,0.93602f,0.96026f,0.00000f,0.83632f,0.84678f,0.85700f,0.87158f,0.87178f,0.87989f,0.88663f,0.88759f,0.88987f,0.89272f,0.90053f,0.90227f,0.90584f,0.90784f,0.90933f,0.91254f,0.91377f,0.91095f,0.91462f,0.91724f,0.91607f,0.91674f,0.91517f,0.91591f,0.91603f,0.91609f,0.91626f,0.91765f,0.92014f,0.92415f,0.92824f,0.95229f,0.00000f,0.81435f,0.82252f,0.83449f,0.84552f,0.84748f,0.85883f,0.86844f,0.87391f,0.86970f,0.88068f,0.87957f,0.88511f,0.89067f,0.89383f,0.90185f,0.89912f,0.90319f,0.90529f,0.90857f,0.90703f,0.90980f,0.90980f,0.91005f,0.91134f,0.90912f,0.91035f,0.91141f,0.91126f,0.91153f,0.91731f,0.92035f,0.94549f,0.00000f,0.78122f,0.79582f,0.80495f,0.81861f,0.82635f,0.83114f,0.84152f,0.84272f,0.85541f,0.86490f,0.86091f,0.87498f,0.87631f,0.88166f,0.88782f,0.88682f,0.88946f,0.89249f,0.89960f,0.89775f,0.90121f,0.90324f,0.90236f,0.90541f,0.90281f,0.90254f,0.90503f,0.90176f,0.90340f,0.90808f,0.91290f,0.93912f,
 0.00000f,0.99458f,0.98358f,0.95187f,0.93858f,0.93003f,0.94586f,0.95001f,0.96276f,0.95902f,0.95916f,0.96448f,0.95894f,0.96016f,0.96324f,0.96309f,0.94788f,0.96427f,0.95394f,0.95522f,0.95827f,0.96463f,0.96892f,0.95878f,0.97486f,0.96374f,0.97471f,0.97148f,0.98609f,0.98360f,0.99374f,0.99600f,1.00000f,0.00000f,0.99099f,0.98567f,0.97764f,0.97462f,0.97038f,0.97051f,0.96877f,0.97171f,0.97078f,0.97097f,0.97090f,0.97042f,0.97136f,0.97155f,0.97279f,0.97244f,0.97337f,0.97426f,0.97465f,0.97616f,0.97694f,0.97896f,0.97984f,0.98193f,0.98363f,0.98550f,0.98788f,0.99036f,0.99242f,0.99524f,0.99739f,1.00000f,0.00000f,0.99278f,0.98490f,0.98168f,0.97924f,0.97785f,0.97759f,0.97472f,0.97523f,0.97395f,0.97315f,0.97417f,0.97362f,0.97416f,0.97416f,0.97461f,0.97474f,0.97566f,0.97609f,0.97771f,0.97862f,0.97935f,0.98102f,0.98268f,0.98385f,0.98592f,0.98733f,0.98943f,0.99119f,0.99339f,0.99537f,0.99775f,1.00000f,0.00000f,0.98557f,0.98509f,0.98085f,0.97906f,0.97651f,0.97718f,0.97447f,0.97477f,0.97312f,0.97351f,0.97286f,0.97379f,0.97411f,0.97330f,0.97474f,0.97593f,0.97543f,0.97655f,0.97833f,0.97922f,0.98063f,0.98129f,0.98278f,0.98448f,0.98589f,0.98752f,0.98943f,0.99148f,0.99331f,0.99547f,0.99776f,1.00000f,0.00000f,0.93953f,0.98286f,0.98269f,0.97930f,0.97726f,0.97701f,0.97525f,0.97493f,0.97465f,0.97332f,0.97347f,0.97425f,0.97345f,0.97336f,0.97479f,0.97490f,0.97601f,0.97645f,0.97776f,0.97924f,0.98046f,0.98126f,0.98268f,0.98423f,0.98567f,0.98746f,0.98926f,0.99127f,0.99326f,0.99544f,0.99777f,1.00000f,0.00000f,0.91886f,0.95902f,0.97754f,0.97901f,0.97720f,0.97587f,0.97409f,0.97403f,0.97393f,0.97298f,0.97220f,0.97418f,0.97260f,0.97377f,0.97430f,0.97475f,0.97621f,0.97648f,0.97703f,0.97850f,0.98055f,0.98139f,0.98280f,0.98455f,0.98551f,0.98749f,0.98942f,0.99116f,0.99322f,0.99542f,0.99772f,1.00000f,0.00000f,0.90201f,0.93186f,0.96355f,0.97390f,0.97500f,0.97504f,0.97383f,0.97346f,0.97299f,0.97213f,0.97264f,0.97220f,0.97214f,0.97382f,0.97356f,0.97459f,0.97560f,0.97654f,0.97696f,0.97806f,0.97996f,0.98100f,0.98235f,0.98413f,0.98576f,0.98722f,0.98933f,0.99130f,0.99337f,0.99537f,0.99757f,1.00000f,0.00000f,0.90967f,0.91676f,0.93541f,0.95635f,0.96785f,0.97328f,0.97231f,0.97270f,0.97392f,0.97235f,0.97274f,0.97265f,0.97312f,0.97304f,0.97318f,0.97436f,0.97552f,0.97610f,0.97740f,0.97849f,0.97933f,0.98132f,0.98222f,0.98446f,0.98580f,0.98721f,0.98921f,0.99100f,0.99321f,0.99535f,0.99761f,1.00000f,0.00000f,0.91441f,0.90958f,0.92199f,0.93909f,0.95574f,0.96353f,0.96761f,0.96978f,0.97105f,0.97161f,0.97223f,0.97252f,0.97208f,0.97203f,0.97387f,0.97434f,0.97464f,0.97639f,0.97730f,0.97852f,0.97941f,0.98080f,0.98207f,0.98406f,0.98554f,0.98718f,0.98888f,0.99094f,0.99301f,0.99532f,0.99765f,1.00000f,0.00000f,0.92880f,0.90204f,0.91601f,0.92560f,0.93813f,0.95059f,0.96122f,0.96390f,0.96781f,0.97078f,0.96962f,0.97132f,0.97150f,0.97298f,0.97363f,0.97386f,0.97456f,0.97514f,0.97656f,0.97819f,0.97917f,0.98005f,0.98204f,0.98418f,0.98546f,0.98711f,0.98898f,0.99107f,0.99323f,0.99549f,0.99764f,1.00000f,0.00000f,0.93548f,0.91105f,0.90900f,0.91774f,0.92504f,0.93926f,0.94680f,0.95350f,0.96122f,0.96547f,0.96767f,0.96955f,0.97123f,0.97088f,0.97207f,0.97336f,0.97475f,0.97562f,0.97658f,0.97744f,0.97905f,0.98015f,0.98179f,0.98348f,0.98484f,0.98696f,0.98873f,0.99068f,0.99297f,0.99506f,0.99750f,1.00000f,0.00000f,0.94113f,0.92181f,0.90746f,0.91020f,0.91913f,0.92849f,0.93457f,0.94529f,0.95133f,0.95711f,0.96073f,0.96560f,0.96789f,0.96870f,0.97131f,0.97320f,0.97398f,0.97466f,0.97601f,0.97726f,0.97841f,0.98009f,0.98145f,0.98380f,0.98486f,0.98695f,0.98871f,0.99113f,0.99301f,0.99498f,0.99751f,1.00000f,0.00000f,0.93884f,0.92486f,0.90988f,0.91383f,0.91991f,0.92452f,0.92718f,0.93431f,0.94080f,0.94829f,0.95278f,0.95961f,0.96311f,0.96458f,0.96872f,0.96896f,0.97266f,0.97338f,0.97537f,0.97748f,0.97861f,0.97930f,0.98116f,0.98312f,0.98541f,0.98674f,0.98843f,0.99042f,0.99280f,0.99487f,0.99722f,1.00000f,0.00000f,0.94884f,0.92834f,0.92463f,0.91432f,0.91853f,0.92119f,0.92584f,0.92787f,0.93241f,0.94117f,0.94554f,0.95052f,0.95686f,0.95871f,0.96368f,0.96670f,0.96931f,0.97157f,0.97396f,0.97607f,0.97761f,0.98020f,0.98036f,0.98299f,0.98428f,0.98676f,0.98848f,0.99051f,0.99265f,0.99484f,0.99722f,1.00000f,0.00000f,0.95438f,0.94010f,0.92986f,0.91853f,0.91610f,0.91626f,0.91887f,0.92202f,0.92725f,0.93284f,0.93737f,0.94399f,0.94978f,0.95269f,0.95857f,0.96087f,0.96610f,0.96849f,0.97159f,0.97371f,0.97523f,0.97758f,0.98021f,0.98226f,0.98442f,0.98601f,0.98807f,0.99050f,0.99236f,0.99484f,0.99706f,1.00000f,0.00000f,0.95286f,0.94448f,0.93168f,0.92498f,0.91966f,0.91902f,0.92018f,0.92221f,0.92653f,0.93092f,0.93053f,0.93617f,0.94290f,0.94528f,0.95138f,0.95558f,0.96007f,0.96206f,0.96806f,0.97023f,0.97268f,0.97530f,0.97841f,0.98146f,0.98357f,0.98537f,0.98814f,0.99018f,0.99214f,0.99503f,0.99711f,1.00000f,0.00000f,0.95636f,0.94786f,0.93142f,0.93051f,0.92530f,0.92319f,0.92479f,0.92494f,0.92612f,0.92495f,0.93552f,0.93724f,0.94108f,0.94120f,0.94481f,0.94956f,0.95276f,0.95883f,0.96344f,0.96716f,0.96989f,0.97350f,0.97544f,0.97850f,0.98173f,0.98413f,0.98652f,0.98975f,0.99174f,0.99470f,0.99723f,1.00000f,0.00000f,0.95594f,0.94897f,0.94234f,0.93161f,0.92382f,0.92628f,0.91781f,0.92206f,0.92624f,0.92538f,0.92967f,0.93140f,0.93650f,0.93766f,0.93786f,0.94585f,0.95038f,0.95333f,0.95591f,0.96079f,0.96590f,0.96981f,0.97158f,0.97610f,0.97951f,0.98236f,0.98553f,0.98820f,0.99084f,0.99394f,0.99697f,1.00000f,0.00000f,0.96091f,0.95397f,0.94513f,0.93735f,0.93164f,0.93078f,0.92715f,0.92517f,0.92290f,0.92865f,0.92706f,0.93084f,0.93236f,0.93692f,0.93522f,0.93885f,0.94316f,0.94747f,0.95274f,0.95719f,0.96157f,0.96444f,0.96828f,0.97258f,0.97657f,0.97857f,0.98338f,0.98601f,0.98937f,0.99308f,0.99663f,1.00000f,0.00000f,0.95866f,0.95725f,0.95338f,0.94663f,0.93864f,0.92946f,0.93085f,0.92757f,0.92331f,0.92570f,0.92583f,0.93234f,0.92882f,0.93358f,0.93424f,0.93814f,0.94001f,0.94480f,0.94857f,0.95204f,0.95621f,0.96080f,0.96341f,0.96712f,0.97082f,0.97465f,0.98030f,0.98372f,0.98751f,0.99020f,0.99482f,0.99995f,0.00000f,0.96471f,0.95987f,0.95205f,0.94383f,0.93987f,0.93370f,0.93306f,0.93081f,0.93020f,0.92837f,0.93084f,0.92852f,0.93254f,0.93242f,0.93373f,0.93501f,0.93714f,0.93864f,0.93998f,0.94617f,0.94870f,0.95386f,0.95885f,0.96345f,0.96678f,0.97097f,0.97591f,0.98064f,0.98399f,0.98818f,0.99229f,0.99969f,0.00000f,0.96749f,0.96309f,0.95516f,0.94897f,0.95143f,0.94001f,0.93510f,0.93203f,0.93156f,0.93094f,0.93527f,0.93088f,0.93241f,0.93240f,0.93262f,0.93480f,0.93340f,0.93760f,0.93730f,0.94100f,0.94437f,0.94937f,0.95330f,0.95829f,0.96163f,0.96657f,0.97043f,0.97579f,0.97976f,0.98446f,0.98955f,0.99888f,0.00000f,0.96376f,0.95504f,0.95437f,0.95047f,0.94608f,0.94362f,0.94287f,0.93664f,0.93357f,0.93109f,0.93262f,0.93058f,0.93009f,0.93137f,0.93335f,0.93447f,0.93346f,0.93786f,0.94040f,0.93913f,0.94095f,0.94405f,0.94775f,0.95274f,0.95638f,0.96271f,0.96657f,0.96950f,0.97599f,0.98030f,0.98524f,0.99703f,0.00000f,0.95847f,0.95867f,0.95942f,0.95316f,0.94981f,0.94404f,0.93989f,0.93914f,0.93533f,0.93047f,0.93353f,0.93087f,0.92574f,0.92787f,0.93308f,0.93320f,0.93273f,0.93518f,0.93263f,0.93595f,0.93860f,0.94467f,0.94546f,0.94759f,0.95256f,0.95601f,0.96075f,0.96505f,0.96963f,0.97492f,0.97902f,0.99331f,0.00000f,0.94082f,0.94765f,0.95006f,0.94915f,0.95078f,0.94220f,0.94020f,0.93857f,0.93249f,0.93150f,0.93767f,0.93628f,0.93079f,0.93107f,0.92878f,0.92926f,0.93040f,0.93310f,0.93258f,0.93574f,0.93652f,0.93730f,0.93963f,0.94281f,0.94724f,0.95118f,0.95368f,0.95876f,0.96443f,0.96732f,0.97223f,0.98868f,0.00000f,0.94191f,0.93728f,0.95090f,0.94451f,0.94430f,0.94056f,0.93939f,0.93131f,0.93637f,0.93276f,0.92951f,0.93414f,0.93414f,0.92801f,0.92822f,0.93196f,0.93064f,0.93235f,0.93145f,0.93321f,0.93564f,0.93799f,0.93651f,0.93803f,0.94176f,0.94373f,0.94941f,0.95280f,0.95692f,0.95923f,0.96545f,0.98394f,0.00000f,0.93542f,0.93814f,0.93482f,0.93813f,0.93508f,0.93591f,0.93544f,0.93390f,0.93194f,0.92917f,0.93485f,0.93208f,0.92890f,0.92589f,0.92508f,0.93109f,0.92891f,0.93313f,0.93152f,0.93069f,0.93420f,0.93521f,0.93441f,0.93589f,0.93633f,0.93772f,0.94051f,0.94560f,0.94880f,0.95279f,0.95664f,0.97790f,0.00000f,0.90965f,0.91586f,0.92923f,0.92337f,0.92659f,0.92968f,0.93349f,0.92882f,0.92630f,0.92439f,0.92210f,0.92290f,0.92463f,0.92190f,0.93200f,0.92427f,0.92430f,0.92522f,0.92917f,0.93076f,0.92969f,0.93045f,0.93253f,0.92919f,0.93075f,0.93362f,0.93293f,0.93639f,0.93912f,0.94479f,0.95056f,0.97024f,0.00000f,0.89962f,0.90305f,0.90375f,0.91368f,0.91264f,0.91146f,0.91370f,0.92158f,0.92154f,0.91617f,0.91973f,0.92455f,0.92126f,0.92340f,0.92471f,0.92242f,0.92143f,0.92752f,0.92301f,0.92972f,0.92209f,0.92485f,0.92703f,0.92905f,0.92591f,0.92691f,0.92570f,0.92986f,0.93596f,0.93490f,0.93991f,0.96386f,0.00000f,0.87490f,0.88146f,0.89224f,0.89396f,0.89666f,0.90219f,0.90242f,0.90345f,0.90638f,0.90312f,0.90905f,0.91286f,0.91762f,0.91296f,0.91579f,0.91887f,0.91552f,0.91986f,0.92233f,0.92349f,0.92583f,0.92329f,0.92238f,0.92251f,0.92015f,0.91915f,0.92127f,0.92132f,0.92382f,0.92946f,0.93228f,0.95616f,0.00000f,0.84782f,0.85954f,0.87189f,0.88003f,0.87504f,0.88504f,0.88396f,0.89293f,0.89917f,0.89971f,0.89842f,0.90352f,0.90596f,0.90598f,0.90063f,0.91244f,0.91092f,0.91230f,0.91717f,0.91588f,0.91559f,0.91402f,0.91452f,0.91508f,0.91534f,0.91125f,0.91433f,0.91618f,0.91533f,0.91969f,0.92376f,0.94964f,0.00000f,0.82054f,0.83674f,0.83984f,0.84393f,0.85730f,0.86256f,0.87146f,0.87751f,0.87919f,0.88195f,0.88513f,0.89530f,0.90107f,0.89608f,0.89735f,0.89720f,0.90250f,0.90490f,0.90374f,0.91302f,0.90586f,0.90943f,0.90848f,0.90980f,0.90648f,0.90575f,0.90540f,0.90686f,0.90786f,0.91146f,0.91758f,0.94024f,0.00000f,0.79256f,0.80389f,0.81401f,0.82472f,0.83368f,0.83776f,0.84950f,0.84862f,0.85966f,0.85884f,0.86652f,0.87066f,0.87626f,0.88249f,0.88146f,0.88455f,0.88706f,0.89463f,0.89469f,0.89123f,0.89717f,0.89823f,0.90099f,0.89979f,0.90129f,0.89849f,0.89912f,0.89685f,0.90077f,0.90274f,0.90798f,0.93463f,
 0.00000f,0.99319f,0.97962f,0.98335f,0.96320f,0.95347f,0.96118f,0.94487f,0.95266f,0.92367f,0.94761f,0.97072f,0.95877f,0.95774f,0.95903f,0.96156f,0.96212f,0.96229f,0.96165f,0.96570f,0.96550f,0.96152f,0.96417f,0.96836f,0.97037f,0.97476f,0.96903f,0.98151f,0.98621f,0.99202f,0.99183f,0.99724f,1.00000f,0.00000f,0.99122f,0.98238f,0.97976f,0.97548f,0.97344f,0.97182f,0.96949f,0.96951f,0.96721f,0.96677f,0.97217f,0.97066f,0.97096f,0.97059f,0.97268f,0.97304f,0.97266f,0.97362f,0.97549f,0.97680f,0.97799f,0.97926f,0.98074f,0.98213f,0.98413f,0.98532f,0.98816f,0.99019f,0.99252f,0.99509f,0.99758f,1.00000f,0.00000f,0.99141f,0.98536f,0.98132f,0.97906f,0.97853f,0.97651f,0.97495f,0.97442f,0.97389f,0.97339f,0.97268f,0.97303f,0.97357f,0.97360f,0.97374f,0.97525f,0.97597f,0.97666f,0.97754f,0.97834f,0.97975f,0.98115f,0.98254f,0.98419f,0.98572f,0.98750f,0.98922f,0.99117f,0.99310f,0.99547f,0.99772f,1.00000f,0.00000f,0.98483f,0.98517f,0.98199f,0.97928f,0.97788f,0.97742f,0.97568f,0.97475f,0.97310f,0.97309f,0.97393f,0.97341f,0.97428f,0.97398f,0.97401f,0.97474f,0.97612f,0.97713f,0.97729f,0.97907f,0.98028f,0.98132f,0.98263f,0.98433f,0.98596f,0.98766f,0.98965f,0.99122f,0.99339f,0.99549f,0.99778f,1.00000f,0.00000f,0.94263f,0.98202f,0.98330f,0.97912f,0.97537f,0.97559f,0.97459f,0.97405f,0.97390f,0.97288f,0.97324f,0.97340f,0.97333f,0.97410f,0.97422f,0.97528f,0.97608f,0.97656f,0.97690f,0.97894f,0.98039f,0.98141f,0.98252f,0.98412f,0.98556f,0.98768f,0.98949f,0.99148f,0.99331f,0.99538f,0.99778f,1.00000f,0.00000f,0.91242f,0.96114f,0.97894f,0.97886f,0.97709f,0.97469f,0.97548f,0.97322f,0.97385f,0.97411f,0.97303f,0.97329f,0.97423f,0.97352f,0.97443f,0.97487f,0.97624f,0.97669f,0.97813f,0.97887f,0.98010f,0.98147f,0.98267f,0.98408f,0.98574f,0.98757f,0.98916f,0.99108f,0.99330f,0.99545f,0.99777f,1.00000f,0.00000f,0.90602f,0.93192f,0.95939f,0.97413f,0.97462f,0.97457f,0.97302f,0.97397f,0.97281f,0.97263f,0.97219f,0.97272f,0.97214f,0.97312f,0.97347f,0.97458f,0.97532f,0.97649f,0.97700f,0.97846f,0.97922f,0.98097f,0.98316f,0.98400f,0.98575f,0.98758f,0.98924f,0.99099f,0.99323f,0.99551f,0.99768f,1.00000f,0.00000f,0.90994f,0.91606f,0.94000f,0.95611f,0.96650f,0.97212f,0.97340f,0.97314f,0.97241f,0.97265f,0.97234f,0.97384f,0.97226f,0.97336f,0.97368f,0.97480f,0.97495f,0.97600f,0.97707f,0.97810f,0.97929f,0.98137f,0.98188f,0.98379f,0.98525f,0.98753f,0.98918f,0.99087f,0.99314f,0.99539f,0.99760f,1.00000f,0.00000f,0.91850f,0.90655f,0.92061f,0.94018f,0.95586f,0.96291f,0.96799f,0.97014f,0.97032f,0.97192f,0.97238f,0.97173f,0.97245f,0.97285f,0.97386f,0.97440f,0.97522f,0.97533f,0.97696f,0.97803f,0.97929f,0.98106f,0.98244f,0.98410f,0.98527f,0.98722f,0.98909f,0.99108f,0.99335f,0.99533f,0.99758f,1.00000f,0.00000f,0.91804f,0.90775f,0.91365f,0.92546f,0.93753f,0.95235f,0.95904f,0.96457f,0.96733f,0.97068f,0.97054f,0.97223f,0.97348f,0.97248f,0.97292f,0.97416f,0.97430f,0.97589f,0.97683f,0.97821f,0.97874f,0.98055f,0.98180f,0.98396f,0.98528f,0.98714f,0.98905f,0.99087f,0.99272f,0.99551f,0.99747f,1.00000f,0.00000f,0.93474f,0.91771f,0.91162f,0.92081f,0.92472f,0.93595f,0.94583f,0.95545f,0.96149f,0.96329f,0.96771f,0.96967f,0.96999f,0.97258f,0.97342f,0.97434f,0.97350f,0.97470f,0.97655f,0.97820f,0.97914f,0.98022f,0.98177f,0.98302f,0.98514f,0.98668f,0.98864f,0.99080f,0.99293f,0.99532f,0.99755f,1.00000f,0.00000f,0.93644f,0.92638f,0.91060f,0.91563f,0.91977f,0.92618f,0.93510f,0.94457f,0.94953f,0.95805f,0.96086f,0.96461f,0.96699f,0.97142f,0.97140f,0.97266f,0.97419f,0.97431f,0.97637f,0.97751f,0.97859f,0.97966f,0.98156f,0.98341f,0.98473f,0.98664f,0.98852f,0.99081f,0.99285f,0.99517f,0.99753f,1.00000f,0.00000f,0.94669f,0.91918f,0.91792f,0.91664f,0.91583f,0.92182f,0.92830f,0.93494f,0.93956f,0.94953f,0.95302f,0.95853f,0.96388f,0.96610f,0.96852f,0.97037f,0.97256f,0.97348f,0.97508f,0.97670f,0.97883f,0.98011f,0.98200f,0.98298f,0.98437f,0.98651f,0.98901f,0.99063f,0.99255f,0.99489f,0.99730f,1.00000f,0.00000f,0.94978f,0.93593f,0.92070f,0.91517f,0.91906f,0.92034f,0.92396f,0.92816f,0.93407f,0.94333f,0.94708f,0.95146f,0.95502f,0.95879f,0.96394f,0.96658f,0.96906f,0.97209f,0.97524f,0.97530f,0.97818f,0.98057f,0.98123f,0.98299f,0.98475f,0.98666f,0.98871f,0.99036f,0.99281f,0.99495f,0.99723f,1.00000f,0.00000f,0.95350f,0.93446f,0.92206f,0.92149f,0.91873f,0.92202f,0.91858f,0.92439f,0.92810f,0.93320f,0.93703f,0.94454f,0.94820f,0.95361f,0.95785f,0.96157f,0.96419f,0.96806f,0.97139f,0.97361f,0.97590f,0.97906f,0.97977f,0.98201f,0.98452f,0.98661f,0.98837f,0.99076f,0.99267f,0.99484f,0.99711f,1.00000f,0.00000f,0.94585f,0.93535f,0.93397f,0.92258f,0.92043f,0.91924f,0.91866f,0.92271f,0.92777f,0.93364f,0.93187f,0.93748f,0.94391f,0.94734f,0.95204f,0.95548f,0.95933f,0.96442f,0.96611f,0.96997f,0.97363f,0.97607f,0.97931f,0.98124f,0.98371f,0.98568f,0.98810f,0.99008f,0.99305f,0.99468f,0.99702f,1.00000f,0.00000f,0.96052f,0.94908f,0.93292f,0.93017f,0.92559f,0.92369f,0.92145f,0.92280f,0.92376f,0.92670f,0.93107f,0.93410f,0.93822f,0.94052f,0.94380f,0.95085f,0.95632f,0.95708f,0.96350f,0.96470f,0.96997f,0.97383f,0.97542f,0.97839f,0.98178f,0.98435f,0.98696f,0.98894f,0.99173f,0.99486f,0.99728f,1.00000f,0.00000f,0.96485f,0.95168f,0.93681f,0.92951f,0.92692f,0.92146f,0.92563f,0.92522f,0.92581f,0.92735f,0.92915f,0.93111f,0.93676f,0.93491f,0.94071f,0.94495f,0.94791f,0.95552f,0.95794f,0.96267f,0.96518f,0.96939f,0.97242f,0.97632f,0.97910f,0.98238f,0.98611f,0.98721f,0.99105f,0.99382f,0.99679f,1.00000f,0.00000f,0.95679f,0.94842f,0.94884f,0.93650f,0.93184f,0.92805f,0.92550f,0.92567f,0.92445f,0.92601f,0.92815f,0.93116f,0.93107f,0.93371f,0.93463f,0.93891f,0.94211f,0.94933f,0.94989f,0.95672f,0.95996f,0.96414f,0.96745f,0.97179f,0.97513f,0.97990f,0.98297f,0.98612f,0.98936f,0.99268f,0.99606f,1.00000f,0.00000f,0.95702f,0.94823f,0.94916f,0.94488f,0.93383f,0.93304f,0.92378f,0.92464f,0.92731f,0.92774f,0.92851f,0.93347f,0.93297f,0.93268f,0.93858f,0.94009f,0.93954f,0.94358f,0.94595f,0.94920f,0.95459f,0.95943f,0.96129f,0.96677f,0.97014f,0.97662f,0.97896f,0.98332f,0.98742f,0.99094f,0.99457f,0.99994f,0.00000f,0.95931f,0.95837f,0.95485f,0.94278f,0.93653f,0.93624f,0.93387f,0.93447f,0.92928f,0.92839f,0.92723f,0.92563f,0.92955f,0.92953f,0.93160f,0.93856f,0.93879f,0.93749f,0.94327f,0.94588f,0.94916f,0.95348f,0.95842f,0.96478f,0.96632f,0.97114f,0.97467f,0.97912f,0.98388f,0.98809f,0.99168f,0.99964f,0.00000f,0.96586f,0.95948f,0.95876f,0.94664f,0.94341f,0.93594f,0.93557f,0.93098f,0.93220f,0.93338f,0.92917f,0.92789f,0.93090f,0.93145f,0.93074f,0.93539f,0.93579f,0.93730f,0.93800f,0.94109f,0.94537f,0.94868f,0.95464f,0.95730f,0.96081f,0.96645f,0.96992f,0.97487f,0.97844f,0.98381f,0.98842f,0.99880f,0.00000f,0.95784f,0.95825f,0.96493f,0.94838f,0.94812f,0.94024f,0.93804f,0.93994f,0.93150f,0.92715f,0.92438f,0.93012f,0.92925f,0.92837f,0.93139f,0.93399f,0.93296f,0.93302f,0.93410f,0.94072f,0.94156f,0.94324f,0.94897f,0.95204f,0.95304f,0.96042f,0.96320f,0.96957f,0.97503f,0.98025f,0.98380f,0.99651f,0.00000f,0.94888f,0.95881f,0.95817f,0.95103f,0.95132f,0.94402f,0.94466f,0.93354f,0.93433f,0.93559f,0.92690f,0.93232f,0.92704f,0.93383f,0.93177f,0.93603f,0.93193f,0.93231f,0.93747f,0.93612f,0.93651f,0.93891f,0.94114f,0.94587f,0.95059f,0.95503f,0.95768f,0.96368f,0.96922f,0.97304f,0.97801f,0.99328f,0.00000f,0.95139f,0.95337f,0.95263f,0.95158f,0.94890f,0.94722f,0.94248f,0.93792f,0.93979f,0.93253f,0.93186f,0.92892f,0.93003f,0.93072f,0.92918f,0.93252f,0.93266f,0.93439f,0.93295f,0.93550f,0.93444f,0.93755f,0.94191f,0.94076f,0.94371f,0.94814f,0.95164f,0.95756f,0.96203f,0.96740f,0.97026f,0.98794f,0.00000f,0.94828f,0.94963f,0.94634f,0.94616f,0.94185f,0.94414f,0.93707f,0.93449f,0.93730f,0.93381f,0.93294f,0.93585f,0.92984f,0.93269f,0.92971f,0.93490f,0.92912f,0.93139f,0.92983f,0.93613f,0.93298f,0.93114f,0.93733f,0.93679f,0.93661f,0.94076f,0.94585f,0.95075f,0.95379f,0.95707f,0.96321f,0.98225f,0.00000f,0.92638f,0.93019f,0.93815f,0.93888f,0.93900f,0.93329f,0.93533f,0.93558f,0.93151f,0.92978f,0.92932f,0.93434f,0.92605f,0.93046f,0.93167f,0.92507f,0.92729f,0.92719f,0.93168f,0.93257f,0.93248f,0.93182f,0.92969f,0.93244f,0.93264f,0.93634f,0.93978f,0.94213f,0.94565f,0.95077f,0.95328f,0.97545f,0.00000f,0.92226f,0.91551f,0.92355f,0.92914f,0.93062f,0.93318f,0.93489f,0.93294f,0.93214f,0.92738f,0.92660f,0.92914f,0.92304f,0.92632f,0.92497f,0.92615f,0.92878f,0.93361f,0.92896f,0.92696f,0.92276f,0.92874f,0.92825f,0.92764f,0.92790f,0.93112f,0.93318f,0.93462f,0.94045f,0.94187f,0.94351f,0.96885f,0.00000f,0.89910f,0.90493f,0.90370f,0.91582f,0.91075f,0.91845f,0.92051f,0.92450f,0.92349f,0.92069f,0.91743f,0.91804f,0.92049f,0.92371f,0.92291f,0.92377f,0.92232f,0.92350f,0.92753f,0.92282f,0.92604f,0.92412f,0.92269f,0.92414f,0.92545f,0.92553f,0.92556f,0.92794f,0.93089f,0.93184f,0.93543f,0.96086f,0.00000f,0.87180f,0.88228f,0.89887f,0.90058f,0.89482f,0.90733f,0.90992f,0.90786f,0.90905f,0.91068f,0.90828f,0.90555f,0.91537f,0.91440f,0.91227f,0.91796f,0.91746f,0.91703f,0.91676f,0.91851f,0.92124f,0.91915f,0.92153f,0.91958f,0.91932f,0.91701f,0.91612f,0.91777f,0.92314f,0.92544f,0.92807f,0.95292f,0.00000f,0.85981f,0.86007f,0.87274f,0.87709f,0.87862f,0.88538f,0.89023f,0.89112f,0.89538f,0.89758f,0.90096f,0.90276f,0.90479f,0.90493f,0.90385f,0.90835f,0.90992f,0.91440f,0.91199f,0.90936f,0.91191f,0.91355f,0.91198f,0.91003f,0.91158f,0.91078f,0.91067f,0.91075f,0.91452f,0.91428f,0.91889f,0.94374f,0.00000f,0.82818f,0.83263f,0.84891f,0.85291f,0.86499f,0.86526f,0.86620f,0.87360f,0.87583f,0.87960f,0.88474f,0.89074f,0.89214f,0.89923f,0.89577f,0.89901f,0.89910f,0.90128f,0.89729f,0.90406f,0.90887f,0.90038f,0.90994f,0.90451f,0.90435f,0.90257f,0.90305f,0.90175f,0.90309f,0.90833f,0.91278f,0.93743f,0.00000f,0.80503f,0.81758f,0.81837f,0.82442f,0.83855f,0.84067f,0.84925f,0.85359f,0.85788f,0.86575f,0.87087f,0.87064f,0.87923f,0.88080f,0.88654f,0.88578f,0.88698f,0.88950f,0.89415f,0.89412f,0.89442f,0.89509f,0.89701f,0.89668f,0.89627f,0.89408f,0.89590f,0.89297f,0.89285f,0.89967f,0.90155f,0.92939f,
 0.00000f,0.99438f,0.98583f,0.97312f,0.95746f,0.94786f,0.95581f,0.95970f,0.95436f,0.94583f,0.95004f,0.93137f,0.93844f,0.95800f,0.96211f,0.96489f,0.96442f,0.96699f,0.96497f,0.96603f,0.96534f,0.96669f,0.97340f,0.97707f,0.97840f,0.97043f,0.97500f,0.98460f,0.98819f,0.99103f,0.99476f,0.99495f,1.00000f,0.00000f,0.99030f,0.98397f,0.97958f,0.97637f,0.97389f,0.97256f,0.97269f,0.96982f,0.96802f,0.96859f,0.96761f,0.96728f,0.96783f,0.97279f,0.97219f,0.97299f,0.97366f,0.97454f,0.97568f,0.97692f,0.97749f,0.97940f,0.98102f,0.98269f,0.98456f,0.98616f,0.98835f,0.99024f,0.99259f,0.99500f,0.99744f,1.00000f,0.00000f,0.99131f,0.98690f,0.98220f,0.97934f,0.97676f,0.97607f,0.97519f,0.97339f,0.97361f,0.97344f,0.97287f,0.97261f,0.97316f,0.97439f,0.97453f,0.97486f,0.97554f,0.97653f,0.97805f,0.97892f,0.98059f,0.98153f,0.98287f,0.98420f,0.98599f,0.98766f,0.98947f,0.99149f,0.99335f,0.99545f,0.99776f,1.00000f,0.00000f,0.98461f,0.98622f,0.98300f,0.97929f,0.97736f,0.97603f,0.97527f,0.97439f,0.97326f,0.97344f,0.97366f,0.97371f,0.97418f,0.97401f,0.97486f,0.97481f,0.97466f,0.97650f,0.97807f,0.97875f,0.97972f,0.98183f,0.98253f,0.98452f,0.98598f,0.98720f,0.98931f,0.99148f,0.99314f,0.99546f,0.99774f,1.00000f,0.00000f,0.94099f,0.98164f,0.98264f,0.97961f,0.97660f,0.97635f,0.97477f,0.97340f,0.97397f,0.97392f,0.97370f,0.97336f,0.97350f,0.97382f,0.97373f,0.97471f,0.97551f,0.97653f,0.97732f,0.97820f,0.98020f,0.98124f,0.98290f,0.98394f,0.98614f,0.98753f,0.98933f,0.99120f,0.99342f,0.99544f,0.99778f,1.00000f,0.00000f,0.91530f,0.95981f,0.97844f,0.97969f,0.97729f,0.97522f,0.97482f,0.97375f,0.97285f,0.97256f,0.97248f,0.97366f,0.97331f,0.97395f,0.97386f,0.97455f,0.97561f,0.97675f,0.97783f,0.97842f,0.98020f,0.98130f,0.98246f,0.98435f,0.98597f,0.98754f,0.98926f,0.99122f,0.99325f,0.99540f,0.99774f,1.00000f,0.00000f,0.90992f,0.93052f,0.96114f,0.97018f,0.97585f,0.97463f,0.97339f,0.97428f,0.97311f,0.97252f,0.97336f,0.97261f,0.97346f,0.97366f,0.97367f,0.97471f,0.97542f,0.97610f,0.97736f,0.97811f,0.97969f,0.98114f,0.98243f,0.98416f,0.98579f,0.98764f,0.98940f,0.99134f,0.99333f,0.99551f,0.99767f,1.00000f,0.00000f,0.91016f,0.91316f,0.93958f,0.95732f,0.96949f,0.97199f,0.97166f,0.97287f,0.97203f,0.97278f,0.97243f,0.97120f,0.97318f,0.97312f,0.97339f,0.97383f,0.97542f,0.97639f,0.97733f,0.97761f,0.97963f,0.98099f,0.98235f,0.98383f,0.98539f,0.98741f,0.98928f,0.99102f,0.99297f,0.99525f,0.99754f,1.00000f,0.00000f,0.91833f,0.91026f,0.92265f,0.93892f,0.95329f,0.96323f,0.96818f,0.97052f,0.97187f,0.97140f,0.97120f,0.97298f,0.97309f,0.97318f,0.97333f,0.97409f,0.97473f,0.97538f,0.97724f,0.97855f,0.97917f,0.98102f,0.98229f,0.98366f,0.98562f,0.98752f,0.98902f,0.99093f,0.99309f,0.99526f,0.99758f,1.00000f,0.00000f,0.92297f,0.91164f,0.91486f,0.92566f,0.93807f,0.94998f,0.95998f,0.96462f,0.96832f,0.96980f,0.97044f,0.97050f,0.97222f,0.97209f,0.97313f,0.97380f,0.97389f,0.97590f,0.97621f,0.97737f,0.97887f,0.98087f,0.98224f,0.98363f,0.98540f,0.98696f,0.98921f,0.99115f,0.99321f,0.99515f,0.99751f,1.00000f,0.00000f,0.92938f,0.91281f,0.91149f,0.91390f,0.92707f,0.93838f,0.94757f,0.95554f,0.95968f,0.96513f,0.96685f,0.97032f,0.96989f,0.97161f,0.97362f,0.97350f,0.97479f,0.97513f,0.97664f,0.97734f,0.97910f,0.98071f,0.98219f,0.98349f,0.98521f,0.98718f,0.98910f,0.99076f,0.99314f,0.99518f,0.99744f,1.00000f,0.00000f,0.93412f,0.91470f,0.91420f,0.91801f,0.91806f,0.92932f,0.93637f,0.94454f,0.95109f,0.95861f,0.96170f,0.96499f,0.96737f,0.96994f,0.97150f,0.97268f,0.97397f,0.97457f,0.97613f,0.97765f,0.97888f,0.98077f,0.98170f,0.98426f,0.98538f,0.98667f,0.98845f,0.99060f,0.99290f,0.99511f,0.99751f,1.00000f,0.00000f,0.94328f,0.92738f,0.91148f,0.91433f,0.91596f,0.92196f,0.93049f,0.93436f,0.94413f,0.94788f,0.95388f,0.95939f,0.96269f,0.96754f,0.96851f,0.97156f,0.97176f,0.97368f,0.97545f,0.97522f,0.97922f,0.98072f,0.98190f,0.98335f,0.98495f,0.98678f,0.98815f,0.99074f,0.99241f,0.99496f,0.99738f,1.00000f,0.00000f,0.94771f,0.92551f,0.91861f,0.91461f,0.91770f,0.91784f,0.92677f,0.93076f,0.93195f,0.93928f,0.94439f,0.95216f,0.95511f,0.95963f,0.96324f,0.96710f,0.96866f,0.97221f,0.97382f,0.97558f,0.97845f,0.97954f,0.98100f,0.98263f,0.98448f,0.98670f,0.98842f,0.99052f,0.99256f,0.99493f,0.99729f,1.00000f,0.00000f,0.95555f,0.94641f,0.93164f,0.91439f,0.91633f,0.91936f,0.92018f,0.92578f,0.92945f,0.93546f,0.93863f,0.94303f,0.94837f,0.95465f,0.95746f,0.96147f,0.96545f,0.96847f,0.97056f,0.97342f,0.97602f,0.97794f,0.98027f,0.98131f,0.98383f,0.98624f,0.98819f,0.99021f,0.99226f,0.99457f,0.99715f,1.00000f,0.00000f,0.94907f,0.94136f,0.93368f,0.92413f,0.92206f,0.91673f,0.91679f,0.92261f,0.92652f,0.92899f,0.93268f,0.94068f,0.94291f,0.94740f,0.95142f,0.95580f,0.95948f,0.96305f,0.96805f,0.97050f,0.97367f,0.97580f,0.97821f,0.98005f,0.98375f,0.98537f,0.98820f,0.99037f,0.99196f,0.99436f,0.99700f,1.00000f,0.00000f,0.95794f,0.94464f,0.93761f,0.92932f,0.92627f,0.92216f,0.92188f,0.92097f,0.92357f,0.92647f,0.93057f,0.93523f,0.93682f,0.94348f,0.94572f,0.94964f,0.95474f,0.95682f,0.96236f,0.96583f,0.96910f,0.97344f,0.97598f,0.97858f,0.98158f,0.98473f,0.98687f,0.98827f,0.99194f,0.99454f,0.99692f,1.00000f,0.00000f,0.95459f,0.95561f,0.94031f,0.92859f,0.92935f,0.92358f,0.92564f,0.92341f,0.92672f,0.92645f,0.93002f,0.93063f,0.93565f,0.93677f,0.93967f,0.94457f,0.94546f,0.95096f,0.95856f,0.96291f,0.96603f,0.96830f,0.97228f,0.97491f,0.97856f,0.98184f,0.98566f,0.98796f,0.99110f,0.99373f,0.99698f,1.00000f,0.00000f,0.96091f,0.95285f,0.94978f,0.93389f,0.93122f,0.93082f,0.92459f,0.92436f,0.92293f,0.92752f,0.92763f,0.93221f,0.93088f,0.93404f,0.93350f,0.93782f,0.94273f,0.94829f,0.95208f,0.95420f,0.95835f,0.96523f,0.96854f,0.97005f,0.97389f,0.97761f,0.98156f,0.98567f,0.98888f,0.99249f,0.99601f,1.00000f,0.00000f,0.96658f,0.95506f,0.94981f,0.93978f,0.93744f,0.93429f,0.92699f,0.92816f,0.92512f,0.92664f,0.93033f,0.92795f,0.92834f,0.93038f,0.93314f,0.93653f,0.93918f,0.94642f,0.94709f,0.95118f,0.95417f,0.96004f,0.96250f,0.96695f,0.97025f,0.97450f,0.97964f,0.98263f,0.98715f,0.99080f,0.99489f,0.99994f,0.00000f,0.95752f,0.95976f,0.94954f,0.94294f,0.94112f,0.93381f,0.93018f,0.92963f,0.93231f,0.92752f,0.92907f,0.92586f,0.93039f,0.93220f,0.93409f,0.93285f,0.93664f,0.93730f,0.94081f,0.94498f,0.94746f,0.95185f,0.95751f,0.96188f,0.96460f,0.97079f,0.97370f,0.97829f,0.98279f,0.98769f,0.99157f,0.99960f,0.00000f,0.95774f,0.95586f,0.95194f,0.94470f,0.94519f,0.93914f,0.93495f,0.92736f,0.93004f,0.92524f,0.92783f,0.93067f,0.92985f,0.92927f,0.93168f,0.93196f,0.93439f,0.93328f,0.93859f,0.94033f,0.94434f,0.94780f,0.95430f,0.95696f,0.96144f,0.96534f,0.96845f,0.97376f,0.97906f,0.98326f,0.98844f,0.99866f,0.00000f,0.96159f,0.96160f,0.95539f,0.95341f,0.94484f,0.94229f,0.93790f,0.93352f,0.93417f,0.93381f,0.92713f,0.93032f,0.93193f,0.92691f,0.93233f,0.93427f,0.93273f,0.93537f,0.93522f,0.93862f,0.93990f,0.94118f,0.94797f,0.95175f,0.95606f,0.96068f,0.96538f,0.97052f,0.97347f,0.97763f,0.98270f,0.99627f,0.00000f,0.95559f,0.96126f,0.95523f,0.95260f,0.95111f,0.94433f,0.94186f,0.93598f,0.93701f,0.93432f,0.93388f,0.93078f,0.93152f,0.92809f,0.93590f,0.93381f,0.93289f,0.93181f,0.93677f,0.93501f,0.93410f,0.93935f,0.93968f,0.94663f,0.94867f,0.95145f,0.95800f,0.96195f,0.96587f,0.97228f,0.97651f,0.99230f,0.00000f,0.95615f,0.95701f,0.95042f,0.94925f,0.95024f,0.94429f,0.94211f,0.93512f,0.93829f,0.93276f,0.93363f,0.93175f,0.92526f,0.93323f,0.93032f,0.92980f,0.93073f,0.92927f,0.93410f,0.93655f,0.93317f,0.93508f,0.93735f,0.94153f,0.94351f,0.94623f,0.95033f,0.95729f,0.95923f,0.96433f,0.96969f,0.98700f,0.00000f,0.94636f,0.95334f,0.95029f,0.94216f,0.94453f,0.94469f,0.93884f,0.93382f,0.93540f,0.93178f,0.93703f,0.93275f,0.93041f,0.92824f,0.93124f,0.92906f,0.92725f,0.92914f,0.93311f,0.93174f,0.93198f,0.93744f,0.93264f,0.93783f,0.93803f,0.94364f,0.94506f,0.94783f,0.95149f,0.95625f,0.95925f,0.98072f,0.00000f,0.93649f,0.93890f,0.93976f,0.94411f,0.94091f,0.93815f,0.93952f,0.93807f,0.93722f,0.92848f,0.93699f,0.93306f,0.92715f,0.93132f,0.92926f,0.93035f,0.92866f,0.93187f,0.92984f,0.92788f,0.92494f,0.93375f,0.92980f,0.92996f,0.93198f,0.93603f,0.93708f,0.94042f,0.94316f,0.94804f,0.95116f,0.97430f,0.00000f,0.91823f,0.92480f,0.92981f,0.92908f,0.93343f,0.93541f,0.91876f,0.92903f,0.93078f,0.92780f,0.92618f,0.92650f,0.93270f,0.92454f,0.92463f,0.92535f,0.92706f,0.92698f,0.92570f,0.92552f,0.92827f,0.92502f,0.92872f,0.92655f,0.92948f,0.92637f,0.92837f,0.93157f,0.93387f,0.93925f,0.94317f,0.96526f,0.00000f,0.90723f,0.91241f,0.90989f,0.92233f,0.92254f,0.91279f,0.92477f,0.92512f,0.92201f,0.92194f,0.92229f,0.92174f,0.92106f,0.92300f,0.91726f,0.92354f,0.92307f,0.92226f,0.92396f,0.92336f,0.92196f,0.92264f,0.92119f,0.92371f,0.91961f,0.91631f,0.92322f,0.92603f,0.92660f,0.92932f,0.93241f,0.95788f,0.00000f,0.88638f,0.89281f,0.89628f,0.90406f,0.90641f,0.90873f,0.90926f,0.90610f,0.90973f,0.91193f,0.91597f,0.90886f,0.91759f,0.91700f,0.91039f,0.91671f,0.91819f,0.91434f,0.92319f,0.91717f,0.91852f,0.91631f,0.91965f,0.91662f,0.91484f,0.91138f,0.91369f,0.91203f,0.91732f,0.92017f,0.92294f,0.95058f,0.00000f,0.86602f,0.87155f,0.87595f,0.87944f,0.88527f,0.88732f,0.88867f,0.89387f,0.89656f,0.90438f,0.90483f,0.90296f,0.89760f,0.90669f,0.90493f,0.90880f,0.90728f,0.91146f,0.91154f,0.91072f,0.91579f,0.91279f,0.90664f,0.90956f,0.90684f,0.90887f,0.90298f,0.90597f,0.90691f,0.91067f,0.91524f,0.94152f,0.00000f,0.83794f,0.84271f,0.85429f,0.86092f,0.86522f,0.87060f,0.87214f,0.87529f,0.87567f,0.88654f,0.89187f,0.89079f,0.89498f,0.89616f,0.89542f,0.89833f,0.89538f,0.89305f,0.90471f,0.90182f,0.90049f,0.90273f,0.90189f,0.90585f,0.89989f,0.89966f,0.89779f,0.89751f,0.90002f,0.90213f,0.90368f,0.93269f,0.00000f,0.80584f,0.81593f,0.82776f,0.83844f,0.84834f,0.84990f,0.85409f,0.85912f,0.85975f,0.86957f,0.86894f,0.86982f,0.88012f,0.87650f,0.88300f,0.88478f,0.88829f,0.88610f,0.88999f,0.89024f,0.89298f,0.89061f,0.89521f,0.89351f,0.89169f,0.88902f,0.89129f,0.88518f,0.89261f,0.89186f,0.89741f,0.92461f,
 0.00000f,0.99596f,0.98965f,0.97936f,0.95694f,0.96017f,0.95925f,0.97032f,0.95026f,0.95424f,0.90631f,0.95021f,0.95663f,0.94802f,0.94194f,0.93410f,0.94562f,0.96628f,0.96652f,0.96847f,0.96713f,0.97446f,0.97422f,0.97268f,0.97750f,0.98110f,0.98305f,0.98313f,0.98207f,0.98921f,0.99394f,0.99739f,1.00000f,0.00000f,0.98982f,0.98365f,0.98050f,0.97408f,0.97309f,0.97192f,0.97167f,0.96970f,0.96776f,0.96422f,0.96818f,0.96908f,0.96712f,0.96843f,0.96872f,0.96896f,0.97381f,0.97406f,0.97586f,0.97770f,0.97823f,0.97978f,0.98100f,0.98327f,0.98493f,0.98685f,0.98873f,0.99042f,0.99271f,0.99518f,0.99758f,1.00000f,0.00000f,0.99151f,0.98603f,0.98146f,0.98011f,0.97691f,0.97614f,0.97443f,0.97398f,0.97426f,0.97348f,0.97365f,0.97379f,0.97300f,0.97437f,0.97374f,0.97471f,0.97591f,0.97566f,0.97826f,0.97902f,0.97985f,0.98122f,0.98249f,0.98462f,0.98577f,0.98759f,0.98927f,0.99147f,0.99337f,0.99550f,0.99776f,1.00000f,0.00000f,0.98221f,0.98606f,0.98282f,0.97907f,0.97724f,0.97642f,0.97582f,0.97391f,0.97399f,0.97392f,0.97404f,0.97224f,0.97404f,0.97434f,0.97381f,0.97480f,0.97576f,0.97680f,0.97779f,0.97834f,0.97977f,0.98125f,0.98251f,0.98445f,0.98604f,0.98758f,0.98957f,0.99117f,0.99336f,0.99550f,0.99776f,1.00000f,0.00000f,0.93991f,0.98245f,0.97946f,0.97745f,0.97829f,0.97585f,0.97496f,0.97349f,0.97403f,0.97378f,0.97359f,0.97288f,0.97346f,0.97426f,0.97422f,0.97518f,0.97582f,0.97644f,0.97721f,0.97904f,0.97996f,0.98087f,0.98296f,0.98413f,0.98585f,0.98736f,0.98925f,0.99135f,0.99325f,0.99546f,0.99774f,1.00000f,0.00000f,0.91595f,0.96042f,0.97885f,0.97724f,0.97675f,0.97610f,0.97483f,0.97450f,0.97338f,0.97217f,0.97244f,0.97287f,0.97356f,0.97352f,0.97348f,0.97521f,0.97533f,0.97715f,0.97718f,0.97809f,0.97976f,0.98113f,0.98258f,0.98419f,0.98561f,0.98727f,0.98940f,0.99137f,0.99318f,0.99536f,0.99781f,1.00000f,0.00000f,0.90790f,0.92977f,0.96184f,0.97225f,0.97523f,0.97522f,0.97362f,0.97345f,0.97318f,0.97228f,0.97362f,0.97316f,0.97348f,0.97379f,0.97410f,0.97478f,0.97566f,0.97597f,0.97822f,0.97848f,0.98002f,0.98141f,0.98271f,0.98381f,0.98585f,0.98776f,0.98930f,0.99118f,0.99326f,0.99542f,0.99765f,1.00000f,0.00000f,0.90936f,0.91480f,0.93600f,0.95813f,0.96754f,0.97107f,0.97325f,0.97416f,0.97127f,0.97077f,0.97273f,0.97207f,0.97271f,0.97350f,0.97332f,0.97470f,0.97514f,0.97538f,0.97774f,0.97829f,0.97937f,0.98160f,0.98257f,0.98402f,0.98584f,0.98740f,0.98909f,0.99113f,0.99302f,0.99540f,0.99758f,1.00000f,0.00000f,0.91750f,0.90790f,0.92498f,0.93725f,0.95335f,0.96402f,0.96744f,0.97006f,0.97214f,0.97200f,0.97121f,0.97191f,0.97200f,0.97369f,0.97322f,0.97433f,0.97454f,0.97574f,0.97674f,0.97837f,0.97957f,0.98093f,0.98206f,0.98364f,0.98588f,0.98693f,0.98925f,0.99100f,0.99323f,0.99544f,0.99761f,1.00000f,0.00000f,0.92663f,0.91029f,0.91506f,0.92661f,0.93647f,0.95180f,0.95960f,0.96401f,0.96703f,0.96944f,0.97078f,0.97044f,0.97207f,0.97345f,0.97349f,0.97326f,0.97500f,0.97515f,0.97674f,0.97815f,0.97932f,0.98007f,0.98229f,0.98394f,0.98530f,0.98715f,0.98915f,0.99094f,0.99299f,0.99527f,0.99746f,1.00000f,0.00000f,0.93071f,0.91243f,0.91495f,0.91904f,0.92742f,0.93684f,0.94240f,0.95515f,0.96247f,0.96514f,0.96708f,0.96810f,0.97047f,0.97166f,0.97381f,0.97273f,0.97359f,0.97613f,0.97674f,0.97814f,0.97852f,0.98063f,0.98173f,0.98367f,0.98507f,0.98696f,0.98884f,0.99099f,0.99286f,0.99486f,0.99745f,1.00000f,0.00000f,0.94125f,0.91956f,0.91239f,0.91446f,0.91877f,0.92642f,0.93779f,0.94365f,0.95006f,0.95917f,0.96147f,0.96507f,0.96777f,0.97005f,0.97013f,0.97279f,0.97364f,0.97567f,0.97688f,0.97698f,0.97858f,0.98046f,0.98147f,0.98363f,0.98486f,0.98620f,0.98851f,0.99083f,0.99264f,0.99521f,0.99723f,1.00000f,0.00000f,0.94480f,0.92328f,0.91425f,0.91390f,0.91854f,0.92455f,0.92703f,0.93704f,0.94144f,0.94827f,0.95443f,0.96083f,0.96259f,0.96512f,0.96936f,0.96924f,0.97335f,0.97472f,0.97605f,0.97642f,0.97853f,0.97969f,0.98149f,0.98338f,0.98456f,0.98680f,0.98881f,0.99094f,0.99279f,0.99490f,0.99739f,1.00000f,0.00000f,0.94389f,0.93389f,0.92682f,0.91544f,0.91591f,0.92315f,0.92479f,0.92725f,0.93505f,0.93945f,0.94648f,0.95120f,0.95588f,0.95959f,0.96357f,0.96722f,0.96956f,0.97202f,0.97374f,0.97538f,0.97783f,0.97925f,0.98225f,0.98297f,0.98507f,0.98658f,0.98830f,0.99014f,0.99247f,0.99492f,0.99724f,1.00000f,0.00000f,0.94974f,0.93667f,0.92739f,0.91767f,0.91727f,0.92484f,0.92214f,0.92620f,0.93091f,0.93290f,0.93836f,0.94373f,0.94749f,0.95211f,0.95808f,0.96272f,0.96518f,0.96901f,0.97111f,0.97281f,0.97620f,0.97840f,0.98052f,0.98185f,0.98396f,0.98626f,0.98801f,0.99021f,0.99249f,0.99503f,0.99708f,1.00000f,0.00000f,0.95451f,0.94237f,0.94244f,0.91970f,0.91920f,0.92151f,0.91986f,0.92413f,0.92441f,0.93014f,0.93324f,0.93825f,0.94099f,0.94558f,0.94969f,0.95446f,0.95838f,0.96432f,0.96738f,0.97068f,0.97267f,0.97549f,0.97760f,0.97960f,0.98328f,0.98519f,0.98831f,0.99011f,0.99227f,0.99469f,0.99686f,1.00000f,0.00000f,0.96238f,0.94671f,0.93490f,0.92897f,0.92745f,0.92310f,0.92243f,0.92569f,0.92548f,0.93279f,0.93047f,0.93457f,0.93368f,0.94243f,0.94544f,0.94956f,0.95466f,0.95625f,0.96244f,0.96563f,0.96984f,0.97269f,0.97462f,0.97850f,0.98159f,0.98441f,0.98660f,0.98932f,0.99217f,0.99440f,0.99711f,1.00000f,0.00000f,0.95850f,0.95064f,0.94377f,0.93709f,0.92659f,0.92495f,0.92609f,0.92179f,0.92574f,0.92738f,0.92753f,0.93417f,0.93433f,0.93660f,0.94084f,0.94114f,0.94938f,0.95229f,0.95716f,0.96209f,0.96457f,0.96872f,0.97093f,0.97566f,0.97945f,0.98182f,0.98494f,0.98805f,0.99116f,0.99379f,0.99681f,1.00000f,0.00000f,0.95747f,0.95957f,0.94475f,0.93391f,0.92881f,0.92701f,0.92498f,0.92374f,0.92234f,0.92458f,0.92876f,0.92769f,0.93124f,0.93710f,0.93836f,0.93995f,0.94004f,0.94854f,0.95098f,0.95540f,0.95833f,0.96201f,0.96707f,0.96972f,0.97507f,0.97841f,0.98086f,0.98517f,0.98935f,0.99222f,0.99628f,1.00000f,0.00000f,0.96053f,0.95459f,0.94831f,0.93965f,0.93603f,0.93049f,0.93017f,0.92701f,0.92504f,0.92703f,0.93034f,0.93104f,0.93336f,0.93304f,0.93621f,0.93654f,0.93750f,0.94228f,0.94267f,0.94762f,0.95405f,0.95966f,0.96412f,0.96608f,0.97087f,0.97428f,0.97814f,0.98348f,0.98637f,0.99103f,0.99466f,0.99994f,0.00000f,0.96280f,0.95511f,0.95497f,0.94371f,0.93840f,0.93580f,0.92832f,0.92813f,0.92482f,0.92480f,0.92901f,0.93003f,0.92881f,0.93318f,0.93202f,0.93302f,0.93863f,0.93478f,0.93890f,0.94360f,0.94944f,0.95277f,0.95955f,0.96216f,0.96622f,0.96933f,0.97396f,0.97860f,0.98285f,0.98648f,0.99190f,0.99957f,0.00000f,0.95923f,0.95665f,0.95840f,0.94072f,0.94118f,0.93966f,0.93497f,0.93069f,0.93062f,0.92778f,0.93079f,0.92719f,0.93498f,0.92987f,0.93267f,0.93056f,0.93313f,0.93239f,0.93816f,0.94306f,0.94413f,0.94696f,0.95235f,0.95421f,0.96251f,0.96266f,0.96913f,0.97426f,0.97902f,0.98235f,0.98707f,0.99851f,0.00000f,0.96398f,0.96388f,0.95977f,0.94562f,0.94500f,0.94144f,0.93900f,0.93676f,0.93151f,0.92896f,0.92883f,0.92437f,0.92938f,0.93116f,0.92969f,0.93480f,0.93011f,0.93630f,0.93530f,0.93876f,0.94072f,0.94266f,0.94662f,0.95097f,0.95211f,0.95854f,0.96246f,0.96895f,0.97270f,0.97737f,0.98170f,0.99618f,0.00000f,0.95739f,0.96196f,0.95448f,0.96249f,0.94605f,0.94048f,0.93853f,0.93475f,0.93373f,0.93454f,0.93156f,0.93175f,0.92928f,0.92819f,0.92450f,0.93089f,0.93399f,0.93331f,0.93472f,0.93464f,0.93557f,0.93702f,0.93901f,0.94454f,0.94786f,0.95090f,0.95811f,0.96096f,0.96523f,0.97081f,0.97445f,0.99201f,0.00000f,0.95615f,0.95767f,0.95282f,0.94820f,0.94822f,0.94039f,0.94274f,0.93487f,0.93676f,0.93304f,0.93360f,0.92787f,0.92997f,0.92924f,0.92835f,0.92833f,0.93311f,0.93137f,0.93303f,0.93185f,0.93284f,0.93480f,0.93534f,0.93928f,0.94196f,0.94380f,0.94837f,0.95468f,0.95847f,0.96264f,0.96668f,0.98615f,0.00000f,0.94724f,0.94861f,0.94692f,0.93926f,0.94377f,0.94174f,0.94072f,0.94026f,0.93361f,0.93542f,0.92889f,0.93379f,0.93067f,0.92941f,0.92859f,0.93130f,0.93026f,0.92594f,0.93081f,0.93228f,0.93374f,0.93308f,0.93168f,0.93642f,0.93511f,0.93961f,0.94441f,0.94608f,0.94991f,0.95365f,0.95839f,0.97973f,0.00000f,0.93886f,0.94279f,0.93988f,0.94501f,0.93231f,0.93432f,0.93633f,0.93239f,0.93736f,0.93038f,0.92838f,0.93028f,0.92889f,0.92771f,0.92787f,0.92595f,0.92833f,0.92959f,0.92790f,0.92822f,0.92825f,0.92703f,0.92870f,0.92996f,0.92849f,0.93288f,0.93438f,0.93928f,0.94375f,0.94359f,0.94999f,0.97257f,0.00000f,0.92337f,0.92702f,0.93109f,0.93362f,0.93042f,0.93320f,0.93324f,0.93157f,0.92772f,0.92799f,0.92454f,0.92650f,0.92857f,0.92663f,0.93179f,0.92672f,0.92405f,0.92472f,0.92565f,0.92602f,0.92453f,0.92687f,0.92568f,0.92375f,0.92479f,0.92352f,0.92643f,0.93115f,0.93261f,0.93658f,0.93969f,0.96378f,0.00000f,0.91228f,0.91274f,0.91747f,0.92185f,0.92301f,0.92459f,0.91680f,0.92189f,0.92662f,0.92307f,0.92244f,0.92313f,0.92512f,0.92078f,0.92513f,0.92035f,0.92373f,0.92453f,0.92271f,0.92345f,0.91927f,0.92214f,0.92193f,0.91746f,0.91777f,0.91756f,0.92099f,0.91943f,0.92344f,0.92494f,0.92968f,0.95558f,0.00000f,0.89219f,0.89845f,0.90197f,0.90823f,0.90451f,0.90466f,0.90730f,0.91189f,0.91386f,0.91501f,0.91407f,0.91173f,0.91385f,0.91634f,0.91340f,0.92103f,0.91496f,0.91498f,0.91543f,0.91598f,0.91211f,0.91808f,0.91694f,0.91421f,0.91265f,0.91242f,0.91152f,0.91275f,0.91379f,0.91800f,0.92020f,0.94676f,0.00000f,0.85884f,0.87631f,0.87736f,0.88110f,0.89364f,0.88987f,0.89642f,0.89568f,0.89841f,0.89901f,0.90641f,0.90561f,0.90573f,0.90707f,0.90629f,0.91357f,0.90967f,0.91097f,0.91034f,0.91158f,0.91051f,0.90985f,0.90589f,0.90870f,0.90763f,0.90604f,0.90466f,0.90402f,0.90340f,0.90647f,0.91038f,0.93614f,0.00000f,0.83862f,0.85050f,0.85661f,0.86307f,0.86920f,0.87355f,0.87775f,0.88250f,0.88188f,0.88552f,0.88837f,0.89008f,0.89130f,0.89240f,0.89686f,0.89842f,0.89734f,0.89828f,0.90023f,0.90066f,0.89961f,0.90119f,0.90180f,0.89930f,0.89713f,0.89507f,0.89553f,0.89240f,0.89405f,0.89792f,0.90187f,0.92934f,0.00000f,0.82063f,0.82361f,0.82776f,0.84170f,0.84231f,0.85164f,0.85330f,0.86575f,0.87009f,0.86750f,0.87195f,0.87829f,0.88047f,0.88000f,0.88421f,0.88326f,0.89032f,0.88894f,0.88741f,0.89248f,0.88987f,0.89005f,0.89006f,0.88690f,0.88927f,0.88966f,0.88448f,0.88163f,0.88361f,0.88543f,0.89015f,0.92106f,
 0.00000f,0.99802f,0.98247f,0.98639f,0.95687f,0.96333f,0.96526f,0.95453f,0.95844f,0.95689f,0.93378f,0.95938f,0.95504f,0.95380f,0.95730f,0.95637f,0.95478f,0.95886f,0.94552f,0.95954f,0.96688f,0.96865f,0.97517f,0.97808f,0.97815f,0.97950f,0.98172f,0.98412f,0.98660f,0.99215f,0.99314f,0.99479f,1.00000f,0.00000f,0.99217f,0.98456f,0.97911f,0.97764f,0.97447f,0.97273f,0.96993f,0.97016f,0.96893f,0.96858f,0.96885f,0.96918f,0.96947f,0.97007f,0.97030f,0.96943f,0.97097f,0.97247f,0.97231f,0.97725f,0.97847f,0.97996f,0.98157f,0.98351f,0.98483f,0.98635f,0.98889f,0.99086f,0.99283f,0.99515f,0.99740f,1.00000f,0.00000f,0.99056f,0.98538f,0.98229f,0.98068f,0.97799f,0.97621f,0.97521f,0.97415f,0.97375f,0.97408f,0.97365f,0.97308f,0.97277f,0.97345f,0.97411f,0.97508f,0.97583f,0.97652f,0.97797f,0.97890f,0.98006f,0.98154f,0.98279f,0.98469f,0.98587f,0.98752f,0.98941f,0.99139f,0.99337f,0.99547f,0.99774f,1.00000f,0.00000f,0.98503f,0.98577f,0.98257f,0.97942f,0.97781f,0.97623f,0.97613f,0.97484f,0.97437f,0.97435f,0.97306f,0.97358f,0.97321f,0.97479f,0.97381f,0.97514f,0.97593f,0.97613f,0.97772f,0.97855f,0.98055f,0.98153f,0.98237f,0.98426f,0.98584f,0.98788f,0.98948f,0.99140f,0.99316f,0.99550f,0.99775f,1.00000f,0.00000f,0.94070f,0.98376f,0.98230f,0.97940f,0.97812f,0.97662f,0.97347f,0.97448f,0.97411f,0.97325f,0.97295f,0.97376f,0.97402f,0.97335f,0.97499f,0.97489f,0.97645f,0.97640f,0.97782f,0.97893f,0.97993f,0.98107f,0.98266f,0.98413f,0.98591f,0.98760f,0.98941f,0.99124f,0.99330f,0.99547f,0.99773f,1.00000f,0.00000f,0.91289f,0.95944f,0.97752f,0.97793f,0.97547f,0.97636f,0.97413f,0.97361f,0.97290f,0.97302f,0.97318f,0.97298f,0.97320f,0.97327f,0.97475f,0.97485f,0.97569f,0.97702f,0.97767f,0.97853f,0.98008f,0.98116f,0.98232f,0.98404f,0.98608f,0.98762f,0.98931f,0.99123f,0.99322f,0.99534f,0.99776f,1.00000f,0.00000f,0.90675f,0.93306f,0.96262f,0.97460f,0.97488f,0.97534f,0.97427f,0.97350f,0.97251f,0.97237f,0.97288f,0.97234f,0.97347f,0.97312f,0.97405f,0.97454f,0.97554f,0.97617f,0.97777f,0.97860f,0.98015f,0.98080f,0.98249f,0.98439f,0.98557f,0.98720f,0.98949f,0.99118f,0.99319f,0.99528f,0.99766f,1.00000f,0.00000f,0.90962f,0.91560f,0.93684f,0.95749f,0.96816f,0.97155f,0.97257f,0.97258f,0.97285f,0.97321f,0.97131f,0.97175f,0.97193f,0.97398f,0.97425f,0.97445f,0.97564f,0.97647f,0.97739f,0.97838f,0.97970f,0.98072f,0.98235f,0.98424f,0.98558f,0.98747f,0.98924f,0.99121f,0.99324f,0.99543f,0.99751f,1.00000f,0.00000f,0.91476f,0.91383f,0.92263f,0.93914f,0.95442f,0.96269f,0.96939f,0.97017f,0.97203f,0.97193f,0.97195f,0.97141f,0.97252f,0.97278f,0.97280f,0.97463f,0.97482f,0.97525f,0.97696f,0.97840f,0.97911f,0.98072f,0.98229f,0.98401f,0.98534f,0.98697f,0.98897f,0.99119f,0.99322f,0.99532f,0.99757f,1.00000f,0.00000f,0.92597f,0.90491f,0.91381f,0.92557f,0.93947f,0.95061f,0.96042f,0.96470f,0.96882f,0.96975f,0.97120f,0.97155f,0.97242f,0.97253f,0.97371f,0.97401f,0.97539f,0.97577f,0.97719f,0.97785f,0.97914f,0.98111f,0.98229f,0.98368f,0.98577f,0.98703f,0.98885f,0.99129f,0.99295f,0.99529f,0.99755f,1.00000f,0.00000f,0.93279f,0.91221f,0.91499f,0.91938f,0.92481f,0.93551f,0.94741f,0.95472f,0.96232f,0.96470f,0.96824f,0.97047f,0.97160f,0.97061f,0.97289f,0.97222f,0.97516f,0.97508f,0.97600f,0.97755f,0.97885f,0.98033f,0.98203f,0.98356f,0.98497f,0.98691f,0.98847f,0.99092f,0.99303f,0.99516f,0.99737f,1.00000f,0.00000f,0.93816f,0.91957f,0.91321f,0.91278f,0.92268f,0.92748f,0.93612f,0.94409f,0.95058f,0.95820f,0.96075f,0.96562f,0.96924f,0.96969f,0.97046f,0.97241f,0.97386f,0.97585f,0.97648f,0.97782f,0.97956f,0.97993f,0.98133f,0.98319f,0.98444f,0.98685f,0.98889f,0.99092f,0.99274f,0.99514f,0.99739f,1.00000f,0.00000f,0.94935f,0.92894f,0.91534f,0.91547f,0.92112f,0.92581f,0.92776f,0.93634f,0.94202f,0.94717f,0.95560f,0.95919f,0.96304f,0.96538f,0.96769f,0.97088f,0.97180f,0.97365f,0.97535f,0.97659f,0.97795f,0.97932f,0.98200f,0.98302f,0.98478f,0.98633f,0.98830f,0.99058f,0.99289f,0.99500f,0.99740f,1.00000f,0.00000f,0.95127f,0.93316f,0.92503f,0.91925f,0.91906f,0.92118f,0.92472f,0.92485f,0.93165f,0.93975f,0.94603f,0.95154f,0.95496f,0.95927f,0.96411f,0.96570f,0.96941f,0.97167f,0.97457f,0.97646f,0.97712f,0.98006f,0.98107f,0.98324f,0.98458f,0.98650f,0.98849f,0.99063f,0.99228f,0.99490f,0.99722f,1.00000f,0.00000f,0.94931f,0.93951f,0.92515f,0.92636f,0.91621f,0.91518f,0.92367f,0.92272f,0.93114f,0.93633f,0.93914f,0.94370f,0.94780f,0.95474f,0.95775f,0.96126f,0.96413f,0.96926f,0.97033f,0.97272f,0.97628f,0.97708f,0.98017f,0.98146f,0.98286f,0.98571f,0.98845f,0.99041f,0.99253f,0.99459f,0.99708f,1.00000f,0.00000f,0.95880f,0.94144f,0.93298f,0.92249f,0.92794f,0.92370f,0.92598f,0.92150f,0.92684f,0.92980f,0.93415f,0.93894f,0.94031f,0.94602f,0.95038f,0.95405f,0.95972f,0.96344f,0.96591f,0.96975f,0.97321f,0.97469f,0.97836f,0.98187f,0.98354f,0.98529f,0.98779f,0.99034f,0.99254f,0.99454f,0.99688f,1.00000f,0.00000f,0.95905f,0.94729f,0.93887f,0.93248f,0.92640f,0.92285f,0.92128f,0.92540f,0.92538f,0.92459f,0.93168f,0.93479f,0.93442f,0.94041f,0.94466f,0.95038f,0.95386f,0.95887f,0.96119f,0.96495f,0.97036f,0.97206f,0.97486f,0.97954f,0.98121f,0.98366f,0.98627f,0.98916f,0.99191f,0.99479f,0.99688f,1.00000f,0.00000f,0.95277f,0.94759f,0.93725f,0.93963f,0.92673f,0.92215f,0.92483f,0.92436f,0.92401f,0.92530f,0.92722f,0.92847f,0.93770f,0.93763f,0.93797f,0.94403f,0.94557f,0.95197f,0.95523f,0.95970f,0.96381f,0.96820f,0.97082f,0.97427f,0.97798f,0.98105f,0.98447f,0.98795f,0.99073f,0.99369f,0.99668f,1.00000f,0.00000f,0.96042f,0.95529f,0.93807f,0.94744f,0.93104f,0.92828f,0.92281f,0.92219f,0.92218f,0.92335f,0.92702f,0.92857f,0.93043f,0.93569f,0.93630f,0.93800f,0.94223f,0.94615f,0.95013f,0.95455f,0.96000f,0.96218f,0.96618f,0.97029f,0.97443f,0.97915f,0.98140f,0.98624f,0.98884f,0.99279f,0.99614f,1.00000f,0.00000f,0.96180f,0.96277f,0.94437f,0.94078f,0.93733f,0.93270f,0.92663f,0.92721f,0.92281f,0.92617f,0.92285f,0.93375f,0.92886f,0.93228f,0.93656f,0.93865f,0.94054f,0.94333f,0.94502f,0.94698f,0.95354f,0.95687f,0.96234f,0.96649f,0.97074f,0.97417f,0.97762f,0.98325f,0.98540f,0.99078f,0.99435f,0.99993f,0.00000f,0.95800f,0.96023f,0.95446f,0.95027f,0.94243f,0.93531f,0.93011f,0.92841f,0.92705f,0.92497f,0.92367f,0.93067f,0.93056f,0.92663f,0.93278f,0.93508f,0.93852f,0.93696f,0.94058f,0.94255f,0.94631f,0.94991f,0.95688f,0.96133f,0.96626f,0.97010f,0.97401f,0.97725f,0.98314f,0.98690f,0.99107f,0.99955f,0.00000f,0.96279f,0.95888f,0.95088f,0.94814f,0.94109f,0.94119f,0.93198f,0.93073f,0.92885f,0.92524f,0.92866f,0.92665f,0.93219f,0.92854f,0.93407f,0.93272f,0.93133f,0.93619f,0.93718f,0.93973f,0.94439f,0.94702f,0.95055f,0.95295f,0.96010f,0.96473f,0.96845f,0.97355f,0.97682f,0.98156f,0.98664f,0.99834f,0.00000f,0.95637f,0.96118f,0.95498f,0.95324f,0.94858f,0.94169f,0.93961f,0.93643f,0.93445f,0.92860f,0.93142f,0.92836f,0.92925f,0.92952f,0.92882f,0.93109f,0.93435f,0.93188f,0.93528f,0.93579f,0.94048f,0.94203f,0.94541f,0.94719f,0.95232f,0.95850f,0.96147f,0.96679f,0.97077f,0.97571f,0.98149f,0.99564f,0.00000f,0.95733f,0.95655f,0.95632f,0.95734f,0.95165f,0.94087f,0.93840f,0.94090f,0.93557f,0.93170f,0.93328f,0.92935f,0.92938f,0.93194f,0.93135f,0.93241f,0.93103f,0.93148f,0.93264f,0.93446f,0.93867f,0.93880f,0.94016f,0.94137f,0.94547f,0.95169f,0.95699f,0.96033f,0.96588f,0.96935f,0.97407f,0.99150f,0.00000f,0.95201f,0.96020f,0.94950f,0.95717f,0.94716f,0.94534f,0.94599f,0.93656f,0.93870f,0.93896f,0.93631f,0.93299f,0.92917f,0.92834f,0.92723f,0.92621f,0.92629f,0.93200f,0.93217f,0.93183f,0.93272f,0.93176f,0.93672f,0.93967f,0.94245f,0.94567f,0.94779f,0.95433f,0.95739f,0.96191f,0.96495f,0.98566f,0.00000f,0.94824f,0.95370f,0.95192f,0.94713f,0.94908f,0.94718f,0.94246f,0.93317f,0.93657f,0.93178f,0.93372f,0.93074f,0.92972f,0.92824f,0.92914f,0.93062f,0.93077f,0.92669f,0.93106f,0.93138f,0.93226f,0.93400f,0.93105f,0.93236f,0.93253f,0.93775f,0.94079f,0.94646f,0.94840f,0.95340f,0.95543f,0.97829f,0.00000f,0.94079f,0.94269f,0.94750f,0.94720f,0.93861f,0.93661f,0.93882f,0.93769f,0.93264f,0.93058f,0.93032f,0.93104f,0.93233f,0.92589f,0.92829f,0.93331f,0.92925f,0.92674f,0.92611f,0.92655f,0.92530f,0.92879f,0.92915f,0.93005f,0.92856f,0.93000f,0.93497f,0.93783f,0.93790f,0.94443f,0.94775f,0.97019f,0.00000f,0.92705f,0.93405f,0.93606f,0.93761f,0.93511f,0.93103f,0.92925f,0.92803f,0.93668f,0.92503f,0.92660f,0.92414f,0.92531f,0.93202f,0.92520f,0.92753f,0.92049f,0.92635f,0.92434f,0.92363f,0.92497f,0.92370f,0.92065f,0.92244f,0.92179f,0.92295f,0.92602f,0.92826f,0.93110f,0.93293f,0.93375f,0.96132f,0.00000f,0.91286f,0.91419f,0.91649f,0.92350f,0.92302f,0.92190f,0.92367f,0.93101f,0.92642f,0.92269f,0.92659f,0.92192f,0.91877f,0.92055f,0.92262f,0.92209f,0.91636f,0.91939f,0.92168f,0.91960f,0.92101f,0.91611f,0.91689f,0.91993f,0.91776f,0.91804f,0.92035f,0.91644f,0.92041f,0.92382f,0.92755f,0.95165f,0.00000f,0.88830f,0.90094f,0.90579f,0.90836f,0.90384f,0.91160f,0.90850f,0.91719f,0.91348f,0.91442f,0.91699f,0.91300f,0.91306f,0.91132f,0.91332f,0.91515f,0.91411f,0.91421f,0.91238f,0.91228f,0.91432f,0.91477f,0.91476f,0.91158f,0.91130f,0.90837f,0.90654f,0.90839f,0.90972f,0.91321f,0.91830f,0.94219f,0.00000f,0.87247f,0.87559f,0.88158f,0.88967f,0.89460f,0.89816f,0.90054f,0.89877f,0.90041f,0.90532f,0.90258f,0.90616f,0.90081f,0.90451f,0.90924f,0.90741f,0.90953f,0.90537f,0.90739f,0.90670f,0.90131f,0.90832f,0.90603f,0.90371f,0.90564f,0.90250f,0.89885f,0.89843f,0.89955f,0.90103f,0.90674f,0.93413f,0.00000f,0.84930f,0.85289f,0.86102f,0.86781f,0.87400f,0.87870f,0.87937f,0.88561f,0.88912f,0.88546f,0.88757f,0.88841f,0.89216f,0.89692f,0.90072f,0.89255f,0.89683f,0.90098f,0.89642f,0.89948f,0.89899f,0.89656f,0.90015f,0.89479f,0.89420f,0.88929f,0.89070f,0.88839f,0.88758f,0.89237f,0.89727f,0.92558f,0.00000f,0.81947f,0.82903f,0.84060f,0.84573f,0.84862f,0.85812f,0.85310f,0.85590f,0.86620f,0.87039f,0.87774f,0.87986f,0.88218f,0.87838f,0.88412f,0.87919f,0.88496f,0.88965f,0.88399f,0.88918f,0.88798f,0.88724f,0.88515f,0.88604f,0.88592f,0.88089f,0.88193f,0.87550f,0.88041f,0.88336f,0.88622f,0.91648f,
 0.00000f,0.99706f,0.98921f,0.97579f,0.95907f,0.95936f,0.97052f,0.97034f,0.96473f,0.93555f,0.93600f,0.95177f,0.94789f,0.95773f,0.95106f,0.95956f,0.95328f,0.95986f,0.95921f,0.96090f,0.96408f,0.94791f,0.97415f,0.97846f,0.98107f,0.98260f,0.98583f,0.98327f,0.98545f,0.99209f,0.99328f,0.99746f,1.00000f,0.00000f,0.99021f,0.98485f,0.97945f,0.97681f,0.97361f,0.97213f,0.97109f,0.97038f,0.96854f,0.96914f,0.96979f,0.96957f,0.96954f,0.96948f,0.97154f,0.97221f,0.97223f,0.97304f,0.97482f,0.97560f,0.97598f,0.97882f,0.98156f,0.98342f,0.98502f,0.98691f,0.98882f,0.99089f,0.99294f,0.99515f,0.99770f,1.00000f,0.00000f,0.99017f,0.98505f,0.98386f,0.97995f,0.97729f,0.97693f,0.97582f,0.97429f,0.97364f,0.97205f,0.97352f,0.97331f,0.97295f,0.97358f,0.97419f,0.97552f,0.97556f,0.97711f,0.97724f,0.97842f,0.97967f,0.98074f,0.98320f,0.98439f,0.98584f,0.98786f,0.98963f,0.99121f,0.99328f,0.99551f,0.99774f,1.00000f,0.00000f,0.98300f,0.98514f,0.98209f,0.97920f,0.97811f,0.97626f,0.97615f,0.97417f,0.97388f,0.97440f,0.97308f,0.97340f,0.97310f,0.97418f,0.97422f,0.97602f,0.97554f,0.97639f,0.97747f,0.97874f,0.98008f,0.98136f,0.98251f,0.98428f,0.98611f,0.98788f,0.98957f,0.99138f,0.99319f,0.99547f,0.99772f,1.00000f,0.00000f,0.94189f,0.98214f,0.98039f,0.97777f,0.97680f,0.97551f,0.97496f,0.97393f,0.97300f,0.97380f,0.97375f,0.97409f,0.97303f,0.97361f,0.97450f,0.97433f,0.97439f,0.97652f,0.97756f,0.97805f,0.98005f,0.98130f,0.98284f,0.98451f,0.98605f,0.98767f,0.98936f,0.99141f,0.99334f,0.99551f,0.99776f,1.00000f,0.00000f,0.91670f,0.95931f,0.97698f,0.97757f,0.97637f,0.97537f,0.97421f,0.97388f,0.97456f,0.97317f,0.97359f,0.97248f,0.97341f,0.97354f,0.97452f,0.97472f,0.97521f,0.97627f,0.97826f,0.97847f,0.98013f,0.98126f,0.98271f,0.98424f,0.98611f,0.98722f,0.98923f,0.99139f,0.99338f,0.99536f,0.99774f,1.00000f,0.00000f,0.90565f,0.93208f,0.95951f,0.97153f,0.97689f,0.97545f,0.97311f,0.97428f,0.97251f,0.97213f,0.97287f,0.97286f,0.97275f,0.97397f,0.97354f,0.97439f,0.97490f,0.97622f,0.97770f,0.97854f,0.98001f,0.98116f,0.98280f,0.98413f,0.98582f,0.98758f,0.98924f,0.99138f,0.99317f,0.99536f,0.99761f,1.00000f,0.00000f,0.90859f,0.91651f,0.93710f,0.95565f,0.96808f,0.97148f,0.97319f,0.97254f,0.97258f,0.97314f,0.97167f,0.97144f,0.97265f,0.97268f,0.97298f,0.97384f,0.97510f,0.97577f,0.97739f,0.97842f,0.97955f,0.98083f,0.98218f,0.98404f,0.98537f,0.98740f,0.98934f,0.99132f,0.99326f,0.99530f,0.99757f,1.00000f,0.00000f,0.91294f,0.90816f,0.92578f,0.93643f,0.95532f,0.96139f,0.96915f,0.97052f,0.97309f,0.97208f,0.97118f,0.97298f,0.97152f,0.97300f,0.97326f,0.97431f,0.97542f,0.97570f,0.97677f,0.97782f,0.97914f,0.98064f,0.98196f,0.98356f,0.98536f,0.98714f,0.98892f,0.99097f,0.99307f,0.99532f,0.99756f,1.00000f,0.00000f,0.92481f,0.90991f,0.91537f,0.92513f,0.93713f,0.95146f,0.95985f,0.96512f,0.96855f,0.96970f,0.97092f,0.97261f,0.97219f,0.97309f,0.97229f,0.97390f,0.97446f,0.97567f,0.97718f,0.97760f,0.97907f,0.98066f,0.98250f,0.98403f,0.98539f,0.98747f,0.98868f,0.99100f,0.99296f,0.99514f,0.99753f,1.00000f,0.00000f,0.92875f,0.91270f,0.91933f,0.91942f,0.92635f,0.93652f,0.94459f,0.95405f,0.95981f,0.96389f,0.96687f,0.97007f,0.97057f,0.97135f,0.97260f,0.97338f,0.97464f,0.97442f,0.97604f,0.97774f,0.97906f,0.97995f,0.98203f,0.98357f,0.98488f,0.98692f,0.98874f,0.99096f,0.99274f,0.99535f,0.99748f,1.00000f,0.00000f,0.93903f,0.92225f,0.91301f,0.91415f,0.92284f,0.92787f,0.93436f,0.94303f,0.94971f,0.95671f,0.96250f,0.96441f,0.96831f,0.97026f,0.97127f,0.97286f,0.97358f,0.97507f,0.97612f,0.97789f,0.97965f,0.98016f,0.98137f,0.98368f,0.98494f,0.98659f,0.98882f,0.99080f,0.99274f,0.99491f,0.99753f,1.00000f,0.00000f,0.94862f,0.92729f,0.91243f,0.91306f,0.91708f,0.92246f,0.92777f,0.93449f,0.94171f,0.94664f,0.95360f,0.95753f,0.96158f,0.96578f,0.96780f,0.97040f,0.97094f,0.97512f,0.97654f,0.97746f,0.97881f,0.97968f,0.98187f,0.98269f,0.98477f,0.98667f,0.98897f,0.99052f,0.99301f,0.99499f,0.99741f,1.00000f,0.00000f,0.94822f,0.93063f,0.92260f,0.91575f,0.91907f,0.91943f,0.92164f,0.93174f,0.93391f,0.93949f,0.94711f,0.95133f,0.95640f,0.96157f,0.96365f,0.96529f,0.96902f,0.97173f,0.97447f,0.97518f,0.97784f,0.98003f,0.98025f,0.98280f,0.98441f,0.98541f,0.98855f,0.99042f,0.99273f,0.99464f,0.99739f,1.00000f,0.00000f,0.95258f,0.93525f,0.92560f,0.91860f,0.91612f,0.92002f,0.91814f,0.92597f,0.93043f,0.93332f,0.93907f,0.94450f,0.95082f,0.95381f,0.95674f,0.96201f,0.96500f,0.96737f,0.97158f,0.97390f,0.97576f,0.97822f,0.98031f,0.98276f,0.98357f,0.98630f,0.98798f,0.99053f,0.99278f,0.99474f,0.99727f,1.00000f,0.00000f,0.95857f,0.94386f,0.93308f,0.92544f,0.92206f,0.92097f,0.91879f,0.92566f,0.93066f,0.93165f,0.93262f,0.93568f,0.94147f,0.94745f,0.95232f,0.95492f,0.95870f,0.96277f,0.96579f,0.97009f,0.97247f,0.97505f,0.97724f,0.97974f,0.98371f,0.98591f,0.98761f,0.99026f,0.99248f,0.99451f,0.99710f,1.00000f,0.00000f,0.95385f,0.94124f,0.93706f,0.92762f,0.92834f,0.92284f,0.92412f,0.92145f,0.92230f,0.93054f,0.92944f,0.93174f,0.93997f,0.93956f,0.94528f,0.94803f,0.95155f,0.95727f,0.96115f,0.96410f,0.96909f,0.97280f,0.97680f,0.97782f,0.98080f,0.98425f,0.98671f,0.98944f,0.99169f,0.99413f,0.99702f,1.00000f,0.00000f,0.95399f,0.94503f,0.93650f,0.93830f,0.92856f,0.92675f,0.92626f,0.92339f,0.92624f,0.92100f,0.92975f,0.93171f,0.93433f,0.93714f,0.93946f,0.94269f,0.94996f,0.95232f,0.95543f,0.95931f,0.96558f,0.96825f,0.97021f,0.97398f,0.97842f,0.98094f,0.98396f,0.98838f,0.99111f,0.99367f,0.99696f,1.00000f,0.00000f,0.95620f,0.95603f,0.95485f,0.93807f,0.93070f,0.92690f,0.92785f,0.92530f,0.91978f,0.92349f,0.92609f,0.92820f,0.92877f,0.93433f,0.93413f,0.93772f,0.94112f,0.94683f,0.94983f,0.95584f,0.96071f,0.96306f,0.96789f,0.97076f,0.97326f,0.97727f,0.98108f,0.98577f,0.98882f,0.99235f,0.99585f,1.00000f,0.00000f,0.96042f,0.95113f,0.94515f,0.93986f,0.93902f,0.93077f,0.92767f,0.92652f,0.92887f,0.93016f,0.93044f,0.92948f,0.92815f,0.93487f,0.93141f,0.93476f,0.93853f,0.94137f,0.94405f,0.94885f,0.95469f,0.95763f,0.96065f,0.96573f,0.96909f,0.97264f,0.97824f,0.98203f,0.98633f,0.99069f,0.99395f,0.99992f,0.00000f,0.96404f,0.95509f,0.95208f,0.94601f,0.94369f,0.93637f,0.93295f,0.92640f,0.93050f,0.93008f,0.92810f,0.92685f,0.92906f,0.92847f,0.93222f,0.93332f,0.93280f,0.93899f,0.94231f,0.94304f,0.94846f,0.95309f,0.95717f,0.96018f,0.96477f,0.96916f,0.97300f,0.97659f,0.98121f,0.98638f,0.99121f,0.99955f,0.00000f,0.96363f,0.96338f,0.95459f,0.95393f,0.94434f,0.94427f,0.94146f,0.93762f,0.92922f,0.92778f,0.92550f,0.93312f,0.92912f,0.93105f,0.93428f,0.93331f,0.93467f,0.93280f,0.93698f,0.93826f,0.94238f,0.94534f,0.94971f,0.95564f,0.95865f,0.96261f,0.96771f,0.97222f,0.97742f,0.98168f,0.98616f,0.99823f,0.00000f,0.96362f,0.95145f,0.96309f,0.95409f,0.94816f,0.94592f,0.93843f,0.93643f,0.93413f,0.93194f,0.92910f,0.92758f,0.92553f,0.93119f,0.92869f,0.93097f,0.92854f,0.93269f,0.93061f,0.93719f,0.93665f,0.93931f,0.94516f,0.94979f,0.95183f,0.95737f,0.96197f,0.96601f,0.97149f,0.97607f,0.97970f,0.99543f,0.00000f,0.95845f,0.95997f,0.96071f,0.95355f,0.95136f,0.93930f,0.93878f,0.93900f,0.93528f,0.93196f,0.93170f,0.93003f,0.93058f,0.92898f,0.93279f,0.93116f,0.93647f,0.93632f,0.92904f,0.93373f,0.93459f,0.93676f,0.93984f,0.94158f,0.94617f,0.94915f,0.95552f,0.95713f,0.96430f,0.96937f,0.97340f,0.99081f,0.00000f,0.95478f,0.95542f,0.95836f,0.95270f,0.94684f,0.94279f,0.93853f,0.93709f,0.93453f,0.93428f,0.93244f,0.92953f,0.92624f,0.93289f,0.93160f,0.92673f,0.92850f,0.93005f,0.93282f,0.93393f,0.93290f,0.93037f,0.93298f,0.93598f,0.93907f,0.94283f,0.95039f,0.95124f,0.95602f,0.96099f,0.96374f,0.98414f,0.00000f,0.95328f,0.95290f,0.95472f,0.94899f,0.95005f,0.93814f,0.94407f,0.93808f,0.93730f,0.93743f,0.93322f,0.92938f,0.92648f,0.93100f,0.93080f,0.92962f,0.92804f,0.92703f,0.92965f,0.92868f,0.93229f,0.92984f,0.92919f,0.93108f,0.93512f,0.93557f,0.94030f,0.94331f,0.94758f,0.95180f,0.95416f,0.97752f,0.00000f,0.94555f,0.94293f,0.94397f,0.94751f,0.94321f,0.94353f,0.93888f,0.93533f,0.93496f,0.92715f,0.93270f,0.92945f,0.92579f,0.92634f,0.92896f,0.92633f,0.92788f,0.92463f,0.92544f,0.92389f,0.92465f,0.92713f,0.92728f,0.92804f,0.92629f,0.92821f,0.93064f,0.93474f,0.93937f,0.94099f,0.94394f,0.96856f,0.00000f,0.92848f,0.93740f,0.93809f,0.93866f,0.93563f,0.93288f,0.93444f,0.93145f,0.93450f,0.93027f,0.93331f,0.92313f,0.92977f,0.92870f,0.92386f,0.92467f,0.92435f,0.92901f,0.92121f,0.92331f,0.92586f,0.92226f,0.92240f,0.92106f,0.91876f,0.92303f,0.92397f,0.92560f,0.92793f,0.92891f,0.93160f,0.95928f,0.00000f,0.92200f,0.91957f,0.92279f,0.92582f,0.92790f,0.92351f,0.92016f,0.92287f,0.92149f,0.92560f,0.92292f,0.92226f,0.91954f,0.91792f,0.92029f,0.92321f,0.92013f,0.92149f,0.91739f,0.92182f,0.92022f,0.91856f,0.91648f,0.91423f,0.91629f,0.91327f,0.91393f,0.91344f,0.91426f,0.91784f,0.92056f,0.94956f,0.00000f,0.89979f,0.90652f,0.90576f,0.91004f,0.91195f,0.90937f,0.91290f,0.91196f,0.91636f,0.91478f,0.91562f,0.91765f,0.91907f,0.91396f,0.91484f,0.91138f,0.91542f,0.91147f,0.91068f,0.91525f,0.91220f,0.90766f,0.91066f,0.90983f,0.90738f,0.90801f,0.90701f,0.90767f,0.90782f,0.91137f,0.91232f,0.94064f,0.00000f,0.87135f,0.88146f,0.88127f,0.88573f,0.89454f,0.89666f,0.89696f,0.89884f,0.89890f,0.90314f,0.90332f,0.90759f,0.90615f,0.90689f,0.90341f,0.90206f,0.90546f,0.90354f,0.90352f,0.90560f,0.90293f,0.90611f,0.90117f,0.90253f,0.89707f,0.89671f,0.89802f,0.89674f,0.89688f,0.89900f,0.89952f,0.93047f,0.00000f,0.85244f,0.86013f,0.86080f,0.87564f,0.88006f,0.87909f,0.88179f,0.88548f,0.88592f,0.88913f,0.89235f,0.89254f,0.89738f,0.89639f,0.89632f,0.89457f,0.89957f,0.89737f,0.89863f,0.89300f,0.89331f,0.89385f,0.89311f,0.89251f,0.89386f,0.88706f,0.88797f,0.88717f,0.88577f,0.88961f,0.89117f,0.91899f,0.00000f,0.82382f,0.83813f,0.83878f,0.84978f,0.85628f,0.85771f,0.86430f,0.86837f,0.87207f,0.87720f,0.87957f,0.87991f,0.87866f,0.88346f,0.88277f,0.88007f,0.88511f,0.88617f,0.88401f,0.88603f,0.88625f,0.88356f,0.88213f,0.88350f,0.88255f,0.87982f,0.88044f,0.87724f,0.87537f,0.87987f,0.88138f,0.91130f,
 0.00000f,0.99228f,0.97303f,0.98530f,0.95771f,0.95014f,0.97363f,0.95999f,0.96705f,0.93793f,0.93937f,0.95704f,0.94752f,0.95663f,0.94976f,0.96949f,0.95702f,0.95408f,0.95916f,0.96407f,0.93633f,0.96940f,0.94803f,0.95683f,0.97756f,0.98034f,0.98487f,0.98627f,0.98788f,0.99300f,0.99380f,0.99744f,1.00000f,0.00000f,0.98895f,0.98264f,0.98127f,0.97572f,0.97487f,0.97308f,0.97146f,0.97141f,0.96889f,0.96987f,0.96936f,0.97009f,0.97014f,0.96990f,0.97142f,0.97168f,0.97290f,0.97388f,0.97545f,0.97403f,0.97749f,0.97780f,0.97919f,0.98199f,0.98307f,0.98728f,0.98892f,0.99081f,0.99320f,0.99506f,0.99762f,1.00000f,0.00000f,0.99125f,0.98637f,0.98303f,0.98003f,0.97755f,0.97609f,0.97613f,0.97423f,0.97390f,0.97288f,0.97382f,0.97402f,0.97355f,0.97331f,0.97494f,0.97529f,0.97546f,0.97645f,0.97826f,0.97887f,0.97970f,0.98118f,0.98266f,0.98442f,0.98581f,0.98766f,0.98955f,0.99135f,0.99327f,0.99548f,0.99773f,1.00000f,0.00000f,0.98411f,0.98506f,0.98256f,0.97887f,0.97724f,0.97695f,0.97569f,0.97388f,0.97402f,0.97343f,0.97329f,0.97302f,0.97291f,0.97357f,0.97456f,0.97524f,0.97607f,0.97645f,0.97753f,0.97911f,0.97968f,0.98101f,0.98281f,0.98425f,0.98582f,0.98751f,0.98932f,0.99149f,0.99323f,0.99540f,0.99780f,1.00000f,0.00000f,0.94255f,0.98099f,0.98122f,0.97939f,0.97733f,0.97629f,0.97515f,0.97352f,0.97416f,0.97278f,0.97241f,0.97339f,0.97394f,0.97416f,0.97383f,0.97480f,0.97538f,0.97647f,0.97808f,0.97901f,0.97984f,0.98143f,0.98290f,0.98411f,0.98574f,0.98761f,0.98932f,0.99138f,0.99340f,0.99538f,0.99772f,1.00000f,0.00000f,0.91770f,0.96141f,0.97749f,0.97793f,0.97794f,0.97519f,0.97432f,0.97368f,0.97256f,0.97379f,0.97237f,0.97294f,0.97328f,0.97367f,0.97387f,0.97443f,0.97486f,0.97641f,0.97728f,0.97902f,0.97965f,0.98126f,0.98268f,0.98397f,0.98578f,0.98759f,0.98927f,0.99154f,0.99344f,0.99548f,0.99774f,1.00000f,0.00000f,0.90708f,0.93038f,0.96114f,0.97376f,0.97438f,0.97465f,0.97448f,0.97333f,0.97321f,0.97181f,0.97210f,0.97242f,0.97294f,0.97441f,0.97395f,0.97536f,0.97563f,0.97660f,0.97756f,0.97855f,0.97959f,0.98130f,0.98248f,0.98419f,0.98560f,0.98731f,0.98935f,0.99124f,0.99322f,0.99539f,0.99767f,1.00000f,0.00000f,0.90915f,0.91703f,0.94039f,0.95788f,0.96743f,0.97281f,0.97443f,0.97304f,0.97325f,0.97216f,0.97203f,0.97179f,0.97222f,0.97334f,0.97414f,0.97450f,0.97639f,0.97645f,0.97700f,0.97843f,0.97967f,0.98060f,0.98228f,0.98380f,0.98562f,0.98756f,0.98917f,0.99113f,0.99312f,0.99521f,0.99748f,1.00000f,0.00000f,0.91945f,0.91071f,0.92278f,0.93914f,0.95343f,0.96418f,0.96595f,0.96986f,0.97075f,0.97145f,0.97280f,0.97243f,0.97198f,0.97294f,0.97382f,0.97353f,0.97515f,0.97583f,0.97675f,0.97852f,0.97946f,0.98102f,0.98239f,0.98344f,0.98497f,0.98701f,0.98913f,0.99097f,0.99310f,0.99528f,0.99753f,1.00000f,0.00000f,0.92440f,0.90926f,0.91525f,0.92647f,0.93724f,0.95190f,0.95951f,0.96391f,0.96926f,0.96931f,0.97179f,0.97251f,0.97190f,0.97221f,0.97425f,0.97413f,0.97478f,0.97577f,0.97672f,0.97799f,0.97898f,0.98066f,0.98214f,0.98362f,0.98528f,0.98707f,0.98871f,0.99110f,0.99286f,0.99531f,0.99749f,1.00000f,0.00000f,0.93510f,0.91443f,0.90821f,0.91726f,0.92708f,0.93557f,0.94611f,0.95507f,0.96061f,0.96575f,0.96801f,0.97007f,0.97115f,0.97206f,0.97241f,0.97480f,0.97495f,0.97587f,0.97718f,0.97723f,0.97935f,0.98003f,0.98178f,0.98374f,0.98540f,0.98719f,0.98905f,0.99101f,0.99299f,0.99509f,0.99747f,1.00000f,0.00000f,0.93499f,0.91916f,0.91475f,0.91452f,0.92190f,0.92910f,0.93739f,0.94599f,0.94924f,0.95618f,0.96140f,0.96310f,0.96828f,0.96889f,0.97015f,0.97260f,0.97367f,0.97507f,0.97636f,0.97780f,0.97887f,0.98014f,0.98122f,0.98314f,0.98473f,0.98658f,0.98855f,0.99097f,0.99281f,0.99534f,0.99743f,1.00000f,0.00000f,0.94777f,0.92490f,0.91798f,0.91776f,0.91522f,0.92157f,0.92797f,0.93643f,0.94274f,0.94694f,0.95372f,0.95782f,0.96248f,0.96656f,0.96753f,0.96940f,0.97196f,0.97327f,0.97589f,0.97753f,0.97914f,0.97913f,0.98223f,0.98320f,0.98447f,0.98658f,0.98866f,0.99032f,0.99263f,0.99501f,0.99719f,1.00000f,0.00000f,0.94651f,0.93443f,0.92525f,0.91920f,0.91791f,0.91778f,0.92330f,0.92984f,0.93178f,0.94126f,0.94609f,0.95134f,0.95515f,0.95870f,0.96389f,0.96794f,0.96823f,0.97182f,0.97445f,0.97617f,0.97721f,0.97914f,0.98162f,0.98292f,0.98486f,0.98635f,0.98840f,0.99076f,0.99222f,0.99489f,0.99734f,1.00000f,0.00000f,0.95364f,0.93641f,0.92996f,0.92359f,0.92381f,0.92189f,0.92181f,0.92536f,0.93173f,0.93385f,0.93958f,0.94304f,0.94877f,0.95435f,0.95958f,0.96243f,0.96402f,0.96960f,0.97040f,0.97440f,0.97623f,0.97921f,0.98100f,0.98164f,0.98444f,0.98678f,0.98846f,0.99036f,0.99250f,0.99466f,0.99734f,1.00000f,0.00000f,0.95141f,0.93997f,0.93356f,0.92153f,0.91934f,0.91707f,0.91639f,0.92589f,0.92713f,0.92886f,0.93317f,0.93921f,0.94078f,0.94729f,0.95183f,0.95522f,0.95824f,0.96414f,0.96620f,0.97023f,0.97175f,0.97575f,0.97754f,0.98026f,0.98368f,0.98553f,0.98735f,0.99013f,0.99203f,0.99484f,0.99696f,1.00000f,0.00000f,0.96086f,0.94596f,0.93937f,0.93214f,0.92426f,0.92177f,0.92008f,0.92454f,0.92765f,0.92858f,0.93274f,0.93333f,0.93676f,0.94219f,0.94466f,0.94915f,0.95356f,0.95732f,0.96068f,0.96310f,0.96913f,0.97038f,0.97591f,0.97820f,0.98070f,0.98365f,0.98692f,0.98939f,0.99190f,0.99453f,0.99698f,1.00000f,0.00000f,0.96252f,0.94928f,0.94363f,0.93950f,0.92799f,0.92592f,0.92077f,0.92340f,0.92045f,0.92888f,0.92778f,0.93179f,0.93493f,0.93713f,0.93727f,0.94506f,0.94538f,0.95243f,0.95421f,0.96006f,0.96370f,0.96643f,0.97128f,0.97513f,0.97778f,0.98111f,0.98476f,0.98761f,0.99062f,0.99391f,0.99665f,1.00000f,0.00000f,0.95731f,0.95378f,0.94597f,0.93876f,0.92760f,0.92793f,0.92708f,0.92123f,0.92773f,0.92694f,0.92636f,0.93133f,0.93235f,0.93205f,0.93486f,0.93937f,0.94351f,0.94511f,0.95100f,0.95361f,0.95976f,0.96224f,0.96529f,0.97005f,0.97436f,0.97699f,0.98127f,0.98577f,0.98940f,0.99244f,0.99610f,1.00000f,0.00000f,0.96085f,0.95868f,0.95161f,0.94226f,0.93632f,0.93508f,0.92580f,0.92626f,0.93057f,0.92422f,0.92547f,0.92552f,0.93158f,0.93046f,0.93388f,0.93587f,0.93690f,0.94106f,0.94499f,0.94830f,0.95286f,0.95706f,0.96177f,0.96613f,0.96754f,0.97272f,0.97712f,0.98101f,0.98566f,0.98971f,0.99433f,0.99992f,0.00000f,0.95939f,0.95890f,0.94804f,0.94605f,0.94009f,0.93569f,0.93158f,0.92818f,0.92814f,0.92964f,0.92555f,0.92471f,0.92654f,0.93251f,0.93474f,0.93519f,0.93478f,0.93645f,0.93797f,0.94272f,0.94629f,0.95270f,0.95564f,0.95946f,0.96321f,0.96953f,0.97197f,0.97647f,0.98136f,0.98600f,0.99101f,0.99951f,0.00000f,0.96349f,0.95941f,0.95436f,0.95244f,0.93959f,0.94161f,0.93487f,0.92584f,0.93009f,0.93164f,0.92765f,0.92920f,0.92591f,0.92599f,0.93104f,0.93170f,0.93369f,0.93245f,0.93388f,0.93915f,0.94362f,0.94446f,0.94806f,0.95477f,0.95855f,0.96277f,0.96751f,0.97241f,0.97602f,0.98144f,0.98572f,0.99826f,0.00000f,0.95983f,0.95733f,0.95749f,0.95436f,0.94690f,0.94122f,0.93646f,0.93305f,0.93210f,0.93470f,0.93040f,0.93115f,0.93207f,0.92778f,0.93075f,0.93061f,0.93170f,0.93659f,0.93458f,0.93201f,0.93865f,0.94161f,0.94188f,0.94928f,0.95021f,0.95699f,0.96113f,0.96554f,0.96999f,0.97482f,0.97914f,0.99523f,0.00000f,0.96405f,0.95731f,0.95730f,0.95720f,0.95993f,0.94703f,0.93988f,0.93831f,0.93705f,0.93514f,0.92748f,0.92725f,0.92993f,0.93190f,0.93207f,0.92764f,0.93262f,0.93197f,0.93057f,0.93197f,0.93482f,0.93551f,0.93940f,0.94109f,0.94644f,0.94921f,0.95426f,0.95834f,0.96333f,0.96820f,0.97204f,0.99063f,0.00000f,0.96120f,0.95782f,0.95667f,0.95531f,0.95417f,0.94325f,0.94324f,0.93592f,0.93569f,0.93164f,0.93024f,0.93184f,0.92792f,0.93404f,0.92888f,0.92566f,0.92928f,0.93318f,0.92844f,0.93006f,0.93348f,0.93270f,0.93292f,0.93463f,0.93811f,0.94161f,0.94636f,0.94950f,0.95326f,0.95759f,0.96129f,0.98378f,0.00000f,0.95700f,0.95652f,0.94472f,0.95276f,0.93906f,0.93973f,0.94237f,0.93601f,0.94146f,0.93530f,0.93033f,0.92952f,0.92702f,0.93231f,0.93086f,0.93099f,0.92753f,0.92708f,0.92902f,0.92658f,0.92906f,0.92837f,0.93049f,0.92961f,0.93431f,0.93464f,0.93764f,0.93999f,0.94367f,0.94811f,0.95110f,0.97511f,0.00000f,0.93827f,0.94985f,0.94543f,0.95036f,0.93922f,0.94566f,0.93779f,0.93924f,0.93791f,0.93034f,0.93468f,0.92781f,0.92804f,0.92763f,0.92526f,0.92700f,0.92612f,0.92735f,0.93019f,0.92531f,0.92726f,0.92605f,0.92439f,0.92452f,0.92638f,0.92678f,0.93226f,0.93294f,0.93535f,0.93915f,0.94107f,0.96704f,0.00000f,0.93123f,0.93996f,0.93624f,0.93840f,0.93245f,0.93453f,0.93856f,0.93417f,0.92726f,0.92967f,0.93063f,0.93021f,0.92708f,0.92069f,0.92716f,0.92216f,0.92646f,0.92481f,0.92065f,0.91870f,0.92179f,0.92193f,0.92099f,0.91857f,0.91794f,0.92043f,0.91981f,0.92337f,0.92654f,0.92591f,0.93155f,0.95675f,0.00000f,0.91805f,0.91995f,0.92779f,0.92415f,0.92344f,0.92339f,0.92743f,0.92499f,0.91945f,0.91884f,0.92425f,0.92031f,0.92463f,0.92465f,0.91895f,0.91777f,0.92000f,0.91956f,0.91385f,0.91809f,0.91723f,0.91534f,0.91579f,0.91352f,0.91295f,0.91498f,0.91517f,0.91631f,0.91444f,0.91805f,0.91908f,0.94728f,0.00000f,0.89671f,0.89828f,0.91380f,0.91995f,0.91581f,0.91030f,0.91461f,0.91576f,0.91276f,0.91843f,0.91338f,0.91792f,0.91504f,0.91708f,0.91457f,0.91086f,0.91498f,0.91136f,0.91003f,0.91042f,0.90870f,0.90767f,0.90788f,0.90608f,0.90493f,0.90253f,0.90434f,0.90321f,0.90334f,0.90487f,0.90886f,0.93670f,0.00000f,0.87591f,0.88155f,0.88955f,0.89289f,0.90308f,0.90021f,0.90392f,0.90248f,0.90630f,0.90265f,0.90614f,0.90125f,0.90528f,0.90878f,0.90782f,0.89996f,0.90436f,0.90558f,0.90161f,0.90042f,0.90547f,0.90180f,0.89819f,0.89903f,0.89728f,0.89865f,0.89217f,0.89465f,0.89198f,0.89753f,0.89613f,0.92602f,0.00000f,0.85860f,0.86782f,0.87632f,0.87038f,0.87704f,0.87662f,0.88903f,0.88142f,0.88854f,0.89339f,0.89391f,0.89174f,0.88625f,0.89225f,0.89391f,0.89603f,0.89769f,0.89440f,0.89329f,0.89176f,0.89157f,0.89524f,0.89238f,0.88861f,0.88661f,0.88536f,0.88101f,0.88078f,0.88298f,0.88273f,0.88583f,0.91587f,0.00000f,0.83068f,0.84239f,0.85183f,0.85115f,0.85806f,0.85956f,0.86182f,0.86738f,0.87157f,0.87615f,0.87370f,0.87807f,0.87757f,0.88031f,0.88097f,0.88072f,0.88691f,0.87995f,0.88655f,0.88177f,0.88022f,0.88023f,0.88016f,0.87755f,0.87738f,0.87387f,0.87458f,0.87077f,0.86982f,0.86948f,0.87651f,0.90494f,
 0.00000f,0.99272f,0.98234f,0.96981f,0.96234f,0.95234f,0.97317f,0.95684f,0.96795f,0.93256f,0.94635f,0.97075f,0.96452f,0.96598f,0.96638f,0.95268f,0.95386f,0.95090f,0.95494f,0.95907f,0.96524f,0.96532f,0.96801f,0.97214f,0.97565f,0.97581f,0.98171f,0.96692f,0.98425f,0.99176f,0.99489f,0.99744f,1.00000f,0.00000f,0.99059f,0.98112f,0.97679f,0.97511f,0.97451f,0.97508f,0.97078f,0.97205f,0.96904f,0.96935f,0.96945f,0.96946f,0.97093f,0.97107f,0.97139f,0.97126f,0.97166f,0.97345f,0.97502f,0.97645f,0.97805f,0.97973f,0.97987f,0.98239f,0.98456f,0.98631f,0.98728f,0.98988f,0.99239f,0.99531f,0.99769f,1.00000f,0.00000f,0.99024f,0.98565f,0.98252f,0.97901f,0.97878f,0.97747f,0.97512f,0.97534f,0.97383f,0.97467f,0.97303f,0.97334f,0.97361f,0.97312f,0.97458f,0.97510f,0.97549f,0.97694f,0.97751f,0.97854f,0.98001f,0.98110f,0.98296f,0.98409f,0.98578f,0.98773f,0.98943f,0.99145f,0.99328f,0.99545f,0.99779f,1.00000f,0.00000f,0.98375f,0.98387f,0.98246f,0.97888f,0.97750f,0.97649f,0.97534f,0.97456f,0.97481f,0.97344f,0.97341f,0.97265f,0.97403f,0.97349f,0.97480f,0.97534f,0.97579f,0.97714f,0.97805f,0.97876f,0.97965f,0.98146f,0.98282f,0.98416f,0.98588f,0.98745f,0.98950f,0.99123f,0.99332f,0.99550f,0.99778f,1.00000f,0.00000f,0.94234f,0.98251f,0.98154f,0.97988f,0.97817f,0.97536f,0.97556f,0.97473f,0.97345f,0.97357f,0.97262f,0.97175f,0.97333f,0.97372f,0.97416f,0.97472f,0.97600f,0.97643f,0.97785f,0.97845f,0.97987f,0.98146f,0.98228f,0.98415f,0.98577f,0.98754f,0.98939f,0.99131f,0.99334f,0.99537f,0.99775f,1.00000f,0.00000f,0.91582f,0.96094f,0.97650f,0.97927f,0.97721f,0.97553f,0.97460f,0.97413f,0.97421f,0.97271f,0.97318f,0.97319f,0.97363f,0.97390f,0.97417f,0.97402f,0.97563f,0.97705f,0.97762f,0.97858f,0.97990f,0.98058f,0.98259f,0.98393f,0.98562f,0.98758f,0.98943f,0.99130f,0.99342f,0.99543f,0.99778f,1.00000f,0.00000f,0.90661f,0.93165f,0.95905f,0.97291f,0.97625f,0.97498f,0.97529f,0.97374f,0.97172f,0.97314f,0.97311f,0.97335f,0.97261f,0.97321f,0.97402f,0.97405f,0.97534f,0.97587f,0.97719f,0.97836f,0.97990f,0.98104f,0.98254f,0.98423f,0.98580f,0.98757f,0.98929f,0.99108f,0.99313f,0.99541f,0.99769f,1.00000f,0.00000f,0.91075f,0.91891f,0.93908f,0.95827f,0.96705f,0.97237f,0.97347f,0.97443f,0.97250f,0.97174f,0.97305f,0.97273f,0.97265f,0.97314f,0.97336f,0.97423f,0.97496f,0.97649f,0.97709f,0.97842f,0.97961f,0.98047f,0.98230f,0.98359f,0.98524f,0.98739f,0.98916f,0.99110f,0.99319f,0.99541f,0.99749f,1.00000f,0.00000f,0.91786f,0.90835f,0.92589f,0.93922f,0.95490f,0.96318f,0.96845f,0.97205f,0.97170f,0.97188f,0.97148f,0.97319f,0.97334f,0.97211f,0.97425f,0.97452f,0.97479f,0.97544f,0.97685f,0.97798f,0.97900f,0.98061f,0.98221f,0.98397f,0.98560f,0.98742f,0.98914f,0.99125f,0.99324f,0.99533f,0.99755f,1.00000f,0.00000f,0.92591f,0.91234f,0.91542f,0.92551f,0.93859f,0.95015f,0.95814f,0.96411f,0.96834f,0.96842f,0.97086f,0.97176f,0.97236f,0.97224f,0.97275f,0.97280f,0.97399f,0.97620f,0.97694f,0.97849f,0.97933f,0.98111f,0.98227f,0.98367f,0.98525f,0.98728f,0.98903f,0.99119f,0.99302f,0.99535f,0.99747f,1.00000f,0.00000f,0.93024f,0.91634f,0.91410f,0.91639f,0.92830f,0.93710f,0.94699f,0.95615f,0.96163f,0.96438f,0.96766f,0.96930f,0.97036f,0.97085f,0.97320f,0.97265f,0.97397f,0.97604f,0.97632f,0.97733f,0.97865f,0.98090f,0.98227f,0.98358f,0.98514f,0.98721f,0.98912f,0.99091f,0.99295f,0.99530f,0.99746f,1.00000f,0.00000f,0.93710f,0.92076f,0.91442f,0.91529f,0.91905f,0.92740f,0.93475f,0.94367f,0.95248f,0.95647f,0.96095f,0.96478f,0.96677f,0.96907f,0.97118f,0.97249f,0.97379f,0.97556f,0.97653f,0.97745f,0.97850f,0.98083f,0.98147f,0.98364f,0.98499f,0.98697f,0.98817f,0.99117f,0.99282f,0.99522f,0.99750f,1.00000f,0.00000f,0.94726f,0.92563f,0.91757f,0.91296f,0.91947f,0.92364f,0.92714f,0.93770f,0.94181f,0.94884f,0.95191f,0.95638f,0.96257f,0.96577f,0.96822f,0.97094f,0.97193f,0.97375f,0.97683f,0.97746f,0.97849f,0.97979f,0.98171f,0.98318f,0.98506f,0.98678f,0.98844f,0.99055f,0.99306f,0.99487f,0.99731f,1.00000f,0.00000f,0.95392f,0.93099f,0.92355f,0.92084f,0.92286f,0.91996f,0.92344f,0.92643f,0.93314f,0.93725f,0.94357f,0.95230f,0.95453f,0.95801f,0.96188f,0.96715f,0.96954f,0.97083f,0.97401f,0.97555f,0.97713f,0.97996f,0.98105f,0.98286f,0.98445f,0.98604f,0.98823f,0.99072f,0.99259f,0.99485f,0.99734f,1.00000f,0.00000f,0.95214f,0.94151f,0.92350f,0.91930f,0.92165f,0.92366f,0.92028f,0.92568f,0.93059f,0.93487f,0.93910f,0.94333f,0.94812f,0.95275f,0.95754f,0.96173f,0.96493f,0.96754f,0.97076f,0.97263f,0.97623f,0.97868f,0.98021f,0.98147f,0.98421f,0.98690f,0.98812f,0.99001f,0.99228f,0.99466f,0.99720f,1.00000f,0.00000f,0.95592f,0.94515f,0.93526f,0.92264f,0.92156f,0.91934f,0.91794f,0.91979f,0.92976f,0.93001f,0.93394f,0.93893f,0.94183f,0.94444f,0.95115f,0.95497f,0.95942f,0.96444f,0.96739f,0.97019f,0.97247f,0.97516f,0.97811f,0.98123f,0.98310f,0.98513f,0.98798f,0.99020f,0.99206f,0.99438f,0.99714f,1.00000f,0.00000f,0.96853f,0.94298f,0.94244f,0.92471f,0.92521f,0.92645f,0.92440f,0.92266f,0.92487f,0.92880f,0.93154f,0.93118f,0.93676f,0.93993f,0.94487f,0.94977f,0.95445f,0.95862f,0.96001f,0.96509f,0.96832f,0.97246f,0.97315f,0.97922f,0.98160f,0.98306f,0.98652f,0.98943f,0.99215f,0.99470f,0.99701f,1.00000f,0.00000f,0.95716f,0.94610f,0.94447f,0.93817f,0.93586f,0.92455f,0.92559f,0.92225f,0.92815f,0.93123f,0.93190f,0.93016f,0.93276f,0.93446f,0.94023f,0.94226f,0.95013f,0.95130f,0.95594f,0.95999f,0.96317f,0.96726f,0.96955f,0.97409f,0.97819f,0.98152f,0.98419f,0.98836f,0.99070f,0.99361f,0.99652f,1.00000f,0.00000f,0.96157f,0.95511f,0.94749f,0.93970f,0.92933f,0.92711f,0.92397f,0.92547f,0.92173f,0.92591f,0.92800f,0.92882f,0.93094f,0.92943f,0.93478f,0.93919f,0.94084f,0.94419f,0.95070f,0.95377f,0.95759f,0.96279f,0.96548f,0.97038f,0.97430f,0.97765f,0.98101f,0.98518f,0.98854f,0.99223f,0.99573f,1.00000f,0.00000f,0.96280f,0.95818f,0.94577f,0.94434f,0.93678f,0.93646f,0.93267f,0.92703f,0.92564f,0.93077f,0.92619f,0.92854f,0.92882f,0.93146f,0.93256f,0.93569f,0.93767f,0.94264f,0.94408f,0.94926f,0.95081f,0.95420f,0.95998f,0.96552f,0.96791f,0.97203f,0.97699f,0.98141f,0.98556f,0.98921f,0.99374f,0.99992f,0.00000f,0.95745f,0.95645f,0.95269f,0.94360f,0.94110f,0.93882f,0.92748f,0.92945f,0.92598f,0.93060f,0.92474f,0.92837f,0.92922f,0.92954f,0.93266f,0.93141f,0.93723f,0.93745f,0.93873f,0.94413f,0.94618f,0.95138f,0.95392f,0.95756f,0.96316f,0.96745f,0.97142f,0.97782f,0.98067f,0.98618f,0.99055f,0.99944f,0.00000f,0.96626f,0.96893f,0.95927f,0.94580f,0.94032f,0.93557f,0.93096f,0.92921f,0.92960f,0.93063f,0.92664f,0.92594f,0.92918f,0.92655f,0.93266f,0.93271f,0.93281f,0.93491f,0.93669f,0.93947f,0.94425f,0.94472f,0.94952f,0.95242f,0.95640f,0.96162f,0.96576f,0.97062f,0.97554f,0.98019f,0.98574f,0.99802f,0.00000f,0.95698f,0.95831f,0.95842f,0.95012f,0.94814f,0.94094f,0.93930f,0.92859f,0.93301f,0.92660f,0.93115f,0.92657f,0.93047f,0.93048f,0.92969f,0.93091f,0.93461f,0.93463f,0.93690f,0.93410f,0.93642f,0.93898f,0.94106f,0.94761f,0.95202f,0.95507f,0.96018f,0.96484f,0.96957f,0.97472f,0.97870f,0.99475f,0.00000f,0.96771f,0.96277f,0.95003f,0.95673f,0.94963f,0.94451f,0.94569f,0.93904f,0.93326f,0.93288f,0.92987f,0.92508f,0.92959f,0.92894f,0.92663f,0.92744f,0.93131f,0.92894f,0.93583f,0.93211f,0.93055f,0.93459f,0.93692f,0.94330f,0.94817f,0.94861f,0.95262f,0.95537f,0.96191f,0.96744f,0.96907f,0.98959f,0.00000f,0.95994f,0.95906f,0.95528f,0.94999f,0.95519f,0.93882f,0.93596f,0.93654f,0.93720f,0.93720f,0.93574f,0.92580f,0.92881f,0.92851f,0.92763f,0.93080f,0.92782f,0.93128f,0.92979f,0.93362f,0.93178f,0.93375f,0.93313f,0.93342f,0.93576f,0.94313f,0.94648f,0.95033f,0.95147f,0.95656f,0.96030f,0.98229f,0.00000f,0.94895f,0.94942f,0.95154f,0.94647f,0.95099f,0.94902f,0.94111f,0.93975f,0.93386f,0.93469f,0.93383f,0.93000f,0.92809f,0.92534f,0.92608f,0.92783f,0.92888f,0.92689f,0.92898f,0.93094f,0.92905f,0.92871f,0.92886f,0.92987f,0.92807f,0.93289f,0.93803f,0.94073f,0.94222f,0.94734f,0.95125f,0.97496f,0.00000f,0.94305f,0.94414f,0.94982f,0.94892f,0.94098f,0.93822f,0.94268f,0.93504f,0.93562f,0.92877f,0.93029f,0.93405f,0.93016f,0.92748f,0.92642f,0.92393f,0.92206f,0.92367f,0.92707f,0.92502f,0.92197f,0.92334f,0.92492f,0.92405f,0.92262f,0.92524f,0.92710f,0.93319f,0.93403f,0.93490f,0.93869f,0.96517f,0.00000f,0.92999f,0.94151f,0.93751f,0.93920f,0.93350f,0.93134f,0.93534f,0.93256f,0.93305f,0.92766f,0.93139f,0.92712f,0.92262f,0.92736f,0.92169f,0.92287f,0.92507f,0.91758f,0.92421f,0.92273f,0.92045f,0.91764f,0.91714f,0.91910f,0.91665f,0.91853f,0.92061f,0.91938f,0.92408f,0.92450f,0.92734f,0.95524f,0.00000f,0.91558f,0.91971f,0.93635f,0.92602f,0.92113f,0.92853f,0.92620f,0.92877f,0.92155f,0.92580f,0.92386f,0.92455f,0.92421f,0.92210f,0.91739f,0.91825f,0.92083f,0.91324f,0.91619f,0.91889f,0.91526f,0.91537f,0.91316f,0.91194f,0.91225f,0.90503f,0.91233f,0.91063f,0.91112f,0.91384f,0.91681f,0.94367f,0.00000f,0.90500f,0.91157f,0.91143f,0.91410f,0.91410f,0.91847f,0.92128f,0.91522f,0.91522f,0.91871f,0.91444f,0.91448f,0.91677f,0.91016f,0.91161f,0.91099f,0.91361f,0.91222f,0.91399f,0.91014f,0.90764f,0.90782f,0.90430f,0.90487f,0.90111f,0.90005f,0.89988f,0.90150f,0.89876f,0.90200f,0.90483f,0.93366f,0.00000f,0.87778f,0.88991f,0.89077f,0.89206f,0.90151f,0.90333f,0.90572f,0.90752f,0.90514f,0.90435f,0.90745f,0.90553f,0.90325f,0.90204f,0.90225f,0.90505f,0.90136f,0.90212f,0.90174f,0.90035f,0.90348f,0.89838f,0.89733f,0.89695f,0.89434f,0.89455f,0.89027f,0.88943f,0.89074f,0.89167f,0.89393f,0.92105f,0.00000f,0.86289f,0.86625f,0.87443f,0.87543f,0.88124f,0.88249f,0.88875f,0.89479f,0.89156f,0.89339f,0.88964f,0.89162f,0.89250f,0.88844f,0.89313f,0.89248f,0.89688f,0.88961f,0.89413f,0.89481f,0.88833f,0.88914f,0.88981f,0.88721f,0.88355f,0.88172f,0.87694f,0.87875f,0.87807f,0.87813f,0.88359f,0.91250f,0.00000f,0.84090f,0.84019f,0.85751f,0.85569f,0.86676f,0.85803f,0.87098f,0.87841f,0.86853f,0.87623f,0.87200f,0.87289f,0.88227f,0.87898f,0.88364f,0.88173f,0.88053f,0.88171f,0.88181f,0.88230f,0.87929f,0.87915f,0.88088f,0.87644f,0.87136f,0.87409f,0.86932f,0.86570f,0.86493f,0.86748f,0.87215f,0.90189f,
 0.00000f,0.99653f,0.97299f,0.95957f,0.95266f,0.95351f,0.95827f,0.94433f,0.95614f,0.96292f,0.96162f,0.97101f,0.93938f,0.96864f,0.96897f,0.96700f,0.96802f,0.95447f,0.95653f,0.95179f,0.96637f,0.96087f,0.97892f,0.96798f,0.97193f,0.97907f,0.98279f,0.98401f,0.97953f,0.98845f,0.98579f,0.99432f,1.00000f,0.00000f,0.98864f,0.98232f,0.97741f,0.97449f,0.97259f,0.97110f,0.96895f,0.97078f,0.97153f,0.97151f,0.97118f,0.96927f,0.97160f,0.97134f,0.97180f,0.97249f,0.97276f,0.97403f,0.97419f,0.97647f,0.97721f,0.97975f,0.98043f,0.98244f,0.98457f,0.98569f,0.98834f,0.98967f,0.99288f,0.99429f,0.99718f,1.00000f,0.00000f,0.99087f,0.98572f,0.98220f,0.97951f,0.97739f,0.97564f,0.97511f,0.97464f,0.97398f,0.97391f,0.97350f,0.97219f,0.97314f,0.97431f,0.97413f,0.97527f,0.97528f,0.97667f,0.97779f,0.97871f,0.97953f,0.98118f,0.98266f,0.98397f,0.98600f,0.98747f,0.98938f,0.99142f,0.99341f,0.99541f,0.99776f,1.00000f,0.00000f,0.98255f,0.98452f,0.98240f,0.97986f,0.97759f,0.97651f,0.97531f,0.97465f,0.97413f,0.97395f,0.97359f,0.97212f,0.97375f,0.97369f,0.97527f,0.97527f,0.97663f,0.97673f,0.97740f,0.97873f,0.97991f,0.98138f,0.98290f,0.98437f,0.98567f,0.98754f,0.98951f,0.99137f,0.99330f,0.99540f,0.99772f,1.00000f,0.00000f,0.94218f,0.98282f,0.98037f,0.97917f,0.97805f,0.97710f,0.97546f,0.97513f,0.97354f,0.97261f,0.97323f,0.97315f,0.97333f,0.97411f,0.97445f,0.97487f,0.97549f,0.97637f,0.97794f,0.97842f,0.98008f,0.98152f,0.98283f,0.98438f,0.98587f,0.98744f,0.98931f,0.99129f,0.99342f,0.99545f,0.99774f,1.00000f,0.00000f,0.91536f,0.95768f,0.97821f,0.97843f,0.97614f,0.97555f,0.97394f,0.97353f,0.97394f,0.97391f,0.97329f,0.97271f,0.97474f,0.97370f,0.97323f,0.97468f,0.97566f,0.97680f,0.97742f,0.97837f,0.97989f,0.98119f,0.98266f,0.98409f,0.98566f,0.98756f,0.98940f,0.99125f,0.99338f,0.99545f,0.99778f,1.00000f,0.00000f,0.90439f,0.93065f,0.96178f,0.97297f,0.97686f,0.97446f,0.97508f,0.97406f,0.97268f,0.97258f,0.97201f,0.97318f,0.97273f,0.97458f,0.97400f,0.97416f,0.97526f,0.97667f,0.97744f,0.97851f,0.97968f,0.98108f,0.98278f,0.98379f,0.98558f,0.98719f,0.98950f,0.99104f,0.99331f,0.99550f,0.99759f,1.00000f,0.00000f,0.90952f,0.91890f,0.93551f,0.95700f,0.96658f,0.97179f,0.97406f,0.97293f,0.97259f,0.97238f,0.97288f,0.97203f,0.97295f,0.97287f,0.97416f,0.97406f,0.97484f,0.97606f,0.97742f,0.97824f,0.97963f,0.98062f,0.98231f,0.98404f,0.98568f,0.98737f,0.98906f,0.99098f,0.99318f,0.99536f,0.99755f,1.00000f,0.00000f,0.91843f,0.91270f,0.92298f,0.93739f,0.95551f,0.96360f,0.96680f,0.96927f,0.97190f,0.97084f,0.97166f,0.97290f,0.97211f,0.97273f,0.97306f,0.97388f,0.97489f,0.97640f,0.97673f,0.97814f,0.97943f,0.98022f,0.98158f,0.98372f,0.98533f,0.98732f,0.98928f,0.99131f,0.99294f,0.99544f,0.99756f,1.00000f,0.00000f,0.92230f,0.91448f,0.91630f,0.92744f,0.93754f,0.94889f,0.95818f,0.96298f,0.96827f,0.96961f,0.97090f,0.97230f,0.97258f,0.97260f,0.97188f,0.97444f,0.97476f,0.97563f,0.97698f,0.97796f,0.97845f,0.98077f,0.98210f,0.98331f,0.98559f,0.98720f,0.98932f,0.99063f,0.99317f,0.99521f,0.99749f,1.00000f,0.00000f,0.93338f,0.91449f,0.91685f,0.92094f,0.92530f,0.93818f,0.94683f,0.95307f,0.96124f,0.96472f,0.96661f,0.96823f,0.97077f,0.97243f,0.97256f,0.97394f,0.97501f,0.97508f,0.97671f,0.97838f,0.97949f,0.98066f,0.98205f,0.98308f,0.98499f,0.98730f,0.98894f,0.99090f,0.99312f,0.99505f,0.99746f,1.00000f,0.00000f,0.93992f,0.91518f,0.91473f,0.91550f,0.92008f,0.92718f,0.93829f,0.94311f,0.95216f,0.95738f,0.96241f,0.96493f,0.96806f,0.96994f,0.97163f,0.97216f,0.97458f,0.97496f,0.97518f,0.97819f,0.97859f,0.97947f,0.98143f,0.98339f,0.98535f,0.98649f,0.98840f,0.99075f,0.99331f,0.99514f,0.99737f,1.00000f,0.00000f,0.94308f,0.92799f,0.91512f,0.91671f,0.91785f,0.92114f,0.92942f,0.93437f,0.94197f,0.94709f,0.95549f,0.95877f,0.96082f,0.96550f,0.96727f,0.97101f,0.97369f,0.97402f,0.97573f,0.97664f,0.97853f,0.97929f,0.98125f,0.98314f,0.98493f,0.98647f,0.98875f,0.99077f,0.99282f,0.99490f,0.99739f,1.00000f,0.00000f,0.94620f,0.93937f,0.92138f,0.92346f,0.92073f,0.92249f,0.92511f,0.93056f,0.93218f,0.93873f,0.94525f,0.95145f,0.95533f,0.96062f,0.96538f,0.96796f,0.96916f,0.97058f,0.97421f,0.97536f,0.97734f,0.97895f,0.98137f,0.98337f,0.98449f,0.98652f,0.98843f,0.99030f,0.99229f,0.99480f,0.99734f,1.00000f,0.00000f,0.95006f,0.93694f,0.92593f,0.92123f,0.92216f,0.92204f,0.92363f,0.92348f,0.92575f,0.93208f,0.93837f,0.94265f,0.94810f,0.95239f,0.95731f,0.96388f,0.96453f,0.96814f,0.97089f,0.97305f,0.97560f,0.97830f,0.98089f,0.98193f,0.98482f,0.98581f,0.98835f,0.99036f,0.99225f,0.99489f,0.99705f,1.00000f,0.00000f,0.96129f,0.93591f,0.93640f,0.92680f,0.92261f,0.91967f,0.92237f,0.92194f,0.92388f,0.93040f,0.93195f,0.93471f,0.94184f,0.94704f,0.94976f,0.95441f,0.95728f,0.96340f,0.96579f,0.97017f,0.97384f,0.97545f,0.97821f,0.98036f,0.98262f,0.98569f,0.98803f,0.99020f,0.99200f,0.99481f,0.99738f,1.00000f,0.00000f,0.95770f,0.94317f,0.93777f,0.93048f,0.92272f,0.92158f,0.92348f,0.92106f,0.92602f,0.92867f,0.93119f,0.93046f,0.93679f,0.93945f,0.94609f,0.94773f,0.95260f,0.95662f,0.96050f,0.96479f,0.96789f,0.97124f,0.97494f,0.97821f,0.98077f,0.98434f,0.98575f,0.98922f,0.99094f,0.99451f,0.99728f,1.00000f,0.00000f,0.95364f,0.95588f,0.94544f,0.92970f,0.92386f,0.92458f,0.92056f,0.92241f,0.92228f,0.92769f,0.93124f,0.93056f,0.93357f,0.93402f,0.93885f,0.94057f,0.94674f,0.95278f,0.95432f,0.96081f,0.96408f,0.96600f,0.97041f,0.97420f,0.97768f,0.98061f,0.98405f,0.98771f,0.99116f,0.99365f,0.99698f,1.00000f,0.00000f,0.95217f,0.95568f,0.94217f,0.93765f,0.93073f,0.92690f,0.92527f,0.92898f,0.91939f,0.92781f,0.92893f,0.93099f,0.93086f,0.93123f,0.93666f,0.93951f,0.94134f,0.94403f,0.94894f,0.95314f,0.95859f,0.96171f,0.96468f,0.97037f,0.97374f,0.97620f,0.98148f,0.98536f,0.98888f,0.99191f,0.99585f,1.00000f,0.00000f,0.96368f,0.95571f,0.94773f,0.94158f,0.94233f,0.93531f,0.92691f,0.92573f,0.92381f,0.92380f,0.92657f,0.92635f,0.92950f,0.93120f,0.93582f,0.93372f,0.93713f,0.94161f,0.94401f,0.94755f,0.95175f,0.95448f,0.96037f,0.96413f,0.96682f,0.97188f,0.97639f,0.98151f,0.98516f,0.98891f,0.99343f,0.99991f,0.00000f,0.95366f,0.96107f,0.95442f,0.94629f,0.94206f,0.94089f,0.93061f,0.93150f,0.92949f,0.92629f,0.92793f,0.93094f,0.92613f,0.93095f,0.92892f,0.93393f,0.93510f,0.93585f,0.93634f,0.94213f,0.94593f,0.95004f,0.95593f,0.95769f,0.96263f,0.96664f,0.97155f,0.97619f,0.98080f,0.98587f,0.99047f,0.99949f,0.00000f,0.95997f,0.96325f,0.95496f,0.94830f,0.93891f,0.94240f,0.93546f,0.92992f,0.92815f,0.92801f,0.92615f,0.92607f,0.93165f,0.92957f,0.92875f,0.93208f,0.93507f,0.93302f,0.93489f,0.93795f,0.94152f,0.94394f,0.94822f,0.95378f,0.95767f,0.96100f,0.96577f,0.97011f,0.97434f,0.98065f,0.98487f,0.99811f,0.00000f,0.97215f,0.96032f,0.95620f,0.95112f,0.95229f,0.93871f,0.93801f,0.93684f,0.93841f,0.92839f,0.93253f,0.92863f,0.92916f,0.92819f,0.92844f,0.92841f,0.93018f,0.93082f,0.93420f,0.93406f,0.93734f,0.93654f,0.94165f,0.94471f,0.95057f,0.95419f,0.95981f,0.96412f,0.96930f,0.97274f,0.97830f,0.99459f,0.00000f,0.96349f,0.95552f,0.96066f,0.95472f,0.94156f,0.94481f,0.94484f,0.93082f,0.93429f,0.92966f,0.93365f,0.92583f,0.93085f,0.92676f,0.92806f,0.93093f,0.92649f,0.92949f,0.92891f,0.93252f,0.93187f,0.93480f,0.93600f,0.94056f,0.94234f,0.94785f,0.95223f,0.95756f,0.95999f,0.96547f,0.97002f,0.98897f,0.00000f,0.96423f,0.96445f,0.95546f,0.95818f,0.94629f,0.94515f,0.94153f,0.93503f,0.93718f,0.93385f,0.93351f,0.93145f,0.92712f,0.92818f,0.92793f,0.93251f,0.92772f,0.92967f,0.92960f,0.92955f,0.92953f,0.92963f,0.93397f,0.93507f,0.93821f,0.93959f,0.94172f,0.94675f,0.95125f,0.95441f,0.95982f,0.98149f,0.00000f,0.94270f,0.95581f,0.95226f,0.95250f,0.94947f,0.93888f,0.94244f,0.94145f,0.93601f,0.93455f,0.93247f,0.92774f,0.92588f,0.92695f,0.92890f,0.92550f,0.92243f,0.92532f,0.92700f,0.92944f,0.92656f,0.93007f,0.92713f,0.92951f,0.93132f,0.93073f,0.93584f,0.94040f,0.94162f,0.94572f,0.94794f,0.97284f,0.00000f,0.94269f,0.95214f,0.94948f,0.94623f,0.94520f,0.94121f,0.94062f,0.93711f,0.93396f,0.93241f,0.93229f,0.92765f,0.92719f,0.92903f,0.92351f,0.92310f,0.91946f,0.92316f,0.92835f,0.92460f,0.92712f,0.92211f,0.92379f,0.92405f,0.91983f,0.92255f,0.92597f,0.92780f,0.93330f,0.93428f,0.93623f,0.96341f,0.00000f,0.93143f,0.93698f,0.94556f,0.94674f,0.93210f,0.93498f,0.93655f,0.93186f,0.92772f,0.93384f,0.93120f,0.92448f,0.92725f,0.92217f,0.92117f,0.91975f,0.92204f,0.92632f,0.92164f,0.91766f,0.92048f,0.91872f,0.91970f,0.91464f,0.91273f,0.91382f,0.91694f,0.91622f,0.92187f,0.92109f,0.92372f,0.95342f,0.00000f,0.91921f,0.92674f,0.93723f,0.93388f,0.93025f,0.92972f,0.92714f,0.92044f,0.92256f,0.92395f,0.92506f,0.92513f,0.91722f,0.91966f,0.92310f,0.91554f,0.91462f,0.91494f,0.91638f,0.91766f,0.91517f,0.91193f,0.91074f,0.91281f,0.90934f,0.90975f,0.90754f,0.90489f,0.90947f,0.90905f,0.91219f,0.94152f,0.00000f,0.90149f,0.91020f,0.92021f,0.91970f,0.91638f,0.91740f,0.91776f,0.91453f,0.92113f,0.91765f,0.91735f,0.91968f,0.91642f,0.91373f,0.91372f,0.91620f,0.90936f,0.91109f,0.90799f,0.91109f,0.90858f,0.90669f,0.90280f,0.90671f,0.90295f,0.89931f,0.89720f,0.89776f,0.89798f,0.89822f,0.89949f,0.93050f,0.00000f,0.88771f,0.89446f,0.89844f,0.90240f,0.90644f,0.90793f,0.90512f,0.91040f,0.90471f,0.89953f,0.90292f,0.90588f,0.90482f,0.90455f,0.90252f,0.90324f,0.90612f,0.90137f,0.90527f,0.90010f,0.90422f,0.90070f,0.89850f,0.89188f,0.89020f,0.88916f,0.88778f,0.88755f,0.88598f,0.88737f,0.89093f,0.91895f,0.00000f,0.86469f,0.87021f,0.87458f,0.88010f,0.88319f,0.89054f,0.88877f,0.89003f,0.89247f,0.89344f,0.89788f,0.89448f,0.89330f,0.88901f,0.89438f,0.89545f,0.89084f,0.89466f,0.89144f,0.89352f,0.89145f,0.88809f,0.88763f,0.88126f,0.88113f,0.87745f,0.87533f,0.87432f,0.87749f,0.87473f,0.87688f,0.90817f,0.00000f,0.83550f,0.85216f,0.85132f,0.85770f,0.86626f,0.86744f,0.86468f,0.87115f,0.87519f,0.87920f,0.87570f,0.87848f,0.87560f,0.87657f,0.87951f,0.88320f,0.88034f,0.88382f,0.87817f,0.88105f,0.87458f,0.87551f,0.87715f,0.87310f,0.86856f,0.86933f,0.86569f,0.86315f,0.86242f,0.86447f,0.86665f,0.89670f,
 0.00000f,0.99861f,0.97430f,0.96731f,0.95954f,0.95147f,0.96109f,0.96157f,0.94804f,0.92006f,0.95608f,0.97299f,0.93604f,0.96839f,0.96824f,0.96412f,0.94395f,0.94665f,0.96431f,0.97184f,0.96505f,0.97309f,0.96682f,0.97100f,0.97882f,0.96982f,0.97787f,0.98663f,0.97243f,0.99221f,0.99483f,0.99609f,1.00000f,0.00000f,0.99099f,0.98320f,0.97780f,0.97564f,0.97327f,0.97145f,0.97086f,0.96766f,0.96693f,0.97135f,0.97222f,0.96764f,0.97213f,0.97191f,0.97174f,0.97233f,0.97327f,0.97427f,0.97509f,0.97627f,0.97748f,0.97910f,0.98041f,0.98235f,0.98374f,0.98637f,0.98830f,0.98940f,0.99259f,0.99513f,0.99747f,1.00000f,0.00000f,0.99065f,0.98743f,0.98167f,0.98039f,0.97881f,0.97662f,0.97483f,0.97508f,0.97338f,0.97357f,0.97364f,0.97295f,0.97419f,0.97446f,0.97409f,0.97527f,0.97578f,0.97675f,0.97774f,0.97885f,0.97949f,0.98103f,0.98270f,0.98415f,0.98577f,0.98744f,0.98940f,0.99137f,0.99330f,0.99543f,0.99781f,1.00000f,0.00000f,0.98321f,0.98584f,0.98233f,0.98019f,0.97681f,0.97749f,0.97458f,0.97413f,0.97305f,0.97429f,0.97339f,0.97287f,0.97420f,0.97354f,0.97428f,0.97526f,0.97561f,0.97699f,0.97789f,0.97832f,0.97987f,0.98153f,0.98287f,0.98448f,0.98605f,0.98754f,0.98928f,0.99128f,0.99321f,0.99546f,0.99776f,1.00000f,0.00000f,0.94184f,0.98231f,0.98165f,0.97900f,0.97705f,0.97558f,0.97519f,0.97509f,0.97346f,0.97313f,0.97285f,0.97308f,0.97236f,0.97439f,0.97455f,0.97555f,0.97506f,0.97661f,0.97705f,0.97916f,0.97935f,0.98159f,0.98223f,0.98394f,0.98571f,0.98763f,0.98945f,0.99107f,0.99340f,0.99542f,0.99778f,1.00000f,0.00000f,0.91509f,0.95965f,0.97823f,0.97828f,0.97736f,0.97508f,0.97640f,0.97372f,0.97466f,0.97296f,0.97288f,0.97374f,0.97321f,0.97367f,0.97496f,0.97465f,0.97587f,0.97648f,0.97713f,0.97937f,0.98022f,0.98103f,0.98267f,0.98427f,0.98571f,0.98756f,0.98964f,0.99139f,0.99321f,0.99550f,0.99771f,1.00000f,0.00000f,0.90437f,0.93159f,0.96129f,0.97207f,0.97515f,0.97681f,0.97484f,0.97316f,0.97219f,0.97275f,0.97318f,0.97303f,0.97211f,0.97375f,0.97370f,0.97424f,0.97508f,0.97635f,0.97723f,0.97856f,0.97972f,0.98106f,0.98256f,0.98395f,0.98535f,0.98756f,0.98935f,0.99100f,0.99322f,0.99540f,0.99760f,1.00000f,0.00000f,0.90750f,0.91764f,0.93627f,0.95630f,0.96833f,0.97225f,0.97288f,0.97290f,0.97275f,0.97289f,0.97114f,0.97258f,0.97244f,0.97342f,0.97298f,0.97400f,0.97494f,0.97595f,0.97714f,0.97850f,0.97969f,0.98099f,0.98216f,0.98389f,0.98547f,0.98736f,0.98931f,0.99116f,0.99328f,0.99531f,0.99749f,1.00000f,0.00000f,0.92013f,0.90872f,0.92143f,0.93851f,0.95406f,0.96362f,0.96822f,0.97095f,0.97139f,0.97153f,0.97173f,0.97257f,0.97344f,0.97248f,0.97420f,0.97319f,0.97494f,0.97616f,0.97666f,0.97852f,0.97948f,0.98109f,0.98273f,0.98417f,0.98563f,0.98734f,0.98918f,0.99117f,0.99323f,0.99533f,0.99755f,1.00000f,0.00000f,0.92501f,0.91129f,0.91463f,0.92682f,0.93964f,0.95096f,0.95965f,0.96630f,0.96751f,0.97010f,0.97004f,0.97292f,0.97174f,0.97326f,0.97272f,0.97395f,0.97486f,0.97499f,0.97654f,0.97779f,0.97921f,0.98070f,0.98185f,0.98361f,0.98544f,0.98685f,0.98871f,0.99112f,0.99310f,0.99519f,0.99756f,1.00000f,0.00000f,0.93092f,0.91645f,0.91068f,0.92227f,0.92575f,0.93783f,0.94691f,0.95366f,0.96019f,0.96346f,0.96764f,0.96985f,0.97048f,0.97187f,0.97292f,0.97393f,0.97483f,0.97495f,0.97653f,0.97728f,0.97935f,0.98058f,0.98220f,0.98357f,0.98523f,0.98692f,0.98893f,0.99120f,0.99280f,0.99524f,0.99749f,1.00000f,0.00000f,0.94222f,0.91202f,0.91863f,0.91276f,0.92114f,0.93042f,0.93698f,0.94294f,0.95049f,0.95832f,0.96032f,0.96521f,0.96830f,0.96936f,0.97159f,0.97220f,0.97363f,0.97492f,0.97729f,0.97768f,0.97901f,0.98065f,0.98201f,0.98378f,0.98503f,0.98668f,0.98853f,0.99044f,0.99273f,0.99500f,0.99744f,1.00000f,0.00000f,0.94776f,0.92890f,0.91923f,0.91735f,0.91916f,0.92541f,0.92830f,0.93792f,0.94239f,0.94817f,0.95380f,0.95793f,0.96406f,0.96562f,0.96765f,0.96989f,0.97274f,0.97339f,0.97614f,0.97685f,0.97849f,0.97937f,0.98186f,0.98324f,0.98477f,0.98640f,0.98870f,0.99068f,0.99274f,0.99490f,0.99726f,1.00000f,0.00000f,0.94950f,0.93245f,0.92479f,0.91266f,0.91630f,0.91955f,0.92794f,0.92723f,0.93265f,0.94035f,0.94505f,0.95121f,0.95416f,0.95994f,0.96292f,0.96603f,0.96913f,0.97211f,0.97303f,0.97612f,0.97723f,0.98013f,0.98102f,0.98286f,0.98453f,0.98627f,0.98850f,0.99037f,0.99256f,0.99506f,0.99725f,1.00000f,0.00000f,0.95246f,0.94501f,0.92268f,0.92218f,0.92173f,0.92186f,0.92415f,0.92731f,0.92964f,0.93236f,0.93907f,0.94114f,0.94881f,0.95389f,0.95567f,0.96109f,0.96505f,0.96838f,0.97055f,0.97296f,0.97600f,0.97756f,0.97896f,0.98270f,0.98455f,0.98651f,0.98777f,0.99035f,0.99244f,0.99507f,0.99745f,1.00000f,0.00000f,0.95632f,0.94730f,0.93513f,0.92455f,0.92190f,0.91908f,0.92277f,0.92384f,0.92462f,0.92963f,0.93346f,0.93655f,0.94044f,0.94625f,0.94989f,0.95447f,0.95838f,0.96312f,0.96601f,0.96954f,0.97310f,0.97517f,0.97726f,0.98026f,0.98201f,0.98551f,0.98768f,0.99006f,0.99266f,0.99440f,0.99710f,1.00000f,0.00000f,0.95680f,0.95066f,0.93344f,0.92700f,0.92351f,0.91935f,0.92573f,0.92647f,0.92547f,0.92519f,0.92908f,0.93111f,0.93546f,0.94031f,0.94395f,0.94861f,0.95286f,0.96025f,0.95972f,0.96514f,0.96796f,0.97239f,0.97402f,0.97694f,0.98039f,0.98305f,0.98650f,0.98887f,0.99168f,0.99440f,0.99731f,1.00000f,0.00000f,0.95853f,0.95545f,0.94249f,0.93369f,0.92445f,0.92075f,0.92150f,0.92744f,0.92621f,0.92743f,0.93002f,0.93163f,0.93263f,0.93446f,0.93691f,0.94269f,0.94782f,0.95006f,0.95489f,0.95910f,0.96365f,0.96753f,0.97065f,0.97417f,0.97726f,0.98170f,0.98437f,0.98681f,0.99115f,0.99404f,0.99685f,1.00000f,0.00000f,0.96157f,0.96120f,0.94990f,0.93848f,0.93338f,0.92338f,0.92977f,0.92209f,0.92312f,0.92383f,0.92669f,0.92647f,0.93225f,0.93590f,0.93335f,0.93830f,0.94029f,0.94577f,0.94788f,0.95255f,0.95801f,0.96158f,0.96398f,0.96970f,0.97280f,0.97738f,0.98107f,0.98498f,0.98892f,0.99256f,0.99565f,1.00000f,0.00000f,0.96202f,0.96041f,0.94960f,0.94217f,0.93394f,0.93437f,0.93152f,0.92694f,0.92214f,0.92662f,0.92487f,0.93171f,0.92948f,0.93066f,0.93370f,0.93415f,0.93696f,0.94118f,0.94362f,0.94692f,0.94901f,0.95455f,0.96064f,0.96416f,0.96882f,0.97323f,0.97711f,0.98023f,0.98521f,0.98941f,0.99383f,0.99991f,0.00000f,0.96619f,0.95157f,0.95264f,0.94491f,0.93870f,0.93554f,0.93178f,0.93059f,0.92667f,0.92659f,0.92711f,0.92701f,0.92770f,0.92903f,0.92983f,0.93067f,0.93547f,0.93809f,0.93883f,0.94243f,0.94795f,0.95142f,0.95453f,0.95728f,0.96139f,0.96734f,0.97144f,0.97520f,0.98127f,0.98448f,0.99012f,0.99942f,0.00000f,0.96929f,0.95890f,0.95370f,0.94427f,0.94026f,0.93636f,0.93527f,0.93231f,0.93035f,0.92912f,0.92942f,0.92929f,0.92995f,0.93016f,0.93031f,0.93323f,0.93125f,0.93578f,0.93487f,0.93446f,0.93892f,0.94186f,0.94813f,0.95139f,0.95601f,0.95956f,0.96398f,0.96992f,0.97435f,0.97936f,0.98440f,0.99787f,0.00000f,0.96097f,0.96331f,0.96043f,0.94999f,0.94948f,0.94434f,0.93853f,0.93737f,0.92775f,0.92947f,0.92606f,0.92789f,0.92777f,0.93228f,0.93150f,0.92905f,0.92988f,0.93182f,0.93407f,0.93347f,0.93445f,0.93675f,0.94224f,0.94481f,0.94885f,0.95327f,0.95750f,0.96315f,0.96659f,0.97222f,0.97657f,0.99451f,0.00000f,0.96272f,0.96674f,0.95603f,0.95552f,0.94952f,0.94421f,0.94351f,0.93816f,0.93312f,0.93134f,0.92697f,0.92929f,0.92926f,0.92983f,0.92844f,0.92866f,0.92930f,0.92928f,0.93150f,0.93786f,0.93337f,0.93314f,0.93769f,0.93966f,0.94339f,0.94868f,0.95014f,0.95504f,0.95953f,0.96481f,0.96982f,0.98845f,0.00000f,0.95323f,0.96565f,0.96031f,0.96198f,0.95789f,0.94377f,0.94029f,0.94056f,0.93826f,0.93312f,0.93396f,0.92826f,0.93224f,0.92791f,0.93278f,0.92543f,0.92551f,0.92740f,0.92756f,0.92967f,0.92789f,0.92511f,0.93161f,0.93106f,0.93592f,0.93932f,0.94071f,0.94637f,0.95136f,0.95116f,0.95687f,0.98126f,0.00000f,0.95478f,0.95497f,0.94976f,0.94973f,0.94884f,0.94466f,0.94456f,0.94045f,0.93713f,0.93433f,0.92983f,0.93203f,0.92472f,0.92495f,0.92465f,0.93086f,0.92597f,0.92694f,0.92649f,0.92707f,0.92566f,0.92563f,0.92624f,0.92980f,0.92746f,0.93107f,0.93390f,0.93750f,0.93987f,0.94219f,0.94635f,0.97176f,0.00000f,0.94463f,0.94441f,0.94892f,0.94030f,0.95037f,0.93991f,0.93422f,0.93740f,0.93502f,0.93354f,0.93387f,0.92976f,0.92607f,0.93035f,0.92368f,0.92448f,0.92362f,0.92503f,0.92703f,0.92346f,0.92544f,0.92216f,0.92203f,0.91878f,0.92231f,0.92199f,0.92542f,0.92873f,0.92859f,0.93115f,0.93406f,0.96208f,0.00000f,0.92729f,0.94339f,0.93880f,0.93744f,0.93633f,0.94292f,0.93168f,0.93647f,0.93385f,0.92532f,0.92745f,0.92773f,0.92475f,0.91933f,0.92297f,0.92124f,0.92291f,0.92267f,0.91953f,0.92004f,0.91757f,0.91865f,0.91519f,0.91711f,0.91348f,0.91477f,0.91630f,0.91614f,0.91809f,0.92040f,0.92048f,0.94906f,0.00000f,0.92608f,0.92705f,0.93144f,0.93643f,0.93274f,0.92831f,0.92905f,0.92585f,0.92460f,0.92819f,0.92510f,0.92305f,0.91956f,0.92322f,0.91452f,0.91832f,0.91727f,0.91804f,0.91271f,0.91395f,0.91680f,0.90933f,0.91154f,0.91077f,0.90767f,0.90353f,0.90445f,0.90512f,0.90564f,0.90810f,0.90935f,0.93829f,0.00000f,0.90677f,0.91323f,0.91795f,0.91727f,0.91598f,0.92510f,0.91693f,0.91567f,0.91631f,0.91411f,0.91389f,0.92063f,0.91474f,0.90920f,0.91731f,0.91431f,0.90981f,0.90494f,0.90906f,0.90605f,0.90539f,0.90408f,0.90107f,0.90234f,0.89894f,0.89581f,0.89336f,0.89073f,0.89411f,0.89654f,0.89850f,0.92708f,0.00000f,0.88609f,0.89034f,0.90016f,0.90469f,0.90728f,0.90930f,0.90803f,0.90971f,0.90952f,0.90306f,0.90709f,0.90535f,0.90664f,0.90412f,0.90124f,0.90514f,0.90167f,0.90364f,0.90185f,0.89717f,0.90189f,0.89599f,0.88926f,0.89067f,0.89131f,0.88623f,0.88366f,0.87965f,0.87972f,0.88406f,0.88469f,0.91556f,0.00000f,0.86271f,0.87317f,0.88350f,0.87649f,0.88832f,0.88839f,0.89311f,0.89126f,0.89938f,0.89489f,0.89126f,0.89656f,0.89386f,0.89633f,0.89814f,0.89205f,0.89194f,0.88675f,0.89078f,0.89062f,0.88541f,0.88441f,0.88127f,0.88099f,0.88177f,0.87696f,0.87318f,0.87127f,0.86995f,0.87124f,0.87224f,0.90374f,0.00000f,0.84555f,0.85145f,0.85955f,0.86366f,0.86009f,0.86991f,0.86897f,0.87923f,0.87656f,0.87808f,0.87542f,0.87885f,0.87757f,0.88117f,0.88251f,0.87972f,0.87696f,0.88217f,0.88008f,0.88226f,0.87770f,0.87336f,0.87343f,0.87016f,0.86822f,0.86313f,0.86076f,0.85912f,0.85765f,0.86087f,0.86040f,0.89169f,
 0.00000f,0.99393f,0.98293f,0.97520f,0.96264f,0.94817f,0.95986f,0.93082f,0.95859f,0.91857f,0.91827f,0.94298f,0.98607f,0.93371f,0.96123f,0.96512f,0.96992f,0.96792f,0.96901f,0.97156f,0.96438f,0.96360f,0.97360f,0.97465f,0.96817f,0.98363f,0.97888f,0.98253f,0.98267f,0.99240f,0.99250f,0.99719f,1.00000f,0.00000f,0.98836f,0.98477f,0.97788f,0.97666f,0.97387f,0.97173f,0.96982f,0.96903f,0.96724f,0.96713f,0.96755f,0.96855f,0.96829f,0.96935f,0.97196f,0.97341f,0.97439f,0.97487f,0.97630f,0.97625f,0.97741f,0.97945f,0.98161f,0.98238f,0.98505f,0.98611f,0.98870f,0.99047f,0.99262f,0.99503f,0.99762f,1.00000f,0.00000f,0.99042f,0.98621f,0.98266f,0.97959f,0.97795f,0.97680f,0.97431f,0.97376f,0.97381f,0.97318f,0.97380f,0.97497f,0.97305f,0.97425f,0.97475f,0.97547f,0.97597f,0.97641f,0.97739f,0.97899f,0.97976f,0.98112f,0.98263f,0.98398f,0.98590f,0.98768f,0.98965f,0.99132f,0.99330f,0.99535f,0.99775f,1.00000f,0.00000f,0.98524f,0.98562f,0.98171f,0.98014f,0.97737f,0.97692f,0.97606f,0.97429f,0.97354f,0.97271f,0.97342f,0.97294f,0.97316f,0.97360f,0.97475f,0.97457f,0.97623f,0.97627f,0.97785f,0.97889f,0.98023f,0.98153f,0.98268f,0.98439f,0.98577f,0.98754f,0.98954f,0.99141f,0.99323f,0.99550f,0.99774f,1.00000f,0.00000f,0.94114f,0.98150f,0.97980f,0.97858f,0.97801f,0.97489f,0.97538f,0.97476f,0.97368f,0.97258f,0.97325f,0.97435f,0.97376f,0.97385f,0.97414f,0.97466f,0.97594f,0.97649f,0.97701f,0.97860f,0.97927f,0.98107f,0.98264f,0.98416f,0.98594f,0.98770f,0.98941f,0.99131f,0.99358f,0.99546f,0.99775f,1.00000f,0.00000f,0.91197f,0.95967f,0.97677f,0.97836f,0.97655f,0.97561f,0.97426f,0.97490f,0.97332f,0.97237f,0.97354f,0.97263f,0.97408f,0.97305f,0.97440f,0.97402f,0.97530f,0.97663f,0.97745f,0.97956f,0.97963f,0.98126f,0.98261f,0.98381f,0.98586f,0.98778f,0.98956f,0.99130f,0.99334f,0.99550f,0.99775f,1.00000f,0.00000f,0.90616f,0.93022f,0.96063f,0.97319f,0.97432f,0.97562f,0.97399f,0.97313f,0.97449f,0.97260f,0.97279f,0.97173f,0.97291f,0.97356f,0.97384f,0.97515f,0.97483f,0.97655f,0.97739f,0.97877f,0.97962f,0.98172f,0.98266f,0.98397f,0.98576f,0.98747f,0.98922f,0.99117f,0.99330f,0.99539f,0.99759f,1.00000f,0.00000f,0.90761f,0.91555f,0.93704f,0.95627f,0.96838f,0.97133f,0.97186f,0.97374f,0.97170f,0.97306f,0.97270f,0.97247f,0.97313f,0.97300f,0.97370f,0.97438f,0.97529f,0.97672f,0.97714f,0.97806f,0.97937f,0.98156f,0.98196f,0.98393f,0.98589f,0.98778f,0.98917f,0.99112f,0.99305f,0.99527f,0.99753f,1.00000f,0.00000f,0.91652f,0.91001f,0.92307f,0.93953f,0.95411f,0.96241f,0.96877f,0.97025f,0.97119f,0.97109f,0.97246f,0.97137f,0.97322f,0.97243f,0.97327f,0.97403f,0.97516f,0.97620f,0.97693f,0.97822f,0.97972f,0.98013f,0.98212f,0.98387f,0.98570f,0.98730f,0.98913f,0.99115f,0.99324f,0.99530f,0.99754f,1.00000f,0.00000f,0.92923f,0.91261f,0.91522f,0.92621f,0.93926f,0.94726f,0.95885f,0.96383f,0.96809f,0.96946f,0.96972f,0.97207f,0.97158f,0.97288f,0.97360f,0.97427f,0.97511f,0.97599f,0.97643f,0.97807f,0.97927f,0.98016f,0.98130f,0.98439f,0.98521f,0.98716f,0.98915f,0.99101f,0.99300f,0.99517f,0.99748f,1.00000f,0.00000f,0.92638f,0.91285f,0.90888f,0.91880f,0.92873f,0.93567f,0.94645f,0.95523f,0.96042f,0.96401f,0.96800f,0.96779f,0.97033f,0.97250f,0.97265f,0.97370f,0.97483f,0.97540f,0.97630f,0.97731f,0.97851f,0.98062f,0.98135f,0.98338f,0.98516f,0.98728f,0.98875f,0.99060f,0.99288f,0.99507f,0.99750f,1.00000f,0.00000f,0.93231f,0.91586f,0.91318f,0.91511f,0.92664f,0.92682f,0.93599f,0.94159f,0.95070f,0.95743f,0.96171f,0.96410f,0.96799f,0.97019f,0.97186f,0.97200f,0.97471f,0.97518f,0.97717f,0.97721f,0.97855f,0.97984f,0.98155f,0.98391f,0.98486f,0.98668f,0.98860f,0.99085f,0.99260f,0.99499f,0.99747f,1.00000f,0.00000f,0.94396f,0.92275f,0.91366f,0.91440f,0.91602f,0.92514f,0.92962f,0.93296f,0.94024f,0.94827f,0.95386f,0.95915f,0.96127f,0.96532f,0.96932f,0.97032f,0.97323f,0.97305f,0.97483f,0.97671f,0.97820f,0.97954f,0.98147f,0.98293f,0.98472f,0.98692f,0.98851f,0.99058f,0.99288f,0.99492f,0.99736f,1.00000f,0.00000f,0.95348f,0.92831f,0.92175f,0.91898f,0.91978f,0.92191f,0.92468f,0.93101f,0.93121f,0.94197f,0.94557f,0.95247f,0.95569f,0.95904f,0.96301f,0.96655f,0.96936f,0.97115f,0.97392f,0.97447f,0.97787f,0.97968f,0.98025f,0.98257f,0.98457f,0.98638f,0.98894f,0.99071f,0.99280f,0.99504f,0.99727f,1.00000f,0.00000f,0.94807f,0.94399f,0.92855f,0.91930f,0.91802f,0.92014f,0.92431f,0.92847f,0.92832f,0.93702f,0.93730f,0.94477f,0.94816f,0.95440f,0.95796f,0.96259f,0.96502f,0.96691f,0.97044f,0.97374f,0.97546f,0.97676f,0.98020f,0.98187f,0.98428f,0.98573f,0.98782f,0.99003f,0.99286f,0.99464f,0.99738f,1.00000f,0.00000f,0.95979f,0.93936f,0.92883f,0.92542f,0.92299f,0.92074f,0.92369f,0.92489f,0.92637f,0.93030f,0.93325f,0.93739f,0.94506f,0.94378f,0.95018f,0.95551f,0.95810f,0.96175f,0.96527f,0.97043f,0.97264f,0.97486f,0.97850f,0.98092f,0.98289f,0.98546f,0.98783f,0.99044f,0.99184f,0.99465f,0.99701f,1.00000f,0.00000f,0.95382f,0.95077f,0.93493f,0.93029f,0.92310f,0.92057f,0.92160f,0.92210f,0.92709f,0.92996f,0.93237f,0.93108f,0.93859f,0.94013f,0.94521f,0.95001f,0.95330f,0.95769f,0.96164f,0.96401f,0.96929f,0.97159f,0.97484f,0.97808f,0.98128f,0.98330f,0.98645f,0.98954f,0.99186f,0.99399f,0.99705f,1.00000f,0.00000f,0.95945f,0.95298f,0.94289f,0.93593f,0.92623f,0.92356f,0.92427f,0.92608f,0.92179f,0.92641f,0.92688f,0.93078f,0.93349f,0.93629f,0.93776f,0.94341f,0.94728f,0.95142f,0.95351f,0.96009f,0.96311f,0.96726f,0.96972f,0.97457f,0.97723f,0.98017f,0.98461f,0.98735f,0.99032f,0.99403f,0.99701f,1.00000f,0.00000f,0.95909f,0.95424f,0.94455f,0.93335f,0.93375f,0.93271f,0.92791f,0.92348f,0.92652f,0.92669f,0.92844f,0.92894f,0.92930f,0.93489f,0.93554f,0.93717f,0.94283f,0.94591f,0.94812f,0.95392f,0.95634f,0.96121f,0.96644f,0.96887f,0.97225f,0.97562f,0.97941f,0.98493f,0.98811f,0.99226f,0.99584f,1.00000f,0.00000f,0.96409f,0.96495f,0.94552f,0.94413f,0.93616f,0.92746f,0.92767f,0.92607f,0.92295f,0.92909f,0.93022f,0.92900f,0.92772f,0.92617f,0.93277f,0.93499f,0.93653f,0.94224f,0.94295f,0.94808f,0.94909f,0.95439f,0.96057f,0.96526f,0.96816f,0.97224f,0.97604f,0.98020f,0.98349f,0.98873f,0.99380f,0.99990f,0.00000f,0.96366f,0.95366f,0.95486f,0.94293f,0.94341f,0.93515f,0.93019f,0.93260f,0.92579f,0.92919f,0.92629f,0.92885f,0.92849f,0.92689f,0.93166f,0.93219f,0.93322f,0.93679f,0.94005f,0.93825f,0.94943f,0.94714f,0.95196f,0.95818f,0.96189f,0.96707f,0.96964f,0.97525f,0.97999f,0.98531f,0.98948f,0.99939f,0.00000f,0.97061f,0.96204f,0.95368f,0.95009f,0.94028f,0.94336f,0.93340f,0.93661f,0.93081f,0.92835f,0.92796f,0.92861f,0.92741f,0.93270f,0.92967f,0.93055f,0.93205f,0.93451f,0.93214f,0.93730f,0.94274f,0.94087f,0.94771f,0.95121f,0.95424f,0.95956f,0.96328f,0.97026f,0.97335f,0.97854f,0.98489f,0.99790f,0.00000f,0.96685f,0.96480f,0.95987f,0.95106f,0.94568f,0.94788f,0.93894f,0.93753f,0.93304f,0.92950f,0.93376f,0.92968f,0.92611f,0.92653f,0.93229f,0.93216f,0.92869f,0.93284f,0.93574f,0.93400f,0.93442f,0.93763f,0.93944f,0.94623f,0.94858f,0.95242f,0.95657f,0.96246f,0.96709f,0.97026f,0.97614f,0.99399f,0.00000f,0.95960f,0.96346f,0.95869f,0.95656f,0.94944f,0.94182f,0.94193f,0.93488f,0.93480f,0.93691f,0.92950f,0.92869f,0.93204f,0.92951f,0.92660f,0.92760f,0.92805f,0.92921f,0.93030f,0.93081f,0.93028f,0.93269f,0.93482f,0.93884f,0.94129f,0.94468f,0.95071f,0.95352f,0.95803f,0.96188f,0.96639f,0.98807f,0.00000f,0.96068f,0.96035f,0.96023f,0.95949f,0.95180f,0.94100f,0.93919f,0.94212f,0.93668f,0.92876f,0.92711f,0.92685f,0.92791f,0.92857f,0.92938f,0.92749f,0.92987f,0.93007f,0.92826f,0.92800f,0.92843f,0.92747f,0.92778f,0.93297f,0.93582f,0.93808f,0.93857f,0.94558f,0.94708f,0.95168f,0.95720f,0.97959f,0.00000f,0.95871f,0.95867f,0.95430f,0.95706f,0.95022f,0.94609f,0.94129f,0.93749f,0.93816f,0.93578f,0.93075f,0.93222f,0.92828f,0.93124f,0.92681f,0.93100f,0.92781f,0.92779f,0.92417f,0.92445f,0.92846f,0.92784f,0.92270f,0.92614f,0.92799f,0.93058f,0.92922f,0.93338f,0.93841f,0.94174f,0.94393f,0.97144f,0.00000f,0.94286f,0.95219f,0.94643f,0.94854f,0.94228f,0.94677f,0.94402f,0.93283f,0.93516f,0.93068f,0.93246f,0.93279f,0.92948f,0.92568f,0.92675f,0.92627f,0.92097f,0.92359f,0.92205f,0.92415f,0.92201f,0.92149f,0.92140f,0.92087f,0.92258f,0.92262f,0.92409f,0.92515f,0.92769f,0.92985f,0.93111f,0.95961f,0.00000f,0.93869f,0.94427f,0.94051f,0.94426f,0.93441f,0.94116f,0.93464f,0.93786f,0.93256f,0.92995f,0.92486f,0.92601f,0.92173f,0.92183f,0.92343f,0.92264f,0.91985f,0.92141f,0.91912f,0.91670f,0.91648f,0.91451f,0.91541f,0.91088f,0.91231f,0.91373f,0.91342f,0.91394f,0.91497f,0.91849f,0.91879f,0.94831f,0.00000f,0.93397f,0.93165f,0.93195f,0.93182f,0.93039f,0.93143f,0.93550f,0.92909f,0.92508f,0.92978f,0.92634f,0.92236f,0.92377f,0.91967f,0.91994f,0.92131f,0.91608f,0.91325f,0.91548f,0.91495f,0.91488f,0.91338f,0.90931f,0.90654f,0.90797f,0.90001f,0.90498f,0.90237f,0.90460f,0.90462f,0.90518f,0.93609f,0.00000f,0.91168f,0.91304f,0.90926f,0.91967f,0.91654f,0.92603f,0.92318f,0.91013f,0.92054f,0.91772f,0.91844f,0.90804f,0.91254f,0.91573f,0.90745f,0.90738f,0.90647f,0.90952f,0.90548f,0.90622f,0.90747f,0.89944f,0.90302f,0.89519f,0.89573f,0.89289f,0.89303f,0.89006f,0.89154f,0.89309f,0.89350f,0.92332f,0.00000f,0.88869f,0.89980f,0.90103f,0.90485f,0.90795f,0.90693f,0.90800f,0.90789f,0.90449f,0.91241f,0.90593f,0.90485f,0.90561f,0.90209f,0.90754f,0.90144f,0.90961f,0.90312f,0.89627f,0.89673f,0.89192f,0.89005f,0.89085f,0.89284f,0.88707f,0.88324f,0.88165f,0.87881f,0.87794f,0.88058f,0.88122f,0.91139f,0.00000f,0.87777f,0.87774f,0.88672f,0.88608f,0.89177f,0.88989f,0.89011f,0.89303f,0.89540f,0.89764f,0.89276f,0.89220f,0.89691f,0.88932f,0.89265f,0.88784f,0.88847f,0.89172f,0.88741f,0.88700f,0.88515f,0.88610f,0.87783f,0.87706f,0.87381f,0.87334f,0.86877f,0.86942f,0.86663f,0.86798f,0.86786f,0.89925f,0.00000f,0.85763f,0.85714f,0.86338f,0.86427f,0.86386f,0.86764f,0.87296f,0.87296f,0.87839f,0.87603f,0.87870f,0.88153f,0.87736f,0.88072f,0.87937f,0.88213f,0.87930f,0.87439f,0.87703f,0.87705f,0.87396f,0.87221f,0.86810f,0.86575f,0.86487f,0.85905f,0.85765f,0.85242f,0.85384f,0.85504f,0.85670f,0.88753f,
 0.00000f,0.99702f,0.97353f,0.97205f,0.96760f,0.94226f,0.95939f,0.95966f,0.95348f,0.94985f,0.94834f,0.96168f,0.95747f,0.95596f,0.93802f,0.95590f,0.95398f,0.96643f,0.96563f,0.96827f,0.95098f,0.95752f,0.97488f,0.97820f,0.98146f,0.98183f,0.98546f,0.98598f,0.98040f,0.98608f,0.99355f,0.99736f,1.00000f,0.00000f,0.98980f,0.98439f,0.97978f,0.97766f,0.97289f,0.97217f,0.97217f,0.97048f,0.96747f,0.96739f,0.96861f,0.96794f,0.96755f,0.96827f,0.97129f,0.97083f,0.97409f,0.97516f,0.97611f,0.97578f,0.97773f,0.97986f,0.98138f,0.98329f,0.98495f,0.98671f,0.98838f,0.99041f,0.99267f,0.99493f,0.99765f,1.00000f,0.00000f,0.99295f,0.98489f,0.98180f,0.97900f,0.97776f,0.97606f,0.97589f,0.97480f,0.97442f,0.97240f,0.97277f,0.97327f,0.97385f,0.97403f,0.97431f,0.97496f,0.97625f,0.97652f,0.97762f,0.97851f,0.97991f,0.98118f,0.98240f,0.98403f,0.98586f,0.98771f,0.98945f,0.99138f,0.99331f,0.99548f,0.99773f,1.00000f,0.00000f,0.98471f,0.98661f,0.98190f,0.98155f,0.97851f,0.97602f,0.97454f,0.97442f,0.97381f,0.97373f,0.97412f,0.97284f,0.97351f,0.97423f,0.97444f,0.97440f,0.97544f,0.97654f,0.97754f,0.97895f,0.98024f,0.98136f,0.98305f,0.98424f,0.98589f,0.98763f,0.98938f,0.99117f,0.99331f,0.99550f,0.99776f,1.00000f,0.00000f,0.94350f,0.98218f,0.98023f,0.97830f,0.97637f,0.97645f,0.97493f,0.97330f,0.97330f,0.97275f,0.97214f,0.97331f,0.97295f,0.97421f,0.97442f,0.97486f,0.97592f,0.97625f,0.97767f,0.97838f,0.97970f,0.98089f,0.98279f,0.98403f,0.98554f,0.98755f,0.98943f,0.99132f,0.99332f,0.99541f,0.99775f,1.00000f,0.00000f,0.91669f,0.96036f,0.97663f,0.97886f,0.97652f,0.97415f,0.97478f,0.97297f,0.97349f,0.97337f,0.97284f,0.97324f,0.97340f,0.97369f,0.97448f,0.97463f,0.97556f,0.97657f,0.97791f,0.97793f,0.97995f,0.98196f,0.98251f,0.98409f,0.98595f,0.98761f,0.98915f,0.99110f,0.99331f,0.99541f,0.99780f,1.00000f,0.00000f,0.91054f,0.93227f,0.96203f,0.97337f,0.97444f,0.97633f,0.97426f,0.97246f,0.97314f,0.97284f,0.97272f,0.97289f,0.97272f,0.97365f,0.97376f,0.97405f,0.97549f,0.97632f,0.97724f,0.97873f,0.98010f,0.98116f,0.98276f,0.98419f,0.98569f,0.98729f,0.98933f,0.99114f,0.99315f,0.99538f,0.99762f,1.00000f,0.00000f,0.90812f,0.91365f,0.93850f,0.95820f,0.96769f,0.97229f,0.97356f,0.97273f,0.97180f,0.97238f,0.97254f,0.97334f,0.97307f,0.97266f,0.97368f,0.97502f,0.97573f,0.97670f,0.97687f,0.97813f,0.97961f,0.98116f,0.98299f,0.98421f,0.98570f,0.98710f,0.98912f,0.99101f,0.99317f,0.99527f,0.99752f,1.00000f,0.00000f,0.92035f,0.91006f,0.92071f,0.93874f,0.95523f,0.96438f,0.96805f,0.97123f,0.97130f,0.97310f,0.97128f,0.97254f,0.97322f,0.97315f,0.97366f,0.97475f,0.97438f,0.97676f,0.97694f,0.97883f,0.97899f,0.98026f,0.98233f,0.98333f,0.98569f,0.98699f,0.98893f,0.99120f,0.99319f,0.99547f,0.99746f,1.00000f,0.00000f,0.92711f,0.91477f,0.91524f,0.92577f,0.93896f,0.95270f,0.95832f,0.96447f,0.96733f,0.96899f,0.97226f,0.97252f,0.97183f,0.97244f,0.97286f,0.97322f,0.97510f,0.97669f,0.97749f,0.97797f,0.97927f,0.98024f,0.98195f,0.98395f,0.98568f,0.98728f,0.98889f,0.99122f,0.99283f,0.99528f,0.99739f,1.00000f,0.00000f,0.93486f,0.91453f,0.91362f,0.92120f,0.92761f,0.93708f,0.94948f,0.95768f,0.96251f,0.96405f,0.96620f,0.96940f,0.97030f,0.97165f,0.97231f,0.97447f,0.97436f,0.97588f,0.97637f,0.97747f,0.97924f,0.98068f,0.98181f,0.98344f,0.98547f,0.98714f,0.98920f,0.99055f,0.99308f,0.99508f,0.99745f,1.00000f,0.00000f,0.93807f,0.91975f,0.91684f,0.91771f,0.92220f,0.92672f,0.93678f,0.94248f,0.95168f,0.95786f,0.96107f,0.96554f,0.96874f,0.97032f,0.97074f,0.97300f,0.97483f,0.97518f,0.97659f,0.97805f,0.97831f,0.98004f,0.98147f,0.98305f,0.98528f,0.98682f,0.98880f,0.99077f,0.99280f,0.99517f,0.99733f,1.00000f,0.00000f,0.94571f,0.93082f,0.91745f,0.91373f,0.92194f,0.92426f,0.92737f,0.93570f,0.94116f,0.94838f,0.95163f,0.95867f,0.96408f,0.96494f,0.96904f,0.96967f,0.97172f,0.97323f,0.97508f,0.97726f,0.97891f,0.98030f,0.98104f,0.98296f,0.98515f,0.98640f,0.98887f,0.99077f,0.99271f,0.99494f,0.99725f,1.00000f,0.00000f,0.95501f,0.93156f,0.92110f,0.91861f,0.91667f,0.92271f,0.92572f,0.92983f,0.93222f,0.93854f,0.94711f,0.95114f,0.95719f,0.95770f,0.96378f,0.96688f,0.96947f,0.97054f,0.97463f,0.97454f,0.97741f,0.97881f,0.98103f,0.98241f,0.98506f,0.98657f,0.98793f,0.98995f,0.99238f,0.99500f,0.99735f,1.00000f,0.00000f,0.95688f,0.93795f,0.92393f,0.92163f,0.92139f,0.92350f,0.92336f,0.92891f,0.92960f,0.93065f,0.93963f,0.94163f,0.94983f,0.95167f,0.95569f,0.95965f,0.96506f,0.96695f,0.96947f,0.97270f,0.97549f,0.97850f,0.98011f,0.98294f,0.98412f,0.98563f,0.98880f,0.99070f,0.99278f,0.99497f,0.99735f,1.00000f,0.00000f,0.95809f,0.93966f,0.93160f,0.92507f,0.92126f,0.92175f,0.92230f,0.92562f,0.92854f,0.92716f,0.93226f,0.93498f,0.94089f,0.94556f,0.94990f,0.95398f,0.95895f,0.96156f,0.96586f,0.96961f,0.97253f,0.97537f,0.97777f,0.97968f,0.98311f,0.98518f,0.98631f,0.99057f,0.99269f,0.99455f,0.99713f,1.00000f,0.00000f,0.95671f,0.95074f,0.93937f,0.93381f,0.92678f,0.92061f,0.92367f,0.92616f,0.92693f,0.92613f,0.92836f,0.93228f,0.93718f,0.94060f,0.94455f,0.94626f,0.95212f,0.95393f,0.96086f,0.96524f,0.96759f,0.97241f,0.97484f,0.97753f,0.98093f,0.98330f,0.98620f,0.98925f,0.99190f,0.99432f,0.99731f,1.00000f,0.00000f,0.95932f,0.94841f,0.94549f,0.93284f,0.93178f,0.92900f,0.92010f,0.92176f,0.92265f,0.92473f,0.93016f,0.93205f,0.93227f,0.93390f,0.94012f,0.94152f,0.94790f,0.95140f,0.95540f,0.95849f,0.96415f,0.96647f,0.96942f,0.97396f,0.97774f,0.98005f,0.98336f,0.98785f,0.99065f,0.99334f,0.99655f,1.00000f,0.00000f,0.96137f,0.95194f,0.94568f,0.93555f,0.92636f,0.92769f,0.92294f,0.92913f,0.92693f,0.92818f,0.92808f,0.93035f,0.92879f,0.93276f,0.93557f,0.93777f,0.94233f,0.94407f,0.94781f,0.95111f,0.95840f,0.96082f,0.96520f,0.96761f,0.97099f,0.97732f,0.98040f,0.98439f,0.98829f,0.99149f,0.99614f,1.00000f,0.00000f,0.95952f,0.95858f,0.94605f,0.93813f,0.93726f,0.93177f,0.92900f,0.92611f,0.92861f,0.92530f,0.92877f,0.92611f,0.92705f,0.93019f,0.93090f,0.93348f,0.93500f,0.94154f,0.94334f,0.94599f,0.95180f,0.95491f,0.95764f,0.96372f,0.96779f,0.97165f,0.97513f,0.98060f,0.98432f,0.98883f,0.99369f,0.99989f,0.00000f,0.96200f,0.96155f,0.95566f,0.94284f,0.94050f,0.93139f,0.93249f,0.93223f,0.92860f,0.92803f,0.93019f,0.92649f,0.92857f,0.92852f,0.92895f,0.93231f,0.93119f,0.93660f,0.93949f,0.93928f,0.94606f,0.95003f,0.95146f,0.95814f,0.96135f,0.96539f,0.97148f,0.97551f,0.98018f,0.98355f,0.98891f,0.99934f,0.00000f,0.95065f,0.96155f,0.96128f,0.94680f,0.93861f,0.94013f,0.93364f,0.92868f,0.92564f,0.92755f,0.92988f,0.92939f,0.92827f,0.92455f,0.92819f,0.92969f,0.93078f,0.93327f,0.93176f,0.93765f,0.93927f,0.94264f,0.94652f,0.95011f,0.95519f,0.95902f,0.96423f,0.96788f,0.97376f,0.97755f,0.98408f,0.99767f,0.00000f,0.95887f,0.96222f,0.95443f,0.94939f,0.94546f,0.94349f,0.93764f,0.93250f,0.93682f,0.92970f,0.93133f,0.92887f,0.93160f,0.92715f,0.92409f,0.92683f,0.93379f,0.93135f,0.93316f,0.93467f,0.93624f,0.93888f,0.94106f,0.94403f,0.94862f,0.95138f,0.95761f,0.96117f,0.96628f,0.97075f,0.97610f,0.99392f,0.00000f,0.96327f,0.95942f,0.95692f,0.95271f,0.95045f,0.94812f,0.94509f,0.93720f,0.93955f,0.93050f,0.92946f,0.92827f,0.92778f,0.92419f,0.92659f,0.92510f,0.92871f,0.92475f,0.93226f,0.93099f,0.93261f,0.93483f,0.93246f,0.93397f,0.94166f,0.94492f,0.94936f,0.95402f,0.95640f,0.96221f,0.96632f,0.98744f,0.00000f,0.96221f,0.96058f,0.95317f,0.95643f,0.95454f,0.95168f,0.94850f,0.93908f,0.93714f,0.93453f,0.92846f,0.93072f,0.92768f,0.92576f,0.92711f,0.93257f,0.92239f,0.92847f,0.92835f,0.92822f,0.92685f,0.92913f,0.92808f,0.92895f,0.93198f,0.93723f,0.93921f,0.94242f,0.94600f,0.95109f,0.95387f,0.97992f,0.00000f,0.95976f,0.95865f,0.96209f,0.95382f,0.94794f,0.94829f,0.94004f,0.94339f,0.93807f,0.93655f,0.92975f,0.93034f,0.92738f,0.93048f,0.92230f,0.92575f,0.92699f,0.92472f,0.92630f,0.92501f,0.92615f,0.92603f,0.92487f,0.92404f,0.92762f,0.92839f,0.92905f,0.93317f,0.93702f,0.93904f,0.94304f,0.96929f,0.00000f,0.95090f,0.95088f,0.94836f,0.95227f,0.94669f,0.94287f,0.94300f,0.93998f,0.93238f,0.93147f,0.92638f,0.92389f,0.92848f,0.92380f,0.92345f,0.92037f,0.92704f,0.92558f,0.92025f,0.91768f,0.92215f,0.91799f,0.91550f,0.91931f,0.92178f,0.92120f,0.92010f,0.92434f,0.92371f,0.92627f,0.92970f,0.95767f,0.00000f,0.94087f,0.94437f,0.94353f,0.93721f,0.94093f,0.93890f,0.94258f,0.93255f,0.93226f,0.93049f,0.92944f,0.92489f,0.92679f,0.92731f,0.92183f,0.91772f,0.91838f,0.91788f,0.91910f,0.91562f,0.91600f,0.91278f,0.91476f,0.90995f,0.91211f,0.90989f,0.91129f,0.91214f,0.91355f,0.91480f,0.91426f,0.94622f,0.00000f,0.93258f,0.92351f,0.93671f,0.93530f,0.93006f,0.93201f,0.92937f,0.92593f,0.93118f,0.92800f,0.92563f,0.92025f,0.91890f,0.91963f,0.91708f,0.91785f,0.91290f,0.91638f,0.91042f,0.91678f,0.91592f,0.90710f,0.90773f,0.90428f,0.90088f,0.90055f,0.89812f,0.90023f,0.90001f,0.90274f,0.90423f,0.93451f,0.00000f,0.91449f,0.92185f,0.91805f,0.91405f,0.91485f,0.92148f,0.91668f,0.91641f,0.91764f,0.92111f,0.91625f,0.90960f,0.91575f,0.91169f,0.90912f,0.90516f,0.91177f,0.91071f,0.90471f,0.90407f,0.90468f,0.89906f,0.90064f,0.89702f,0.89485f,0.88932f,0.89045f,0.88870f,0.88833f,0.88781f,0.88803f,0.92107f,0.00000f,0.89127f,0.89879f,0.90465f,0.89931f,0.90494f,0.90815f,0.90664f,0.90274f,0.90690f,0.90646f,0.91087f,0.90525f,0.90801f,0.90554f,0.90149f,0.90175f,0.90466f,0.90214f,0.89834f,0.89486f,0.89613f,0.88829f,0.89208f,0.88345f,0.88608f,0.87921f,0.87618f,0.87710f,0.87158f,0.87644f,0.87758f,0.90960f,0.00000f,0.87046f,0.87863f,0.88462f,0.88721f,0.89135f,0.89248f,0.88931f,0.89483f,0.89698f,0.89235f,0.89303f,0.89150f,0.89294f,0.89280f,0.89527f,0.89658f,0.89197f,0.89065f,0.88590f,0.88812f,0.88182f,0.87988f,0.87947f,0.87952f,0.87334f,0.87238f,0.86529f,0.86348f,0.86387f,0.86339f,0.86298f,0.89538f,0.00000f,0.85393f,0.85508f,0.86428f,0.86057f,0.87208f,0.87007f,0.87412f,0.87802f,0.87911f,0.87917f,0.88599f,0.88199f,0.87817f,0.87453f,0.88145f,0.87883f,0.87255f,0.87653f,0.87599f,0.86836f,0.86984f,0.86812f,0.86288f,0.86227f,0.85998f,0.85876f,0.85181f,0.85124f,0.84866f,0.85175f,0.85175f,0.88552f,
 0.00000f,0.99282f,0.98784f,0.98157f,0.95587f,0.95592f,0.95214f,0.95711f,0.95700f,0.94890f,0.95172f,0.96251f,0.92129f,0.95826f,0.95975f,0.96291f,0.93778f,0.94130f,0.97225f,0.96474f,0.94927f,0.96855f,0.97910f,0.97877f,0.98080f,0.97332f,0.98306f,0.98320f,0.98769f,0.99193f,0.99412f,0.99747f,1.00000f,0.00000f,0.99032f,0.98359f,0.98030f,0.97667f,0.97306f,0.97365f,0.97139f,0.97089f,0.96856f,0.96944f,0.97017f,0.96737f,0.96857f,0.96775f,0.97048f,0.97026f,0.97126f,0.97565f,0.97550f,0.97494f,0.97775f,0.98036f,0.98170f,0.98308f,0.98419f,0.98672f,0.98834f,0.99059f,0.99293f,0.99509f,0.99768f,1.00000f,0.00000f,0.99010f,0.98617f,0.98220f,0.97957f,0.97768f,0.97522f,0.97537f,0.97389f,0.97300f,0.97295f,0.97350f,0.97357f,0.97296f,0.97366f,0.97368f,0.97442f,0.97548f,0.97694f,0.97757f,0.97939f,0.98023f,0.98100f,0.98276f,0.98465f,0.98594f,0.98750f,0.98957f,0.99140f,0.99328f,0.99546f,0.99773f,1.00000f,0.00000f,0.98557f,0.98571f,0.98309f,0.98059f,0.97769f,0.97623f,0.97449f,0.97483f,0.97469f,0.97318f,0.97331f,0.97380f,0.97332f,0.97452f,0.97494f,0.97489f,0.97565f,0.97708f,0.97822f,0.97836f,0.97968f,0.98154f,0.98270f,0.98399f,0.98565f,0.98767f,0.98941f,0.99133f,0.99330f,0.99550f,0.99772f,1.00000f,0.00000f,0.94439f,0.98106f,0.98134f,0.97952f,0.97566f,0.97527f,0.97474f,0.97350f,0.97283f,0.97336f,0.97302f,0.97315f,0.97280f,0.97269f,0.97469f,0.97481f,0.97557f,0.97675f,0.97793f,0.97845f,0.97998f,0.98112f,0.98244f,0.98445f,0.98589f,0.98779f,0.98943f,0.99129f,0.99336f,0.99552f,0.99780f,1.00000f,0.00000f,0.91402f,0.96025f,0.97689f,0.97717f,0.97639f,0.97536f,0.97462f,0.97387f,0.97385f,0.97412f,0.97357f,0.97368f,0.97355f,0.97351f,0.97461f,0.97541f,0.97522f,0.97694f,0.97783f,0.97850f,0.97989f,0.98096f,0.98257f,0.98424f,0.98613f,0.98765f,0.98941f,0.99130f,0.99323f,0.99538f,0.99776f,1.00000f,0.00000f,0.90508f,0.93095f,0.96105f,0.97379f,0.97524f,0.97586f,0.97445f,0.97342f,0.97327f,0.97229f,0.97276f,0.97279f,0.97369f,0.97241f,0.97397f,0.97446f,0.97611f,0.97658f,0.97779f,0.97901f,0.97974f,0.98108f,0.98262f,0.98409f,0.98575f,0.98721f,0.98918f,0.99110f,0.99333f,0.99539f,0.99763f,1.00000f,0.00000f,0.91366f,0.91970f,0.93742f,0.95864f,0.97128f,0.97295f,0.97335f,0.97249f,0.97308f,0.97242f,0.97246f,0.97283f,0.97261f,0.97381f,0.97352f,0.97442f,0.97555f,0.97594f,0.97703f,0.97849f,0.97968f,0.98056f,0.98225f,0.98398f,0.98553f,0.98723f,0.98926f,0.99098f,0.99311f,0.99530f,0.99759f,1.00000f,0.00000f,0.92026f,0.91324f,0.92420f,0.93956f,0.95626f,0.96386f,0.96853f,0.96959f,0.97253f,0.97138f,0.97255f,0.97189f,0.97233f,0.97231f,0.97276f,0.97401f,0.97458f,0.97587f,0.97660f,0.97775f,0.97978f,0.98044f,0.98196f,0.98392f,0.98551f,0.98735f,0.98945f,0.99082f,0.99302f,0.99527f,0.99756f,1.00000f,0.00000f,0.92393f,0.91117f,0.91619f,0.92638f,0.93888f,0.95176f,0.95765f,0.96478f,0.96714f,0.97068f,0.97113f,0.97204f,0.97149f,0.97355f,0.97452f,0.97327f,0.97497f,0.97578f,0.97627f,0.97809f,0.97894f,0.98082f,0.98188f,0.98339f,0.98525f,0.98715f,0.98914f,0.99107f,0.99291f,0.99518f,0.99750f,1.00000f,0.00000f,0.93590f,0.91310f,0.91064f,0.91812f,0.92505f,0.93608f,0.94684f,0.95669f,0.96020f,0.96487f,0.96677f,0.97214f,0.96999f,0.97138f,0.97293f,0.97325f,0.97454f,0.97595f,0.97609f,0.97679f,0.97890f,0.97990f,0.98200f,0.98348f,0.98523f,0.98702f,0.98921f,0.99115f,0.99275f,0.99515f,0.99752f,1.00000f,0.00000f,0.93760f,0.92409f,0.91034f,0.91542f,0.92291f,0.92827f,0.93993f,0.94582f,0.95100f,0.95447f,0.96145f,0.96610f,0.96885f,0.96983f,0.97162f,0.97271f,0.97326f,0.97532f,0.97609f,0.97728f,0.97826f,0.98067f,0.98136f,0.98307f,0.98534f,0.98693f,0.98906f,0.99055f,0.99265f,0.99481f,0.99742f,1.00000f,0.00000f,0.94403f,0.92924f,0.92024f,0.91522f,0.91961f,0.92516f,0.92823f,0.93475f,0.94202f,0.94810f,0.95276f,0.95813f,0.96135f,0.96487f,0.96884f,0.96967f,0.97151f,0.97345f,0.97500f,0.97777f,0.97838f,0.98018f,0.98147f,0.98314f,0.98449f,0.98711f,0.98889f,0.99053f,0.99262f,0.99491f,0.99728f,1.00000f,0.00000f,0.95417f,0.93291f,0.92324f,0.91522f,0.91775f,0.92262f,0.92591f,0.93181f,0.93448f,0.94129f,0.94876f,0.95158f,0.95505f,0.95990f,0.96470f,0.96584f,0.96878f,0.97194f,0.97374f,0.97527f,0.97715f,0.97813f,0.98160f,0.98303f,0.98522f,0.98660f,0.98848f,0.99014f,0.99218f,0.99472f,0.99748f,1.00000f,0.00000f,0.95490f,0.94198f,0.92635f,0.91778f,0.91670f,0.91917f,0.92390f,0.92625f,0.92943f,0.93372f,0.93754f,0.94115f,0.94993f,0.95288f,0.95549f,0.96047f,0.96472f,0.96806f,0.97073f,0.97297f,0.97514f,0.97755f,0.98000f,0.98183f,0.98345f,0.98555f,0.98785f,0.99016f,0.99239f,0.99503f,0.99733f,1.00000f,0.00000f,0.94987f,0.94425f,0.92996f,0.92603f,0.92022f,0.92083f,0.92338f,0.92472f,0.92861f,0.93148f,0.93364f,0.93826f,0.93988f,0.94460f,0.94905f,0.95447f,0.95868f,0.96313f,0.96486f,0.96853f,0.97223f,0.97499f,0.97786f,0.98159f,0.98358f,0.98560f,0.98777f,0.99024f,0.99237f,0.99481f,0.99707f,1.00000f,0.00000f,0.95264f,0.94494f,0.93819f,0.92801f,0.92413f,0.91873f,0.92293f,0.91997f,0.92703f,0.92761f,0.93059f,0.93331f,0.93296f,0.94063f,0.94300f,0.94710f,0.95244f,0.95576f,0.96101f,0.96254f,0.96887f,0.96964f,0.97412f,0.97765f,0.98107f,0.98392f,0.98608f,0.98855f,0.99157f,0.99430f,0.99707f,1.00000f,0.00000f,0.95885f,0.95029f,0.94624f,0.93571f,0.92892f,0.92503f,0.92362f,0.92435f,0.92658f,0.92832f,0.92957f,0.93173f,0.93014f,0.93865f,0.94186f,0.94486f,0.94569f,0.95014f,0.95220f,0.95772f,0.96280f,0.96626f,0.96986f,0.97393f,0.97773f,0.98092f,0.98431f,0.98733f,0.99051f,0.99369f,0.99661f,1.00000f,0.00000f,0.95847f,0.95660f,0.94635f,0.94148f,0.93293f,0.93137f,0.93221f,0.92636f,0.92530f,0.92633f,0.92931f,0.93118f,0.93035f,0.93457f,0.93477f,0.93533f,0.94149f,0.94631f,0.94848f,0.95339f,0.95695f,0.96078f,0.96477f,0.96923f,0.97269f,0.97650f,0.98163f,0.98425f,0.98816f,0.99201f,0.99554f,1.00000f,0.00000f,0.96148f,0.95442f,0.94857f,0.94467f,0.93236f,0.92923f,0.92976f,0.92468f,0.92269f,0.92893f,0.92427f,0.92794f,0.92977f,0.93441f,0.93256f,0.93351f,0.93610f,0.93901f,0.94313f,0.94409f,0.95168f,0.95569f,0.95913f,0.96175f,0.96798f,0.97143f,0.97586f,0.97931f,0.98403f,0.98833f,0.99346f,0.99989f,0.00000f,0.96121f,0.95542f,0.95151f,0.95040f,0.93661f,0.93564f,0.93342f,0.93144f,0.92973f,0.93027f,0.92607f,0.92725f,0.92726f,0.93046f,0.93156f,0.93268f,0.92941f,0.93673f,0.93849f,0.94041f,0.94329f,0.94838f,0.95387f,0.95778f,0.96106f,0.96470f,0.96930f,0.97526f,0.97792f,0.98443f,0.98913f,0.99934f,0.00000f,0.96708f,0.96007f,0.95762f,0.95591f,0.94301f,0.94461f,0.93675f,0.93131f,0.93150f,0.92459f,0.92628f,0.92617f,0.92860f,0.93005f,0.92722f,0.93037f,0.93397f,0.93148f,0.93523f,0.93643f,0.94211f,0.94234f,0.94816f,0.95207f,0.95316f,0.95912f,0.96276f,0.96829f,0.97370f,0.97840f,0.98262f,0.99761f,0.00000f,0.95967f,0.95509f,0.96021f,0.95321f,0.94792f,0.94119f,0.94004f,0.93190f,0.92745f,0.92939f,0.92427f,0.92605f,0.92696f,0.92634f,0.92964f,0.92699f,0.92975f,0.93194f,0.92856f,0.93238f,0.93467f,0.93578f,0.93714f,0.94403f,0.94847f,0.95029f,0.95568f,0.95943f,0.96567f,0.96973f,0.97538f,0.99380f,0.00000f,0.95543f,0.95908f,0.96020f,0.95620f,0.94682f,0.94782f,0.93619f,0.93356f,0.93362f,0.93322f,0.93035f,0.93237f,0.92593f,0.92838f,0.92746f,0.93038f,0.93353f,0.92667f,0.92794f,0.93126f,0.92987f,0.93584f,0.93299f,0.93750f,0.93796f,0.94518f,0.94570f,0.95048f,0.95580f,0.95997f,0.96561f,0.98722f,0.00000f,0.95730f,0.96246f,0.96121f,0.95659f,0.94287f,0.94701f,0.94373f,0.94022f,0.93252f,0.93019f,0.92928f,0.93473f,0.93060f,0.92812f,0.92515f,0.92579f,0.93007f,0.92989f,0.92406f,0.92747f,0.92583f,0.92666f,0.92979f,0.93122f,0.93274f,0.93742f,0.94013f,0.94154f,0.94567f,0.95012f,0.95373f,0.97825f,0.00000f,0.95620f,0.95644f,0.95687f,0.94895f,0.94442f,0.94834f,0.93656f,0.93571f,0.93976f,0.93115f,0.93625f,0.93079f,0.93187f,0.93037f,0.92503f,0.92387f,0.92581f,0.92449f,0.92381f,0.92525f,0.92048f,0.92558f,0.92486f,0.92404f,0.92501f,0.92576f,0.92880f,0.93349f,0.93521f,0.93856f,0.93932f,0.96911f,0.00000f,0.94605f,0.94799f,0.94864f,0.95091f,0.94591f,0.94760f,0.94252f,0.93618f,0.93598f,0.93059f,0.93300f,0.93052f,0.93053f,0.92287f,0.92433f,0.92457f,0.92541f,0.92331f,0.92071f,0.91683f,0.91858f,0.91743f,0.92180f,0.92084f,0.91592f,0.91705f,0.92129f,0.91979f,0.92340f,0.92406f,0.92657f,0.95756f,0.00000f,0.94181f,0.94307f,0.94872f,0.94022f,0.93962f,0.93818f,0.93672f,0.94174f,0.93523f,0.92789f,0.92582f,0.93090f,0.92578f,0.91737f,0.92127f,0.92239f,0.92167f,0.91907f,0.91770f,0.91738f,0.91032f,0.91444f,0.91081f,0.91049f,0.90763f,0.90853f,0.90794f,0.91027f,0.91011f,0.91194f,0.91383f,0.94469f,0.00000f,0.92932f,0.93037f,0.93594f,0.93361f,0.92875f,0.93125f,0.93104f,0.92900f,0.93089f,0.92527f,0.92334f,0.92312f,0.91790f,0.92180f,0.91888f,0.91846f,0.91617f,0.91432f,0.91473f,0.90900f,0.91157f,0.90905f,0.90620f,0.90302f,0.90013f,0.89791f,0.90046f,0.89839f,0.89733f,0.90037f,0.90057f,0.93043f,0.00000f,0.91558f,0.91464f,0.91684f,0.92107f,0.92508f,0.92306f,0.92464f,0.92619f,0.91980f,0.91803f,0.91932f,0.91572f,0.91597f,0.91819f,0.91276f,0.91297f,0.90658f,0.90798f,0.90562f,0.90210f,0.90233f,0.89859f,0.89664f,0.89680f,0.89419f,0.89040f,0.88536f,0.88475f,0.88473f,0.88476f,0.88511f,0.91828f,0.00000f,0.90021f,0.89919f,0.91254f,0.90873f,0.91309f,0.90739f,0.90611f,0.90834f,0.90424f,0.90855f,0.90790f,0.90215f,0.90498f,0.90609f,0.89942f,0.90300f,0.90099f,0.89758f,0.89546f,0.89043f,0.89203f,0.88857f,0.88693f,0.88546f,0.88482f,0.87831f,0.87640f,0.87519f,0.87233f,0.87274f,0.87222f,0.90434f,0.00000f,0.88028f,0.88169f,0.87963f,0.88828f,0.89215f,0.89375f,0.89649f,0.89261f,0.89617f,0.89564f,0.89742f,0.89375f,0.89075f,0.89073f,0.88966f,0.89115f,0.89108f,0.89106f,0.88316f,0.88687f,0.88369f,0.87722f,0.87764f,0.87067f,0.86821f,0.86335f,0.86355f,0.85844f,0.85846f,0.86063f,0.85907f,0.89304f,0.00000f,0.86065f,0.85601f,0.86536f,0.86414f,0.87385f,0.87561f,0.87335f,0.87909f,0.88017f,0.88090f,0.87780f,0.87864f,0.88208f,0.87964f,0.87355f,0.87774f,0.87911f,0.87291f,0.87112f,0.87077f,0.86731f,0.86909f,0.86401f,0.85897f,0.85688f,0.85380f,0.84944f,0.84852f,0.84562f,0.84437f,0.84843f,0.88047f,
 0.00000f,0.99735f,0.98383f,0.97990f,0.97382f,0.96336f,0.97184f,0.96789f,0.95857f,0.94523f,0.95500f,0.94791f,0.95008f,0.95397f,0.93176f,0.92787f,0.95871f,0.95850f,0.93886f,0.96581f,0.95381f,0.96240f,0.98034f,0.96085f,0.97639f,0.98320f,0.98547f,0.98579f,0.98984f,0.98559f,0.99482f,0.99740f,1.00000f,0.00000f,0.99080f,0.98490f,0.97948f,0.97755f,0.97484f,0.97391f,0.97052f,0.97019f,0.97031f,0.96904f,0.96897f,0.96948f,0.96948f,0.96835f,0.96979f,0.97105f,0.97226f,0.97192f,0.97411f,0.97530f,0.97695f,0.98080f,0.98017f,0.98196f,0.98529f,0.98706f,0.98880f,0.99104f,0.99242f,0.99514f,0.99764f,1.00000f,0.00000f,0.99053f,0.98494f,0.98220f,0.98027f,0.97639f,0.97673f,0.97551f,0.97436f,0.97385f,0.97275f,0.97376f,0.97317f,0.97310f,0.97355f,0.97451f,0.97466f,0.97540f,0.97652f,0.97727f,0.97838f,0.98014f,0.98149f,0.98266f,0.98429f,0.98615f,0.98729f,0.98943f,0.99118f,0.99327f,0.99561f,0.99773f,1.00000f,0.00000f,0.98387f,0.98410f,0.98281f,0.97968f,0.97698f,0.97633f,0.97496f,0.97505f,0.97409f,0.97329f,0.97384f,0.97328f,0.97346f,0.97391f,0.97450f,0.97468f,0.97583f,0.97636f,0.97757f,0.97903f,0.98003f,0.98117f,0.98291f,0.98445f,0.98588f,0.98765f,0.98940f,0.99129f,0.99337f,0.99542f,0.99775f,1.00000f,0.00000f,0.94168f,0.98193f,0.98285f,0.97899f,0.97722f,0.97628f,0.97545f,0.97355f,0.97383f,0.97339f,0.97292f,0.97313f,0.97356f,0.97398f,0.97394f,0.97560f,0.97588f,0.97602f,0.97766f,0.97901f,0.97980f,0.98150f,0.98250f,0.98447f,0.98603f,0.98764f,0.98948f,0.99131f,0.99334f,0.99553f,0.99773f,1.00000f,0.00000f,0.91456f,0.95919f,0.97752f,0.97964f,0.97630f,0.97579f,0.97500f,0.97273f,0.97418f,0.97310f,0.97317f,0.97342f,0.97323f,0.97316f,0.97433f,0.97473f,0.97584f,0.97668f,0.97793f,0.97943f,0.97993f,0.98051f,0.98258f,0.98424f,0.98586f,0.98758f,0.98944f,0.99110f,0.99345f,0.99549f,0.99778f,1.00000f,0.00000f,0.90890f,0.93143f,0.95975f,0.97343f,0.97609f,0.97467f,0.97391f,0.97386f,0.97267f,0.97292f,0.97340f,0.97313f,0.97392f,0.97322f,0.97392f,0.97469f,0.97529f,0.97677f,0.97730f,0.97894f,0.97984f,0.98107f,0.98256f,0.98437f,0.98535f,0.98752f,0.98924f,0.99135f,0.99338f,0.99531f,0.99761f,1.00000f,0.00000f,0.91279f,0.91573f,0.93920f,0.95823f,0.96759f,0.97248f,0.97415f,0.97374f,0.97289f,0.97231f,0.97161f,0.97188f,0.97292f,0.97284f,0.97362f,0.97356f,0.97454f,0.97675f,0.97651f,0.97812f,0.97962f,0.98110f,0.98212f,0.98394f,0.98557f,0.98750f,0.98915f,0.99123f,0.99325f,0.99531f,0.99744f,1.00000f,0.00000f,0.92018f,0.91361f,0.92152f,0.93846f,0.95421f,0.96340f,0.97008f,0.97119f,0.97167f,0.97302f,0.97249f,0.97192f,0.97248f,0.97359f,0.97494f,0.97481f,0.97445f,0.97618f,0.97740f,0.97783f,0.97954f,0.98041f,0.98281f,0.98363f,0.98552f,0.98725f,0.98904f,0.99085f,0.99309f,0.99543f,0.99755f,1.00000f,0.00000f,0.93048f,0.91591f,0.91929f,0.92753f,0.94007f,0.95247f,0.96009f,0.96362f,0.96824f,0.96971f,0.97105f,0.97203f,0.97269f,0.97300f,0.97412f,0.97367f,0.97480f,0.97552f,0.97716f,0.97744f,0.97892f,0.98031f,0.98216f,0.98360f,0.98543f,0.98729f,0.98932f,0.99103f,0.99301f,0.99522f,0.99759f,1.00000f,0.00000f,0.93431f,0.91804f,0.91260f,0.92036f,0.92922f,0.93714f,0.94857f,0.95348f,0.96129f,0.96399f,0.96899f,0.96998f,0.97054f,0.97062f,0.97409f,0.97409f,0.97398f,0.97628f,0.97697f,0.97790f,0.97859f,0.97998f,0.98162f,0.98371f,0.98453f,0.98715f,0.98883f,0.99061f,0.99302f,0.99530f,0.99758f,1.00000f,0.00000f,0.94410f,0.91337f,0.90901f,0.91748f,0.91916f,0.92958f,0.93641f,0.94251f,0.94989f,0.95622f,0.96338f,0.96464f,0.96869f,0.96933f,0.97074f,0.97291f,0.97363f,0.97502f,0.97605f,0.97751f,0.97870f,0.98034f,0.98268f,0.98292f,0.98461f,0.98639f,0.98841f,0.99069f,0.99281f,0.99510f,0.99733f,1.00000f,0.00000f,0.94906f,0.92034f,0.91980f,0.91750f,0.91972f,0.92712f,0.92863f,0.93656f,0.94082f,0.94739f,0.95280f,0.95890f,0.96262f,0.96535f,0.96881f,0.97078f,0.97325f,0.97418f,0.97596f,0.97725f,0.97780f,0.98011f,0.98112f,0.98385f,0.98501f,0.98691f,0.98876f,0.99069f,0.99278f,0.99503f,0.99739f,1.00000f,0.00000f,0.94761f,0.93395f,0.91977f,0.91998f,0.91642f,0.92222f,0.92545f,0.92944f,0.93385f,0.94161f,0.94459f,0.94923f,0.95649f,0.96082f,0.96377f,0.96633f,0.96783f,0.97126f,0.97362f,0.97587f,0.97707f,0.97877f,0.98062f,0.98242f,0.98502f,0.98716f,0.98781f,0.99072f,0.99263f,0.99479f,0.99725f,1.00000f,0.00000f,0.95263f,0.93471f,0.92459f,0.92207f,0.91910f,0.91742f,0.92041f,0.92700f,0.92983f,0.93352f,0.93717f,0.94479f,0.94655f,0.95258f,0.95771f,0.95822f,0.96379f,0.96761f,0.97190f,0.97278f,0.97554f,0.97837f,0.98031f,0.98342f,0.98511f,0.98647f,0.98858f,0.99069f,0.99234f,0.99502f,0.99706f,1.00000f,0.00000f,0.96068f,0.94997f,0.93106f,0.92502f,0.92238f,0.91996f,0.92204f,0.92452f,0.92704f,0.92952f,0.93374f,0.93794f,0.93969f,0.94703f,0.94920f,0.95464f,0.95810f,0.96128f,0.96549f,0.96844f,0.97127f,0.97504f,0.97704f,0.98068f,0.98259f,0.98544f,0.98782f,0.99017f,0.99256f,0.99456f,0.99697f,1.00000f,0.00000f,0.95641f,0.95078f,0.94247f,0.93067f,0.92562f,0.92222f,0.92182f,0.92488f,0.92285f,0.92866f,0.93107f,0.93336f,0.93320f,0.93750f,0.94420f,0.94795f,0.94998f,0.95763f,0.95969f,0.96405f,0.96876f,0.97073f,0.97565f,0.97761f,0.98082f,0.98332f,0.98707f,0.98919f,0.99139f,0.99405f,0.99701f,1.00000f,0.00000f,0.96205f,0.94504f,0.93872f,0.93738f,0.92665f,0.92467f,0.92585f,0.92520f,0.92264f,0.92545f,0.92491f,0.93053f,0.93072f,0.93657f,0.93911f,0.94174f,0.94524f,0.94904f,0.95457f,0.95955f,0.96182f,0.96653f,0.96937f,0.97319f,0.97630f,0.98052f,0.98379f,0.98712f,0.99008f,0.99357f,0.99661f,1.00000f,0.00000f,0.96246f,0.95501f,0.94426f,0.93331f,0.93220f,0.92807f,0.92931f,0.92385f,0.92516f,0.92771f,0.92734f,0.92977f,0.92965f,0.93230f,0.93485f,0.93786f,0.94060f,0.94240f,0.94775f,0.95236f,0.95769f,0.96022f,0.96385f,0.96840f,0.97158f,0.97629f,0.97870f,0.98361f,0.98849f,0.99207f,0.99575f,1.00000f,0.00000f,0.95705f,0.95395f,0.94294f,0.93819f,0.93055f,0.93324f,0.93263f,0.92480f,0.92752f,0.92881f,0.92884f,0.92623f,0.92887f,0.93130f,0.93241f,0.93033f,0.93470f,0.93782f,0.94281f,0.94707f,0.94914f,0.95378f,0.95812f,0.96350f,0.96582f,0.97139f,0.97467f,0.97952f,0.98349f,0.98826f,0.99329f,0.99988f,0.00000f,0.96370f,0.95768f,0.94751f,0.94695f,0.93934f,0.92974f,0.93340f,0.93046f,0.92673f,0.92613f,0.92465f,0.92545f,0.92865f,0.92936f,0.92993f,0.92938f,0.93116f,0.93582f,0.93772f,0.94141f,0.94379f,0.94709f,0.95261f,0.95674f,0.96090f,0.96473f,0.96979f,0.97388f,0.97844f,0.98363f,0.98874f,0.99932f,0.00000f,0.96582f,0.96479f,0.95627f,0.94804f,0.94375f,0.94694f,0.93991f,0.93234f,0.92718f,0.92949f,0.92396f,0.92530f,0.92711f,0.92829f,0.93083f,0.93144f,0.92907f,0.93118f,0.93212f,0.93431f,0.93900f,0.94162f,0.94453f,0.95032f,0.95542f,0.95829f,0.96215f,0.96703f,0.97131f,0.97730f,0.98256f,0.99759f,0.00000f,0.95950f,0.95414f,0.96228f,0.95263f,0.94539f,0.94273f,0.93685f,0.93884f,0.92982f,0.92754f,0.93125f,0.92714f,0.92706f,0.92377f,0.92874f,0.93174f,0.92758f,0.92900f,0.93227f,0.93366f,0.93311f,0.93616f,0.93985f,0.94088f,0.94779f,0.95184f,0.95512f,0.95971f,0.96504f,0.96986f,0.97442f,0.99331f,0.00000f,0.96654f,0.95733f,0.95703f,0.95887f,0.94784f,0.94485f,0.93862f,0.93391f,0.93923f,0.93193f,0.92911f,0.93139f,0.92419f,0.93098f,0.92975f,0.92580f,0.92972f,0.92864f,0.92885f,0.92816f,0.92916f,0.93257f,0.93109f,0.93688f,0.94131f,0.94339f,0.94777f,0.95022f,0.95470f,0.95958f,0.96341f,0.98651f,0.00000f,0.95687f,0.96260f,0.95930f,0.94819f,0.95055f,0.94446f,0.94751f,0.93614f,0.93082f,0.93113f,0.93600f,0.93107f,0.93053f,0.93143f,0.93039f,0.92834f,0.92358f,0.92435f,0.92470f,0.92675f,0.92406f,0.92831f,0.92898f,0.92818f,0.93166f,0.93452f,0.93755f,0.94220f,0.94423f,0.94898f,0.95119f,0.97807f,0.00000f,0.95462f,0.96039f,0.95680f,0.95279f,0.95443f,0.94712f,0.94335f,0.94036f,0.93867f,0.93470f,0.92974f,0.92625f,0.92466f,0.92363f,0.92491f,0.92285f,0.92603f,0.92094f,0.92539f,0.92172f,0.91963f,0.92388f,0.92102f,0.92503f,0.92444f,0.92526f,0.92604f,0.93273f,0.93437f,0.93683f,0.93862f,0.96799f,0.00000f,0.95196f,0.95177f,0.94947f,0.95399f,0.95258f,0.94993f,0.94487f,0.94130f,0.93393f,0.93608f,0.92949f,0.92711f,0.92313f,0.92461f,0.92034f,0.92143f,0.92313f,0.92256f,0.91962f,0.91769f,0.91697f,0.91749f,0.91872f,0.91383f,0.91382f,0.91775f,0.91974f,0.91876f,0.92102f,0.92329f,0.92451f,0.95511f,0.00000f,0.94079f,0.94947f,0.94477f,0.94837f,0.94184f,0.94244f,0.93862f,0.93288f,0.93474f,0.93069f,0.92888f,0.92803f,0.92925f,0.92107f,0.92203f,0.91981f,0.91641f,0.91556f,0.91564f,0.91644f,0.91563f,0.91329f,0.91118f,0.91177f,0.90893f,0.90862f,0.90691f,0.90843f,0.90860f,0.90954f,0.90978f,0.94170f,0.00000f,0.92464f,0.94151f,0.93482f,0.93163f,0.94016f,0.92998f,0.92924f,0.92224f,0.92171f,0.93204f,0.92241f,0.92555f,0.91610f,0.91767f,0.91570f,0.91707f,0.92000f,0.91407f,0.91180f,0.91047f,0.90525f,0.90858f,0.90154f,0.90241f,0.90119f,0.89716f,0.89815f,0.89405f,0.89477f,0.89638f,0.89534f,0.92869f,0.00000f,0.91453f,0.92473f,0.91672f,0.91670f,0.92536f,0.92794f,0.92630f,0.91932f,0.92036f,0.91514f,0.91545f,0.91384f,0.91359f,0.90931f,0.91439f,0.90720f,0.90765f,0.90585f,0.90482f,0.90637f,0.89765f,0.89979f,0.89565f,0.89439f,0.89169f,0.88550f,0.88562f,0.88035f,0.88113f,0.88303f,0.88210f,0.91542f,0.00000f,0.90252f,0.89769f,0.90633f,0.91014f,0.90533f,0.90944f,0.91247f,0.91450f,0.90090f,0.90929f,0.90447f,0.90924f,0.90031f,0.90244f,0.90030f,0.90585f,0.89899f,0.89705f,0.89765f,0.88998f,0.89558f,0.88945f,0.88695f,0.88301f,0.87958f,0.87710f,0.87355f,0.87059f,0.86879f,0.86982f,0.86880f,0.90119f,0.00000f,0.87135f,0.89072f,0.89220f,0.88686f,0.89603f,0.89614f,0.89750f,0.89235f,0.88917f,0.90197f,0.88919f,0.89798f,0.89580f,0.89354f,0.89164f,0.89198f,0.89141f,0.88916f,0.88335f,0.88220f,0.87973f,0.87834f,0.87702f,0.87081f,0.86643f,0.86510f,0.86177f,0.85521f,0.85469f,0.85389f,0.85623f,0.88885f,0.00000f,0.86340f,0.87105f,0.86855f,0.86742f,0.87188f,0.87642f,0.86852f,0.87806f,0.87454f,0.88436f,0.87639f,0.87704f,0.87602f,0.87603f,0.87853f,0.87744f,0.87664f,0.87298f,0.87063f,0.87060f,0.86441f,0.86398f,0.85817f,0.85678f,0.85539f,0.85148f,0.84595f,0.84309f,0.84242f,0.83971f,0.84156f,0.87467f,
 0.00000f,0.99504f,0.98607f,0.98246f,0.97608f,0.95866f,0.94818f,0.96070f,0.94669f,0.95967f,0.94980f,0.95239f,0.94833f,0.94644f,0.94824f,0.95873f,0.95084f,0.96234f,0.95700f,0.94671f,0.97029f,0.97084f,0.96142f,0.96808f,0.97956f,0.98156f,0.98562f,0.98710f,0.97939f,0.99082f,0.99236f,0.99646f,1.00000f,0.00000f,0.99043f,0.98487f,0.98167f,0.97713f,0.97466f,0.97210f,0.97136f,0.97082f,0.97198f,0.96858f,0.97034f,0.97034f,0.96874f,0.96991f,0.97086f,0.97030f,0.97085f,0.97305f,0.97410f,0.97586f,0.97672f,0.97858f,0.98053f,0.98343f,0.98472f,0.98699f,0.98915f,0.99009f,0.99278f,0.99518f,0.99752f,1.00000f,0.00000f,0.99041f,0.98597f,0.98312f,0.97931f,0.97819f,0.97641f,0.97483f,0.97446f,0.97439f,0.97399f,0.97288f,0.97344f,0.97276f,0.97366f,0.97479f,0.97470f,0.97589f,0.97656f,0.97778f,0.97856f,0.97995f,0.98101f,0.98269f,0.98463f,0.98586f,0.98778f,0.98937f,0.99126f,0.99340f,0.99544f,0.99776f,1.00000f,0.00000f,0.98444f,0.98734f,0.98190f,0.97933f,0.97698f,0.97614f,0.97491f,0.97497f,0.97358f,0.97434f,0.97322f,0.97330f,0.97371f,0.97368f,0.97449f,0.97489f,0.97615f,0.97648f,0.97773f,0.97859f,0.97984f,0.98158f,0.98284f,0.98427f,0.98610f,0.98760f,0.98929f,0.99126f,0.99339f,0.99550f,0.99780f,1.00000f,0.00000f,0.94553f,0.98117f,0.98040f,0.97928f,0.97862f,0.97652f,0.97503f,0.97342f,0.97419f,0.97282f,0.97287f,0.97209f,0.97334f,0.97394f,0.97466f,0.97529f,0.97540f,0.97664f,0.97725f,0.97855f,0.98011f,0.98132f,0.98279f,0.98427f,0.98602f,0.98745f,0.98924f,0.99133f,0.99331f,0.99546f,0.99778f,1.00000f,0.00000f,0.91405f,0.96096f,0.97727f,0.97976f,0.97654f,0.97613f,0.97473f,0.97421f,0.97285f,0.97261f,0.97269f,0.97294f,0.97361f,0.97399f,0.97431f,0.97524f,0.97540f,0.97623f,0.97722f,0.97810f,0.97986f,0.98106f,0.98182f,0.98415f,0.98554f,0.98748f,0.98921f,0.99116f,0.99328f,0.99541f,0.99771f,1.00000f,0.00000f,0.90232f,0.93333f,0.96112f,0.97268f,0.97531f,0.97584f,0.97410f,0.97267f,0.97305f,0.97327f,0.97209f,0.97156f,0.97353f,0.97362f,0.97473f,0.97492f,0.97532f,0.97660f,0.97705f,0.97837f,0.97983f,0.98112f,0.98219f,0.98386f,0.98614f,0.98738f,0.98924f,0.99137f,0.99326f,0.99544f,0.99760f,1.00000f,0.00000f,0.91045f,0.91706f,0.93613f,0.95795f,0.96783f,0.97242f,0.97245f,0.97352f,0.97153f,0.97156f,0.97294f,0.97244f,0.97298f,0.97276f,0.97321f,0.97469f,0.97493f,0.97611f,0.97752f,0.97824f,0.97915f,0.98056f,0.98192f,0.98361f,0.98588f,0.98726f,0.98930f,0.99112f,0.99310f,0.99537f,0.99749f,1.00000f,0.00000f,0.91784f,0.91054f,0.92412f,0.94029f,0.95376f,0.96415f,0.96874f,0.97087f,0.97127f,0.97195f,0.97246f,0.97216f,0.97314f,0.97312f,0.97405f,0.97444f,0.97559f,0.97550f,0.97715f,0.97822f,0.97941f,0.98112f,0.98180f,0.98402f,0.98574f,0.98748f,0.98936f,0.99098f,0.99329f,0.99526f,0.99766f,1.00000f,0.00000f,0.92793f,0.91362f,0.91532f,0.92839f,0.93876f,0.95164f,0.95925f,0.96498f,0.96784f,0.96892f,0.97268f,0.97050f,0.97155f,0.97218f,0.97374f,0.97504f,0.97438f,0.97578f,0.97594f,0.97772f,0.97939f,0.98139f,0.98258f,0.98353f,0.98501f,0.98722f,0.98914f,0.99116f,0.99306f,0.99529f,0.99759f,1.00000f,0.00000f,0.93108f,0.91514f,0.91143f,0.91916f,0.92578f,0.93799f,0.94984f,0.95665f,0.95887f,0.96525f,0.96775f,0.97037f,0.97125f,0.97264f,0.97329f,0.97387f,0.97516f,0.97547f,0.97724f,0.97800f,0.97866f,0.97978f,0.98160f,0.98385f,0.98548f,0.98675f,0.98908f,0.99072f,0.99298f,0.99518f,0.99739f,1.00000f,0.00000f,0.93780f,0.92227f,0.91593f,0.91572f,0.92452f,0.92724f,0.93662f,0.94337f,0.95091f,0.95634f,0.96156f,0.96312f,0.96844f,0.96914f,0.97083f,0.97233f,0.97435f,0.97627f,0.97606f,0.97756f,0.97947f,0.97970f,0.98082f,0.98345f,0.98595f,0.98657f,0.98874f,0.99080f,0.99271f,0.99527f,0.99749f,1.00000f,0.00000f,0.94623f,0.92410f,0.91626f,0.91499f,0.91830f,0.92493f,0.93027f,0.93302f,0.94403f,0.95003f,0.95398f,0.95788f,0.96210f,0.96417f,0.96713f,0.97069f,0.97295f,0.97258f,0.97448f,0.97757f,0.97871f,0.97946f,0.98087f,0.98291f,0.98530f,0.98702f,0.98864f,0.99063f,0.99266f,0.99493f,0.99734f,1.00000f,0.00000f,0.95034f,0.92841f,0.92038f,0.91848f,0.91783f,0.92194f,0.92434f,0.93233f,0.93374f,0.93925f,0.94764f,0.95063f,0.95459f,0.95924f,0.96178f,0.96726f,0.96988f,0.97173f,0.97333f,0.97521f,0.97758f,0.97841f,0.98156f,0.98371f,0.98413f,0.98700f,0.98872f,0.99055f,0.99247f,0.99470f,0.99715f,1.00000f,0.00000f,0.95588f,0.93176f,0.92943f,0.91998f,0.92007f,0.92076f,0.92378f,0.92932f,0.92953f,0.93460f,0.93861f,0.94208f,0.94782f,0.95341f,0.95566f,0.95964f,0.96464f,0.96740f,0.97108f,0.97260f,0.97476f,0.97836f,0.97886f,0.98245f,0.98463f,0.98591f,0.98863f,0.99074f,0.99272f,0.99505f,0.99718f,1.00000f,0.00000f,0.95362f,0.94392f,0.93124f,0.93088f,0.92079f,0.91680f,0.92517f,0.92415f,0.92979f,0.93031f,0.93408f,0.93403f,0.94066f,0.94393f,0.95221f,0.95521f,0.95801f,0.96130f,0.96487f,0.96791f,0.97099f,0.97486f,0.97743f,0.98044f,0.98373f,0.98512f,0.98694f,0.99020f,0.99265f,0.99482f,0.99704f,1.00000f,0.00000f,0.95273f,0.94764f,0.94155f,0.93174f,0.92766f,0.92186f,0.92525f,0.92283f,0.92649f,0.92781f,0.92909f,0.93275f,0.93395f,0.93900f,0.94462f,0.94637f,0.95175f,0.95595f,0.95954f,0.96375f,0.96679f,0.97046f,0.97497f,0.97792f,0.98073f,0.98348f,0.98703f,0.98915f,0.99200f,0.99427f,0.99729f,1.00000f,0.00000f,0.95319f,0.95268f,0.94374f,0.93178f,0.93113f,0.92846f,0.92434f,0.92283f,0.92458f,0.92653f,0.92574f,0.93189f,0.93051f,0.93465f,0.93604f,0.94331f,0.94723f,0.94940f,0.95352f,0.95734f,0.96266f,0.96600f,0.96945f,0.97294f,0.97693f,0.98001f,0.98371f,0.98686f,0.99033f,0.99375f,0.99673f,1.00000f,0.00000f,0.95885f,0.95809f,0.94220f,0.94376f,0.94072f,0.92311f,0.92984f,0.92150f,0.92085f,0.92848f,0.92415f,0.93112f,0.93111f,0.93570f,0.93276f,0.93808f,0.94353f,0.94360f,0.94766f,0.95280f,0.95645f,0.95858f,0.96538f,0.96871f,0.97245f,0.97630f,0.97872f,0.98408f,0.98790f,0.99064f,0.99576f,1.00000f,0.00000f,0.96973f,0.95242f,0.94648f,0.94195f,0.93789f,0.93123f,0.92914f,0.92923f,0.92560f,0.92808f,0.92385f,0.92768f,0.92870f,0.93123f,0.93074f,0.93405f,0.93701f,0.93814f,0.94239f,0.94663f,0.95141f,0.95533f,0.95868f,0.96260f,0.96622f,0.97128f,0.97485f,0.97873f,0.98253f,0.98776f,0.99298f,0.99989f,0.00000f,0.96042f,0.95246f,0.95430f,0.95075f,0.93900f,0.93608f,0.93596f,0.93241f,0.92871f,0.92600f,0.92606f,0.92859f,0.92751f,0.92798f,0.93075f,0.93128f,0.93440f,0.93619f,0.93737f,0.94142f,0.94414f,0.94779f,0.95378f,0.95630f,0.96092f,0.96392f,0.96777f,0.97403f,0.97879f,0.98313f,0.98884f,0.99928f,0.00000f,0.96503f,0.96188f,0.94890f,0.94923f,0.94018f,0.93683f,0.93775f,0.93415f,0.92762f,0.92696f,0.93080f,0.92733f,0.92632f,0.92795f,0.92903f,0.93101f,0.93353f,0.93099f,0.93207f,0.93765f,0.93723f,0.94158f,0.94373f,0.94870f,0.95420f,0.95836f,0.96114f,0.96807f,0.97386f,0.97592f,0.98131f,0.99737f,0.00000f,0.96210f,0.96042f,0.96180f,0.95568f,0.94511f,0.93971f,0.94487f,0.93360f,0.93113f,0.92912f,0.93112f,0.92966f,0.92660f,0.92745f,0.93263f,0.93002f,0.92866f,0.92835f,0.93203f,0.93275f,0.93422f,0.93468f,0.93742f,0.94145f,0.94528f,0.95215f,0.95466f,0.95851f,0.96274f,0.96787f,0.97378f,0.99320f,0.00000f,0.96928f,0.96148f,0.95723f,0.95042f,0.95151f,0.94492f,0.94177f,0.93723f,0.93012f,0.92726f,0.93027f,0.92827f,0.92968f,0.92740f,0.92717f,0.92698f,0.92697f,0.92617f,0.92749f,0.93057f,0.93104f,0.92970f,0.93556f,0.93357f,0.93630f,0.94385f,0.94471f,0.95085f,0.95352f,0.95806f,0.96225f,0.98627f,0.00000f,0.95932f,0.95565f,0.95587f,0.95510f,0.95483f,0.95024f,0.94981f,0.93816f,0.93833f,0.94432f,0.93250f,0.92730f,0.92952f,0.92410f,0.92766f,0.92613f,0.92538f,0.92455f,0.92509f,0.92467f,0.92641f,0.92677f,0.92695f,0.93011f,0.93099f,0.93248f,0.93637f,0.93889f,0.94370f,0.94749f,0.94974f,0.97753f,0.00000f,0.95340f,0.96296f,0.96081f,0.95487f,0.94624f,0.94361f,0.94109f,0.93797f,0.93027f,0.93691f,0.93625f,0.92693f,0.93037f,0.92536f,0.92562f,0.92421f,0.92402f,0.92444f,0.92614f,0.92167f,0.92159f,0.92182f,0.92047f,0.92381f,0.92101f,0.92266f,0.92776f,0.93042f,0.93158f,0.93313f,0.93715f,0.96590f,0.00000f,0.94886f,0.95306f,0.95286f,0.95538f,0.95077f,0.94659f,0.94048f,0.93913f,0.93768f,0.93365f,0.93072f,0.92937f,0.92707f,0.92244f,0.92683f,0.92298f,0.92063f,0.92150f,0.91898f,0.92027f,0.91968f,0.91376f,0.91594f,0.91354f,0.91401f,0.91570f,0.91948f,0.91928f,0.92040f,0.92171f,0.92402f,0.95342f,0.00000f,0.94283f,0.94630f,0.95231f,0.93988f,0.93472f,0.94119f,0.93776f,0.93168f,0.93155f,0.93113f,0.92867f,0.92354f,0.92152f,0.92069f,0.92463f,0.92083f,0.91794f,0.91728f,0.91493f,0.91036f,0.91213f,0.90737f,0.91222f,0.90992f,0.90941f,0.90481f,0.90532f,0.90618f,0.90423f,0.90780f,0.90720f,0.93945f,0.00000f,0.93179f,0.93575f,0.93609f,0.93609f,0.93385f,0.93211f,0.92915f,0.93116f,0.92757f,0.92967f,0.93028f,0.91687f,0.92203f,0.92182f,0.91784f,0.91191f,0.91273f,0.90942f,0.90796f,0.91052f,0.90443f,0.90358f,0.90380f,0.90172f,0.89608f,0.89634f,0.89472f,0.89234f,0.89281f,0.89301f,0.89509f,0.92603f,0.00000f,0.92145f,0.91875f,0.92959f,0.92690f,0.92069f,0.92522f,0.92586f,0.91874f,0.92090f,0.91678f,0.91985f,0.91635f,0.91367f,0.91286f,0.90860f,0.91050f,0.90680f,0.90750f,0.90313f,0.89914f,0.89714f,0.89416f,0.89639f,0.89230f,0.88890f,0.88644f,0.88079f,0.88113f,0.87851f,0.88040f,0.87952f,0.91246f,0.00000f,0.90334f,0.90165f,0.90774f,0.91337f,0.90678f,0.91669f,0.91042f,0.91041f,0.91535f,0.90767f,0.91186f,0.90470f,0.90536f,0.90457f,0.90178f,0.89362f,0.89897f,0.89802f,0.89144f,0.88930f,0.88941f,0.88625f,0.88234f,0.87768f,0.87263f,0.87747f,0.87033f,0.86763f,0.86141f,0.86523f,0.86432f,0.89901f,0.00000f,0.88134f,0.88852f,0.89388f,0.88897f,0.89021f,0.89972f,0.89881f,0.89576f,0.89940f,0.89257f,0.89497f,0.88976f,0.89139f,0.89450f,0.89238f,0.88791f,0.88802f,0.88374f,0.87901f,0.87746f,0.88047f,0.87667f,0.87045f,0.87013f,0.86508f,0.86250f,0.85654f,0.85333f,0.85107f,0.84946f,0.85152f,0.88463f,0.00000f,0.86519f,0.87361f,0.87518f,0.87736f,0.88014f,0.87754f,0.88104f,0.87428f,0.87979f,0.87803f,0.87931f,0.88570f,0.88520f,0.87901f,0.87305f,0.87281f,0.87071f,0.87051f,0.86754f,0.86752f,0.86786f,0.86328f,0.85858f,0.85734f,0.85067f,0.84246f,0.84289f,0.84147f,0.83723f,0.83583f,0.83727f,0.87121f,
 0.00000f,0.99989f,0.98466f,0.97358f,0.97701f,0.95855f,0.95814f,0.94810f,0.96519f,0.95701f,0.95368f,0.96954f,0.96321f,0.96772f,0.94993f,0.96113f,0.96946f,0.96541f,0.96142f,0.95434f,0.96529f,0.96973f,0.95033f,0.96613f,0.96975f,0.98270f,0.98028f,0.98616f,0.99051f,0.98953f,0.99537f,0.99627f,1.00000f,0.00000f,0.99115f,0.98451f,0.97933f,0.97654f,0.97574f,0.97202f,0.96982f,0.97201f,0.96937f,0.97042f,0.97086f,0.97021f,0.97079f,0.96972f,0.97133f,0.97247f,0.97279f,0.97358f,0.97482f,0.97658f,0.97714f,0.97804f,0.98004f,0.98187f,0.98494f,0.98628f,0.98912f,0.99128f,0.99292f,0.99532f,0.99748f,1.00000f,0.00000f,0.98981f,0.98653f,0.98225f,0.97960f,0.97664f,0.97701f,0.97489f,0.97467f,0.97443f,0.97370f,0.97361f,0.97301f,0.97234f,0.97324f,0.97463f,0.97466f,0.97606f,0.97654f,0.97781f,0.97850f,0.97968f,0.98131f,0.98260f,0.98444f,0.98600f,0.98756f,0.98956f,0.99158f,0.99338f,0.99553f,0.99773f,1.00000f,0.00000f,0.98262f,0.98540f,0.98232f,0.97938f,0.97796f,0.97571f,0.97460f,0.97372f,0.97465f,0.97337f,0.97363f,0.97290f,0.97358f,0.97430f,0.97476f,0.97492f,0.97567f,0.97735f,0.97733f,0.97877f,0.97938f,0.98102f,0.98299f,0.98429f,0.98582f,0.98777f,0.98962f,0.99147f,0.99335f,0.99546f,0.99773f,1.00000f,0.00000f,0.94237f,0.98319f,0.98185f,0.97859f,0.97838f,0.97626f,0.97552f,0.97507f,0.97261f,0.97315f,0.97316f,0.97337f,0.97427f,0.97430f,0.97468f,0.97504f,0.97563f,0.97636f,0.97780f,0.97807f,0.97995f,0.98092f,0.98258f,0.98419f,0.98581f,0.98749f,0.98927f,0.99146f,0.99315f,0.99543f,0.99777f,1.00000f,0.00000f,0.91403f,0.96075f,0.97809f,0.97756f,0.97727f,0.97518f,0.97472f,0.97402f,0.97292f,0.97374f,0.97265f,0.97333f,0.97385f,0.97387f,0.97410f,0.97456f,0.97577f,0.97620f,0.97760f,0.97856f,0.97982f,0.98084f,0.98259f,0.98405f,0.98592f,0.98753f,0.98921f,0.99112f,0.99325f,0.99542f,0.99769f,1.00000f,0.00000f,0.91026f,0.93333f,0.96205f,0.97305f,0.97755f,0.97562f,0.97517f,0.97344f,0.97408f,0.97173f,0.97320f,0.97251f,0.97234f,0.97356f,0.97424f,0.97407f,0.97506f,0.97608f,0.97705f,0.97874f,0.97963f,0.98086f,0.98206f,0.98400f,0.98550f,0.98757f,0.98935f,0.99116f,0.99337f,0.99546f,0.99757f,1.00000f,0.00000f,0.91146f,0.91656f,0.93828f,0.95799f,0.96785f,0.97096f,0.97258f,0.97280f,0.97328f,0.97223f,0.97298f,0.97221f,0.97273f,0.97414f,0.97408f,0.97496f,0.97546f,0.97605f,0.97751f,0.97845f,0.97932f,0.98143f,0.98257f,0.98395f,0.98576f,0.98720f,0.98924f,0.99114f,0.99322f,0.99545f,0.99756f,1.00000f,0.00000f,0.91560f,0.91365f,0.92244f,0.93790f,0.95603f,0.96522f,0.96932f,0.97192f,0.97206f,0.97310f,0.97061f,0.97156f,0.97177f,0.97339f,0.97310f,0.97479f,0.97447f,0.97645f,0.97681f,0.97779f,0.97874f,0.98099f,0.98229f,0.98340f,0.98574f,0.98721f,0.98915f,0.99106f,0.99298f,0.99537f,0.99766f,1.00000f,0.00000f,0.92804f,0.90769f,0.91701f,0.92691f,0.94057f,0.95130f,0.95873f,0.96525f,0.96758f,0.96894f,0.97105f,0.97060f,0.97250f,0.97134f,0.97308f,0.97390f,0.97475f,0.97553f,0.97695f,0.97826f,0.97878f,0.98051f,0.98221f,0.98381f,0.98547f,0.98728f,0.98915f,0.99108f,0.99314f,0.99529f,0.99755f,1.00000f,0.00000f,0.93314f,0.91564f,0.91178f,0.92015f,0.92702f,0.93939f,0.94783f,0.95502f,0.96026f,0.96454f,0.96660f,0.96973f,0.97252f,0.97201f,0.97270f,0.97364f,0.97461f,0.97582f,0.97672f,0.97797f,0.97933f,0.98051f,0.98144f,0.98323f,0.98506f,0.98699f,0.98864f,0.99127f,0.99303f,0.99509f,0.99733f,1.00000f,0.00000f,0.94169f,0.91804f,0.91084f,0.91754f,0.92419f,0.92937f,0.93626f,0.94440f,0.95152f,0.95509f,0.96137f,0.96515f,0.96679f,0.96907f,0.96993f,0.97277f,0.97486f,0.97539f,0.97635f,0.97744f,0.97948f,0.98038f,0.98205f,0.98341f,0.98551f,0.98707f,0.98889f,0.99094f,0.99280f,0.99509f,0.99741f,1.00000f,0.00000f,0.95033f,0.92375f,0.91871f,0.91459f,0.91938f,0.92401f,0.92726f,0.93665f,0.94051f,0.94818f,0.95301f,0.95737f,0.96195f,0.96508f,0.96689f,0.97072f,0.97256f,0.97395f,0.97507f,0.97710f,0.97837f,0.97999f,0.98112f,0.98215f,0.98461f,0.98725f,0.98859f,0.99054f,0.99307f,0.99498f,0.99725f,1.00000f,0.00000f,0.95395f,0.93208f,0.92178f,0.91726f,0.91893f,0.91879f,0.92613f,0.93205f,0.93220f,0.93988f,0.94566f,0.94895f,0.95426f,0.95967f,0.96444f,0.96642f,0.96852f,0.97166f,0.97319f,0.97614f,0.97733f,0.97887f,0.98075f,0.98256f,0.98411f,0.98712f,0.98801f,0.99032f,0.99273f,0.99494f,0.99725f,1.00000f,0.00000f,0.96049f,0.93996f,0.93008f,0.92171f,0.91240f,0.92055f,0.92191f,0.92780f,0.93077f,0.93102f,0.93705f,0.94366f,0.94649f,0.95033f,0.95617f,0.96051f,0.96538f,0.96715f,0.97106f,0.97133f,0.97601f,0.97777f,0.98106f,0.98202f,0.98390f,0.98628f,0.98811f,0.99042f,0.99225f,0.99477f,0.99698f,1.00000f,0.00000f,0.96140f,0.94509f,0.93043f,0.92561f,0.92184f,0.92245f,0.92199f,0.92515f,0.93020f,0.93142f,0.93386f,0.93784f,0.94073f,0.94309f,0.95201f,0.95370f,0.95852f,0.96169f,0.96644f,0.96843f,0.97367f,0.97522f,0.97792f,0.98054f,0.98276f,0.98486f,0.98824f,0.98978f,0.99191f,0.99470f,0.99701f,1.00000f,0.00000f,0.96431f,0.95408f,0.93947f,0.93266f,0.92809f,0.92522f,0.92033f,0.92314f,0.92470f,0.92912f,0.92859f,0.93258f,0.93331f,0.94096f,0.94430f,0.94803f,0.95250f,0.95695f,0.96172f,0.96431f,0.96657f,0.96921f,0.97472f,0.97781f,0.98059f,0.98396f,0.98640f,0.98888f,0.99200f,0.99387f,0.99694f,1.00000f,0.00000f,0.95296f,0.95188f,0.94736f,0.93774f,0.93127f,0.92417f,0.92707f,0.92574f,0.93002f,0.92460f,0.92797f,0.93033f,0.93264f,0.93249f,0.93699f,0.94349f,0.94718f,0.94977f,0.95510f,0.95897f,0.96100f,0.96489f,0.96941f,0.97132f,0.97736f,0.98109f,0.98347f,0.98736f,0.98965f,0.99321f,0.99655f,1.00000f,0.00000f,0.96070f,0.95636f,0.94790f,0.93467f,0.93263f,0.92830f,0.93061f,0.92303f,0.92586f,0.92305f,0.92432f,0.92866f,0.93122f,0.93349f,0.93390f,0.93821f,0.94067f,0.94305f,0.94773f,0.95126f,0.95643f,0.95991f,0.96399f,0.96877f,0.97223f,0.97583f,0.98088f,0.98467f,0.98761f,0.99198f,0.99552f,1.00000f,0.00000f,0.96132f,0.95264f,0.94859f,0.94436f,0.93874f,0.92945f,0.92981f,0.92499f,0.92917f,0.92503f,0.92528f,0.92953f,0.92706f,0.92762f,0.93372f,0.93374f,0.93460f,0.93579f,0.94263f,0.94446f,0.94691f,0.95354f,0.95700f,0.96167f,0.96596f,0.96987f,0.97517f,0.97909f,0.98379f,0.98805f,0.99309f,0.99988f,0.00000f,0.95925f,0.95118f,0.95548f,0.94024f,0.94218f,0.93844f,0.93637f,0.93138f,0.93068f,0.92883f,0.92628f,0.92601f,0.92823f,0.92941f,0.93032f,0.93242f,0.93317f,0.93197f,0.93711f,0.94020f,0.94380f,0.94756f,0.95156f,0.95374f,0.95859f,0.96292f,0.96950f,0.97442f,0.97773f,0.98266f,0.98851f,0.99929f,0.00000f,0.96273f,0.96030f,0.94750f,0.95373f,0.94119f,0.93905f,0.93904f,0.93447f,0.93355f,0.93042f,0.92655f,0.92766f,0.92545f,0.92579f,0.92840f,0.92869f,0.93257f,0.93136f,0.93245f,0.93645f,0.93803f,0.94112f,0.94250f,0.94792f,0.95276f,0.95680f,0.96186f,0.96659f,0.97075f,0.97642f,0.98153f,0.99731f,0.00000f,0.96006f,0.96198f,0.96067f,0.94942f,0.95052f,0.93997f,0.93776f,0.93832f,0.93146f,0.92933f,0.92725f,0.92451f,0.92650f,0.93132f,0.92996f,0.93237f,0.92797f,0.93185f,0.92658f,0.93097f,0.93465f,0.93492f,0.93822f,0.94088f,0.94461f,0.95150f,0.95368f,0.95740f,0.96407f,0.96807f,0.97235f,0.99326f,0.00000f,0.96071f,0.95898f,0.95753f,0.94797f,0.95379f,0.94739f,0.94446f,0.93717f,0.93424f,0.92438f,0.92712f,0.92583f,0.93263f,0.92666f,0.92408f,0.92770f,0.93012f,0.92679f,0.92663f,0.92375f,0.92506f,0.93060f,0.93079f,0.93430f,0.93674f,0.94118f,0.94377f,0.94980f,0.95425f,0.95725f,0.96253f,0.98578f,0.00000f,0.96396f,0.96040f,0.95877f,0.95359f,0.95220f,0.94802f,0.93991f,0.93491f,0.93557f,0.93088f,0.93034f,0.93207f,0.93230f,0.92817f,0.93017f,0.92284f,0.92722f,0.92480f,0.92323f,0.92516f,0.92397f,0.92715f,0.92443f,0.93086f,0.93219f,0.93226f,0.93497f,0.93860f,0.94245f,0.94487f,0.94771f,0.97646f,0.00000f,0.95504f,0.95294f,0.96372f,0.95498f,0.94935f,0.94695f,0.94446f,0.94190f,0.93577f,0.93513f,0.93060f,0.92551f,0.92634f,0.92749f,0.92265f,0.91966f,0.92222f,0.92344f,0.92016f,0.92191f,0.92073f,0.92056f,0.91681f,0.92298f,0.91875f,0.92357f,0.92514f,0.92726f,0.93005f,0.93098f,0.93421f,0.96503f,0.00000f,0.95115f,0.95038f,0.95198f,0.94702f,0.94414f,0.94480f,0.94412f,0.94104f,0.93609f,0.93302f,0.93208f,0.92240f,0.92542f,0.92675f,0.92381f,0.92024f,0.92007f,0.92387f,0.91937f,0.91874f,0.91678f,0.91320f,0.91563f,0.91267f,0.91366f,0.91517f,0.91524f,0.91480f,0.91797f,0.91841f,0.91972f,0.95214f,0.00000f,0.94062f,0.94836f,0.94472f,0.94612f,0.94616f,0.94362f,0.93799f,0.93219f,0.93358f,0.93421f,0.93033f,0.92639f,0.92304f,0.92130f,0.92027f,0.91679f,0.91840f,0.91738f,0.91659f,0.91426f,0.91631f,0.91431f,0.91243f,0.90782f,0.90416f,0.90481f,0.90335f,0.90407f,0.90451f,0.90522f,0.90488f,0.93779f,0.00000f,0.92951f,0.93974f,0.94242f,0.93588f,0.94058f,0.93845f,0.93391f,0.92780f,0.93335f,0.92661f,0.92847f,0.92030f,0.92194f,0.92125f,0.91352f,0.91116f,0.91262f,0.90667f,0.91342f,0.90310f,0.90582f,0.90361f,0.89783f,0.89728f,0.89762f,0.89301f,0.89082f,0.89147f,0.89216f,0.89041f,0.88973f,0.92300f,0.00000f,0.92090f,0.91902f,0.92436f,0.92946f,0.92470f,0.92530f,0.92289f,0.91962f,0.92329f,0.92119f,0.91505f,0.91318f,0.91435f,0.91306f,0.90966f,0.90703f,0.90494f,0.90400f,0.90344f,0.89860f,0.89728f,0.89639f,0.89323f,0.88831f,0.88766f,0.88280f,0.87799f,0.87647f,0.87785f,0.87494f,0.87699f,0.90850f,0.00000f,0.90624f,0.90363f,0.90784f,0.90875f,0.91336f,0.91582f,0.90952f,0.91052f,0.91027f,0.91326f,0.90420f,0.90588f,0.90116f,0.90675f,0.90038f,0.90187f,0.89796f,0.89916f,0.89472f,0.89264f,0.89199f,0.88287f,0.88323f,0.87641f,0.87214f,0.86906f,0.86529f,0.86473f,0.86110f,0.86102f,0.86126f,0.89478f,0.00000f,0.89273f,0.88618f,0.89697f,0.89461f,0.89870f,0.90563f,0.89517f,0.89822f,0.90007f,0.89387f,0.90146f,0.89717f,0.89083f,0.89143f,0.88664f,0.88634f,0.88631f,0.88229f,0.88252f,0.87825f,0.87621f,0.87465f,0.86637f,0.86302f,0.86201f,0.85837f,0.85139f,0.84901f,0.84723f,0.84891f,0.84589f,0.88208f,0.00000f,0.86213f,0.86314f,0.87028f,0.87870f,0.87497f,0.88063f,0.88077f,0.87978f,0.88378f,0.87770f,0.88168f,0.88069f,0.88071f,0.87692f,0.87611f,0.87356f,0.87236f,0.87165f,0.86921f,0.86473f,0.86399f,0.85653f,0.85205f,0.85185f,0.84767f,0.84331f,0.84035f,0.83590f,0.83371f,0.83052f,0.83163f,0.86777f,
 0.00000f,0.99510f,0.98271f,0.98003f,0.97136f,0.95882f,0.97184f,0.95689f,0.96248f,0.95752f,0.95657f,0.96966f,0.96179f,0.96334f,0.95671f,0.96774f,0.94970f,0.95339f,0.95696f,0.95043f,0.96045f,0.96013f,0.97093f,0.95626f,0.95641f,0.98255f,0.96914f,0.98665f,0.98147f,0.99104f,0.99406f,0.99723f,1.00000f,0.00000f,0.99108f,0.98423f,0.98080f,0.97894f,0.97419f,0.97340f,0.97197f,0.97107f,0.97097f,0.97017f,0.96971f,0.96998f,0.97136f,0.97116f,0.97176f,0.97178f,0.97280f,0.97354f,0.97501f,0.97614f,0.97790f,0.97971f,0.97960f,0.98186f,0.98440f,0.98552f,0.98812f,0.98999f,0.99296f,0.99514f,0.99764f,1.00000f,0.00000f,0.99049f,0.98554f,0.98235f,0.98024f,0.97941f,0.97569f,0.97619f,0.97319f,0.97360f,0.97297f,0.97274f,0.97281f,0.97374f,0.97360f,0.97451f,0.97399f,0.97644f,0.97588f,0.97755f,0.97848f,0.97952f,0.98119f,0.98263f,0.98445f,0.98597f,0.98779f,0.98940f,0.99132f,0.99320f,0.99538f,0.99777f,1.00000f,0.00000f,0.98561f,0.98454f,0.98198f,0.97935f,0.97774f,0.97558f,0.97580f,0.97533f,0.97396f,0.97392f,0.97371f,0.97312f,0.97381f,0.97320f,0.97400f,0.97474f,0.97593f,0.97692f,0.97789f,0.97912f,0.98018f,0.98167f,0.98272f,0.98431f,0.98591f,0.98742f,0.98956f,0.99141f,0.99335f,0.99539f,0.99781f,1.00000f,0.00000f,0.94342f,0.98166f,0.98102f,0.97991f,0.97952f,0.97525f,0.97494f,0.97389f,0.97325f,0.97379f,0.97339f,0.97302f,0.97291f,0.97378f,0.97392f,0.97467f,0.97560f,0.97667f,0.97684f,0.97861f,0.97955f,0.98091f,0.98250f,0.98420f,0.98574f,0.98769f,0.98921f,0.99124f,0.99336f,0.99547f,0.99780f,1.00000f,0.00000f,0.91513f,0.95825f,0.97709f,0.97933f,0.97676f,0.97624f,0.97514f,0.97478f,0.97384f,0.97275f,0.97339f,0.97241f,0.97381f,0.97325f,0.97441f,0.97534f,0.97516f,0.97642f,0.97754f,0.97875f,0.97997f,0.98122f,0.98283f,0.98424f,0.98552f,0.98758f,0.98931f,0.99132f,0.99319f,0.99551f,0.99774f,1.00000f,0.00000f,0.90863f,0.93078f,0.96004f,0.97245f,0.97544f,0.97456f,0.97349f,0.97270f,0.97320f,0.97197f,0.97295f,0.97218f,0.97216f,0.97365f,0.97490f,0.97435f,0.97550f,0.97654f,0.97759f,0.97851f,0.98002f,0.98098f,0.98251f,0.98420f,0.98554f,0.98741f,0.98938f,0.99136f,0.99340f,0.99533f,0.99766f,1.00000f,0.00000f,0.90668f,0.91418f,0.93810f,0.95964f,0.96767f,0.97345f,0.97424f,0.97291f,0.97270f,0.97217f,0.97279f,0.97243f,0.97231f,0.97357f,0.97404f,0.97432f,0.97499f,0.97598f,0.97712f,0.97809f,0.97916f,0.98108f,0.98258f,0.98395f,0.98554f,0.98759f,0.98927f,0.99122f,0.99302f,0.99536f,0.99758f,1.00000f,0.00000f,0.91674f,0.91179f,0.92638f,0.94148f,0.95448f,0.96279f,0.96690f,0.97166f,0.97203f,0.97080f,0.97228f,0.97243f,0.97268f,0.97300f,0.97307f,0.97395f,0.97460f,0.97581f,0.97663f,0.97858f,0.97954f,0.98049f,0.98210f,0.98405f,0.98530f,0.98717f,0.98894f,0.99086f,0.99315f,0.99518f,0.99755f,1.00000f,0.00000f,0.92650f,0.90873f,0.91693f,0.92866f,0.94064f,0.94993f,0.95864f,0.96467f,0.96763f,0.97048f,0.97108f,0.97170f,0.97175f,0.97211f,0.97376f,0.97392f,0.97550f,0.97582f,0.97672f,0.97796f,0.97856f,0.98041f,0.98263f,0.98365f,0.98515f,0.98730f,0.98903f,0.99111f,0.99313f,0.99536f,0.99750f,1.00000f,0.00000f,0.94003f,0.92031f,0.91382f,0.91957f,0.92731f,0.94069f,0.94572f,0.95711f,0.95917f,0.96543f,0.96880f,0.96922f,0.97131f,0.97117f,0.97183f,0.97352f,0.97450f,0.97528f,0.97563f,0.97772f,0.97859f,0.98043f,0.98212f,0.98286f,0.98526f,0.98768f,0.98848f,0.99078f,0.99273f,0.99528f,0.99751f,1.00000f,0.00000f,0.93582f,0.92009f,0.91433f,0.91667f,0.92454f,0.92978f,0.93792f,0.94525f,0.95152f,0.95742f,0.96129f,0.96403f,0.96846f,0.97070f,0.97122f,0.97312f,0.97465f,0.97524f,0.97576f,0.97806f,0.97921f,0.98005f,0.98126f,0.98358f,0.98515f,0.98690f,0.98857f,0.99066f,0.99290f,0.99495f,0.99735f,1.00000f,0.00000f,0.93794f,0.93126f,0.91659f,0.91159f,0.92019f,0.92351f,0.92853f,0.93804f,0.94179f,0.94829f,0.95550f,0.95696f,0.96245f,0.96480f,0.96807f,0.97047f,0.97372f,0.97400f,0.97583f,0.97693f,0.97827f,0.97999f,0.98099f,0.98315f,0.98512f,0.98655f,0.98850f,0.99108f,0.99256f,0.99497f,0.99737f,1.00000f,0.00000f,0.94834f,0.92900f,0.92499f,0.91644f,0.92065f,0.92070f,0.92522f,0.92933f,0.93548f,0.93857f,0.94375f,0.94930f,0.95476f,0.95984f,0.96348f,0.96580f,0.96871f,0.97181f,0.97358f,0.97625f,0.97672f,0.97949f,0.98089f,0.98343f,0.98507f,0.98662f,0.98849f,0.99024f,0.99258f,0.99492f,0.99732f,1.00000f,0.00000f,0.95951f,0.94174f,0.92292f,0.92283f,0.91980f,0.91710f,0.92533f,0.92715f,0.93141f,0.93404f,0.93861f,0.94206f,0.94776f,0.95209f,0.95804f,0.95898f,0.96402f,0.96861f,0.96979f,0.97357f,0.97582f,0.97781f,0.97949f,0.98227f,0.98427f,0.98607f,0.98829f,0.99005f,0.99258f,0.99495f,0.99712f,1.00000f,0.00000f,0.95093f,0.94065f,0.93839f,0.92838f,0.91959f,0.92039f,0.91614f,0.92507f,0.92945f,0.93118f,0.93265f,0.93615f,0.94255f,0.94632f,0.94914f,0.95402f,0.95745f,0.96170f,0.96638f,0.96835f,0.97214f,0.97449f,0.97793f,0.97951f,0.98160f,0.98527f,0.98783f,0.99011f,0.99260f,0.99481f,0.99688f,1.00000f,0.00000f,0.96075f,0.94914f,0.94121f,0.93384f,0.93116f,0.92144f,0.92150f,0.92503f,0.92784f,0.92606f,0.92890f,0.93329f,0.93667f,0.93962f,0.94401f,0.94768f,0.95126f,0.95678f,0.96005f,0.96200f,0.96628f,0.97055f,0.97480f,0.97718f,0.97931f,0.98282f,0.98575f,0.98896f,0.99210f,0.99420f,0.99684f,1.00000f,0.00000f,0.95229f,0.95310f,0.94535f,0.93789f,0.92824f,0.92931f,0.92852f,0.92349f,0.92531f,0.92593f,0.92836f,0.93485f,0.93334f,0.93443f,0.93526f,0.94271f,0.94624f,0.94974f,0.95324f,0.95604f,0.96235f,0.96582f,0.96985f,0.97313f,0.97663f,0.98037f,0.98376f,0.98657f,0.98956f,0.99354f,0.99701f,1.00000f,0.00000f,0.96728f,0.94946f,0.94669f,0.93629f,0.93693f,0.93238f,0.92752f,0.92457f,0.92321f,0.92659f,0.92506f,0.93283f,0.92884f,0.93092f,0.93475f,0.93952f,0.94102f,0.94601f,0.94839f,0.95252f,0.95522f,0.95917f,0.96328f,0.96766f,0.97245f,0.97600f,0.97917f,0.98446f,0.98760f,0.99133f,0.99586f,1.00000f,0.00000f,0.96852f,0.95627f,0.94962f,0.94014f,0.93965f,0.93242f,0.92917f,0.92726f,0.92534f,0.92662f,0.92816f,0.92799f,0.92975f,0.92943f,0.93053f,0.93359f,0.93684f,0.93937f,0.94126f,0.94528f,0.94673f,0.95316f,0.95754f,0.96265f,0.96629f,0.97019f,0.97402f,0.97834f,0.98333f,0.98852f,0.99320f,0.99987f,0.00000f,0.96598f,0.95657f,0.95243f,0.94312f,0.93904f,0.94029f,0.92623f,0.92960f,0.92550f,0.92389f,0.92156f,0.92880f,0.92931f,0.93134f,0.92709f,0.93332f,0.92991f,0.92979f,0.93607f,0.93960f,0.94216f,0.94812f,0.95013f,0.95551f,0.95702f,0.96324f,0.96786f,0.97324f,0.97895f,0.98199f,0.98822f,0.99919f,0.00000f,0.96218f,0.96563f,0.95595f,0.95044f,0.94949f,0.94009f,0.93253f,0.93037f,0.93150f,0.92612f,0.92885f,0.92684f,0.92475f,0.92846f,0.92891f,0.92967f,0.92731f,0.93103f,0.93350f,0.93516f,0.93742f,0.94053f,0.94475f,0.94873f,0.95218f,0.95637f,0.95978f,0.96610f,0.97046f,0.97667f,0.98177f,0.99714f,0.00000f,0.96746f,0.95996f,0.95531f,0.95767f,0.94115f,0.94767f,0.93800f,0.93798f,0.92949f,0.93069f,0.92871f,0.93027f,0.92528f,0.93138f,0.92850f,0.92705f,0.92850f,0.92733f,0.93228f,0.92995f,0.93302f,0.93427f,0.93740f,0.94163f,0.94627f,0.94725f,0.95171f,0.95861f,0.96159f,0.96705f,0.97200f,0.99251f,0.00000f,0.96029f,0.95929f,0.96003f,0.94660f,0.94800f,0.94467f,0.94002f,0.94111f,0.94014f,0.93342f,0.92952f,0.92601f,0.92907f,0.92574f,0.92787f,0.92544f,0.92424f,0.92319f,0.92621f,0.92861f,0.92749f,0.92479f,0.93160f,0.93375f,0.93650f,0.94352f,0.94491f,0.94791f,0.95202f,0.95653f,0.96006f,0.98488f,0.00000f,0.96455f,0.95829f,0.95991f,0.95194f,0.95668f,0.94923f,0.95112f,0.93970f,0.93889f,0.93506f,0.92877f,0.93055f,0.92431f,0.92363f,0.92391f,0.92577f,0.93057f,0.92565f,0.92289f,0.92553f,0.92353f,0.92617f,0.92765f,0.92877f,0.92951f,0.93229f,0.93449f,0.93695f,0.94134f,0.94397f,0.94728f,0.97551f,0.00000f,0.95291f,0.95451f,0.95364f,0.95585f,0.94299f,0.95580f,0.94077f,0.94368f,0.93796f,0.93265f,0.93470f,0.92541f,0.92403f,0.92052f,0.92393f,0.92347f,0.92620f,0.92585f,0.92477f,0.92160f,0.92046f,0.91896f,0.92121f,0.91992f,0.91997f,0.92369f,0.92411f,0.92463f,0.92911f,0.92977f,0.93430f,0.96366f,0.00000f,0.94748f,0.95435f,0.95807f,0.94910f,0.95134f,0.95046f,0.94043f,0.94492f,0.93909f,0.93218f,0.92539f,0.92416f,0.92525f,0.92528f,0.92526f,0.92467f,0.91954f,0.91736f,0.92051f,0.91599f,0.91529f,0.91683f,0.91588f,0.91246f,0.91248f,0.90835f,0.91409f,0.91445f,0.91602f,0.91772f,0.91855f,0.95084f,0.00000f,0.94318f,0.95184f,0.94864f,0.94663f,0.94056f,0.94218f,0.92874f,0.93987f,0.93212f,0.93175f,0.92828f,0.92634f,0.92973f,0.92077f,0.92451f,0.91942f,0.91942f,0.91504f,0.91086f,0.90948f,0.91247f,0.90882f,0.91024f,0.90504f,0.90650f,0.90173f,0.90073f,0.90163f,0.90388f,0.90332f,0.90168f,0.93641f,0.00000f,0.93465f,0.93636f,0.93904f,0.93888f,0.93881f,0.93282f,0.93320f,0.92961f,0.92520f,0.91925f,0.92800f,0.92518f,0.92172f,0.91615f,0.91441f,0.91004f,0.91027f,0.90582f,0.90704f,0.90417f,0.90448f,0.90664f,0.89924f,0.89784f,0.89588f,0.89105f,0.89027f,0.88799f,0.88767f,0.88856f,0.88671f,0.92161f,0.00000f,0.92307f,0.92277f,0.92248f,0.92860f,0.92449f,0.92865f,0.91951f,0.91945f,0.92296f,0.91650f,0.91861f,0.91463f,0.91581f,0.91074f,0.90958f,0.90635f,0.90564f,0.90319f,0.89819f,0.89743f,0.89182f,0.89370f,0.89123f,0.88671f,0.88517f,0.87565f,0.87488f,0.87462f,0.87409f,0.87267f,0.87106f,0.90549f,0.00000f,0.90409f,0.91183f,0.90823f,0.91518f,0.91023f,0.91401f,0.91749f,0.90768f,0.91347f,0.91001f,0.90857f,0.90551f,0.90148f,0.90548f,0.90133f,0.90061f,0.89689f,0.89778f,0.89248f,0.88567f,0.89010f,0.88569f,0.88171f,0.87279f,0.87237f,0.86628f,0.86259f,0.86164f,0.86096f,0.85875f,0.85789f,0.89211f,0.00000f,0.88113f,0.89049f,0.89432f,0.89882f,0.89476f,0.89797f,0.90238f,0.89780f,0.89671f,0.89516f,0.89400f,0.89720f,0.89152f,0.89315f,0.88446f,0.88544f,0.88136f,0.88528f,0.87931f,0.87825f,0.87438f,0.86889f,0.87017f,0.86311f,0.85995f,0.85754f,0.84907f,0.84555f,0.84302f,0.84224f,0.84057f,0.87621f,0.00000f,0.86998f,0.87227f,0.87404f,0.87665f,0.88075f,0.87520f,0.88525f,0.88112f,0.88647f,0.88462f,0.88096f,0.87958f,0.87994f,0.87849f,0.87477f,0.87645f,0.86873f,0.86962f,0.86825f,0.85880f,0.86075f,0.85616f,0.85119f,0.85244f,0.84496f,0.84272f,0.83691f,0.83248f,0.82980f,0.82825f,0.82593f,0.86539f,
 0.00000f,0.99769f,0.98514f,0.97814f,0.97239f,0.96477f,0.97713f,0.96471f,0.95946f,0.95588f,0.95954f,0.96740f,0.96194f,0.95092f,0.95655f,0.96917f,0.94702f,0.96598f,0.96628f,0.96574f,0.95305f,0.95393f,0.97322f,0.96396f,0.96810f,0.97368f,0.96614f,0.98135f,0.97937f,0.98396f,0.99390f,0.99753f,1.00000f,0.00000f,0.99240f,0.98583f,0.98117f,0.97841f,0.97502f,0.97389f,0.97204f,0.97148f,0.97064f,0.97192f,0.97080f,0.97085f,0.96994f,0.97036f,0.97245f,0.97227f,0.97300f,0.97455f,0.97502f,0.97600f,0.97737f,0.97922f,0.98052f,0.98254f,0.98379f,0.98526f,0.98795f,0.98987f,0.99228f,0.99507f,0.99765f,1.00000f,0.00000f,0.99145f,0.98590f,0.98288f,0.98000f,0.97768f,0.97629f,0.97518f,0.97531f,0.97463f,0.97359f,0.97221f,0.97393f,0.97341f,0.97380f,0.97440f,0.97514f,0.97611f,0.97689f,0.97756f,0.97856f,0.97970f,0.98175f,0.98278f,0.98405f,0.98592f,0.98753f,0.98923f,0.99141f,0.99335f,0.99545f,0.99770f,1.00000f,0.00000f,0.98392f,0.98575f,0.98165f,0.97999f,0.97910f,0.97698f,0.97453f,0.97458f,0.97356f,0.97429f,0.97378f,0.97334f,0.97352f,0.97452f,0.97419f,0.97476f,0.97536f,0.97630f,0.97815f,0.97826f,0.97991f,0.98127f,0.98273f,0.98407f,0.98567f,0.98787f,0.98932f,0.99132f,0.99337f,0.99550f,0.99778f,1.00000f,0.00000f,0.94200f,0.98316f,0.98294f,0.97910f,0.97692f,0.97595f,0.97492f,0.97394f,0.97372f,0.97325f,0.97379f,0.97317f,0.97412f,0.97400f,0.97441f,0.97492f,0.97564f,0.97653f,0.97781f,0.97924f,0.97985f,0.98129f,0.98234f,0.98409f,0.98581f,0.98737f,0.98964f,0.99143f,0.99342f,0.99542f,0.99772f,1.00000f,0.00000f,0.91620f,0.95886f,0.97596f,0.97681f,0.97724f,0.97538f,0.97449f,0.97433f,0.97363f,0.97292f,0.97163f,0.97274f,0.97436f,0.97334f,0.97479f,0.97532f,0.97553f,0.97611f,0.97702f,0.97807f,0.97987f,0.98107f,0.98257f,0.98414f,0.98568f,0.98756f,0.98927f,0.99122f,0.99345f,0.99545f,0.99771f,1.00000f,0.00000f,0.90929f,0.93011f,0.96107f,0.97316f,0.97580f,0.97439f,0.97380f,0.97317f,0.97308f,0.97254f,0.97115f,0.97376f,0.97285f,0.97310f,0.97382f,0.97473f,0.97551f,0.97654f,0.97773f,0.97796f,0.97967f,0.98116f,0.98290f,0.98408f,0.98554f,0.98761f,0.98937f,0.99122f,0.99320f,0.99543f,0.99765f,1.00000f,0.00000f,0.91085f,0.91567f,0.93829f,0.95770f,0.97033f,0.97376f,0.97178f,0.97335f,0.97233f,0.97147f,0.97336f,0.97317f,0.97294f,0.97352f,0.97441f,0.97383f,0.97547f,0.97681f,0.97748f,0.97805f,0.97951f,0.98079f,0.98243f,0.98389f,0.98584f,0.98699f,0.98932f,0.99124f,0.99335f,0.99527f,0.99755f,1.00000f,0.00000f,0.92121f,0.91078f,0.92551f,0.94141f,0.95577f,0.96364f,0.96870f,0.97160f,0.97199f,0.97152f,0.97142f,0.97252f,0.97284f,0.97247f,0.97445f,0.97447f,0.97530f,0.97575f,0.97676f,0.97822f,0.97969f,0.98063f,0.98227f,0.98331f,0.98583f,0.98722f,0.98900f,0.99099f,0.99313f,0.99534f,0.99760f,1.00000f,0.00000f,0.92471f,0.91364f,0.91517f,0.92911f,0.93975f,0.94960f,0.96148f,0.96521f,0.96737f,0.97059f,0.97155f,0.97227f,0.97185f,0.97220f,0.97370f,0.97426f,0.97379f,0.97621f,0.97626f,0.97818f,0.97905f,0.98082f,0.98182f,0.98354f,0.98535f,0.98701f,0.98882f,0.99092f,0.99289f,0.99538f,0.99744f,1.00000f,0.00000f,0.93895f,0.91793f,0.91117f,0.91804f,0.92970f,0.93908f,0.94578f,0.95502f,0.95961f,0.96547f,0.96829f,0.96843f,0.97158f,0.97275f,0.97324f,0.97381f,0.97454f,0.97490f,0.97689f,0.97755f,0.97931f,0.98055f,0.98192f,0.98274f,0.98505f,0.98671f,0.98866f,0.99113f,0.99311f,0.99533f,0.99756f,1.00000f,0.00000f,0.94662f,0.92038f,0.91376f,0.91625f,0.92252f,0.93293f,0.93708f,0.94487f,0.95148f,0.95783f,0.96111f,0.96493f,0.96847f,0.97072f,0.97184f,0.97291f,0.97328f,0.97573f,0.97710f,0.97667f,0.97817f,0.98080f,0.98173f,0.98299f,0.98416f,0.98710f,0.98858f,0.99099f,0.99290f,0.99502f,0.99737f,1.00000f,0.00000f,0.94616f,0.92392f,0.91977f,0.91501f,0.92105f,0.92315f,0.92973f,0.93389f,0.94388f,0.95055f,0.95351f,0.95857f,0.96376f,0.96561f,0.96767f,0.97053f,0.97135f,0.97407f,0.97593f,0.97681f,0.97883f,0.98086f,0.98148f,0.98374f,0.98495f,0.98641f,0.98912f,0.99043f,0.99296f,0.99512f,0.99726f,1.00000f,0.00000f,0.95157f,0.92667f,0.92337f,0.91569f,0.92209f,0.91976f,0.92718f,0.92978f,0.93485f,0.93789f,0.94346f,0.95132f,0.95403f,0.96047f,0.96251f,0.96686f,0.96770f,0.97146f,0.97371f,0.97475f,0.97724f,0.97881f,0.98133f,0.98324f,0.98468f,0.98637f,0.98881f,0.99089f,0.99235f,0.99447f,0.99737f,1.00000f,0.00000f,0.94709f,0.94046f,0.92951f,0.92410f,0.92047f,0.92263f,0.92315f,0.93041f,0.92934f,0.93537f,0.93910f,0.94496f,0.94798f,0.95406f,0.95582f,0.95971f,0.96574f,0.96770f,0.97030f,0.97337f,0.97548f,0.97784f,0.98030f,0.98206f,0.98415f,0.98638f,0.98838f,0.99026f,0.99216f,0.99476f,0.99732f,1.00000f,0.00000f,0.95439f,0.94332f,0.93725f,0.92702f,0.91688f,0.92142f,0.92491f,0.92386f,0.92530f,0.92822f,0.93166f,0.93760f,0.93902f,0.94451f,0.94988f,0.95263f,0.95808f,0.96272f,0.96458f,0.96946f,0.97068f,0.97386f,0.97799f,0.98008f,0.98293f,0.98501f,0.98782f,0.98979f,0.99206f,0.99486f,0.99705f,1.00000f,0.00000f,0.95643f,0.95105f,0.93786f,0.93226f,0.92254f,0.92449f,0.92339f,0.92654f,0.92534f,0.92912f,0.92963f,0.93188f,0.93803f,0.94201f,0.94372f,0.94928f,0.95086f,0.95411f,0.95833f,0.96353f,0.96692f,0.97097f,0.97462f,0.97713f,0.98038f,0.98240f,0.98539f,0.98947f,0.99207f,0.99443f,0.99725f,1.00000f,0.00000f,0.96015f,0.95481f,0.94233f,0.93694f,0.92793f,0.92790f,0.93042f,0.92305f,0.92273f,0.92704f,0.92788f,0.93065f,0.93104f,0.93261f,0.93905f,0.93950f,0.94616f,0.94889f,0.95362f,0.95600f,0.96083f,0.96481f,0.96955f,0.97298f,0.97730f,0.98033f,0.98310f,0.98764f,0.99045f,0.99353f,0.99695f,1.00000f,0.00000f,0.96190f,0.95245f,0.94238f,0.94138f,0.92924f,0.92946f,0.92738f,0.92523f,0.92437f,0.92562f,0.92695f,0.92832f,0.93163f,0.93561f,0.93296f,0.93735f,0.94203f,0.94253f,0.94688f,0.95183f,0.95536f,0.96071f,0.96170f,0.96783f,0.97042f,0.97576f,0.97874f,0.98322f,0.98723f,0.99118f,0.99533f,1.00000f,0.00000f,0.95454f,0.95574f,0.95043f,0.94920f,0.94363f,0.93292f,0.92772f,0.92904f,0.92930f,0.92694f,0.92469f,0.92980f,0.93098f,0.92962f,0.93229f,0.93146f,0.93459f,0.93681f,0.93869f,0.94205f,0.94867f,0.95257f,0.95873f,0.95985f,0.96425f,0.96887f,0.97300f,0.97890f,0.98402f,0.98745f,0.99270f,0.99987f,0.00000f,0.97193f,0.95778f,0.95272f,0.94305f,0.93918f,0.93895f,0.93311f,0.92801f,0.92749f,0.92486f,0.92558f,0.92934f,0.92713f,0.92616f,0.92616f,0.93334f,0.93183f,0.93461f,0.93734f,0.93948f,0.94069f,0.94686f,0.95150f,0.95512f,0.95793f,0.96262f,0.96790f,0.97273f,0.97783f,0.98239f,0.98855f,0.99922f,0.00000f,0.96283f,0.95898f,0.96013f,0.94609f,0.94276f,0.93911f,0.93864f,0.93017f,0.92922f,0.92978f,0.92635f,0.92972f,0.92982f,0.92654f,0.92811f,0.92958f,0.92966f,0.93080f,0.93449f,0.93384f,0.93705f,0.93813f,0.94283f,0.94900f,0.95320f,0.95636f,0.95968f,0.96595f,0.96947f,0.97517f,0.98101f,0.99721f,0.00000f,0.96200f,0.96015f,0.96035f,0.94880f,0.94593f,0.93866f,0.93870f,0.94237f,0.92666f,0.92839f,0.92736f,0.92746f,0.92264f,0.92659f,0.92755f,0.92811f,0.92683f,0.92686f,0.92824f,0.92792f,0.93109f,0.93312f,0.93757f,0.94020f,0.94292f,0.94790f,0.95340f,0.95713f,0.96136f,0.96570f,0.97076f,0.99264f,0.00000f,0.96227f,0.95927f,0.95878f,0.94694f,0.94830f,0.94764f,0.93804f,0.93892f,0.93553f,0.93125f,0.92948f,0.92385f,0.92731f,0.92494f,0.93000f,0.92616f,0.92278f,0.92586f,0.92554f,0.92812f,0.92640f,0.92946f,0.93192f,0.93382f,0.93651f,0.93937f,0.94351f,0.94673f,0.95094f,0.95530f,0.95947f,0.98509f,0.00000f,0.96419f,0.96161f,0.95517f,0.95326f,0.94504f,0.94767f,0.95060f,0.93426f,0.93782f,0.93978f,0.93496f,0.92958f,0.93041f,0.92533f,0.92296f,0.92384f,0.92442f,0.92630f,0.92725f,0.92526f,0.92285f,0.92343f,0.92588f,0.92477f,0.92772f,0.93070f,0.93230f,0.93857f,0.94061f,0.94248f,0.94636f,0.97483f,0.00000f,0.95535f,0.95921f,0.95595f,0.95326f,0.94610f,0.94497f,0.94684f,0.93989f,0.93173f,0.93585f,0.93481f,0.93018f,0.92525f,0.92522f,0.92027f,0.91958f,0.92116f,0.92460f,0.92104f,0.92198f,0.92053f,0.91991f,0.91687f,0.92221f,0.91683f,0.91968f,0.92203f,0.92508f,0.92660f,0.92972f,0.93174f,0.96271f,0.00000f,0.95320f,0.95220f,0.94972f,0.94850f,0.95050f,0.94495f,0.94290f,0.94182f,0.93735f,0.93668f,0.93033f,0.92085f,0.93149f,0.92530f,0.92163f,0.91787f,0.92038f,0.91859f,0.91555f,0.91845f,0.91650f,0.91512f,0.91451f,0.91033f,0.90921f,0.90905f,0.91135f,0.91089f,0.91319f,0.91337f,0.91590f,0.94744f,0.00000f,0.94828f,0.95356f,0.94858f,0.94698f,0.94637f,0.94213f,0.93735f,0.93522f,0.93424f,0.92957f,0.92963f,0.92590f,0.92637f,0.91842f,0.92188f,0.91589f,0.92050f,0.91191f,0.91240f,0.91369f,0.90661f,0.90956f,0.90444f,0.90123f,0.90342f,0.90125f,0.90216f,0.89865f,0.89856f,0.89936f,0.89878f,0.93567f,0.00000f,0.93346f,0.93862f,0.94512f,0.93497f,0.94080f,0.93906f,0.93212f,0.93711f,0.92948f,0.92861f,0.92350f,0.91882f,0.91792f,0.92170f,0.91522f,0.91737f,0.91028f,0.90700f,0.90566f,0.90640f,0.90193f,0.90232f,0.89850f,0.89632f,0.89610f,0.88982f,0.88882f,0.88401f,0.88582f,0.88475f,0.88330f,0.91955f,0.00000f,0.92425f,0.92911f,0.92814f,0.93308f,0.93369f,0.92924f,0.92547f,0.92036f,0.91816f,0.91770f,0.91031f,0.91168f,0.91617f,0.91020f,0.91043f,0.90674f,0.90171f,0.90439f,0.90416f,0.89674f,0.89489f,0.89520f,0.89114f,0.88588f,0.88177f,0.87871f,0.87439f,0.87120f,0.86972f,0.87035f,0.86978f,0.90474f,0.00000f,0.90965f,0.90900f,0.91063f,0.91556f,0.91083f,0.91499f,0.91366f,0.91093f,0.90712f,0.91409f,0.91194f,0.90467f,0.90344f,0.90738f,0.89861f,0.90402f,0.89566f,0.89089f,0.89058f,0.88444f,0.88415f,0.88331f,0.87909f,0.87293f,0.86850f,0.86510f,0.86217f,0.85708f,0.85876f,0.85459f,0.85350f,0.88939f,0.00000f,0.88004f,0.89433f,0.89213f,0.89877f,0.89695f,0.90103f,0.90264f,0.90288f,0.89419f,0.90216f,0.90416f,0.89302f,0.89020f,0.88977f,0.88473f,0.88697f,0.88479f,0.88295f,0.88315f,0.87357f,0.87221f,0.87202f,0.86470f,0.86205f,0.85739f,0.85213f,0.84715f,0.84172f,0.84041f,0.84077f,0.84085f,0.87480f,0.00000f,0.86589f,0.87522f,0.87749f,0.87748f,0.87777f,0.88125f,0.88163f,0.88595f,0.88714f,0.88185f,0.88004f,0.87976f,0.87831f,0.87802f,0.87258f,0.87625f,0.86826f,0.86661f,0.86631f,0.86015f,0.85614f,0.85776f,0.84800f,0.84423f,0.83932f,0.83348f,0.83286f,0.82686f,0.82691f,0.82357f,0.82436f,0.85947f,
 0.00000f,0.99309f,0.98131f,0.98469f,0.96794f,0.95471f,0.97598f,0.97771f,0.95723f,0.95264f,0.95003f,0.96439f,0.96356f,0.96448f,0.96666f,0.94856f,0.96682f,0.96305f,0.94594f,0.96731f,0.95570f,0.95685f,0.97511f,0.97788f,0.96465f,0.97689f,0.97641f,0.98158f,0.99058f,0.99314f,0.99098f,0.99417f,1.00000f,0.00000f,0.98967f,0.98498f,0.98146f,0.97889f,0.97573f,0.97426f,0.97330f,0.97153f,0.96977f,0.97035f,0.97072f,0.97117f,0.97060f,0.97228f,0.97093f,0.97208f,0.97273f,0.97417f,0.97508f,0.97632f,0.97679f,0.97936f,0.98171f,0.98208f,0.98471f,0.98649f,0.98851f,0.99077f,0.99294f,0.99497f,0.99721f,1.00000f,0.00000f,0.99298f,0.98564f,0.98283f,0.98053f,0.97855f,0.97735f,0.97501f,0.97478f,0.97296f,0.97432f,0.97402f,0.97306f,0.97349f,0.97332f,0.97414f,0.97512f,0.97596f,0.97705f,0.97797f,0.97906f,0.97972f,0.98125f,0.98267f,0.98427f,0.98594f,0.98758f,0.98936f,0.99147f,0.99338f,0.99550f,0.99775f,1.00000f,0.00000f,0.98413f,0.98611f,0.98280f,0.97860f,0.97773f,0.97567f,0.97642f,0.97552f,0.97367f,0.97320f,0.97283f,0.97320f,0.97302f,0.97448f,0.97434f,0.97514f,0.97574f,0.97717f,0.97782f,0.97897f,0.97974f,0.98188f,0.98284f,0.98388f,0.98594f,0.98767f,0.98963f,0.99127f,0.99327f,0.99549f,0.99774f,1.00000f,0.00000f,0.94296f,0.98076f,0.98252f,0.97961f,0.97767f,0.97637f,0.97461f,0.97508f,0.97391f,0.97263f,0.97277f,0.97279f,0.97309f,0.97464f,0.97364f,0.97466f,0.97552f,0.97670f,0.97759f,0.97894f,0.97969f,0.98138f,0.98277f,0.98382f,0.98611f,0.98774f,0.98949f,0.99120f,0.99336f,0.99538f,0.99777f,1.00000f,0.00000f,0.91351f,0.96128f,0.97770f,0.97881f,0.97710f,0.97426f,0.97522f,0.97457f,0.97290f,0.97252f,0.97293f,0.97276f,0.97302f,0.97289f,0.97400f,0.97535f,0.97521f,0.97630f,0.97794f,0.97858f,0.97972f,0.98116f,0.98250f,0.98422f,0.98595f,0.98745f,0.98911f,0.99118f,0.99322f,0.99542f,0.99773f,1.00000f,0.00000f,0.90586f,0.92875f,0.96063f,0.97396f,0.97586f,0.97572f,0.97486f,0.97384f,0.97270f,0.97308f,0.97355f,0.97231f,0.97241f,0.97388f,0.97456f,0.97472f,0.97574f,0.97609f,0.97747f,0.97841f,0.97960f,0.98140f,0.98262f,0.98400f,0.98578f,0.98782f,0.98909f,0.99111f,0.99331f,0.99540f,0.99761f,1.00000f,0.00000f,0.90855f,0.91657f,0.93862f,0.95729f,0.96647f,0.97189f,0.97261f,0.97289f,0.97250f,0.97272f,0.97228f,0.97262f,0.97284f,0.97287f,0.97373f,0.97304f,0.97597f,0.97657f,0.97717f,0.97805f,0.98000f,0.98085f,0.98197f,0.98415f,0.98589f,0.98730f,0.98938f,0.99116f,0.99314f,0.99531f,0.99755f,1.00000f,0.00000f,0.91513f,0.91059f,0.92491f,0.94058f,0.95526f,0.96428f,0.96781f,0.97156f,0.97175f,0.97263f,0.97238f,0.97122f,0.97212f,0.97307f,0.97338f,0.97433f,0.97511f,0.97598f,0.97739f,0.97864f,0.97970f,0.98121f,0.98171f,0.98377f,0.98527f,0.98727f,0.98911f,0.99105f,0.99319f,0.99521f,0.99751f,1.00000f,0.00000f,0.92149f,0.91337f,0.91735f,0.92721f,0.94135f,0.95459f,0.95900f,0.96388f,0.96878f,0.96893f,0.96915f,0.97031f,0.97144f,0.97230f,0.97238f,0.97417f,0.97515f,0.97648f,0.97723f,0.97792f,0.97866f,0.98077f,0.98246f,0.98353f,0.98570f,0.98712f,0.98867f,0.99110f,0.99291f,0.99533f,0.99744f,1.00000f,0.00000f,0.93433f,0.91390f,0.92070f,0.92479f,0.92990f,0.93755f,0.94895f,0.95395f,0.96065f,0.96581f,0.96635f,0.96849f,0.97250f,0.97165f,0.97284f,0.97393f,0.97441f,0.97598f,0.97674f,0.97771f,0.97852f,0.98010f,0.98205f,0.98356f,0.98481f,0.98690f,0.98871f,0.99092f,0.99316f,0.99499f,0.99742f,1.00000f,0.00000f,0.94006f,0.92380f,0.91569f,0.91506f,0.92337f,0.93097f,0.93745f,0.94463f,0.95194f,0.95505f,0.96208f,0.96335f,0.96788f,0.97099f,0.97110f,0.97318f,0.97363f,0.97604f,0.97632f,0.97842f,0.97845f,0.98054f,0.98190f,0.98271f,0.98484f,0.98665f,0.98883f,0.99107f,0.99307f,0.99505f,0.99738f,1.00000f,0.00000f,0.94564f,0.92812f,0.91979f,0.91666f,0.92058f,0.92306f,0.93006f,0.93467f,0.93904f,0.94796f,0.95184f,0.95676f,0.96199f,0.96642f,0.96823f,0.97043f,0.97147f,0.97359f,0.97538f,0.97639f,0.97810f,0.98092f,0.98187f,0.98296f,0.98538f,0.98642f,0.98822f,0.99076f,0.99220f,0.99508f,0.99739f,1.00000f,0.00000f,0.94793f,0.93147f,0.92083f,0.92156f,0.91932f,0.92143f,0.92617f,0.93024f,0.93330f,0.94066f,0.94579f,0.95207f,0.95501f,0.95806f,0.96134f,0.96549f,0.96844f,0.97136f,0.97275f,0.97499f,0.97673f,0.97897f,0.98108f,0.98262f,0.98487f,0.98655f,0.98848f,0.99051f,0.99231f,0.99496f,0.99741f,1.00000f,0.00000f,0.95736f,0.93624f,0.92765f,0.92431f,0.91756f,0.92171f,0.92724f,0.92795f,0.92878f,0.93354f,0.93830f,0.94287f,0.94849f,0.95107f,0.95691f,0.95989f,0.96524f,0.96903f,0.97034f,0.97225f,0.97665f,0.97662f,0.97940f,0.98223f,0.98334f,0.98664f,0.98860f,0.99069f,0.99222f,0.99511f,0.99733f,1.00000f,0.00000f,0.94865f,0.94128f,0.93121f,0.92357f,0.91833f,0.92096f,0.92345f,0.92410f,0.92661f,0.92975f,0.93238f,0.93612f,0.93921f,0.94393f,0.95104f,0.95391f,0.95756f,0.96218f,0.96479f,0.96859f,0.97153f,0.97448f,0.97718f,0.98054f,0.98329f,0.98506f,0.98794f,0.98945f,0.99230f,0.99494f,0.99710f,1.00000f,0.00000f,0.95165f,0.94590f,0.93759f,0.92754f,0.92476f,0.92252f,0.92323f,0.92581f,0.92564f,0.92775f,0.93137f,0.93447f,0.93490f,0.93985f,0.94528f,0.94865f,0.95130f,0.95524f,0.95992f,0.96300f,0.96586f,0.97142f,0.97325f,0.97694f,0.98010f,0.98292f,0.98593f,0.98856f,0.99236f,0.99406f,0.99687f,1.00000f,0.00000f,0.95639f,0.95297f,0.94310f,0.93392f,0.93021f,0.92435f,0.92404f,0.92242f,0.92509f,0.92524f,0.92886f,0.92879f,0.93754f,0.93669f,0.93853f,0.94203f,0.94517f,0.94886f,0.95238f,0.95779f,0.96184f,0.96558f,0.97006f,0.97278f,0.97646f,0.98039f,0.98345f,0.98733f,0.99036f,0.99320f,0.99712f,1.00000f,0.00000f,0.95661f,0.95220f,0.94628f,0.93781f,0.93685f,0.92989f,0.92743f,0.93121f,0.92559f,0.92656f,0.92376f,0.92727f,0.93221f,0.93078f,0.93283f,0.93819f,0.94035f,0.94129f,0.94448f,0.95061f,0.95329f,0.95954f,0.96276f,0.96657f,0.97191f,0.97495f,0.97952f,0.98312f,0.98731f,0.99129f,0.99526f,1.00000f,0.00000f,0.95800f,0.95465f,0.95143f,0.94458f,0.93822f,0.93493f,0.93323f,0.93044f,0.92315f,0.93037f,0.93011f,0.92779f,0.92709f,0.93287f,0.93275f,0.93278f,0.93452f,0.93695f,0.94137f,0.94364f,0.94711f,0.95476f,0.95699f,0.96076f,0.96484f,0.97048f,0.97238f,0.97905f,0.98262f,0.98814f,0.99267f,0.99987f,0.00000f,0.96126f,0.95826f,0.96260f,0.95103f,0.93907f,0.93127f,0.92796f,0.93221f,0.92887f,0.93171f,0.92851f,0.92778f,0.92449f,0.92334f,0.93072f,0.93146f,0.92943f,0.92982f,0.93592f,0.94049f,0.94218f,0.94691f,0.94995f,0.95463f,0.95708f,0.96463f,0.96801f,0.97228f,0.97658f,0.98175f,0.98767f,0.99919f,0.00000f,0.96793f,0.96124f,0.95505f,0.94908f,0.94683f,0.93735f,0.93921f,0.93707f,0.92964f,0.92788f,0.92752f,0.92369f,0.92611f,0.92692f,0.92680f,0.93185f,0.92866f,0.93329f,0.93031f,0.93439f,0.93690f,0.94005f,0.94251f,0.94716f,0.94989f,0.95632f,0.95896f,0.96398f,0.96852f,0.97313f,0.98059f,0.99718f,0.00000f,0.96415f,0.96246f,0.96202f,0.95127f,0.94164f,0.94241f,0.93544f,0.93933f,0.93407f,0.92761f,0.93081f,0.93025f,0.92607f,0.92497f,0.92622f,0.93028f,0.92888f,0.92598f,0.92945f,0.92818f,0.92773f,0.93232f,0.93841f,0.93907f,0.94306f,0.94812f,0.95158f,0.95496f,0.96061f,0.96668f,0.97114f,0.99196f,0.00000f,0.96519f,0.96174f,0.95436f,0.95110f,0.94651f,0.94224f,0.93899f,0.93829f,0.93281f,0.93042f,0.92625f,0.93188f,0.92904f,0.92517f,0.92836f,0.92891f,0.92610f,0.92878f,0.92660f,0.92230f,0.92688f,0.92661f,0.93101f,0.93284f,0.93436f,0.94013f,0.94313f,0.94733f,0.95094f,0.95363f,0.95859f,0.98481f,0.00000f,0.95877f,0.95877f,0.96210f,0.95994f,0.95571f,0.95028f,0.93873f,0.94157f,0.93684f,0.93139f,0.92688f,0.93053f,0.92596f,0.92382f,0.92647f,0.92755f,0.92248f,0.92116f,0.92359f,0.92215f,0.92440f,0.92330f,0.92417f,0.92731f,0.92860f,0.93017f,0.93202f,0.93526f,0.93796f,0.94176f,0.94376f,0.97442f,0.00000f,0.95824f,0.95590f,0.96195f,0.95693f,0.95213f,0.94893f,0.94250f,0.93955f,0.93626f,0.93254f,0.92789f,0.92946f,0.92740f,0.92689f,0.92013f,0.92794f,0.92361f,0.92305f,0.92106f,0.92052f,0.92190f,0.91641f,0.91607f,0.91744f,0.91847f,0.91885f,0.92289f,0.92340f,0.92679f,0.92629f,0.92883f,0.96112f,0.00000f,0.95477f,0.96516f,0.95597f,0.94940f,0.94804f,0.94807f,0.94561f,0.93808f,0.94362f,0.93765f,0.93199f,0.92501f,0.92432f,0.92404f,0.92085f,0.91905f,0.92086f,0.92041f,0.91899f,0.91867f,0.91347f,0.91450f,0.91381f,0.91308f,0.90979f,0.90868f,0.90955f,0.91129f,0.91349f,0.91293f,0.91184f,0.94784f,0.00000f,0.95046f,0.95059f,0.95100f,0.94884f,0.94555f,0.93849f,0.93984f,0.93921f,0.93099f,0.93162f,0.92301f,0.92313f,0.92174f,0.92124f,0.91763f,0.91688f,0.91709f,0.91526f,0.91373f,0.90833f,0.90721f,0.91062f,0.90322f,0.90588f,0.89868f,0.89849f,0.89586f,0.89840f,0.89763f,0.89741f,0.89806f,0.93102f,0.00000f,0.93115f,0.94060f,0.94003f,0.94112f,0.94114f,0.93749f,0.93638f,0.93151f,0.93306f,0.93118f,0.93218f,0.92561f,0.92050f,0.91182f,0.91599f,0.91328f,0.91017f,0.90997f,0.90613f,0.90345f,0.89917f,0.90041f,0.89694f,0.89666f,0.89325f,0.88768f,0.88338f,0.88360f,0.88165f,0.88055f,0.88171f,0.91592f,0.00000f,0.92364f,0.92203f,0.92404f,0.92948f,0.93220f,0.92327f,0.92646f,0.92287f,0.92629f,0.92029f,0.91181f,0.91615f,0.91625f,0.91165f,0.91280f,0.90168f,0.89999f,0.90166f,0.89769f,0.89791f,0.89244f,0.89347f,0.88690f,0.88678f,0.88377f,0.87539f,0.87141f,0.87036f,0.86622f,0.86569f,0.86606f,0.90201f,0.00000f,0.90933f,0.90786f,0.91457f,0.91998f,0.91675f,0.91474f,0.91029f,0.91289f,0.91587f,0.91576f,0.90792f,0.90250f,0.90462f,0.90236f,0.89981f,0.90071f,0.89518f,0.89036f,0.89030f,0.88407f,0.88208f,0.88264f,0.87603f,0.87561f,0.86799f,0.86041f,0.85556f,0.85313f,0.85281f,0.85109f,0.85239f,0.88419f,0.00000f,0.88757f,0.89499f,0.89630f,0.89688f,0.89914f,0.89829f,0.90235f,0.89610f,0.89533f,0.90248f,0.89504f,0.89708f,0.88961f,0.88569f,0.88713f,0.88646f,0.88265f,0.88171f,0.87978f,0.87313f,0.86970f,0.86743f,0.86580f,0.85846f,0.85512f,0.84852f,0.84452f,0.83915f,0.83824f,0.83792f,0.83614f,0.87136f,0.00000f,0.87008f,0.87346f,0.88376f,0.88372f,0.88473f,0.88401f,0.87995f,0.89168f,0.88382f,0.87998f,0.88231f,0.88272f,0.87846f,0.87798f,0.87528f,0.87178f,0.86913f,0.86595f,0.86046f,0.85787f,0.85669f,0.85597f,0.84708f,0.84426f,0.83641f,0.83023f,0.82807f,0.82704f,0.82331f,0.81923f,0.81993f,0.85670f,
};

// 33x33 average albedo table calculated using moonshine_devtools/compute_albedo for the following ranges:
// F*Reflection + (1-F)*Transmission
// EtaT/EtaI [1, 3]
// Roughness [0,1]
const float
TransmissionAlbedo::mOneMinusEavg[COMP*COMP] =
{
 0.03107f,0.01817f,0.01573f,0.01560f,0.01573f,0.01608f,0.01636f,0.01674f,0.01718f,0.01778f,0.01842f,0.01896f,0.01969f,0.02023f,0.02092f,0.02157f,0.02242f,0.02373f,0.02504f,0.02675f,0.02940f,0.03169f,0.03503f,0.03780f,0.04288f,0.04738f,0.05271f,0.05847f,0.06464f,0.07125f,0.07858f,0.08516f,0.09264f,
 0.03025f,0.01820f,0.01570f,0.01564f,0.01579f,0.01603f,0.01649f,0.01696f,0.01756f,0.01835f,0.01913f,0.01971f,0.02074f,0.02151f,0.02232f,0.02334f,0.02388f,0.02489f,0.02600f,0.02749f,0.02901f,0.03067f,0.03341f,0.03612f,0.03962f,0.04354f,0.04793f,0.05448f,0.05956f,0.06666f,0.07335f,0.08089f,0.08925f,
 0.03012f,0.01825f,0.01570f,0.01562f,0.01571f,0.01607f,0.01644f,0.01697f,0.01754f,0.01841f,0.01932f,0.02024f,0.02113f,0.02235f,0.02342f,0.02440f,0.02554f,0.02663f,0.02759f,0.02874f,0.02992f,0.03195f,0.03343f,0.03628f,0.03951f,0.04326f,0.04792f,0.05255f,0.05835f,0.06454f,0.07166f,0.07898f,0.08790f,
 0.02896f,0.01803f,0.01573f,0.01565f,0.01582f,0.01613f,0.01642f,0.01689f,0.01762f,0.01840f,0.01938f,0.02041f,0.02161f,0.02285f,0.02391f,0.02551f,0.02683f,0.02771f,0.02878f,0.03028f,0.03132f,0.03262f,0.03504f,0.03682f,0.03964f,0.04329f,0.04732f,0.05232f,0.05777f,0.06386f,0.07114f,0.07878f,0.08689f,
 0.02935f,0.01797f,0.01569f,0.01563f,0.01583f,0.01612f,0.01654f,0.01693f,0.01765f,0.01843f,0.01962f,0.02047f,0.02180f,0.02327f,0.02447f,0.02601f,0.02748f,0.02881f,0.03027f,0.03176f,0.03311f,0.03471f,0.03595f,0.03767f,0.04052f,0.04358f,0.04792f,0.05222f,0.05817f,0.06448f,0.07117f,0.07841f,0.08803f,
 0.03006f,0.01799f,0.01567f,0.01566f,0.01575f,0.01599f,0.01647f,0.01692f,0.01765f,0.01862f,0.01959f,0.02069f,0.02207f,0.02366f,0.02483f,0.02638f,0.02816f,0.02949f,0.03081f,0.03235f,0.03367f,0.03534f,0.03717f,0.03943f,0.04195f,0.04496f,0.04871f,0.05341f,0.05908f,0.06507f,0.07290f,0.07970f,0.08887f,
 0.02811f,0.01792f,0.01575f,0.01567f,0.01578f,0.01614f,0.01640f,0.01711f,0.01763f,0.01856f,0.01961f,0.02096f,0.02219f,0.02375f,0.02511f,0.02670f,0.02806f,0.03006f,0.03155f,0.03346f,0.03473f,0.03681f,0.03800f,0.04067f,0.04296f,0.04679f,0.04951f,0.05388f,0.05984f,0.06625f,0.07282f,0.08217f,0.09007f,
 0.02922f,0.01795f,0.01560f,0.01571f,0.01572f,0.01601f,0.01653f,0.01700f,0.01764f,0.01836f,0.01967f,0.02093f,0.02229f,0.02381f,0.02535f,0.02706f,0.02891f,0.03027f,0.03193f,0.03389f,0.03598f,0.03738f,0.03954f,0.04189f,0.04445f,0.04756f,0.05157f,0.05558f,0.06184f,0.06809f,0.07454f,0.08313f,0.09190f,
 0.02742f,0.01783f,0.01569f,0.01562f,0.01583f,0.01606f,0.01647f,0.01698f,0.01781f,0.01851f,0.01970f,0.02085f,0.02217f,0.02401f,0.02544f,0.02752f,0.02903f,0.03085f,0.03272f,0.03442f,0.03669f,0.03851f,0.04066f,0.04282f,0.04560f,0.04891f,0.05254f,0.05669f,0.06243f,0.06838f,0.07616f,0.08432f,0.09433f,
 0.02777f,0.01784f,0.01572f,0.01570f,0.01584f,0.01597f,0.01644f,0.01700f,0.01761f,0.01861f,0.01979f,0.02112f,0.02257f,0.02393f,0.02580f,0.02746f,0.02940f,0.03113f,0.03337f,0.03500f,0.03761f,0.03950f,0.04148f,0.04405f,0.04620f,0.04964f,0.05338f,0.05842f,0.06432f,0.07024f,0.07778f,0.08635f,0.09667f,
 0.02858f,0.01800f,0.01575f,0.01563f,0.01582f,0.01607f,0.01646f,0.01702f,0.01778f,0.01856f,0.01972f,0.02115f,0.02249f,0.02417f,0.02578f,0.02772f,0.02972f,0.03164f,0.03352f,0.03553f,0.03786f,0.04017f,0.04252f,0.04469f,0.04785f,0.05063f,0.05476f,0.05911f,0.06520f,0.07245f,0.07958f,0.08793f,0.09867f,
 0.02715f,0.01772f,0.01571f,0.01568f,0.01577f,0.01612f,0.01649f,0.01707f,0.01780f,0.01849f,0.01966f,0.02088f,0.02252f,0.02401f,0.02600f,0.02790f,0.02962f,0.03152f,0.03399f,0.03622f,0.03804f,0.04065f,0.04345f,0.04513f,0.04939f,0.05186f,0.05616f,0.06070f,0.06669f,0.07354f,0.08119f,0.09127f,0.10101f,
 0.02765f,0.01771f,0.01559f,0.01561f,0.01576f,0.01595f,0.01641f,0.01703f,0.01771f,0.01864f,0.01980f,0.02096f,0.02246f,0.02422f,0.02611f,0.02792f,0.03027f,0.03199f,0.03449f,0.03635f,0.03882f,0.04179f,0.04418f,0.04642f,0.05036f,0.05337f,0.05729f,0.06282f,0.06807f,0.07521f,0.08312f,0.09264f,0.10302f,
 0.02620f,0.01762f,0.01568f,0.01567f,0.01579f,0.01612f,0.01636f,0.01707f,0.01766f,0.01867f,0.01974f,0.02126f,0.02243f,0.02403f,0.02606f,0.02797f,0.02982f,0.03266f,0.03442f,0.03703f,0.03941f,0.04183f,0.04492f,0.04749f,0.05037f,0.05484f,0.05886f,0.06371f,0.06860f,0.07723f,0.08547f,0.09465f,0.10522f,
 0.02916f,0.01766f,0.01568f,0.01557f,0.01570f,0.01601f,0.01652f,0.01698f,0.01775f,0.01864f,0.01987f,0.02117f,0.02274f,0.02432f,0.02638f,0.02839f,0.03001f,0.03261f,0.03478f,0.03727f,0.04002f,0.04248f,0.04502f,0.04834f,0.05196f,0.05553f,0.05978f,0.06560f,0.07124f,0.07834f,0.08673f,0.09587f,0.10794f,
 0.02718f,0.01762f,0.01576f,0.01557f,0.01580f,0.01593f,0.01655f,0.01706f,0.01774f,0.01864f,0.01983f,0.02118f,0.02264f,0.02410f,0.02624f,0.02811f,0.03033f,0.03241f,0.03528f,0.03753f,0.04026f,0.04304f,0.04642f,0.04927f,0.05261f,0.05680f,0.06162f,0.06634f,0.07270f,0.08044f,0.08936f,0.09890f,0.11006f,
 0.02582f,0.01754f,0.01562f,0.01565f,0.01583f,0.01603f,0.01644f,0.01711f,0.01773f,0.01872f,0.01975f,0.02100f,0.02253f,0.02440f,0.02628f,0.02829f,0.03043f,0.03271f,0.03571f,0.03779f,0.04112f,0.04391f,0.04640f,0.04999f,0.05378f,0.05765f,0.06214f,0.06816f,0.07458f,0.08227f,0.09123f,0.10128f,0.11268f,
 0.02685f,0.01777f,0.01566f,0.01566f,0.01583f,0.01612f,0.01635f,0.01705f,0.01777f,0.01867f,0.01984f,0.02115f,0.02247f,0.02422f,0.02617f,0.02854f,0.03036f,0.03271f,0.03587f,0.03802f,0.04123f,0.04411f,0.04730f,0.05114f,0.05511f,0.05880f,0.06405f,0.06929f,0.07577f,0.08357f,0.09207f,0.10310f,0.11470f,
 0.02574f,0.01770f,0.01561f,0.01559f,0.01574f,0.01606f,0.01643f,0.01699f,0.01778f,0.01850f,0.01982f,0.02105f,0.02257f,0.02431f,0.02636f,0.02828f,0.03055f,0.03344f,0.03597f,0.03835f,0.04156f,0.04473f,0.04804f,0.05114f,0.05522f,0.05966f,0.06448f,0.07102f,0.07734f,0.08584f,0.09481f,0.10535f,0.11727f,
 0.02597f,0.01757f,0.01567f,0.01561f,0.01586f,0.01603f,0.01646f,0.01713f,0.01783f,0.01859f,0.01992f,0.02108f,0.02267f,0.02435f,0.02618f,0.02840f,0.03080f,0.03317f,0.03612f,0.03867f,0.04167f,0.04462f,0.04863f,0.05181f,0.05636f,0.06051f,0.06628f,0.07172f,0.07963f,0.08698f,0.09744f,0.10720f,0.11847f,
 0.02786f,0.01772f,0.01559f,0.01569f,0.01580f,0.01609f,0.01643f,0.01697f,0.01782f,0.01858f,0.01966f,0.02119f,0.02267f,0.02429f,0.02579f,0.02857f,0.03071f,0.03340f,0.03603f,0.03913f,0.04223f,0.04563f,0.04871f,0.05232f,0.05730f,0.06206f,0.06672f,0.07345f,0.08032f,0.08932f,0.09866f,0.11019f,0.12240f,
 0.02672f,0.01767f,0.01561f,0.01561f,0.01586f,0.01604f,0.01649f,0.01704f,0.01766f,0.01864f,0.01974f,0.02113f,0.02255f,0.02452f,0.02621f,0.02841f,0.03068f,0.03313f,0.03648f,0.03927f,0.04251f,0.04591f,0.04928f,0.05344f,0.05766f,0.06259f,0.06859f,0.07485f,0.08212f,0.09097f,0.10071f,0.11208f,0.12402f,
 0.02635f,0.01762f,0.01570f,0.01560f,0.01577f,0.01605f,0.01644f,0.01706f,0.01783f,0.01868f,0.01970f,0.02107f,0.02262f,0.02419f,0.02628f,0.02860f,0.03109f,0.03371f,0.03646f,0.03977f,0.04274f,0.04622f,0.05000f,0.05436f,0.05857f,0.06371f,0.06961f,0.07644f,0.08379f,0.09176f,0.10185f,0.11355f,0.12643f,
 0.02681f,0.01764f,0.01562f,0.01564f,0.01587f,0.01593f,0.01652f,0.01710f,0.01765f,0.01862f,0.01974f,0.02103f,0.02248f,0.02444f,0.02624f,0.02869f,0.03113f,0.03361f,0.03669f,0.03967f,0.04301f,0.04686f,0.05073f,0.05409f,0.05971f,0.06470f,0.07036f,0.07744f,0.08495f,0.09436f,0.10439f,0.11563f,0.12825f,
 0.02610f,0.01759f,0.01561f,0.01564f,0.01584f,0.01605f,0.01644f,0.01702f,0.01775f,0.01866f,0.01991f,0.02109f,0.02274f,0.02430f,0.02619f,0.02851f,0.03075f,0.03370f,0.03667f,0.04002f,0.04341f,0.04692f,0.05071f,0.05552f,0.06049f,0.06520f,0.07118f,0.07888f,0.08567f,0.09642f,0.10605f,0.11817f,0.13113f,
 0.02604f,0.01761f,0.01567f,0.01562f,0.01593f,0.01606f,0.01645f,0.01696f,0.01767f,0.01853f,0.01963f,0.02097f,0.02260f,0.02442f,0.02632f,0.02880f,0.03107f,0.03381f,0.03696f,0.04033f,0.04361f,0.04783f,0.05111f,0.05610f,0.06116f,0.06624f,0.07324f,0.08015f,0.08786f,0.09788f,0.10814f,0.11954f,0.13365f,
 0.02646f,0.01766f,0.01568f,0.01560f,0.01588f,0.01596f,0.01641f,0.01696f,0.01777f,0.01860f,0.01980f,0.02102f,0.02264f,0.02419f,0.02655f,0.02857f,0.03148f,0.03365f,0.03650f,0.04044f,0.04386f,0.04734f,0.05252f,0.05653f,0.06149f,0.06730f,0.07374f,0.08133f,0.08896f,0.09855f,0.10991f,0.12223f,0.13563f,
 0.02777f,0.01774f,0.01571f,0.01564f,0.01576f,0.01600f,0.01638f,0.01707f,0.01759f,0.01847f,0.01968f,0.02118f,0.02241f,0.02439f,0.02632f,0.02878f,0.03123f,0.03427f,0.03726f,0.04085f,0.04449f,0.04828f,0.05238f,0.05703f,0.06235f,0.06850f,0.07517f,0.08202f,0.09081f,0.10058f,0.11126f,0.12350f,0.13846f,
 0.02701f,0.01768f,0.01564f,0.01562f,0.01582f,0.01615f,0.01645f,0.01712f,0.01758f,0.01853f,0.01962f,0.02105f,0.02265f,0.02433f,0.02633f,0.02880f,0.03127f,0.03426f,0.03714f,0.04071f,0.04413f,0.04842f,0.05269f,0.05794f,0.06288f,0.06925f,0.07548f,0.08397f,0.09222f,0.10185f,0.11382f,0.12620f,0.14012f,
 0.02518f,0.01749f,0.01565f,0.01564f,0.01580f,0.01608f,0.01648f,0.01692f,0.01771f,0.01861f,0.01967f,0.02098f,0.02272f,0.02446f,0.02656f,0.02855f,0.03118f,0.03412f,0.03727f,0.04130f,0.04468f,0.04887f,0.05310f,0.05879f,0.06375f,0.07083f,0.07715f,0.08405f,0.09372f,0.10372f,0.11509f,0.12822f,0.14266f,
 0.02764f,0.01752f,0.01573f,0.01559f,0.01590f,0.01601f,0.01647f,0.01700f,0.01776f,0.01855f,0.01974f,0.02091f,0.02253f,0.02440f,0.02636f,0.02885f,0.03141f,0.03413f,0.03722f,0.04106f,0.04548f,0.04916f,0.05364f,0.05920f,0.06413f,0.07086f,0.07785f,0.08560f,0.09524f,0.10609f,0.11650f,0.13050f,0.14426f,
 0.02744f,0.01748f,0.01561f,0.01565f,0.01576f,0.01612f,0.01647f,0.01693f,0.01763f,0.01857f,0.01968f,0.02092f,0.02238f,0.02447f,0.02611f,0.02904f,0.03131f,0.03441f,0.03770f,0.04153f,0.04531f,0.04938f,0.05478f,0.05969f,0.06491f,0.07206f,0.07946f,0.08727f,0.09602f,0.10654f,0.11844f,0.13146f,0.14721f,
 0.02566f,0.01728f,0.01557f,0.01560f,0.01582f,0.01612f,0.01639f,0.01704f,0.01765f,0.01860f,0.01963f,0.02091f,0.02270f,0.02452f,0.02629f,0.02902f,0.03146f,0.03421f,0.03799f,0.04123f,0.04559f,0.05002f,0.05480f,0.05998f,0.06594f,0.07260f,0.07945f,0.08866f,0.09746f,0.10813f,0.12036f,0.13367f,0.14879f,
};

// 33x33x33 table calculated using moonshine_devtools/compute_albedo for the following ranges:
// F*Reflection + (1-F)*Transmission
// EtaT/EtaI [1/3, 1]
// Roughness [0,1]
// CosTheta [0, 1]
const float
TransmissionAlbedo::mEExit[COMP*COMP*COMP] =
{
 0.00000f,1.00000f,0.99721f,1.00000f,0.99012f,0.96855f,0.99085f,0.99424f,0.98321f,0.96543f,0.94144f,0.94252f,0.95532f,0.94831f,0.93802f,0.95437f,0.95174f,0.95074f,0.94960f,0.94810f,0.96542f,0.96908f,0.96786f,0.96896f,0.97045f,0.96792f,0.97360f,0.98179f,0.98429f,0.98865f,0.99157f,0.99500f,1.00000f,0.00000f,0.99273f,0.98666f,0.98280f,0.98057f,0.97940f,0.97680f,0.97646f,0.97537f,0.97328f,0.96715f,0.96792f,0.97086f,0.96929f,0.96812f,0.97123f,0.97175f,0.97174f,0.97306f,0.97291f,0.97702f,0.97771f,0.97898f,0.98039f,0.98208f,0.98346f,0.98568f,0.98846f,0.99024f,0.99270f,0.99491f,0.99741f,1.00000f,0.00000f,0.98936f,0.98665f,0.98263f,0.97994f,0.97823f,0.97558f,0.97471f,0.97472f,0.97426f,0.97288f,0.97275f,0.97350f,0.97282f,0.97451f,0.97476f,0.97476f,0.97595f,0.97654f,0.97778f,0.97883f,0.97987f,0.98115f,0.98273f,0.98384f,0.98554f,0.98760f,0.98943f,0.99134f,0.99328f,0.99541f,0.99775f,1.00000f,0.00000f,0.98389f,0.98559f,0.98119f,0.98057f,0.97861f,0.97574f,0.97561f,0.97430f,0.97349f,0.97374f,0.97267f,0.97255f,0.97412f,0.97435f,0.97390f,0.97522f,0.97532f,0.97672f,0.97808f,0.97890f,0.98041f,0.98187f,0.98296f,0.98436f,0.98577f,0.98781f,0.98937f,0.99134f,0.99327f,0.99547f,0.99775f,1.00000f,0.00000f,0.93939f,0.98245f,0.98263f,0.97965f,0.97708f,0.97653f,0.97370f,0.97374f,0.97207f,0.97454f,0.97339f,0.97432f,0.97321f,0.97457f,0.97359f,0.97555f,0.97591f,0.97703f,0.97748f,0.97835f,0.98050f,0.98115f,0.98266f,0.98428f,0.98570f,0.98737f,0.98944f,0.99140f,0.99331f,0.99549f,0.99775f,1.00000f,0.00000f,0.90851f,0.95434f,0.97535f,0.97716f,0.97628f,0.97543f,0.97433f,0.97385f,0.96981f,0.97313f,0.97304f,0.97311f,0.97246f,0.97306f,0.97423f,0.97450f,0.97543f,0.97648f,0.97752f,0.97851f,0.97974f,0.98142f,0.98281f,0.98442f,0.98550f,0.98762f,0.98936f,0.99123f,0.99329f,0.99541f,0.99772f,1.00000f,0.00000f,0.90063f,0.92459f,0.95546f,0.97098f,0.97547f,0.97566f,0.97419f,0.96905f,0.96621f,0.97269f,0.97182f,0.97186f,0.97298f,0.97354f,0.97386f,0.97447f,0.97515f,0.97697f,0.97705f,0.97865f,0.97998f,0.98114f,0.98276f,0.98400f,0.98589f,0.98725f,0.98922f,0.99114f,0.99337f,0.99544f,0.99755f,1.00000f,0.00000f,0.89985f,0.90411f,0.92622f,0.95116f,0.96432f,0.97005f,0.96801f,0.96018f,0.96433f,0.97200f,0.97160f,0.97361f,0.97277f,0.97347f,0.97311f,0.97439f,0.97522f,0.97617f,0.97767f,0.97832f,0.97953f,0.98104f,0.98235f,0.98356f,0.98598f,0.98776f,0.98940f,0.99121f,0.99315f,0.99518f,0.99749f,1.00000f,0.00000f,0.90840f,0.89807f,0.90934f,0.92496f,0.94401f,0.95218f,0.94945f,0.94450f,0.96116f,0.97162f,0.97099f,0.97268f,0.97295f,0.97271f,0.97369f,0.97382f,0.97497f,0.97646f,0.97695f,0.97828f,0.97901f,0.98093f,0.98259f,0.98377f,0.98563f,0.98740f,0.98914f,0.99113f,0.99324f,0.99535f,0.99760f,1.00000f,0.00000f,0.92051f,0.89967f,0.89915f,0.90436f,0.92063f,0.91756f,0.91436f,0.92521f,0.95040f,0.96619f,0.96955f,0.97122f,0.97218f,0.97194f,0.97281f,0.97339f,0.97457f,0.97629f,0.97632f,0.97835f,0.97920f,0.98035f,0.98221f,0.98377f,0.98548f,0.98718f,0.98888f,0.99097f,0.99312f,0.99519f,0.99737f,1.00000f,0.00000f,0.92420f,0.89737f,0.89182f,0.89333f,0.88092f,0.87058f,0.87023f,0.89661f,0.93183f,0.95318f,0.96209f,0.96649f,0.97043f,0.96995f,0.97295f,0.97290f,0.97392f,0.97611f,0.97660f,0.97843f,0.97896f,0.98057f,0.98193f,0.98314f,0.98554f,0.98708f,0.98903f,0.99058f,0.99321f,0.99523f,0.99739f,1.00000f,0.00000f,0.93438f,0.90185f,0.88162f,0.86109f,0.83732f,0.82188f,0.83174f,0.87268f,0.90960f,0.93582f,0.94819f,0.95779f,0.96195f,0.96657f,0.97072f,0.97182f,0.97385f,0.97546f,0.97593f,0.97809f,0.97807f,0.98024f,0.98128f,0.98387f,0.98488f,0.98660f,0.98913f,0.99089f,0.99292f,0.99499f,0.99750f,1.00000f,0.00000f,0.93695f,0.89385f,0.86017f,0.82415f,0.78650f,0.77830f,0.80666f,0.84834f,0.88312f,0.91359f,0.93023f,0.94180f,0.94734f,0.95746f,0.96213f,0.96539f,0.97039f,0.97318f,0.97449f,0.97692f,0.97804f,0.97920f,0.98138f,0.98305f,0.98473f,0.98665f,0.98842f,0.99063f,0.99293f,0.99472f,0.99736f,1.00000f,0.00000f,0.90483f,0.86261f,0.82102f,0.76713f,0.73370f,0.74292f,0.77557f,0.82405f,0.86369f,0.89252f,0.91031f,0.92403f,0.93338f,0.94254f,0.95038f,0.95424f,0.96327f,0.96652f,0.97196f,0.97617f,0.97836f,0.97953f,0.98177f,0.98272f,0.98465f,0.98640f,0.98903f,0.99118f,0.99241f,0.99491f,0.99739f,1.00000f,0.00000f,0.87176f,0.81904f,0.75681f,0.71009f,0.68326f,0.70511f,0.75321f,0.80306f,0.84777f,0.87450f,0.89207f,0.90542f,0.91830f,0.92686f,0.93468f,0.94154f,0.95128f,0.96061f,0.96672f,0.97176f,0.97587f,0.97848f,0.98120f,0.98263f,0.98400f,0.98606f,0.98802f,0.99078f,0.99262f,0.99458f,0.99726f,1.00000f,0.00000f,0.82035f,0.75126f,0.69626f,0.64986f,0.64152f,0.68136f,0.73358f,0.78611f,0.82539f,0.85477f,0.87391f,0.89025f,0.89951f,0.91149f,0.91917f,0.92892f,0.93835f,0.94734f,0.95856f,0.96701f,0.97266f,0.97760f,0.97985f,0.98348f,0.98426f,0.98618f,0.98792f,0.99030f,0.99226f,0.99476f,0.99721f,1.00000f,0.00000f,0.74998f,0.68620f,0.62052f,0.59053f,0.60746f,0.65827f,0.71342f,0.76787f,0.80708f,0.83826f,0.86363f,0.87517f,0.88550f,0.89644f,0.90705f,0.91568f,0.92835f,0.93743f,0.94907f,0.95934f,0.96638f,0.97387f,0.97810f,0.98123f,0.98352f,0.98606f,0.98780f,0.99044f,0.99221f,0.99433f,0.99689f,1.00000f,0.00000f,0.67910f,0.61083f,0.56214f,0.54437f,0.57413f,0.63428f,0.69224f,0.74749f,0.79209f,0.82314f,0.84553f,0.86221f,0.87528f,0.88581f,0.89317f,0.90184f,0.91400f,0.92602f,0.94095f,0.95155f,0.95848f,0.96763f,0.97486f,0.97760f,0.98148f,0.98512f,0.98748f,0.99004f,0.99210f,0.99438f,0.99702f,1.00000f,0.00000f,0.59793f,0.54042f,0.49816f,0.50065f,0.54941f,0.61349f,0.67863f,0.72854f,0.77433f,0.80567f,0.83059f,0.84859f,0.86189f,0.87706f,0.88680f,0.89240f,0.90602f,0.91573f,0.92943f,0.94232f,0.95127f,0.95981f,0.97029f,0.97514f,0.98069f,0.98368f,0.98720f,0.98966f,0.99130f,0.99470f,0.99679f,1.00000f,0.00000f,0.52444f,0.47404f,0.44412f,0.46395f,0.52695f,0.59461f,0.65951f,0.71843f,0.75884f,0.79228f,0.81476f,0.83465f,0.85181f,0.86415f,0.87418f,0.88820f,0.89605f,0.91012f,0.92132f,0.93162f,0.94366f,0.95433f,0.96208f,0.97101f,0.97824f,0.98160f,0.98658f,0.98809f,0.99162f,0.99392f,0.99655f,1.00000f,0.00000f,0.46303f,0.41820f,0.40720f,0.43610f,0.50538f,0.57641f,0.64331f,0.69626f,0.73906f,0.77515f,0.80370f,0.82370f,0.84053f,0.85929f,0.86897f,0.88370f,0.88924f,0.89875f,0.91393f,0.92075f,0.93472f,0.94482f,0.95334f,0.96355f,0.97132f,0.97784f,0.98427f,0.98810f,0.99129f,0.99340f,0.99621f,1.00000f,0.00000f,0.40787f,0.36435f,0.36456f,0.41314f,0.48489f,0.55897f,0.62599f,0.68207f,0.72640f,0.76059f,0.79067f,0.81290f,0.82868f,0.84432f,0.85808f,0.87576f,0.88603f,0.89536f,0.90618f,0.91792f,0.92901f,0.93904f,0.94768f,0.95615f,0.96500f,0.97349f,0.98058f,0.98674f,0.99094f,0.99307f,0.99675f,0.99999f,0.00000f,0.35708f,0.32015f,0.33611f,0.39035f,0.46683f,0.53942f,0.60904f,0.66642f,0.70761f,0.74560f,0.77675f,0.80336f,0.81866f,0.83473f,0.85110f,0.86582f,0.87748f,0.89028f,0.90495f,0.91124f,0.92402f,0.92983f,0.94100f,0.95235f,0.95890f,0.96730f,0.97693f,0.98283f,0.98792f,0.99304f,0.99602f,0.99995f,0.00000f,0.30907f,0.28310f,0.30776f,0.37105f,0.44885f,0.52721f,0.59214f,0.64700f,0.69319f,0.72990f,0.75970f,0.79003f,0.81633f,0.82928f,0.84182f,0.86060f,0.87266f,0.88712f,0.90151f,0.90715f,0.91870f,0.92807f,0.93669f,0.94655f,0.95324f,0.96329f,0.97041f,0.98019f,0.98739f,0.99298f,0.99674f,0.99994f,0.00000f,0.27049f,0.25290f,0.28629f,0.35332f,0.43177f,0.50973f,0.57172f,0.63216f,0.67438f,0.71399f,0.74534f,0.77313f,0.79743f,0.81442f,0.83002f,0.85410f,0.86666f,0.87956f,0.89321f,0.90578f,0.91430f,0.92389f,0.93035f,0.93887f,0.94505f,0.95767f,0.96405f,0.97473f,0.98266f,0.98960f,0.99504f,0.99984f,0.00000f,0.23531f,0.22374f,0.26751f,0.33981f,0.41650f,0.48881f,0.55073f,0.61023f,0.66101f,0.69923f,0.73249f,0.76324f,0.78493f,0.80754f,0.82393f,0.84400f,0.86250f,0.87523f,0.88910f,0.90166f,0.91063f,0.92027f,0.92511f,0.93191f,0.94122f,0.94984f,0.95787f,0.96706f,0.97778f,0.98636f,0.99563f,0.99978f,0.00000f,0.20425f,0.20234f,0.25254f,0.32249f,0.40051f,0.47341f,0.54114f,0.59232f,0.64298f,0.68199f,0.71851f,0.74949f,0.77350f,0.79814f,0.81355f,0.83522f,0.85060f,0.86837f,0.88148f,0.89502f,0.90644f,0.91697f,0.92029f,0.92994f,0.93338f,0.94476f,0.94998f,0.96400f,0.97194f,0.98350f,0.99360f,0.99965f,0.00000f,0.18125f,0.18459f,0.23477f,0.31091f,0.38646f,0.45695f,0.51826f,0.57845f,0.62222f,0.66506f,0.70329f,0.73248f,0.75643f,0.78277f,0.80895f,0.82336f,0.84782f,0.86176f,0.87506f,0.89072f,0.90078f,0.90976f,0.91695f,0.92592f,0.93066f,0.93711f,0.94794f,0.95439f,0.96790f,0.97940f,0.99086f,0.99961f,0.00000f,0.15678f,0.16769f,0.22657f,0.29862f,0.37206f,0.44311f,0.50682f,0.55678f,0.60902f,0.64943f,0.69011f,0.71730f,0.74583f,0.77759f,0.79507f,0.81699f,0.83873f,0.85374f,0.86643f,0.88023f,0.89506f,0.90520f,0.91627f,0.92177f,0.92585f,0.93376f,0.94005f,0.94952f,0.96133f,0.97308f,0.98793f,0.99945f,0.00000f,0.13964f,0.15668f,0.21222f,0.28509f,0.36112f,0.42849f,0.48746f,0.54572f,0.59338f,0.63142f,0.66774f,0.70556f,0.73326f,0.76056f,0.78560f,0.80662f,0.82810f,0.84736f,0.85935f,0.87483f,0.88525f,0.90185f,0.91394f,0.91771f,0.92483f,0.92891f,0.93341f,0.94259f,0.95340f,0.96830f,0.98396f,0.99912f,0.00000f,0.12562f,0.14410f,0.20246f,0.27559f,0.34598f,0.41175f,0.47172f,0.52479f,0.57568f,0.61718f,0.65583f,0.69231f,0.71668f,0.75099f,0.77507f,0.79322f,0.81672f,0.83815f,0.85557f,0.86858f,0.87929f,0.89211f,0.90491f,0.91411f,0.92191f,0.92318f,0.92835f,0.93514f,0.94980f,0.95893f,0.97730f,0.99896f,0.00000f,0.11145f,0.13603f,0.19570f,0.26411f,0.33372f,0.39769f,0.45833f,0.51340f,0.55666f,0.60168f,0.63880f,0.67693f,0.71109f,0.73287f,0.76184f,0.78529f,0.80710f,0.83089f,0.84556f,0.86619f,0.87735f,0.88942f,0.90059f,0.90909f,0.91380f,0.92676f,0.92628f,0.92875f,0.94043f,0.95472f,0.97095f,0.99858f,0.00000f,0.09900f,0.12618f,0.18694f,0.25400f,0.32295f,0.38381f,0.44254f,0.49642f,0.54206f,0.58426f,0.62582f,0.66146f,0.69845f,0.72611f,0.75062f,0.77508f,0.79935f,0.81961f,0.83731f,0.85256f,0.86757f,0.88610f,0.89653f,0.90553f,0.91331f,0.91959f,0.92122f,0.92524f,0.93372f,0.94595f,0.96410f,0.99839f,
 0.00000f,1.00000f,0.99716f,1.00000f,0.99159f,0.96913f,0.99065f,0.99382f,0.98450f,0.96662f,0.96551f,0.99872f,0.98107f,0.98843f,0.98566f,0.96350f,0.95436f,0.94660f,0.96015f,0.95883f,0.96088f,0.95741f,0.97265f,0.97353f,0.97674f,0.97623f,0.97863f,0.97727f,0.98204f,0.98636f,0.99385f,0.99706f,1.00000f,0.00000f,0.99231f,0.98601f,0.98281f,0.98058f,0.97845f,0.97625f,0.97792f,0.97586f,0.97402f,0.97299f,0.97463f,0.97406f,0.97487f,0.97309f,0.97249f,0.97097f,0.97054f,0.97301f,0.97457f,0.97533f,0.97612f,0.97992f,0.98112f,0.98297f,0.98429f,0.98607f,0.98769f,0.98999f,0.99222f,0.99515f,0.99757f,1.00000f,0.00000f,0.99278f,0.98590f,0.98204f,0.97949f,0.97552f,0.97692f,0.97522f,0.97421f,0.97336f,0.97399f,0.97315f,0.97376f,0.97344f,0.97404f,0.97483f,0.97461f,0.97581f,0.97682f,0.97782f,0.97839f,0.97981f,0.98110f,0.98292f,0.98421f,0.98589f,0.98751f,0.98948f,0.99136f,0.99335f,0.99544f,0.99778f,1.00000f,0.00000f,0.98229f,0.98694f,0.98238f,0.97943f,0.97877f,0.97651f,0.97538f,0.97391f,0.97417f,0.97338f,0.97413f,0.97304f,0.97341f,0.97321f,0.97396f,0.97490f,0.97576f,0.97689f,0.97753f,0.97873f,0.97986f,0.98157f,0.98238f,0.98454f,0.98603f,0.98775f,0.98951f,0.99128f,0.99342f,0.99550f,0.99775f,1.00000f,0.00000f,0.93665f,0.98173f,0.98268f,0.98030f,0.97644f,0.97599f,0.97457f,0.97384f,0.97377f,0.97305f,0.97362f,0.97397f,0.97319f,0.97400f,0.97494f,0.97470f,0.97538f,0.97637f,0.97781f,0.97883f,0.98026f,0.98070f,0.98275f,0.98420f,0.98586f,0.98754f,0.98914f,0.99133f,0.99337f,0.99552f,0.99776f,1.00000f,0.00000f,0.91027f,0.95972f,0.97914f,0.97790f,0.97773f,0.97553f,0.97513f,0.97389f,0.97434f,0.97383f,0.97311f,0.97302f,0.97036f,0.97329f,0.97435f,0.97530f,0.97573f,0.97645f,0.97685f,0.97828f,0.97940f,0.98137f,0.98281f,0.98390f,0.98577f,0.98760f,0.98937f,0.99118f,0.99327f,0.99540f,0.99776f,1.00000f,0.00000f,0.89557f,0.92298f,0.95440f,0.97082f,0.97606f,0.97532f,0.97450f,0.97347f,0.97153f,0.97290f,0.97183f,0.97210f,0.96894f,0.97325f,0.97320f,0.97397f,0.97548f,0.97607f,0.97719f,0.97876f,0.97966f,0.98134f,0.98237f,0.98395f,0.98551f,0.98732f,0.98937f,0.99112f,0.99348f,0.99537f,0.99762f,1.00000f,0.00000f,0.90700f,0.90384f,0.92776f,0.95027f,0.96640f,0.96955f,0.97159f,0.97182f,0.97272f,0.97269f,0.97175f,0.96710f,0.96704f,0.97233f,0.97344f,0.97443f,0.97523f,0.97594f,0.97695f,0.97796f,0.97940f,0.98089f,0.98231f,0.98383f,0.98559f,0.98732f,0.98928f,0.99109f,0.99303f,0.99535f,0.99756f,1.00000f,0.00000f,0.90794f,0.89358f,0.90498f,0.92648f,0.94339f,0.95696f,0.96490f,0.97024f,0.97036f,0.97067f,0.96572f,0.96111f,0.96535f,0.97131f,0.97318f,0.97435f,0.97477f,0.97580f,0.97660f,0.97792f,0.97942f,0.98077f,0.98210f,0.98353f,0.98546f,0.98713f,0.98930f,0.99107f,0.99304f,0.99530f,0.99758f,1.00000f,0.00000f,0.91467f,0.89784f,0.89665f,0.90508f,0.91926f,0.93536f,0.95084f,0.95883f,0.96301f,0.96046f,0.95315f,0.95335f,0.96331f,0.96948f,0.97206f,0.97416f,0.97409f,0.97588f,0.97651f,0.97793f,0.97901f,0.98090f,0.98214f,0.98368f,0.98531f,0.98723f,0.98859f,0.99095f,0.99298f,0.99531f,0.99759f,1.00000f,0.00000f,0.92416f,0.90226f,0.89158f,0.89547f,0.90509f,0.91555f,0.92829f,0.93855f,0.94290f,0.93641f,0.93653f,0.94468f,0.95940f,0.96754f,0.97208f,0.97271f,0.97395f,0.97570f,0.97592f,0.97738f,0.97892f,0.98037f,0.98165f,0.98335f,0.98495f,0.98691f,0.98872f,0.99082f,0.99306f,0.99531f,0.99745f,1.00000f,0.00000f,0.92580f,0.90602f,0.89470f,0.89080f,0.89580f,0.90235f,0.90688f,0.91044f,0.90661f,0.90694f,0.91379f,0.93020f,0.95123f,0.96222f,0.96758f,0.97150f,0.97471f,0.97628f,0.97601f,0.97745f,0.97899f,0.97925f,0.98178f,0.98339f,0.98512f,0.98703f,0.98880f,0.99063f,0.99286f,0.99499f,0.99730f,1.00000f,0.00000f,0.93323f,0.91073f,0.89973f,0.89138f,0.88528f,0.88806f,0.88176f,0.87669f,0.86848f,0.87161f,0.88692f,0.90981f,0.93280f,0.95004f,0.95992f,0.96549f,0.96910f,0.97180f,0.97533f,0.97555f,0.97829f,0.98016f,0.98164f,0.98312f,0.98448f,0.98666f,0.98906f,0.99046f,0.99252f,0.99514f,0.99745f,1.00000f,0.00000f,0.94089f,0.91167f,0.90272f,0.88864f,0.88134f,0.86688f,0.85244f,0.83842f,0.83000f,0.83738f,0.86238f,0.89018f,0.91490f,0.93236f,0.94716f,0.95460f,0.95974f,0.96512f,0.96859f,0.97176f,0.97557f,0.97818f,0.98084f,0.98199f,0.98518f,0.98705f,0.98836f,0.99070f,0.99258f,0.99500f,0.99723f,1.00000f,0.00000f,0.93983f,0.91875f,0.90330f,0.87982f,0.85980f,0.83577f,0.81151f,0.79719f,0.79558f,0.81415f,0.83692f,0.86795f,0.89511f,0.91425f,0.92777f,0.93802f,0.94777f,0.95374f,0.95975f,0.96505f,0.96859f,0.97484f,0.97787f,0.98155f,0.98450f,0.98688f,0.98839f,0.99030f,0.99237f,0.99449f,0.99728f,1.00000f,0.00000f,0.93620f,0.91099f,0.88735f,0.85727f,0.82800f,0.80152f,0.76847f,0.75855f,0.76856f,0.78721f,0.81717f,0.84706f,0.87685f,0.89257f,0.91029f,0.92272f,0.93153f,0.93915f,0.94735f,0.95396f,0.95980f,0.96499f,0.97136f,0.97544f,0.98133f,0.98481f,0.98776f,0.99037f,0.99206f,0.99479f,0.99693f,1.00000f,0.00000f,0.92343f,0.89879f,0.86219f,0.82691f,0.78464f,0.75643f,0.73140f,0.72633f,0.74122f,0.76878f,0.80319f,0.83158f,0.85548f,0.87490f,0.89212f,0.90724f,0.91568f,0.92244f,0.93142f,0.93852f,0.94675f,0.95385f,0.96182f,0.96873f,0.97586f,0.98152f,0.98682f,0.99013f,0.99248f,0.99446f,0.99679f,1.00000f,0.00000f,0.89456f,0.85849f,0.82227f,0.77563f,0.74164f,0.70315f,0.68988f,0.69439f,0.72127f,0.74871f,0.78653f,0.81565f,0.84071f,0.85882f,0.87492f,0.88719f,0.89940f,0.90550f,0.91240f,0.92465f,0.93148f,0.93921f,0.95002f,0.96110f,0.96858f,0.97767f,0.98358f,0.98846f,0.99167f,0.99407f,0.99717f,1.00000f,0.00000f,0.85824f,0.81910f,0.77609f,0.73180f,0.68874f,0.65305f,0.65610f,0.67231f,0.69968f,0.73676f,0.76980f,0.79815f,0.82682f,0.84411f,0.85619f,0.87197f,0.88077f,0.88844f,0.89665f,0.90431f,0.91742f,0.92412f,0.93555f,0.94437f,0.95763f,0.97008f,0.97974f,0.98789f,0.99105f,0.99453f,0.99683f,1.00000f,0.00000f,0.80745f,0.76724f,0.72414f,0.68013f,0.63524f,0.61554f,0.62174f,0.64924f,0.67961f,0.72044f,0.75525f,0.78157f,0.81168f,0.82903f,0.84295f,0.85848f,0.86754f,0.87539f,0.88261f,0.88986f,0.89908f,0.90759f,0.92044f,0.93217f,0.94597f,0.96097f,0.97133f,0.98091f,0.98875f,0.99360f,0.99696f,1.00000f,0.00000f,0.75605f,0.71250f,0.66886f,0.61951f,0.58965f,0.57875f,0.59672f,0.62785f,0.66610f,0.70463f,0.73940f,0.77213f,0.79428f,0.81714f,0.82891f,0.84220f,0.85303f,0.86143f,0.86897f,0.87438f,0.88579f,0.89405f,0.90685f,0.91788f,0.93410f,0.94898f,0.96587f,0.97429f,0.98445f,0.99262f,0.99578f,0.99998f,0.00000f,0.69523f,0.65449f,0.60733f,0.56846f,0.54075f,0.54815f,0.57407f,0.60758f,0.65091f,0.69089f,0.72792f,0.75317f,0.78358f,0.80456f,0.81458f,0.83379f,0.84257f,0.85000f,0.85563f,0.86007f,0.87220f,0.88054f,0.89482f,0.90637f,0.92023f,0.93690f,0.95463f,0.96659f,0.97909f,0.98883f,0.99547f,0.99995f,0.00000f,0.63840f,0.59832f,0.55407f,0.51930f,0.50573f,0.51560f,0.54683f,0.59302f,0.63803f,0.67346f,0.71555f,0.74682f,0.77153f,0.78957f,0.80286f,0.81608f,0.83039f,0.83548f,0.84542f,0.85327f,0.86295f,0.86635f,0.87880f,0.89329f,0.90678f,0.92454f,0.94152f,0.95588f,0.97155f,0.98396f,0.99416f,0.99986f,0.00000f,0.58202f,0.54165f,0.50329f,0.47218f,0.47190f,0.49479f,0.53241f,0.57340f,0.61798f,0.66040f,0.69817f,0.72812f,0.75504f,0.77901f,0.79157f,0.80641f,0.82057f,0.83003f,0.83771f,0.84555f,0.85281f,0.86076f,0.87084f,0.88235f,0.89729f,0.91158f,0.92673f,0.94501f,0.96109f,0.97806f,0.99098f,0.99957f,0.00000f,0.53185f,0.48773f,0.45332f,0.43368f,0.43823f,0.46945f,0.51325f,0.56091f,0.60326f,0.64911f,0.68725f,0.71480f,0.74139f,0.75959f,0.77857f,0.79237f,0.80780f,0.81954f,0.82888f,0.83820f,0.84736f,0.85095f,0.86086f,0.87442f,0.88742f,0.90314f,0.91723f,0.93472f,0.95042f,0.96923f,0.98540f,0.99933f,0.00000f,0.48132f,0.44558f,0.40922f,0.39902f,0.41362f,0.44928f,0.49625f,0.54658f,0.59176f,0.63369f,0.67054f,0.69804f,0.72734f,0.75336f,0.77079f,0.78513f,0.79793f,0.81104f,0.82253f,0.82591f,0.83699f,0.84547f,0.85709f,0.86592f,0.87932f,0.89094f,0.90245f,0.92034f,0.93923f,0.95866f,0.97817f,0.99890f,0.00000f,0.43644f,0.40047f,0.37007f,0.36860f,0.39430f,0.43084f,0.48125f,0.53009f,0.57794f,0.61215f,0.65663f,0.68756f,0.72130f,0.73850f,0.75738f,0.77435f,0.78953f,0.80797f,0.81537f,0.82092f,0.83265f,0.83916f,0.85050f,0.85681f,0.87188f,0.88046f,0.89434f,0.90888f,0.92632f,0.94687f,0.96921f,0.99835f,0.00000f,0.39348f,0.36099f,0.34007f,0.34198f,0.36898f,0.41113f,0.46252f,0.51390f,0.56188f,0.60417f,0.63920f,0.67378f,0.69847f,0.72237f,0.74383f,0.76232f,0.77476f,0.79338f,0.80334f,0.81664f,0.82272f,0.83159f,0.84202f,0.85555f,0.85757f,0.86808f,0.88620f,0.89726f,0.91235f,0.93507f,0.95978f,0.99715f,0.00000f,0.35601f,0.32296f,0.30995f,0.32121f,0.35386f,0.39978f,0.44869f,0.49821f,0.54680f,0.58796f,0.62265f,0.65826f,0.68748f,0.71490f,0.73502f,0.74960f,0.77175f,0.78403f,0.79647f,0.81013f,0.81909f,0.82841f,0.83572f,0.84744f,0.85601f,0.86539f,0.87413f,0.88770f,0.89870f,0.91942f,0.94731f,0.99624f,0.00000f,0.31610f,0.29420f,0.27992f,0.29769f,0.33500f,0.38269f,0.43811f,0.48092f,0.52754f,0.57117f,0.61152f,0.63993f,0.67247f,0.70014f,0.71584f,0.74245f,0.75752f,0.77276f,0.79013f,0.80036f,0.81332f,0.82397f,0.83284f,0.84212f,0.84905f,0.85962f,0.86422f,0.87645f,0.88920f,0.90641f,0.93348f,0.99469f,0.00000f,0.28964f,0.26389f,0.26019f,0.28005f,0.32048f,0.37101f,0.42053f,0.47272f,0.51634f,0.56271f,0.59773f,0.62723f,0.65521f,0.68142f,0.70410f,0.72565f,0.74858f,0.76757f,0.77974f,0.79467f,0.80570f,0.81434f,0.82885f,0.83750f,0.84436f,0.84929f,0.85976f,0.86632f,0.87970f,0.89321f,0.91947f,0.99219f,0.00000f,0.26169f,0.24026f,0.23940f,0.26479f,0.30756f,0.35929f,0.40533f,0.45717f,0.49899f,0.54209f,0.58035f,0.61453f,0.64491f,0.67051f,0.69625f,0.71562f,0.73660f,0.75314f,0.77251f,0.78513f,0.80202f,0.81452f,0.82112f,0.83379f,0.84037f,0.84752f,0.85443f,0.85574f,0.86553f,0.87766f,0.90106f,0.98832f,0.00000f,0.23754f,0.21710f,0.22145f,0.25011f,0.29531f,0.34295f,0.39397f,0.44174f,0.48966f,0.52892f,0.56411f,0.60049f,0.63326f,0.65766f,0.68822f,0.70918f,0.72651f,0.74944f,0.76314f,0.77622f,0.79078f,0.80927f,0.81167f,0.82459f,0.83685f,0.83910f,0.84855f,0.84801f,0.85348f,0.86263f,0.88615f,0.97851f,
 0.00000f,1.00000f,0.99608f,1.00000f,0.99045f,0.96892f,0.99183f,0.99428f,0.98414f,0.96616f,0.96490f,0.99716f,0.98252f,0.98876f,0.98573f,0.98527f,0.98259f,0.95888f,0.95928f,0.95239f,0.96778f,0.96807f,0.96661f,0.96815f,0.96607f,0.97485f,0.97849f,0.98255f,0.98478f,0.98714f,0.99242f,0.99345f,1.00000f,0.00000f,0.99127f,0.98731f,0.98371f,0.98064f,0.97782f,0.97704f,0.97675f,0.97569f,0.97376f,0.97390f,0.97420f,0.97308f,0.97474f,0.97431f,0.97498f,0.97602f,0.97282f,0.97292f,0.97321f,0.97649f,0.97804f,0.97879f,0.97964f,0.98118f,0.98433f,0.98616f,0.98808f,0.99038f,0.99234f,0.99487f,0.99721f,1.00000f,0.00000f,0.99017f,0.98552f,0.98467f,0.98051f,0.97758f,0.97601f,0.97518f,0.97461f,0.97433f,0.97255f,0.97328f,0.97347f,0.97432f,0.97405f,0.97421f,0.97575f,0.97608f,0.97696f,0.97745f,0.97879f,0.97996f,0.98125f,0.98255f,0.98414f,0.98562f,0.98771f,0.98942f,0.99128f,0.99318f,0.99528f,0.99774f,1.00000f,0.00000f,0.98237f,0.98516f,0.98322f,0.97930f,0.97894f,0.97673f,0.97426f,0.97416f,0.97514f,0.97403f,0.97318f,0.97347f,0.97363f,0.97359f,0.97430f,0.97486f,0.97607f,0.97648f,0.97764f,0.97899f,0.98032f,0.98161f,0.98254f,0.98405f,0.98595f,0.98738f,0.98958f,0.99137f,0.99343f,0.99549f,0.99775f,1.00000f,0.00000f,0.94034f,0.98298f,0.98119f,0.98037f,0.97808f,0.97666f,0.97473f,0.97417f,0.97374f,0.97231f,0.97400f,0.97379f,0.97395f,0.97395f,0.97489f,0.97434f,0.97568f,0.97732f,0.97798f,0.97876f,0.97990f,0.98109f,0.98266f,0.98441f,0.98599f,0.98736f,0.98931f,0.99133f,0.99336f,0.99546f,0.99777f,1.00000f,0.00000f,0.90425f,0.95737f,0.97754f,0.97768f,0.97748f,0.97641f,0.97474f,0.97408f,0.97409f,0.97279f,0.97310f,0.97249f,0.97375f,0.97364f,0.97325f,0.97329f,0.97536f,0.97611f,0.97750f,0.97904f,0.98006f,0.98144f,0.98250f,0.98391f,0.98570f,0.98770f,0.98937f,0.99117f,0.99321f,0.99540f,0.99775f,1.00000f,0.00000f,0.89677f,0.92173f,0.95770f,0.96989f,0.97421f,0.97562f,0.97457f,0.97398f,0.97370f,0.97308f,0.97259f,0.97224f,0.97304f,0.97248f,0.97288f,0.97234f,0.97497f,0.97658f,0.97758f,0.97874f,0.98011f,0.98123f,0.98247f,0.98394f,0.98557f,0.98743f,0.98911f,0.99123f,0.99324f,0.99545f,0.99758f,1.00000f,0.00000f,0.90374f,0.90550f,0.92794f,0.94984f,0.96577f,0.97040f,0.97196f,0.97258f,0.97244f,0.97166f,0.97223f,0.97189f,0.97344f,0.97201f,0.96933f,0.97099f,0.97509f,0.97596f,0.97732f,0.97831f,0.98006f,0.98063f,0.98283f,0.98355f,0.98553f,0.98721f,0.98945f,0.99130f,0.99321f,0.99531f,0.99751f,1.00000f,0.00000f,0.90984f,0.89526f,0.90744f,0.92479f,0.94208f,0.95520f,0.96593f,0.96870f,0.97064f,0.97121f,0.97276f,0.97200f,0.97097f,0.96821f,0.96663f,0.97129f,0.97468f,0.97629f,0.97682f,0.97775f,0.97978f,0.98014f,0.98247f,0.98416f,0.98567f,0.98713f,0.98949f,0.99119f,0.99316f,0.99534f,0.99761f,1.00000f,0.00000f,0.92057f,0.89766f,0.89579f,0.90678f,0.91949f,0.93769f,0.95015f,0.95707f,0.96515f,0.96850f,0.96870f,0.97018f,0.96608f,0.96031f,0.96257f,0.96964f,0.97380f,0.97534f,0.97708f,0.97775f,0.97899f,0.98030f,0.98236f,0.98346f,0.98546f,0.98707f,0.98901f,0.99107f,0.99295f,0.99513f,0.99758f,1.00000f,0.00000f,0.92718f,0.90388f,0.89106f,0.89652f,0.90396f,0.91426f,0.92626f,0.93966f,0.94964f,0.95736f,0.96122f,0.95784f,0.95458f,0.95435f,0.95947f,0.96821f,0.97324f,0.97551f,0.97659f,0.97774f,0.97922f,0.98019f,0.98201f,0.98352f,0.98508f,0.98728f,0.98878f,0.99105f,0.99303f,0.99525f,0.99751f,1.00000f,0.00000f,0.93030f,0.90326f,0.89677f,0.88941f,0.89475f,0.90306f,0.91288f,0.92249f,0.93385f,0.93988f,0.94033f,0.93929f,0.93772f,0.94056f,0.95331f,0.96465f,0.97053f,0.97534f,0.97613f,0.97674f,0.97828f,0.98111f,0.98119f,0.98342f,0.98490f,0.98701f,0.98885f,0.99101f,0.99279f,0.99493f,0.99755f,1.00000f,0.00000f,0.93787f,0.91068f,0.90006f,0.88954f,0.88930f,0.89680f,0.90293f,0.90439f,0.90861f,0.91115f,0.91076f,0.90925f,0.91479f,0.92702f,0.94403f,0.95639f,0.96634f,0.97202f,0.97411f,0.97586f,0.97798f,0.98006f,0.98209f,0.98345f,0.98548f,0.98695f,0.98892f,0.99047f,0.99262f,0.99512f,0.99741f,1.00000f,0.00000f,0.93859f,0.91464f,0.89780f,0.89518f,0.89035f,0.88870f,0.88786f,0.88695f,0.88414f,0.88151f,0.87576f,0.87891f,0.89005f,0.91001f,0.92745f,0.94606f,0.95606f,0.96388f,0.96865f,0.97157f,0.97499f,0.97872f,0.98057f,0.98299f,0.98437f,0.98640f,0.98862f,0.99053f,0.99245f,0.99501f,0.99740f,1.00000f,0.00000f,0.94530f,0.92177f,0.90143f,0.89705f,0.88465f,0.88244f,0.87227f,0.86341f,0.85722f,0.84266f,0.84143f,0.84852f,0.86412f,0.88733f,0.90991f,0.92872f,0.94302f,0.95320f,0.95922f,0.96481f,0.97235f,0.97340f,0.97718f,0.98052f,0.98299f,0.98616f,0.98812f,0.98981f,0.99238f,0.99475f,0.99721f,1.00000f,0.00000f,0.94863f,0.93119f,0.90654f,0.89451f,0.87778f,0.86736f,0.85537f,0.83674f,0.81739f,0.80845f,0.80926f,0.82070f,0.84555f,0.86738f,0.89053f,0.91013f,0.92368f,0.93511f,0.94488f,0.95173f,0.95923f,0.96670f,0.96957f,0.97577f,0.97886f,0.98375f,0.98707f,0.98995f,0.99269f,0.99492f,0.99728f,1.00000f,0.00000f,0.94604f,0.92883f,0.90179f,0.88586f,0.87015f,0.84891f,0.82363f,0.80366f,0.78511f,0.78154f,0.78254f,0.79789f,0.82122f,0.84576f,0.87169f,0.89071f,0.90529f,0.91802f,0.92829f,0.93615f,0.94577f,0.95256f,0.95890f,0.96530f,0.97169f,0.97771f,0.98294f,0.98812f,0.99235f,0.99444f,0.99711f,1.00000f,0.00000f,0.93691f,0.91536f,0.89812f,0.87267f,0.84847f,0.81724f,0.78901f,0.76681f,0.75063f,0.75278f,0.76035f,0.77814f,0.80599f,0.83144f,0.84849f,0.86879f,0.88718f,0.89949f,0.90891f,0.92109f,0.92680f,0.93587f,0.94288f,0.95141f,0.95961f,0.96929f,0.97560f,0.98347f,0.99014f,0.99388f,0.99663f,1.00000f,0.00000f,0.93081f,0.90605f,0.87811f,0.84400f,0.81519f,0.77930f,0.74793f,0.73005f,0.71623f,0.72674f,0.74131f,0.76458f,0.78823f,0.81266f,0.83451f,0.85349f,0.86885f,0.88066f,0.89384f,0.90150f,0.91128f,0.92020f,0.92489f,0.93613f,0.94451f,0.95502f,0.96428f,0.97563f,0.98495f,0.99286f,0.99683f,1.00000f,0.00000f,0.90381f,0.87959f,0.84382f,0.81387f,0.77243f,0.73704f,0.70590f,0.69029f,0.68940f,0.70110f,0.72766f,0.75128f,0.77257f,0.80371f,0.82172f,0.83478f,0.85291f,0.86442f,0.87226f,0.88288f,0.89160f,0.90055f,0.90953f,0.91749f,0.92881f,0.93841f,0.95278f,0.96481f,0.97899f,0.98970f,0.99569f,1.00000f,0.00000f,0.87421f,0.84603f,0.80583f,0.77282f,0.72880f,0.68990f,0.66578f,0.65542f,0.66236f,0.67832f,0.70800f,0.73436f,0.75974f,0.78666f,0.80985f,0.82257f,0.83979f,0.84879f,0.85874f,0.86903f,0.87442f,0.88328f,0.88895f,0.89953f,0.91226f,0.92289f,0.93438f,0.95207f,0.96858f,0.98302f,0.99380f,0.99997f,0.00000f,0.82712f,0.79483f,0.76267f,0.72283f,0.68406f,0.64625f,0.62628f,0.62314f,0.64153f,0.66252f,0.69087f,0.72269f,0.74746f,0.77357f,0.79234f,0.80906f,0.82396f,0.83519f,0.84472f,0.85168f,0.85427f,0.86449f,0.87078f,0.88305f,0.89464f,0.90360f,0.91949f,0.93602f,0.95622f,0.97365f,0.98999f,0.99985f,0.00000f,0.78646f,0.75335f,0.71066f,0.67669f,0.63386f,0.61091f,0.59572f,0.59757f,0.61866f,0.65159f,0.67628f,0.70912f,0.73943f,0.76209f,0.78140f,0.79743f,0.80882f,0.82079f,0.82756f,0.83684f,0.84181f,0.85020f,0.86058f,0.86699f,0.87543f,0.88538f,0.90226f,0.91881f,0.94168f,0.96239f,0.98377f,0.99957f,0.00000f,0.74148f,0.70865f,0.66387f,0.62548f,0.58788f,0.56853f,0.56561f,0.57744f,0.59816f,0.63457f,0.66359f,0.69418f,0.72586f,0.74929f,0.76902f,0.78254f,0.79110f,0.80676f,0.81723f,0.82506f,0.82852f,0.83600f,0.84140f,0.85231f,0.85959f,0.87076f,0.88772f,0.90378f,0.92534f,0.94934f,0.97476f,0.99918f,0.00000f,0.69071f,0.65255f,0.61331f,0.57295f,0.54846f,0.52938f,0.53809f,0.55967f,0.58944f,0.61038f,0.65174f,0.68493f,0.71119f,0.73192f,0.75395f,0.77035f,0.78008f,0.79451f,0.80473f,0.80914f,0.81634f,0.82596f,0.82775f,0.83573f,0.84249f,0.85406f,0.86889f,0.88807f,0.91124f,0.93416f,0.96211f,0.99816f,0.00000f,0.64301f,0.60718f,0.57058f,0.52811f,0.51035f,0.50206f,0.51871f,0.53437f,0.56809f,0.60150f,0.63564f,0.67212f,0.69362f,0.72016f,0.73887f,0.75733f,0.77175f,0.78457f,0.79221f,0.79808f,0.80633f,0.81188f,0.81974f,0.82306f,0.83075f,0.84005f,0.85352f,0.87040f,0.88956f,0.91548f,0.94814f,0.99661f,0.00000f,0.59711f,0.55979f,0.51861f,0.48478f,0.46849f,0.47096f,0.48888f,0.51793f,0.55109f,0.59155f,0.62624f,0.65234f,0.67857f,0.70687f,0.72355f,0.74118f,0.75671f,0.77440f,0.78293f,0.78815f,0.79763f,0.80303f,0.80620f,0.81299f,0.81668f,0.82951f,0.84027f,0.85635f,0.87718f,0.89673f,0.93077f,0.99472f,0.00000f,0.54444f,0.51729f,0.47591f,0.45092f,0.43745f,0.44578f,0.47253f,0.50139f,0.53565f,0.57236f,0.60574f,0.64100f,0.66834f,0.69189f,0.71209f,0.73549f,0.74383f,0.75962f,0.77142f,0.78167f,0.78815f,0.78987f,0.80091f,0.80700f,0.80924f,0.81810f,0.82739f,0.83768f,0.85828f,0.88300f,0.90907f,0.99166f,0.00000f,0.49840f,0.46957f,0.43685f,0.41631f,0.40891f,0.42365f,0.45244f,0.48297f,0.52558f,0.55973f,0.59545f,0.62525f,0.65948f,0.68048f,0.69845f,0.71714f,0.73575f,0.74507f,0.75828f,0.76654f,0.77827f,0.78307f,0.79421f,0.79784f,0.79985f,0.80935f,0.81526f,0.82632f,0.84097f,0.86034f,0.89208f,0.98491f,0.00000f,0.46085f,0.43242f,0.40512f,0.38386f,0.38065f,0.40338f,0.43372f,0.46704f,0.50393f,0.54589f,0.58288f,0.61229f,0.64371f,0.66961f,0.68601f,0.70874f,0.72340f,0.73813f,0.74894f,0.76417f,0.76680f,0.77937f,0.78248f,0.78798f,0.79550f,0.79861f,0.80086f,0.81179f,0.82778f,0.84327f,0.87151f,0.97006f,0.00000f,0.42450f,0.39545f,0.36686f,0.35377f,0.36026f,0.38517f,0.41828f,0.45498f,0.49271f,0.53164f,0.56675f,0.59687f,0.62466f,0.65432f,0.67279f,0.69214f,0.71026f,0.72642f,0.73725f,0.74939f,0.76305f,0.76921f,0.77652f,0.78306f,0.78830f,0.79133f,0.79458f,0.80072f,0.81243f,0.82528f,0.84680f,0.95400f,0.00000f,0.39155f,0.36306f,0.33799f,0.32926f,0.33782f,0.36509f,0.39996f,0.44186f,0.48161f,0.51850f,0.55288f,0.58277f,0.61438f,0.63433f,0.66198f,0.68202f,0.70392f,0.71598f,0.73052f,0.74175f,0.75418f,0.75850f,0.77072f,0.77503f,0.77824f,0.78177f,0.78791f,0.78992f,0.79725f,0.80396f,0.82281f,0.93134f,0.00000f,0.35568f,0.33066f,0.31108f,0.30815f,0.32008f,0.34786f,0.38671f,0.42738f,0.46669f,0.49957f,0.53842f,0.56923f,0.59555f,0.62371f,0.65009f,0.66830f,0.68730f,0.70365f,0.72116f,0.73083f,0.74673f,0.75290f,0.76241f,0.77087f,0.77071f,0.77753f,0.77693f,0.77934f,0.78182f,0.79117f,0.79793f,0.90339f,
 0.00000f,1.00000f,0.99814f,1.00000f,0.99155f,0.96887f,0.99030f,0.99308f,0.98444f,0.96707f,0.96490f,0.99820f,0.98173f,0.98868f,0.98462f,0.98421f,0.98177f,0.97244f,0.97136f,0.97007f,0.95703f,0.96713f,0.96589f,0.96453f,0.96810f,0.97881f,0.97960f,0.98215f,0.98458f,0.98900f,0.99150f,0.99635f,1.00000f,0.00000f,0.99130f,0.98679f,0.98259f,0.97992f,0.97742f,0.97761f,0.97649f,0.97521f,0.97326f,0.97336f,0.97458f,0.97331f,0.97516f,0.97517f,0.97541f,0.97566f,0.97549f,0.97652f,0.97696f,0.97546f,0.97792f,0.97857f,0.97975f,0.98172f,0.98501f,0.98616f,0.98833f,0.99031f,0.99261f,0.99471f,0.99750f,1.00000f,0.00000f,0.99179f,0.98729f,0.98276f,0.98018f,0.97802f,0.97714f,0.97493f,0.97536f,0.97500f,0.97384f,0.97334f,0.97335f,0.97263f,0.97371f,0.97497f,0.97505f,0.97535f,0.97677f,0.97768f,0.97844f,0.97970f,0.98115f,0.98282f,0.98439f,0.98597f,0.98752f,0.98929f,0.99130f,0.99326f,0.99532f,0.99775f,1.00000f,0.00000f,0.98227f,0.98344f,0.98236f,0.98054f,0.97742f,0.97628f,0.97484f,0.97331f,0.97383f,0.97357f,0.97340f,0.97273f,0.97371f,0.97429f,0.97421f,0.97457f,0.97639f,0.97728f,0.97784f,0.97859f,0.98008f,0.98139f,0.98239f,0.98439f,0.98623f,0.98745f,0.98939f,0.99145f,0.99328f,0.99546f,0.99780f,1.00000f,0.00000f,0.93667f,0.98210f,0.98027f,0.97941f,0.97711f,0.97600f,0.97732f,0.97394f,0.97438f,0.97358f,0.97268f,0.97300f,0.97390f,0.97408f,0.97446f,0.97533f,0.97528f,0.97661f,0.97699f,0.97879f,0.97989f,0.98115f,0.98243f,0.98392f,0.98558f,0.98745f,0.98941f,0.99141f,0.99331f,0.99553f,0.99772f,1.00000f,0.00000f,0.91144f,0.95520f,0.97509f,0.97818f,0.97602f,0.97570f,0.97379f,0.97336f,0.97243f,0.97254f,0.97271f,0.97301f,0.97369f,0.97326f,0.97416f,0.97510f,0.97591f,0.97496f,0.97709f,0.97878f,0.97982f,0.98119f,0.98263f,0.98409f,0.98563f,0.98774f,0.98946f,0.99126f,0.99326f,0.99552f,0.99773f,1.00000f,0.00000f,0.90159f,0.92206f,0.95530f,0.97057f,0.97572f,0.97460f,0.97431f,0.97359f,0.97307f,0.97264f,0.97191f,0.97215f,0.97276f,0.97356f,0.97394f,0.97505f,0.97488f,0.97419f,0.97798f,0.97835f,0.97976f,0.98131f,0.98244f,0.98405f,0.98564f,0.98743f,0.98938f,0.99129f,0.99347f,0.99557f,0.99762f,1.00000f,0.00000f,0.90212f,0.90788f,0.92749f,0.95079f,0.96417f,0.97134f,0.97311f,0.97195f,0.97243f,0.97185f,0.97316f,0.97237f,0.97222f,0.97338f,0.97468f,0.97421f,0.97201f,0.97388f,0.97748f,0.97846f,0.97990f,0.98099f,0.98255f,0.98394f,0.98505f,0.98731f,0.98931f,0.99138f,0.99319f,0.99538f,0.99758f,1.00000f,0.00000f,0.90632f,0.89339f,0.90769f,0.92742f,0.94254f,0.95857f,0.96364f,0.96898f,0.96991f,0.97359f,0.97136f,0.97280f,0.97328f,0.97216f,0.97282f,0.97092f,0.96951f,0.97345f,0.97621f,0.97787f,0.97929f,0.98092f,0.98229f,0.98387f,0.98551f,0.98722f,0.98909f,0.99122f,0.99308f,0.99526f,0.99762f,1.00000f,0.00000f,0.92137f,0.89584f,0.89798f,0.90590f,0.92317f,0.94039f,0.94933f,0.95831f,0.96409f,0.96645f,0.96917f,0.97181f,0.97099f,0.97220f,0.96971f,0.96543f,0.96642f,0.97196f,0.97627f,0.97824f,0.97890f,0.98079f,0.98236f,0.98356f,0.98528f,0.98727f,0.98896f,0.99094f,0.99283f,0.99523f,0.99747f,1.00000f,0.00000f,0.92520f,0.89903f,0.89228f,0.89650f,0.90729f,0.91507f,0.92878f,0.94225f,0.95080f,0.95751f,0.96238f,0.96627f,0.96706f,0.96673f,0.96326f,0.95985f,0.96478f,0.97194f,0.97567f,0.97792f,0.97884f,0.98053f,0.98144f,0.98378f,0.98559f,0.98695f,0.98880f,0.99083f,0.99282f,0.99515f,0.99750f,1.00000f,0.00000f,0.93342f,0.90418f,0.88903f,0.89039f,0.89840f,0.90338f,0.91064f,0.92171f,0.93257f,0.94372f,0.95095f,0.95571f,0.95562f,0.95243f,0.95096f,0.95347f,0.96084f,0.96998f,0.97459f,0.97756f,0.97930f,0.98042f,0.98134f,0.98331f,0.98499f,0.98681f,0.98876f,0.99097f,0.99283f,0.99501f,0.99748f,1.00000f,0.00000f,0.93473f,0.91075f,0.89442f,0.88974f,0.88928f,0.89628f,0.89887f,0.90700f,0.91351f,0.92368f,0.93133f,0.93508f,0.93422f,0.93380f,0.93260f,0.94321f,0.95414f,0.96597f,0.97319f,0.97522f,0.97839f,0.97943f,0.98105f,0.98349f,0.98525f,0.98688f,0.98908f,0.99048f,0.99269f,0.99484f,0.99730f,1.00000f,0.00000f,0.94096f,0.91211f,0.90238f,0.89600f,0.88954f,0.88669f,0.89344f,0.89850f,0.89987f,0.90454f,0.90670f,0.90522f,0.90345f,0.90671f,0.91459f,0.92895f,0.94459f,0.95724f,0.96641f,0.97081f,0.97354f,0.97819f,0.98051f,0.98397f,0.98421f,0.98606f,0.98866f,0.99099f,0.99228f,0.99475f,0.99704f,1.00000f,0.00000f,0.94406f,0.91952f,0.90406f,0.89721f,0.88474f,0.88931f,0.88720f,0.88375f,0.88647f,0.88126f,0.87484f,0.87280f,0.87248f,0.88122f,0.89549f,0.91045f,0.93026f,0.94417f,0.95449f,0.96369f,0.96936f,0.97303f,0.97835f,0.98130f,0.98371f,0.98639f,0.98794f,0.99067f,0.99232f,0.99488f,0.99707f,1.00000f,0.00000f,0.94125f,0.93156f,0.91721f,0.89597f,0.89110f,0.87935f,0.88043f,0.87043f,0.86046f,0.85323f,0.84465f,0.83819f,0.84242f,0.85672f,0.87303f,0.89132f,0.90955f,0.92709f,0.94192f,0.94995f,0.95741f,0.96438f,0.96943f,0.97456f,0.97884f,0.98254f,0.98607f,0.98987f,0.99227f,0.99443f,0.99716f,1.00000f,0.00000f,0.94257f,0.93687f,0.91217f,0.90049f,0.89070f,0.87115f,0.86573f,0.85139f,0.83850f,0.82527f,0.81102f,0.80917f,0.81660f,0.83203f,0.84942f,0.87151f,0.89198f,0.91058f,0.92268f,0.93475f,0.94333f,0.95217f,0.95686f,0.96419f,0.97051f,0.97642f,0.98116f,0.98692f,0.99157f,0.99449f,0.99710f,1.00000f,0.00000f,0.94817f,0.93018f,0.91676f,0.90150f,0.88384f,0.86319f,0.85083f,0.83009f,0.80926f,0.79025f,0.78196f,0.78426f,0.79460f,0.80836f,0.83150f,0.85338f,0.87483f,0.88816f,0.90431f,0.91659f,0.92826f,0.93438f,0.94483f,0.94967f,0.95885f,0.96583f,0.97240f,0.98026f,0.98712f,0.99355f,0.99711f,1.00000f,0.00000f,0.94477f,0.92917f,0.90939f,0.89077f,0.86877f,0.84323f,0.82092f,0.79588f,0.77243f,0.75816f,0.75683f,0.75868f,0.77651f,0.79294f,0.81630f,0.83861f,0.85511f,0.87113f,0.88491f,0.89951f,0.90768f,0.91820f,0.92479f,0.93405f,0.94333f,0.94997f,0.95937f,0.96893f,0.97868f,0.98908f,0.99589f,1.00000f,0.00000f,0.92781f,0.91147f,0.90127f,0.87028f,0.84371f,0.81584f,0.79411f,0.75830f,0.73974f,0.72924f,0.73035f,0.74310f,0.75687f,0.77802f,0.80013f,0.81844f,0.83669f,0.85315f,0.86540f,0.87963f,0.88795f,0.89616f,0.90642f,0.91464f,0.92510f,0.93187f,0.94403f,0.95457f,0.96821f,0.98085f,0.99365f,0.99999f,0.00000f,0.92298f,0.90089f,0.87663f,0.84356f,0.81430f,0.78240f,0.74902f,0.72524f,0.70920f,0.69885f,0.71107f,0.72136f,0.74067f,0.76657f,0.78783f,0.80584f,0.82464f,0.83694f,0.84957f,0.85878f,0.87250f,0.87528f,0.88705f,0.89370f,0.90257f,0.91339f,0.92734f,0.93716f,0.95431f,0.96952f,0.98817f,0.99994f,0.00000f,0.89696f,0.87199f,0.84037f,0.80604f,0.77494f,0.74363f,0.71433f,0.68496f,0.67344f,0.68066f,0.69030f,0.70888f,0.72834f,0.75343f,0.77465f,0.79487f,0.81214f,0.82157f,0.83507f,0.84314f,0.85179f,0.86181f,0.86585f,0.87441f,0.88291f,0.89256f,0.90438f,0.91874f,0.93410f,0.95443f,0.97882f,0.99974f,0.00000f,0.86359f,0.83491f,0.80602f,0.76901f,0.73393f,0.70022f,0.67149f,0.65718f,0.65027f,0.65578f,0.66832f,0.69207f,0.71049f,0.73953f,0.76306f,0.77819f,0.79664f,0.80873f,0.81885f,0.82942f,0.83284f,0.84149f,0.84780f,0.85494f,0.86142f,0.87375f,0.88204f,0.89798f,0.91486f,0.93847f,0.96663f,0.99922f,0.00000f,0.82197f,0.80019f,0.76484f,0.72961f,0.69527f,0.65618f,0.63578f,0.62145f,0.62669f,0.63881f,0.65363f,0.67964f,0.70648f,0.72768f,0.75374f,0.76587f,0.78194f,0.79394f,0.80677f,0.81005f,0.81861f,0.82535f,0.82805f,0.83503f,0.84359f,0.85242f,0.86120f,0.87616f,0.89416f,0.91689f,0.94836f,0.99775f,0.00000f,0.79383f,0.75481f,0.71857f,0.68281f,0.65598f,0.62038f,0.59972f,0.59291f,0.60087f,0.61797f,0.63899f,0.66263f,0.69075f,0.71595f,0.73473f,0.75321f,0.76872f,0.77798f,0.79116f,0.79990f,0.80167f,0.80979f,0.81505f,0.82083f,0.82572f,0.83365f,0.84138f,0.85668f,0.87433f,0.89805f,0.92902f,0.99606f,0.00000f,0.74455f,0.72030f,0.67589f,0.64404f,0.60754f,0.58067f,0.56917f,0.56670f,0.57775f,0.59962f,0.62581f,0.65061f,0.67939f,0.69737f,0.72307f,0.74334f,0.75694f,0.76449f,0.77524f,0.78128f,0.79072f,0.79611f,0.79985f,0.80346f,0.80958f,0.81747f,0.82365f,0.83491f,0.84981f,0.87535f,0.90819f,0.99214f,0.00000f,0.70164f,0.67166f,0.63214f,0.59989f,0.56536f,0.54286f,0.53284f,0.54058f,0.55831f,0.58770f,0.61077f,0.63789f,0.66121f,0.68384f,0.70652f,0.73111f,0.74529f,0.75287f,0.76788f,0.77344f,0.77670f,0.78266f,0.78368f,0.78828f,0.79512f,0.79770f,0.80565f,0.81631f,0.83215f,0.85183f,0.88644f,0.98281f,0.00000f,0.66427f,0.62886f,0.58747f,0.55094f,0.52657f,0.51227f,0.50572f,0.52197f,0.54579f,0.56983f,0.60165f,0.62540f,0.65128f,0.67012f,0.69918f,0.71279f,0.72701f,0.73793f,0.75025f,0.75772f,0.76673f,0.77226f,0.77187f,0.77809f,0.77816f,0.78364f,0.78827f,0.79876f,0.81172f,0.82919f,0.85884f,0.96661f,0.00000f,0.61729f,0.58582f,0.54637f,0.51499f,0.48619f,0.47882f,0.48421f,0.50240f,0.52582f,0.55319f,0.58136f,0.61221f,0.63513f,0.66223f,0.68656f,0.69901f,0.71357f,0.72589f,0.73678f,0.75142f,0.75458f,0.75800f,0.76390f,0.76655f,0.76879f,0.77035f,0.77493f,0.77912f,0.78939f,0.80678f,0.83101f,0.94590f,0.00000f,0.58026f,0.53991f,0.50762f,0.47702f,0.45380f,0.45285f,0.46065f,0.48285f,0.51042f,0.53665f,0.56828f,0.59745f,0.62094f,0.64769f,0.66867f,0.68633f,0.70073f,0.71529f,0.72680f,0.73551f,0.73923f,0.74730f,0.75237f,0.75276f,0.75912f,0.76107f,0.76169f,0.76512f,0.77531f,0.78612f,0.80461f,0.92022f,0.00000f,0.53843f,0.49779f,0.47096f,0.44090f,0.42499f,0.42844f,0.44226f,0.46410f,0.49439f,0.52250f,0.55196f,0.58093f,0.60558f,0.63083f,0.65363f,0.67279f,0.68924f,0.70312f,0.71449f,0.72368f,0.73456f,0.74065f,0.74506f,0.74540f,0.74911f,0.74993f,0.75032f,0.74874f,0.75278f,0.76009f,0.77581f,0.88749f,0.00000f,0.49723f,0.46741f,0.43569f,0.41135f,0.39870f,0.40515f,0.41695f,0.44503f,0.47800f,0.50998f,0.54090f,0.56761f,0.59566f,0.62267f,0.64048f,0.66158f,0.67113f,0.69001f,0.69705f,0.71490f,0.72101f,0.73283f,0.73590f,0.73675f,0.73758f,0.74281f,0.73875f,0.73282f,0.73908f,0.73739f,0.74763f,0.85620f,0.00000f,0.45887f,0.42813f,0.40267f,0.37699f,0.37108f,0.38627f,0.40460f,0.43105f,0.46178f,0.49692f,0.52299f,0.55221f,0.58248f,0.60486f,0.62912f,0.64671f,0.66354f,0.67615f,0.68979f,0.70186f,0.71030f,0.72083f,0.72500f,0.72947f,0.73207f,0.73172f,0.72837f,0.73370f,0.72260f,0.71881f,0.71533f,0.81506f,
 0.00000f,1.00000f,0.99668f,1.00000f,0.99074f,0.96839f,0.99040f,0.99400f,0.98489f,0.96693f,0.96366f,0.99781f,0.98131f,0.98950f,0.98514f,0.98394f,0.98151f,0.97211f,0.97075f,0.97657f,0.97258f,0.96992f,0.96865f,0.96383f,0.97485f,0.97438f,0.97644f,0.97684f,0.98095f,0.99079f,0.99263f,0.99618f,1.00000f,0.00000f,0.99251f,0.98420f,0.98465f,0.98152f,0.97744f,0.97666f,0.97653f,0.97500f,0.97311f,0.97305f,0.97501f,0.97412f,0.97435f,0.97433f,0.97590f,0.97563f,0.97552f,0.97675f,0.97825f,0.97891f,0.97868f,0.97911f,0.97939f,0.98286f,0.98447f,0.98590f,0.98807f,0.98989f,0.99298f,0.99507f,0.99751f,1.00000f,0.00000f,0.99075f,0.98679f,0.98338f,0.97917f,0.97804f,0.97732f,0.97551f,0.97498f,0.97420f,0.97416f,0.97346f,0.97332f,0.97429f,0.97389f,0.97377f,0.97491f,0.97586f,0.97633f,0.97832f,0.97866f,0.98012f,0.98154f,0.98181f,0.98421f,0.98618f,0.98749f,0.98938f,0.99151f,0.99332f,0.99547f,0.99777f,1.00000f,0.00000f,0.98116f,0.98533f,0.98251f,0.97893f,0.97788f,0.97687f,0.97534f,0.97445f,0.97470f,0.97349f,0.97388f,0.97311f,0.97303f,0.97420f,0.97469f,0.97461f,0.97573f,0.97653f,0.97788f,0.97935f,0.97993f,0.98111f,0.98277f,0.98411f,0.98589f,0.98773f,0.98952f,0.99129f,0.99326f,0.99548f,0.99777f,1.00000f,0.00000f,0.93873f,0.98230f,0.98237f,0.98017f,0.97690f,0.97537f,0.97416f,0.97502f,0.97326f,0.97339f,0.97377f,0.97319f,0.97334f,0.97411f,0.97399f,0.97496f,0.97481f,0.97674f,0.97782f,0.97837f,0.97986f,0.98122f,0.98249f,0.98422f,0.98574f,0.98785f,0.98915f,0.99131f,0.99343f,0.99536f,0.99780f,1.00000f,0.00000f,0.90848f,0.95471f,0.97619f,0.97779f,0.97664f,0.97488f,0.97513f,0.97415f,0.97306f,0.97307f,0.97350f,0.97156f,0.97197f,0.97405f,0.97513f,0.97475f,0.97562f,0.97608f,0.97720f,0.97811f,0.98003f,0.98044f,0.98253f,0.98425f,0.98586f,0.98776f,0.98926f,0.99123f,0.99331f,0.99549f,0.99771f,1.00000f,0.00000f,0.89939f,0.91983f,0.95134f,0.97003f,0.97556f,0.97453f,0.97412f,0.97394f,0.97224f,0.97259f,0.97196f,0.97317f,0.97228f,0.97365f,0.97413f,0.97489f,0.97482f,0.97686f,0.97582f,0.97770f,0.97932f,0.98081f,0.98283f,0.98438f,0.98573f,0.98731f,0.98920f,0.99118f,0.99328f,0.99540f,0.99756f,1.00000f,0.00000f,0.90479f,0.90395f,0.92359f,0.94954f,0.96385f,0.97099f,0.97349f,0.97209f,0.97317f,0.97196f,0.97246f,0.97274f,0.97330f,0.97369f,0.97335f,0.97487f,0.97564f,0.97540f,0.97438f,0.97733f,0.97942f,0.98081f,0.98276f,0.98418f,0.98566f,0.98727f,0.98918f,0.99117f,0.99318f,0.99532f,0.99754f,1.00000f,0.00000f,0.91075f,0.89917f,0.90834f,0.92361f,0.94461f,0.95616f,0.96244f,0.96858f,0.97159f,0.97162f,0.97211f,0.97290f,0.97181f,0.97306f,0.97375f,0.97380f,0.97488f,0.97271f,0.97253f,0.97666f,0.97984f,0.98086f,0.98241f,0.98363f,0.98527f,0.98714f,0.98918f,0.99112f,0.99325f,0.99537f,0.99758f,1.00000f,0.00000f,0.91383f,0.89591f,0.89734f,0.90939f,0.92166f,0.93736f,0.94813f,0.95840f,0.96455f,0.96753f,0.97118f,0.97202f,0.97277f,0.97292f,0.97290f,0.97355f,0.97058f,0.96817f,0.97087f,0.97574f,0.97927f,0.98087f,0.98218f,0.98390f,0.98536f,0.98697f,0.98901f,0.99107f,0.99308f,0.99511f,0.99755f,1.00000f,0.00000f,0.92169f,0.90230f,0.89177f,0.89847f,0.90429f,0.91505f,0.93066f,0.94120f,0.94987f,0.95760f,0.96292f,0.96665f,0.96935f,0.97083f,0.97107f,0.96797f,0.96611f,0.96435f,0.97031f,0.97476f,0.97859f,0.97956f,0.98178f,0.98338f,0.98576f,0.98679f,0.98898f,0.99064f,0.99264f,0.99507f,0.99746f,1.00000f,0.00000f,0.92763f,0.90427f,0.88857f,0.89217f,0.89752f,0.90482f,0.91326f,0.92227f,0.93265f,0.94267f,0.95202f,0.95693f,0.96230f,0.96473f,0.96248f,0.95987f,0.95842f,0.96090f,0.96850f,0.97572f,0.97817f,0.98036f,0.98195f,0.98321f,0.98528f,0.98688f,0.98878f,0.99048f,0.99286f,0.99514f,0.99742f,1.00000f,0.00000f,0.94162f,0.91237f,0.89511f,0.89046f,0.88724f,0.89531f,0.89734f,0.90851f,0.91868f,0.92558f,0.93364f,0.94170f,0.94672f,0.94692f,0.94794f,0.94461f,0.94797f,0.95603f,0.96521f,0.97300f,0.97660f,0.97951f,0.98210f,0.98284f,0.98538f,0.98652f,0.98867f,0.99072f,0.99275f,0.99514f,0.99721f,1.00000f,0.00000f,0.93756f,0.91974f,0.90677f,0.88834f,0.88756f,0.89226f,0.88992f,0.89761f,0.90029f,0.90838f,0.91288f,0.92011f,0.92464f,0.92119f,0.92269f,0.92764f,0.93551f,0.94641f,0.95924f,0.96897f,0.97542f,0.97722f,0.98061f,0.98256f,0.98476f,0.98696f,0.98790f,0.99028f,0.99224f,0.99482f,0.99729f,1.00000f,0.00000f,0.94019f,0.92294f,0.90871f,0.89918f,0.89242f,0.88353f,0.88575f,0.89065f,0.88832f,0.89258f,0.89424f,0.89584f,0.89464f,0.89317f,0.89577f,0.90601f,0.91842f,0.93365f,0.94785f,0.96007f,0.96791f,0.97216f,0.97741f,0.97941f,0.98318f,0.98543f,0.98825f,0.99097f,0.99265f,0.99480f,0.99719f,1.00000f,0.00000f,0.94571f,0.92978f,0.91089f,0.90287f,0.89139f,0.88365f,0.88563f,0.87993f,0.88049f,0.87500f,0.87092f,0.86538f,0.86248f,0.86218f,0.86937f,0.88052f,0.90134f,0.91702f,0.93329f,0.94526f,0.95469f,0.96440f,0.96905f,0.97417f,0.97802f,0.98307f,0.98659f,0.99011f,0.99265f,0.99487f,0.99718f,1.00000f,0.00000f,0.94839f,0.92781f,0.91642f,0.90584f,0.89486f,0.88100f,0.87826f,0.87275f,0.86475f,0.85221f,0.84639f,0.83427f,0.83218f,0.83554f,0.84241f,0.86202f,0.87965f,0.89736f,0.91671f,0.92971f,0.94024f,0.94986f,0.95718f,0.96525f,0.97103f,0.97670f,0.98228f,0.98619f,0.99052f,0.99477f,0.99683f,1.00000f,0.00000f,0.94879f,0.93513f,0.91613f,0.89728f,0.89487f,0.88019f,0.87364f,0.86035f,0.84380f,0.82869f,0.81622f,0.80923f,0.80745f,0.80905f,0.82245f,0.83739f,0.85860f,0.87503f,0.89949f,0.91081f,0.92304f,0.93405f,0.94255f,0.95105f,0.95806f,0.96612f,0.97265f,0.97927f,0.98617f,0.99129f,0.99660f,1.00000f,0.00000f,0.94657f,0.93419f,0.92230f,0.90635f,0.88883f,0.87308f,0.85648f,0.83873f,0.82051f,0.80449f,0.78926f,0.78116f,0.78244f,0.78825f,0.80504f,0.81999f,0.84139f,0.86010f,0.87618f,0.89077f,0.90334f,0.91572f,0.92490f,0.93315f,0.94343f,0.94938f,0.95822f,0.96769f,0.97614f,0.98557f,0.99410f,1.00000f,0.00000f,0.93799f,0.93548f,0.91867f,0.89967f,0.87668f,0.85398f,0.83533f,0.81382f,0.78603f,0.76712f,0.75681f,0.75232f,0.76182f,0.76977f,0.78672f,0.80746f,0.82459f,0.84053f,0.85774f,0.87119f,0.88424f,0.89564f,0.90505f,0.91345f,0.92267f,0.93278f,0.94051f,0.95181f,0.96159f,0.97480f,0.98792f,0.99999f,0.00000f,0.93730f,0.91969f,0.90306f,0.87994f,0.85840f,0.83384f,0.80835f,0.78286f,0.75148f,0.73800f,0.73072f,0.73057f,0.74709f,0.75567f,0.77330f,0.79052f,0.80961f,0.82618f,0.83998f,0.85344f,0.86456f,0.87697f,0.88263f,0.89107f,0.90236f,0.91141f,0.92251f,0.93187f,0.94516f,0.95976f,0.97854f,0.99990f,0.00000f,0.92471f,0.90567f,0.88817f,0.85758f,0.83455f,0.80353f,0.77735f,0.74523f,0.72690f,0.70879f,0.70754f,0.71190f,0.72266f,0.74173f,0.76082f,0.77737f,0.79322f,0.81162f,0.82328f,0.83662f,0.84443f,0.85354f,0.86423f,0.87147f,0.87968f,0.88629f,0.89649f,0.90897f,0.92306f,0.94024f,0.96323f,0.99956f,0.00000f,0.91086f,0.88548f,0.86075f,0.83575f,0.80266f,0.76853f,0.73245f,0.71048f,0.69219f,0.68238f,0.68655f,0.69510f,0.70896f,0.72669f,0.74593f,0.76511f,0.77999f,0.79817f,0.81200f,0.82068f,0.83069f,0.83503f,0.84322f,0.85104f,0.85660f,0.86386f,0.87503f,0.88429f,0.90112f,0.91993f,0.94226f,0.99834f,0.00000f,0.88019f,0.86639f,0.82998f,0.79963f,0.77089f,0.73232f,0.70380f,0.67777f,0.66299f,0.66140f,0.66648f,0.67662f,0.69251f,0.71530f,0.73685f,0.75446f,0.77000f,0.78839f,0.79519f,0.80655f,0.81346f,0.81854f,0.82615f,0.83073f,0.83653f,0.84363f,0.84853f,0.86230f,0.87563f,0.89223f,0.92431f,0.99570f,0.00000f,0.84883f,0.81956f,0.79156f,0.76532f,0.73104f,0.69641f,0.66520f,0.64222f,0.63908f,0.63331f,0.64411f,0.66310f,0.68437f,0.70048f,0.72650f,0.74196f,0.75459f,0.76957f,0.77922f,0.79168f,0.80001f,0.79892f,0.80530f,0.81048f,0.81361f,0.81916f,0.82718f,0.83887f,0.85197f,0.86775f,0.89701f,0.98996f,0.00000f,0.82220f,0.78860f,0.75609f,0.72073f,0.68359f,0.64987f,0.62624f,0.61005f,0.60644f,0.61805f,0.62828f,0.64811f,0.66901f,0.68729f,0.70889f,0.72825f,0.74670f,0.75849f,0.76982f,0.77393f,0.78370f,0.78813f,0.79079f,0.79348f,0.79711f,0.80134f,0.80646f,0.81343f,0.82705f,0.84248f,0.86971f,0.97484f,0.00000f,0.78215f,0.74745f,0.71479f,0.68163f,0.64511f,0.61184f,0.58878f,0.58196f,0.58386f,0.59656f,0.61320f,0.63352f,0.65517f,0.67687f,0.69687f,0.71614f,0.72757f,0.74287f,0.74975f,0.76207f,0.76791f,0.77005f,0.77405f,0.77846f,0.77659f,0.78166f,0.78168f,0.79335f,0.80226f,0.81635f,0.83908f,0.95426f,0.00000f,0.73696f,0.70946f,0.67373f,0.64290f,0.60548f,0.57758f,0.55798f,0.55827f,0.56440f,0.57778f,0.59397f,0.61827f,0.64306f,0.66763f,0.68441f,0.70022f,0.71761f,0.72876f,0.73921f,0.74634f,0.75293f,0.75760f,0.76157f,0.76069f,0.76309f,0.76289f,0.76477f,0.77263f,0.77671f,0.78859f,0.81079f,0.92712f,0.00000f,0.69791f,0.66443f,0.63959f,0.59648f,0.56875f,0.54330f,0.53046f,0.52779f,0.54424f,0.56296f,0.58492f,0.60490f,0.62862f,0.64731f,0.66940f,0.68674f,0.70130f,0.71692f,0.72925f,0.73598f,0.73612f,0.73740f,0.74579f,0.74805f,0.74639f,0.74932f,0.74730f,0.75085f,0.75222f,0.75987f,0.77387f,0.89608f,0.00000f,0.66062f,0.62682f,0.59458f,0.56058f,0.52539f,0.51257f,0.49829f,0.51153f,0.52349f,0.54308f,0.56879f,0.59093f,0.61521f,0.63599f,0.65633f,0.67550f,0.69033f,0.70368f,0.70951f,0.72176f,0.72453f,0.72843f,0.73374f,0.73511f,0.73299f,0.73120f,0.73229f,0.72850f,0.73514f,0.73567f,0.73786f,0.85944f,0.00000f,0.62327f,0.58017f,0.55494f,0.52003f,0.49956f,0.48068f,0.47773f,0.48683f,0.50883f,0.52827f,0.55244f,0.57629f,0.59941f,0.62583f,0.64131f,0.66071f,0.67736f,0.68721f,0.69860f,0.70550f,0.71167f,0.72296f,0.72319f,0.72025f,0.72026f,0.72134f,0.71152f,0.71681f,0.71430f,0.71000f,0.70599f,0.81898f,0.00000f,0.58180f,0.54248f,0.51816f,0.48110f,0.46230f,0.45145f,0.45604f,0.46718f,0.49157f,0.51695f,0.53967f,0.56312f,0.58328f,0.61094f,0.63140f,0.64858f,0.66233f,0.67431f,0.68738f,0.69757f,0.70003f,0.70413f,0.71013f,0.71444f,0.71065f,0.70776f,0.70494f,0.69987f,0.69026f,0.68400f,0.67327f,0.77660f,0.00000f,0.54443f,0.51143f,0.48198f,0.44981f,0.43194f,0.42501f,0.43094f,0.44758f,0.47161f,0.49701f,0.52427f,0.54583f,0.57368f,0.59229f,0.61385f,0.63231f,0.64812f,0.66419f,0.67293f,0.68164f,0.68982f,0.69695f,0.69746f,0.69867f,0.69884f,0.69485f,0.68855f,0.68069f,0.67354f,0.65906f,0.63707f,0.73317f,
 0.00000f,1.00000f,0.99728f,1.00000f,0.99059f,0.96771f,0.99076f,0.99327f,0.98430f,0.96536f,0.96404f,0.99787f,0.98119f,0.98876f,0.98548f,0.98416f,0.98224f,0.97221f,0.97103f,0.97612f,0.97344f,0.97254f,0.97726f,0.97549f,0.97435f,0.97436f,0.97377f,0.98218f,0.98519f,0.98801f,0.99255f,0.99373f,1.00000f,0.00000f,0.99118f,0.98487f,0.98350f,0.97984f,0.97720f,0.97893f,0.97619f,0.97626f,0.97333f,0.97364f,0.97437f,0.97356f,0.97461f,0.97443f,0.97522f,0.97575f,0.97487f,0.97653f,0.97837f,0.97824f,0.97924f,0.97977f,0.98135f,0.98267f,0.98389f,0.98574f,0.98846f,0.99039f,0.99259f,0.99502f,0.99724f,1.00000f,0.00000f,0.99075f,0.98549f,0.98165f,0.98044f,0.97758f,0.97669f,0.97570f,0.97455f,0.97398f,0.97401f,0.97396f,0.97300f,0.97382f,0.97397f,0.97492f,0.97522f,0.97524f,0.97643f,0.97768f,0.97924f,0.98027f,0.98119f,0.98290f,0.98446f,0.98604f,0.98754f,0.98944f,0.99118f,0.99337f,0.99550f,0.99772f,1.00000f,0.00000f,0.98395f,0.98680f,0.98154f,0.97982f,0.97798f,0.97677f,0.97544f,0.97429f,0.97446f,0.97289f,0.97300f,0.97330f,0.97365f,0.97444f,0.97449f,0.97516f,0.97565f,0.97647f,0.97746f,0.97905f,0.98033f,0.98100f,0.98262f,0.98419f,0.98598f,0.98754f,0.98942f,0.99120f,0.99346f,0.99553f,0.99767f,1.00000f,0.00000f,0.93894f,0.98195f,0.98227f,0.97909f,0.97854f,0.97680f,0.97468f,0.97261f,0.97300f,0.97354f,0.97326f,0.97386f,0.97375f,0.97381f,0.97477f,0.97470f,0.97559f,0.97614f,0.97798f,0.97893f,0.98012f,0.98131f,0.98281f,0.98400f,0.98610f,0.98748f,0.98941f,0.99153f,0.99339f,0.99541f,0.99776f,1.00000f,0.00000f,0.90813f,0.95610f,0.97718f,0.97916f,0.97707f,0.97542f,0.97393f,0.97411f,0.97388f,0.97324f,0.97267f,0.97341f,0.97365f,0.97361f,0.97440f,0.97564f,0.97570f,0.97738f,0.97740f,0.97846f,0.97903f,0.98096f,0.98279f,0.98427f,0.98573f,0.98751f,0.98938f,0.99117f,0.99330f,0.99542f,0.99777f,1.00000f,0.00000f,0.90112f,0.92204f,0.95452f,0.97083f,0.97640f,0.97421f,0.97343f,0.97356f,0.97227f,0.97257f,0.97231f,0.97262f,0.97304f,0.97354f,0.97381f,0.97500f,0.97555f,0.97625f,0.97754f,0.97789f,0.97852f,0.98099f,0.98248f,0.98444f,0.98584f,0.98731f,0.98952f,0.99108f,0.99325f,0.99533f,0.99762f,1.00000f,0.00000f,0.90358f,0.90687f,0.92544f,0.94799f,0.96351f,0.97219f,0.97351f,0.97356f,0.97321f,0.97260f,0.97311f,0.97267f,0.97301f,0.97393f,0.97385f,0.97503f,0.97570f,0.97621f,0.97700f,0.97577f,0.97828f,0.98087f,0.98216f,0.98391f,0.98562f,0.98727f,0.98898f,0.99115f,0.99312f,0.99533f,0.99747f,1.00000f,0.00000f,0.90967f,0.89500f,0.91053f,0.92685f,0.94248f,0.95775f,0.96613f,0.96892f,0.97193f,0.97137f,0.97289f,0.97202f,0.97129f,0.97290f,0.97324f,0.97414f,0.97491f,0.97595f,0.97457f,0.97366f,0.97766f,0.98028f,0.98238f,0.98396f,0.98575f,0.98721f,0.98937f,0.99110f,0.99319f,0.99535f,0.99755f,1.00000f,0.00000f,0.91373f,0.89576f,0.89306f,0.90846f,0.92120f,0.93636f,0.94822f,0.95799f,0.96364f,0.96896f,0.96991f,0.97074f,0.97220f,0.97264f,0.97333f,0.97386f,0.97539f,0.97342f,0.97133f,0.97268f,0.97632f,0.97999f,0.98206f,0.98370f,0.98502f,0.98736f,0.98859f,0.99104f,0.99282f,0.99495f,0.99738f,1.00000f,0.00000f,0.92989f,0.89525f,0.89211f,0.89402f,0.90380f,0.91720f,0.93070f,0.93949f,0.95058f,0.95905f,0.96465f,0.96912f,0.96859f,0.97200f,0.97215f,0.97353f,0.97189f,0.96813f,0.96679f,0.96927f,0.97585f,0.97982f,0.98208f,0.98307f,0.98507f,0.98694f,0.98896f,0.99081f,0.99306f,0.99527f,0.99748f,1.00000f,0.00000f,0.93115f,0.90903f,0.89160f,0.88952f,0.89520f,0.90256f,0.91195f,0.92311f,0.93477f,0.94266f,0.95069f,0.95661f,0.96160f,0.96630f,0.96808f,0.96846f,0.96532f,0.96194f,0.96332f,0.96974f,0.97540f,0.97954f,0.98180f,0.98361f,0.98491f,0.98711f,0.98880f,0.99052f,0.99287f,0.99527f,0.99734f,1.00000f,0.00000f,0.93120f,0.91688f,0.90017f,0.88974f,0.89036f,0.89455f,0.89762f,0.90536f,0.91647f,0.92547f,0.93223f,0.94437f,0.95089f,0.95497f,0.95385f,0.95576f,0.95250f,0.95447f,0.96097f,0.96681f,0.97397f,0.97873f,0.98104f,0.98259f,0.98467f,0.98673f,0.98854f,0.99038f,0.99262f,0.99493f,0.99726f,1.00000f,0.00000f,0.94032f,0.91575f,0.89900f,0.89473f,0.88868f,0.88972f,0.89140f,0.89392f,0.90247f,0.91053f,0.91561f,0.92506f,0.92961f,0.93506f,0.93399f,0.93415f,0.93646f,0.94183f,0.95088f,0.96204f,0.97042f,0.97696f,0.97956f,0.98287f,0.98410f,0.98680f,0.98865f,0.98974f,0.99258f,0.99472f,0.99739f,1.00000f,0.00000f,0.94525f,0.92088f,0.91012f,0.89477f,0.88777f,0.88535f,0.88615f,0.88807f,0.89260f,0.89809f,0.89847f,0.90645f,0.90850f,0.90565f,0.90893f,0.91046f,0.91737f,0.92797f,0.94022f,0.95346f,0.96465f,0.97207f,0.97729f,0.98059f,0.98292f,0.98555f,0.98820f,0.99035f,0.99261f,0.99517f,0.99718f,1.00000f,0.00000f,0.94318f,0.92756f,0.91255f,0.90035f,0.88355f,0.88939f,0.88140f,0.88652f,0.88766f,0.88596f,0.88512f,0.88412f,0.87905f,0.88101f,0.88183f,0.88417f,0.89408f,0.90933f,0.92429f,0.93897f,0.95259f,0.96213f,0.96966f,0.97432f,0.97891f,0.98250f,0.98604f,0.98936f,0.99267f,0.99469f,0.99729f,1.00000f,0.00000f,0.94306f,0.93344f,0.91320f,0.89978f,0.89542f,0.88522f,0.88246f,0.87954f,0.87549f,0.86783f,0.86546f,0.85877f,0.85357f,0.84932f,0.85191f,0.86159f,0.87071f,0.89016f,0.90628f,0.92354f,0.93694f,0.94691f,0.95544f,0.96522f,0.97162f,0.97639f,0.98140f,0.98658f,0.99042f,0.99377f,0.99723f,1.00000f,0.00000f,0.94470f,0.93091f,0.91938f,0.90283f,0.89794f,0.88578f,0.88019f,0.86989f,0.86476f,0.85431f,0.83929f,0.83221f,0.82662f,0.82500f,0.82682f,0.83660f,0.85212f,0.87078f,0.88667f,0.90466f,0.91985f,0.93185f,0.94160f,0.95032f,0.95727f,0.96625f,0.97208f,0.97987f,0.98471f,0.99097f,0.99657f,1.00000f,0.00000f,0.94397f,0.92967f,0.92982f,0.90760f,0.90274f,0.88363f,0.87511f,0.85975f,0.84658f,0.83717f,0.81583f,0.80509f,0.79940f,0.80052f,0.80266f,0.81867f,0.83188f,0.84773f,0.87009f,0.88743f,0.89879f,0.91188f,0.92216f,0.93157f,0.94260f,0.95114f,0.96037f,0.96683f,0.97550f,0.98349f,0.99308f,1.00000f,0.00000f,0.94866f,0.93539f,0.92875f,0.90470f,0.89610f,0.87570f,0.86534f,0.84137f,0.82395f,0.80882f,0.78851f,0.78219f,0.77460f,0.77513f,0.78882f,0.80322f,0.81588f,0.83224f,0.84762f,0.86510f,0.87841f,0.89010f,0.90244f,0.91273f,0.92203f,0.93126f,0.94047f,0.95007f,0.96176f,0.97321f,0.98482f,0.99998f,0.00000f,0.94666f,0.93581f,0.91496f,0.90344f,0.88706f,0.86219f,0.84256f,0.82041f,0.79589f,0.77651f,0.76591f,0.75329f,0.75252f,0.75878f,0.77004f,0.78312f,0.79982f,0.81811f,0.83037f,0.84752f,0.86300f,0.87281f,0.88044f,0.89071f,0.90311f,0.90897f,0.92040f,0.93118f,0.94183f,0.95494f,0.97170f,0.99983f,0.00000f,0.94112f,0.92187f,0.90415f,0.89649f,0.86801f,0.84629f,0.82191f,0.79377f,0.76685f,0.75137f,0.73550f,0.73039f,0.73280f,0.74416f,0.75302f,0.76954f,0.78721f,0.80318f,0.81684f,0.82889f,0.84424f,0.85413f,0.86042f,0.86747f,0.87678f,0.88350f,0.89496f,0.90695f,0.91937f,0.93464f,0.95273f,0.99926f,0.00000f,0.92818f,0.90973f,0.89519f,0.87454f,0.84691f,0.81758f,0.78886f,0.76376f,0.74135f,0.71767f,0.70831f,0.71027f,0.71614f,0.72690f,0.74183f,0.75671f,0.77108f,0.78739f,0.80400f,0.81518f,0.82492f,0.83019f,0.83990f,0.84872f,0.85356f,0.85974f,0.87167f,0.87844f,0.89262f,0.90775f,0.92971f,0.99710f,0.00000f,0.91749f,0.88570f,0.86384f,0.84586f,0.81602f,0.79023f,0.75389f,0.72519f,0.70468f,0.68999f,0.68344f,0.68905f,0.70307f,0.71389f,0.73046f,0.75036f,0.76236f,0.77085f,0.78973f,0.79613f,0.80911f,0.81350f,0.81643f,0.82488f,0.83180f,0.83522f,0.84408f,0.85303f,0.86560f,0.88116f,0.90388f,0.98931f,0.00000f,0.88752f,0.87075f,0.85501f,0.81562f,0.78074f,0.75204f,0.71778f,0.69362f,0.67295f,0.66343f,0.66337f,0.67088f,0.68221f,0.70017f,0.71706f,0.73410f,0.74890f,0.76002f,0.77103f,0.78165f,0.78871f,0.79630f,0.79945f,0.80257f,0.80842f,0.81260f,0.81789f,0.82763f,0.83840f,0.85167f,0.87165f,0.97118f,0.00000f,0.85852f,0.83517f,0.81411f,0.78690f,0.74729f,0.71536f,0.67873f,0.66004f,0.64811f,0.64128f,0.64597f,0.65843f,0.66863f,0.68582f,0.70426f,0.72188f,0.73724f,0.74518f,0.75772f,0.76663f,0.77603f,0.78024f,0.78300f,0.78234f,0.78619f,0.78871f,0.79295f,0.80092f,0.80831f,0.82098f,0.83944f,0.94755f,0.00000f,0.83572f,0.80996f,0.77661f,0.74249f,0.70872f,0.67794f,0.64732f,0.63019f,0.61942f,0.61569f,0.62977f,0.63595f,0.65755f,0.67688f,0.68760f,0.70443f,0.71839f,0.73260f,0.74677f,0.75454f,0.75822f,0.76194f,0.76668f,0.76597f,0.76764f,0.76764f,0.77301f,0.77414f,0.78385f,0.79001f,0.80328f,0.91794f,0.00000f,0.79919f,0.77089f,0.74238f,0.70524f,0.67204f,0.63222f,0.61305f,0.60164f,0.59555f,0.59901f,0.60733f,0.62392f,0.63729f,0.65600f,0.67704f,0.69651f,0.71074f,0.72256f,0.73274f,0.73709f,0.74024f,0.74524f,0.75133f,0.74820f,0.75120f,0.74817f,0.74933f,0.74819f,0.75478f,0.75937f,0.76686f,0.88135f,0.00000f,0.76424f,0.73006f,0.70402f,0.67340f,0.62718f,0.60157f,0.58363f,0.56987f,0.56851f,0.57902f,0.58977f,0.60483f,0.62702f,0.64498f,0.66509f,0.68055f,0.69771f,0.70462f,0.71829f,0.72299f,0.72472f,0.73072f,0.73106f,0.73415f,0.72934f,0.73244f,0.72872f,0.72613f,0.72683f,0.72965f,0.72596f,0.84346f,0.00000f,0.72653f,0.69408f,0.65994f,0.63219f,0.59277f,0.56650f,0.55046f,0.54270f,0.54770f,0.56098f,0.57695f,0.59267f,0.61226f,0.62824f,0.65053f,0.66800f,0.68174f,0.69160f,0.70549f,0.70733f,0.71553f,0.71586f,0.71739f,0.71730f,0.71269f,0.71381f,0.70868f,0.70294f,0.70318f,0.69706f,0.68803f,0.79853f,0.00000f,0.69297f,0.65735f,0.62459f,0.59121f,0.55571f,0.53379f,0.51937f,0.51936f,0.51970f,0.54151f,0.55837f,0.57705f,0.59872f,0.61852f,0.63282f,0.65100f,0.66460f,0.67137f,0.69004f,0.69814f,0.70046f,0.69657f,0.70281f,0.70191f,0.69963f,0.69550f,0.68678f,0.68713f,0.67802f,0.66704f,0.64644f,0.75481f,0.00000f,0.64953f,0.62270f,0.58557f,0.55400f,0.51682f,0.50432f,0.49307f,0.49780f,0.50672f,0.52269f,0.54015f,0.55928f,0.58086f,0.60574f,0.61961f,0.63784f,0.65284f,0.66179f,0.67394f,0.68100f,0.68841f,0.68958f,0.68638f,0.68782f,0.69042f,0.68538f,0.67652f,0.66798f,0.65618f,0.64052f,0.61335f,0.70712f,0.00000f,0.61043f,0.58366f,0.55013f,0.51788f,0.49599f,0.47649f,0.47013f,0.47112f,0.48545f,0.50281f,0.52626f,0.54426f,0.56885f,0.58832f,0.60762f,0.62540f,0.64227f,0.64753f,0.65630f,0.66730f,0.67236f,0.67319f,0.67696f,0.67427f,0.67468f,0.66649f,0.66216f,0.64710f,0.63390f,0.61230f,0.57050f,0.65808f,
 0.00000f,1.00000f,0.99860f,1.00000f,0.99056f,0.96829f,0.99072f,0.99364f,0.98457f,0.96513f,0.96360f,0.99801f,0.98169f,0.98783f,0.98619f,0.98495f,0.98192f,0.97180f,0.97170f,0.97648f,0.97294f,0.97240f,0.99110f,0.97818f,0.96841f,0.97932f,0.98018f,0.98074f,0.98327f,0.98974f,0.99147f,0.99678f,1.00000f,0.00000f,0.99214f,0.98859f,0.98237f,0.97976f,0.97700f,0.97734f,0.97749f,0.97484f,0.97300f,0.97278f,0.97517f,0.97389f,0.97419f,0.97507f,0.97489f,0.97598f,0.97454f,0.97614f,0.97877f,0.97844f,0.97886f,0.98167f,0.98129f,0.98187f,0.98486f,0.98662f,0.98818f,0.99012f,0.99271f,0.99505f,0.99756f,1.00000f,0.00000f,0.99115f,0.98467f,0.98244f,0.97999f,0.97866f,0.97812f,0.97653f,0.97475f,0.97401f,0.97281f,0.97333f,0.97313f,0.97383f,0.97435f,0.97405f,0.97501f,0.97550f,0.97641f,0.97798f,0.97869f,0.97946f,0.98134f,0.98266f,0.98424f,0.98587f,0.98747f,0.98956f,0.99126f,0.99333f,0.99546f,0.99771f,1.00000f,0.00000f,0.98502f,0.98390f,0.98284f,0.97933f,0.97842f,0.97664f,0.97419f,0.97358f,0.97422f,0.97378f,0.97289f,0.97385f,0.97333f,0.97472f,0.97476f,0.97472f,0.97591f,0.97703f,0.97723f,0.97854f,0.98005f,0.98119f,0.98269f,0.98441f,0.98603f,0.98723f,0.98949f,0.99126f,0.99331f,0.99547f,0.99771f,1.00000f,0.00000f,0.93786f,0.98234f,0.98169f,0.97893f,0.97685f,0.97499f,0.97434f,0.97449f,0.97365f,0.97267f,0.97364f,0.97329f,0.97416f,0.97370f,0.97507f,0.97456f,0.97587f,0.97681f,0.97778f,0.97865f,0.97960f,0.98149f,0.98257f,0.98438f,0.98580f,0.98745f,0.98941f,0.99145f,0.99331f,0.99555f,0.99777f,1.00000f,0.00000f,0.90571f,0.95596f,0.97498f,0.97936f,0.97706f,0.97555f,0.97396f,0.97460f,0.97423f,0.97256f,0.97341f,0.97235f,0.97318f,0.97326f,0.97443f,0.97416f,0.97579f,0.97643f,0.97723f,0.97815f,0.98019f,0.98038f,0.98268f,0.98397f,0.98564f,0.98732f,0.98921f,0.99129f,0.99324f,0.99544f,0.99771f,1.00000f,0.00000f,0.89861f,0.92364f,0.95661f,0.97059f,0.97539f,0.97484f,0.97423f,0.97392f,0.97277f,0.97365f,0.97284f,0.97292f,0.97217f,0.97355f,0.97395f,0.97504f,0.97548f,0.97669f,0.97748f,0.97856f,0.97954f,0.97998f,0.98240f,0.98390f,0.98560f,0.98783f,0.98934f,0.99121f,0.99332f,0.99543f,0.99762f,1.00000f,0.00000f,0.90105f,0.90399f,0.92739f,0.94921f,0.96399f,0.97006f,0.97339f,0.97174f,0.97139f,0.97081f,0.97289f,0.97182f,0.97278f,0.97317f,0.97324f,0.97462f,0.97489f,0.97567f,0.97701f,0.97843f,0.97790f,0.97910f,0.98196f,0.98444f,0.98586f,0.98740f,0.98932f,0.99118f,0.99313f,0.99529f,0.99756f,1.00000f,0.00000f,0.91258f,0.89665f,0.91076f,0.92609f,0.94502f,0.95722f,0.96478f,0.96868f,0.97156f,0.97120f,0.97149f,0.97127f,0.97259f,0.97383f,0.97319f,0.97435f,0.97576f,0.97588f,0.97789f,0.97643f,0.97590f,0.97881f,0.98209f,0.98409f,0.98523f,0.98726f,0.98929f,0.99111f,0.99309f,0.99518f,0.99749f,1.00000f,0.00000f,0.91777f,0.89359f,0.90049f,0.90658f,0.92151f,0.93721f,0.94927f,0.95920f,0.96433f,0.96636f,0.97047f,0.97106f,0.97077f,0.97290f,0.97321f,0.97379f,0.97451f,0.97660f,0.97539f,0.97305f,0.97394f,0.97882f,0.98154f,0.98365f,0.98539f,0.98724f,0.98901f,0.99101f,0.99301f,0.99519f,0.99754f,1.00000f,0.00000f,0.92783f,0.89874f,0.88987f,0.89972f,0.90289f,0.91923f,0.92830f,0.94121f,0.95013f,0.95801f,0.96269f,0.96701f,0.96889f,0.97320f,0.97307f,0.97371f,0.97391f,0.97360f,0.97180f,0.96965f,0.97318f,0.97734f,0.98129f,0.98311f,0.98505f,0.98704f,0.98883f,0.99066f,0.99256f,0.99519f,0.99755f,1.00000f,0.00000f,0.92983f,0.90499f,0.88828f,0.89497f,0.89578f,0.90146f,0.91113f,0.92036f,0.93397f,0.94303f,0.95081f,0.95736f,0.96125f,0.96500f,0.96846f,0.97109f,0.96963f,0.96868f,0.96639f,0.96649f,0.97064f,0.97709f,0.98023f,0.98291f,0.98500f,0.98675f,0.98899f,0.99092f,0.99293f,0.99494f,0.99737f,1.00000f,0.00000f,0.93628f,0.91038f,0.89503f,0.89062f,0.88787f,0.89137f,0.89535f,0.90320f,0.91707f,0.92586f,0.93477f,0.94486f,0.94983f,0.95640f,0.95929f,0.96091f,0.96191f,0.95885f,0.95949f,0.96327f,0.96893f,0.97674f,0.98098f,0.98276f,0.98467f,0.98669f,0.98916f,0.99050f,0.99244f,0.99503f,0.99731f,1.00000f,0.00000f,0.94603f,0.91507f,0.90181f,0.89191f,0.88588f,0.88649f,0.89005f,0.89997f,0.90341f,0.90913f,0.91751f,0.92668f,0.93308f,0.93753f,0.94321f,0.94422f,0.94219f,0.94541f,0.95017f,0.95687f,0.96601f,0.97260f,0.97971f,0.98255f,0.98422f,0.98663f,0.98906f,0.99068f,0.99283f,0.99491f,0.99694f,1.00000f,0.00000f,0.94343f,0.92403f,0.90159f,0.89484f,0.89210f,0.88623f,0.88770f,0.88797f,0.89056f,0.89664f,0.90152f,0.90947f,0.91338f,0.92061f,0.92119f,0.92223f,0.92424f,0.92826f,0.93698f,0.94651f,0.95863f,0.96802f,0.97538f,0.97987f,0.98266f,0.98639f,0.98845f,0.99026f,0.99215f,0.99481f,0.99714f,1.00000f,0.00000f,0.94478f,0.93200f,0.91075f,0.90168f,0.88953f,0.88756f,0.88166f,0.88651f,0.88377f,0.88716f,0.88986f,0.89254f,0.89209f,0.89464f,0.89302f,0.89549f,0.89816f,0.90648f,0.91863f,0.93368f,0.94744f,0.95740f,0.96725f,0.97396f,0.97929f,0.98246f,0.98719f,0.99031f,0.99199f,0.99479f,0.99717f,1.00000f,0.00000f,0.94339f,0.92749f,0.91840f,0.90217f,0.89470f,0.88594f,0.88595f,0.88368f,0.88036f,0.88135f,0.87641f,0.87350f,0.87265f,0.86697f,0.86536f,0.86738f,0.87528f,0.88488f,0.89969f,0.91873f,0.93342f,0.94478f,0.95496f,0.96280f,0.97138f,0.97617f,0.98100f,0.98673f,0.99064f,0.99407f,0.99706f,1.00000f,0.00000f,0.94636f,0.93850f,0.91903f,0.90859f,0.89703f,0.89158f,0.88439f,0.87631f,0.87672f,0.86885f,0.85927f,0.85465f,0.84637f,0.83836f,0.83736f,0.84048f,0.85014f,0.86446f,0.88104f,0.89776f,0.91458f,0.92523f,0.93841f,0.94802f,0.95767f,0.96443f,0.97217f,0.97892f,0.98556f,0.99089f,0.99614f,1.00000f,0.00000f,0.94283f,0.93768f,0.91765f,0.90750f,0.89973f,0.89445f,0.88166f,0.87348f,0.86303f,0.85042f,0.84039f,0.82657f,0.81848f,0.81401f,0.81258f,0.81864f,0.82787f,0.84607f,0.86220f,0.88064f,0.89501f,0.91187f,0.92189f,0.93120f,0.94096f,0.94998f,0.95766f,0.96714f,0.97508f,0.98370f,0.99188f,1.00000f,0.00000f,0.94418f,0.94527f,0.92617f,0.91252f,0.90280f,0.88414f,0.86949f,0.86495f,0.84854f,0.82985f,0.81252f,0.80065f,0.79085f,0.78747f,0.79329f,0.79971f,0.81080f,0.82705f,0.84293f,0.86349f,0.87335f,0.88634f,0.90073f,0.91170f,0.92247f,0.93088f,0.93923f,0.94858f,0.96094f,0.97055f,0.98379f,0.99997f,0.00000f,0.94523f,0.94165f,0.93050f,0.91107f,0.89690f,0.87974f,0.86521f,0.84729f,0.82898f,0.80649f,0.78504f,0.77870f,0.76830f,0.76827f,0.77707f,0.78180f,0.79613f,0.80825f,0.82578f,0.84301f,0.85609f,0.86878f,0.87764f,0.89096f,0.89849f,0.90835f,0.91897f,0.92871f,0.93948f,0.95319f,0.96856f,0.99975f,0.00000f,0.94607f,0.93890f,0.92306f,0.90296f,0.88447f,0.86445f,0.84418f,0.82185f,0.80148f,0.78054f,0.76193f,0.75265f,0.74809f,0.74838f,0.75746f,0.76704f,0.78054f,0.79759f,0.80673f,0.82411f,0.83540f,0.84733f,0.85853f,0.86497f,0.87529f,0.88331f,0.89381f,0.90479f,0.91530f,0.93077f,0.94802f,0.99876f,0.00000f,0.94149f,0.92440f,0.91324f,0.89259f,0.86863f,0.84685f,0.82531f,0.79902f,0.77595f,0.75189f,0.73870f,0.72315f,0.72618f,0.73336f,0.74031f,0.75567f,0.76971f,0.78185f,0.80080f,0.80818f,0.81534f,0.82850f,0.83904f,0.84446f,0.85200f,0.85985f,0.86896f,0.87757f,0.88828f,0.90056f,0.92100f,0.99266f,0.00000f,0.92867f,0.91222f,0.88643f,0.87139f,0.84695f,0.83040f,0.79788f,0.76420f,0.74308f,0.72515f,0.70944f,0.70145f,0.70638f,0.71418f,0.72988f,0.74089f,0.75656f,0.76806f,0.78335f,0.78886f,0.80339f,0.81092f,0.81508f,0.82047f,0.82837f,0.83526f,0.84140f,0.84844f,0.85888f,0.87129f,0.88951f,0.97500f,0.00000f,0.90583f,0.90127f,0.87198f,0.84892f,0.82794f,0.80111f,0.76517f,0.74069f,0.71147f,0.69720f,0.68765f,0.68288f,0.69109f,0.69732f,0.71896f,0.72730f,0.74467f,0.75604f,0.76897f,0.77922f,0.78341f,0.79043f,0.79415f,0.80011f,0.80534f,0.80786f,0.81229f,0.81804f,0.82818f,0.83839f,0.85442f,0.95077f,0.00000f,0.89535f,0.87624f,0.85281f,0.82072f,0.79025f,0.75857f,0.73216f,0.69901f,0.67572f,0.66891f,0.66845f,0.66572f,0.67473f,0.68573f,0.70246f,0.71167f,0.73019f,0.74244f,0.75485f,0.76184f,0.76452f,0.77438f,0.77551f,0.77960f,0.78162f,0.78236f,0.78784f,0.78980f,0.79855f,0.80355f,0.81468f,0.92142f,0.00000f,0.87296f,0.84478f,0.81803f,0.79225f,0.75399f,0.72646f,0.70107f,0.67319f,0.65259f,0.64496f,0.64716f,0.64881f,0.66246f,0.67558f,0.68642f,0.70572f,0.71703f,0.73072f,0.74368f,0.74600f,0.75062f,0.75942f,0.76319f,0.76086f,0.76004f,0.75772f,0.75923f,0.76305f,0.76726f,0.77119f,0.77578f,0.88284f,0.00000f,0.84281f,0.81718f,0.79021f,0.75694f,0.72397f,0.68924f,0.66270f,0.63886f,0.62558f,0.62003f,0.62340f,0.63452f,0.64559f,0.66188f,0.67887f,0.69292f,0.70381f,0.71519f,0.72770f,0.73087f,0.73547f,0.74120f,0.74098f,0.73999f,0.73984f,0.74096f,0.73766f,0.73477f,0.73925f,0.73616f,0.73230f,0.83936f,0.00000f,0.80660f,0.78713f,0.75454f,0.72476f,0.68467f,0.65642f,0.62548f,0.60664f,0.59922f,0.59996f,0.60132f,0.62072f,0.62679f,0.64544f,0.66107f,0.67285f,0.68724f,0.70403f,0.70761f,0.71496f,0.72344f,0.72619f,0.72466f,0.72271f,0.72174f,0.71711f,0.71378f,0.70569f,0.70876f,0.70327f,0.68892f,0.79212f,0.00000f,0.77559f,0.74822f,0.71655f,0.68285f,0.64684f,0.61953f,0.59324f,0.57843f,0.57320f,0.57411f,0.58474f,0.60293f,0.61302f,0.63323f,0.64735f,0.66536f,0.67445f,0.68852f,0.69565f,0.69793f,0.70432f,0.70799f,0.70833f,0.70839f,0.69823f,0.69775f,0.69382f,0.68489f,0.67771f,0.66669f,0.64786f,0.74460f,0.00000f,0.74301f,0.70864f,0.68189f,0.64183f,0.61055f,0.58225f,0.56912f,0.55149f,0.55117f,0.56160f,0.56837f,0.58514f,0.60038f,0.62051f,0.63363f,0.65042f,0.65920f,0.66800f,0.67719f,0.68457f,0.68978f,0.69265f,0.68911f,0.68740f,0.69010f,0.67833f,0.67635f,0.65975f,0.65264f,0.63132f,0.60554f,0.69510f,0.00000f,0.70963f,0.67534f,0.63811f,0.61178f,0.57784f,0.55163f,0.53441f,0.52300f,0.52544f,0.53704f,0.55125f,0.56952f,0.58320f,0.60043f,0.62053f,0.63028f,0.64027f,0.65347f,0.66154f,0.66789f,0.67595f,0.67307f,0.67693f,0.67378f,0.66818f,0.66322f,0.65133f,0.63600f,0.62457f,0.59777f,0.56077f,0.64547f,0.00000f,0.67108f,0.64002f,0.60068f,0.57869f,0.53993f,0.52270f,0.50333f,0.49967f,0.50872f,0.51960f,0.53583f,0.55324f,0.56704f,0.59060f,0.60095f,0.61277f,0.63017f,0.64110f,0.64647f,0.65669f,0.65810f,0.66552f,0.65862f,0.66159f,0.65749f,0.64389f,0.62898f,0.62295f,0.59458f,0.56741f,0.52213f,0.59718f,
 0.00000f,1.00000f,0.99717f,1.00000f,0.98997f,0.96890f,0.99091f,0.99448f,0.98369f,0.96688f,0.96471f,0.99806f,0.98096f,0.98961f,0.98534f,0.98454f,0.98210f,0.97220f,0.97107f,0.97621f,0.97334f,0.97162f,0.99105f,0.98695f,0.97872f,0.97715f,0.97344f,0.98376f,0.98560f,0.98959f,0.99213f,0.99636f,1.00000f,0.00000f,0.99416f,0.98686f,0.98445f,0.98165f,0.97868f,0.97721f,0.97563f,0.97557f,0.97358f,0.97266f,0.97501f,0.97344f,0.97500f,0.97441f,0.97574f,0.97546f,0.97459f,0.97628f,0.97845f,0.97887f,0.97920f,0.98204f,0.98307f,0.98345f,0.98441f,0.98531f,0.98873f,0.99062f,0.99269f,0.99487f,0.99739f,1.00000f,0.00000f,0.99153f,0.98594f,0.98324f,0.98058f,0.97914f,0.97573f,0.97577f,0.97433f,0.97557f,0.97317f,0.97455f,0.97360f,0.97443f,0.97444f,0.97430f,0.97538f,0.97584f,0.97631f,0.97738f,0.97890f,0.98034f,0.98176f,0.98337f,0.98407f,0.98604f,0.98755f,0.98923f,0.99136f,0.99330f,0.99543f,0.99774f,1.00000f,0.00000f,0.98331f,0.98492f,0.98107f,0.97916f,0.97739f,0.97599f,0.97587f,0.97444f,0.97425f,0.97308f,0.97286f,0.97287f,0.97291f,0.97367f,0.97401f,0.97488f,0.97602f,0.97676f,0.97811f,0.97880f,0.97988f,0.98126f,0.98310f,0.98397f,0.98603f,0.98755f,0.98955f,0.99148f,0.99325f,0.99541f,0.99767f,1.00000f,0.00000f,0.93952f,0.98073f,0.98037f,0.97924f,0.97756f,0.97608f,0.97488f,0.97468f,0.97346f,0.97351f,0.97333f,0.97400f,0.97320f,0.97404f,0.97492f,0.97544f,0.97589f,0.97684f,0.97746f,0.97917f,0.97998f,0.98158f,0.98259f,0.98427f,0.98584f,0.98744f,0.98938f,0.99132f,0.99336f,0.99537f,0.99775f,1.00000f,0.00000f,0.90568f,0.95689f,0.97810f,0.97763f,0.97826f,0.97557f,0.97473f,0.97350f,0.97204f,0.97290f,0.97272f,0.97257f,0.97332f,0.97345f,0.97409f,0.97517f,0.97579f,0.97667f,0.97760f,0.97873f,0.98023f,0.98182f,0.98198f,0.98443f,0.98582f,0.98756f,0.98938f,0.99125f,0.99334f,0.99553f,0.99775f,1.00000f,0.00000f,0.89821f,0.92237f,0.95344f,0.97184f,0.97491f,0.97597f,0.97319f,0.97297f,0.97362f,0.97195f,0.97348f,0.97251f,0.97322f,0.97313f,0.97355f,0.97481f,0.97469f,0.97673f,0.97728f,0.97817f,0.97916f,0.98060f,0.98177f,0.98398f,0.98592f,0.98761f,0.98927f,0.99119f,0.99325f,0.99530f,0.99759f,1.00000f,0.00000f,0.90331f,0.90476f,0.93015f,0.95029f,0.96564f,0.97175f,0.97289f,0.97280f,0.97303f,0.97211f,0.97186f,0.97210f,0.97314f,0.97385f,0.97414f,0.97418f,0.97481f,0.97634f,0.97666f,0.97829f,0.97921f,0.97945f,0.98098f,0.98372f,0.98581f,0.98758f,0.98907f,0.99115f,0.99330f,0.99518f,0.99758f,1.00000f,0.00000f,0.90436f,0.89924f,0.90755f,0.92434f,0.94520f,0.95705f,0.96559f,0.97018f,0.97244f,0.97129f,0.97188f,0.97153f,0.97221f,0.97255f,0.97364f,0.97461f,0.97476f,0.97636f,0.97697f,0.97778f,0.97832f,0.97783f,0.98088f,0.98405f,0.98571f,0.98713f,0.98903f,0.99113f,0.99331f,0.99533f,0.99760f,1.00000f,0.00000f,0.91414f,0.89475f,0.89836f,0.90823f,0.92099f,0.93713f,0.95051f,0.95781f,0.96429f,0.96904f,0.97114f,0.97172f,0.97274f,0.97303f,0.97354f,0.97470f,0.97493f,0.97595f,0.97683f,0.97699f,0.97605f,0.97581f,0.98053f,0.98349f,0.98536f,0.98758f,0.98919f,0.99116f,0.99301f,0.99521f,0.99752f,1.00000f,0.00000f,0.92836f,0.89863f,0.89637f,0.89719f,0.90610f,0.91584f,0.93065f,0.94096f,0.95052f,0.95774f,0.96144f,0.96648f,0.96906f,0.97228f,0.97179f,0.97333f,0.97439f,0.97548f,0.97474f,0.97334f,0.97245f,0.97478f,0.98001f,0.98302f,0.98528f,0.98680f,0.98873f,0.99081f,0.99304f,0.99520f,0.99748f,1.00000f,0.00000f,0.93098f,0.90408f,0.89505f,0.88619f,0.89377f,0.90300f,0.91150f,0.92322f,0.93386f,0.94252f,0.95011f,0.95671f,0.96110f,0.96709f,0.97027f,0.97138f,0.97426f,0.97243f,0.97079f,0.96870f,0.96859f,0.97289f,0.97948f,0.98259f,0.98470f,0.98681f,0.98823f,0.99089f,0.99277f,0.99486f,0.99743f,1.00000f,0.00000f,0.93469f,0.91202f,0.89473f,0.89081f,0.89243f,0.89501f,0.89682f,0.90600f,0.91739f,0.92424f,0.93523f,0.94210f,0.94914f,0.95551f,0.96088f,0.96473f,0.96673f,0.96479f,0.96343f,0.96303f,0.96646f,0.97286f,0.97824f,0.98191f,0.98515f,0.98621f,0.98857f,0.99047f,0.99270f,0.99489f,0.99739f,1.00000f,0.00000f,0.93390f,0.91785f,0.89763f,0.88889f,0.88693f,0.88792f,0.89308f,0.89702f,0.89883f,0.91034f,0.91770f,0.92541f,0.93408f,0.94163f,0.94854f,0.95049f,0.95246f,0.95147f,0.95198f,0.95570f,0.96189f,0.97012f,0.97688f,0.98187f,0.98450f,0.98628f,0.98877f,0.99034f,0.99238f,0.99493f,0.99721f,1.00000f,0.00000f,0.94153f,0.92676f,0.90493f,0.89704f,0.88756f,0.88642f,0.88476f,0.89141f,0.88939f,0.89776f,0.90277f,0.91349f,0.91525f,0.92331f,0.92709f,0.92791f,0.93064f,0.93300f,0.93758f,0.94503f,0.95423f,0.96395f,0.97310f,0.97863f,0.98326f,0.98640f,0.98799f,0.99027f,0.99289f,0.99478f,0.99696f,1.00000f,0.00000f,0.94284f,0.93028f,0.91248f,0.90188f,0.89190f,0.88938f,0.88435f,0.88088f,0.88591f,0.88746f,0.89025f,0.89574f,0.89769f,0.90138f,0.90258f,0.90391f,0.90478f,0.90947f,0.91952f,0.93064f,0.94284f,0.95420f,0.96436f,0.97200f,0.97840f,0.98278f,0.98668f,0.98980f,0.99221f,0.99480f,0.99716f,1.00000f,0.00000f,0.94498f,0.93281f,0.91298f,0.90216f,0.89135f,0.88749f,0.88260f,0.88302f,0.88348f,0.88001f,0.88086f,0.88217f,0.87610f,0.87576f,0.87688f,0.87716f,0.87949f,0.88696f,0.89677f,0.91090f,0.92663f,0.94071f,0.95261f,0.96050f,0.96872f,0.97735f,0.98083f,0.98652f,0.99064f,0.99442f,0.99682f,1.00000f,0.00000f,0.94378f,0.93543f,0.91795f,0.90872f,0.90248f,0.88913f,0.88647f,0.87599f,0.87856f,0.87634f,0.86899f,0.86132f,0.85690f,0.85033f,0.85156f,0.85108f,0.85584f,0.86526f,0.87942f,0.89477f,0.90829f,0.92401f,0.93828f,0.94902f,0.95663f,0.96387f,0.97350f,0.97859f,0.98528f,0.99040f,0.99618f,1.00000f,0.00000f,0.94837f,0.93267f,0.92218f,0.90916f,0.90699f,0.89083f,0.88175f,0.87957f,0.87343f,0.86435f,0.86092f,0.84619f,0.83598f,0.82884f,0.82624f,0.82373f,0.83121f,0.84229f,0.85886f,0.87445f,0.89266f,0.90419f,0.91981f,0.92894f,0.94090f,0.94881f,0.95920f,0.96630f,0.97450f,0.98320f,0.99186f,1.00000f,0.00000f,0.94705f,0.93867f,0.93002f,0.91560f,0.90712f,0.89394f,0.88270f,0.86945f,0.86076f,0.85117f,0.83392f,0.82125f,0.81208f,0.80220f,0.80260f,0.80192f,0.81168f,0.82346f,0.83925f,0.85409f,0.86909f,0.88334f,0.89973f,0.91032f,0.91898f,0.93077f,0.94147f,0.94963f,0.95935f,0.96993f,0.98234f,0.99996f,0.00000f,0.94306f,0.93455f,0.92673f,0.91020f,0.90423f,0.89337f,0.87984f,0.85675f,0.84822f,0.82479f,0.80482f,0.79921f,0.78527f,0.77968f,0.78269f,0.78578f,0.79612f,0.80268f,0.82322f,0.83680f,0.85297f,0.86413f,0.87674f,0.88853f,0.89756f,0.90719f,0.91784f,0.92859f,0.93919f,0.95219f,0.96630f,0.99963f,0.00000f,0.93790f,0.93690f,0.92521f,0.91411f,0.89951f,0.87797f,0.86515f,0.84726f,0.82611f,0.80091f,0.78836f,0.77462f,0.76286f,0.76376f,0.76484f,0.76841f,0.77888f,0.79132f,0.80453f,0.81853f,0.83098f,0.84432f,0.85647f,0.86506f,0.87793f,0.88215f,0.89468f,0.90363f,0.91420f,0.92824f,0.94287f,0.99731f,0.00000f,0.94371f,0.93019f,0.92211f,0.90455f,0.89119f,0.86723f,0.84993f,0.82720f,0.79926f,0.77938f,0.75957f,0.74883f,0.74239f,0.74084f,0.74836f,0.75415f,0.76546f,0.78064f,0.78992f,0.80078f,0.81450f,0.82691f,0.83437f,0.84241f,0.84907f,0.85676f,0.86640f,0.87660f,0.88604f,0.89690f,0.91388f,0.98547f,0.00000f,0.94206f,0.93208f,0.91221f,0.89310f,0.86973f,0.85073f,0.82431f,0.80134f,0.77792f,0.75026f,0.73037f,0.72708f,0.72368f,0.72175f,0.72668f,0.74214f,0.75197f,0.76468f,0.77781f,0.78830f,0.79794f,0.80334f,0.81565f,0.81657f,0.82431f,0.83276f,0.84102f,0.84468f,0.85510f,0.86616f,0.88113f,0.96451f,0.00000f,0.92883f,0.91750f,0.90028f,0.87636f,0.85412f,0.82258f,0.79648f,0.76883f,0.74838f,0.72672f,0.70911f,0.70465f,0.70249f,0.70730f,0.71666f,0.72608f,0.74071f,0.75286f,0.76245f,0.77378f,0.77983f,0.78916f,0.79080f,0.79797f,0.80110f,0.80313f,0.80844f,0.81487f,0.82127f,0.83078f,0.83990f,0.93522f,0.00000f,0.91518f,0.89705f,0.87713f,0.85582f,0.82355f,0.79212f,0.76363f,0.73898f,0.71673f,0.69677f,0.68993f,0.68530f,0.68608f,0.69104f,0.70356f,0.71645f,0.72342f,0.73834f,0.75195f,0.75874f,0.76140f,0.76824f,0.77407f,0.77632f,0.77499f,0.78061f,0.78186f,0.78430f,0.78987f,0.79271f,0.80025f,0.89794f,0.00000f,0.90362f,0.87356f,0.85270f,0.82095f,0.79492f,0.76703f,0.73377f,0.70974f,0.68654f,0.66853f,0.65992f,0.66543f,0.66720f,0.68019f,0.68997f,0.70349f,0.71130f,0.72427f,0.72825f,0.74468f,0.74601f,0.75536f,0.75514f,0.75472f,0.75330f,0.75144f,0.75472f,0.75767f,0.75547f,0.75613f,0.75155f,0.85638f,0.00000f,0.87671f,0.84836f,0.82250f,0.79444f,0.76283f,0.73614f,0.70143f,0.68286f,0.65528f,0.64811f,0.64455f,0.64095f,0.65321f,0.66477f,0.67580f,0.68832f,0.69908f,0.71098f,0.71911f,0.72520f,0.73027f,0.73578f,0.73473f,0.73589f,0.73301f,0.72887f,0.72785f,0.72539f,0.72361f,0.71810f,0.70736f,0.80651f,0.00000f,0.84702f,0.81832f,0.78472f,0.76271f,0.73227f,0.70015f,0.66447f,0.64216f,0.62702f,0.62164f,0.61957f,0.62633f,0.63486f,0.64896f,0.66657f,0.67548f,0.68646f,0.69866f,0.70365f,0.71016f,0.71401f,0.71776f,0.71963f,0.71021f,0.71171f,0.70361f,0.70463f,0.69572f,0.68968f,0.67783f,0.65776f,0.75743f,0.00000f,0.81378f,0.79199f,0.75566f,0.72913f,0.69336f,0.66858f,0.63365f,0.61269f,0.60250f,0.59936f,0.60174f,0.60608f,0.61806f,0.63011f,0.64739f,0.66090f,0.67096f,0.68294f,0.68719f,0.69289f,0.69712f,0.70115f,0.69842f,0.69468f,0.69136f,0.68590f,0.67886f,0.66868f,0.65589f,0.64001f,0.61298f,0.70574f,0.00000f,0.78333f,0.76104f,0.72435f,0.69274f,0.65841f,0.62440f,0.60405f,0.58693f,0.57861f,0.57763f,0.58437f,0.59147f,0.60440f,0.62482f,0.63750f,0.64676f,0.65222f,0.66473f,0.67160f,0.67613f,0.68155f,0.68582f,0.68189f,0.68049f,0.67166f,0.66246f,0.65185f,0.64523f,0.62588f,0.60451f,0.56631f,0.65568f,0.00000f,0.75329f,0.72000f,0.68695f,0.65577f,0.62555f,0.59246f,0.56982f,0.55242f,0.55356f,0.55245f,0.56453f,0.57335f,0.59054f,0.60571f,0.61506f,0.63538f,0.64004f,0.64733f,0.65987f,0.66394f,0.66582f,0.66303f,0.66593f,0.66351f,0.65479f,0.64592f,0.63394f,0.61892f,0.59682f,0.56465f,0.52362f,0.60545f,0.00000f,0.71912f,0.69357f,0.65616f,0.62142f,0.59109f,0.56285f,0.53814f,0.53167f,0.52783f,0.53753f,0.54676f,0.56312f,0.57121f,0.58704f,0.59916f,0.61241f,0.63026f,0.63498f,0.64131f,0.64880f,0.65028f,0.65111f,0.65020f,0.64211f,0.63633f,0.62900f,0.61111f,0.59154f,0.56220f,0.53241f,0.48371f,0.55760f,
 0.00000f,1.00000f,0.99820f,1.00000f,0.98922f,0.96760f,0.99013f,0.99359f,0.98431f,0.96605f,0.96486f,0.99864f,0.98273f,0.98873f,0.98630f,0.98413f,0.98259f,0.97275f,0.97115f,0.97617f,0.97338f,0.97157f,0.99109f,0.98756f,0.98600f,0.97990f,0.97935f,0.97959f,0.98749f,0.99042f,0.99352f,0.99624f,1.00000f,0.00000f,0.99268f,0.98746f,0.98313f,0.98168f,0.97731f,0.97727f,0.97618f,0.97565f,0.97258f,0.97356f,0.97447f,0.97412f,0.97485f,0.97397f,0.97573f,0.97574f,0.97558f,0.97619f,0.97810f,0.97852f,0.97907f,0.98198f,0.98217f,0.98433f,0.98478f,0.98637f,0.98798f,0.99082f,0.99297f,0.99509f,0.99752f,1.00000f,0.00000f,0.99008f,0.98460f,0.98316f,0.98119f,0.97743f,0.97644f,0.97619f,0.97513f,0.97384f,0.97335f,0.97339f,0.97332f,0.97443f,0.97442f,0.97389f,0.97471f,0.97623f,0.97703f,0.97732f,0.97893f,0.98012f,0.98132f,0.98295f,0.98454f,0.98603f,0.98735f,0.98926f,0.99122f,0.99333f,0.99541f,0.99770f,1.00000f,0.00000f,0.98386f,0.98645f,0.98102f,0.97934f,0.97749f,0.97693f,0.97452f,0.97462f,0.97420f,0.97409f,0.97378f,0.97292f,0.97422f,0.97390f,0.97433f,0.97518f,0.97622f,0.97727f,0.97738f,0.97892f,0.98033f,0.98124f,0.98261f,0.98404f,0.98587f,0.98753f,0.98936f,0.99119f,0.99332f,0.99553f,0.99776f,1.00000f,0.00000f,0.94146f,0.98174f,0.98090f,0.97939f,0.97765f,0.97488f,0.97487f,0.97427f,0.97306f,0.97365f,0.97326f,0.97298f,0.97286f,0.97386f,0.97406f,0.97524f,0.97558f,0.97715f,0.97787f,0.97806f,0.97984f,0.98090f,0.98238f,0.98440f,0.98582f,0.98761f,0.98953f,0.99139f,0.99339f,0.99551f,0.99773f,1.00000f,0.00000f,0.91041f,0.95712f,0.97660f,0.97866f,0.97705f,0.97380f,0.97466f,0.97361f,0.97282f,0.97312f,0.97331f,0.97316f,0.97455f,0.97385f,0.97413f,0.97457f,0.97603f,0.97650f,0.97728f,0.97858f,0.97982f,0.98052f,0.98273f,0.98390f,0.98563f,0.98727f,0.98944f,0.99140f,0.99330f,0.99550f,0.99771f,1.00000f,0.00000f,0.89852f,0.92482f,0.95355f,0.97253f,0.97323f,0.97578f,0.97428f,0.97398f,0.97286f,0.97223f,0.97261f,0.97273f,0.97314f,0.97395f,0.97444f,0.97384f,0.97525f,0.97699f,0.97644f,0.97841f,0.97951f,0.98131f,0.98210f,0.98333f,0.98570f,0.98749f,0.98915f,0.99136f,0.99340f,0.99541f,0.99760f,1.00000f,0.00000f,0.90210f,0.90743f,0.92597f,0.95069f,0.96490f,0.97047f,0.97198f,0.97273f,0.97254f,0.97269f,0.97230f,0.97197f,0.97310f,0.97381f,0.97444f,0.97392f,0.97505f,0.97587f,0.97719f,0.97849f,0.97967f,0.98049f,0.98040f,0.98368f,0.98582f,0.98747f,0.98921f,0.99115f,0.99310f,0.99537f,0.99751f,1.00000f,0.00000f,0.91178f,0.89956f,0.90661f,0.92380f,0.94310f,0.95804f,0.96479f,0.96955f,0.97130f,0.97243f,0.97258f,0.97199f,0.97177f,0.97257f,0.97210f,0.97418f,0.97526f,0.97591f,0.97630f,0.97775f,0.97935f,0.97944f,0.97934f,0.98264f,0.98526f,0.98720f,0.98935f,0.99103f,0.99312f,0.99526f,0.99758f,1.00000f,0.00000f,0.92088f,0.89365f,0.89758f,0.90940f,0.92114f,0.93701f,0.94841f,0.95792f,0.96376f,0.96927f,0.97081f,0.97133f,0.97215f,0.97229f,0.97374f,0.97412f,0.97503f,0.97579f,0.97691f,0.97808f,0.97784f,0.97671f,0.97868f,0.98226f,0.98548f,0.98723f,0.98911f,0.99111f,0.99297f,0.99521f,0.99757f,1.00000f,0.00000f,0.92343f,0.89706f,0.89323f,0.89697f,0.90466f,0.91542f,0.93045f,0.94218f,0.95090f,0.95738f,0.96305f,0.96631f,0.96806f,0.97010f,0.97298f,0.97391f,0.97460f,0.97588f,0.97667f,0.97553f,0.97396f,0.97459f,0.97765f,0.98238f,0.98510f,0.98711f,0.98904f,0.99081f,0.99301f,0.99522f,0.99738f,1.00000f,0.00000f,0.93063f,0.90540f,0.89341f,0.89284f,0.89766f,0.90535f,0.91452f,0.92156f,0.93195f,0.94165f,0.95269f,0.95618f,0.96309f,0.96557f,0.96826f,0.97201f,0.97350f,0.97459f,0.97363f,0.97180f,0.97060f,0.97220f,0.97674f,0.98125f,0.98480f,0.98682f,0.98885f,0.99046f,0.99287f,0.99500f,0.99741f,1.00000f,0.00000f,0.93166f,0.91417f,0.90082f,0.88888f,0.89117f,0.89448f,0.89706f,0.90650f,0.91611f,0.92717f,0.93745f,0.94142f,0.95026f,0.95539f,0.96260f,0.96715f,0.96784f,0.96920f,0.96672f,0.96627f,0.96633f,0.97044f,0.97636f,0.98090f,0.98439f,0.98661f,0.98870f,0.99016f,0.99277f,0.99488f,0.99747f,1.00000f,0.00000f,0.93621f,0.91596f,0.89581f,0.88706f,0.89128f,0.88867f,0.89362f,0.89668f,0.90342f,0.90895f,0.91642f,0.92612f,0.93483f,0.94127f,0.94875f,0.95259f,0.95719f,0.95659f,0.95582f,0.95450f,0.96117f,0.96656f,0.97382f,0.98058f,0.98353f,0.98638f,0.98805f,0.99055f,0.99282f,0.99484f,0.99720f,1.00000f,0.00000f,0.94330f,0.92190f,0.90733f,0.89459f,0.88999f,0.88739f,0.88735f,0.88803f,0.89116f,0.89735f,0.90349f,0.91267f,0.91789f,0.92650f,0.93070f,0.93634f,0.93723f,0.93656f,0.94111f,0.94290f,0.95265f,0.96094f,0.97008f,0.97845f,0.98246f,0.98502f,0.98778f,0.99013f,0.99250f,0.99473f,0.99683f,1.00000f,0.00000f,0.94245f,0.92569f,0.90590f,0.89818f,0.89072f,0.88081f,0.88319f,0.88419f,0.88356f,0.88763f,0.89368f,0.89398f,0.90174f,0.90730f,0.90757f,0.91125f,0.91246f,0.91637f,0.91972f,0.92688f,0.94021f,0.95087f,0.96294f,0.97141f,0.97794f,0.98286f,0.98609f,0.98938f,0.99206f,0.99454f,0.99702f,1.00000f,0.00000f,0.94994f,0.92909f,0.91728f,0.90339f,0.89508f,0.89083f,0.88860f,0.88417f,0.88222f,0.88130f,0.88021f,0.88255f,0.88403f,0.88518f,0.88487f,0.88678f,0.88453f,0.88905f,0.90066f,0.90838f,0.92364f,0.93674f,0.94920f,0.96149f,0.96805f,0.97746f,0.98124f,0.98684f,0.99037f,0.99400f,0.99668f,1.00000f,0.00000f,0.94600f,0.93045f,0.92088f,0.90631f,0.90005f,0.88407f,0.87917f,0.88500f,0.87754f,0.87779f,0.87373f,0.87190f,0.86561f,0.86342f,0.85858f,0.85960f,0.85983f,0.86823f,0.87737f,0.89052f,0.90684f,0.92010f,0.93573f,0.94703f,0.95731f,0.96324f,0.97254f,0.97882f,0.98525f,0.99142f,0.99578f,1.00000f,0.00000f,0.94624f,0.93477f,0.92297f,0.91329f,0.90133f,0.89525f,0.88816f,0.87706f,0.87337f,0.87286f,0.86041f,0.85699f,0.84897f,0.84185f,0.83709f,0.83221f,0.83477f,0.84052f,0.85830f,0.86814f,0.88564f,0.89998f,0.91482f,0.92941f,0.94039f,0.94825f,0.95675f,0.96630f,0.97529f,0.98468f,0.99226f,1.00000f,0.00000f,0.94620f,0.94314f,0.92768f,0.91479f,0.90833f,0.89943f,0.88466f,0.87850f,0.86693f,0.85700f,0.84574f,0.83376f,0.82338f,0.81663f,0.81377f,0.81375f,0.81179f,0.82415f,0.83533f,0.85094f,0.86591f,0.88186f,0.89497f,0.90873f,0.92030f,0.93029f,0.94080f,0.95062f,0.95984f,0.96959f,0.98096f,0.99994f,0.00000f,0.94092f,0.93860f,0.92648f,0.91557f,0.90889f,0.89207f,0.88168f,0.86686f,0.85761f,0.84422f,0.83086f,0.81382f,0.80395f,0.79292f,0.79184f,0.79177f,0.79558f,0.80330f,0.81711f,0.83249f,0.84725f,0.85963f,0.87392f,0.88546f,0.89787f,0.91020f,0.91634f,0.92792f,0.94007f,0.95048f,0.96386f,0.99929f,0.00000f,0.94281f,0.93763f,0.93274f,0.91482f,0.91225f,0.89447f,0.87160f,0.85780f,0.84156f,0.82382f,0.80643f,0.79188f,0.78044f,0.77146f,0.77136f,0.77131f,0.78280f,0.79252f,0.80312f,0.81520f,0.82719f,0.84006f,0.85025f,0.86346f,0.87549f,0.88099f,0.89107f,0.90356f,0.91529f,0.92482f,0.93996f,0.99428f,0.00000f,0.95397f,0.92731f,0.92685f,0.91447f,0.89972f,0.87648f,0.86164f,0.84641f,0.82474f,0.79995f,0.77957f,0.76567f,0.75564f,0.75333f,0.75067f,0.75986f,0.76420f,0.77644f,0.78942f,0.79978f,0.80952f,0.81985f,0.82998f,0.84208f,0.85009f,0.85693f,0.86465f,0.87183f,0.88546f,0.89460f,0.91135f,0.98077f,0.00000f,0.94033f,0.93145f,0.92188f,0.90275f,0.88232f,0.86535f,0.84277f,0.82890f,0.80281f,0.77578f,0.75622f,0.74536f,0.73718f,0.73482f,0.73606f,0.74455f,0.75226f,0.76378f,0.77591f,0.78330f,0.79438f,0.80242f,0.80929f,0.81544f,0.82418f,0.82904f,0.83760f,0.84430f,0.85145f,0.86342f,0.87056f,0.95638f,0.00000f,0.92783f,0.92534f,0.91462f,0.89485f,0.87744f,0.85056f,0.82628f,0.79737f,0.77121f,0.75152f,0.73534f,0.72413f,0.71666f,0.71775f,0.72234f,0.73034f,0.73806f,0.74888f,0.76378f,0.76325f,0.77800f,0.78149f,0.79231f,0.79579f,0.79984f,0.80156f,0.80535f,0.81187f,0.81555f,0.82354f,0.83036f,0.92415f,0.00000f,0.92543f,0.90326f,0.89516f,0.87552f,0.85321f,0.82322f,0.79801f,0.76625f,0.74279f,0.72221f,0.71204f,0.69903f,0.69939f,0.69902f,0.70981f,0.71224f,0.73114f,0.73482f,0.74328f,0.75462f,0.75946f,0.76498f,0.77088f,0.77111f,0.77443f,0.77498f,0.77724f,0.77938f,0.78138f,0.78505f,0.78521f,0.88502f,0.00000f,0.91602f,0.89867f,0.88247f,0.85131f,0.81708f,0.80401f,0.76965f,0.73968f,0.71442f,0.69796f,0.68553f,0.67499f,0.68018f,0.68574f,0.69498f,0.70437f,0.71301f,0.72603f,0.72701f,0.73869f,0.74396f,0.74697f,0.74930f,0.75181f,0.75231f,0.75090f,0.74708f,0.74994f,0.74611f,0.74733f,0.73552f,0.83705f,0.00000f,0.88818f,0.87307f,0.85147f,0.82972f,0.79817f,0.76929f,0.73761f,0.71087f,0.68819f,0.67454f,0.65769f,0.66397f,0.66414f,0.66757f,0.67967f,0.68769f,0.69275f,0.70746f,0.71745f,0.72350f,0.72834f,0.73285f,0.73136f,0.73192f,0.73052f,0.72278f,0.71974f,0.71854f,0.71381f,0.70091f,0.68716f,0.78756f,0.00000f,0.87450f,0.84981f,0.82478f,0.79209f,0.76460f,0.73178f,0.70694f,0.67807f,0.65850f,0.64145f,0.63622f,0.64131f,0.64717f,0.65147f,0.66558f,0.67704f,0.68590f,0.69605f,0.69860f,0.70689f,0.70881f,0.71330f,0.71186f,0.70944f,0.70628f,0.69733f,0.69327f,0.68842f,0.67503f,0.66293f,0.63465f,0.73385f,0.00000f,0.84408f,0.81781f,0.79160f,0.75564f,0.72481f,0.70430f,0.66956f,0.64750f,0.62472f,0.61730f,0.61986f,0.62477f,0.62745f,0.64062f,0.65293f,0.66229f,0.67192f,0.67769f,0.68435f,0.69288f,0.69637f,0.69141f,0.69404f,0.68702f,0.68453f,0.68008f,0.66880f,0.65543f,0.64195f,0.61951f,0.58759f,0.68159f,0.00000f,0.81814f,0.78558f,0.76475f,0.73302f,0.69946f,0.66612f,0.64286f,0.61476f,0.61062f,0.59883f,0.59796f,0.60333f,0.60977f,0.62001f,0.63717f,0.64812f,0.65393f,0.66161f,0.67511f,0.67500f,0.67107f,0.67979f,0.67596f,0.67063f,0.66750f,0.65422f,0.63795f,0.62677f,0.60317f,0.57889f,0.53972f,0.62763f,0.00000f,0.79262f,0.75250f,0.73498f,0.70493f,0.66139f,0.63238f,0.60425f,0.58734f,0.57420f,0.57752f,0.57782f,0.58711f,0.59510f,0.60738f,0.61844f,0.63028f,0.63864f,0.64498f,0.65504f,0.65923f,0.65966f,0.65810f,0.66101f,0.65362f,0.64394f,0.63113f,0.61714f,0.60044f,0.57395f,0.53574f,0.49401f,0.57759f,0.00000f,0.76027f,0.73024f,0.69494f,0.66434f,0.62944f,0.60141f,0.57430f,0.55820f,0.55368f,0.54937f,0.55767f,0.56833f,0.57769f,0.59266f,0.60249f,0.61142f,0.62189f,0.62659f,0.63388f,0.64058f,0.64307f,0.64326f,0.63503f,0.63136f,0.62253f,0.61093f,0.59083f,0.57217f,0.54284f,0.50114f,0.45661f,0.52837f,
 0.00000f,1.00000f,0.99842f,1.00000f,0.99087f,0.96976f,0.99048f,0.99342f,0.98388f,0.96682f,0.96430f,0.99805f,0.98159f,0.98781f,0.98601f,0.98467f,0.98231f,0.97181f,0.97149f,0.97674f,0.97279f,0.97184f,0.99150f,0.98739f,0.98579f,0.98683f,0.98175f,0.98276f,0.98451f,0.98758f,0.99299f,0.99670f,1.00000f,0.00000f,0.99252f,0.98814f,0.98422f,0.98078f,0.97757f,0.97696f,0.97602f,0.97535f,0.97372f,0.97414f,0.97372f,0.97388f,0.97475f,0.97515f,0.97567f,0.97516f,0.97538f,0.97637f,0.97827f,0.97874f,0.97893f,0.98161f,0.98312f,0.98414f,0.98584f,0.98685f,0.98814f,0.99034f,0.99244f,0.99505f,0.99757f,1.00000f,0.00000f,0.98959f,0.98789f,0.98197f,0.97914f,0.97826f,0.97687f,0.97629f,0.97482f,0.97375f,0.97295f,0.97338f,0.97318f,0.97341f,0.97431f,0.97439f,0.97575f,0.97582f,0.97667f,0.97784f,0.97888f,0.98002f,0.98095f,0.98271f,0.98444f,0.98576f,0.98751f,0.98942f,0.99120f,0.99333f,0.99552f,0.99773f,1.00000f,0.00000f,0.98249f,0.98611f,0.98232f,0.98055f,0.97824f,0.97690f,0.97457f,0.97428f,0.97322f,0.97365f,0.97281f,0.97311f,0.97282f,0.97460f,0.97394f,0.97485f,0.97656f,0.97636f,0.97775f,0.97819f,0.98035f,0.98147f,0.98274f,0.98415f,0.98586f,0.98763f,0.98928f,0.99134f,0.99319f,0.99537f,0.99778f,1.00000f,0.00000f,0.93754f,0.98075f,0.98211f,0.97914f,0.97880f,0.97621f,0.97372f,0.97510f,0.97381f,0.97330f,0.97274f,0.97250f,0.97313f,0.97344f,0.97407f,0.97554f,0.97591f,0.97633f,0.97748f,0.97820f,0.98024f,0.98121f,0.98272f,0.98429f,0.98598f,0.98760f,0.98914f,0.99138f,0.99332f,0.99543f,0.99778f,1.00000f,0.00000f,0.90897f,0.95682f,0.97548f,0.97846f,0.97736f,0.97683f,0.97475f,0.97480f,0.97335f,0.97363f,0.97352f,0.97262f,0.97314f,0.97423f,0.97433f,0.97456f,0.97523f,0.97610f,0.97771f,0.97886f,0.97999f,0.98141f,0.98265f,0.98418f,0.98575f,0.98755f,0.98952f,0.99153f,0.99322f,0.99540f,0.99776f,1.00000f,0.00000f,0.90063f,0.92295f,0.95339f,0.97117f,0.97604f,0.97594f,0.97307f,0.97301f,0.97337f,0.97201f,0.97272f,0.97300f,0.97337f,0.97423f,0.97388f,0.97460f,0.97571f,0.97627f,0.97731f,0.97845f,0.98002f,0.98085f,0.98264f,0.98324f,0.98566f,0.98760f,0.98910f,0.99111f,0.99327f,0.99549f,0.99764f,1.00000f,0.00000f,0.90793f,0.90591f,0.92830f,0.94718f,0.96379f,0.97023f,0.97344f,0.97298f,0.97383f,0.97213f,0.97142f,0.97271f,0.97258f,0.97279f,0.97389f,0.97496f,0.97504f,0.97565f,0.97754f,0.97770f,0.97978f,0.98144f,0.98227f,0.98251f,0.98541f,0.98744f,0.98919f,0.99084f,0.99328f,0.99529f,0.99751f,1.00000f,0.00000f,0.91388f,0.89490f,0.90891f,0.92427f,0.94224f,0.95662f,0.96620f,0.96976f,0.97203f,0.97260f,0.97261f,0.97280f,0.97173f,0.97239f,0.97288f,0.97457f,0.97568f,0.97604f,0.97715f,0.97861f,0.98036f,0.98046f,0.98027f,0.98155f,0.98510f,0.98726f,0.98912f,0.99108f,0.99320f,0.99532f,0.99755f,1.00000f,0.00000f,0.91294f,0.89429f,0.89587f,0.90904f,0.92350f,0.93530f,0.95111f,0.95866f,0.96387f,0.96779f,0.96983f,0.97195f,0.97179f,0.97164f,0.97292f,0.97335f,0.97446f,0.97536f,0.97707f,0.97798f,0.97950f,0.97899f,0.97809f,0.98140f,0.98499f,0.98687f,0.98898f,0.99127f,0.99316f,0.99514f,0.99753f,1.00000f,0.00000f,0.91914f,0.90117f,0.89196f,0.89578f,0.90454f,0.92039f,0.92909f,0.94086f,0.95069f,0.95792f,0.96304f,0.96623f,0.96957f,0.97090f,0.97170f,0.97352f,0.97355f,0.97569f,0.97722f,0.97721f,0.97699f,0.97487f,0.97627f,0.98033f,0.98482f,0.98720f,0.98863f,0.99128f,0.99283f,0.99511f,0.99749f,1.00000f,0.00000f,0.92803f,0.90610f,0.89266f,0.89435f,0.89433f,0.90209f,0.91099f,0.91928f,0.93280f,0.94301f,0.94958f,0.95668f,0.96158f,0.96596f,0.96915f,0.97145f,0.97358f,0.97505f,0.97540f,0.97498f,0.97305f,0.97174f,0.97478f,0.97937f,0.98402f,0.98659f,0.98859f,0.99058f,0.99296f,0.99512f,0.99735f,1.00000f,0.00000f,0.93409f,0.90390f,0.89964f,0.88847f,0.88687f,0.89351f,0.89818f,0.90914f,0.91298f,0.92556f,0.93347f,0.94143f,0.94884f,0.95552f,0.96114f,0.96611f,0.96894f,0.97140f,0.97100f,0.96862f,0.96845f,0.96830f,0.97343f,0.97939f,0.98373f,0.98668f,0.98854f,0.99049f,0.99251f,0.99496f,0.99734f,1.00000f,0.00000f,0.93761f,0.91513f,0.90100f,0.89476f,0.89333f,0.88558f,0.88854f,0.89817f,0.90112f,0.91066f,0.91698f,0.92517f,0.93445f,0.94247f,0.94942f,0.95546f,0.95840f,0.96107f,0.95999f,0.96050f,0.96183f,0.96557f,0.97170f,0.97795f,0.98279f,0.98647f,0.98874f,0.99001f,0.99269f,0.99498f,0.99713f,1.00000f,0.00000f,0.94177f,0.91624f,0.90655f,0.89328f,0.88825f,0.88781f,0.88309f,0.89108f,0.89103f,0.89714f,0.90481f,0.90813f,0.91700f,0.92701f,0.93159f,0.93769f,0.94129f,0.94413f,0.94413f,0.94645f,0.95248f,0.95801f,0.96784f,0.97518f,0.98169f,0.98493f,0.98739f,0.99052f,0.99250f,0.99465f,0.99692f,1.00000f,0.00000f,0.93899f,0.92520f,0.90885f,0.89497f,0.88896f,0.87999f,0.88424f,0.88615f,0.88982f,0.88835f,0.89519f,0.89527f,0.90489f,0.90962f,0.91581f,0.91604f,0.91980f,0.92090f,0.92432f,0.93005f,0.93692f,0.94832f,0.95929f,0.96862f,0.97659f,0.98222f,0.98602f,0.98980f,0.99235f,0.99450f,0.99700f,1.00000f,0.00000f,0.94773f,0.93231f,0.91390f,0.90470f,0.89710f,0.88473f,0.88362f,0.88089f,0.88467f,0.88049f,0.88497f,0.88654f,0.88683f,0.88775f,0.88953f,0.89177f,0.89515f,0.89644f,0.90282f,0.90984f,0.92122f,0.93570f,0.94678f,0.95833f,0.96788f,0.97546f,0.98173f,0.98585f,0.99056f,0.99438f,0.99709f,1.00000f,0.00000f,0.94197f,0.93354f,0.91994f,0.90986f,0.89718f,0.89561f,0.88846f,0.88160f,0.87704f,0.87942f,0.87615f,0.87431f,0.87570f,0.87197f,0.86918f,0.86968f,0.86856f,0.87552f,0.87774f,0.88834f,0.90086f,0.91715f,0.92879f,0.94534f,0.95523f,0.96367f,0.97177f,0.97810f,0.98460f,0.99089f,0.99611f,1.00000f,0.00000f,0.95476f,0.94043f,0.92358f,0.90917f,0.89916f,0.89781f,0.88506f,0.88086f,0.87725f,0.87097f,0.86568f,0.86098f,0.85827f,0.84770f,0.84477f,0.84320f,0.84054f,0.84753f,0.85730f,0.86866f,0.88387f,0.89655f,0.91178f,0.92665f,0.93684f,0.94821f,0.95837f,0.96661f,0.97489f,0.98317f,0.99178f,1.00000f,0.00000f,0.94893f,0.93374f,0.93189f,0.92218f,0.90547f,0.89536f,0.89211f,0.88197f,0.87339f,0.86445f,0.85445f,0.84771f,0.83564f,0.82693f,0.82629f,0.81698f,0.82043f,0.82503f,0.83673f,0.84811f,0.86621f,0.87937f,0.89156f,0.90579f,0.91944f,0.92876f,0.93957f,0.94978f,0.96029f,0.97047f,0.98112f,0.99990f,0.00000f,0.94659f,0.93965f,0.93448f,0.92016f,0.90664f,0.89232f,0.87789f,0.87890f,0.86617f,0.85142f,0.83929f,0.82826f,0.81606f,0.80800f,0.80178f,0.79781f,0.79822f,0.80427f,0.81684f,0.83109f,0.84168f,0.85649f,0.86842f,0.88375f,0.89674f,0.90655f,0.91691f,0.92825f,0.94059f,0.95141f,0.96438f,0.99844f,0.00000f,0.94909f,0.94037f,0.93105f,0.91912f,0.90483f,0.89493f,0.88357f,0.86814f,0.85857f,0.83555f,0.82412f,0.80606f,0.79452f,0.78354f,0.78007f,0.78055f,0.78401f,0.79136f,0.80113f,0.81259f,0.82367f,0.83622f,0.84637f,0.86328f,0.87321f,0.88265f,0.89154f,0.90123f,0.91400f,0.92450f,0.93916f,0.99224f,0.00000f,0.94992f,0.94198f,0.92718f,0.91893f,0.90912f,0.89065f,0.87333f,0.85863f,0.84224f,0.81838f,0.80066f,0.78570f,0.77354f,0.76267f,0.76162f,0.76524f,0.76679f,0.77647f,0.78587f,0.79560f,0.80648f,0.81869f,0.83020f,0.83947f,0.84738f,0.85574f,0.86335f,0.87398f,0.88423f,0.89390f,0.90753f,0.97684f,0.00000f,0.94408f,0.92983f,0.92413f,0.91033f,0.89975f,0.88121f,0.85919f,0.84206f,0.82477f,0.79782f,0.77861f,0.76218f,0.74590f,0.74395f,0.73910f,0.74903f,0.75238f,0.75968f,0.77290f,0.78306f,0.79139f,0.79632f,0.80736f,0.81495f,0.82282f,0.82929f,0.83660f,0.84027f,0.85040f,0.86025f,0.86951f,0.95081f,0.00000f,0.94059f,0.93190f,0.91707f,0.90126f,0.88467f,0.86346f,0.84134f,0.81860f,0.80069f,0.77246f,0.75416f,0.74073f,0.72881f,0.73069f,0.72855f,0.73252f,0.74240f,0.75332f,0.75915f,0.76909f,0.77554f,0.77998f,0.78869f,0.79241f,0.79826f,0.80086f,0.80646f,0.80869f,0.81644f,0.82015f,0.82255f,0.91634f,0.00000f,0.93150f,0.92368f,0.90507f,0.88378f,0.86953f,0.84466f,0.82261f,0.78874f,0.77240f,0.74777f,0.73252f,0.72042f,0.71084f,0.71160f,0.70812f,0.71551f,0.72754f,0.73662f,0.74121f,0.75229f,0.75692f,0.76510f,0.76688f,0.77192f,0.77323f,0.77753f,0.77694f,0.77837f,0.77672f,0.77565f,0.77567f,0.87464f,0.00000f,0.92341f,0.91446f,0.89651f,0.87055f,0.84602f,0.81997f,0.79180f,0.76809f,0.74183f,0.72153f,0.70571f,0.69909f,0.69339f,0.68965f,0.70257f,0.70556f,0.71856f,0.72129f,0.72918f,0.73517f,0.74309f,0.74566f,0.74463f,0.75008f,0.75018f,0.74840f,0.74612f,0.74467f,0.74072f,0.73299f,0.72402f,0.82396f,0.00000f,0.91117f,0.88960f,0.87061f,0.84270f,0.82063f,0.79249f,0.76577f,0.73774f,0.71541f,0.69686f,0.68291f,0.67596f,0.67137f,0.67393f,0.68788f,0.68863f,0.70040f,0.70540f,0.71499f,0.71927f,0.72646f,0.72806f,0.72964f,0.72652f,0.72512f,0.71730f,0.71250f,0.71261f,0.70236f,0.69055f,0.67079f,0.77207f,0.00000f,0.88800f,0.87384f,0.85316f,0.83211f,0.79824f,0.76480f,0.73095f,0.70663f,0.68308f,0.66699f,0.65836f,0.65498f,0.65682f,0.66044f,0.66980f,0.67739f,0.68404f,0.69213f,0.69866f,0.70512f,0.70558f,0.70722f,0.70889f,0.70596f,0.69842f,0.69265f,0.68495f,0.67715f,0.66210f,0.64602f,0.61686f,0.71826f,0.00000f,0.86301f,0.85165f,0.82369f,0.78864f,0.76349f,0.73528f,0.69902f,0.67865f,0.64939f,0.63869f,0.63555f,0.63071f,0.64008f,0.64504f,0.65617f,0.66154f,0.67293f,0.68009f,0.68555f,0.68800f,0.69029f,0.69168f,0.68887f,0.68438f,0.67721f,0.66593f,0.65602f,0.64188f,0.62614f,0.59966f,0.56841f,0.66298f,0.00000f,0.84195f,0.82162f,0.79778f,0.76017f,0.72679f,0.69788f,0.67075f,0.64666f,0.63052f,0.61767f,0.61214f,0.61690f,0.61895f,0.63028f,0.64151f,0.64774f,0.65702f,0.66357f,0.66539f,0.67061f,0.66982f,0.67236f,0.66866f,0.66208f,0.65714f,0.64393f,0.62647f,0.61193f,0.58860f,0.55761f,0.51590f,0.61324f,0.00000f,0.81533f,0.78484f,0.76691f,0.72998f,0.70219f,0.66691f,0.63834f,0.61502f,0.60413f,0.59514f,0.59266f,0.59453f,0.60219f,0.60958f,0.62031f,0.63424f,0.63679f,0.64513f,0.64872f,0.65081f,0.65285f,0.65489f,0.65262f,0.64591f,0.62918f,0.61973f,0.60147f,0.58216f,0.55431f,0.51789f,0.47592f,0.55820f,0.00000f,0.79530f,0.76104f,0.73219f,0.69915f,0.66537f,0.63566f,0.60864f,0.58623f,0.57380f,0.56875f,0.56959f,0.57687f,0.58478f,0.59920f,0.60463f,0.61502f,0.62182f,0.62431f,0.62905f,0.63315f,0.63788f,0.63741f,0.62548f,0.62214f,0.61275f,0.59594f,0.57855f,0.54448f,0.51634f,0.47646f,0.43487f,0.51155f,
 0.00000f,1.00000f,0.99837f,1.00000f,0.99035f,0.96895f,0.99159f,0.99463f,0.98371f,0.96619f,0.96539f,0.99824f,0.98192f,0.98977f,0.98592f,0.98488f,0.98168f,0.97260f,0.97129f,0.97670f,0.97274f,0.97217f,0.99153f,0.98777f,0.98551f,0.98730f,0.98283f,0.98334f,0.98645f,0.98920f,0.99139f,0.99521f,1.00000f,0.00000f,0.99252f,0.98713f,0.98290f,0.98046f,0.97801f,0.97664f,0.97660f,0.97545f,0.97277f,0.97338f,0.97508f,0.97456f,0.97453f,0.97402f,0.97559f,0.97613f,0.97547f,0.97567f,0.97826f,0.97826f,0.97899f,0.98196f,0.98266f,0.98439f,0.98581f,0.98684f,0.98852f,0.99058f,0.99268f,0.99477f,0.99727f,1.00000f,0.00000f,0.99008f,0.98557f,0.98295f,0.97971f,0.97719f,0.97672f,0.97587f,0.97446f,0.97285f,0.97438f,0.97293f,0.97446f,0.97387f,0.97395f,0.97459f,0.97487f,0.97538f,0.97690f,0.97795f,0.97875f,0.97957f,0.98100f,0.98320f,0.98398f,0.98615f,0.98759f,0.98931f,0.99129f,0.99338f,0.99540f,0.99772f,1.00000f,0.00000f,0.98569f,0.98485f,0.98280f,0.97914f,0.97755f,0.97567f,0.97455f,0.97405f,0.97388f,0.97450f,0.97280f,0.97364f,0.97295f,0.97403f,0.97387f,0.97561f,0.97582f,0.97687f,0.97783f,0.97938f,0.97960f,0.98137f,0.98270f,0.98422f,0.98614f,0.98768f,0.98941f,0.99150f,0.99341f,0.99551f,0.99777f,1.00000f,0.00000f,0.93976f,0.98129f,0.98129f,0.97871f,0.97784f,0.97578f,0.97369f,0.97437f,0.97331f,0.97452f,0.97255f,0.97343f,0.97404f,0.97392f,0.97424f,0.97459f,0.97580f,0.97689f,0.97761f,0.97870f,0.97971f,0.98132f,0.98241f,0.98420f,0.98586f,0.98753f,0.98946f,0.99137f,0.99333f,0.99543f,0.99772f,1.00000f,0.00000f,0.90628f,0.95663f,0.97728f,0.97839f,0.97666f,0.97423f,0.97513f,0.97380f,0.97337f,0.97308f,0.97327f,0.97322f,0.97214f,0.97352f,0.97426f,0.97424f,0.97562f,0.97655f,0.97744f,0.97896f,0.97991f,0.98131f,0.98246f,0.98442f,0.98518f,0.98768f,0.98928f,0.99144f,0.99331f,0.99552f,0.99776f,1.00000f,0.00000f,0.89829f,0.92057f,0.95645f,0.97174f,0.97444f,0.97485f,0.97326f,0.97285f,0.97337f,0.97247f,0.97239f,0.97290f,0.97299f,0.97323f,0.97398f,0.97476f,0.97525f,0.97687f,0.97714f,0.97862f,0.97983f,0.98079f,0.98226f,0.98424f,0.98526f,0.98753f,0.98930f,0.99133f,0.99343f,0.99542f,0.99764f,1.00000f,0.00000f,0.89917f,0.90369f,0.92619f,0.94884f,0.96386f,0.96952f,0.97256f,0.97372f,0.97334f,0.97231f,0.97205f,0.97303f,0.97242f,0.97381f,0.97406f,0.97412f,0.97512f,0.97709f,0.97703f,0.97803f,0.97951f,0.98103f,0.98219f,0.98287f,0.98488f,0.98716f,0.98926f,0.99114f,0.99331f,0.99528f,0.99752f,1.00000f,0.00000f,0.90796f,0.89384f,0.90750f,0.92307f,0.94438f,0.95796f,0.96495f,0.96847f,0.97147f,0.97120f,0.97085f,0.97283f,0.97293f,0.97308f,0.97321f,0.97426f,0.97500f,0.97553f,0.97696f,0.97763f,0.97935f,0.98085f,0.98170f,0.98122f,0.98429f,0.98706f,0.98922f,0.99107f,0.99332f,0.99523f,0.99756f,1.00000f,0.00000f,0.91731f,0.89744f,0.89781f,0.90941f,0.92151f,0.93552f,0.95124f,0.95738f,0.96404f,0.96920f,0.96956f,0.97249f,0.97198f,0.97297f,0.97297f,0.97373f,0.97494f,0.97578f,0.97682f,0.97844f,0.97837f,0.98027f,0.97988f,0.97991f,0.98396f,0.98690f,0.98905f,0.99080f,0.99296f,0.99524f,0.99760f,1.00000f,0.00000f,0.91903f,0.90347f,0.89948f,0.89910f,0.90660f,0.91864f,0.93058f,0.94111f,0.95050f,0.95796f,0.96266f,0.96780f,0.96929f,0.97132f,0.97170f,0.97286f,0.97426f,0.97488f,0.97624f,0.97766f,0.97814f,0.97823f,0.97693f,0.97850f,0.98341f,0.98696f,0.98888f,0.99066f,0.99292f,0.99506f,0.99737f,1.00000f,0.00000f,0.92853f,0.90916f,0.89154f,0.89320f,0.89574f,0.90078f,0.91043f,0.92202f,0.93196f,0.94149f,0.94917f,0.95722f,0.96457f,0.96619f,0.96884f,0.97028f,0.97254f,0.97473f,0.97593f,0.97611f,0.97545f,0.97389f,0.97512f,0.97752f,0.98323f,0.98601f,0.98867f,0.99075f,0.99264f,0.99502f,0.99725f,1.00000f,0.00000f,0.93661f,0.90858f,0.90171f,0.88822f,0.88853f,0.89612f,0.90004f,0.90581f,0.91772f,0.92286f,0.93310f,0.94059f,0.95002f,0.95689f,0.96065f,0.96713f,0.96880f,0.97179f,0.97357f,0.97139f,0.97039f,0.97029f,0.97230f,0.97696f,0.98251f,0.98629f,0.98876f,0.99060f,0.99275f,0.99499f,0.99733f,1.00000f,0.00000f,0.93537f,0.91909f,0.90520f,0.88628f,0.88322f,0.89190f,0.89293f,0.89812f,0.90309f,0.90954f,0.91484f,0.92697f,0.93549f,0.94068f,0.94643f,0.95512f,0.95813f,0.96327f,0.96530f,0.96313f,0.96302f,0.96566f,0.96943f,0.97598f,0.98113f,0.98626f,0.98819f,0.99040f,0.99231f,0.99493f,0.99716f,1.00000f,0.00000f,0.94529f,0.92223f,0.90452f,0.89328f,0.89055f,0.89102f,0.88539f,0.89140f,0.89180f,0.89463f,0.90152f,0.91308f,0.91673f,0.92631f,0.93284f,0.93983f,0.94627f,0.94741f,0.94849f,0.94948f,0.95406f,0.95786f,0.96566f,0.97322f,0.97977f,0.98455f,0.98752f,0.99020f,0.99250f,0.99488f,0.99740f,1.00000f,0.00000f,0.94709f,0.92810f,0.91230f,0.90132f,0.89436f,0.88892f,0.88097f,0.88425f,0.88519f,0.88734f,0.89128f,0.89500f,0.90292f,0.90922f,0.91565f,0.91897f,0.92381f,0.92555f,0.92785f,0.93214f,0.93786f,0.94707f,0.95809f,0.96689f,0.97556f,0.98212f,0.98679f,0.98977f,0.99209f,0.99439f,0.99711f,1.00000f,0.00000f,0.94008f,0.93365f,0.91243f,0.90575f,0.89275f,0.88834f,0.88546f,0.88156f,0.87879f,0.87645f,0.88294f,0.88956f,0.88758f,0.89510f,0.89554f,0.89766f,0.90070f,0.90210f,0.90671f,0.91192f,0.92025f,0.93215f,0.94451f,0.95791f,0.96575f,0.97364f,0.98045f,0.98613f,0.99023f,0.99432f,0.99681f,1.00000f,0.00000f,0.94576f,0.93172f,0.92236f,0.90268f,0.89834f,0.89106f,0.88597f,0.88294f,0.87890f,0.87706f,0.87882f,0.88102f,0.87880f,0.87441f,0.87330f,0.87315f,0.87568f,0.87561f,0.88196f,0.89134f,0.90189f,0.91418f,0.92815f,0.94217f,0.95271f,0.96295f,0.97071f,0.97878f,0.98518f,0.99054f,0.99593f,1.00000f,0.00000f,0.94665f,0.93023f,0.92315f,0.91554f,0.89852f,0.89336f,0.88727f,0.88463f,0.87497f,0.87314f,0.87263f,0.86753f,0.86549f,0.85682f,0.85267f,0.84722f,0.84997f,0.85289f,0.85851f,0.87002f,0.88096f,0.89498f,0.91034f,0.92479f,0.93743f,0.94888f,0.95737f,0.96729f,0.97437f,0.98362f,0.99217f,1.00000f,0.00000f,0.95113f,0.94203f,0.92921f,0.91990f,0.90829f,0.89685f,0.88620f,0.88480f,0.87766f,0.87349f,0.86443f,0.85380f,0.84470f,0.84221f,0.83054f,0.82896f,0.82557f,0.83009f,0.84016f,0.84794f,0.86182f,0.87540f,0.88743f,0.90361f,0.91579f,0.92727f,0.93787f,0.94820f,0.95875f,0.96993f,0.98117f,0.99976f,0.00000f,0.94866f,0.93476f,0.93151f,0.91379f,0.91280f,0.89848f,0.88817f,0.87831f,0.86909f,0.85746f,0.84984f,0.83692f,0.82757f,0.81825f,0.81259f,0.80603f,0.80787f,0.80935f,0.81779f,0.82840f,0.84141f,0.85700f,0.86822f,0.88346f,0.89456f,0.90548f,0.91660f,0.92753f,0.93904f,0.95055f,0.96382f,0.99772f,0.00000f,0.94730f,0.93967f,0.92860f,0.91472f,0.90936f,0.89656f,0.88367f,0.87683f,0.85742f,0.85056f,0.83235f,0.81607f,0.80695f,0.79597f,0.78940f,0.78788f,0.79036f,0.79230f,0.80296f,0.81131f,0.82117f,0.83529f,0.84734f,0.86012f,0.87333f,0.88208f,0.89165f,0.90140f,0.91236f,0.92430f,0.93813f,0.99078f,0.00000f,0.94250f,0.94021f,0.93355f,0.91863f,0.91242f,0.89681f,0.87913f,0.86373f,0.85141f,0.82891f,0.81802f,0.80113f,0.78603f,0.77353f,0.76900f,0.76950f,0.77412f,0.77544f,0.78561f,0.79346f,0.80465f,0.81812f,0.82807f,0.83502f,0.84655f,0.85309f,0.86414f,0.87462f,0.88081f,0.89455f,0.90398f,0.97437f,0.00000f,0.94603f,0.93699f,0.92723f,0.91720f,0.90565f,0.89023f,0.87132f,0.85151f,0.83581f,0.81396f,0.79282f,0.77902f,0.76600f,0.75412f,0.75192f,0.75333f,0.75854f,0.76374f,0.77181f,0.77730f,0.79050f,0.79636f,0.80652f,0.81395f,0.82026f,0.82957f,0.83410f,0.84215f,0.84896f,0.85884f,0.86381f,0.94669f,0.00000f,0.94303f,0.93226f,0.92187f,0.91043f,0.89604f,0.87842f,0.85726f,0.83805f,0.81048f,0.78942f,0.77450f,0.76040f,0.74126f,0.73647f,0.73001f,0.73893f,0.74460f,0.75107f,0.75842f,0.76585f,0.77462f,0.78166f,0.78426f,0.79145f,0.79467f,0.79878f,0.80444f,0.80784f,0.81320f,0.81446f,0.81669f,0.91025f,0.00000f,0.93364f,0.93245f,0.91665f,0.90307f,0.88304f,0.85565f,0.83531f,0.81514f,0.78925f,0.76974f,0.74912f,0.73458f,0.72664f,0.72057f,0.71888f,0.72411f,0.73065f,0.73421f,0.74619f,0.75028f,0.75863f,0.76141f,0.76609f,0.76851f,0.76795f,0.77312f,0.77352f,0.77192f,0.77320f,0.77212f,0.76414f,0.86646f,0.00000f,0.92121f,0.91238f,0.89772f,0.88580f,0.86263f,0.83429f,0.81258f,0.79353f,0.76245f,0.73878f,0.72498f,0.71216f,0.70491f,0.70533f,0.70699f,0.70928f,0.71734f,0.72154f,0.73027f,0.74105f,0.74011f,0.74709f,0.74122f,0.74600f,0.74516f,0.74285f,0.74020f,0.73712f,0.73628f,0.72718f,0.71287f,0.81694f,0.00000f,0.91959f,0.90617f,0.87937f,0.86332f,0.84514f,0.82310f,0.78701f,0.76049f,0.74286f,0.71123f,0.69597f,0.68811f,0.68408f,0.68817f,0.69108f,0.69433f,0.70005f,0.71029f,0.71305f,0.72494f,0.72104f,0.72481f,0.72565f,0.72730f,0.72116f,0.72038f,0.71138f,0.70416f,0.69537f,0.67821f,0.65792f,0.76369f,0.00000f,0.90045f,0.88607f,0.86679f,0.84162f,0.82024f,0.78778f,0.76261f,0.73421f,0.71339f,0.69251f,0.67782f,0.67031f,0.66631f,0.67251f,0.67384f,0.68098f,0.68779f,0.69488f,0.70190f,0.70576f,0.70521f,0.70520f,0.70373f,0.70172f,0.69648f,0.68888f,0.67706f,0.66830f,0.65365f,0.63140f,0.60655f,0.70723f,0.00000f,0.88502f,0.86686f,0.84038f,0.82015f,0.78434f,0.76733f,0.72648f,0.70942f,0.68147f,0.66193f,0.65334f,0.64514f,0.64941f,0.65289f,0.65562f,0.66299f,0.66783f,0.67679f,0.67744f,0.68662f,0.69125f,0.68707f,0.68401f,0.67914f,0.67047f,0.65733f,0.64723f,0.63512f,0.61253f,0.59030f,0.55446f,0.65326f,0.00000f,0.85991f,0.84349f,0.82361f,0.78913f,0.75926f,0.72865f,0.69510f,0.67273f,0.64890f,0.63770f,0.63200f,0.62761f,0.62960f,0.63651f,0.63965f,0.64798f,0.65411f,0.65988f,0.66889f,0.66728f,0.66690f,0.66588f,0.66515f,0.65689f,0.64639f,0.63456f,0.61902f,0.59866f,0.57439f,0.53917f,0.50559f,0.60052f,0.00000f,0.84025f,0.80794f,0.78368f,0.76384f,0.73145f,0.69818f,0.67125f,0.64123f,0.62428f,0.61702f,0.60693f,0.61197f,0.60899f,0.61964f,0.62748f,0.63525f,0.63716f,0.64340f,0.64974f,0.65274f,0.64642f,0.64729f,0.64635f,0.63506f,0.62635f,0.60564f,0.59499f,0.56774f,0.53332f,0.49782f,0.45956f,0.54699f,0.00000f,0.81245f,0.78160f,0.75984f,0.72977f,0.69610f,0.66254f,0.63133f,0.60810f,0.59527f,0.58632f,0.58584f,0.58474f,0.59166f,0.60171f,0.60590f,0.61726f,0.62622f,0.63060f,0.63252f,0.62706f,0.63426f,0.62504f,0.62200f,0.61365f,0.59866f,0.58064f,0.56283f,0.53349f,0.49623f,0.45823f,0.41828f,0.49686f,
 0.00000f,1.00000f,0.99890f,1.00000f,0.99079f,0.96813f,0.99074f,0.99427f,0.98353f,0.96592f,0.96485f,0.99794f,0.98146f,0.98929f,0.98575f,0.98380f,0.98211f,0.97191f,0.97134f,0.97591f,0.97315f,0.97205f,0.99086f,0.98738f,0.98565f,0.98723f,0.99174f,0.98277f,0.98846f,0.98995f,0.99229f,0.99654f,1.00000f,0.00000f,0.99275f,0.98749f,0.98295f,0.98092f,0.97823f,0.97603f,0.97656f,0.97534f,0.97345f,0.97335f,0.97505f,0.97377f,0.97424f,0.97532f,0.97597f,0.97561f,0.97475f,0.97598f,0.97817f,0.97876f,0.97919f,0.98202f,0.98280f,0.98422f,0.98582f,0.98797f,0.98842f,0.99100f,0.99273f,0.99495f,0.99751f,1.00000f,0.00000f,0.99155f,0.98585f,0.98232f,0.98041f,0.97733f,0.97687f,0.97454f,0.97506f,0.97303f,0.97361f,0.97330f,0.97248f,0.97378f,0.97366f,0.97435f,0.97572f,0.97553f,0.97690f,0.97833f,0.97866f,0.98021f,0.98144f,0.98308f,0.98426f,0.98587f,0.98762f,0.98962f,0.99135f,0.99320f,0.99548f,0.99778f,1.00000f,0.00000f,0.98153f,0.98481f,0.98214f,0.97993f,0.97869f,0.97626f,0.97510f,0.97452f,0.97386f,0.97323f,0.97294f,0.97334f,0.97369f,0.97417f,0.97412f,0.97528f,0.97571f,0.97648f,0.97788f,0.97837f,0.97976f,0.98158f,0.98269f,0.98443f,0.98582f,0.98766f,0.98929f,0.99123f,0.99346f,0.99545f,0.99778f,1.00000f,0.00000f,0.93808f,0.98147f,0.98144f,0.97990f,0.97762f,0.97578f,0.97553f,0.97412f,0.97446f,0.97331f,0.97218f,0.97299f,0.97314f,0.97351f,0.97408f,0.97491f,0.97585f,0.97686f,0.97752f,0.97892f,0.98018f,0.98161f,0.98277f,0.98449f,0.98600f,0.98755f,0.98930f,0.99121f,0.99331f,0.99552f,0.99775f,1.00000f,0.00000f,0.90800f,0.95654f,0.97637f,0.97757f,0.97709f,0.97531f,0.97406f,0.97369f,0.97294f,0.97298f,0.97275f,0.97324f,0.97318f,0.97357f,0.97397f,0.97477f,0.97556f,0.97595f,0.97733f,0.97919f,0.97958f,0.98122f,0.98300f,0.98412f,0.98557f,0.98723f,0.98930f,0.99119f,0.99330f,0.99543f,0.99778f,1.00000f,0.00000f,0.89244f,0.92183f,0.95550f,0.97087f,0.97592f,0.97414f,0.97272f,0.97280f,0.97186f,0.97355f,0.97290f,0.97247f,0.97262f,0.97315f,0.97426f,0.97458f,0.97559f,0.97647f,0.97774f,0.97880f,0.97974f,0.98115f,0.98276f,0.98405f,0.98529f,0.98696f,0.98919f,0.99115f,0.99334f,0.99544f,0.99761f,1.00000f,0.00000f,0.89906f,0.90243f,0.92768f,0.95032f,0.96468f,0.97064f,0.97334f,0.97085f,0.97342f,0.97178f,0.97139f,0.97292f,0.97264f,0.97217f,0.97374f,0.97439f,0.97569f,0.97630f,0.97689f,0.97873f,0.97984f,0.98107f,0.98256f,0.98358f,0.98394f,0.98711f,0.98922f,0.99110f,0.99312f,0.99525f,0.99756f,1.00000f,0.00000f,0.91190f,0.89640f,0.90616f,0.92346f,0.94308f,0.95468f,0.96485f,0.96863f,0.97208f,0.97088f,0.97225f,0.97270f,0.97320f,0.97237f,0.97371f,0.97416f,0.97522f,0.97651f,0.97610f,0.97833f,0.97923f,0.98087f,0.98238f,0.98202f,0.98322f,0.98694f,0.98929f,0.99075f,0.99313f,0.99525f,0.99754f,1.00000f,0.00000f,0.91814f,0.89779f,0.89845f,0.90893f,0.92030f,0.93673f,0.95072f,0.95962f,0.96332f,0.96710f,0.97029f,0.97147f,0.97268f,0.97243f,0.97387f,0.97420f,0.97460f,0.97636f,0.97718f,0.97847f,0.97981f,0.98094f,0.98119f,0.97998f,0.98280f,0.98670f,0.98869f,0.99101f,0.99294f,0.99529f,0.99743f,1.00000f,0.00000f,0.92918f,0.90060f,0.89530f,0.89972f,0.90715f,0.91712f,0.93186f,0.93869f,0.95187f,0.95875f,0.96295f,0.96788f,0.96961f,0.97059f,0.97202f,0.97301f,0.97486f,0.97588f,0.97656f,0.97769f,0.97915f,0.97864f,0.97788f,0.97862f,0.98205f,0.98633f,0.98901f,0.99084f,0.99311f,0.99517f,0.99751f,1.00000f,0.00000f,0.92963f,0.90386f,0.88990f,0.88997f,0.90090f,0.90044f,0.91217f,0.92367f,0.93288f,0.94255f,0.95121f,0.95656f,0.96203f,0.96640f,0.96895f,0.97239f,0.97381f,0.97543f,0.97576f,0.97796f,0.97725f,0.97637f,0.97567f,0.97734f,0.98148f,0.98601f,0.98899f,0.99064f,0.99301f,0.99509f,0.99746f,1.00000f,0.00000f,0.93729f,0.90533f,0.89504f,0.89143f,0.88830f,0.89556f,0.89896f,0.90747f,0.91495f,0.92348f,0.93479f,0.94152f,0.95144f,0.95653f,0.96140f,0.96480f,0.96872f,0.97203f,0.97482f,0.97422f,0.97298f,0.97277f,0.97170f,0.97531f,0.98139f,0.98535f,0.98876f,0.99026f,0.99286f,0.99492f,0.99730f,1.00000f,0.00000f,0.94039f,0.91894f,0.90566f,0.89256f,0.88899f,0.88791f,0.89246f,0.89602f,0.90155f,0.90914f,0.91691f,0.92679f,0.93353f,0.94199f,0.94981f,0.95451f,0.96019f,0.96459f,0.96676f,0.96717f,0.96725f,0.96740f,0.96959f,0.97439f,0.98035f,0.98532f,0.98789f,0.99069f,0.99261f,0.99482f,0.99701f,1.00000f,0.00000f,0.94414f,0.92236f,0.90501f,0.89490f,0.88742f,0.88679f,0.88886f,0.89018f,0.89105f,0.89570f,0.89985f,0.90890f,0.91809f,0.92359f,0.93374f,0.94283f,0.94700f,0.94984f,0.95086f,0.95310f,0.95625f,0.95805f,0.96434f,0.97234f,0.97850f,0.98385f,0.98751f,0.98996f,0.99280f,0.99502f,0.99708f,1.00000f,0.00000f,0.93772f,0.92473f,0.91270f,0.89527f,0.88918f,0.88845f,0.88269f,0.88546f,0.88594f,0.88646f,0.89250f,0.89581f,0.90293f,0.91071f,0.91680f,0.92229f,0.92765f,0.92913f,0.93193f,0.93601f,0.93993f,0.94591f,0.95538f,0.96463f,0.97402f,0.98124f,0.98603f,0.98896f,0.99229f,0.99467f,0.99736f,1.00000f,0.00000f,0.94151f,0.93177f,0.91403f,0.90262f,0.89425f,0.88787f,0.88836f,0.88052f,0.88059f,0.88243f,0.88341f,0.88743f,0.89193f,0.89800f,0.89667f,0.90307f,0.90486f,0.90390f,0.91092f,0.91596f,0.92083f,0.93186f,0.94321f,0.95388f,0.96542f,0.97436f,0.98112f,0.98568f,0.99074f,0.99390f,0.99684f,1.00000f,0.00000f,0.94278f,0.93462f,0.92428f,0.91116f,0.89616f,0.89226f,0.88700f,0.88338f,0.88186f,0.87682f,0.87827f,0.88102f,0.88127f,0.88051f,0.87871f,0.88018f,0.88250f,0.87977f,0.88326f,0.89284f,0.90251f,0.91352f,0.92854f,0.94118f,0.95275f,0.96265f,0.97072f,0.97852f,0.98468f,0.99092f,0.99573f,1.00000f,0.00000f,0.94135f,0.93629f,0.91870f,0.91225f,0.89976f,0.89207f,0.88629f,0.88436f,0.87650f,0.87551f,0.87359f,0.86934f,0.86801f,0.86212f,0.86198f,0.85812f,0.85468f,0.85664f,0.86374f,0.87136f,0.87808f,0.89451f,0.90897f,0.92300f,0.93440f,0.94773f,0.95646f,0.96591f,0.97468f,0.98374f,0.99117f,1.00000f,0.00000f,0.94548f,0.93680f,0.92627f,0.91336f,0.90932f,0.89413f,0.88717f,0.88072f,0.88158f,0.87193f,0.86341f,0.85734f,0.85328f,0.84761f,0.83691f,0.83378f,0.83060f,0.83363f,0.83918f,0.84652f,0.86279f,0.87627f,0.88790f,0.90432f,0.91520f,0.92727f,0.93979f,0.94698f,0.95989f,0.97019f,0.98119f,0.99961f,0.00000f,0.94154f,0.93476f,0.93022f,0.92193f,0.91120f,0.90326f,0.89319f,0.87994f,0.87600f,0.86375f,0.85693f,0.84369f,0.83501f,0.82934f,0.81981f,0.81343f,0.81307f,0.81480f,0.81887f,0.83039f,0.84180f,0.85627f,0.86827f,0.88061f,0.89278f,0.90534f,0.91746f,0.92633f,0.93734f,0.95162f,0.96369f,0.99733f,0.00000f,0.94730f,0.93757f,0.93205f,0.92072f,0.91310f,0.89916f,0.88879f,0.87575f,0.86777f,0.85416f,0.84481f,0.82757f,0.82006f,0.80676f,0.79718f,0.79229f,0.79474f,0.80182f,0.80101f,0.81301f,0.82231f,0.83420f,0.84526f,0.85781f,0.86574f,0.88094f,0.89142f,0.90071f,0.91387f,0.92544f,0.93822f,0.99027f,0.00000f,0.93912f,0.94342f,0.93236f,0.92092f,0.91555f,0.89795f,0.88633f,0.87442f,0.86149f,0.83873f,0.82803f,0.80833f,0.79902f,0.78323f,0.78305f,0.77495f,0.77381f,0.77915f,0.78402f,0.79681f,0.80260f,0.81711f,0.82567f,0.83630f,0.84336f,0.85247f,0.86358f,0.87049f,0.88096f,0.89241f,0.90468f,0.97309f,0.00000f,0.93686f,0.93730f,0.93255f,0.91466f,0.90391f,0.89626f,0.87813f,0.86458f,0.84371f,0.82694f,0.80782f,0.79180f,0.77890f,0.76784f,0.76253f,0.75930f,0.76220f,0.76825f,0.77369f,0.78058f,0.79022f,0.79811f,0.80351f,0.81389f,0.82059f,0.82659f,0.83424f,0.84211f,0.84751f,0.85504f,0.86287f,0.94490f,0.00000f,0.94038f,0.93854f,0.92708f,0.91475f,0.90282f,0.88517f,0.86901f,0.84333f,0.82712f,0.80810f,0.78597f,0.77111f,0.75172f,0.74723f,0.74629f,0.74319f,0.74494f,0.74948f,0.75776f,0.76573f,0.76963f,0.77558f,0.78115f,0.78826f,0.79484f,0.79832f,0.80158f,0.80619f,0.80887f,0.81127f,0.81314f,0.90758f,0.00000f,0.93306f,0.92961f,0.92094f,0.90749f,0.89494f,0.86932f,0.85067f,0.82807f,0.80625f,0.78531f,0.76125f,0.75003f,0.73622f,0.72933f,0.72681f,0.73162f,0.72913f,0.73694f,0.74507f,0.74636f,0.76100f,0.76304f,0.76498f,0.76839f,0.76814f,0.76659f,0.76883f,0.77096f,0.77243f,0.76825f,0.75893f,0.86406f,0.00000f,0.93147f,0.91635f,0.91570f,0.89949f,0.87752f,0.85738f,0.83044f,0.81675f,0.78239f,0.76250f,0.74235f,0.72764f,0.71455f,0.71378f,0.71376f,0.71477f,0.71611f,0.72262f,0.72912f,0.73438f,0.73581f,0.74039f,0.74366f,0.74197f,0.74250f,0.74178f,0.74016f,0.73322f,0.72898f,0.71720f,0.70490f,0.81225f,0.00000f,0.92481f,0.91275f,0.90592f,0.88425f,0.85598f,0.83907f,0.81306f,0.78561f,0.75708f,0.73351f,0.71193f,0.70723f,0.69673f,0.69383f,0.69363f,0.70029f,0.70419f,0.70967f,0.71411f,0.71965f,0.72281f,0.72823f,0.72139f,0.71947f,0.71989f,0.71284f,0.70665f,0.70038f,0.68690f,0.67238f,0.64979f,0.75781f,0.00000f,0.92072f,0.88796f,0.88472f,0.85444f,0.84467f,0.80831f,0.78864f,0.75168f,0.72412f,0.70752f,0.68889f,0.68770f,0.68292f,0.67955f,0.68322f,0.68407f,0.69248f,0.69629f,0.69742f,0.70384f,0.70695f,0.70246f,0.70263f,0.69635f,0.69044f,0.68410f,0.66882f,0.65748f,0.64397f,0.62258f,0.59593f,0.70087f,0.00000f,0.89370f,0.87532f,0.85955f,0.83921f,0.81066f,0.78647f,0.75435f,0.73097f,0.70916f,0.68346f,0.67018f,0.66085f,0.65705f,0.66182f,0.66324f,0.66828f,0.67295f,0.67758f,0.68096f,0.68703f,0.68223f,0.68216f,0.67824f,0.67522f,0.66587f,0.65820f,0.64241f,0.62557f,0.60196f,0.57594f,0.54674f,0.64574f,0.00000f,0.87897f,0.85637f,0.83822f,0.80795f,0.78247f,0.75831f,0.72017f,0.70016f,0.66922f,0.65575f,0.64937f,0.63791f,0.63967f,0.64381f,0.64828f,0.64859f,0.65601f,0.66478f,0.66369f,0.67077f,0.66979f,0.66060f,0.66019f,0.65219f,0.64374f,0.62565f,0.61243f,0.59043f,0.55939f,0.52572f,0.49635f,0.59001f,0.00000f,0.86053f,0.83195f,0.81013f,0.77704f,0.74810f,0.72090f,0.69376f,0.66228f,0.64500f,0.63144f,0.62388f,0.61577f,0.61992f,0.62208f,0.62767f,0.63842f,0.63777f,0.64189f,0.64595f,0.64501f,0.64086f,0.64057f,0.63702f,0.62978f,0.61548f,0.59938f,0.58115f,0.55237f,0.52057f,0.48508f,0.45073f,0.53662f,0.00000f,0.83480f,0.81207f,0.77980f,0.74793f,0.71716f,0.69117f,0.66434f,0.63718f,0.61796f,0.60295f,0.59871f,0.60030f,0.60049f,0.60661f,0.61007f,0.61715f,0.62521f,0.62537f,0.62620f,0.62290f,0.62864f,0.61660f,0.61829f,0.60752f,0.59043f,0.57306f,0.54705f,0.52016f,0.47977f,0.44511f,0.40543f,0.48899f,
 0.00000f,1.00000f,0.99766f,0.99988f,0.99069f,0.96885f,0.99110f,0.99386f,0.98427f,0.96583f,0.96392f,0.99757f,0.98174f,0.98883f,0.98515f,0.98421f,0.98239f,0.97200f,0.97106f,0.97618f,0.97355f,0.97220f,0.99118f,0.98738f,0.98593f,0.98693f,0.99132f,0.98606f,0.98517f,0.98854f,0.99328f,0.99595f,1.00000f,0.00000f,0.99177f,0.98581f,0.98418f,0.98019f,0.97710f,0.97846f,0.97631f,0.97593f,0.97481f,0.97253f,0.97509f,0.97376f,0.97450f,0.97499f,0.97522f,0.97621f,0.97520f,0.97632f,0.97762f,0.97847f,0.97876f,0.98208f,0.98265f,0.98442f,0.98574f,0.98816f,0.98910f,0.99030f,0.99278f,0.99509f,0.99750f,1.00000f,0.00000f,0.98978f,0.98501f,0.98291f,0.98054f,0.97775f,0.97626f,0.97576f,0.97428f,0.97441f,0.97391f,0.97358f,0.97365f,0.97420f,0.97414f,0.97394f,0.97520f,0.97661f,0.97656f,0.97798f,0.97858f,0.98090f,0.98134f,0.98320f,0.98441f,0.98592f,0.98771f,0.98953f,0.99143f,0.99350f,0.99547f,0.99775f,1.00000f,0.00000f,0.98181f,0.98590f,0.98230f,0.98012f,0.97764f,0.97590f,0.97409f,0.97351f,0.97296f,0.97377f,0.97378f,0.97316f,0.97339f,0.97410f,0.97504f,0.97520f,0.97583f,0.97697f,0.97817f,0.97908f,0.98004f,0.98148f,0.98262f,0.98472f,0.98607f,0.98780f,0.98932f,0.99113f,0.99329f,0.99543f,0.99776f,1.00000f,0.00000f,0.94054f,0.98220f,0.98053f,0.97921f,0.97840f,0.97461f,0.97521f,0.97450f,0.97300f,0.97379f,0.97341f,0.97332f,0.97326f,0.97333f,0.97371f,0.97538f,0.97649f,0.97641f,0.97751f,0.97914f,0.97934f,0.98117f,0.98308f,0.98461f,0.98601f,0.98777f,0.98943f,0.99129f,0.99327f,0.99554f,0.99776f,1.00000f,0.00000f,0.90763f,0.95570f,0.97660f,0.97774f,0.97747f,0.97581f,0.97517f,0.97331f,0.97523f,0.97257f,0.97295f,0.97288f,0.97428f,0.97467f,0.97497f,0.97437f,0.97562f,0.97646f,0.97756f,0.97867f,0.98010f,0.98142f,0.98263f,0.98445f,0.98556f,0.98707f,0.98950f,0.99117f,0.99331f,0.99550f,0.99778f,1.00000f,0.00000f,0.89945f,0.92204f,0.95407f,0.97022f,0.97531f,0.97404f,0.97384f,0.97456f,0.97262f,0.97326f,0.97281f,0.97322f,0.97322f,0.97287f,0.97445f,0.97453f,0.97590f,0.97652f,0.97830f,0.97782f,0.97973f,0.98156f,0.98243f,0.98418f,0.98585f,0.98659f,0.98947f,0.99094f,0.99327f,0.99540f,0.99763f,1.00000f,0.00000f,0.90302f,0.90720f,0.92403f,0.95204f,0.96476f,0.97077f,0.97338f,0.97301f,0.97278f,0.97304f,0.97271f,0.97240f,0.97286f,0.97302f,0.97322f,0.97424f,0.97454f,0.97679f,0.97691f,0.97796f,0.97983f,0.98087f,0.98227f,0.98354f,0.98495f,0.98676f,0.98917f,0.99125f,0.99304f,0.99539f,0.99751f,1.00000f,0.00000f,0.91120f,0.89909f,0.90393f,0.92514f,0.94315f,0.95762f,0.96642f,0.96961f,0.97068f,0.97101f,0.97155f,0.97315f,0.97303f,0.97272f,0.97344f,0.97420f,0.97617f,0.97593f,0.97665f,0.97791f,0.97969f,0.98092f,0.98219f,0.98363f,0.98353f,0.98590f,0.98881f,0.99091f,0.99316f,0.99534f,0.99756f,1.00000f,0.00000f,0.91199f,0.89646f,0.89940f,0.90675f,0.92119f,0.93671f,0.94965f,0.95805f,0.96306f,0.96873f,0.97028f,0.97172f,0.97259f,0.97214f,0.97312f,0.97387f,0.97422f,0.97651f,0.97694f,0.97730f,0.97886f,0.98098f,0.98162f,0.98142f,0.98206f,0.98595f,0.98886f,0.99118f,0.99318f,0.99533f,0.99744f,1.00000f,0.00000f,0.93252f,0.89840f,0.89700f,0.89437f,0.90473f,0.91842f,0.92936f,0.94092f,0.95139f,0.95655f,0.96439f,0.96716f,0.96939f,0.97154f,0.97232f,0.97289f,0.97353f,0.97526f,0.97627f,0.97759f,0.97895f,0.98004f,0.97953f,0.97882f,0.98126f,0.98613f,0.98869f,0.99079f,0.99291f,0.99497f,0.99744f,1.00000f,0.00000f,0.93787f,0.90595f,0.89272f,0.89179f,0.89495f,0.90270f,0.91197f,0.92314f,0.93471f,0.94281f,0.95016f,0.95791f,0.96245f,0.96757f,0.96816f,0.97177f,0.97259f,0.97565f,0.97629f,0.97748f,0.97816f,0.97717f,0.97612f,0.97686f,0.98046f,0.98555f,0.98841f,0.99119f,0.99272f,0.99512f,0.99739f,1.00000f,0.00000f,0.93747f,0.90690f,0.90023f,0.88905f,0.89270f,0.89426f,0.89884f,0.90799f,0.91602f,0.92625f,0.93604f,0.94242f,0.94778f,0.95529f,0.96101f,0.96620f,0.96895f,0.97238f,0.97384f,0.97546f,0.97534f,0.97408f,0.97346f,0.97508f,0.97924f,0.98535f,0.98824f,0.99061f,0.99248f,0.99511f,0.99723f,1.00000f,0.00000f,0.94522f,0.91419f,0.89561f,0.88807f,0.89204f,0.88758f,0.89256f,0.89683f,0.90175f,0.90571f,0.91830f,0.92589f,0.93516f,0.94303f,0.94921f,0.95501f,0.96074f,0.96397f,0.96680f,0.96942f,0.96849f,0.96850f,0.96961f,0.97306f,0.97852f,0.98415f,0.98814f,0.99043f,0.99230f,0.99472f,0.99707f,1.00000f,0.00000f,0.94835f,0.92079f,0.90290f,0.90003f,0.88618f,0.88486f,0.88710f,0.88792f,0.89095f,0.89573f,0.90276f,0.91012f,0.91849f,0.92488f,0.93565f,0.94083f,0.94689f,0.95126f,0.95416f,0.95581f,0.95804f,0.95844f,0.96359f,0.97047f,0.97763f,0.98254f,0.98791f,0.98975f,0.99269f,0.99474f,0.99718f,1.00000f,0.00000f,0.94671f,0.92982f,0.91289f,0.89637f,0.89082f,0.88353f,0.88696f,0.88595f,0.88412f,0.88531f,0.89047f,0.89531f,0.90463f,0.91050f,0.91861f,0.92531f,0.92971f,0.93196f,0.93778f,0.93784f,0.94145f,0.94641f,0.95491f,0.96421f,0.97200f,0.97974f,0.98609f,0.98919f,0.99213f,0.99487f,0.99710f,1.00000f,0.00000f,0.95044f,0.92789f,0.91908f,0.90053f,0.89570f,0.88509f,0.87963f,0.88352f,0.88292f,0.87988f,0.88375f,0.88594f,0.88795f,0.89377f,0.89906f,0.90543f,0.90766f,0.91121f,0.91141f,0.91818f,0.92249f,0.93050f,0.94275f,0.95160f,0.96388f,0.97232f,0.98034f,0.98585f,0.99065f,0.99394f,0.99672f,1.00000f,0.00000f,0.94393f,0.93064f,0.92035f,0.91032f,0.90011f,0.89061f,0.88324f,0.87768f,0.87903f,0.87625f,0.87831f,0.88022f,0.88148f,0.88267f,0.88094f,0.88552f,0.88474f,0.88570f,0.89247f,0.89214f,0.90241f,0.91302f,0.92484f,0.93859f,0.95202f,0.96180f,0.97184f,0.97809f,0.98507f,0.99104f,0.99590f,1.00000f,0.00000f,0.94864f,0.93793f,0.93028f,0.91259f,0.90096f,0.88893f,0.88591f,0.88014f,0.87423f,0.87543f,0.87322f,0.87290f,0.87231f,0.87139f,0.86478f,0.86000f,0.86073f,0.86405f,0.86271f,0.86902f,0.87949f,0.89309f,0.90466f,0.92067f,0.93538f,0.94606f,0.95774f,0.96640f,0.97423f,0.98330f,0.99183f,0.99999f,0.00000f,0.94745f,0.93292f,0.92792f,0.92037f,0.90508f,0.89316f,0.89060f,0.88069f,0.87916f,0.87208f,0.87489f,0.86650f,0.85966f,0.85408f,0.84688f,0.83984f,0.83793f,0.83970f,0.84126f,0.84906f,0.86384f,0.87283f,0.88472f,0.90281f,0.91392f,0.92669f,0.93713f,0.94930f,0.95809f,0.96911f,0.98133f,0.99955f,0.00000f,0.94735f,0.93609f,0.93001f,0.91702f,0.90381f,0.90159f,0.89146f,0.88601f,0.87618f,0.86841f,0.86115f,0.84970f,0.84360f,0.83882f,0.82667f,0.82036f,0.81883f,0.81476f,0.82465f,0.82582f,0.84101f,0.85419f,0.86617f,0.88059f,0.88990f,0.90715f,0.91571f,0.92834f,0.93835f,0.95064f,0.96377f,0.99706f,0.00000f,0.94005f,0.94639f,0.93187f,0.92109f,0.90906f,0.90175f,0.89157f,0.87908f,0.86969f,0.86272f,0.85146f,0.83862f,0.82677f,0.81431f,0.80642f,0.79905f,0.79388f,0.80012f,0.80088f,0.80962f,0.82152f,0.83390f,0.84746f,0.85714f,0.86624f,0.87812f,0.89057f,0.89991f,0.91138f,0.92611f,0.93805f,0.98944f,0.00000f,0.94828f,0.93354f,0.93291f,0.91460f,0.91295f,0.90488f,0.88481f,0.87841f,0.86487f,0.85049f,0.83350f,0.82727f,0.80887f,0.79343f,0.79067f,0.78242f,0.77943f,0.78276f,0.78949f,0.79800f,0.80320f,0.81065f,0.82371f,0.83338f,0.84208f,0.85387f,0.85972f,0.87117f,0.88013f,0.89168f,0.90226f,0.97159f,0.00000f,0.94286f,0.94493f,0.92961f,0.91868f,0.91355f,0.90122f,0.87807f,0.87228f,0.85438f,0.83821f,0.81786f,0.80500f,0.78644f,0.77303f,0.77481f,0.76507f,0.76736f,0.76920f,0.77368f,0.78166f,0.78655f,0.79310f,0.80330f,0.81006f,0.81896f,0.83028f,0.83280f,0.83806f,0.85047f,0.85241f,0.85917f,0.94333f,0.00000f,0.93987f,0.94147f,0.93170f,0.91227f,0.90914f,0.89725f,0.87686f,0.85759f,0.83365f,0.81558f,0.79758f,0.77902f,0.76597f,0.75605f,0.75268f,0.75115f,0.75048f,0.75269f,0.75954f,0.76569f,0.76854f,0.77656f,0.78407f,0.78800f,0.79171f,0.79874f,0.80209f,0.80475f,0.80845f,0.80972f,0.80984f,0.90519f,0.00000f,0.93915f,0.93443f,0.91625f,0.91213f,0.90346f,0.87987f,0.86182f,0.84051f,0.81680f,0.79603f,0.78105f,0.75551f,0.74605f,0.74009f,0.73518f,0.73329f,0.73485f,0.74038f,0.74552f,0.74989f,0.75505f,0.75916f,0.76392f,0.76849f,0.76749f,0.76782f,0.77070f,0.76833f,0.76629f,0.76443f,0.75681f,0.86031f,0.00000f,0.93597f,0.92199f,0.91433f,0.91201f,0.88013f,0.86466f,0.84305f,0.82131f,0.79662f,0.77505f,0.75573f,0.73908f,0.72839f,0.72095f,0.72029f,0.71879f,0.71934f,0.72388f,0.72997f,0.73354f,0.74414f,0.74231f,0.74421f,0.74246f,0.74129f,0.74027f,0.73505f,0.73264f,0.72582f,0.71508f,0.70151f,0.80910f,0.00000f,0.92791f,0.91430f,0.91033f,0.88840f,0.86556f,0.85459f,0.82420f,0.79618f,0.76883f,0.75071f,0.73318f,0.71209f,0.70754f,0.70669f,0.70194f,0.70615f,0.70863f,0.71057f,0.71462f,0.71731f,0.72334f,0.72037f,0.72031f,0.72005f,0.71632f,0.70559f,0.70063f,0.69372f,0.67841f,0.66483f,0.64481f,0.75219f,0.00000f,0.91858f,0.90932f,0.88716f,0.86828f,0.85123f,0.82658f,0.79319f,0.77329f,0.75067f,0.72357f,0.70980f,0.69842f,0.68887f,0.68966f,0.68784f,0.69043f,0.69050f,0.69741f,0.70573f,0.70325f,0.70099f,0.70440f,0.70382f,0.69635f,0.68809f,0.67941f,0.66997f,0.65342f,0.63887f,0.61653f,0.58979f,0.69505f,0.00000f,0.90726f,0.88709f,0.87475f,0.85123f,0.83024f,0.80152f,0.77456f,0.74584f,0.72021f,0.70177f,0.68915f,0.67498f,0.66804f,0.66402f,0.66525f,0.67484f,0.67777f,0.68044f,0.68540f,0.68476f,0.68158f,0.67632f,0.68090f,0.67377f,0.66080f,0.64838f,0.63706f,0.61756f,0.59354f,0.56660f,0.53697f,0.63939f,0.00000f,0.89509f,0.87742f,0.85248f,0.82561f,0.79066f,0.77134f,0.74594f,0.71791f,0.69397f,0.67344f,0.66061f,0.65146f,0.65172f,0.64731f,0.65180f,0.66090f,0.65889f,0.66303f,0.66392f,0.66352f,0.66465f,0.65913f,0.65495f,0.64975f,0.63570f,0.62285f,0.60256f,0.57727f,0.55192f,0.51955f,0.48948f,0.58441f,0.00000f,0.87178f,0.84683f,0.82816f,0.79822f,0.76750f,0.74520f,0.71641f,0.68802f,0.66392f,0.64666f,0.63808f,0.63163f,0.62408f,0.63185f,0.63331f,0.64085f,0.64184f,0.64157f,0.64987f,0.64863f,0.64493f,0.63816f,0.63162f,0.62754f,0.61259f,0.59077f,0.56946f,0.54180f,0.50554f,0.47303f,0.44229f,0.53165f,0.00000f,0.85326f,0.82195f,0.80025f,0.76865f,0.74659f,0.71851f,0.68226f,0.65904f,0.63527f,0.62302f,0.61097f,0.60657f,0.60824f,0.61080f,0.61341f,0.62164f,0.62335f,0.62524f,0.62364f,0.62723f,0.62525f,0.61745f,0.61257f,0.60174f,0.58572f,0.56490f,0.53543f,0.50158f,0.47055f,0.42987f,0.40236f,0.48542f,
 0.00000f,1.00000f,0.99900f,1.00000f,0.99248f,0.96888f,0.99066f,0.99396f,0.98370f,0.96635f,0.96344f,0.99840f,0.98129f,0.98840f,0.98563f,0.98511f,0.98245f,0.97279f,0.97147f,0.97632f,0.97341f,0.97257f,0.99163f,0.98741f,0.98608f,0.98683f,0.99134f,0.98897f,0.98707f,0.98908f,0.99409f,0.99702f,1.00000f,0.00000f,0.99190f,0.98642f,0.98288f,0.97948f,0.97858f,0.97633f,0.97565f,0.97492f,0.97372f,0.97362f,0.97460f,0.97409f,0.97436f,0.97473f,0.97541f,0.97608f,0.97520f,0.97645f,0.97799f,0.97797f,0.97963f,0.98193f,0.98313f,0.98433f,0.98598f,0.98793f,0.98913f,0.99067f,0.99270f,0.99520f,0.99755f,1.00000f,0.00000f,0.99133f,0.98531f,0.98224f,0.98037f,0.97720f,0.97649f,0.97497f,0.97437f,0.97325f,0.97446f,0.97452f,0.97379f,0.97354f,0.97361f,0.97483f,0.97460f,0.97629f,0.97643f,0.97785f,0.97936f,0.97977f,0.98162f,0.98285f,0.98438f,0.98595f,0.98765f,0.98923f,0.99134f,0.99330f,0.99540f,0.99772f,1.00000f,0.00000f,0.98228f,0.98515f,0.98239f,0.98077f,0.97806f,0.97653f,0.97455f,0.97377f,0.97407f,0.97358f,0.97364f,0.97368f,0.97320f,0.97397f,0.97471f,0.97513f,0.97608f,0.97679f,0.97724f,0.97901f,0.97985f,0.98124f,0.98256f,0.98442f,0.98550f,0.98772f,0.98936f,0.99118f,0.99325f,0.99550f,0.99776f,1.00000f,0.00000f,0.93700f,0.98060f,0.98193f,0.97816f,0.97774f,0.97608f,0.97506f,0.97524f,0.97342f,0.97329f,0.97395f,0.97325f,0.97245f,0.97413f,0.97444f,0.97526f,0.97550f,0.97635f,0.97772f,0.97862f,0.98021f,0.98161f,0.98283f,0.98407f,0.98579f,0.98758f,0.98921f,0.99139f,0.99322f,0.99545f,0.99776f,1.00000f,0.00000f,0.90841f,0.95316f,0.97693f,0.97626f,0.97648f,0.97692f,0.97466f,0.97362f,0.97272f,0.97227f,0.97280f,0.97266f,0.97372f,0.97418f,0.97414f,0.97439f,0.97659f,0.97675f,0.97729f,0.97819f,0.97985f,0.98084f,0.98270f,0.98410f,0.98564f,0.98726f,0.98923f,0.99123f,0.99325f,0.99549f,0.99773f,1.00000f,0.00000f,0.89831f,0.92170f,0.95595f,0.96991f,0.97469f,0.97591f,0.97533f,0.97388f,0.97368f,0.97351f,0.97316f,0.97402f,0.97360f,0.97395f,0.97348f,0.97403f,0.97577f,0.97646f,0.97809f,0.97876f,0.97974f,0.98112f,0.98205f,0.98416f,0.98575f,0.98677f,0.98910f,0.99125f,0.99317f,0.99546f,0.99761f,1.00000f,0.00000f,0.89976f,0.90103f,0.92498f,0.95111f,0.96304f,0.97111f,0.97241f,0.97299f,0.97364f,0.97287f,0.97214f,0.97299f,0.97315f,0.97273f,0.97373f,0.97435f,0.97534f,0.97546f,0.97651f,0.97833f,0.97970f,0.98074f,0.98247f,0.98356f,0.98568f,0.98615f,0.98916f,0.99108f,0.99328f,0.99525f,0.99756f,1.00000f,0.00000f,0.90975f,0.89354f,0.90535f,0.92594f,0.94389f,0.95748f,0.96435f,0.97025f,0.97095f,0.97112f,0.97202f,0.97206f,0.97274f,0.97389f,0.97358f,0.97520f,0.97598f,0.97597f,0.97702f,0.97775f,0.97975f,0.98070f,0.98218f,0.98400f,0.98428f,0.98523f,0.98867f,0.99094f,0.99316f,0.99541f,0.99762f,1.00000f,0.00000f,0.91154f,0.89534f,0.89806f,0.90873f,0.92176f,0.93603f,0.95133f,0.95892f,0.96293f,0.96778f,0.97019f,0.97125f,0.97206f,0.97279f,0.97279f,0.97415f,0.97438f,0.97594f,0.97693f,0.97777f,0.97899f,0.97995f,0.98259f,0.98306f,0.98198f,0.98515f,0.98857f,0.99073f,0.99312f,0.99509f,0.99742f,1.00000f,0.00000f,0.92121f,0.90205f,0.89829f,0.89836f,0.90607f,0.91736f,0.93002f,0.94032f,0.94923f,0.95729f,0.96224f,0.96711f,0.96996f,0.97078f,0.97266f,0.97339f,0.97518f,0.97551f,0.97649f,0.97739f,0.97917f,0.98031f,0.98055f,0.98049f,0.98081f,0.98439f,0.98844f,0.99083f,0.99318f,0.99531f,0.99751f,1.00000f,0.00000f,0.93076f,0.90388f,0.89345f,0.89117f,0.89331f,0.89919f,0.90892f,0.92222f,0.93275f,0.94236f,0.94962f,0.95655f,0.96176f,0.96572f,0.96989f,0.97194f,0.97347f,0.97435f,0.97607f,0.97759f,0.97824f,0.97869f,0.97800f,0.97766f,0.97985f,0.98379f,0.98791f,0.99037f,0.99300f,0.99502f,0.99752f,1.00000f,0.00000f,0.93121f,0.90714f,0.89431f,0.89268f,0.88842f,0.89437f,0.90050f,0.90801f,0.91670f,0.92482f,0.93393f,0.94317f,0.95017f,0.95746f,0.96142f,0.96598f,0.96920f,0.97121f,0.97462f,0.97761f,0.97678f,0.97661f,0.97501f,0.97479f,0.97820f,0.98377f,0.98802f,0.99075f,0.99262f,0.99499f,0.99748f,1.00000f,0.00000f,0.93479f,0.91671f,0.90267f,0.89070f,0.88968f,0.88712f,0.89018f,0.89778f,0.90144f,0.90854f,0.91728f,0.92767f,0.93542f,0.94258f,0.94835f,0.95539f,0.96078f,0.96547f,0.96919f,0.97089f,0.97092f,0.97101f,0.97087f,0.97299f,0.97793f,0.98320f,0.98765f,0.99043f,0.99257f,0.99509f,0.99727f,1.00000f,0.00000f,0.94756f,0.92350f,0.90820f,0.89736f,0.88993f,0.88546f,0.88710f,0.88779f,0.89341f,0.89289f,0.89993f,0.91175f,0.91856f,0.92453f,0.93454f,0.93960f,0.94744f,0.95464f,0.95704f,0.95845f,0.96026f,0.96028f,0.96526f,0.96924f,0.97523f,0.98263f,0.98687f,0.99004f,0.99261f,0.99479f,0.99705f,1.00000f,0.00000f,0.94377f,0.92558f,0.91013f,0.89927f,0.89255f,0.88482f,0.88221f,0.88421f,0.88770f,0.88619f,0.89092f,0.89537f,0.90593f,0.91070f,0.91845f,0.92489f,0.93112f,0.93389f,0.93871f,0.94343f,0.94420f,0.94757f,0.95505f,0.96347f,0.97117f,0.97828f,0.98451f,0.98919f,0.99274f,0.99471f,0.99719f,1.00000f,0.00000f,0.94379f,0.92768f,0.91542f,0.90742f,0.89321f,0.88785f,0.88634f,0.88258f,0.88377f,0.88452f,0.88672f,0.88794f,0.89042f,0.89547f,0.90158f,0.90466f,0.90954f,0.91414f,0.91739f,0.91974f,0.92473f,0.93415f,0.94138f,0.95188f,0.96270f,0.97249f,0.98047f,0.98555f,0.99149f,0.99402f,0.99676f,1.00000f,0.00000f,0.94875f,0.93375f,0.92419f,0.90479f,0.89582f,0.88582f,0.88581f,0.87885f,0.87924f,0.87916f,0.87724f,0.88257f,0.88051f,0.87957f,0.88432f,0.88310f,0.88755f,0.89041f,0.89249f,0.89848f,0.90353f,0.91358f,0.92676f,0.93677f,0.94781f,0.96130f,0.97042f,0.97700f,0.98469f,0.99092f,0.99615f,1.00000f,0.00000f,0.94341f,0.93487f,0.92740f,0.91305f,0.89926f,0.89194f,0.88360f,0.88200f,0.87879f,0.87572f,0.87377f,0.87279f,0.86859f,0.87006f,0.86714f,0.86938f,0.86570f,0.86449f,0.87038f,0.87360f,0.88287f,0.89744f,0.90790f,0.91841f,0.93284f,0.94559f,0.95663f,0.96528f,0.97581f,0.98276f,0.99123f,0.99999f,0.00000f,0.94438f,0.93929f,0.92623f,0.92190f,0.90554f,0.89619f,0.88905f,0.88131f,0.87948f,0.86999f,0.86825f,0.86627f,0.86164f,0.85465f,0.85008f,0.84442f,0.84367f,0.84364f,0.84406f,0.85196f,0.86135f,0.87354f,0.88629f,0.89958f,0.91237f,0.92512f,0.93711f,0.94885f,0.95998f,0.96938f,0.98202f,0.99950f,0.00000f,0.94366f,0.94127f,0.92544f,0.92377f,0.90706f,0.90106f,0.88899f,0.88301f,0.87728f,0.86649f,0.86102f,0.85677f,0.84474f,0.84002f,0.83337f,0.82658f,0.82478f,0.82268f,0.82689f,0.83316f,0.83913f,0.85299f,0.86511f,0.87669f,0.89120f,0.90323f,0.91557f,0.92750f,0.93742f,0.95046f,0.96296f,0.99690f,0.00000f,0.94270f,0.94028f,0.93249f,0.92395f,0.91488f,0.90321f,0.89115f,0.87946f,0.87335f,0.86399f,0.85820f,0.84427f,0.83359f,0.82124f,0.81217f,0.80841f,0.80453f,0.80599f,0.80758f,0.81230f,0.82258f,0.83298f,0.84456f,0.85569f,0.86588f,0.87753f,0.88853f,0.90241f,0.91279f,0.92318f,0.93670f,0.98889f,0.00000f,0.94454f,0.93311f,0.92535f,0.93015f,0.91067f,0.89826f,0.89294f,0.88411f,0.87262f,0.85302f,0.84177f,0.82899f,0.81542f,0.80795f,0.79696f,0.78791f,0.78734f,0.78733f,0.78673f,0.79323f,0.80609f,0.81231f,0.82410f,0.83410f,0.84516f,0.85074f,0.85969f,0.86781f,0.88006f,0.89158f,0.90182f,0.97087f,0.00000f,0.94199f,0.93690f,0.92669f,0.91872f,0.91486f,0.90022f,0.88502f,0.87507f,0.85990f,0.84424f,0.82458f,0.81394f,0.79499f,0.78919f,0.77711f,0.77132f,0.77176f,0.77350f,0.77486f,0.78389f,0.78764f,0.79466f,0.80184f,0.81005f,0.81875f,0.82582f,0.83165f,0.83875f,0.84491f,0.85232f,0.85815f,0.94140f,0.00000f,0.93674f,0.94065f,0.93355f,0.92305f,0.90975f,0.89069f,0.87822f,0.86403f,0.84568f,0.82542f,0.81471f,0.79468f,0.78146f,0.76966f,0.76109f,0.75667f,0.75476f,0.75438f,0.76381f,0.76922f,0.77270f,0.77811f,0.78272f,0.78567f,0.79113f,0.79535f,0.80176f,0.80349f,0.80806f,0.81047f,0.80791f,0.90345f,0.00000f,0.94069f,0.93144f,0.92347f,0.91111f,0.90141f,0.89175f,0.86968f,0.85614f,0.83363f,0.81271f,0.79375f,0.77351f,0.75437f,0.74553f,0.74268f,0.74123f,0.73959f,0.73958f,0.74793f,0.75173f,0.75546f,0.76040f,0.76230f,0.76192f,0.76348f,0.76707f,0.76727f,0.76533f,0.76660f,0.76005f,0.75350f,0.85687f,0.00000f,0.93360f,0.93619f,0.91662f,0.90946f,0.89532f,0.87465f,0.85458f,0.82968f,0.80911f,0.78949f,0.76765f,0.75491f,0.73793f,0.73117f,0.72479f,0.72687f,0.72490f,0.72880f,0.72888f,0.73269f,0.74021f,0.74340f,0.74251f,0.74071f,0.74045f,0.73811f,0.73319f,0.73026f,0.72125f,0.71209f,0.69492f,0.80516f,0.00000f,0.92932f,0.92728f,0.91079f,0.89568f,0.88635f,0.86312f,0.83564f,0.81260f,0.78616f,0.76530f,0.74558f,0.72710f,0.72112f,0.71297f,0.71075f,0.70860f,0.70589f,0.71603f,0.71747f,0.71800f,0.72102f,0.72254f,0.72048f,0.71426f,0.71328f,0.70810f,0.69933f,0.69236f,0.67575f,0.66070f,0.64118f,0.74915f,0.00000f,0.92855f,0.91268f,0.89883f,0.88387f,0.86298f,0.84161f,0.82179f,0.79436f,0.76761f,0.74175f,0.72047f,0.71140f,0.70311f,0.69573f,0.69057f,0.69612f,0.69776f,0.70056f,0.70561f,0.70178f,0.70537f,0.70123f,0.70192f,0.69463f,0.69285f,0.67715f,0.66479f,0.65026f,0.63421f,0.61105f,0.58502f,0.69288f,0.00000f,0.91257f,0.90100f,0.89002f,0.86208f,0.83709f,0.81741f,0.79091f,0.76012f,0.73502f,0.71239f,0.69808f,0.68604f,0.67923f,0.67468f,0.67597f,0.67696f,0.68164f,0.68117f,0.68313f,0.68289f,0.68285f,0.68200f,0.67627f,0.66776f,0.65997f,0.64673f,0.62681f,0.61316f,0.58389f,0.55932f,0.53221f,0.63626f,0.00000f,0.89318f,0.87952f,0.86334f,0.84840f,0.81644f,0.79184f,0.76640f,0.73597f,0.70855f,0.68514f,0.67665f,0.66362f,0.66121f,0.65824f,0.65912f,0.66038f,0.66022f,0.66267f,0.66460f,0.66715f,0.66238f,0.65815f,0.65010f,0.64115f,0.62967f,0.61570f,0.59542f,0.56915f,0.54186f,0.51294f,0.48417f,0.58021f,0.00000f,0.88383f,0.86111f,0.84194f,0.82156f,0.78944f,0.76532f,0.73452f,0.71004f,0.67692f,0.66300f,0.64854f,0.64042f,0.63929f,0.63946f,0.64417f,0.63904f,0.64341f,0.64641f,0.64944f,0.64268f,0.64520f,0.63540f,0.62828f,0.61813f,0.60480f,0.58798f,0.56125f,0.53094f,0.50104f,0.46727f,0.43618f,0.52684f,0.00000f,0.86210f,0.84034f,0.81864f,0.79225f,0.76333f,0.72844f,0.70293f,0.67157f,0.65431f,0.64130f,0.62153f,0.62320f,0.61800f,0.62020f,0.62223f,0.62177f,0.62782f,0.62788f,0.62956f,0.62546f,0.62045f,0.61675f,0.60353f,0.59365f,0.57663f,0.55264f,0.52732f,0.49459f,0.46107f,0.42538f,0.39435f,0.47881f,
 0.00000f,1.00000f,0.99871f,1.00000f,0.98972f,0.96915f,0.99099f,0.99406f,0.98334f,0.96679f,0.96441f,0.99777f,0.98213f,0.98894f,0.98553f,0.98492f,0.98233f,0.97221f,0.97133f,0.97619f,0.97288f,0.97243f,0.99137f,0.98758f,0.98589f,0.98748f,0.99169f,0.99076f,0.98826f,0.99026f,0.99181f,0.99653f,1.00000f,0.00000f,0.99267f,0.98823f,0.98372f,0.98078f,0.97661f,0.97749f,0.97625f,0.97592f,0.97310f,0.97336f,0.97486f,0.97352f,0.97493f,0.97446f,0.97593f,0.97636f,0.97548f,0.97679f,0.97820f,0.97857f,0.97894f,0.98190f,0.98311f,0.98427f,0.98585f,0.98797f,0.98982f,0.99079f,0.99288f,0.99485f,0.99753f,1.00000f,0.00000f,0.99290f,0.98507f,0.98195f,0.97958f,0.97825f,0.97804f,0.97520f,0.97554f,0.97412f,0.97332f,0.97436f,0.97380f,0.97434f,0.97432f,0.97480f,0.97493f,0.97584f,0.97627f,0.97790f,0.97881f,0.97953f,0.98118f,0.98284f,0.98400f,0.98606f,0.98746f,0.98954f,0.99121f,0.99339f,0.99535f,0.99776f,1.00000f,0.00000f,0.98515f,0.98513f,0.98090f,0.97913f,0.97658f,0.97727f,0.97456f,0.97483f,0.97315f,0.97355f,0.97290f,0.97382f,0.97414f,0.97333f,0.97478f,0.97597f,0.97547f,0.97659f,0.97837f,0.97926f,0.98067f,0.98133f,0.98281f,0.98451f,0.98592f,0.98755f,0.98944f,0.99148f,0.99331f,0.99546f,0.99776f,1.00000f,0.00000f,0.93596f,0.98243f,0.98271f,0.97933f,0.97728f,0.97703f,0.97528f,0.97495f,0.97467f,0.97333f,0.97348f,0.97426f,0.97347f,0.97337f,0.97481f,0.97491f,0.97602f,0.97646f,0.97777f,0.97925f,0.98047f,0.98127f,0.98269f,0.98424f,0.98568f,0.98746f,0.98905f,0.99127f,0.99326f,0.99544f,0.99777f,1.00000f,0.00000f,0.91310f,0.95489f,0.97665f,0.97897f,0.97721f,0.97588f,0.97411f,0.97404f,0.97393f,0.97299f,0.97221f,0.97418f,0.97261f,0.97377f,0.97431f,0.97476f,0.97621f,0.97648f,0.97704f,0.97851f,0.98056f,0.98140f,0.98280f,0.98456f,0.98551f,0.98749f,0.98914f,0.99116f,0.99322f,0.99542f,0.99772f,1.00000f,0.00000f,0.89550f,0.92325f,0.95848f,0.97198f,0.97459f,0.97502f,0.97383f,0.97347f,0.97299f,0.97214f,0.97264f,0.97221f,0.97214f,0.97383f,0.97356f,0.97459f,0.97560f,0.97655f,0.97696f,0.97806f,0.97996f,0.98100f,0.98235f,0.98414f,0.98576f,0.98703f,0.98897f,0.99130f,0.99336f,0.99537f,0.99757f,1.00000f,0.00000f,0.90249f,0.90517f,0.92435f,0.94924f,0.96419f,0.97181f,0.97188f,0.97265f,0.97392f,0.97235f,0.97274f,0.97265f,0.97313f,0.97304f,0.97318f,0.97437f,0.97552f,0.97610f,0.97741f,0.97849f,0.97933f,0.98132f,0.98222f,0.98446f,0.98581f,0.98616f,0.98879f,0.99100f,0.99321f,0.99535f,0.99761f,1.00000f,0.00000f,0.90727f,0.89690f,0.90719f,0.92615f,0.94595f,0.95713f,0.96428f,0.96819f,0.97045f,0.97148f,0.97221f,0.97252f,0.97208f,0.97203f,0.97387f,0.97434f,0.97464f,0.97639f,0.97730f,0.97852f,0.97941f,0.98080f,0.98207f,0.98406f,0.98486f,0.98552f,0.98832f,0.99093f,0.99301f,0.99532f,0.99765f,1.00000f,0.00000f,0.92128f,0.88853f,0.89888f,0.90743f,0.92148f,0.93708f,0.95198f,0.95756f,0.96411f,0.96877f,0.96868f,0.97101f,0.97142f,0.97296f,0.97363f,0.97386f,0.97456f,0.97514f,0.97656f,0.97819f,0.97917f,0.98005f,0.98204f,0.98364f,0.98322f,0.98446f,0.98829f,0.99105f,0.99323f,0.99549f,0.99764f,1.00000f,0.00000f,0.92739f,0.89782f,0.89060f,0.89632f,0.90296f,0.91933f,0.92980f,0.93953f,0.95107f,0.95858f,0.96305f,0.96675f,0.96974f,0.97020f,0.97183f,0.97330f,0.97474f,0.97562f,0.97658f,0.97744f,0.97905f,0.98015f,0.98134f,0.98149f,0.98086f,0.98370f,0.98794f,0.99064f,0.99297f,0.99506f,0.99750f,1.00000f,0.00000f,0.93290f,0.90781f,0.88808f,0.88715f,0.89374f,0.90328f,0.91034f,0.92365f,0.93316f,0.94259f,0.94964f,0.95737f,0.96202f,0.96473f,0.96895f,0.97187f,0.97336f,0.97444f,0.97595f,0.97725f,0.97841f,0.97970f,0.97920f,0.97911f,0.97907f,0.98304f,0.98752f,0.99103f,0.99301f,0.99498f,0.99751f,1.00000f,0.00000f,0.92982f,0.91067f,0.89064f,0.88980f,0.89301f,0.89494f,0.89806f,0.90617f,0.91475f,0.92518f,0.93330f,0.94340f,0.95021f,0.95460f,0.96134f,0.96370f,0.96922f,0.97110f,0.97410f,0.97692f,0.97778f,0.97672f,0.97646f,0.97607f,0.97835f,0.98280f,0.98720f,0.99022f,0.99280f,0.99487f,0.99722f,1.00000f,0.00000f,0.93959f,0.91405f,0.90529f,0.88989f,0.88971f,0.89013f,0.89286f,0.89468f,0.90003f,0.91051f,0.91752f,0.92521f,0.93525f,0.94049f,0.94830f,0.95492f,0.96017f,0.96464f,0.96901f,0.97183f,0.97215f,0.97300f,0.97141f,0.97369f,0.97657f,0.98231f,0.98693f,0.99020f,0.99262f,0.99484f,0.99722f,1.00000f,0.00000f,0.94423f,0.92519f,0.90997f,0.89382f,0.88658f,0.88361f,0.88300f,0.88477f,0.88966f,0.89552f,0.90127f,0.91026f,0.91873f,0.92465f,0.93436f,0.93967f,0.94836f,0.95348f,0.95869f,0.96001f,0.96154f,0.96252f,0.96527f,0.96860f,0.97469f,0.98106f,0.98617f,0.99003f,0.99230f,0.99484f,0.99706f,1.00000f,0.00000f,0.94177f,0.92913f,0.91164f,0.90000f,0.88997f,0.88475f,0.88247f,0.88278f,0.88540f,0.88877f,0.88848f,0.89534f,0.90339f,0.90841f,0.91695f,0.92409f,0.93137f,0.93430f,0.94187f,0.94417f,0.94651f,0.94918f,0.95489f,0.96251f,0.97048f,0.97759f,0.98465f,0.98899f,0.99187f,0.99500f,0.99711f,1.00000f,0.00000f,0.94467f,0.93197f,0.91137f,0.90524f,0.89550f,0.88920f,0.88663f,0.88337f,0.88197f,0.87890f,0.88901f,0.89003f,0.89492f,0.89561f,0.90049f,0.90697f,0.91068f,0.91620f,0.92088f,0.92438f,0.92788f,0.93372f,0.94094f,0.95127f,0.96093f,0.97065f,0.97873f,0.98582f,0.99003f,0.99420f,0.99716f,1.00000f,0.00000f,0.94336f,0.93246f,0.92148f,0.90627f,0.89368f,0.89154f,0.87885f,0.87910f,0.88082f,0.87680f,0.87898f,0.87917f,0.88414f,0.88440f,0.88345f,0.88989f,0.89290f,0.89340f,0.89512f,0.90015f,0.90668f,0.91531f,0.92361f,0.93565f,0.94930f,0.95968f,0.97004f,0.97788f,0.98455f,0.99071f,0.99605f,1.00000f,0.00000f,0.94731f,0.93633f,0.92371f,0.91153f,0.90089f,0.89556f,0.88760f,0.88143f,0.87485f,0.87760f,0.87282f,0.87385f,0.87332f,0.87515f,0.86889f,0.86889f,0.86860f,0.86905f,0.87311f,0.87899f,0.88593f,0.89382f,0.90476f,0.91824f,0.93217f,0.94298f,0.95610f,0.96505f,0.97449f,0.98346f,0.99191f,0.99999f,0.00000f,0.94401f,0.93912f,0.93120f,0.92015f,0.90710f,0.89333f,0.89067f,0.88280f,0.87387f,0.87304f,0.86883f,0.87120f,0.86180f,0.86059f,0.85462f,0.85223f,0.84818f,0.84893f,0.85143f,0.85589f,0.86363f,0.87304f,0.88510f,0.89775f,0.90964f,0.92317f,0.93724f,0.94833f,0.95877f,0.96884f,0.98125f,0.99950f,0.00000f,0.94891f,0.94089f,0.92892f,0.91640f,0.90784f,0.89686f,0.89214f,0.88517f,0.87923f,0.87171f,0.86840f,0.85887f,0.85563f,0.84605f,0.83853f,0.83179f,0.82864f,0.82467f,0.82425f,0.83187f,0.83990f,0.85225f,0.86361f,0.87760f,0.88843f,0.90164f,0.91445f,0.92727f,0.93742f,0.94989f,0.96230f,0.99678f,0.00000f,0.95084f,0.94360f,0.93085f,0.92023f,0.91887f,0.90253f,0.89197f,0.88399f,0.87673f,0.87005f,0.86510f,0.85109f,0.84181f,0.83152f,0.82152f,0.81665f,0.80678f,0.80765f,0.80754f,0.81410f,0.82104f,0.83278f,0.84213f,0.85456f,0.86541f,0.87709f,0.88714f,0.90022f,0.91099f,0.92252f,0.93739f,0.98905f,0.00000f,0.94758f,0.93473f,0.93019f,0.92182f,0.91185f,0.90414f,0.89767f,0.88471f,0.87211f,0.86066f,0.85135f,0.83699f,0.82430f,0.81016f,0.80608f,0.79693f,0.78999f,0.79160f,0.79504f,0.79719f,0.80373f,0.81181f,0.82166f,0.83221f,0.84116f,0.85321f,0.86198f,0.86918f,0.88248f,0.89145f,0.90164f,0.97107f,0.00000f,0.94424f,0.94001f,0.93571f,0.92463f,0.91522f,0.90262f,0.89046f,0.88164f,0.86514f,0.84927f,0.83784f,0.81988f,0.80202f,0.78915f,0.78485f,0.77951f,0.77393f,0.77466f,0.77465f,0.78048f,0.78750f,0.79865f,0.80326f,0.80807f,0.81853f,0.82424f,0.83208f,0.83740f,0.84647f,0.85266f,0.85676f,0.94135f,0.00000f,0.93083f,0.93252f,0.92893f,0.92201f,0.91495f,0.89777f,0.88657f,0.87195f,0.85195f,0.83614f,0.82446f,0.80706f,0.78806f,0.77578f,0.76248f,0.75956f,0.75814f,0.76096f,0.76211f,0.76658f,0.77207f,0.77603f,0.78230f,0.78658f,0.79228f,0.79820f,0.79910f,0.80363f,0.80743f,0.80749f,0.80661f,0.90383f,0.00000f,0.93877f,0.92750f,0.93455f,0.91794f,0.90679f,0.89282f,0.87718f,0.85296f,0.84197f,0.82041f,0.79953f,0.78519f,0.77024f,0.75337f,0.74855f,0.74756f,0.74435f,0.74627f,0.74550f,0.75171f,0.75580f,0.76107f,0.76060f,0.76256f,0.76628f,0.76529f,0.76815f,0.76711f,0.76320f,0.75685f,0.75302f,0.85731f,0.00000f,0.94196f,0.93591f,0.92207f,0.91446f,0.89747f,0.88144f,0.86410f,0.84598f,0.82272f,0.79893f,0.78593f,0.76499f,0.74663f,0.73550f,0.72957f,0.73061f,0.72615f,0.73288f,0.73317f,0.73426f,0.74163f,0.74325f,0.74224f,0.74101f,0.73881f,0.73681f,0.73221f,0.72681f,0.72032f,0.70716f,0.69441f,0.80194f,0.00000f,0.92665f,0.92172f,0.92284f,0.90098f,0.88714f,0.86983f,0.85541f,0.82976f,0.79968f,0.77644f,0.75360f,0.74075f,0.72708f,0.71781f,0.71833f,0.71110f,0.71129f,0.71295f,0.72086f,0.71901f,0.72044f,0.72116f,0.71925f,0.71394f,0.71197f,0.70659f,0.69538f,0.68762f,0.67117f,0.65402f,0.63874f,0.74884f,0.00000f,0.92942f,0.91929f,0.90203f,0.89420f,0.87182f,0.84767f,0.82343f,0.80923f,0.77900f,0.75169f,0.73274f,0.71935f,0.70945f,0.70280f,0.70029f,0.69586f,0.69746f,0.70255f,0.69910f,0.70687f,0.69835f,0.69887f,0.69789f,0.69642f,0.68486f,0.67479f,0.65952f,0.64493f,0.62992f,0.60393f,0.58178f,0.69098f,0.00000f,0.91829f,0.90727f,0.89696f,0.87460f,0.85184f,0.83146f,0.80359f,0.77546f,0.75217f,0.72148f,0.70834f,0.69962f,0.69164f,0.68212f,0.68095f,0.68368f,0.67858f,0.68261f,0.68351f,0.68727f,0.68635f,0.68114f,0.67459f,0.66899f,0.65625f,0.64409f,0.62505f,0.60342f,0.58098f,0.55647f,0.53054f,0.63154f,0.00000f,0.90404f,0.89371f,0.88128f,0.86296f,0.82832f,0.80521f,0.77611f,0.75478f,0.72976f,0.70679f,0.68744f,0.67674f,0.67057f,0.66443f,0.65771f,0.66744f,0.66528f,0.66587f,0.66746f,0.66729f,0.66237f,0.65586f,0.65009f,0.63934f,0.62744f,0.60946f,0.58928f,0.56487f,0.53440f,0.50932f,0.47832f,0.57672f,0.00000f,0.88925f,0.87776f,0.85366f,0.82744f,0.80464f,0.77780f,0.75284f,0.72366f,0.69632f,0.67627f,0.66359f,0.65689f,0.65607f,0.64599f,0.64498f,0.64482f,0.64862f,0.64953f,0.64431f,0.64830f,0.64069f,0.63576f,0.62505f,0.61604f,0.59792f,0.57786f,0.55165f,0.52442f,0.49252f,0.46375f,0.43163f,0.52385f,0.00000f,0.87127f,0.85418f,0.83156f,0.80647f,0.77992f,0.74873f,0.72007f,0.68651f,0.67174f,0.64731f,0.63801f,0.62890f,0.62824f,0.62956f,0.62292f,0.62443f,0.62701f,0.62895f,0.62504f,0.62038f,0.61723f,0.61266f,0.60354f,0.58545f,0.57205f,0.54425f,0.51934f,0.48475f,0.45273f,0.42086f,0.38846f,0.47600f,
 0.00000f,1.00000f,0.99680f,1.00000f,0.99085f,0.96766f,0.99067f,0.99357f,0.98357f,0.96705f,0.96436f,0.99888f,0.98151f,0.98913f,0.98575f,0.98328f,0.98151f,0.97252f,0.97074f,0.97628f,0.97366f,0.97228f,0.99131f,0.98743f,0.98553f,0.98669f,0.99147f,0.99094f,0.98514f,0.99119f,0.99286f,0.99588f,1.00000f,0.00000f,0.99284f,0.98531f,0.98381f,0.98009f,0.97720f,0.97661f,0.97660f,0.97590f,0.97316f,0.97309f,0.97556f,0.97389f,0.97442f,0.97381f,0.97608f,0.97585f,0.97440f,0.97602f,0.97831f,0.97887f,0.97944f,0.98179f,0.98285f,0.98419f,0.98568f,0.98792f,0.98959f,0.99043f,0.99295f,0.99507f,0.99743f,1.00000f,0.00000f,0.99151f,0.98554f,0.98157f,0.97936f,0.97885f,0.97689f,0.97536f,0.97485f,0.97434f,0.97386f,0.97283f,0.97321f,0.97373f,0.97375f,0.97390f,0.97542f,0.97613f,0.97683f,0.97770f,0.97849f,0.97989f,0.98129f,0.98266f,0.98431f,0.98583f,0.98761f,0.98932f,0.99118f,0.99313f,0.99548f,0.99773f,1.00000f,0.00000f,0.98442f,0.98520f,0.98203f,0.97933f,0.97795f,0.97749f,0.97576f,0.97484f,0.97319f,0.97319f,0.97396f,0.97345f,0.97432f,0.97401f,0.97404f,0.97478f,0.97614f,0.97716f,0.97732f,0.97910f,0.98030f,0.98135f,0.98265f,0.98435f,0.98598f,0.98767f,0.98967f,0.99122f,0.99340f,0.99549f,0.99778f,1.00000f,0.00000f,0.93895f,0.98162f,0.98332f,0.97914f,0.97539f,0.97561f,0.97462f,0.97407f,0.97393f,0.97291f,0.97325f,0.97340f,0.97333f,0.97411f,0.97423f,0.97529f,0.97609f,0.97657f,0.97691f,0.97895f,0.98039f,0.98141f,0.98253f,0.98412f,0.98556f,0.98768f,0.98949f,0.99148f,0.99332f,0.99538f,0.99778f,1.00000f,0.00000f,0.90639f,0.95700f,0.97802f,0.97882f,0.97709f,0.97470f,0.97550f,0.97323f,0.97386f,0.97413f,0.97303f,0.97330f,0.97424f,0.97352f,0.97443f,0.97487f,0.97624f,0.97670f,0.97813f,0.97887f,0.98010f,0.98148f,0.98268f,0.98409f,0.98575f,0.98758f,0.98884f,0.99108f,0.99330f,0.99545f,0.99777f,1.00000f,0.00000f,0.89925f,0.92328f,0.95409f,0.97222f,0.97418f,0.97455f,0.97302f,0.97398f,0.97281f,0.97263f,0.97220f,0.97272f,0.97215f,0.97313f,0.97347f,0.97458f,0.97532f,0.97649f,0.97701f,0.97846f,0.97922f,0.98097f,0.98317f,0.98400f,0.98575f,0.98758f,0.98854f,0.99099f,0.99323f,0.99551f,0.99768f,1.00000f,0.00000f,0.90286f,0.90474f,0.92965f,0.94886f,0.96275f,0.97064f,0.97301f,0.97309f,0.97241f,0.97265f,0.97234f,0.97384f,0.97226f,0.97336f,0.97369f,0.97480f,0.97495f,0.97601f,0.97707f,0.97810f,0.97929f,0.98137f,0.98188f,0.98379f,0.98525f,0.98711f,0.98827f,0.99086f,0.99314f,0.99539f,0.99760f,1.00000f,0.00000f,0.91114f,0.89421f,0.90556f,0.92698f,0.94633f,0.95655f,0.96467f,0.96867f,0.96971f,0.97177f,0.97237f,0.97173f,0.97245f,0.97285f,0.97386f,0.97440f,0.97522f,0.97534f,0.97696f,0.97803f,0.97929f,0.98107f,0.98244f,0.98410f,0.98507f,0.98579f,0.98789f,0.99105f,0.99335f,0.99533f,0.99758f,1.00000f,0.00000f,0.91049f,0.89436f,0.89617f,0.90764f,0.92062f,0.93901f,0.94947f,0.95841f,0.96358f,0.96866f,0.96960f,0.97191f,0.97340f,0.97247f,0.97292f,0.97416f,0.97430f,0.97589f,0.97684f,0.97821f,0.97874f,0.98055f,0.98180f,0.98394f,0.98403f,0.98435f,0.98770f,0.99080f,0.99272f,0.99551f,0.99747f,1.00000f,0.00000f,0.92693f,0.90422f,0.89310f,0.89925f,0.90278f,0.91557f,0.92848f,0.94200f,0.95161f,0.95633f,0.96310f,0.96695f,0.96845f,0.97190f,0.97317f,0.97428f,0.97349f,0.97470f,0.97655f,0.97820f,0.97914f,0.98022f,0.98176f,0.98170f,0.98186f,0.98316f,0.98704f,0.99065f,0.99293f,0.99532f,0.99755f,1.00000f,0.00000f,0.92780f,0.91209f,0.89148f,0.89247f,0.89430f,0.90023f,0.91085f,0.92291f,0.93145f,0.94365f,0.94957f,0.95644f,0.96122f,0.96764f,0.96913f,0.97141f,0.97357f,0.97408f,0.97631f,0.97749f,0.97859f,0.97958f,0.98021f,0.97997f,0.97966f,0.98237f,0.98689f,0.99060f,0.99284f,0.99517f,0.99753f,1.00000f,0.00000f,0.93764f,0.90546f,0.89863f,0.89310f,0.88852f,0.89259f,0.89910f,0.90684f,0.91331f,0.92681f,0.93319f,0.94217f,0.95112f,0.95601f,0.96122f,0.96522f,0.96897f,0.97125f,0.97391f,0.97616f,0.97847f,0.97844f,0.97818f,0.97686f,0.97771f,0.98172f,0.98710f,0.99029f,0.99252f,0.99489f,0.99730f,1.00000f,0.00000f,0.94018f,0.92131f,0.90150f,0.89102f,0.89043f,0.88904f,0.89131f,0.89432f,0.90166f,0.91299f,0.91879f,0.92653f,0.93331f,0.94038f,0.94915f,0.95445f,0.95967f,0.96528f,0.97050f,0.97156f,0.97401f,0.97531f,0.97359f,0.97419f,0.97648f,0.98106f,0.98657f,0.98988f,0.99275f,0.99495f,0.99723f,1.00000f,0.00000f,0.94324f,0.91956f,0.90244f,0.89655f,0.88968f,0.88948f,0.88373f,0.88750f,0.89066f,0.89655f,0.90178f,0.91114f,0.91749f,0.92585f,0.93336f,0.94070f,0.94640f,0.95353f,0.95927f,0.96197f,0.96363f,0.96541f,0.96626f,0.96900f,0.97439f,0.98057f,0.98594f,0.99006f,0.99257f,0.99484f,0.99711f,1.00000f,0.00000f,0.93487f,0.92001f,0.91377f,0.89773f,0.89106f,0.88551f,0.88100f,0.88287f,0.88661f,0.89165f,0.88965f,0.89704f,0.90504f,0.91000f,0.91809f,0.92451f,0.93123f,0.93898f,0.94102f,0.94538f,0.95018f,0.95222f,0.95650f,0.96248f,0.96924f,0.97679f,0.98439f,0.98875f,0.99273f,0.99464f,0.99702f,1.00000f,0.00000f,0.94897f,0.93296f,0.91245f,0.90484f,0.89574f,0.88952f,0.88373f,0.88149f,0.88023f,0.88041f,0.88444f,0.88728f,0.89205f,0.89436f,0.89982f,0.90872f,0.91560f,0.91695f,0.92334f,0.92487f,0.92951f,0.93565f,0.94203f,0.95127f,0.96134f,0.97009f,0.97883f,0.98461f,0.98996f,0.99438f,0.99720f,1.00000f,0.00000f,0.95197f,0.93487f,0.91608f,0.90400f,0.89663f,0.88691f,0.88631f,0.88216f,0.88014f,0.87931f,0.87781f,0.87918f,0.88471f,0.88249f,0.88806f,0.89127f,0.89315f,0.89934f,0.90067f,0.90538f,0.90840f,0.91652f,0.92469f,0.93604f,0.94705f,0.95963f,0.96996f,0.97657f,0.98485f,0.99082f,0.99592f,1.00000f,0.00000f,0.94321f,0.93109f,0.92738f,0.91020f,0.90100f,0.89294f,0.88616f,0.88145f,0.87709f,0.87511f,0.87488f,0.87562f,0.87313f,0.87329f,0.87121f,0.87124f,0.87073f,0.87612f,0.87341f,0.88080f,0.88633f,0.89518f,0.90508f,0.91793f,0.93096f,0.94380f,0.95527f,0.96484f,0.97421f,0.98298f,0.99154f,0.99999f,0.00000f,0.94208f,0.92961f,0.92642f,0.91837f,0.90315f,0.89710f,0.88330f,0.88021f,0.87814f,0.87508f,0.87238f,0.87283f,0.86787f,0.86238f,0.86258f,0.85816f,0.85146f,0.85092f,0.85218f,0.85625f,0.86334f,0.87459f,0.88186f,0.89748f,0.90976f,0.92537f,0.93597f,0.94766f,0.95907f,0.96994f,0.98115f,0.99949f,0.00000f,0.94342f,0.93916f,0.93180f,0.91525f,0.90410f,0.89962f,0.89312f,0.88878f,0.87875f,0.87359f,0.86642f,0.85893f,0.85552f,0.84776f,0.84233f,0.84016f,0.83390f,0.82756f,0.83314f,0.83553f,0.84229f,0.85180f,0.86421f,0.87885f,0.88827f,0.90167f,0.91338f,0.92536f,0.93797f,0.95037f,0.96207f,0.99684f,0.00000f,0.94954f,0.93951f,0.93477f,0.91860f,0.91064f,0.89848f,0.89339f,0.88350f,0.87833f,0.87342f,0.86211f,0.85254f,0.84412f,0.83547f,0.82503f,0.81894f,0.81323f,0.81227f,0.81267f,0.81568f,0.82290f,0.83182f,0.84398f,0.85402f,0.86402f,0.87703f,0.88824f,0.90010f,0.91003f,0.92330f,0.93533f,0.98920f,0.00000f,0.94129f,0.93769f,0.94018f,0.91958f,0.91386f,0.90172f,0.89381f,0.88896f,0.87308f,0.86069f,0.84743f,0.84144f,0.82874f,0.81684f,0.80752f,0.80283f,0.79567f,0.79240f,0.79329f,0.80308f,0.80560f,0.81282f,0.82344f,0.83119f,0.83759f,0.85012f,0.85819f,0.87008f,0.88057f,0.89209f,0.90091f,0.97047f,0.00000f,0.93337f,0.93934f,0.93388f,0.92228f,0.91681f,0.90376f,0.89671f,0.87827f,0.86826f,0.85844f,0.83724f,0.82961f,0.80903f,0.80465f,0.79118f,0.78684f,0.77824f,0.77662f,0.78344f,0.78301f,0.78724f,0.79346f,0.79930f,0.80797f,0.81588f,0.82419f,0.82784f,0.83782f,0.84578f,0.85114f,0.85628f,0.94070f,0.00000f,0.94008f,0.93743f,0.93122f,0.92409f,0.91425f,0.90441f,0.88940f,0.87411f,0.86421f,0.84410f,0.82848f,0.80975f,0.79588f,0.78139f,0.77065f,0.76790f,0.76504f,0.76568f,0.76422f,0.77059f,0.77110f,0.77777f,0.78479f,0.78625f,0.78918f,0.79405f,0.79727f,0.80312f,0.80572f,0.80778f,0.80554f,0.90198f,0.00000f,0.94312f,0.93849f,0.92860f,0.91992f,0.90674f,0.89904f,0.87912f,0.86360f,0.84854f,0.82972f,0.80925f,0.79673f,0.77593f,0.76687f,0.75561f,0.75505f,0.74764f,0.74997f,0.74773f,0.75551f,0.75516f,0.75556f,0.76352f,0.76202f,0.76175f,0.76289f,0.76543f,0.76681f,0.76144f,0.75419f,0.75305f,0.85415f,0.00000f,0.93020f,0.92613f,0.92539f,0.91555f,0.90183f,0.88478f,0.87052f,0.85145f,0.83099f,0.80875f,0.78952f,0.77480f,0.75582f,0.74614f,0.74094f,0.73229f,0.73149f,0.73068f,0.73513f,0.73981f,0.74013f,0.74189f,0.73847f,0.73773f,0.73584f,0.73591f,0.73102f,0.72324f,0.71520f,0.70633f,0.69449f,0.80202f,0.00000f,0.93697f,0.91889f,0.91718f,0.90737f,0.89441f,0.88021f,0.86331f,0.83941f,0.81608f,0.79241f,0.76713f,0.75281f,0.73501f,0.72740f,0.71914f,0.71847f,0.72036f,0.72550f,0.72157f,0.71930f,0.71612f,0.72208f,0.71825f,0.71241f,0.71054f,0.70649f,0.69723f,0.68301f,0.67033f,0.65291f,0.63580f,0.74587f,0.00000f,0.92587f,0.91935f,0.90347f,0.89872f,0.87202f,0.86107f,0.83994f,0.81660f,0.79351f,0.76610f,0.74291f,0.72829f,0.71570f,0.71133f,0.70711f,0.70549f,0.70163f,0.70381f,0.70705f,0.70147f,0.70312f,0.70023f,0.69372f,0.68968f,0.68543f,0.67427f,0.65872f,0.64254f,0.62213f,0.60227f,0.57732f,0.68828f,0.00000f,0.91132f,0.90533f,0.90451f,0.88543f,0.85671f,0.84422f,0.82013f,0.78877f,0.76784f,0.74071f,0.72077f,0.70407f,0.69712f,0.68991f,0.68396f,0.68671f,0.68748f,0.68259f,0.68298f,0.68384f,0.68550f,0.67878f,0.67379f,0.66463f,0.65582f,0.64066f,0.62071f,0.60036f,0.57779f,0.55362f,0.52621f,0.63071f,0.00000f,0.91375f,0.89578f,0.88652f,0.86263f,0.83867f,0.81565f,0.79089f,0.76513f,0.73897f,0.71717f,0.70022f,0.68579f,0.67566f,0.66933f,0.66609f,0.66749f,0.66773f,0.67017f,0.66691f,0.66355f,0.66230f,0.65566f,0.64623f,0.63787f,0.62426f,0.60410f,0.58548f,0.55748f,0.53202f,0.50218f,0.47585f,0.57561f,0.00000f,0.89512f,0.87598f,0.86549f,0.84198f,0.82455f,0.79267f,0.76146f,0.73589f,0.70991f,0.68597f,0.67364f,0.66365f,0.65699f,0.65718f,0.64949f,0.65038f,0.64706f,0.64761f,0.64399f,0.64474f,0.64380f,0.62756f,0.62554f,0.60920f,0.59466f,0.57150f,0.54667f,0.51770f,0.48520f,0.45784f,0.43061f,0.52264f,0.00000f,0.88587f,0.86924f,0.83985f,0.81560f,0.79435f,0.76358f,0.73474f,0.70877f,0.68153f,0.66383f,0.65045f,0.64191f,0.63495f,0.63153f,0.63427f,0.63231f,0.62917f,0.62825f,0.62736f,0.62445f,0.61875f,0.60737f,0.59915f,0.58303f,0.56371f,0.53698f,0.51075f,0.47719f,0.44497f,0.41617f,0.38513f,0.47182f,
 0.00000f,1.00000f,0.99684f,1.00000f,0.99038f,0.96875f,0.99073f,0.99320f,0.98464f,0.96572f,0.96453f,0.99809f,0.98167f,0.98931f,0.98573f,0.98529f,0.98154f,0.97187f,0.97159f,0.97649f,0.97316f,0.97163f,0.99142f,0.98757f,0.98587f,0.98697f,0.99163f,0.99083f,0.98934f,0.99128f,0.99269f,0.99599f,1.00000f,0.00000f,0.99174f,0.98643f,0.98335f,0.98089f,0.97786f,0.97816f,0.97749f,0.97471f,0.97361f,0.97372f,0.97546f,0.97422f,0.97451f,0.97566f,0.97512f,0.97556f,0.97514f,0.97619f,0.97808f,0.97873f,0.97894f,0.98151f,0.98252f,0.98407f,0.98599f,0.98796f,0.98946f,0.99091f,0.99292f,0.99503f,0.99748f,1.00000f,0.00000f,0.99142f,0.98706f,0.98243f,0.97962f,0.97707f,0.97638f,0.97548f,0.97375f,0.97400f,0.97385f,0.97329f,0.97303f,0.97360f,0.97453f,0.97468f,0.97501f,0.97566f,0.97666f,0.97818f,0.97906f,0.98072f,0.98165f,0.98298f,0.98429f,0.98608f,0.98775f,0.98954f,0.99149f,0.99337f,0.99546f,0.99776f,1.00000f,0.00000f,0.98417f,0.98625f,0.98305f,0.97934f,0.97741f,0.97609f,0.97533f,0.97446f,0.97333f,0.97352f,0.97375f,0.97379f,0.97426f,0.97404f,0.97489f,0.97484f,0.97469f,0.97653f,0.97809f,0.97877f,0.97974f,0.98185f,0.98256f,0.98454f,0.98600f,0.98722f,0.98933f,0.99146f,0.99315f,0.99546f,0.99774f,1.00000f,0.00000f,0.93743f,0.98122f,0.98265f,0.97963f,0.97662f,0.97637f,0.97479f,0.97342f,0.97399f,0.97395f,0.97372f,0.97338f,0.97353f,0.97383f,0.97374f,0.97472f,0.97552f,0.97654f,0.97732f,0.97821f,0.98021f,0.98125f,0.98291f,0.98394f,0.98615f,0.98754f,0.98933f,0.99120f,0.99343f,0.99545f,0.99778f,1.00000f,0.00000f,0.90959f,0.95570f,0.97753f,0.97965f,0.97730f,0.97523f,0.97483f,0.97376f,0.97287f,0.97257f,0.97250f,0.97367f,0.97332f,0.97396f,0.97387f,0.97455f,0.97561f,0.97676f,0.97783f,0.97842f,0.98021f,0.98131f,0.98247f,0.98435f,0.98597f,0.98754f,0.98917f,0.99121f,0.99325f,0.99540f,0.99774f,1.00000f,0.00000f,0.90324f,0.92204f,0.95588f,0.96835f,0.97542f,0.97460f,0.97339f,0.97429f,0.97312f,0.97253f,0.97337f,0.97262f,0.97346f,0.97366f,0.97367f,0.97471f,0.97542f,0.97611f,0.97736f,0.97811f,0.97969f,0.98114f,0.98243f,0.98416f,0.98579f,0.98764f,0.98883f,0.99131f,0.99333f,0.99551f,0.99767f,1.00000f,0.00000f,0.90315f,0.90183f,0.92886f,0.95004f,0.96583f,0.97046f,0.97126f,0.97283f,0.97203f,0.97278f,0.97243f,0.97120f,0.97318f,0.97312f,0.97339f,0.97384f,0.97542f,0.97639f,0.97733f,0.97761f,0.97963f,0.98099f,0.98235f,0.98383f,0.98539f,0.98728f,0.98826f,0.99095f,0.99297f,0.99525f,0.99754f,1.00000f,0.00000f,0.91098f,0.89740f,0.90807f,0.92536f,0.94323f,0.95693f,0.96488f,0.96899f,0.97130f,0.97128f,0.97119f,0.97298f,0.97309f,0.97318f,0.97333f,0.97409f,0.97473f,0.97539f,0.97724f,0.97855f,0.97917f,0.98102f,0.98229f,0.98366f,0.98562f,0.98660f,0.98736f,0.99082f,0.99309f,0.99526f,0.99758f,1.00000f,0.00000f,0.91516f,0.89817f,0.89751f,0.90729f,0.92159f,0.93677f,0.95035f,0.95834f,0.96450f,0.96787f,0.96953f,0.97018f,0.97215f,0.97208f,0.97313f,0.97380f,0.97389f,0.97590f,0.97621f,0.97737f,0.97887f,0.98087f,0.98224f,0.98363f,0.98463f,0.98444f,0.98728f,0.99096f,0.99321f,0.99515f,0.99751f,1.00000f,0.00000f,0.92135f,0.89908f,0.89311f,0.89189f,0.90530f,0.91859f,0.93027f,0.94212f,0.94961f,0.95822f,0.96218f,0.96757f,0.96847f,0.97094f,0.97338f,0.97343f,0.97478f,0.97513f,0.97664f,0.97734f,0.97911f,0.98071f,0.98219f,0.98284f,0.98277f,0.98351f,0.98695f,0.99044f,0.99314f,0.99518f,0.99744f,1.00000f,0.00000f,0.92553f,0.90074f,0.89531f,0.89510f,0.89300f,0.90399f,0.91259f,0.92337f,0.93333f,0.94397f,0.95016f,0.95694f,0.96164f,0.96628f,0.96915f,0.97133f,0.97340f,0.97433f,0.97606f,0.97763f,0.97888f,0.98077f,0.98089f,0.98200f,0.98044f,0.98193f,0.98608f,0.99021f,0.99288f,0.99511f,0.99751f,1.00000f,0.00000f,0.93434f,0.91333f,0.89153f,0.89021f,0.88826f,0.89272f,0.90122f,0.90691f,0.91835f,0.92525f,0.93455f,0.94312f,0.95019f,0.95792f,0.96119f,0.96646f,0.96823f,0.97169f,0.97420f,0.97471f,0.97904f,0.97983f,0.97897f,0.97804f,0.97866f,0.98123f,0.98568f,0.99017f,0.99237f,0.99496f,0.99738f,1.00000f,0.00000f,0.93802f,0.91116f,0.89886f,0.89006f,0.88929f,0.88644f,0.89299f,0.89705f,0.89984f,0.90912f,0.91630f,0.92718f,0.93354f,0.94151f,0.94827f,0.95538f,0.95942f,0.96532f,0.96891f,0.97215f,0.97529f,0.97514f,0.97461f,0.97494f,0.97607f,0.98058f,0.98551f,0.98979f,0.99247f,0.99493f,0.99729f,1.00000f,0.00000f,0.94518f,0.93135f,0.91208f,0.88919f,0.88730f,0.88621f,0.88471f,0.88903f,0.89179f,0.89845f,0.90324f,0.90916f,0.91724f,0.92707f,0.93314f,0.94016f,0.94810f,0.95400f,0.95880f,0.96282f,0.96515f,0.96656f,0.96799f,0.96856f,0.97381f,0.97949f,0.98502f,0.98925f,0.99210f,0.99456f,0.99715f,1.00000f,0.00000f,0.93812f,0.92586f,0.91351f,0.89914f,0.89264f,0.88357f,0.87954f,0.88281f,0.88485f,0.88681f,0.89115f,0.90020f,0.90400f,0.91120f,0.91724f,0.92426f,0.93191f,0.93760f,0.94497f,0.94811f,0.95128f,0.95371f,0.95663f,0.96146f,0.96930f,0.97608f,0.98358f,0.98868f,0.99153f,0.99430f,0.99700f,1.00000f,0.00000f,0.94599f,0.92882f,0.91693f,0.90406f,0.89648f,0.88868f,0.88377f,0.87955f,0.88006f,0.88062f,0.88461f,0.88875f,0.89039f,0.89829f,0.90263f,0.90823f,0.91591f,0.91742f,0.92506f,0.92773f,0.93242f,0.93653f,0.94282f,0.95069f,0.96015f,0.96991f,0.97800f,0.98350f,0.99009f,0.99402f,0.99685f,1.00000f,0.00000f,0.94178f,0.93911f,0.91933f,0.90234f,0.89884f,0.88863f,0.88604f,0.88138f,0.88110f,0.87756f,0.87949f,0.87864f,0.88363f,0.88455f,0.88724f,0.89283f,0.89238f,0.89792f,0.90275f,0.90794f,0.91162f,0.91654f,0.92511f,0.93503f,0.94661f,0.95783f,0.96917f,0.97728f,0.98478f,0.99061f,0.99606f,1.00000f,0.00000f,0.94722f,0.93540f,0.92846f,0.90800f,0.90048f,0.89565f,0.88473f,0.88023f,0.87578f,0.87744f,0.87421f,0.87687f,0.87364f,0.87504f,0.87150f,0.87177f,0.87401f,0.87868f,0.87947f,0.88078f,0.88622f,0.89802f,0.90637f,0.91559f,0.92848f,0.94068f,0.95275f,0.96451f,0.97398f,0.98296f,0.99129f,0.99999f,0.00000f,0.95206f,0.93645f,0.92773f,0.91307f,0.90654f,0.89855f,0.88631f,0.88341f,0.87655f,0.87450f,0.87377f,0.86857f,0.86494f,0.86209f,0.85928f,0.85855f,0.85568f,0.85928f,0.85538f,0.86007f,0.86564f,0.87614f,0.88465f,0.89807f,0.90914f,0.92264f,0.93681f,0.94695f,0.95962f,0.97019f,0.98186f,0.99950f,0.00000f,0.94164f,0.94046f,0.92603f,0.91548f,0.90911f,0.89754f,0.88893f,0.88353f,0.88218f,0.87344f,0.86886f,0.86104f,0.85859f,0.85316f,0.84847f,0.83972f,0.83611f,0.83232f,0.83350f,0.83749f,0.84292f,0.85088f,0.86413f,0.87505f,0.88680f,0.90195f,0.91172f,0.92492f,0.93717f,0.95018f,0.96236f,0.99693f,0.00000f,0.94101f,0.93573f,0.92722f,0.91634f,0.91191f,0.90191f,0.89282f,0.88051f,0.87796f,0.86666f,0.86267f,0.85783f,0.84700f,0.83850f,0.83045f,0.82213f,0.81739f,0.81266f,0.81543f,0.81743f,0.82349f,0.83039f,0.84423f,0.85389f,0.86542f,0.87631f,0.88635f,0.89821f,0.91197f,0.92230f,0.93627f,0.98888f,0.00000f,0.94438f,0.94086f,0.93017f,0.92410f,0.91106f,0.90366f,0.89431f,0.88384f,0.87714f,0.86975f,0.85450f,0.84704f,0.83843f,0.82186f,0.81600f,0.80709f,0.79989f,0.79834f,0.79725f,0.80026f,0.80582f,0.81165f,0.82228f,0.83201f,0.84164f,0.85136f,0.86161f,0.87260f,0.87998f,0.88926f,0.90073f,0.97056f,0.00000f,0.93959f,0.94130f,0.93107f,0.92342f,0.91718f,0.90446f,0.89573f,0.88196f,0.87376f,0.86133f,0.84861f,0.83444f,0.82076f,0.80662f,0.80182f,0.79218f,0.78547f,0.78053f,0.78584f,0.78275f,0.78623f,0.79447f,0.79835f,0.81004f,0.81498f,0.82145f,0.82927f,0.83611f,0.84238f,0.84979f,0.85572f,0.94155f,0.00000f,0.94346f,0.93986f,0.92728f,0.92138f,0.91618f,0.90284f,0.89119f,0.87590f,0.86667f,0.84759f,0.83501f,0.81795f,0.79969f,0.79226f,0.78001f,0.77267f,0.76855f,0.76619f,0.76898f,0.77214f,0.77151f,0.77581f,0.77936f,0.78705f,0.79080f,0.79363f,0.79683f,0.80288f,0.80287f,0.80632f,0.80574f,0.90119f,0.00000f,0.93918f,0.94093f,0.93120f,0.91556f,0.91042f,0.90111f,0.88310f,0.86469f,0.85184f,0.83360f,0.82081f,0.80184f,0.78466f,0.77081f,0.76398f,0.75573f,0.75052f,0.74920f,0.75582f,0.75480f,0.75486f,0.76246f,0.76045f,0.76454f,0.76324f,0.76794f,0.76649f,0.76266f,0.75888f,0.75438f,0.74742f,0.85462f,0.00000f,0.93764f,0.93240f,0.92667f,0.92207f,0.90605f,0.89085f,0.88003f,0.86138f,0.84323f,0.81587f,0.80629f,0.78542f,0.76484f,0.75626f,0.74504f,0.74363f,0.73916f,0.73853f,0.73880f,0.73604f,0.73375f,0.74308f,0.73920f,0.73739f,0.73612f,0.73459f,0.73001f,0.72285f,0.71445f,0.70420f,0.69299f,0.80168f,0.00000f,0.93002f,0.92750f,0.92162f,0.90930f,0.90015f,0.88478f,0.85120f,0.84120f,0.82569f,0.79709f,0.77557f,0.75940f,0.75218f,0.73517f,0.72763f,0.72500f,0.72296f,0.72061f,0.72152f,0.72084f,0.72297f,0.71911f,0.71975f,0.71401f,0.71168f,0.70109f,0.69191f,0.68261f,0.66723f,0.65205f,0.63465f,0.74376f,0.00000f,0.93140f,0.92544f,0.90776f,0.90565f,0.88844f,0.86042f,0.85237f,0.82688f,0.80192f,0.77797f,0.75894f,0.74090f,0.72296f,0.71711f,0.70872f,0.70854f,0.70448f,0.70602f,0.70595f,0.70469f,0.70264f,0.69884f,0.69448f,0.69049f,0.67884f,0.66607f,0.65725f,0.64067f,0.62049f,0.59835f,0.57796f,0.68634f,0.00000f,0.92441f,0.91659f,0.90265f,0.89337f,0.87483f,0.85289f,0.82592f,0.79647f,0.77629f,0.75316f,0.73755f,0.71513f,0.70877f,0.69925f,0.68804f,0.68890f,0.68927f,0.68582f,0.69091f,0.68479f,0.68323f,0.67696f,0.67289f,0.66301f,0.65078f,0.63109f,0.61685f,0.59391f,0.57379f,0.54900f,0.52435f,0.62795f,0.00000f,0.91834f,0.90614f,0.88884f,0.87068f,0.85108f,0.82502f,0.80016f,0.77663f,0.75401f,0.73274f,0.71512f,0.69634f,0.68022f,0.67996f,0.67387f,0.67269f,0.66855f,0.67166f,0.66980f,0.66571f,0.66515f,0.65526f,0.64126f,0.63313f,0.61734f,0.60399f,0.57652f,0.55273f,0.52470f,0.50004f,0.47758f,0.57231f,0.00000f,0.90287f,0.88753f,0.87434f,0.85554f,0.82608f,0.80666f,0.77674f,0.74870f,0.71934f,0.70399f,0.68738f,0.67377f,0.66506f,0.65981f,0.65347f,0.65588f,0.65140f,0.64495f,0.65109f,0.64462f,0.63642f,0.63106f,0.62089f,0.60969f,0.58675f,0.56714f,0.54017f,0.51123f,0.48402f,0.45491f,0.42619f,0.51978f,0.00000f,0.88197f,0.86965f,0.85229f,0.83365f,0.81256f,0.78037f,0.75010f,0.72016f,0.69317f,0.67776f,0.66087f,0.64683f,0.64361f,0.63618f,0.63317f,0.63226f,0.63258f,0.62940f,0.62550f,0.61914f,0.61460f,0.60474f,0.59563f,0.57882f,0.55781f,0.52629f,0.50088f,0.46695f,0.44147f,0.40953f,0.38417f,0.47040f,
 0.00000f,1.00000f,0.99815f,1.00000f,0.99074f,0.96883f,0.99067f,0.99347f,0.98339f,0.96728f,0.96425f,0.99824f,0.98315f,0.98914f,0.98631f,0.98487f,0.98167f,0.97173f,0.97133f,0.97608f,0.97315f,0.97252f,0.99150f,0.98687f,0.98588f,0.98726f,0.99166f,0.99068f,0.99330f,0.98877f,0.99346f,0.99672f,1.00000f,0.00000f,0.99115f,0.98590f,0.98382f,0.97890f,0.97703f,0.97650f,0.97673f,0.97510f,0.97276f,0.97235f,0.97495f,0.97355f,0.97353f,0.97491f,0.97542f,0.97570f,0.97507f,0.97564f,0.97798f,0.97908f,0.97899f,0.98183f,0.98277f,0.98433f,0.98598f,0.98789f,0.98967f,0.99155f,0.99251f,0.99518f,0.99756f,1.00000f,0.00000f,0.99160f,0.98619f,0.98165f,0.98037f,0.97717f,0.97643f,0.97473f,0.97431f,0.97462f,0.97389f,0.97401f,0.97415f,0.97337f,0.97474f,0.97410f,0.97509f,0.97602f,0.97578f,0.97838f,0.97913f,0.97995f,0.98132f,0.98259f,0.98468f,0.98584f,0.98767f,0.98934f,0.99151f,0.99335f,0.99550f,0.99777f,1.00000f,0.00000f,0.98175f,0.98609f,0.98286f,0.97912f,0.97729f,0.97649f,0.97587f,0.97397f,0.97405f,0.97399f,0.97411f,0.97231f,0.97411f,0.97442f,0.97389f,0.97488f,0.97578f,0.97682f,0.97782f,0.97836f,0.97979f,0.98127f,0.98253f,0.98447f,0.98606f,0.98759f,0.98958f,0.99112f,0.99336f,0.99550f,0.99776f,1.00000f,0.00000f,0.93631f,0.98204f,0.97947f,0.97747f,0.97831f,0.97587f,0.97498f,0.97351f,0.97405f,0.97380f,0.97362f,0.97290f,0.97348f,0.97428f,0.97424f,0.97520f,0.97583f,0.97644f,0.97722f,0.97905f,0.97996f,0.98088f,0.98297f,0.98413f,0.98585f,0.98737f,0.98925f,0.99121f,0.99325f,0.99546f,0.99774f,1.00000f,0.00000f,0.91017f,0.95631f,0.97800f,0.97719f,0.97676f,0.97611f,0.97484f,0.97451f,0.97339f,0.97218f,0.97245f,0.97288f,0.97357f,0.97352f,0.97349f,0.97521f,0.97534f,0.97716f,0.97718f,0.97809f,0.97976f,0.98114f,0.98258f,0.98419f,0.98561f,0.98727f,0.98940f,0.99119f,0.99318f,0.99536f,0.99781f,1.00000f,0.00000f,0.90095f,0.92117f,0.95647f,0.97029f,0.97477f,0.97519f,0.97363f,0.97345f,0.97318f,0.97229f,0.97362f,0.97316f,0.97349f,0.97380f,0.97411f,0.97478f,0.97567f,0.97597f,0.97822f,0.97849f,0.98002f,0.98141f,0.98271f,0.98381f,0.98585f,0.98776f,0.98912f,0.99099f,0.99326f,0.99542f,0.99765f,1.00000f,0.00000f,0.90245f,0.90352f,0.92495f,0.95087f,0.96384f,0.96960f,0.97285f,0.97411f,0.97127f,0.97077f,0.97273f,0.97207f,0.97271f,0.97351f,0.97332f,0.97470f,0.97514f,0.97538f,0.97774f,0.97829f,0.97937f,0.98160f,0.98257f,0.98402f,0.98584f,0.98740f,0.98823f,0.99090f,0.99302f,0.99540f,0.99758f,1.00000f,0.00000f,0.90981f,0.89505f,0.90988f,0.92375f,0.94327f,0.95760f,0.96408f,0.96847f,0.97158f,0.97187f,0.97119f,0.97191f,0.97200f,0.97369f,0.97322f,0.97434f,0.97454f,0.97574f,0.97674f,0.97837f,0.97957f,0.98093f,0.98206f,0.98364f,0.98588f,0.98630f,0.98761f,0.99072f,0.99323f,0.99544f,0.99761f,1.00000f,0.00000f,0.91882f,0.89682f,0.89775f,0.90797f,0.91984f,0.93866f,0.95011f,0.95779f,0.96326f,0.96751f,0.96980f,0.97008f,0.97200f,0.97344f,0.97349f,0.97327f,0.97500f,0.97515f,0.97674f,0.97815f,0.97932f,0.98007f,0.98229f,0.98394f,0.98497f,0.98539f,0.98687f,0.99055f,0.99299f,0.99527f,0.99746f,1.00000f,0.00000f,0.92288f,0.89841f,0.89610f,0.89747f,0.90572f,0.91668f,0.92514f,0.94148f,0.95246f,0.95823f,0.96252f,0.96535f,0.96898f,0.97098f,0.97360f,0.97266f,0.97358f,0.97613f,0.97674f,0.97814f,0.97852f,0.98063f,0.98173f,0.98346f,0.98357f,0.98381f,0.98599f,0.99046f,0.99284f,0.99486f,0.99745f,1.00000f,0.00000f,0.93266f,0.90556f,0.89315f,0.89117f,0.89292f,0.90082f,0.91381f,0.92238f,0.93184f,0.94484f,0.95023f,0.95669f,0.96217f,0.96642f,0.96783f,0.97154f,0.97306f,0.97548f,0.97682f,0.97696f,0.97858f,0.98046f,0.98130f,0.98185f,0.98108f,0.98131f,0.98553f,0.99010f,0.99260f,0.99521f,0.99723f,1.00000f,0.00000f,0.93599f,0.90921f,0.89476f,0.88960f,0.89035f,0.89542f,0.89743f,0.90944f,0.91584f,0.92530f,0.93518f,0.94527f,0.95003f,0.95508f,0.96241f,0.96410f,0.97006f,0.97261f,0.97484f,0.97587f,0.97831f,0.97932f,0.97958f,0.97964f,0.97846f,0.98089f,0.98549f,0.99006f,0.99270f,0.99490f,0.99739f,1.00000f,0.00000f,0.93441f,0.91941f,0.90717f,0.89074f,0.88681f,0.89140f,0.89174f,0.89369f,0.90260f,0.90892f,0.91902f,0.92612f,0.93442f,0.94159f,0.94893f,0.95568f,0.96032f,0.96537f,0.96879f,0.97219f,0.97500f,0.97565f,0.97672f,0.97544f,0.97724f,0.97986f,0.98480f,0.98903f,0.99231f,0.99492f,0.99724f,1.00000f,0.00000f,0.93926f,0.92182f,0.90766f,0.89283f,0.88789f,0.89243f,0.88633f,0.88897f,0.89378f,0.89621f,0.90362f,0.91020f,0.91706f,0.92454f,0.93397f,0.94216f,0.94752f,0.95467f,0.95975f,0.96318f,0.96622f,0.96844f,0.96930f,0.97012f,0.97349f,0.97858f,0.98395f,0.98886f,0.99224f,0.99501f,0.99708f,1.00000f,0.00000f,0.94333f,0.92674f,0.92250f,0.89473f,0.88919f,0.88846f,0.88317f,0.88397f,0.88347f,0.88828f,0.89159f,0.89786f,0.90200f,0.90876f,0.91513f,0.92332f,0.93056f,0.94056f,0.94557f,0.95020f,0.95158f,0.95505f,0.95736f,0.96144f,0.96815f,0.97509f,0.98259f,0.98811f,0.99173f,0.99462f,0.99686f,1.00000f,0.00000f,0.95017f,0.93066f,0.91385f,0.90343f,0.89713f,0.88889f,0.88434f,0.88404f,0.88117f,0.88701f,0.88399f,0.88785f,0.88736f,0.89781f,0.90281f,0.90916f,0.91628f,0.91887f,0.92649f,0.92972f,0.93470f,0.93792f,0.94278f,0.95064f,0.95988f,0.96821f,0.97712f,0.98461f,0.99033f,0.99387f,0.99702f,1.00000f,0.00000f,0.94565f,0.93396f,0.92259f,0.91156f,0.89678f,0.89037f,0.88740f,0.87923f,0.88032f,0.87776f,0.87773f,0.88305f,0.88156f,0.88498f,0.88922f,0.88976f,0.89806f,0.90019f,0.90427f,0.90905f,0.91204f,0.91930f,0.92517f,0.93560f,0.94658f,0.95700f,0.96806f,0.97699f,0.98464f,0.99056f,0.99594f,1.00000f,0.00000f,0.94392f,0.94238f,0.92328f,0.90768f,0.89834f,0.89156f,0.88559f,0.87983f,0.87458f,0.87405f,0.87546f,0.87184f,0.87402f,0.87830f,0.87897f,0.87771f,0.87402f,0.88107f,0.88073f,0.88523f,0.88910f,0.89600f,0.90636f,0.91491f,0.92991f,0.94191f,0.95195f,0.96395f,0.97476f,0.98263f,0.99147f,0.99999f,0.00000f,0.94566f,0.93591f,0.92601f,0.91251f,0.90472f,0.89427f,0.88985f,0.88239f,0.87686f,0.87483f,0.87586f,0.87256f,0.87115f,0.86661f,0.86449f,0.86147f,0.85637f,0.85750f,0.85480f,0.85853f,0.86772f,0.87686f,0.88858f,0.89711f,0.90984f,0.92199f,0.93481f,0.94834f,0.95881f,0.97023f,0.98112f,0.99950f,0.00000f,0.94710f,0.93564f,0.93191f,0.91589f,0.90659f,0.89889f,0.88681f,0.88250f,0.87481f,0.87085f,0.86994f,0.86590f,0.85926f,0.85825f,0.84903f,0.84338f,0.84212f,0.83422f,0.83581f,0.83966f,0.84668f,0.85370f,0.86761f,0.87646f,0.88855f,0.89972f,0.91321f,0.92604f,0.93768f,0.94932f,0.96303f,0.99691f,0.00000f,0.94259f,0.93619f,0.93371f,0.91252f,0.90829f,0.90208f,0.89310f,0.88438f,0.87898f,0.87033f,0.86755f,0.85667f,0.85717f,0.84255f,0.83672f,0.82668f,0.82176f,0.81462f,0.81846f,0.82280f,0.82654f,0.83287f,0.84450f,0.85168f,0.86790f,0.87351f,0.88760f,0.90013f,0.91363f,0.92178f,0.93583f,0.98874f,0.00000f,0.94688f,0.94274f,0.93478f,0.91636f,0.91106f,0.90224f,0.89606f,0.88836f,0.87650f,0.86604f,0.85809f,0.84351f,0.83789f,0.83090f,0.81693f,0.81478f,0.80260f,0.80379f,0.80167f,0.80436f,0.80868f,0.81449f,0.82197f,0.83123f,0.83819f,0.84862f,0.85912f,0.87056f,0.87967f,0.89043f,0.90005f,0.97068f,0.00000f,0.94075f,0.94163f,0.92988f,0.93269f,0.91202f,0.90082f,0.89277f,0.88200f,0.87341f,0.86424f,0.85080f,0.83879f,0.82484f,0.81157f,0.79729f,0.79551f,0.79301f,0.78694f,0.78490f,0.78695f,0.78873f,0.79297f,0.79828f,0.80820f,0.81473f,0.81952f,0.83190f,0.83625f,0.84350f,0.85076f,0.85479f,0.94069f,0.00000f,0.94287f,0.93938f,0.92955f,0.91963f,0.91369f,0.89999f,0.89434f,0.87718f,0.86803f,0.85437f,0.83937f,0.82135f,0.81020f,0.79589f,0.78351f,0.77605f,0.77465f,0.77140f,0.77053f,0.77232f,0.77213f,0.77722f,0.77972f,0.78401f,0.78895f,0.79111f,0.79643f,0.80150f,0.80405f,0.80309f,0.80246f,0.90149f,0.00000f,0.93827f,0.93405f,0.92725f,0.91339f,0.90975f,0.89903f,0.88826f,0.87592f,0.85539f,0.84416f,0.82036f,0.81068f,0.79065f,0.77994f,0.76714f,0.76189f,0.75786f,0.75269f,0.75699f,0.75712f,0.75922f,0.75990f,0.75940f,0.76354f,0.76146f,0.76418f,0.76630f,0.76232f,0.75809f,0.75339f,0.74899f,0.85605f,0.00000f,0.93754f,0.93568f,0.92521f,0.92161f,0.89876f,0.88930f,0.88055f,0.86097f,0.84870f,0.82353f,0.80450f,0.79136f,0.77263f,0.76279f,0.75148f,0.74119f,0.74169f,0.73924f,0.73835f,0.73925f,0.74003f,0.73846f,0.73887f,0.73944f,0.73281f,0.73259f,0.72706f,0.72223f,0.71467f,0.70187f,0.68987f,0.80182f,0.00000f,0.93293f,0.92794f,0.92264f,0.91410f,0.89777f,0.88684f,0.87065f,0.85333f,0.82916f,0.80790f,0.78528f,0.77070f,0.75519f,0.74420f,0.73780f,0.73147f,0.72515f,0.72451f,0.72248f,0.72423f,0.72190f,0.72201f,0.71878f,0.71244f,0.70809f,0.69959f,0.69029f,0.68100f,0.66656f,0.65109f,0.63518f,0.74373f,0.00000f,0.93393f,0.92370f,0.91597f,0.90667f,0.89254f,0.87599f,0.84855f,0.83309f,0.81409f,0.79162f,0.76551f,0.74985f,0.74069f,0.72696f,0.72091f,0.71063f,0.71203f,0.71245f,0.70932f,0.70847f,0.69969f,0.70142f,0.69603f,0.68428f,0.67966f,0.66593f,0.65330f,0.63599f,0.61657f,0.59431f,0.57545f,0.68579f,0.00000f,0.92684f,0.92084f,0.90967f,0.89713f,0.87396f,0.85362f,0.83288f,0.81264f,0.78849f,0.76868f,0.74629f,0.72586f,0.71458f,0.70512f,0.69494f,0.69861f,0.69289f,0.68817f,0.68929f,0.68878f,0.67876f,0.67943f,0.67108f,0.66134f,0.64918f,0.63276f,0.61321f,0.59238f,0.57015f,0.54534f,0.52289f,0.62780f,0.00000f,0.90658f,0.90920f,0.89236f,0.87390f,0.86312f,0.83520f,0.81414f,0.78753f,0.76357f,0.73946f,0.72294f,0.70764f,0.69436f,0.68619f,0.67851f,0.68240f,0.67474f,0.67458f,0.67100f,0.66892f,0.66328f,0.65685f,0.64355f,0.63259f,0.62093f,0.59955f,0.57802f,0.54999f,0.52332f,0.49663f,0.47304f,0.57091f,0.00000f,0.90141f,0.89422f,0.87773f,0.85974f,0.83872f,0.81635f,0.78888f,0.76468f,0.73764f,0.71454f,0.69604f,0.68151f,0.67140f,0.66270f,0.66142f,0.65921f,0.65394f,0.65342f,0.65025f,0.64436f,0.63776f,0.63145f,0.62037f,0.60605f,0.58410f,0.56083f,0.53732f,0.50482f,0.48036f,0.44799f,0.42439f,0.52148f,0.00000f,0.89841f,0.87895f,0.85418f,0.84016f,0.80990f,0.79123f,0.76065f,0.74193f,0.71192f,0.68532f,0.67107f,0.66005f,0.65284f,0.64313f,0.64096f,0.63669f,0.63683f,0.63284f,0.62811f,0.62341f,0.61352f,0.60579f,0.59283f,0.57380f,0.55414f,0.52916f,0.49601f,0.46313f,0.43430f,0.40626f,0.38334f,0.47010f,
 0.00000f,1.00000f,0.99788f,1.00000f,0.99057f,0.96949f,0.98947f,0.99411f,0.98296f,0.96517f,0.96536f,0.99782f,0.98120f,0.98900f,0.98644f,0.98394f,0.98154f,0.97184f,0.97113f,0.97688f,0.97289f,0.97225f,0.99135f,0.98754f,0.98560f,0.98639f,0.99138f,0.99100f,0.99309f,0.98984f,0.99372f,0.99690f,1.00000f,0.00000f,0.99354f,0.98698f,0.98202f,0.98138f,0.97762f,0.97741f,0.97569f,0.97482f,0.97275f,0.97334f,0.97431f,0.97342f,0.97425f,0.97472f,0.97567f,0.97499f,0.97501f,0.97654f,0.97783f,0.97841f,0.97927f,0.98183f,0.98280f,0.98468f,0.98584f,0.98756f,0.98962f,0.99161f,0.99275f,0.99513f,0.99765f,1.00000f,0.00000f,0.99065f,0.98553f,0.98248f,0.98091f,0.97825f,0.97649f,0.97549f,0.97444f,0.97405f,0.97439f,0.97397f,0.97334f,0.97307f,0.97376f,0.97443f,0.97539f,0.97614f,0.97681f,0.97828f,0.97899f,0.98016f,0.98163f,0.98287f,0.98476f,0.98593f,0.98758f,0.98945f,0.99143f,0.99337f,0.99548f,0.99775f,1.00000f,0.00000f,0.98460f,0.98580f,0.98260f,0.97947f,0.97785f,0.97628f,0.97618f,0.97489f,0.97443f,0.97441f,0.97312f,0.97364f,0.97327f,0.97486f,0.97387f,0.97520f,0.97599f,0.97619f,0.97778f,0.97857f,0.98057f,0.98155f,0.98239f,0.98427f,0.98586f,0.98790f,0.98948f,0.99140f,0.99315f,0.99550f,0.99775f,1.00000f,0.00000f,0.93695f,0.98334f,0.98231f,0.97942f,0.97814f,0.97663f,0.97349f,0.97450f,0.97413f,0.97327f,0.97297f,0.97377f,0.97404f,0.97337f,0.97501f,0.97491f,0.97647f,0.97642f,0.97784f,0.97893f,0.97993f,0.98108f,0.98266f,0.98414f,0.98591f,0.98760f,0.98941f,0.99111f,0.99330f,0.99547f,0.99773f,1.00000f,0.00000f,0.90701f,0.95532f,0.97668f,0.97789f,0.97547f,0.97637f,0.97414f,0.97362f,0.97291f,0.97303f,0.97319f,0.97299f,0.97320f,0.97328f,0.97475f,0.97485f,0.97570f,0.97703f,0.97768f,0.97854f,0.98008f,0.98116f,0.98232f,0.98404f,0.98608f,0.98763f,0.98931f,0.99091f,0.99322f,0.99534f,0.99776f,1.00000f,0.00000f,0.89995f,0.92454f,0.95736f,0.97278f,0.97443f,0.97531f,0.97428f,0.97350f,0.97252f,0.97237f,0.97288f,0.97234f,0.97347f,0.97312f,0.97406f,0.97454f,0.97554f,0.97617f,0.97778f,0.97861f,0.98016f,0.98080f,0.98249f,0.98439f,0.98558f,0.98720f,0.98949f,0.99074f,0.99320f,0.99528f,0.99766f,1.00000f,0.00000f,0.90273f,0.90427f,0.92608f,0.95019f,0.96438f,0.97011f,0.97217f,0.97254f,0.97285f,0.97321f,0.97132f,0.97175f,0.97193f,0.97399f,0.97426f,0.97446f,0.97565f,0.97647f,0.97739f,0.97838f,0.97970f,0.98072f,0.98235f,0.98424f,0.98558f,0.98747f,0.98865f,0.99074f,0.99324f,0.99543f,0.99751f,1.00000f,0.00000f,0.90743f,0.90103f,0.90735f,0.92526f,0.94427f,0.95643f,0.96603f,0.96865f,0.97147f,0.97180f,0.97194f,0.97141f,0.97253f,0.97278f,0.97280f,0.97463f,0.97483f,0.97525f,0.97697f,0.97840f,0.97911f,0.98072f,0.98229f,0.98401f,0.98534f,0.98675f,0.98738f,0.99062f,0.99322f,0.99532f,0.99757f,1.00000f,0.00000f,0.91829f,0.89145f,0.89609f,0.90697f,0.92296f,0.93743f,0.95080f,0.95846f,0.96495f,0.96778f,0.97029f,0.97123f,0.97235f,0.97252f,0.97371f,0.97401f,0.97540f,0.97577f,0.97719f,0.97785f,0.97914f,0.98111f,0.98229f,0.98368f,0.98576f,0.98567f,0.98637f,0.99049f,0.99293f,0.99529f,0.99755f,1.00000f,0.00000f,0.92477f,0.89865f,0.89628f,0.89806f,0.90330f,0.91551f,0.93049f,0.94088f,0.95227f,0.95779f,0.96379f,0.96780f,0.97008f,0.96991f,0.97264f,0.97216f,0.97515f,0.97508f,0.97600f,0.97755f,0.97885f,0.98033f,0.98203f,0.98356f,0.98406f,0.98408f,0.98529f,0.98999f,0.99298f,0.99516f,0.99737f,1.00000f,0.00000f,0.92944f,0.90574f,0.89380f,0.88939f,0.89736f,0.90166f,0.91210f,0.92253f,0.93259f,0.94384f,0.94939f,0.95732f,0.96367f,0.96599f,0.96823f,0.97121f,0.97327f,0.97564f,0.97641f,0.97781f,0.97956f,0.97993f,0.98133f,0.98204f,0.98141f,0.98235f,0.98518f,0.98975f,0.99265f,0.99514f,0.99739f,1.00000f,0.00000f,0.94020f,0.91442f,0.89571f,0.89128f,0.89381f,0.89643f,0.89826f,0.90785f,0.91644f,0.92453f,0.93627f,0.94330f,0.95033f,0.95538f,0.96046f,0.96591f,0.96853f,0.97154f,0.97416f,0.97603f,0.97775f,0.97917f,0.98057f,0.97961f,0.97892f,0.98019f,0.98438f,0.98941f,0.99273f,0.99500f,0.99740f,1.00000f,0.00000f,0.94158f,0.91880f,0.90556f,0.89501f,0.89043f,0.88917f,0.89168f,0.89143f,0.89912f,0.90963f,0.91816f,0.92685f,0.93337f,0.94095f,0.94930f,0.95367f,0.96054f,0.96521f,0.96991f,0.97343f,0.97488f,0.97738f,0.97684f,0.97731f,0.97694f,0.97960f,0.98430f,0.98914f,0.99203f,0.99489f,0.99722f,1.00000f,0.00000f,0.93877f,0.92445f,0.90494f,0.90189f,0.88719f,0.88169f,0.88853f,0.88617f,0.89367f,0.89921f,0.90373f,0.91059f,0.91692f,0.92778f,0.93325f,0.94057f,0.94703f,0.95536f,0.95908f,0.96328f,0.96786f,0.96837f,0.97080f,0.97085f,0.97253f,0.97774f,0.98394f,0.98876f,0.99217f,0.99456f,0.99708f,1.00000f,0.00000f,0.94776f,0.92605f,0.91290f,0.89741f,0.89800f,0.88993f,0.88870f,0.88276f,0.88552f,0.88784f,0.89266f,0.89858f,0.90167f,0.90974f,0.91600f,0.92322f,0.93260f,0.93966f,0.94483f,0.94996f,0.95420f,0.95600f,0.95934f,0.96402f,0.96873f,0.97505f,0.98202f,0.98788f,0.99185f,0.99445f,0.99688f,1.00000f,0.00000f,0.94692f,0.93133f,0.91790f,0.90705f,0.89644f,0.88900f,0.88277f,0.88377f,0.88197f,0.87849f,0.88511f,0.88884f,0.88832f,0.89583f,0.90164f,0.90962f,0.91574f,0.92303f,0.92640f,0.93090f,0.93656f,0.93961f,0.94441f,0.95322f,0.95906f,0.96722f,0.97610f,0.98402f,0.98992f,0.99420f,0.99679f,1.00000f,0.00000f,0.93972f,0.93077f,0.91632f,0.91429f,0.89586f,0.88750f,0.88564f,0.88179f,0.87855f,0.87670f,0.87660f,0.87674f,0.88583f,0.88605f,0.88644f,0.89457f,0.89480f,0.90223f,0.90445f,0.90977f,0.91476f,0.91952f,0.92538f,0.93472f,0.94447f,0.95607f,0.96657f,0.97659f,0.98405f,0.99049f,0.99574f,1.00000f,0.00000f,0.94656f,0.93786f,0.91628f,0.92100f,0.90028f,0.89282f,0.88298f,0.87893f,0.87510f,0.87203f,0.87380f,0.87335f,0.87319f,0.87838f,0.87665f,0.87664f,0.87926f,0.88103f,0.88370f,0.88675f,0.89203f,0.89706f,0.90592f,0.91588f,0.92858f,0.94230f,0.95218f,0.96515f,0.97387f,0.98329f,0.99145f,0.99999f,0.00000f,0.94728f,0.94422f,0.92194f,0.91393f,0.90621f,0.89719f,0.88614f,0.88257f,0.87475f,0.87319f,0.86685f,0.87612f,0.86675f,0.86821f,0.86785f,0.86571f,0.86317f,0.86253f,0.86024f,0.86001f,0.86870f,0.87575f,0.88629f,0.89725f,0.91032f,0.92205f,0.93461f,0.94861f,0.95726f,0.97046f,0.98085f,0.99950f,0.00000f,0.94210f,0.94079f,0.93113f,0.92270f,0.91025f,0.89848f,0.88883f,0.88286f,0.87708f,0.87097f,0.86516f,0.86827f,0.86272f,0.85302f,0.85245f,0.84775f,0.84644f,0.83964f,0.84086f,0.84109f,0.84596f,0.85131f,0.86504f,0.87606f,0.88858f,0.90145f,0.91304f,0.92388f,0.93788f,0.95031f,0.96279f,0.99697f,0.00000f,0.94604f,0.93844f,0.92643f,0.91938f,0.90844f,0.90376f,0.88972f,0.88404f,0.87772f,0.86885f,0.86698f,0.85719f,0.85535f,0.84507f,0.84148f,0.83208f,0.82380f,0.82499f,0.82068f,0.82224f,0.82818f,0.83267f,0.84194f,0.84970f,0.86568f,0.87741f,0.88740f,0.89911f,0.90914f,0.92105f,0.93599f,0.98900f,0.00000f,0.93881f,0.94010f,0.93001f,0.92413f,0.91505f,0.90291f,0.89634f,0.88775f,0.88009f,0.86794f,0.86325f,0.85169f,0.84308f,0.83266f,0.82168f,0.81666f,0.81198f,0.80367f,0.80571f,0.80265f,0.81015f,0.81505f,0.82216f,0.82832f,0.83697f,0.84980f,0.85811f,0.86999f,0.87691f,0.88884f,0.90121f,0.97068f,0.00000f,0.94025f,0.93527f,0.93111f,0.92792f,0.91708f,0.90174f,0.89324f,0.88926f,0.87619f,0.86523f,0.85494f,0.84264f,0.83105f,0.81852f,0.80772f,0.80011f,0.79167f,0.78732f,0.78883f,0.78783f,0.79408f,0.79592f,0.80012f,0.80511f,0.81210f,0.82159f,0.83063f,0.83722f,0.84417f,0.84940f,0.85486f,0.94049f,0.00000f,0.93719f,0.94080f,0.92520f,0.92836f,0.91249f,0.90479f,0.89888f,0.88032f,0.87397f,0.86389f,0.84906f,0.83112f,0.81345f,0.80078f,0.79118f,0.77779f,0.77219f,0.77516f,0.77284f,0.77340f,0.77439f,0.77555f,0.78256f,0.78641f,0.79165f,0.79394f,0.79606f,0.80201f,0.80254f,0.80336f,0.80275f,0.90252f,0.00000f,0.93831f,0.93892f,0.93144f,0.92023f,0.91575f,0.90561f,0.89186f,0.87313f,0.86347f,0.84478f,0.83206f,0.81313f,0.79840f,0.78369f,0.77613f,0.76757f,0.76160f,0.75597f,0.75665f,0.75791f,0.76106f,0.76195f,0.75984f,0.76102f,0.75984f,0.76287f,0.76292f,0.76306f,0.75708f,0.75511f,0.74694f,0.85507f,0.00000f,0.93734f,0.93352f,0.93178f,0.92389f,0.90576f,0.89234f,0.88373f,0.86964f,0.84775f,0.82994f,0.81344f,0.79881f,0.78237f,0.76531f,0.75638f,0.75392f,0.74759f,0.74053f,0.73896f,0.73896f,0.73883f,0.74138f,0.74115f,0.74005f,0.73487f,0.73190f,0.72851f,0.72126f,0.71111f,0.70252f,0.69140f,0.80133f,0.00000f,0.93352f,0.93362f,0.92714f,0.91781f,0.90348f,0.88578f,0.87105f,0.85266f,0.84400f,0.81196f,0.79502f,0.77545f,0.76243f,0.75509f,0.73987f,0.73474f,0.72574f,0.72848f,0.72439f,0.72515f,0.72573f,0.72013f,0.71528f,0.71076f,0.70646f,0.69927f,0.69012f,0.67777f,0.66386f,0.64824f,0.63052f,0.74443f,0.00000f,0.93136f,0.92324f,0.91403f,0.90949f,0.89428f,0.87539f,0.86116f,0.84757f,0.82103f,0.79327f,0.77839f,0.75701f,0.74086f,0.73163f,0.72341f,0.71741f,0.70865f,0.71016f,0.71100f,0.70690f,0.70632f,0.69652f,0.69200f,0.68933f,0.67901f,0.66626f,0.65331f,0.63331f,0.61516f,0.59601f,0.57511f,0.68737f,0.00000f,0.91951f,0.92148f,0.91265f,0.89880f,0.87728f,0.86316f,0.83930f,0.82408f,0.79765f,0.77504f,0.75770f,0.73542f,0.72334f,0.70815f,0.70235f,0.69925f,0.69411f,0.69204f,0.68968f,0.68427f,0.68413f,0.67838f,0.67171f,0.65999f,0.64778f,0.62966f,0.60839f,0.58968f,0.56608f,0.54242f,0.52548f,0.62518f,0.00000f,0.91812f,0.90732f,0.89656f,0.88472f,0.86883f,0.84892f,0.82732f,0.80009f,0.77534f,0.75296f,0.72944f,0.71530f,0.69536f,0.68986f,0.68761f,0.68077f,0.67975f,0.67236f,0.67036f,0.66622f,0.65623f,0.65552f,0.64513f,0.63091f,0.61886f,0.59644f,0.57087f,0.54441f,0.51933f,0.49445f,0.47073f,0.57216f,0.00000f,0.91059f,0.89518f,0.88228f,0.86807f,0.84703f,0.82515f,0.80057f,0.77781f,0.74818f,0.72313f,0.70403f,0.68860f,0.67897f,0.67499f,0.66936f,0.65952f,0.65781f,0.65793f,0.64797f,0.64713f,0.63696f,0.62753f,0.61898f,0.59978f,0.57997f,0.55338f,0.53102f,0.50302f,0.47412f,0.44758f,0.42579f,0.51879f,0.00000f,0.89421f,0.88281f,0.86935f,0.84651f,0.82364f,0.80395f,0.76622f,0.74110f,0.71844f,0.69694f,0.68404f,0.67087f,0.65939f,0.64831f,0.64599f,0.63687f,0.63751f,0.63712f,0.62898f,0.62365f,0.61340f,0.60413f,0.58815f,0.56989f,0.54948f,0.52172f,0.49417f,0.46078f,0.43428f,0.40441f,0.38265f,0.47075f,
 0.00000f,1.00000f,0.99721f,1.00000f,0.99067f,0.96907f,0.99166f,0.99388f,0.98422f,0.96690f,0.96500f,0.99755f,0.98106f,0.98940f,0.98501f,0.98436f,0.98202f,0.97193f,0.97105f,0.97605f,0.97277f,0.97220f,0.99188f,0.98726f,0.98588f,0.98699f,0.99159f,0.99094f,0.99325f,0.99064f,0.99119f,0.99682f,1.00000f,0.00000f,0.99102f,0.98626f,0.98323f,0.98061f,0.97668f,0.97617f,0.97538f,0.97479f,0.97340f,0.97388f,0.97487f,0.97383f,0.97454f,0.97481f,0.97578f,0.97617f,0.97490f,0.97600f,0.97840f,0.97887f,0.97924f,0.98215f,0.98265f,0.98448f,0.98573f,0.98793f,0.98958f,0.99157f,0.99285f,0.99473f,0.99766f,1.00000f,0.00000f,0.99024f,0.98519f,0.98403f,0.98015f,0.97752f,0.97718f,0.97608f,0.97455f,0.97393f,0.97232f,0.97379f,0.97359f,0.97323f,0.97386f,0.97441f,0.97580f,0.97582f,0.97737f,0.97748f,0.97868f,0.97991f,0.98096f,0.98327f,0.98446f,0.98589f,0.98791f,0.98968f,0.99125f,0.99328f,0.99548f,0.99774f,1.00000f,0.00000f,0.98252f,0.98516f,0.98213f,0.97925f,0.97815f,0.97631f,0.97620f,0.97422f,0.97394f,0.97446f,0.97314f,0.97345f,0.97315f,0.97423f,0.97427f,0.97608f,0.97559f,0.97644f,0.97752f,0.97879f,0.98013f,0.98141f,0.98253f,0.98429f,0.98612f,0.98789f,0.98957f,0.99139f,0.99320f,0.99546f,0.99772f,1.00000f,0.00000f,0.93830f,0.98169f,0.98040f,0.97778f,0.97682f,0.97553f,0.97497f,0.97394f,0.97302f,0.97382f,0.97377f,0.97411f,0.97305f,0.97363f,0.97451f,0.97435f,0.97441f,0.97654f,0.97758f,0.97807f,0.98007f,0.98131f,0.98284f,0.98451f,0.98605f,0.98767f,0.98936f,0.99141f,0.99334f,0.99551f,0.99776f,1.00000f,0.00000f,0.91071f,0.95492f,0.97602f,0.97753f,0.97637f,0.97538f,0.97422f,0.97388f,0.97457f,0.97318f,0.97360f,0.97249f,0.97341f,0.97354f,0.97452f,0.97472f,0.97521f,0.97628f,0.97826f,0.97847f,0.98013f,0.98127f,0.98272f,0.98424f,0.98612f,0.98722f,0.98923f,0.99117f,0.99338f,0.99536f,0.99774f,1.00000f,0.00000f,0.89880f,0.92339f,0.95420f,0.96950f,0.97648f,0.97542f,0.97312f,0.97428f,0.97252f,0.97213f,0.97287f,0.97286f,0.97275f,0.97397f,0.97354f,0.97439f,0.97491f,0.97623f,0.97770f,0.97854f,0.98002f,0.98117f,0.98280f,0.98413f,0.98582f,0.98758f,0.98924f,0.99072f,0.99317f,0.99536f,0.99761f,1.00000f,0.00000f,0.90142f,0.90469f,0.92585f,0.94825f,0.96425f,0.97004f,0.97283f,0.97248f,0.97258f,0.97314f,0.97167f,0.97144f,0.97265f,0.97268f,0.97298f,0.97384f,0.97511f,0.97577f,0.97739f,0.97843f,0.97955f,0.98083f,0.98218f,0.98404f,0.98537f,0.98740f,0.98906f,0.99052f,0.99325f,0.99530f,0.99757f,1.00000f,0.00000f,0.90508f,0.89515f,0.91056f,0.92268f,0.94541f,0.95502f,0.96575f,0.96901f,0.97253f,0.97193f,0.97117f,0.97298f,0.97152f,0.97300f,0.97326f,0.97431f,0.97542f,0.97570f,0.97677f,0.97782f,0.97914f,0.98064f,0.98196f,0.98356f,0.98536f,0.98714f,0.98783f,0.99004f,0.99305f,0.99532f,0.99756f,1.00000f,0.00000f,0.91719f,0.89612f,0.89732f,0.90650f,0.92048f,0.93814f,0.95039f,0.95879f,0.96475f,0.96772f,0.96999f,0.97231f,0.97212f,0.97308f,0.97229f,0.97390f,0.97446f,0.97567f,0.97718f,0.97760f,0.97907f,0.98066f,0.98250f,0.98403f,0.98539f,0.98635f,0.98653f,0.98982f,0.99292f,0.99514f,0.99753f,1.00000f,0.00000f,0.92038f,0.89876f,0.90026f,0.89748f,0.90422f,0.91628f,0.92740f,0.94018f,0.94968f,0.95698f,0.96249f,0.96744f,0.96912f,0.97061f,0.97238f,0.97332f,0.97463f,0.97442f,0.97604f,0.97774f,0.97906f,0.97995f,0.98203f,0.98357f,0.98417f,0.98452f,0.98572f,0.98969f,0.99265f,0.99535f,0.99748f,1.00000f,0.00000f,0.93028f,0.90815f,0.89378f,0.89088f,0.89730f,0.90228f,0.91003f,0.92189f,0.93136f,0.94219f,0.95129f,0.95611f,0.96269f,0.96666f,0.96901f,0.97168f,0.97295f,0.97486f,0.97606f,0.97788f,0.97965f,0.98016f,0.98137f,0.98298f,0.98264f,0.98260f,0.98513f,0.98928f,0.99259f,0.99491f,0.99753f,1.00000f,0.00000f,0.93962f,0.91289f,0.89295f,0.88910f,0.88921f,0.89343f,0.89855f,0.90630f,0.91599f,0.92332f,0.93433f,0.94146f,0.94892f,0.95608f,0.96054f,0.96544f,0.96752f,0.97298f,0.97535f,0.97691f,0.97860f,0.97961f,0.98108f,0.98013f,0.98039f,0.98113f,0.98463f,0.98898f,0.99278f,0.99499f,0.99741f,1.00000f,0.00000f,0.93857f,0.91595f,0.90271f,0.89106f,0.88993f,0.88779f,0.88819f,0.89787f,0.90156f,0.90874f,0.91921f,0.92670f,0.93514f,0.94361f,0.94881f,0.95331f,0.96018f,0.96518f,0.96977f,0.97197f,0.97583f,0.97804f,0.97689f,0.97755f,0.97751f,0.97863f,0.98365f,0.98856f,0.99239f,0.99461f,0.99739f,1.00000f,0.00000f,0.94197f,0.92036f,0.90562f,0.89359f,0.88666f,0.88708f,0.88309f,0.88879f,0.89260f,0.89663f,0.90406f,0.91185f,0.92080f,0.92649f,0.93225f,0.94139f,0.94760f,0.95309f,0.96023f,0.96546f,0.96801f,0.97048f,0.97134f,0.97308f,0.97394f,0.97800f,0.98309f,0.98844f,0.99231f,0.99469f,0.99727f,1.00000f,0.00000f,0.94723f,0.92803f,0.91255f,0.90047f,0.89202f,0.88682f,0.88132f,0.88652f,0.88946f,0.88950f,0.89069f,0.89480f,0.90244f,0.91132f,0.91968f,0.92431f,0.93151f,0.93879f,0.94499f,0.95111f,0.95444f,0.95719f,0.95912f,0.96295f,0.96928f,0.97500f,0.98150f,0.98746f,0.99170f,0.99439f,0.99710f,1.00000f,0.00000f,0.94185f,0.92504f,0.91597f,0.90217f,0.89781f,0.88867f,0.88630f,0.87958f,0.87843f,0.88526f,0.88257f,0.88483f,0.89433f,0.89484f,0.90312f,0.90786f,0.91360f,0.92183f,0.92743f,0.93163f,0.93763f,0.94226f,0.94729f,0.95114f,0.95899f,0.96766f,0.97641f,0.98387f,0.98948f,0.99353f,0.99694f,1.00000f,0.00000f,0.94103f,0.92829f,0.91527f,0.91267f,0.89773f,0.89168f,0.88698f,0.88155f,0.88069f,0.87204f,0.87954f,0.88053f,0.88333f,0.88558f,0.88832f,0.89324f,0.90129f,0.90341f,0.90657f,0.91053f,0.91778f,0.92166f,0.92689f,0.93511f,0.94640f,0.95585f,0.96584f,0.97702f,0.98463f,0.99049f,0.99607f,1.00000f,0.00000f,0.94248f,0.93834f,0.93289f,0.91162f,0.89952f,0.89178f,0.88789f,0.88149f,0.87232f,0.87196f,0.87231f,0.87367f,0.87165f,0.87710f,0.87571f,0.87737f,0.88046f,0.88468f,0.88567f,0.88925f,0.89559f,0.89932f,0.90936f,0.91782f,0.92695f,0.94043f,0.95174f,0.96442f,0.97371f,0.98295f,0.99136f,0.99999f,0.00000f,0.94553f,0.93232f,0.92230f,0.91298f,0.90775f,0.89451f,0.88736f,0.88184f,0.88057f,0.87831f,0.87521f,0.87106f,0.86672f,0.87091f,0.86382f,0.86427f,0.86313f,0.86268f,0.86217f,0.86612f,0.87244f,0.87887f,0.88512f,0.89702f,0.90901f,0.92033f,0.93468f,0.94696f,0.95858f,0.97054f,0.98083f,0.99946f,0.00000f,0.94831f,0.93583f,0.92864f,0.91847f,0.91178f,0.89961f,0.89194f,0.88034f,0.88116f,0.87607f,0.86935f,0.86452f,0.86324f,0.85779f,0.85431f,0.84796f,0.84273f,0.84487f,0.84388f,0.84434f,0.85043f,0.85693f,0.86587f,0.87505f,0.88763f,0.90046f,0.91139f,0.92370f,0.93543f,0.94965f,0.96334f,0.99689f,0.00000f,0.94689f,0.94291f,0.92991f,0.92560f,0.91130f,0.90616f,0.89961f,0.89108f,0.87725f,0.87121f,0.86407f,0.86577f,0.85599f,0.84996f,0.84518f,0.83591f,0.82992f,0.82126f,0.82415f,0.82411f,0.82819f,0.83421f,0.84237f,0.85477f,0.86394f,0.87420f,0.88753f,0.89786f,0.91125f,0.92309f,0.93497f,0.98880f,0.00000f,0.94611f,0.93038f,0.93799f,0.92433f,0.91428f,0.90767f,0.89525f,0.88828f,0.88037f,0.87188f,0.86210f,0.85290f,0.84190f,0.83724f,0.82608f,0.81940f,0.80895f,0.80745f,0.80141f,0.80697f,0.80930f,0.81220f,0.82415f,0.83239f,0.83826f,0.84944f,0.86033f,0.86846f,0.88036f,0.88981f,0.89873f,0.97013f,0.00000f,0.94085f,0.93887f,0.93529f,0.92429f,0.91644f,0.89957f,0.89406f,0.88834f,0.87737f,0.86583f,0.85764f,0.84539f,0.83566f,0.82192f,0.81481f,0.80441f,0.80195f,0.79696f,0.78642f,0.79094f,0.79115f,0.79688f,0.80278f,0.80614f,0.81524f,0.82024f,0.82941f,0.83424f,0.84308f,0.85045f,0.85564f,0.94076f,0.00000f,0.93970f,0.93562f,0.93349f,0.92299f,0.91325f,0.90261f,0.89159f,0.88324f,0.87081f,0.86168f,0.84739f,0.83235f,0.81638f,0.81034f,0.79850f,0.78413f,0.77915f,0.77728f,0.77734f,0.77764f,0.77664f,0.77440f,0.78047f,0.78436f,0.78779f,0.79162f,0.79871f,0.79915f,0.80262f,0.80637f,0.80327f,0.90172f,0.00000f,0.94186f,0.93665f,0.93261f,0.92175f,0.91647f,0.89709f,0.89413f,0.87846f,0.86803f,0.85400f,0.83492f,0.81708f,0.80125f,0.79072f,0.78302f,0.77223f,0.76503f,0.76180f,0.75888f,0.75862f,0.76344f,0.75930f,0.75907f,0.76099f,0.76316f,0.76127f,0.76293f,0.76178f,0.75808f,0.75404f,0.74637f,0.85368f,0.00000f,0.94087f,0.93256f,0.92716f,0.92344f,0.91030f,0.90160f,0.88598f,0.87041f,0.85507f,0.83317f,0.82183f,0.80298f,0.78288f,0.77127f,0.76410f,0.75219f,0.74930f,0.74390f,0.74285f,0.74105f,0.74055f,0.74157f,0.74138f,0.74047f,0.73315f,0.73118f,0.72655f,0.71844f,0.71310f,0.70057f,0.69082f,0.80012f,0.00000f,0.93192f,0.93453f,0.92759f,0.91909f,0.90536f,0.88943f,0.87760f,0.85995f,0.84640f,0.82288f,0.80578f,0.77783f,0.77285f,0.75926f,0.74362f,0.73836f,0.73353f,0.73574f,0.72319f,0.72745f,0.72515f,0.72055f,0.71784f,0.71169f,0.70414f,0.69858f,0.68937f,0.67805f,0.66231f,0.64691f,0.62985f,0.74266f,0.00000f,0.93805f,0.92710f,0.91997f,0.91038f,0.90052f,0.87980f,0.86098f,0.84418f,0.82102f,0.80686f,0.78392f,0.76525f,0.74896f,0.73443f,0.72846f,0.72293f,0.71543f,0.71325f,0.70930f,0.70933f,0.70587f,0.70024f,0.69525f,0.68672f,0.67537f,0.66223f,0.64716f,0.62892f,0.61005f,0.59295f,0.57405f,0.68473f,0.00000f,0.92860f,0.92487f,0.91066f,0.90045f,0.88656f,0.86530f,0.84813f,0.82376f,0.80682f,0.78092f,0.76012f,0.74464f,0.72863f,0.71473f,0.70823f,0.70251f,0.70117f,0.69318f,0.68809f,0.69138f,0.68221f,0.67432f,0.66935f,0.66038f,0.64514f,0.62919f,0.60780f,0.58833f,0.56329f,0.54443f,0.52135f,0.62770f,0.00000f,0.91350f,0.91041f,0.89381f,0.88141f,0.86966f,0.85016f,0.82679f,0.80436f,0.77918f,0.75893f,0.73940f,0.72580f,0.70592f,0.69764f,0.68892f,0.68364f,0.68237f,0.67357f,0.67107f,0.66652f,0.65819f,0.65601f,0.64025f,0.62934f,0.60886f,0.59076f,0.56960f,0.54325f,0.52057f,0.49212f,0.47004f,0.57066f,0.00000f,0.91035f,0.90132f,0.88224f,0.87593f,0.85672f,0.83297f,0.80588f,0.78260f,0.75429f,0.73378f,0.71383f,0.69732f,0.68816f,0.67802f,0.67428f,0.66467f,0.66278f,0.65761f,0.65182f,0.64276f,0.63594f,0.62779f,0.61455f,0.60007f,0.57989f,0.55345f,0.52972f,0.50216f,0.47239f,0.44725f,0.42163f,0.51838f,0.00000f,0.89552f,0.89156f,0.86758f,0.85571f,0.83713f,0.80887f,0.78118f,0.75934f,0.73530f,0.71593f,0.69224f,0.67486f,0.66120f,0.65822f,0.65185f,0.64093f,0.63917f,0.63372f,0.63100f,0.62386f,0.61376f,0.60142f,0.58466f,0.56968f,0.54462f,0.51551f,0.49069f,0.46195f,0.42833f,0.40506f,0.38046f,0.46695f,
 0.00000f,1.00000f,0.99910f,1.00000f,0.99138f,0.96811f,0.98998f,0.99320f,0.98424f,0.96660f,0.96494f,0.99739f,0.98272f,0.98851f,0.98534f,0.98449f,0.98277f,0.97166f,0.97114f,0.97646f,0.97289f,0.97230f,0.99108f,0.98730f,0.98604f,0.98715f,0.99183f,0.99070f,0.99327f,0.99161f,0.99288f,0.99697f,1.00000f,0.00000f,0.99154f,0.98717f,0.98313f,0.98084f,0.97898f,0.97598f,0.97598f,0.97511f,0.97272f,0.97351f,0.97467f,0.97419f,0.97427f,0.97499f,0.97560f,0.97565f,0.97556f,0.97631f,0.97855f,0.97855f,0.97906f,0.98202f,0.98264f,0.98412f,0.98558f,0.98823f,0.98962f,0.99138f,0.99306f,0.99488f,0.99763f,1.00000f,0.00000f,0.99140f,0.98664f,0.98318f,0.98024f,0.97777f,0.97634f,0.97637f,0.97446f,0.97415f,0.97313f,0.97409f,0.97428f,0.97379f,0.97355f,0.97516f,0.97552f,0.97568f,0.97663f,0.97848f,0.97909f,0.97990f,0.98138f,0.98285f,0.98458f,0.98596f,0.98771f,0.98960f,0.99138f,0.99327f,0.99547f,0.99773f,1.00000f,0.00000f,0.98368f,0.98511f,0.98259f,0.97891f,0.97728f,0.97699f,0.97574f,0.97393f,0.97407f,0.97348f,0.97334f,0.97307f,0.97296f,0.97361f,0.97461f,0.97529f,0.97611f,0.97649f,0.97757f,0.97915f,0.97972f,0.98105f,0.98284f,0.98428f,0.98585f,0.98753f,0.98933f,0.99149f,0.99323f,0.99539f,0.99780f,1.00000f,0.00000f,0.93875f,0.98050f,0.98123f,0.97940f,0.97735f,0.97631f,0.97517f,0.97353f,0.97418f,0.97279f,0.97243f,0.97341f,0.97396f,0.97418f,0.97385f,0.97481f,0.97540f,0.97648f,0.97809f,0.97902f,0.97985f,0.98144f,0.98291f,0.98411f,0.98575f,0.98762f,0.98932f,0.99138f,0.99340f,0.99538f,0.99772f,1.00000f,0.00000f,0.91193f,0.95715f,0.97662f,0.97788f,0.97794f,0.97519f,0.97432f,0.97369f,0.97257f,0.97380f,0.97238f,0.97295f,0.97329f,0.97368f,0.97387f,0.97444f,0.97486f,0.97642f,0.97729f,0.97902f,0.97966f,0.98127f,0.98269f,0.98398f,0.98579f,0.98759f,0.98928f,0.99145f,0.99344f,0.99548f,0.99774f,1.00000f,0.00000f,0.90004f,0.92147f,0.95566f,0.97180f,0.97396f,0.97463f,0.97448f,0.97333f,0.97321f,0.97181f,0.97210f,0.97242f,0.97294f,0.97442f,0.97395f,0.97536f,0.97563f,0.97660f,0.97757f,0.97855f,0.97960f,0.98130f,0.98248f,0.98419f,0.98560f,0.98731f,0.98935f,0.99091f,0.99321f,0.99539f,0.99767f,1.00000f,0.00000f,0.90186f,0.90535f,0.92940f,0.95060f,0.96369f,0.97133f,0.97406f,0.97299f,0.97325f,0.97216f,0.97203f,0.97180f,0.97223f,0.97334f,0.97414f,0.97450f,0.97639f,0.97645f,0.97700f,0.97843f,0.97967f,0.98060f,0.98229f,0.98380f,0.98562f,0.98756f,0.98905f,0.99018f,0.99309f,0.99521f,0.99748f,1.00000f,0.00000f,0.91181f,0.89766f,0.90723f,0.92542f,0.94324f,0.95802f,0.96259f,0.96828f,0.97021f,0.97131f,0.97278f,0.97243f,0.97198f,0.97294f,0.97382f,0.97353f,0.97515f,0.97583f,0.97675f,0.97852f,0.97946f,0.98102f,0.98239f,0.98344f,0.98497f,0.98701f,0.98812f,0.98972f,0.99305f,0.99528f,0.99753f,1.00000f,0.00000f,0.91661f,0.89503f,0.89739f,0.90702f,0.92000f,0.93866f,0.94976f,0.95788f,0.96566f,0.96726f,0.97089f,0.97218f,0.97183f,0.97220f,0.97425f,0.97413f,0.97478f,0.97577f,0.97672f,0.97799f,0.97898f,0.98066f,0.98214f,0.98362f,0.98528f,0.98657f,0.98680f,0.98938f,0.99279f,0.99531f,0.99749f,1.00000f,0.00000f,0.92664f,0.90048f,0.88966f,0.89538f,0.90508f,0.91523f,0.92827f,0.94133f,0.95071f,0.95888f,0.96352f,0.96730f,0.96966f,0.97138f,0.97215f,0.97475f,0.97494f,0.97587f,0.97718f,0.97723f,0.97935f,0.98004f,0.98178f,0.98374f,0.98504f,0.98547f,0.98604f,0.98924f,0.99281f,0.99509f,0.99746f,1.00000f,0.00000f,0.92617f,0.90494f,0.89547f,0.89109f,0.89573f,0.90371f,0.91338f,0.92463f,0.93086f,0.94141f,0.95038f,0.95467f,0.96273f,0.96496f,0.96789f,0.97139f,0.97307f,0.97487f,0.97630f,0.97779f,0.97887f,0.98014f,0.98122f,0.98282f,0.98307f,0.98290f,0.98425f,0.98905f,0.99257f,0.99534f,0.99743f,1.00000f,0.00000f,0.93845f,0.91057f,0.89814f,0.89302f,0.88708f,0.89182f,0.89894f,0.90823f,0.91679f,0.92381f,0.93426f,0.94175f,0.94969f,0.95684f,0.96055f,0.96436f,0.96858f,0.97126f,0.97480f,0.97705f,0.97893f,0.97908f,0.98186f,0.98119f,0.98055f,0.98132f,0.98424f,0.98823f,0.99228f,0.99499f,0.99719f,1.00000f,0.00000f,0.93639f,0.91950f,0.90494f,0.89415f,0.88896f,0.88593f,0.88997f,0.89600f,0.89932f,0.91041f,0.91803f,0.92664f,0.93379f,0.94060f,0.94938f,0.95623f,0.95930f,0.96540f,0.96975f,0.97302f,0.97516f,0.97744f,0.97903f,0.97873f,0.97857f,0.97939f,0.98332f,0.98859f,0.99170f,0.99485f,0.99734f,1.00000f,0.00000f,0.94294f,0.92139f,0.90958f,0.89859f,0.89494f,0.88890f,0.88607f,0.88838f,0.89447f,0.89738f,0.90451f,0.90958f,0.91861f,0.92730f,0.93588f,0.94228f,0.94670f,0.95557f,0.95900f,0.96595f,0.96928f,0.97195f,0.97314f,0.97277f,0.97509f,0.97818f,0.98308f,0.98784f,0.99188f,0.99458f,0.99734f,1.00000f,0.00000f,0.94011f,0.92446f,0.91317f,0.89666f,0.88948f,0.88334f,0.87909f,0.88629f,0.88597f,0.88695f,0.89220f,0.89903f,0.90255f,0.91093f,0.91861f,0.92486f,0.93158f,0.94076f,0.94589f,0.95184f,0.95434f,0.95939f,0.96094f,0.96392f,0.96977f,0.97454f,0.98008f,0.98694f,0.99103f,0.99469f,0.99695f,1.00000f,0.00000f,0.94882f,0.92963f,0.91858f,0.90686f,0.89399f,0.88729f,0.88148f,0.88315f,0.88329f,0.88329f,0.88683f,0.88736f,0.89107f,0.89827f,0.90170f,0.90882f,0.91592f,0.92263f,0.92785f,0.93135f,0.93869f,0.94017f,0.94744f,0.95314f,0.95935f,0.96666f,0.97578f,0.98347f,0.98962f,0.99386f,0.99689f,1.00000f,0.00000f,0.94969f,0.93251f,0.92274f,0.91360f,0.89759f,0.89153f,0.88210f,0.88081f,0.87432f,0.88090f,0.87785f,0.88060f,0.88287f,0.88619f,0.88662f,0.89613f,0.89678f,0.90499f,0.90686f,0.91280f,0.91708f,0.92088f,0.92796f,0.93570f,0.94529f,0.95603f,0.96674f,0.97609f,0.98375f,0.99082f,0.99574f,1.00000f,0.00000f,0.94342f,0.93636f,0.92431f,0.91238f,0.89666f,0.89291f,0.88783f,0.87766f,0.88081f,0.87654f,0.87320f,0.87665f,0.87625f,0.87409f,0.87680f,0.88017f,0.88356f,0.88358f,0.88795f,0.89025f,0.89575f,0.90083f,0.90669f,0.91732f,0.92858f,0.93898f,0.95198f,0.96446f,0.97470f,0.98288f,0.99164f,0.99999f,0.00000f,0.94589f,0.94013f,0.92881f,0.91481f,0.90501f,0.89976f,0.88543f,0.88215f,0.88215f,0.87224f,0.87044f,0.86710f,0.87120f,0.86778f,0.86829f,0.86575f,0.86348f,0.86557f,0.86627f,0.86797f,0.87216f,0.87953f,0.88860f,0.89810f,0.90631f,0.92007f,0.93395f,0.94621f,0.95830f,0.96959f,0.98149f,0.99950f,0.00000f,0.94333f,0.93933f,0.92459f,0.91811f,0.90762f,0.89884f,0.89046f,0.88250f,0.87827f,0.87606f,0.86803f,0.86289f,0.86094f,0.86188f,0.85947f,0.85379f,0.84794f,0.84528f,0.84365f,0.84498f,0.85028f,0.85830f,0.86515f,0.87558f,0.88561f,0.90094f,0.91171f,0.92345f,0.93662f,0.94998f,0.96299f,0.99686f,0.00000f,0.94665f,0.93892f,0.93017f,0.92399f,0.90710f,0.90463f,0.89249f,0.87926f,0.87877f,0.87593f,0.86645f,0.86337f,0.85232f,0.84709f,0.84357f,0.83717f,0.83277f,0.82557f,0.82422f,0.82733f,0.83256f,0.83559f,0.84164f,0.85542f,0.86443f,0.87476f,0.88655f,0.89992f,0.91087f,0.92294f,0.93645f,0.98848f,0.00000f,0.94208f,0.93634f,0.93207f,0.92500f,0.91302f,0.90256f,0.89300f,0.88471f,0.87884f,0.87598f,0.86440f,0.85760f,0.85132f,0.83624f,0.82949f,0.82335f,0.81705f,0.81593f,0.81029f,0.80529f,0.81295f,0.81851f,0.81912f,0.83275f,0.83774f,0.84986f,0.85896f,0.86886f,0.87890f,0.88879f,0.89901f,0.97056f,0.00000f,0.94620f,0.93587f,0.93169f,0.92714f,0.92510f,0.90747f,0.89527f,0.88836f,0.87944f,0.87129f,0.85581f,0.84565f,0.83777f,0.82838f,0.81805f,0.80532f,0.80217f,0.79693f,0.79129f,0.79141f,0.79462f,0.79600f,0.80412f,0.80682f,0.81580f,0.82059f,0.82988f,0.83669f,0.84406f,0.85028f,0.85607f,0.94119f,0.00000f,0.94507f,0.93763f,0.93196f,0.92554f,0.91935f,0.90293f,0.89639f,0.88188f,0.87364f,0.86127f,0.84863f,0.83973f,0.82301f,0.81412f,0.80003f,0.78981f,0.78432f,0.78451f,0.77456f,0.77577f,0.78103f,0.77871f,0.78138f,0.78254f,0.78708f,0.79194f,0.79881f,0.79780f,0.80147f,0.80230f,0.80127f,0.90235f,0.00000f,0.94470f,0.93879f,0.92286f,0.92476f,0.90543f,0.89920f,0.89406f,0.87891f,0.87301f,0.85487f,0.83640f,0.82182f,0.80694f,0.79726f,0.78724f,0.77800f,0.76763f,0.76415f,0.76272f,0.75889f,0.76151f,0.76059f,0.76089f,0.76028f,0.76622f,0.76234f,0.76102f,0.75837f,0.75494f,0.74912f,0.74590f,0.85423f,0.00000f,0.93196f,0.93800f,0.92745f,0.92587f,0.90651f,0.90449f,0.88644f,0.87653f,0.86209f,0.84051f,0.82935f,0.80717f,0.79305f,0.77845f,0.76627f,0.75945f,0.75149f,0.75141f,0.74990f,0.74350f,0.74564f,0.74277f,0.74047f,0.73565f,0.73568f,0.72750f,0.72833f,0.71962f,0.70993f,0.70054f,0.68855f,0.80093f,0.00000f,0.93287f,0.93576f,0.92459f,0.91887f,0.90253f,0.89334f,0.88341f,0.86587f,0.84228f,0.82821f,0.81096f,0.79026f,0.77417f,0.75633f,0.75415f,0.74111f,0.74012f,0.73382f,0.72899f,0.72490f,0.72456f,0.72263f,0.71823f,0.71156f,0.70250f,0.69687f,0.68594f,0.67530f,0.66467f,0.64398f,0.63324f,0.74219f,0.00000f,0.93073f,0.92464f,0.92327f,0.90861f,0.89567f,0.88248f,0.87058f,0.84974f,0.82539f,0.80579f,0.79012f,0.77003f,0.75914f,0.74860f,0.73266f,0.72219f,0.72143f,0.71738f,0.70721f,0.71044f,0.70544f,0.69879f,0.69361f,0.68715f,0.67578f,0.66343f,0.65214f,0.63272f,0.61183f,0.59543f,0.57443f,0.68617f,0.00000f,0.92247f,0.91274f,0.91869f,0.91141f,0.89083f,0.86840f,0.85429f,0.83467f,0.80551f,0.79406f,0.76635f,0.75427f,0.73668f,0.72522f,0.71423f,0.70385f,0.70417f,0.69809f,0.69214f,0.68790f,0.68384f,0.67225f,0.66839f,0.65480f,0.64237f,0.62570f,0.60608f,0.58498f,0.56646f,0.54271f,0.52216f,0.62704f,0.00000f,0.91600f,0.90799f,0.90322f,0.88965f,0.88088f,0.85801f,0.83701f,0.81288f,0.79328f,0.76537f,0.74781f,0.72835f,0.71382f,0.70646f,0.69607f,0.68444f,0.68296f,0.67862f,0.66955f,0.66493f,0.66492f,0.65259f,0.64013f,0.62873f,0.60997f,0.59209f,0.56591f,0.54401f,0.51698f,0.49448f,0.46670f,0.56948f,0.00000f,0.91261f,0.90843f,0.89992f,0.87185f,0.85692f,0.83367f,0.81876f,0.78633f,0.76515f,0.74539f,0.72419f,0.70453f,0.68545f,0.68300f,0.67403f,0.67042f,0.66430f,0.65881f,0.65182f,0.64376f,0.63426f,0.62853f,0.61333f,0.59466f,0.57430f,0.55053f,0.52213f,0.49650f,0.47380f,0.44683f,0.42248f,0.51814f,0.00000f,0.89907f,0.89345f,0.88328f,0.85644f,0.84094f,0.81406f,0.78703f,0.76339f,0.73842f,0.71897f,0.69783f,0.68218f,0.66830f,0.66079f,0.65454f,0.64568f,0.64791f,0.63472f,0.63520f,0.62317f,0.61030f,0.59934f,0.58268f,0.56214f,0.54069f,0.51257f,0.48846f,0.45581f,0.43036f,0.40000f,0.37965f,0.46588f,
 0.00000f,1.00000f,0.99828f,1.00000f,0.98998f,0.96792f,0.98977f,0.99365f,0.98461f,0.96544f,0.96457f,0.99876f,0.98105f,0.98894f,0.98582f,0.98447f,0.98215f,0.97232f,0.97132f,0.97597f,0.97364f,0.97252f,0.99159f,0.98761f,0.98571f,0.98669f,0.99152f,0.99100f,0.99318f,0.99326f,0.99281f,0.99532f,1.00000f,0.00000f,0.99281f,0.98490f,0.98275f,0.98157f,0.97822f,0.97739f,0.97613f,0.97519f,0.97308f,0.97290f,0.97384f,0.97343f,0.97509f,0.97495f,0.97523f,0.97525f,0.97452f,0.97644f,0.97851f,0.97874f,0.97951f,0.98238f,0.98247f,0.98443f,0.98589f,0.98814f,0.98959f,0.99158f,0.99291f,0.99502f,0.99740f,1.00000f,0.00000f,0.99038f,0.98591f,0.98285f,0.97941f,0.97903f,0.97769f,0.97535f,0.97557f,0.97406f,0.97490f,0.97326f,0.97358f,0.97383f,0.97335f,0.97480f,0.97533f,0.97569f,0.97714f,0.97770f,0.97872f,0.98015f,0.98127f,0.98312f,0.98424f,0.98590f,0.98786f,0.98953f,0.99153f,0.99320f,0.99543f,0.99777f,1.00000f,0.00000f,0.98324f,0.98392f,0.98252f,0.97896f,0.97755f,0.97654f,0.97538f,0.97461f,0.97485f,0.97349f,0.97345f,0.97269f,0.97408f,0.97354f,0.97485f,0.97539f,0.97583f,0.97719f,0.97809f,0.97880f,0.97968f,0.98150f,0.98285f,0.98419f,0.98590f,0.98747f,0.98952f,0.99125f,0.99330f,0.99550f,0.99778f,1.00000f,0.00000f,0.93863f,0.98205f,0.98156f,0.97991f,0.97819f,0.97537f,0.97557f,0.97475f,0.97346f,0.97359f,0.97264f,0.97176f,0.97334f,0.97374f,0.97418f,0.97473f,0.97601f,0.97645f,0.97786f,0.97846f,0.97988f,0.98147f,0.98229f,0.98416f,0.98578f,0.98755f,0.98940f,0.99132f,0.99331f,0.99537f,0.99775f,1.00000f,0.00000f,0.90967f,0.95671f,0.97553f,0.97923f,0.97722f,0.97553f,0.97461f,0.97413f,0.97422f,0.97272f,0.97318f,0.97319f,0.97364f,0.97390f,0.97417f,0.97403f,0.97564f,0.97705f,0.97763f,0.97859f,0.97991f,0.98058f,0.98260f,0.98394f,0.98562f,0.98758f,0.98943f,0.99131f,0.99338f,0.99543f,0.99778f,1.00000f,0.00000f,0.89962f,0.92279f,0.95355f,0.97091f,0.97580f,0.97496f,0.97529f,0.97375f,0.97173f,0.97315f,0.97311f,0.97335f,0.97262f,0.97321f,0.97403f,0.97405f,0.97534f,0.97587f,0.97719f,0.97836f,0.97990f,0.98104f,0.98254f,0.98423f,0.98580f,0.98757f,0.98929f,0.99082f,0.99310f,0.99541f,0.99769f,1.00000f,0.00000f,0.90329f,0.90656f,0.92765f,0.95077f,0.96327f,0.97086f,0.97309f,0.97438f,0.97250f,0.97174f,0.97305f,0.97273f,0.97265f,0.97314f,0.97336f,0.97423f,0.97496f,0.97649f,0.97709f,0.97842f,0.97961f,0.98047f,0.98230f,0.98359f,0.98524f,0.98739f,0.98916f,0.99025f,0.99312f,0.99541f,0.99749f,1.00000f,0.00000f,0.91030f,0.89499f,0.91050f,0.92541f,0.94469f,0.95679f,0.96498f,0.97052f,0.97111f,0.97176f,0.97147f,0.97319f,0.97334f,0.97211f,0.97426f,0.97452f,0.97480f,0.97544f,0.97685f,0.97798f,0.97900f,0.98061f,0.98221f,0.98397f,0.98560f,0.98742f,0.98872f,0.98968f,0.99313f,0.99533f,0.99755f,1.00000f,0.00000f,0.91800f,0.89841f,0.89720f,0.90661f,0.92133f,0.93672f,0.94832f,0.95757f,0.96462f,0.96639f,0.96994f,0.97145f,0.97227f,0.97223f,0.97275f,0.97280f,0.97399f,0.97620f,0.97694f,0.97849f,0.97933f,0.98111f,0.98227f,0.98367f,0.98525f,0.98711f,0.98748f,0.98918f,0.99284f,0.99535f,0.99747f,1.00000f,0.00000f,0.92189f,0.90232f,0.89489f,0.89433f,0.90619f,0.91687f,0.92947f,0.94277f,0.95156f,0.95744f,0.96328f,0.96660f,0.96897f,0.97020f,0.97296f,0.97259f,0.97396f,0.97604f,0.97632f,0.97733f,0.97865f,0.98090f,0.98227f,0.98358f,0.98507f,0.98576f,0.98624f,0.98892f,0.99262f,0.99530f,0.99746f,1.00000f,0.00000f,0.92802f,0.90638f,0.89489f,0.89181f,0.89319f,0.90215f,0.91039f,0.92228f,0.93392f,0.94189f,0.94966f,0.95685f,0.96130f,0.96532f,0.96890f,0.97124f,0.97323f,0.97536f,0.97647f,0.97743f,0.97849f,0.98083f,0.98147f,0.98362f,0.98381f,0.98414f,0.98448f,0.98886f,0.99247f,0.99520f,0.99750f,1.00000f,0.00000f,0.93786f,0.91115f,0.89754f,0.88841f,0.89172f,0.89381f,0.89756f,0.90915f,0.91596f,0.92624f,0.93219f,0.94014f,0.94996f,0.95606f,0.96103f,0.96599f,0.96858f,0.97157f,0.97562f,0.97693f,0.97830f,0.97973f,0.98160f,0.98184f,0.98191f,0.98162f,0.98403f,0.98804f,0.99255f,0.99485f,0.99731f,1.00000f,0.00000f,0.94406f,0.91590f,0.90283f,0.89600f,0.89358f,0.88815f,0.89014f,0.89249f,0.90105f,0.90668f,0.91591f,0.92770f,0.93337f,0.93982f,0.94739f,0.95577f,0.96078f,0.96413f,0.96934f,0.97261f,0.97530f,0.97865f,0.97886f,0.97915f,0.97881f,0.97950f,0.98328f,0.98806f,0.99196f,0.99480f,0.99734f,1.00000f,0.00000f,0.94158f,0.92654f,0.90332f,0.89384f,0.89128f,0.89004f,0.88397f,0.88832f,0.89322f,0.89806f,0.90372f,0.91053f,0.91802f,0.92574f,0.93404f,0.94136f,0.94796f,0.95351f,0.95956f,0.96406f,0.96957f,0.97216f,0.97326f,0.97379f,0.97528f,0.97879f,0.98198f,0.98713f,0.99148f,0.99455f,0.99720f,1.00000f,0.00000f,0.94432f,0.92921f,0.91507f,0.89766f,0.89121f,0.88614f,0.88014f,0.87995f,0.88878f,0.88851f,0.89209f,0.89924f,0.90354f,0.90852f,0.91789f,0.92489f,0.93245f,0.94110f,0.94774f,0.95247f,0.95591f,0.95955f,0.96313f,0.96643f,0.96922f,0.97393f,0.98063f,0.98661f,0.99089f,0.99418f,0.99714f,1.00000f,0.00000f,0.95629f,0.92692f,0.92197f,0.89893f,0.89466f,0.89243f,0.88602f,0.88123f,0.88123f,0.88364f,0.88546f,0.88404f,0.89046f,0.89569f,0.90329f,0.91014f,0.91795f,0.92399f,0.92804f,0.93482f,0.93934f,0.94455f,0.94518f,0.95428f,0.95989f,0.96641f,0.97508f,0.98327f,0.98973f,0.99403f,0.99692f,1.00000f,0.00000f,0.94412f,0.92922f,0.92292f,0.91218f,0.90509f,0.88967f,0.88646f,0.87998f,0.88247f,0.88352f,0.88207f,0.87937f,0.88123f,0.88377f,0.88988f,0.89338f,0.90355f,0.90487f,0.91009f,0.91504f,0.91814f,0.92384f,0.92778f,0.93665f,0.94630f,0.95586f,0.96540f,0.97659f,0.98368f,0.99045f,0.99563f,1.00000f,0.00000f,0.94760f,0.93759f,0.92543f,0.91275f,0.89831f,0.89163f,0.88458f,0.88188f,0.87460f,0.87660f,0.87486f,0.87403f,0.87420f,0.87219f,0.87766f,0.88137f,0.88209f,0.88451f,0.88981f,0.89244f,0.89546f,0.90211f,0.90776f,0.91868f,0.92969f,0.94009f,0.95182f,0.96355f,0.97308f,0.98305f,0.99137f,0.99999f,0.00000f,0.94785f,0.93977f,0.92308f,0.91772f,0.90585f,0.90095f,0.89250f,0.88272f,0.87728f,0.87872f,0.87071f,0.87055f,0.86883f,0.86913f,0.86756f,0.86765f,0.86663f,0.86943f,0.86696f,0.87125f,0.87195f,0.87663f,0.88712f,0.89852f,0.90717f,0.91929f,0.93304f,0.94647f,0.95764f,0.96878f,0.98083f,0.99948f,0.00000f,0.94165f,0.93699f,0.92930f,0.91568f,0.90885f,0.90247f,0.88638f,0.88433f,0.87579f,0.87675f,0.86765f,0.86729f,0.86484f,0.85998f,0.85924f,0.85195f,0.85155f,0.84827f,0.84666f,0.84941f,0.85101f,0.85762f,0.86456f,0.87483f,0.88675f,0.89839f,0.90994f,0.92588f,0.93577f,0.94977f,0.96273f,0.99685f,0.00000f,0.94905f,0.94820f,0.93459f,0.91712f,0.90752f,0.89800f,0.88887f,0.88209f,0.87861f,0.87509f,0.86650f,0.86102f,0.85780f,0.84913f,0.84857f,0.84211f,0.83472f,0.83063f,0.82997f,0.83051f,0.83364f,0.83553f,0.84539f,0.85317f,0.86113f,0.87377f,0.88537f,0.89812f,0.91039f,0.92271f,0.93589f,0.98871f,0.00000f,0.93933f,0.93720f,0.93282f,0.92038f,0.91405f,0.90240f,0.89626f,0.88039f,0.87928f,0.86874f,0.86724f,0.85488f,0.85126f,0.84311f,0.83369f,0.82628f,0.82119f,0.81710f,0.81394f,0.81115f,0.81319f,0.81669f,0.82085f,0.83171f,0.84021f,0.84822f,0.85840f,0.86960f,0.87972f,0.89084f,0.89934f,0.97024f,0.00000f,0.94966f,0.94085f,0.92442f,0.92649f,0.91501f,0.90489f,0.90106f,0.88913f,0.87704f,0.87071f,0.85855f,0.84511f,0.84039f,0.82948f,0.81503f,0.80790f,0.80448f,0.79711f,0.79940f,0.79304f,0.79454f,0.79664f,0.80091f,0.81169f,0.81910f,0.82224f,0.82739f,0.83325f,0.84216f,0.85098f,0.85414f,0.94089f,0.00000f,0.94332f,0.93870f,0.93014f,0.91989f,0.92039f,0.89829f,0.88994f,0.88282f,0.87777f,0.86875f,0.85680f,0.83544f,0.82597f,0.81375f,0.80359f,0.79712f,0.78638f,0.78557f,0.77984f,0.78269f,0.78033f,0.78349f,0.78194f,0.78235f,0.78563f,0.79476f,0.79898f,0.80136f,0.79964f,0.80167f,0.80239f,0.90202f,0.00000f,0.93503f,0.93152f,0.92821f,0.91812f,0.91656f,0.90843f,0.89322f,0.88373f,0.86656f,0.85759f,0.84546f,0.82605f,0.81241f,0.79709f,0.78591f,0.77825f,0.77302f,0.76492f,0.76575f,0.76545f,0.76401f,0.76171f,0.76414f,0.76137f,0.75889f,0.76181f,0.76456f,0.76181f,0.75523f,0.75126f,0.74777f,0.85444f,0.00000f,0.93545f,0.93070f,0.93083f,0.92320f,0.90855f,0.89775f,0.89263f,0.87430f,0.86289f,0.84315f,0.82982f,0.81842f,0.79851f,0.78410f,0.77264f,0.76178f,0.75106f,0.75207f,0.75060f,0.74663f,0.74152f,0.74081f,0.74062f,0.73810f,0.73224f,0.73051f,0.72290f,0.72123f,0.70918f,0.69937f,0.68833f,0.79952f,0.00000f,0.92963f,0.93554f,0.92451f,0.91827f,0.90391f,0.89109f,0.88347f,0.86759f,0.85265f,0.83003f,0.81906f,0.79669f,0.77743f,0.76836f,0.75124f,0.74675f,0.74295f,0.73061f,0.73462f,0.73248f,0.72540f,0.72143f,0.71444f,0.71232f,0.70374f,0.69833f,0.68737f,0.67423f,0.66349f,0.64517f,0.63225f,0.74354f,0.00000f,0.92587f,0.92272f,0.93112f,0.91039f,0.89338f,0.88784f,0.87203f,0.85770f,0.83059f,0.81937f,0.79627f,0.77929f,0.76338f,0.74904f,0.73672f,0.72855f,0.72468f,0.71249f,0.71482f,0.71345f,0.70607f,0.69999f,0.69405f,0.68483f,0.67640f,0.65741f,0.64975f,0.62776f,0.61010f,0.59342f,0.57617f,0.68398f,0.00000f,0.92726f,0.92623f,0.91458f,0.90412f,0.89098f,0.87960f,0.86301f,0.83805f,0.81592f,0.79789f,0.77338f,0.75789f,0.74257f,0.72756f,0.71641f,0.70826f,0.70750f,0.70030f,0.69919f,0.69160f,0.68195f,0.67672f,0.66627f,0.65529f,0.63981f,0.62250f,0.60481f,0.58454f,0.56026f,0.54064f,0.52015f,0.62725f,0.00000f,0.91361f,0.91626f,0.90250f,0.88935f,0.88247f,0.86301f,0.84442f,0.82187f,0.79772f,0.77236f,0.75715f,0.73444f,0.71849f,0.70666f,0.69763f,0.69217f,0.68508f,0.67975f,0.67420f,0.66931f,0.66530f,0.65152f,0.63962f,0.62561f,0.60632f,0.58803f,0.56386f,0.54085f,0.51566f,0.49200f,0.46680f,0.56961f,0.00000f,0.91493f,0.90371f,0.89698f,0.87821f,0.86308f,0.84147f,0.82478f,0.80345f,0.77408f,0.75561f,0.72715f,0.71101f,0.69731f,0.68344f,0.68009f,0.67104f,0.66931f,0.65615f,0.65619f,0.64857f,0.63351f,0.62605f,0.61416f,0.59488f,0.57315f,0.54881f,0.52176f,0.49541f,0.47129f,0.44357f,0.41830f,0.51698f,0.00000f,0.90803f,0.88978f,0.89058f,0.86246f,0.85357f,0.81797f,0.79898f,0.78026f,0.74366f,0.72839f,0.70303f,0.68452f,0.67733f,0.66439f,0.66048f,0.65223f,0.64290f,0.63901f,0.63342f,0.62299f,0.60972f,0.59982f,0.58633f,0.56409f,0.53395f,0.51302f,0.48233f,0.45466f,0.42686f,0.40031f,0.38050f,0.46720f,
 0.00000f,1.00000f,0.99699f,0.99892f,0.99054f,0.96833f,0.99102f,0.99361f,0.98465f,0.96608f,0.96421f,0.99867f,0.98110f,0.98827f,0.98547f,0.98442f,0.98303f,0.97211f,0.97089f,0.97685f,0.97307f,0.97132f,0.99086f,0.98734f,0.98655f,0.98696f,0.99150f,0.99062f,0.99332f,0.99497f,0.99316f,0.99631f,1.00000f,0.00000f,0.99038f,0.98653f,0.98263f,0.98024f,0.97789f,0.97735f,0.97586f,0.97591f,0.97301f,0.97318f,0.97483f,0.97360f,0.97498f,0.97463f,0.97536f,0.97604f,0.97535f,0.97716f,0.97775f,0.97844f,0.97902f,0.98215f,0.98247f,0.98406f,0.98577f,0.98806f,0.98943f,0.99150f,0.99343f,0.99503f,0.99751f,1.00000f,0.00000f,0.99101f,0.98595f,0.98250f,0.97985f,0.97778f,0.97607f,0.97552f,0.97492f,0.97420f,0.97411f,0.97372f,0.97241f,0.97335f,0.97450f,0.97433f,0.97547f,0.97547f,0.97685f,0.97796f,0.97888f,0.97969f,0.98134f,0.98280f,0.98406f,0.98610f,0.98758f,0.98947f,0.99150f,0.99347f,0.99543f,0.99777f,1.00000f,0.00000f,0.98207f,0.98457f,0.98246f,0.97993f,0.97767f,0.97659f,0.97539f,0.97469f,0.97418f,0.97399f,0.97363f,0.97216f,0.97379f,0.97373f,0.97531f,0.97531f,0.97667f,0.97677f,0.97743f,0.97877f,0.97994f,0.98141f,0.98292f,0.98439f,0.98569f,0.98756f,0.98953f,0.99138f,0.99322f,0.99541f,0.99772f,1.00000f,0.00000f,0.93828f,0.98237f,0.98038f,0.97919f,0.97808f,0.97713f,0.97548f,0.97515f,0.97355f,0.97262f,0.97325f,0.97316f,0.97335f,0.97413f,0.97446f,0.97488f,0.97550f,0.97639f,0.97795f,0.97843f,0.98008f,0.98153f,0.98283f,0.98439f,0.98588f,0.98744f,0.98932f,0.99130f,0.99336f,0.99545f,0.99774f,1.00000f,0.00000f,0.90898f,0.95351f,0.97731f,0.97839f,0.97615f,0.97556f,0.97395f,0.97354f,0.97394f,0.97392f,0.97329f,0.97272f,0.97475f,0.97371f,0.97324f,0.97468f,0.97566f,0.97681f,0.97742f,0.97837f,0.97990f,0.98120f,0.98266f,0.98410f,0.98566f,0.98757f,0.98940f,0.99125f,0.99327f,0.99545f,0.99778f,1.00000f,0.00000f,0.89707f,0.92145f,0.95618f,0.97104f,0.97640f,0.97444f,0.97508f,0.97406f,0.97268f,0.97258f,0.97201f,0.97318f,0.97273f,0.97458f,0.97400f,0.97416f,0.97526f,0.97668f,0.97744f,0.97851f,0.97968f,0.98109f,0.98278f,0.98379f,0.98558f,0.98719f,0.98950f,0.99089f,0.99318f,0.99550f,0.99759f,1.00000f,0.00000f,0.90201f,0.90701f,0.92431f,0.94934f,0.96285f,0.97026f,0.97363f,0.97288f,0.97259f,0.97238f,0.97289f,0.97203f,0.97295f,0.97288f,0.97416f,0.97406f,0.97484f,0.97607f,0.97742f,0.97824f,0.97963f,0.98062f,0.98231f,0.98404f,0.98568f,0.98738f,0.98906f,0.99034f,0.99298f,0.99536f,0.99755f,1.00000f,0.00000f,0.91073f,0.89912f,0.90764f,0.92328f,0.94547f,0.95714f,0.96321f,0.96767f,0.97137f,0.97072f,0.97165f,0.97290f,0.97211f,0.97273f,0.97306f,0.97388f,0.97489f,0.97640f,0.97673f,0.97814f,0.97943f,0.98022f,0.98158f,0.98372f,0.98533f,0.98732f,0.98901f,0.99001f,0.99271f,0.99544f,0.99756f,1.00000f,0.00000f,0.91423f,0.90060f,0.89792f,0.90854f,0.92027f,0.93534f,0.94838f,0.95660f,0.96454f,0.96771f,0.96994f,0.97201f,0.97250f,0.97259f,0.97188f,0.97444f,0.97476f,0.97564f,0.97698f,0.97796f,0.97845f,0.98077f,0.98210f,0.98331f,0.98559f,0.98718f,0.98816f,0.98855f,0.99284f,0.99521f,0.99749f,1.00000f,0.00000f,0.92494f,0.90030f,0.89778f,0.89908f,0.90290f,0.91788f,0.92969f,0.93915f,0.95123f,0.95775f,0.96197f,0.96554f,0.96935f,0.97176f,0.97233f,0.97388f,0.97500f,0.97508f,0.97671f,0.97838f,0.97949f,0.98066f,0.98205f,0.98308f,0.98499f,0.98635f,0.98635f,0.98838f,0.99268f,0.99504f,0.99746f,1.00000f,0.00000f,0.93100f,0.90052f,0.89492f,0.89207f,0.89393f,0.90140f,0.91431f,0.92175f,0.93405f,0.94292f,0.95141f,0.95686f,0.96243f,0.96636f,0.96935f,0.97090f,0.97389f,0.97476f,0.97511f,0.97818f,0.97859f,0.97947f,0.98143f,0.98339f,0.98467f,0.98365f,0.98485f,0.98806f,0.99274f,0.99512f,0.99737f,1.00000f,0.00000f,0.93379f,0.91345f,0.89521f,0.89141f,0.89020f,0.89113f,0.89978f,0.90615f,0.91610f,0.92400f,0.93630f,0.94281f,0.94821f,0.95584f,0.96010f,0.96602f,0.97038f,0.97196f,0.97458f,0.97611f,0.97833f,0.97923f,0.98124f,0.98216f,0.98217f,0.98221f,0.98381f,0.98766f,0.99216f,0.99486f,0.99739f,1.00000f,0.00000f,0.93614f,0.92429f,0.90122f,0.89821f,0.89130f,0.89030f,0.89185f,0.89684f,0.89910f,0.90818f,0.91711f,0.92690f,0.93426f,0.94300f,0.95087f,0.95675f,0.96051f,0.96398f,0.96971f,0.97222f,0.97538f,0.97788f,0.97980f,0.98038f,0.97922f,0.98026f,0.98320f,0.98719f,0.99143f,0.99472f,0.99734f,1.00000f,0.00000f,0.93946f,0.92142f,0.90577f,0.89604f,0.89251f,0.88914f,0.88769f,0.88571f,0.88751f,0.89557f,0.90329f,0.90973f,0.91775f,0.92532f,0.93400f,0.94386f,0.94759f,0.95453f,0.96012f,0.96454f,0.96925f,0.97204f,0.97475f,0.97466f,0.97697f,0.97814f,0.98232f,0.98709f,0.99125f,0.99474f,0.99704f,1.00000f,0.00000f,0.94961f,0.92056f,0.91608f,0.90128f,0.89267f,0.88538f,0.88512f,0.88224f,0.88281f,0.88893f,0.89080f,0.89386f,0.90306f,0.91115f,0.91655f,0.92411f,0.93038f,0.94027f,0.94597f,0.95321f,0.95865f,0.96118f,0.96345f,0.96626f,0.96898f,0.97484f,0.98000f,0.98621f,0.99076f,0.99457f,0.99737f,1.00000f,0.00000f,0.94549f,0.92689f,0.91669f,0.90529f,0.89273f,0.88695f,0.88508f,0.87958f,0.88207f,0.88353f,0.88498f,0.88348f,0.89159f,0.89523f,0.90411f,0.90723f,0.91565f,0.92297f,0.92916f,0.93464f,0.93995f,0.94375f,0.94883f,0.95439f,0.95974f,0.96785f,0.97407f,0.98257f,0.98827f,0.99376f,0.99718f,1.00000f,0.00000f,0.94064f,0.93843f,0.92360f,0.90375f,0.89378f,0.88997f,0.88145f,0.87918f,0.87674f,0.87978f,0.88133f,0.87947f,0.88250f,0.88272f,0.88863f,0.89188f,0.89909f,0.90742f,0.90964f,0.91736f,0.92091f,0.92326f,0.93031f,0.93730f,0.94631f,0.95526f,0.96522f,0.97532f,0.98428f,0.99048f,0.99604f,1.00000f,0.00000f,0.93842f,0.93797f,0.92018f,0.91177f,0.89998f,0.89117f,0.88556f,0.88574f,0.87224f,0.87744f,0.87586f,0.87617f,0.87466f,0.87369f,0.87939f,0.88189f,0.88300f,0.88515f,0.88882f,0.89309f,0.89867f,0.90382f,0.90850f,0.91872f,0.92909f,0.93810f,0.95136f,0.96403f,0.97372f,0.98253f,0.99149f,0.99999f,0.00000f,0.94885f,0.93704f,0.92500f,0.91461f,0.91082f,0.89917f,0.88641f,0.88099f,0.87508f,0.87239f,0.87148f,0.86779f,0.86904f,0.86915f,0.87175f,0.86753f,0.86717f,0.86912f,0.86837f,0.87205f,0.87394f,0.87896f,0.88962f,0.89776f,0.90693f,0.91938f,0.93239f,0.94657f,0.95782f,0.96882f,0.98048f,0.99949f,0.00000f,0.93767f,0.94177f,0.93079f,0.91835f,0.90987f,0.90430f,0.88918f,0.88600f,0.87912f,0.87221f,0.87031f,0.87015f,0.86190f,0.86312f,0.85621f,0.85524f,0.85218f,0.84957f,0.84582f,0.84984f,0.85326f,0.85845f,0.86822f,0.87421f,0.88675f,0.89769f,0.91079f,0.92395f,0.93697f,0.94980f,0.96261f,0.99690f,0.00000f,0.94337f,0.94273f,0.93080f,0.91941f,0.90607f,0.90437f,0.89331f,0.88364f,0.87748f,0.87255f,0.86642f,0.86043f,0.86121f,0.85290f,0.84516f,0.84196f,0.83994f,0.83193f,0.83105f,0.83075f,0.83377f,0.83730f,0.84518f,0.85508f,0.86456f,0.87421f,0.88664f,0.89721f,0.90923f,0.92329f,0.93592f,0.98900f,0.00000f,0.95421f,0.93889f,0.93069f,0.92114f,0.91802f,0.90024f,0.89413f,0.88871f,0.88640f,0.87057f,0.86922f,0.85869f,0.85187f,0.84261f,0.83503f,0.82684f,0.82076f,0.81421f,0.81394f,0.81258f,0.81592f,0.81494f,0.82262f,0.82969f,0.84008f,0.84826f,0.85868f,0.86863f,0.87960f,0.88890f,0.89955f,0.97035f,0.00000f,0.94525f,0.93334f,0.93477f,0.92446f,0.90639f,0.90539f,0.90035f,0.88121f,0.87961f,0.86797f,0.86420f,0.84821f,0.84450f,0.83060f,0.82225f,0.81402f,0.80334f,0.80216f,0.79509f,0.79718f,0.79431f,0.79973f,0.80254f,0.80930f,0.81375f,0.82240f,0.82945f,0.83661f,0.84226f,0.84887f,0.85508f,0.94060f,0.00000f,0.94688f,0.94347f,0.92986f,0.92772f,0.91064f,0.90454f,0.89603f,0.88284f,0.87758f,0.86631f,0.85713f,0.84534f,0.82888f,0.81661f,0.80651f,0.80214f,0.79095f,0.78672f,0.78225f,0.78022f,0.77930f,0.77809f,0.78423f,0.78677f,0.79071f,0.78993f,0.79281f,0.79764f,0.80093f,0.80041f,0.80488f,0.90080f,0.00000f,0.92804f,0.93687f,0.92837f,0.92379f,0.91557f,0.89842f,0.89543f,0.88678f,0.87067f,0.86013f,0.84636f,0.82706f,0.81350f,0.80378f,0.79455f,0.78051f,0.77139f,0.76967f,0.76711f,0.76793f,0.76392f,0.76481f,0.76200f,0.76380f,0.76422f,0.76223f,0.76316f,0.76046f,0.75656f,0.75262f,0.74771f,0.85323f,0.00000f,0.93314f,0.93780f,0.92898f,0.92053f,0.91299f,0.89979f,0.89203f,0.87725f,0.86260f,0.84890f,0.83341f,0.81503f,0.80132f,0.78975f,0.77510f,0.76447f,0.75341f,0.75183f,0.75336f,0.74787f,0.74954f,0.74251f,0.74005f,0.73806f,0.73212f,0.72778f,0.72431f,0.71590f,0.71185f,0.69839f,0.68880f,0.79972f,0.00000f,0.93004f,0.92974f,0.93153f,0.92509f,0.90260f,0.89465f,0.88533f,0.86718f,0.84943f,0.84146f,0.82150f,0.79931f,0.78693f,0.76828f,0.75648f,0.74718f,0.74324f,0.74356f,0.73522f,0.72912f,0.72809f,0.72533f,0.72096f,0.70866f,0.70344f,0.69530f,0.68455f,0.67130f,0.65922f,0.64481f,0.63123f,0.74712f,0.00000f,0.92729f,0.92780f,0.93094f,0.91859f,0.90269f,0.89016f,0.87349f,0.85057f,0.83655f,0.82085f,0.80272f,0.78418f,0.76245f,0.75175f,0.74608f,0.72874f,0.72343f,0.71929f,0.71577f,0.71221f,0.70773f,0.70089f,0.69313f,0.68642f,0.67645f,0.66141f,0.64347f,0.62510f,0.61128f,0.59191f,0.57219f,0.68458f,0.00000f,0.92158f,0.92257f,0.92326f,0.91009f,0.89308f,0.87831f,0.86188f,0.83967f,0.82589f,0.80323f,0.78174f,0.76519f,0.74784f,0.73623f,0.72274f,0.71881f,0.70723f,0.70230f,0.69613f,0.69497f,0.68502f,0.67736f,0.66641f,0.66102f,0.64104f,0.62377f,0.60354f,0.58238f,0.56195f,0.54061f,0.52011f,0.62826f,0.00000f,0.92206f,0.91906f,0.90993f,0.89975f,0.88763f,0.87032f,0.84822f,0.83166f,0.80249f,0.77582f,0.76015f,0.74236f,0.72739f,0.71552f,0.70229f,0.69418f,0.68914f,0.68201f,0.68343f,0.66990f,0.66771f,0.65661f,0.64253f,0.62374f,0.60487f,0.58544f,0.56265f,0.53948f,0.51579f,0.49270f,0.46950f,0.56855f,0.00000f,0.91264f,0.90580f,0.89506f,0.88244f,0.86657f,0.85281f,0.82722f,0.80441f,0.78055f,0.75779f,0.74147f,0.72005f,0.70577f,0.68954f,0.68361f,0.67834f,0.66535f,0.66374f,0.65503f,0.64851f,0.63937f,0.62655f,0.61376f,0.59057f,0.57162f,0.54686f,0.51954f,0.49317f,0.47218f,0.44250f,0.42120f,0.51667f,0.00000f,0.89786f,0.90034f,0.88200f,0.86609f,0.85327f,0.82942f,0.79803f,0.78045f,0.75857f,0.73576f,0.70845f,0.69689f,0.67882f,0.67054f,0.66077f,0.65610f,0.64850f,0.64227f,0.63363f,0.62649f,0.60776f,0.59642f,0.58159f,0.55807f,0.53483f,0.51037f,0.48025f,0.45356f,0.42653f,0.40308f,0.38048f,0.46554f,
 0.00000f,1.00000f,0.99789f,1.00000f,0.99045f,0.96853f,0.99193f,0.99412f,0.98415f,0.96723f,0.96367f,0.99932f,0.98191f,0.98835f,0.98560f,0.98500f,0.98266f,0.97232f,0.97072f,0.97616f,0.97322f,0.97232f,0.99150f,0.98753f,0.98582f,0.98682f,0.99145f,0.99076f,0.99328f,0.99496f,0.99331f,0.99629f,1.00000f,0.00000f,0.99253f,0.98672f,0.98309f,0.98083f,0.97787f,0.97776f,0.97731f,0.97508f,0.97342f,0.97371f,0.97488f,0.97307f,0.97486f,0.97452f,0.97540f,0.97613f,0.97577f,0.97630f,0.97774f,0.97823f,0.97888f,0.98189f,0.98262f,0.98415f,0.98530f,0.98781f,0.98935f,0.99165f,0.99345f,0.99511f,0.99747f,1.00000f,0.00000f,0.99077f,0.98763f,0.98197f,0.98070f,0.97917f,0.97701f,0.97523f,0.97552f,0.97382f,0.97374f,0.97383f,0.97315f,0.97438f,0.97465f,0.97429f,0.97545f,0.97593f,0.97691f,0.97790f,0.97900f,0.97963f,0.98117f,0.98283f,0.98426f,0.98587f,0.98753f,0.98946f,0.99145f,0.99336f,0.99544f,0.99781f,1.00000f,0.00000f,0.98273f,0.98588f,0.98238f,0.98026f,0.97688f,0.97757f,0.97467f,0.97421f,0.97313f,0.97433f,0.97343f,0.97291f,0.97424f,0.97358f,0.97432f,0.97530f,0.97564f,0.97703f,0.97792f,0.97835f,0.97990f,0.98156f,0.98289f,0.98450f,0.98607f,0.98756f,0.98930f,0.99129f,0.99322f,0.99546f,0.99776f,1.00000f,0.00000f,0.93813f,0.98188f,0.98167f,0.97902f,0.97707f,0.97560f,0.97522f,0.97511f,0.97349f,0.97315f,0.97286f,0.97309f,0.97238f,0.97440f,0.97456f,0.97556f,0.97507f,0.97662f,0.97706f,0.97917f,0.97936f,0.98160f,0.98224f,0.98395f,0.98572f,0.98764f,0.98945f,0.99107f,0.99332f,0.99542f,0.99778f,1.00000f,0.00000f,0.90868f,0.95540f,0.97724f,0.97824f,0.97737f,0.97509f,0.97641f,0.97373f,0.97467f,0.97296f,0.97289f,0.97375f,0.97322f,0.97367f,0.97496f,0.97466f,0.97588f,0.97648f,0.97714f,0.97937f,0.98022f,0.98103f,0.98267f,0.98427f,0.98571f,0.98756f,0.98964f,0.99139f,0.99299f,0.99550f,0.99771f,1.00000f,0.00000f,0.89723f,0.92246f,0.95569f,0.97005f,0.97471f,0.97679f,0.97485f,0.97317f,0.97219f,0.97275f,0.97318f,0.97303f,0.97212f,0.97376f,0.97370f,0.97424f,0.97508f,0.97635f,0.97723f,0.97857f,0.97972f,0.98106f,0.98256f,0.98396f,0.98536f,0.98756f,0.98935f,0.99100f,0.99296f,0.99540f,0.99761f,1.00000f,0.00000f,0.89968f,0.90529f,0.92457f,0.94864f,0.96453f,0.97079f,0.97245f,0.97285f,0.97275f,0.97289f,0.97114f,0.97258f,0.97244f,0.97342f,0.97299f,0.97400f,0.97494f,0.97595f,0.97714f,0.97850f,0.97969f,0.98099f,0.98216f,0.98389f,0.98547f,0.98736f,0.98931f,0.99071f,0.99289f,0.99531f,0.99749f,1.00000f,0.00000f,0.91234f,0.89490f,0.90545f,0.92461f,0.94365f,0.95707f,0.96467f,0.96946f,0.97084f,0.97141f,0.97171f,0.97257f,0.97344f,0.97248f,0.97420f,0.97319f,0.97494f,0.97616f,0.97667f,0.97852f,0.97949f,0.98110f,0.98273f,0.98417f,0.98563f,0.98734f,0.98909f,0.98990f,0.99274f,0.99533f,0.99755f,1.00000f,0.00000f,0.91694f,0.89714f,0.89622f,0.90736f,0.92285f,0.93732f,0.94999f,0.96024f,0.96375f,0.96812f,0.96908f,0.97259f,0.97167f,0.97325f,0.97272f,0.97395f,0.97486f,0.97499f,0.97654f,0.97779f,0.97921f,0.98070f,0.98186f,0.98361f,0.98544f,0.98685f,0.98783f,0.98917f,0.99261f,0.99519f,0.99756f,1.00000f,0.00000f,0.92254f,0.90222f,0.89128f,0.89976f,0.90317f,0.91716f,0.92984f,0.93999f,0.95017f,0.95639f,0.96302f,0.96727f,0.96910f,0.97120f,0.97269f,0.97386f,0.97482f,0.97495f,0.97653f,0.97728f,0.97935f,0.98058f,0.98220f,0.98357f,0.98523f,0.98622f,0.98672f,0.98878f,0.99223f,0.99523f,0.99749f,1.00000f,0.00000f,0.93323f,0.89748f,0.89867f,0.88897f,0.89497f,0.90436f,0.91281f,0.92160f,0.93240f,0.94351f,0.94923f,0.95701f,0.96279f,0.96584f,0.96934f,0.97097f,0.97306f,0.97472f,0.97724f,0.97767f,0.97901f,0.98065f,0.98201f,0.98378f,0.98458f,0.98442f,0.98503f,0.98734f,0.99187f,0.99497f,0.99744f,1.00000f,0.00000f,0.93829f,0.91420f,0.89898f,0.89253f,0.89097f,0.89575f,0.89805f,0.91004f,0.91623f,0.92550f,0.93457f,0.94226f,0.95146f,0.95600f,0.96066f,0.96485f,0.96946f,0.97152f,0.97497f,0.97632f,0.97831f,0.97932f,0.98185f,0.98258f,0.98266f,0.98275f,0.98384f,0.98742f,0.99187f,0.99482f,0.99726f,1.00000f,0.00000f,0.93923f,0.91737f,0.90434f,0.88726f,0.88657f,0.88778f,0.89391f,0.89361f,0.90068f,0.90997f,0.91701f,0.92670f,0.93307f,0.94207f,0.94821f,0.95453f,0.96010f,0.96567f,0.96824f,0.97315f,0.97519f,0.97907f,0.97987f,0.98027f,0.98009f,0.98073f,0.98281f,0.98683f,0.99155f,0.99494f,0.99725f,1.00000f,0.00000f,0.94181f,0.92937f,0.90193f,0.89663f,0.89174f,0.88835f,0.88884f,0.88939f,0.89200f,0.89541f,0.90385f,0.90783f,0.91860f,0.92678f,0.93202f,0.94074f,0.94819f,0.95480f,0.95978f,0.96463f,0.96973f,0.97214f,0.97336f,0.97602f,0.97703f,0.97885f,0.98148f,0.98665f,0.99129f,0.99488f,0.99744f,1.00000f,0.00000f,0.94473f,0.93136f,0.91484f,0.89833f,0.89113f,0.88500f,0.88491f,0.88391f,0.88316f,0.88760f,0.89273f,0.89635f,0.90180f,0.91029f,0.91636f,0.92468f,0.93162f,0.93991f,0.94567f,0.95274f,0.95851f,0.96129f,0.96395f,0.96725f,0.96923f,0.97494f,0.97974f,0.98599f,0.99115f,0.99407f,0.99708f,1.00000f,0.00000f,0.94461f,0.93425f,0.91262f,0.90141f,0.89342f,0.88445f,0.88733f,0.88486f,0.88165f,0.87920f,0.88290f,0.88485f,0.88943f,0.89603f,0.90218f,0.90880f,0.91612f,0.92646f,0.92873f,0.93681f,0.94076f,0.94656f,0.94897f,0.95329f,0.95944f,0.96648f,0.97480f,0.98203f,0.98889f,0.99358f,0.99721f,1.00000f,0.00000f,0.94538f,0.93817f,0.92093f,0.90750f,0.89324f,0.88570f,0.88255f,0.88484f,0.88059f,0.87925f,0.88026f,0.88066f,0.88051f,0.88341f,0.88653f,0.89392f,0.90117f,0.90441f,0.91059f,0.91647f,0.92116f,0.92768f,0.93147f,0.93898f,0.94586f,0.95611f,0.96534f,0.97422f,0.98422f,0.99061f,0.99598f,1.00000f,0.00000f,0.94733f,0.94337f,0.92778f,0.91212f,0.90252f,0.88809f,0.88948f,0.87827f,0.87650f,0.87333f,0.87375f,0.87186f,0.87589f,0.87904f,0.87569f,0.88128f,0.88341f,0.88832f,0.88873f,0.89399f,0.89990f,0.90519f,0.90913f,0.91877f,0.92826f,0.93998f,0.95082f,0.96355f,0.97358f,0.98326f,0.99115f,0.99999f,0.00000f,0.94711f,0.94210f,0.92673f,0.91491f,0.90259f,0.89849f,0.89080f,0.88243f,0.87353f,0.87448f,0.86924f,0.87483f,0.86931f,0.86791f,0.86949f,0.86832f,0.86848f,0.86875f,0.87078f,0.87229f,0.87510f,0.88140f,0.89087f,0.89835f,0.90917f,0.92142f,0.93343f,0.94473f,0.95798f,0.96999f,0.98120f,0.99949f,0.00000f,0.95001f,0.93197f,0.92891f,0.91682f,0.90655f,0.89867f,0.89082f,0.88489f,0.87740f,0.87277f,0.87081f,0.86642f,0.86362f,0.86126f,0.85776f,0.85460f,0.85498f,0.85304f,0.85093f,0.85199f,0.85622f,0.86298f,0.86780f,0.87579f,0.88518f,0.89923f,0.91092f,0.92340f,0.93789f,0.94811f,0.96328f,0.99683f,0.00000f,0.95211f,0.93819f,0.92927f,0.91583f,0.90711f,0.89881f,0.89342f,0.88563f,0.87922f,0.87394f,0.86959f,0.86516f,0.86133f,0.85587f,0.84960f,0.84546f,0.83818f,0.83707f,0.83228f,0.82863f,0.83328f,0.83677f,0.84566f,0.85241f,0.86319f,0.87350f,0.88399f,0.89788f,0.90962f,0.92261f,0.93517f,0.98889f,0.00000f,0.94265f,0.94173f,0.93506f,0.91993f,0.91582f,0.90580f,0.89486f,0.88903f,0.87518f,0.87175f,0.86317f,0.85868f,0.85178f,0.84849f,0.84000f,0.83024f,0.82338f,0.82010f,0.81817f,0.81272f,0.81514f,0.81769f,0.82548f,0.83020f,0.83906f,0.84791f,0.85826f,0.86878f,0.87662f,0.88839f,0.89753f,0.97049f,0.00000f,0.94405f,0.94460f,0.92966f,0.92473f,0.91461f,0.90426f,0.89906f,0.88887f,0.87728f,0.87042f,0.85774f,0.85365f,0.84559f,0.83619f,0.82410f,0.81659f,0.80996f,0.80357f,0.80075f,0.80530f,0.79919f,0.79906f,0.80507f,0.80881f,0.81614f,0.82341f,0.82752f,0.83480f,0.84299f,0.84905f,0.85568f,0.94018f,0.00000f,0.93540f,0.94399f,0.93423f,0.93179f,0.92241f,0.90331f,0.89473f,0.88937f,0.87934f,0.86704f,0.85884f,0.84349f,0.83671f,0.82077f,0.81581f,0.79994f,0.79140f,0.78710f,0.78269f,0.78275f,0.77924f,0.77561f,0.78349f,0.78184f,0.79078f,0.79317f,0.79399f,0.79885f,0.80336f,0.79849f,0.80234f,0.90114f,0.00000f,0.93958f,0.93498f,0.92503f,0.92041f,0.91378f,0.90436f,0.89802f,0.88600f,0.87376f,0.86054f,0.84643f,0.83705f,0.81597f,0.80458f,0.79433f,0.79074f,0.77839f,0.77177f,0.77003f,0.76688f,0.76368f,0.76425f,0.76120f,0.76488f,0.76091f,0.76168f,0.76236f,0.76045f,0.75523f,0.74995f,0.74820f,0.85297f,0.00000f,0.93437f,0.92874f,0.92822f,0.91395f,0.91718f,0.89910f,0.88535f,0.87971f,0.86672f,0.85288f,0.83919f,0.82233f,0.80504f,0.79406f,0.77749f,0.77061f,0.76017f,0.75716f,0.75797f,0.74857f,0.74734f,0.74346f,0.74120f,0.73559f,0.73456f,0.72737f,0.72430f,0.71859f,0.70671f,0.69783f,0.68947f,0.80249f,0.00000f,0.92357f,0.93488f,0.92316f,0.91528f,0.90587f,0.90338f,0.88164f,0.87457f,0.85920f,0.83542f,0.82196f,0.80520f,0.78789f,0.77220f,0.76379f,0.75274f,0.74728f,0.74180f,0.73508f,0.73221f,0.72731f,0.72421f,0.71765f,0.71466f,0.70302f,0.69533f,0.68680f,0.67311f,0.66022f,0.64424f,0.62955f,0.74209f,0.00000f,0.93214f,0.92635f,0.92334f,0.91990f,0.90554f,0.89039f,0.87658f,0.85979f,0.84205f,0.82866f,0.80714f,0.79020f,0.77070f,0.75953f,0.74214f,0.73793f,0.73160f,0.72384f,0.71777f,0.71297f,0.71171f,0.69858f,0.69519f,0.68731f,0.67285f,0.65797f,0.64183f,0.62638f,0.60881f,0.59089f,0.57403f,0.68537f,0.00000f,0.92502f,0.92355f,0.91844f,0.90615f,0.89300f,0.88740f,0.86427f,0.84354f,0.82610f,0.80428f,0.78420f,0.77280f,0.75244f,0.73494f,0.73108f,0.72036f,0.71124f,0.70230f,0.69815f,0.69295f,0.68620f,0.67641f,0.66674f,0.65755f,0.64208f,0.61855f,0.59920f,0.57989f,0.55814f,0.54064f,0.51879f,0.62574f,0.00000f,0.91785f,0.91076f,0.91115f,0.90211f,0.88853f,0.87345f,0.85179f,0.83228f,0.81361f,0.78349f,0.76739f,0.74468f,0.73063f,0.71804f,0.70496f,0.70170f,0.69336f,0.68603f,0.67963f,0.66814f,0.66552f,0.65052f,0.63611f,0.62135f,0.60489f,0.58303f,0.56016f,0.53379f,0.51160f,0.49098f,0.46883f,0.56811f,0.00000f,0.90830f,0.90690f,0.90259f,0.87894f,0.87415f,0.85161f,0.83563f,0.80937f,0.79429f,0.76565f,0.74191f,0.72826f,0.70779f,0.69996f,0.69373f,0.67807f,0.67296f,0.66213f,0.65679f,0.65090f,0.63711f,0.62502f,0.60762f,0.58986f,0.57111f,0.54376f,0.52117f,0.49223f,0.46754f,0.44553f,0.41806f,0.51310f,0.00000f,0.90697f,0.89950f,0.88889f,0.87413f,0.84769f,0.83387f,0.80786f,0.79530f,0.76662f,0.73926f,0.71598f,0.70382f,0.68753f,0.67611f,0.66640f,0.65774f,0.64889f,0.64536f,0.63554f,0.62968f,0.61440f,0.59587f,0.57730f,0.55613f,0.53190f,0.50281f,0.47885f,0.45138f,0.42755f,0.40179f,0.37522f,0.46398f,
 0.00000f,1.00000f,0.99836f,1.00000f,0.99073f,0.96986f,0.99147f,0.99437f,0.98429f,0.96686f,0.96470f,0.99728f,0.98099f,0.98883f,0.98608f,0.98442f,0.98233f,0.97197f,0.97068f,0.97608f,0.97343f,0.97197f,0.99130f,0.98746f,0.98575f,0.98674f,0.99168f,0.99089f,0.99331f,0.99488f,0.99333f,0.99651f,1.00000f,0.00000f,0.99017f,0.98756f,0.98210f,0.98111f,0.97783f,0.97654f,0.97730f,0.97523f,0.97351f,0.97333f,0.97508f,0.97270f,0.97512f,0.97350f,0.97511f,0.97531f,0.97561f,0.97625f,0.97826f,0.97835f,0.97886f,0.98192f,0.98331f,0.98418f,0.98604f,0.98783f,0.98983f,0.99145f,0.99335f,0.99514f,0.99756f,1.00000f,0.00000f,0.99053f,0.98640f,0.98290f,0.97988f,0.97824f,0.97716f,0.97469f,0.97414f,0.97420f,0.97357f,0.97421f,0.97534f,0.97331f,0.97444f,0.97493f,0.97564f,0.97612f,0.97656f,0.97754f,0.97913f,0.97989f,0.98125f,0.98274f,0.98409f,0.98599f,0.98777f,0.98972f,0.99138f,0.99334f,0.99536f,0.99775f,1.00000f,0.00000f,0.98477f,0.98566f,0.98176f,0.98019f,0.97743f,0.97699f,0.97613f,0.97437f,0.97362f,0.97279f,0.97350f,0.97301f,0.97320f,0.97363f,0.97479f,0.97460f,0.97626f,0.97629f,0.97787f,0.97892f,0.98026f,0.98155f,0.98270f,0.98442f,0.98579f,0.98756f,0.98955f,0.99142f,0.99324f,0.99550f,0.99774f,1.00000f,0.00000f,0.93735f,0.98104f,0.97981f,0.97860f,0.97803f,0.97491f,0.97540f,0.97478f,0.97371f,0.97260f,0.97328f,0.97437f,0.97377f,0.97386f,0.97415f,0.97467f,0.97595f,0.97650f,0.97702f,0.97861f,0.97928f,0.98107f,0.98265f,0.98417f,0.98595f,0.98770f,0.98942f,0.99131f,0.99356f,0.99546f,0.99775f,1.00000f,0.00000f,0.90543f,0.95546f,0.97589f,0.97832f,0.97656f,0.97562f,0.97427f,0.97490f,0.97333f,0.97238f,0.97355f,0.97263f,0.97409f,0.97305f,0.97441f,0.97403f,0.97530f,0.97664f,0.97746f,0.97956f,0.97964f,0.98126f,0.98261f,0.98382f,0.98586f,0.98778f,0.98956f,0.99130f,0.99300f,0.99550f,0.99775f,1.00000f,0.00000f,0.89912f,0.92066f,0.95510f,0.97123f,0.97393f,0.97559f,0.97400f,0.97314f,0.97450f,0.97261f,0.97279f,0.97173f,0.97291f,0.97357f,0.97384f,0.97515f,0.97483f,0.97655f,0.97739f,0.97877f,0.97962f,0.98172f,0.98266f,0.98397f,0.98576f,0.98747f,0.98922f,0.99117f,0.99288f,0.99539f,0.99759f,1.00000f,0.00000f,0.90007f,0.90318f,0.92549f,0.94850f,0.96450f,0.96985f,0.97144f,0.97370f,0.97170f,0.97307f,0.97270f,0.97247f,0.97313f,0.97301f,0.97370f,0.97438f,0.97529f,0.97672f,0.97714f,0.97807f,0.97937f,0.98156f,0.98196f,0.98393f,0.98589f,0.98778f,0.98917f,0.99081f,0.99247f,0.99526f,0.99753f,1.00000f,0.00000f,0.90859f,0.89607f,0.90725f,0.92520f,0.94365f,0.95585f,0.96524f,0.96876f,0.97065f,0.97097f,0.97245f,0.97137f,0.97322f,0.97243f,0.97327f,0.97403f,0.97516f,0.97620f,0.97693f,0.97822f,0.97972f,0.98013f,0.98212f,0.98387f,0.98570f,0.98730f,0.98913f,0.99027f,0.99256f,0.99529f,0.99754f,1.00000f,0.00000f,0.92086f,0.89844f,0.89636f,0.90680f,0.92198f,0.93350f,0.94901f,0.95735f,0.96435f,0.96747f,0.96890f,0.97178f,0.97150f,0.97287f,0.97360f,0.97427f,0.97511f,0.97598f,0.97643f,0.97807f,0.97927f,0.98016f,0.98130f,0.98439f,0.98521f,0.98716f,0.98862f,0.98928f,0.99215f,0.99516f,0.99748f,1.00000f,0.00000f,0.91778f,0.89853f,0.88925f,0.89630f,0.90597f,0.91468f,0.92847f,0.94157f,0.95041f,0.95708f,0.96361f,0.96517f,0.96886f,0.97187f,0.97243f,0.97365f,0.97482f,0.97539f,0.97630f,0.97731f,0.97851f,0.98062f,0.98135f,0.98338f,0.98516f,0.98681f,0.98698f,0.98799f,0.99212f,0.99506f,0.99750f,1.00000f,0.00000f,0.92319f,0.90060f,0.89294f,0.89082f,0.90059f,0.90086f,0.91135f,0.92011f,0.93243f,0.94293f,0.95041f,0.95594f,0.96248f,0.96654f,0.96971f,0.97079f,0.97413f,0.97498f,0.97711f,0.97719f,0.97855f,0.97984f,0.98155f,0.98391f,0.98458f,0.98477f,0.98522f,0.98780f,0.99161f,0.99494f,0.99747f,1.00000f,0.00000f,0.93453f,0.90751f,0.89334f,0.88926f,0.88754f,0.89466f,0.90000f,0.90467f,0.91355f,0.92574f,0.93430f,0.94333f,0.94874f,0.95592f,0.96223f,0.96531f,0.97014f,0.97101f,0.97362f,0.97617f,0.97799f,0.97947f,0.98146f,0.98266f,0.98278f,0.98322f,0.98391f,0.98713f,0.99183f,0.99483f,0.99736f,1.00000f,0.00000f,0.94324f,0.91318f,0.90106f,0.89303f,0.89014f,0.88952f,0.89114f,0.89748f,0.89819f,0.91132f,0.91803f,0.92792f,0.93452f,0.94133f,0.94841f,0.95538f,0.96043f,0.96476f,0.96932f,0.97124f,0.97589f,0.97866f,0.97938f,0.98061f,0.98049f,0.98130f,0.98369f,0.98691f,0.99151f,0.99487f,0.99727f,1.00000f,0.00000f,0.93731f,0.92816f,0.90810f,0.89441f,0.88802f,0.88670f,0.88842f,0.89084f,0.89014f,0.90054f,0.90144f,0.91215f,0.91803f,0.92752f,0.93452f,0.94254f,0.94819f,0.95296f,0.95935f,0.96536f,0.96938f,0.97179f,0.97511f,0.97587f,0.97768f,0.97830f,0.98160f,0.98627f,0.99154f,0.99439f,0.99737f,1.00000f,0.00000f,0.94826f,0.92328f,0.90814f,0.89963f,0.89192f,0.88607f,0.88572f,0.88490f,0.88475f,0.88848f,0.89132f,0.89669f,0.90689f,0.90745f,0.91642f,0.92525f,0.93083f,0.93857f,0.94585f,0.95399f,0.95835f,0.96166f,0.96577f,0.96840f,0.97071f,0.97400f,0.97983f,0.98576f,0.99008f,0.99429f,0.99699f,1.00000f,0.00000f,0.94146f,0.93396f,0.91369f,0.90430f,0.89259f,0.88575f,0.88327f,0.88079f,0.88379f,0.88488f,0.88583f,0.88428f,0.89387f,0.89600f,0.90281f,0.91068f,0.91662f,0.92449f,0.93127f,0.93604f,0.94294f,0.94611f,0.95046f,0.95541f,0.96130f,0.96717f,0.97466f,0.98280f,0.98883f,0.99309f,0.99693f,1.00000f,0.00000f,0.94629f,0.93572f,0.92129f,0.90956f,0.89526f,0.88842f,0.88440f,0.88311f,0.87579f,0.87795f,0.87660f,0.87957f,0.88272f,0.88539f,0.88752f,0.89525f,0.90110f,0.90646f,0.91021f,0.91782f,0.92289f,0.92748f,0.93170f,0.93938f,0.94626f,0.95516f,0.96549f,0.97462f,0.98311f,0.99067f,0.99610f,1.00000f,0.00000f,0.94516f,0.93596f,0.92213f,0.90631f,0.90272f,0.89786f,0.88778f,0.87973f,0.87955f,0.87647f,0.87609f,0.87415f,0.87400f,0.87796f,0.87872f,0.88048f,0.88611f,0.89011f,0.89080f,0.89753f,0.89900f,0.90495f,0.91232f,0.91957f,0.92858f,0.93820f,0.94958f,0.96248f,0.97268f,0.98288f,0.99164f,0.99999f,0.00000f,0.94898f,0.94650f,0.92287f,0.91718f,0.90469f,0.89116f,0.88711f,0.88097f,0.87451f,0.87730f,0.87538f,0.87114f,0.86736f,0.86412f,0.86946f,0.87041f,0.86916f,0.87309f,0.87075f,0.87627f,0.87660f,0.88153f,0.89071f,0.90007f,0.90895f,0.92002f,0.93261f,0.94465f,0.95640f,0.96888f,0.98137f,0.99950f,0.00000f,0.94735f,0.93366f,0.93091f,0.91513f,0.91186f,0.89804f,0.88914f,0.88655f,0.87621f,0.87603f,0.86931f,0.86905f,0.86499f,0.85956f,0.86152f,0.85749f,0.85404f,0.85421f,0.85341f,0.85004f,0.86124f,0.85870f,0.86556f,0.87893f,0.88720f,0.90010f,0.90930f,0.92373f,0.93623f,0.95046f,0.96302f,0.99680f,0.00000f,0.95304f,0.94148f,0.92923f,0.92123f,0.90733f,0.90594f,0.89110f,0.88989f,0.87935f,0.87249f,0.86891f,0.86512f,0.85933f,0.85957f,0.85023f,0.84594f,0.84179f,0.83741f,0.83001f,0.83448f,0.83936f,0.83728f,0.84705f,0.85402f,0.86154f,0.87403f,0.88428f,0.89917f,0.90782f,0.92201f,0.93723f,0.98882f,0.00000f,0.94838f,0.94295f,0.93437f,0.92136f,0.91123f,0.90977f,0.89554f,0.88934f,0.88096f,0.87310f,0.87187f,0.86174f,0.85050f,0.84478f,0.84274f,0.83582f,0.82420f,0.82351f,0.82157f,0.81777f,0.81526f,0.82019f,0.82313f,0.83428f,0.83991f,0.84716f,0.85649f,0.86879f,0.87920f,0.88736f,0.89963f,0.97005f,0.00000f,0.94084f,0.94109f,0.93179f,0.92588f,0.91421f,0.90216f,0.89751f,0.88586f,0.87996f,0.87637f,0.86262f,0.85409f,0.84870f,0.83815f,0.82543f,0.81800f,0.80988f,0.80503f,0.80234f,0.80000f,0.79855f,0.80204f,0.80343f,0.81069f,0.81475f,0.82002f,0.82916f,0.83391f,0.84149f,0.84679f,0.85276f,0.94043f,0.00000f,0.94333f,0.93845f,0.93385f,0.92863f,0.91626f,0.90076f,0.89387f,0.89053f,0.87933f,0.86380f,0.85370f,0.84469f,0.83525f,0.82504f,0.81745f,0.80326f,0.79877f,0.79406f,0.78698f,0.78492f,0.78180f,0.78139f,0.78103f,0.78734f,0.79047f,0.79228f,0.79185f,0.79769f,0.79737f,0.80200f,0.80542f,0.90154f,0.00000f,0.94284f,0.93804f,0.92965f,0.92749f,0.91543f,0.90597f,0.89425f,0.88292f,0.87628f,0.86453f,0.84897f,0.83838f,0.82229f,0.81522f,0.80044f,0.79439f,0.78251f,0.77654f,0.76856f,0.76793f,0.76889f,0.76761f,0.75913f,0.76373f,0.76275f,0.76138f,0.75739f,0.75625f,0.75552f,0.75184f,0.74763f,0.85393f,0.00000f,0.93115f,0.93568f,0.92462f,0.92122f,0.90906f,0.90699f,0.89582f,0.87559f,0.86826f,0.85262f,0.84220f,0.82856f,0.81220f,0.79481f,0.78611f,0.77759f,0.76245f,0.75937f,0.75309f,0.75327f,0.74908f,0.74476f,0.74301f,0.73796f,0.73816f,0.73092f,0.72527f,0.71514f,0.70704f,0.69912f,0.68786f,0.79995f,0.00000f,0.93382f,0.93390f,0.92433f,0.92098f,0.90342f,0.90274f,0.88604f,0.87772f,0.86022f,0.84379f,0.82288f,0.80905f,0.79163f,0.77689f,0.76811f,0.75762f,0.74711f,0.74234f,0.73912f,0.73244f,0.72886f,0.72249f,0.71764f,0.70798f,0.70289f,0.69682f,0.68321f,0.67085f,0.65713f,0.64661f,0.63333f,0.74150f,0.00000f,0.93858f,0.92999f,0.92256f,0.91443f,0.90393f,0.89375f,0.88531f,0.86593f,0.84573f,0.83293f,0.81175f,0.79173f,0.77862f,0.76178f,0.74831f,0.74215f,0.73312f,0.72365f,0.72072f,0.71629f,0.71279f,0.70630f,0.69509f,0.68448f,0.67500f,0.65562f,0.64535f,0.62727f,0.60886f,0.59171f,0.57269f,0.68421f,0.00000f,0.92723f,0.92010f,0.90863f,0.90870f,0.89266f,0.88928f,0.87285f,0.84130f,0.83452f,0.81362f,0.79412f,0.76864f,0.75325f,0.74611f,0.72625f,0.71803f,0.71230f,0.70640f,0.69986f,0.69317f,0.68983f,0.67459f,0.66917f,0.65152f,0.63914f,0.62140f,0.60101f,0.57947f,0.56125f,0.54101f,0.51942f,0.62517f,0.00000f,0.91702f,0.92011f,0.90917f,0.90071f,0.89028f,0.87148f,0.85481f,0.83393f,0.80951f,0.79614f,0.77171f,0.75323f,0.73631f,0.72161f,0.71757f,0.70084f,0.70316f,0.68953f,0.67853f,0.67153f,0.65876f,0.64930f,0.63879f,0.62470f,0.60558f,0.58201f,0.56090f,0.53570f,0.51003f,0.49092f,0.47029f,0.57042f,0.00000f,0.92123f,0.90978f,0.90602f,0.88832f,0.87773f,0.85601f,0.83365f,0.81723f,0.79411f,0.77163f,0.74922f,0.72956f,0.71723f,0.70048f,0.69120f,0.67786f,0.67097f,0.66825f,0.65753f,0.64890f,0.63653f,0.62639f,0.60617f,0.58958f,0.56255f,0.54316f,0.51725f,0.49367f,0.46775f,0.44176f,0.42076f,0.51641f,0.00000f,0.91512f,0.90182f,0.89241f,0.87555f,0.85380f,0.83428f,0.81770f,0.78926f,0.76933f,0.74457f,0.72318f,0.70763f,0.69171f,0.67913f,0.66536f,0.66350f,0.65374f,0.64180f,0.63431f,0.62578f,0.61019f,0.59628f,0.57774f,0.55318f,0.53124f,0.50254f,0.47717f,0.44806f,0.42380f,0.40017f,0.37730f,0.46523f,
 0.00000f,1.00000f,0.99862f,1.00000f,0.99091f,0.96841f,0.99039f,0.99393f,0.98266f,0.96788f,0.96419f,0.99904f,0.98171f,0.98807f,0.98519f,0.98405f,0.98139f,0.97158f,0.97146f,0.97584f,0.97330f,0.97179f,0.99080f,0.98711f,0.98601f,0.98678f,0.99130f,0.99087f,0.99321f,0.99498f,0.99212f,0.99651f,1.00000f,0.00000f,0.99115f,0.98724f,0.98411f,0.98172f,0.97681f,0.97736f,0.97736f,0.97619f,0.97285f,0.97263f,0.97506f,0.97300f,0.97425f,0.97470f,0.97535f,0.97547f,0.97494f,0.97687f,0.97838f,0.97836f,0.97953f,0.98217f,0.98272f,0.98426f,0.98606f,0.98790f,0.98972f,0.99159f,0.99342f,0.99492f,0.99758f,1.00000f,0.00000f,0.99305f,0.98506f,0.98202f,0.97926f,0.97805f,0.97637f,0.97619f,0.97513f,0.97477f,0.97276f,0.97313f,0.97363f,0.97422f,0.97439f,0.97452f,0.97516f,0.97639f,0.97666f,0.97775f,0.97865f,0.98003f,0.98130f,0.98249f,0.98412f,0.98595f,0.98779f,0.98952f,0.99143f,0.99335f,0.99547f,0.99774f,1.00000f,0.00000f,0.98416f,0.98664f,0.98194f,0.98160f,0.97857f,0.97608f,0.97460f,0.97449f,0.97388f,0.97380f,0.97419f,0.97291f,0.97358f,0.97430f,0.97448f,0.97443f,0.97548f,0.97657f,0.97756f,0.97897f,0.98026f,0.98138f,0.98307f,0.98426f,0.98591f,0.98764f,0.98940f,0.99118f,0.99332f,0.99550f,0.99776f,1.00000f,0.00000f,0.93927f,0.98176f,0.98024f,0.97832f,0.97639f,0.97647f,0.97495f,0.97332f,0.97332f,0.97277f,0.97216f,0.97333f,0.97297f,0.97424f,0.97444f,0.97487f,0.97592f,0.97626f,0.97768f,0.97839f,0.97971f,0.98090f,0.98279f,0.98404f,0.98555f,0.98755f,0.98943f,0.99132f,0.99332f,0.99541f,0.99775f,1.00000f,0.00000f,0.91016f,0.95590f,0.97565f,0.97881f,0.97653f,0.97416f,0.97479f,0.97298f,0.97350f,0.97338f,0.97285f,0.97325f,0.97341f,0.97370f,0.97448f,0.97463f,0.97556f,0.97657f,0.97791f,0.97794f,0.97995f,0.98196f,0.98252f,0.98409f,0.98595f,0.98761f,0.98915f,0.99110f,0.99318f,0.99541f,0.99780f,1.00000f,0.00000f,0.90319f,0.92302f,0.95650f,0.97132f,0.97399f,0.97631f,0.97427f,0.97247f,0.97314f,0.97285f,0.97272f,0.97289f,0.97272f,0.97365f,0.97376f,0.97405f,0.97550f,0.97632f,0.97724f,0.97873f,0.98010f,0.98117f,0.98276f,0.98419f,0.98569f,0.98729f,0.98933f,0.99114f,0.99269f,0.99538f,0.99762f,1.00000f,0.00000f,0.90021f,0.90115f,0.92699f,0.95064f,0.96377f,0.97091f,0.97316f,0.97268f,0.97180f,0.97238f,0.97254f,0.97335f,0.97307f,0.97267f,0.97368f,0.97503f,0.97573f,0.97670f,0.97687f,0.97813f,0.97961f,0.98116f,0.98300f,0.98421f,0.98570f,0.98710f,0.98912f,0.99094f,0.99239f,0.99527f,0.99752f,1.00000f,0.00000f,0.91226f,0.89634f,0.90462f,0.92447f,0.94483f,0.95785f,0.96469f,0.96965f,0.97076f,0.97298f,0.97127f,0.97254f,0.97322f,0.97315f,0.97367f,0.97475f,0.97438f,0.97676f,0.97694f,0.97883f,0.97900f,0.98026f,0.98233f,0.98333f,0.98569f,0.98699f,0.98893f,0.99037f,0.99233f,0.99546f,0.99746f,1.00000f,0.00000f,0.91889f,0.90063f,0.89627f,0.90622f,0.92153f,0.93911f,0.94829f,0.95799f,0.96358f,0.96702f,0.97136f,0.97224f,0.97176f,0.97243f,0.97286f,0.97322f,0.97510f,0.97669f,0.97749f,0.97797f,0.97927f,0.98024f,0.98195f,0.98395f,0.98568f,0.98728f,0.98853f,0.98970f,0.99194f,0.99526f,0.99739f,1.00000f,0.00000f,0.92625f,0.89981f,0.89385f,0.89906f,0.90479f,0.91588f,0.93226f,0.94375f,0.95256f,0.95703f,0.96164f,0.96671f,0.96892f,0.97102f,0.97204f,0.97442f,0.97435f,0.97588f,0.97637f,0.97747f,0.97924f,0.98068f,0.98181f,0.98344f,0.98547f,0.98699f,0.98773f,0.98801f,0.99185f,0.99503f,0.99745f,1.00000f,0.00000f,0.92912f,0.90510f,0.89626f,0.89336f,0.89580f,0.90053f,0.91239f,0.92068f,0.93351f,0.94353f,0.95003f,0.95778f,0.96318f,0.96685f,0.96848f,0.97178f,0.97423f,0.97499f,0.97653f,0.97804f,0.97831f,0.98004f,0.98147f,0.98305f,0.98520f,0.98552f,0.98555f,0.98770f,0.99144f,0.99509f,0.99733f,1.00000f,0.00000f,0.93574f,0.91572f,0.89643f,0.88811f,0.89352f,0.89436f,0.89753f,0.90741f,0.91519f,0.92537f,0.93233f,0.94299f,0.95160f,0.95535f,0.96218f,0.96471f,0.96854f,0.97116f,0.97394f,0.97670f,0.97873f,0.98024f,0.98102f,0.98284f,0.98386f,0.98334f,0.98449f,0.98696f,0.99130f,0.99480f,0.99725f,1.00000f,0.00000f,0.94457f,0.91618f,0.90073f,0.89336f,0.88682f,0.89030f,0.89144f,0.89582f,0.89884f,0.90769f,0.91915f,0.92645f,0.93614f,0.93969f,0.94923f,0.95554f,0.96119f,0.96401f,0.97013f,0.97130f,0.97558f,0.97776f,0.98036f,0.98090f,0.98167f,0.98182f,0.98252f,0.98594f,0.99087f,0.99480f,0.99734f,1.00000f,0.00000f,0.94610f,0.92198f,0.90304f,0.89631f,0.89090f,0.88975f,0.88733f,0.89200f,0.89215f,0.89306f,0.90448f,0.90840f,0.91991f,0.92476f,0.93216f,0.93952f,0.94856f,0.95317f,0.95842f,0.96455f,0.96954f,0.97392f,0.97569f,0.97781f,0.97790f,0.97861f,0.98272f,0.98626f,0.99126f,0.99466f,0.99733f,1.00000f,0.00000f,0.94636f,0.92370f,0.91062f,0.89894f,0.89039f,0.88719f,0.88462f,0.88604f,0.88710f,0.88496f,0.89131f,0.89436f,0.90248f,0.90980f,0.91696f,0.92404f,0.93299f,0.93888f,0.94604f,0.95336f,0.95860f,0.96268f,0.96606f,0.96777f,0.97153f,0.97462f,0.97787f,0.98587f,0.99081f,0.99415f,0.99710f,1.00000f,0.00000f,0.94442f,0.93390f,0.91838f,0.90758f,0.89552f,0.88584f,0.88482f,0.88420f,0.88269f,0.88059f,0.88201f,0.88676f,0.89167f,0.89643f,0.90269f,0.90706f,0.91567f,0.92043f,0.93112f,0.93825f,0.94214f,0.94882f,0.95219f,0.95638f,0.96119f,0.96626f,0.97428f,0.98213f,0.98850f,0.99343f,0.99719f,1.00000f,0.00000f,0.94631f,0.93093f,0.92355f,0.90680f,0.90079f,0.89350f,0.88033f,0.87821f,0.87684f,0.87703f,0.88018f,0.88125f,0.88092f,0.88284f,0.89062f,0.89310f,0.90185f,0.90744f,0.91317f,0.91743f,0.92458f,0.92794f,0.93277f,0.93992f,0.94816f,0.95505f,0.96400f,0.97526f,0.98314f,0.98990f,0.99561f,1.00000f,0.00000f,0.94719f,0.93381f,0.92313f,0.90824f,0.89524f,0.89148f,0.88278f,0.88542f,0.87914f,0.87819f,0.87489f,0.87541f,0.87255f,0.87578f,0.87888f,0.88134f,0.88669f,0.88821f,0.89182f,0.89525f,0.90313f,0.90648f,0.91274f,0.91863f,0.92759f,0.94099f,0.95083f,0.96271f,0.97278f,0.98222f,0.99169f,0.99999f,0.00000f,0.94430f,0.93955f,0.92302f,0.91100f,0.90555f,0.89527f,0.88858f,0.88179f,0.88043f,0.87291f,0.87451f,0.86848f,0.86726f,0.86809f,0.86743f,0.86919f,0.86879f,0.87404f,0.87397f,0.87413f,0.88031f,0.88500f,0.88951f,0.89935f,0.90884f,0.92131f,0.93193f,0.94572f,0.95727f,0.96856f,0.98121f,0.99949f,0.00000f,0.94585f,0.94180f,0.93224f,0.91520f,0.90802f,0.89457f,0.89084f,0.88675f,0.87882f,0.87372f,0.87351f,0.86569f,0.86564f,0.86112f,0.85925f,0.85894f,0.85423f,0.85650f,0.85481f,0.85368f,0.85942f,0.86381f,0.86626f,0.87775f,0.88715f,0.89749f,0.91260f,0.92406f,0.93747f,0.94785f,0.96245f,0.99686f,0.00000f,0.93359f,0.94065f,0.93647f,0.91762f,0.90541f,0.90224f,0.89187f,0.88225f,0.87453f,0.87176f,0.87137f,0.86615f,0.86011f,0.85118f,0.84928f,0.84546f,0.84144f,0.83865f,0.83380f,0.83685f,0.83615f,0.84038f,0.84794f,0.85380f,0.86333f,0.87277f,0.88599f,0.89702f,0.91066f,0.92086f,0.93675f,0.98887f,0.00000f,0.94026f,0.94053f,0.92865f,0.91933f,0.91095f,0.90469f,0.89442f,0.88451f,0.88426f,0.87231f,0.86895f,0.86125f,0.85808f,0.84661f,0.83572f,0.83188f,0.83299f,0.82454f,0.82176f,0.81870f,0.82022f,0.82170f,0.82607f,0.83180f,0.84090f,0.84654f,0.85908f,0.86719f,0.87850f,0.88876f,0.89913f,0.97045f,0.00000f,0.94477f,0.93711f,0.93058f,0.92206f,0.91488f,0.90798f,0.90113f,0.88768f,0.88510f,0.87049f,0.86214f,0.85542f,0.84700f,0.83507f,0.82671f,0.81740f,0.81417f,0.80515f,0.80689f,0.80180f,0.80289f,0.80503f,0.80219f,0.80507f,0.81558f,0.82093f,0.82866f,0.83602f,0.83984f,0.84701f,0.85469f,0.94030f,0.00000f,0.94393f,0.93810f,0.92679f,0.92517f,0.91903f,0.91091f,0.90219f,0.88749f,0.87954f,0.86989f,0.85655f,0.85021f,0.83708f,0.82381f,0.81463f,0.81197f,0.79420f,0.79577f,0.78898f,0.78540f,0.78379f,0.78358f,0.78291f,0.78400f,0.78770f,0.79357f,0.79469f,0.79695f,0.79759f,0.80225f,0.80241f,0.90297f,0.00000f,0.94349f,0.93795f,0.93696f,0.92391f,0.91229f,0.90797f,0.89323f,0.88987f,0.87621f,0.86816f,0.84943f,0.83949f,0.82539f,0.81697f,0.79842f,0.79293f,0.78398f,0.77743f,0.77251f,0.76912f,0.76806f,0.76635f,0.76420f,0.76270f,0.76285f,0.76086f,0.75652f,0.75776f,0.75428f,0.75127f,0.74894f,0.85482f,0.00000f,0.93894f,0.93386f,0.92635f,0.92435f,0.91258f,0.90312f,0.89569f,0.88380f,0.86660f,0.85495f,0.83854f,0.82192f,0.81335f,0.79756f,0.78549f,0.77340f,0.77265f,0.76419f,0.75594f,0.74959f,0.75039f,0.74311f,0.73947f,0.73817f,0.73943f,0.73003f,0.72222f,0.71788f,0.70687f,0.69904f,0.68826f,0.79869f,0.00000f,0.93483f,0.93311f,0.92566f,0.91341f,0.91006f,0.90002f,0.89386f,0.87399f,0.86064f,0.84494f,0.83180f,0.81225f,0.79897f,0.78646f,0.76916f,0.75510f,0.74965f,0.74430f,0.74093f,0.73342f,0.72971f,0.72292f,0.71911f,0.70891f,0.70364f,0.69345f,0.68486f,0.67263f,0.65705f,0.64510f,0.62818f,0.74216f,0.00000f,0.93505f,0.91923f,0.92602f,0.91678f,0.90244f,0.89392f,0.88016f,0.86333f,0.85360f,0.83446f,0.81707f,0.79460f,0.77877f,0.76870f,0.75281f,0.74513f,0.73324f,0.73099f,0.71860f,0.71960f,0.71555f,0.70030f,0.69584f,0.68328f,0.67159f,0.65654f,0.64050f,0.62440f,0.60980f,0.59173f,0.57250f,0.68381f,0.00000f,0.92805f,0.92806f,0.91609f,0.90145f,0.89165f,0.88607f,0.86580f,0.84890f,0.83449f,0.81764f,0.79485f,0.77264f,0.76055f,0.74766f,0.73355f,0.71988f,0.71874f,0.71097f,0.70135f,0.69583f,0.68891f,0.67513f,0.66871f,0.65460f,0.63840f,0.61749f,0.60104f,0.58056f,0.55908f,0.53834f,0.52117f,0.62599f,0.00000f,0.91696f,0.91551f,0.91237f,0.89469f,0.88716f,0.87414f,0.85675f,0.83309f,0.81652f,0.79615f,0.77818f,0.75494f,0.74452f,0.72682f,0.71549f,0.70346f,0.69882f,0.69191f,0.68136f,0.67333f,0.66607f,0.64969f,0.63915f,0.61918f,0.60406f,0.58152f,0.55696f,0.53853f,0.51073f,0.49018f,0.46601f,0.56811f,0.00000f,0.90958f,0.90767f,0.90059f,0.89137f,0.87887f,0.85947f,0.83891f,0.81980f,0.79779f,0.77381f,0.75205f,0.73457f,0.71946f,0.70350f,0.69669f,0.69346f,0.67856f,0.66857f,0.65965f,0.65213f,0.63883f,0.62323f,0.60687f,0.58953f,0.56508f,0.54101f,0.51605f,0.49298f,0.46872f,0.44447f,0.42054f,0.51442f,0.00000f,0.90873f,0.89746f,0.89188f,0.87000f,0.86305f,0.83956f,0.81852f,0.79899f,0.77561f,0.75379f,0.73536f,0.71457f,0.69462f,0.68058f,0.67457f,0.66461f,0.65164f,0.64578f,0.63510f,0.62017f,0.60930f,0.59503f,0.57482f,0.55203f,0.52870f,0.50345f,0.47384f,0.45264f,0.42433f,0.39958f,0.37751f,0.46753f,
 0.00000f,1.00000f,0.99758f,1.00000f,0.99039f,0.96848f,0.99089f,0.99373f,0.98357f,0.96570f,0.96511f,0.99851f,0.98160f,0.98850f,0.98595f,0.98442f,0.98184f,0.97209f,0.97079f,0.97619f,0.97311f,0.97217f,0.99137f,0.98740f,0.98579f,0.98672f,0.99161f,0.99094f,0.99325f,0.99507f,0.99281f,0.99655f,1.00000f,0.00000f,0.99217f,0.98553f,0.98334f,0.98076f,0.97699f,0.97780f,0.97690f,0.97608f,0.97343f,0.97381f,0.97583f,0.97356f,0.97433f,0.97419f,0.97550f,0.97575f,0.97526f,0.97629f,0.97816f,0.97807f,0.97910f,0.98186f,0.98283f,0.98409f,0.98564f,0.98812f,0.98946f,0.99157f,0.99344f,0.99500f,0.99758f,1.00000f,0.00000f,0.99019f,0.98634f,0.98241f,0.97981f,0.97794f,0.97549f,0.97567f,0.97420f,0.97331f,0.97327f,0.97382f,0.97390f,0.97327f,0.97399f,0.97399f,0.97473f,0.97576f,0.97702f,0.97771f,0.97952f,0.98034f,0.98110f,0.98285f,0.98474f,0.98601f,0.98757f,0.98963f,0.99145f,0.99331f,0.99545f,0.99773f,1.00000f,0.00000f,0.98509f,0.98574f,0.98313f,0.98063f,0.97775f,0.97628f,0.97455f,0.97490f,0.97475f,0.97324f,0.97337f,0.97386f,0.97338f,0.97459f,0.97500f,0.97496f,0.97571f,0.97710f,0.97825f,0.97839f,0.97970f,0.98156f,0.98272f,0.98401f,0.98566f,0.98768f,0.98942f,0.99134f,0.99330f,0.99550f,0.99772f,1.00000f,0.00000f,0.94037f,0.98057f,0.98135f,0.97954f,0.97568f,0.97529f,0.97476f,0.97352f,0.97285f,0.97338f,0.97304f,0.97317f,0.97282f,0.97271f,0.97471f,0.97483f,0.97559f,0.97676f,0.97794f,0.97845f,0.97998f,0.98112f,0.98245f,0.98446f,0.98589f,0.98780f,0.98943f,0.99129f,0.99336f,0.99552f,0.99780f,1.00000f,0.00000f,0.90732f,0.95563f,0.97592f,0.97713f,0.97639f,0.97537f,0.97462f,0.97388f,0.97385f,0.97413f,0.97358f,0.97369f,0.97356f,0.97352f,0.97462f,0.97542f,0.97523f,0.97694f,0.97783f,0.97850f,0.97989f,0.98097f,0.98257f,0.98425f,0.98613f,0.98765f,0.98942f,0.99130f,0.99314f,0.99538f,0.99776f,1.00000f,0.00000f,0.89742f,0.92152f,0.95544f,0.97179f,0.97482f,0.97583f,0.97445f,0.97343f,0.97327f,0.97230f,0.97276f,0.97280f,0.97369f,0.97241f,0.97397f,0.97446f,0.97612f,0.97658f,0.97780f,0.97901f,0.97975f,0.98108f,0.98263f,0.98409f,0.98575f,0.98721f,0.98918f,0.99110f,0.99290f,0.99539f,0.99763f,1.00000f,0.00000f,0.90594f,0.90717f,0.92543f,0.95119f,0.96751f,0.97143f,0.97295f,0.97243f,0.97308f,0.97242f,0.97246f,0.97283f,0.97261f,0.97382f,0.97352f,0.97442f,0.97555f,0.97594f,0.97703f,0.97849f,0.97968f,0.98056f,0.98225f,0.98398f,0.98553f,0.98723f,0.98926f,0.99098f,0.99227f,0.99530f,0.99759f,1.00000f,0.00000f,0.91202f,0.89941f,0.90777f,0.92514f,0.94570f,0.95739f,0.96519f,0.96794f,0.97196f,0.97125f,0.97254f,0.97189f,0.97233f,0.97231f,0.97276f,0.97401f,0.97458f,0.97587f,0.97660f,0.97775f,0.97978f,0.98044f,0.98196f,0.98392f,0.98551f,0.98735f,0.98945f,0.99026f,0.99198f,0.99526f,0.99756f,1.00000f,0.00000f,0.91570f,0.89665f,0.89723f,0.90715f,0.92143f,0.93804f,0.94774f,0.95828f,0.96344f,0.96873f,0.97020f,0.97175f,0.97143f,0.97353f,0.97452f,0.97327f,0.97497f,0.97578f,0.97627f,0.97809f,0.97894f,0.98082f,0.98188f,0.98339f,0.98525f,0.98715f,0.98893f,0.98929f,0.99150f,0.99515f,0.99750f,1.00000f,0.00000f,0.92683f,0.89799f,0.89068f,0.89554f,0.90207f,0.91510f,0.92924f,0.94302f,0.94989f,0.95796f,0.96224f,0.96964f,0.96863f,0.97074f,0.97269f,0.97319f,0.97453f,0.97595f,0.97609f,0.97679f,0.97890f,0.97990f,0.98200f,0.98349f,0.98523f,0.98696f,0.98805f,0.98841f,0.99133f,0.99508f,0.99752f,1.00000f,0.00000f,0.92829f,0.90891f,0.88982f,0.89121f,0.89662f,0.90177f,0.91524f,0.92438f,0.93289f,0.93977f,0.95061f,0.95804f,0.96319f,0.96623f,0.96948f,0.97152f,0.97264f,0.97513f,0.97603f,0.97726f,0.97825f,0.98067f,0.98136f,0.98307f,0.98533f,0.98581f,0.98675f,0.98716f,0.99105f,0.99470f,0.99742f,1.00000f,0.00000f,0.93386f,0.91410f,0.89939f,0.88965f,0.89095f,0.89480f,0.89796f,0.90580f,0.91581f,0.92487f,0.93318f,0.94231f,0.94890f,0.95524f,0.96200f,0.96496f,0.96836f,0.97144f,0.97384f,0.97726f,0.97820f,0.98012f,0.98146f,0.98313f,0.98344f,0.98462f,0.98522f,0.98675f,0.99107f,0.99473f,0.99727f,1.00000f,0.00000f,0.94365f,0.91749f,0.90228f,0.88926f,0.88764f,0.89028f,0.89153f,0.89783f,0.90195f,0.91053f,0.92118f,0.92705f,0.93370f,0.94225f,0.95043f,0.95424f,0.95987f,0.96538f,0.96926f,0.97213f,0.97515f,0.97711f,0.98106f,0.98185f,0.98236f,0.98258f,0.98317f,0.98626f,0.99049f,0.99445f,0.99748f,1.00000f,0.00000f,0.94360f,0.92571f,0.90514f,0.89178f,0.88621f,0.88503f,0.88754f,0.88851f,0.89122f,0.89637f,0.90203f,0.90808f,0.92007f,0.92600f,0.93192f,0.94038f,0.94793f,0.95442f,0.96005f,0.96486f,0.96913f,0.97317f,0.97588f,0.97691f,0.97792f,0.97917f,0.98153f,0.98589f,0.99053f,0.99464f,0.99731f,1.00000f,0.00000f,0.93794f,0.92799f,0.90906f,0.89949f,0.88889f,0.88604f,0.88535f,0.88455f,0.88696f,0.88987f,0.89179f,0.89804f,0.90160f,0.90877f,0.91655f,0.92501f,0.93222f,0.94048f,0.94513f,0.95246f,0.95896f,0.96243f,0.96663f,0.97079f,0.97300f,0.97572f,0.97981f,0.98515f,0.99028f,0.99428f,0.99704f,1.00000f,0.00000f,0.94016f,0.92815f,0.91718f,0.90149f,0.89272f,0.88377f,0.88422f,0.87768f,0.88293f,0.88202f,0.88443f,0.88679f,0.88743f,0.89649f,0.90149f,0.90749f,0.91640f,0.92238f,0.93166f,0.93498f,0.94349f,0.94558f,0.95210f,0.95702f,0.96165f,0.96849f,0.97387f,0.98121f,0.98816f,0.99327f,0.99693f,1.00000f,0.00000f,0.94545f,0.93301f,0.92421f,0.90920f,0.89754f,0.88932f,0.88377f,0.88115f,0.88030f,0.87993f,0.87981f,0.88122f,0.87934f,0.88850f,0.89324f,0.89702f,0.89926f,0.90645f,0.90968f,0.91691f,0.92434f,0.92916f,0.93494f,0.94085f,0.94790f,0.95568f,0.96501f,0.97449f,0.98291f,0.99038f,0.99573f,1.00000f,0.00000f,0.94443f,0.93870f,0.92381f,0.91486f,0.90176f,0.89639f,0.89208f,0.88260f,0.87779f,0.87597f,0.87633f,0.87625f,0.87431f,0.87843f,0.87901f,0.87866f,0.88581f,0.89126f,0.89343f,0.89898f,0.90305f,0.90773f,0.91329f,0.92204f,0.92973f,0.94064f,0.95313f,0.96240f,0.97287f,0.98247f,0.99132f,0.99999f,0.00000f,0.94657f,0.93530f,0.92537f,0.91748f,0.89973f,0.89318f,0.88923f,0.87962f,0.87439f,0.87727f,0.86890f,0.87006f,0.86981f,0.87334f,0.86917f,0.86997f,0.87080f,0.87240f,0.87395f,0.87386f,0.88104f,0.88709f,0.89209f,0.89852f,0.91134f,0.92083f,0.93310f,0.94454f,0.95693f,0.96865f,0.98128f,0.99948f,0.00000f,0.94495f,0.93567f,0.92775f,0.92247f,0.90424f,0.89819f,0.89201f,0.88652f,0.87940f,0.87683f,0.86877f,0.86706f,0.86431f,0.86496f,0.86184f,0.85938f,0.85310f,0.85833f,0.85623f,0.85619f,0.85810f,0.86324f,0.87069f,0.87901f,0.88774f,0.89688f,0.90979f,0.92290f,0.93495f,0.94988f,0.96193f,0.99684f,0.00000f,0.94931f,0.93911f,0.93263f,0.92736f,0.90953f,0.90633f,0.89474f,0.88490f,0.88047f,0.86878f,0.86707f,0.86317f,0.86228f,0.85811f,0.84994f,0.84831f,0.84729f,0.83882f,0.83901f,0.83670f,0.84154f,0.84254f,0.84927f,0.85704f,0.86206f,0.87558f,0.88514f,0.89767f,0.91033f,0.92365f,0.93501f,0.98907f,0.00000f,0.94108f,0.93319f,0.93435f,0.92330f,0.91359f,0.90215f,0.89655f,0.88367f,0.87479f,0.87203f,0.86246f,0.85957f,0.85453f,0.84701f,0.84494f,0.83451f,0.83011f,0.82758f,0.81753f,0.81709f,0.81969f,0.82129f,0.82202f,0.83332f,0.84064f,0.84657f,0.85679f,0.86695f,0.87735f,0.88820f,0.89962f,0.97039f,0.00000f,0.93627f,0.93648f,0.93393f,0.92501f,0.91161f,0.90741f,0.89129f,0.88400f,0.87992f,0.87294f,0.86510f,0.86037f,0.84642f,0.84090f,0.83013f,0.82425f,0.82262f,0.80867f,0.80316f,0.80540f,0.80140f,0.80642f,0.80460f,0.81082f,0.81243f,0.82301f,0.82510f,0.83281f,0.84124f,0.84836f,0.85488f,0.94082f,0.00000f,0.93850f,0.93990f,0.93454f,0.92575f,0.90737f,0.90613f,0.89828f,0.88892f,0.87643f,0.86700f,0.85751f,0.85561f,0.84134f,0.82960f,0.81591f,0.80932f,0.80415f,0.79605f,0.78750f,0.78724f,0.78385f,0.78277f,0.78710f,0.78680f,0.78854f,0.79455f,0.79644f,0.79888f,0.80085f,0.80310f,0.80264f,0.90085f,0.00000f,0.93934f,0.93568f,0.93110f,0.91801f,0.90845f,0.90743f,0.88944f,0.88247f,0.88028f,0.86311f,0.85813f,0.84164f,0.83176f,0.82036f,0.80385f,0.79414f,0.78532f,0.77795f,0.77516f,0.77149f,0.76493f,0.76851f,0.76560f,0.76435f,0.76089f,0.75863f,0.75827f,0.76007f,0.75511f,0.75189f,0.74507f,0.85439f,0.00000f,0.93251f,0.92982f,0.92528f,0.92231f,0.91184f,0.90671f,0.89516f,0.87984f,0.87255f,0.85540f,0.84634f,0.83113f,0.81876f,0.79983f,0.78880f,0.78110f,0.77283f,0.76403f,0.75925f,0.75070f,0.74877f,0.74425f,0.74586f,0.74171f,0.73175f,0.72690f,0.72391f,0.71137f,0.70633f,0.69638f,0.68704f,0.80161f,0.00000f,0.93464f,0.93056f,0.93031f,0.91658f,0.90820f,0.89891f,0.88857f,0.88394f,0.86617f,0.84724f,0.82947f,0.82083f,0.79992f,0.78133f,0.77467f,0.76574f,0.75594f,0.75083f,0.74285f,0.73762f,0.72715f,0.72419f,0.71897f,0.71056f,0.70139f,0.69439f,0.68115f,0.66928f,0.65829f,0.64404f,0.62931f,0.74327f,0.00000f,0.93090f,0.92503f,0.92432f,0.91458f,0.90124f,0.89357f,0.88329f,0.86782f,0.85649f,0.83610f,0.81685f,0.80095f,0.78297f,0.77273f,0.75959f,0.74995f,0.73664f,0.73182f,0.72541f,0.71500f,0.71209f,0.70557f,0.69614f,0.68217f,0.67001f,0.65505f,0.64360f,0.62624f,0.60823f,0.59281f,0.57420f,0.68387f,0.00000f,0.92756f,0.91917f,0.91283f,0.90826f,0.90197f,0.88821f,0.87619f,0.86293f,0.84064f,0.81950f,0.80272f,0.78183f,0.76962f,0.75627f,0.74080f,0.73104f,0.71691f,0.71267f,0.70524f,0.69473f,0.68692f,0.67854f,0.66694f,0.65637f,0.63893f,0.61825f,0.59887f,0.57913f,0.55956f,0.54055f,0.51881f,0.62543f,0.00000f,0.92429f,0.91450f,0.91755f,0.90321f,0.89529f,0.87539f,0.85597f,0.84219f,0.81646f,0.80227f,0.78126f,0.75807f,0.74548f,0.73338f,0.71908f,0.71119f,0.70229f,0.69132f,0.68220f,0.67142f,0.66516f,0.65311f,0.63753f,0.62046f,0.60305f,0.58039f,0.55800f,0.53533f,0.51070f,0.49171f,0.46730f,0.56963f,0.00000f,0.91866f,0.90974f,0.89496f,0.89089f,0.87782f,0.86200f,0.84556f,0.82099f,0.80327f,0.77957f,0.76240f,0.73900f,0.72395f,0.70753f,0.69649f,0.68780f,0.68068f,0.67148f,0.65911f,0.65208f,0.63969f,0.62339f,0.60913f,0.58474f,0.56420f,0.53634f,0.51762f,0.49096f,0.46784f,0.44287f,0.42165f,0.51661f,0.00000f,0.91300f,0.89557f,0.89125f,0.87343f,0.86390f,0.84744f,0.82152f,0.80346f,0.78203f,0.75844f,0.73528f,0.71358f,0.70483f,0.68844f,0.67361f,0.66707f,0.66117f,0.64438f,0.63723f,0.62618f,0.61002f,0.59637f,0.57455f,0.55016f,0.52581f,0.49979f,0.47514f,0.44788f,0.42419f,0.39824f,0.37804f,0.46681f,
 0.00000f,1.00000f,0.99766f,1.00000f,0.99152f,0.96968f,0.99168f,0.99458f,0.98353f,0.96776f,0.96460f,0.99861f,0.98203f,0.98893f,0.98493f,0.98482f,0.98185f,0.97250f,0.97096f,0.97644f,0.97286f,0.97220f,0.99187f,0.98786f,0.98564f,0.98700f,0.99134f,0.99090f,0.99338f,0.99506f,0.99340f,0.99697f,1.00000f,0.00000f,0.99187f,0.98712f,0.98274f,0.98111f,0.97770f,0.97748f,0.97509f,0.97491f,0.97394f,0.97252f,0.97433f,0.97374f,0.97534f,0.97460f,0.97582f,0.97596f,0.97513f,0.97592f,0.97793f,0.97877f,0.97926f,0.98202f,0.98293f,0.98390f,0.98571f,0.98785f,0.98948f,0.99146f,0.99329f,0.99505f,0.99759f,1.00000f,0.00000f,0.99060f,0.98508f,0.98239f,0.98048f,0.97664f,0.97697f,0.97577f,0.97464f,0.97412f,0.97304f,0.97404f,0.97343f,0.97340f,0.97383f,0.97480f,0.97496f,0.97566f,0.97678f,0.97753f,0.97859f,0.98030f,0.98158f,0.98275f,0.98437f,0.98621f,0.98736f,0.98948f,0.99122f,0.99330f,0.99561f,0.99773f,1.00000f,0.00000f,0.98335f,0.98413f,0.98285f,0.97972f,0.97703f,0.97638f,0.97501f,0.97511f,0.97414f,0.97334f,0.97390f,0.97334f,0.97351f,0.97397f,0.97456f,0.97473f,0.97588f,0.97641f,0.97761f,0.97907f,0.98005f,0.98119f,0.98293f,0.98447f,0.98590f,0.98766f,0.98941f,0.99130f,0.99338f,0.99542f,0.99775f,1.00000f,0.00000f,0.93779f,0.98144f,0.98286f,0.97900f,0.97723f,0.97629f,0.97547f,0.97357f,0.97385f,0.97341f,0.97294f,0.97315f,0.97357f,0.97399f,0.97395f,0.97561f,0.97590f,0.97604f,0.97768f,0.97902f,0.97980f,0.98150f,0.98250f,0.98448f,0.98603f,0.98764f,0.98948f,0.99131f,0.99334f,0.99553f,0.99773f,1.00000f,0.00000f,0.90792f,0.95448f,0.97650f,0.97961f,0.97631f,0.97579f,0.97501f,0.97274f,0.97419f,0.97311f,0.97318f,0.97343f,0.97324f,0.97317f,0.97434f,0.97474f,0.97585f,0.97669f,0.97794f,0.97944f,0.97994f,0.98051f,0.98259f,0.98424f,0.98586f,0.98758f,0.98944f,0.99110f,0.99345f,0.99549f,0.99778f,1.00000f,0.00000f,0.90115f,0.92165f,0.95396f,0.97142f,0.97565f,0.97465f,0.97391f,0.97387f,0.97267f,0.97293f,0.97340f,0.97314f,0.97392f,0.97322f,0.97393f,0.97470f,0.97529f,0.97678f,0.97730f,0.97894f,0.97984f,0.98107f,0.98257f,0.98437f,0.98535f,0.98752f,0.98924f,0.99135f,0.99298f,0.99531f,0.99761f,1.00000f,0.00000f,0.90487f,0.90301f,0.92748f,0.95062f,0.96375f,0.97103f,0.97376f,0.97369f,0.97289f,0.97231f,0.97161f,0.97188f,0.97292f,0.97284f,0.97363f,0.97356f,0.97455f,0.97675f,0.97652f,0.97812f,0.97962f,0.98110f,0.98213f,0.98394f,0.98557f,0.98750f,0.98915f,0.99123f,0.99251f,0.99530f,0.99744f,1.00000f,0.00000f,0.91210f,0.89929f,0.90512f,0.92351f,0.94355f,0.95678f,0.96669f,0.96956f,0.97112f,0.97289f,0.97247f,0.97192f,0.97248f,0.97359f,0.97494f,0.97481f,0.97445f,0.97618f,0.97740f,0.97783f,0.97954f,0.98041f,0.98281f,0.98363f,0.98552f,0.98725f,0.98905f,0.99051f,0.99212f,0.99540f,0.99755f,1.00000f,0.00000f,0.92205f,0.90132f,0.90063f,0.90790f,0.92214f,0.93858f,0.95009f,0.95708f,0.96451f,0.96776f,0.97013f,0.97174f,0.97263f,0.97299f,0.97412f,0.97367f,0.97481f,0.97552f,0.97716f,0.97744f,0.97892f,0.98031f,0.98216f,0.98360f,0.98543f,0.98729f,0.98926f,0.98973f,0.99164f,0.99517f,0.99759f,1.00000f,0.00000f,0.92541f,0.90305f,0.89281f,0.89739f,0.90583f,0.91597f,0.93092f,0.93950f,0.95115f,0.95697f,0.96453f,0.96735f,0.96911f,0.96994f,0.97386f,0.97403f,0.97397f,0.97628f,0.97697f,0.97790f,0.97859f,0.97998f,0.98162f,0.98371f,0.98453f,0.98715f,0.98795f,0.98853f,0.99155f,0.99520f,0.99758f,1.00000f,0.00000f,0.93480f,0.89812f,0.88805f,0.89269f,0.89238f,0.90289f,0.91185f,0.92042f,0.93185f,0.94172f,0.95240f,0.95664f,0.96319f,0.96572f,0.96844f,0.97171f,0.97302f,0.97480f,0.97599f,0.97749f,0.97870f,0.98034f,0.98268f,0.98292f,0.98461f,0.98557f,0.98623f,0.98781f,0.99120f,0.99493f,0.99733f,1.00000f,0.00000f,0.93913f,0.90498f,0.89906f,0.89231f,0.89058f,0.89698f,0.89819f,0.90767f,0.91468f,0.92484f,0.93338f,0.94298f,0.95019f,0.95579f,0.96196f,0.96597f,0.97006f,0.97228f,0.97483f,0.97669f,0.97760f,0.98004f,0.98111f,0.98385f,0.98424f,0.98462f,0.98499f,0.98695f,0.99096f,0.99478f,0.99738f,1.00000f,0.00000f,0.93716f,0.91828f,0.89888f,0.89437f,0.88661f,0.88933f,0.89095f,0.89545f,0.90114f,0.91108f,0.91692f,0.92447f,0.93520f,0.94285f,0.94930f,0.95532f,0.95917f,0.96491f,0.96910f,0.97297f,0.97511f,0.97776f,0.98018f,0.98139f,0.98266f,0.98341f,0.98324f,0.98695f,0.99078f,0.99444f,0.99724f,1.00000f,0.00000f,0.94133f,0.91886f,0.90324f,0.89571f,0.88874f,0.88341f,0.88328f,0.88905f,0.89123f,0.89627f,0.90210f,0.91180f,0.91635f,0.92575f,0.93439f,0.93797f,0.94718f,0.95417f,0.96149f,0.96466f,0.96940f,0.97417f,0.97646f,0.97913f,0.98018f,0.98036f,0.98257f,0.98625f,0.99033f,0.99460f,0.99703f,1.00000f,0.00000f,0.94891f,0.93353f,0.90963f,0.89917f,0.89161f,0.88463f,0.88385f,0.88404f,0.88558f,0.88727f,0.89183f,0.89820f,0.90127f,0.91176f,0.91646f,0.92488f,0.93155f,0.93831f,0.94659f,0.95231f,0.95776f,0.96317f,0.96616f,0.97001f,0.97229f,0.97577f,0.97953f,0.98520f,0.99025f,0.99397f,0.99692f,1.00000f,0.00000f,0.94381f,0.93391f,0.92053f,0.90410f,0.89505f,0.88674f,0.88266f,0.88309f,0.87812f,0.88290f,0.88476f,0.88733f,0.88750f,0.89348f,0.90299f,0.90904f,0.91370f,0.92434f,0.93036f,0.93696f,0.94458f,0.94769f,0.95471f,0.95711f,0.96227f,0.96722f,0.97515f,0.98195f,0.98785f,0.99294f,0.99684f,1.00000f,0.00000f,0.94857f,0.92732f,0.91648f,0.91080f,0.89551f,0.88895f,0.88608f,0.88251f,0.87616f,0.87697f,0.87475f,0.87973f,0.87975f,0.88590f,0.88985f,0.89366f,0.89926f,0.90564f,0.91316f,0.92050f,0.92341f,0.92967f,0.93513f,0.94080f,0.94781f,0.95614f,0.96510f,0.97382f,0.98215f,0.99015f,0.99567f,1.00000f,0.00000f,0.94800f,0.93689f,0.92197f,0.90619f,0.90050f,0.89157f,0.88900f,0.88007f,0.87737f,0.87691f,0.87407f,0.87511f,0.87406f,0.87575f,0.87812f,0.88302f,0.88560f,0.88785f,0.89410f,0.89870f,0.90483f,0.90833f,0.91459f,0.92154f,0.92951f,0.94028f,0.94931f,0.96137f,0.97331f,0.98284f,0.99155f,0.99999f,0.00000f,0.94186f,0.93499f,0.91970f,0.91069f,0.89935f,0.89662f,0.89211f,0.88004f,0.87935f,0.87660f,0.87397f,0.86879f,0.86869f,0.87037f,0.87042f,0.86657f,0.86963f,0.87125f,0.87517f,0.87848f,0.88029f,0.88515f,0.89375f,0.90200f,0.90915f,0.92116f,0.93192f,0.94455f,0.95629f,0.96907f,0.98092f,0.99948f,0.00000f,0.94744f,0.93762f,0.92357f,0.91862f,0.90671f,0.89227f,0.89128f,0.88523f,0.87670f,0.87324f,0.86680f,0.86486f,0.86510f,0.86383f,0.86123f,0.85705f,0.85543f,0.85933f,0.85742f,0.85894f,0.86013f,0.86339f,0.87134f,0.87997f,0.88827f,0.89846f,0.91049f,0.92337f,0.93493f,0.94868f,0.96220f,0.99690f,0.00000f,0.94850f,0.94391f,0.93121f,0.91929f,0.91079f,0.90956f,0.89763f,0.88566f,0.87611f,0.87443f,0.86440f,0.86216f,0.86075f,0.85746f,0.85646f,0.85123f,0.84235f,0.84086f,0.83611f,0.83640f,0.84096f,0.84251f,0.84863f,0.85705f,0.86627f,0.87485f,0.88372f,0.89720f,0.90890f,0.92251f,0.93535f,0.98872f,0.00000f,0.94122f,0.93212f,0.93634f,0.92290f,0.91095f,0.90348f,0.89376f,0.89139f,0.87739f,0.87062f,0.87025f,0.86071f,0.85559f,0.84586f,0.84278f,0.83968f,0.82928f,0.82495f,0.82642f,0.82321f,0.81947f,0.82372f,0.82864f,0.83109f,0.84190f,0.84945f,0.85826f,0.86758f,0.87861f,0.88886f,0.90035f,0.97049f,0.00000f,0.94742f,0.93460f,0.93057f,0.92796f,0.91231f,0.90448f,0.89425f,0.88442f,0.88545f,0.87277f,0.86381f,0.86069f,0.84647f,0.84465f,0.83446f,0.82218f,0.82043f,0.81357f,0.80833f,0.80274f,0.80152f,0.80486f,0.80388f,0.81035f,0.81627f,0.82226f,0.82706f,0.83267f,0.84083f,0.84808f,0.85284f,0.94030f,0.00000f,0.93798f,0.93978f,0.93204f,0.91712f,0.91441f,0.90370f,0.90149f,0.88467f,0.87427f,0.86730f,0.86656f,0.85287f,0.84293f,0.83574f,0.82447f,0.81340f,0.80116f,0.79447f,0.78973f,0.78970f,0.78525f,0.78779f,0.78750f,0.78663f,0.79050f,0.79220f,0.79514f,0.79899f,0.80054f,0.80211f,0.80242f,0.90064f,0.00000f,0.93720f,0.93866f,0.93060f,0.92183f,0.91923f,0.90627f,0.89677f,0.88765f,0.87821f,0.86766f,0.85345f,0.83851f,0.82692f,0.81530f,0.80608f,0.79550f,0.78918f,0.77855f,0.77864f,0.77092f,0.76393f,0.76713f,0.76416f,0.76609f,0.76244f,0.76055f,0.75682f,0.76090f,0.75564f,0.75244f,0.74619f,0.85488f,0.00000f,0.93815f,0.93315f,0.92616f,0.92507f,0.91795f,0.90885f,0.89800f,0.88631f,0.86987f,0.86254f,0.84471f,0.83148f,0.81500f,0.80430f,0.79074f,0.78315f,0.77304f,0.76794f,0.75859f,0.75405f,0.74898f,0.74752f,0.74470f,0.73591f,0.73196f,0.72904f,0.72485f,0.71444f,0.70560f,0.69862f,0.68922f,0.79940f,0.00000f,0.93247f,0.93578f,0.92511f,0.92325f,0.91004f,0.90351f,0.89123f,0.87666f,0.86646f,0.85134f,0.83555f,0.82106f,0.80700f,0.78738f,0.77610f,0.76565f,0.75406f,0.74661f,0.74295f,0.73824f,0.73400f,0.72686f,0.71968f,0.71531f,0.70510f,0.69691f,0.68167f,0.67060f,0.65699f,0.64383f,0.62796f,0.74246f,0.00000f,0.92414f,0.93547f,0.92208f,0.91213f,0.91233f,0.89108f,0.88130f,0.86341f,0.84806f,0.84537f,0.81738f,0.80750f,0.78323f,0.77337f,0.75897f,0.75145f,0.74460f,0.73418f,0.72674f,0.72012f,0.71047f,0.70561f,0.69233f,0.68486f,0.67217f,0.65713f,0.64434f,0.62202f,0.60675f,0.59096f,0.57033f,0.68441f,0.00000f,0.92391f,0.92757f,0.91137f,0.90242f,0.90118f,0.89310f,0.87902f,0.85739f,0.84127f,0.81966f,0.80217f,0.78342f,0.77009f,0.75167f,0.74499f,0.73140f,0.72320f,0.71292f,0.70482f,0.70036f,0.68780f,0.68022f,0.66713f,0.65340f,0.63867f,0.61804f,0.60096f,0.57902f,0.55883f,0.53752f,0.51875f,0.62630f,0.00000f,0.92467f,0.91095f,0.91003f,0.90373f,0.88616f,0.87878f,0.86349f,0.84886f,0.81798f,0.80298f,0.78169f,0.76898f,0.74596f,0.73201f,0.72032f,0.71566f,0.70448f,0.69034f,0.68566f,0.67182f,0.66779f,0.65119f,0.63923f,0.61961f,0.60211f,0.58179f,0.55560f,0.53427f,0.51245f,0.48951f,0.46648f,0.56839f,0.00000f,0.90653f,0.91637f,0.90662f,0.88764f,0.88174f,0.86541f,0.84885f,0.82419f,0.80027f,0.79036f,0.75859f,0.74781f,0.73028f,0.71596f,0.70068f,0.69452f,0.68487f,0.67557f,0.65975f,0.65043f,0.63879f,0.62332f,0.60805f,0.58653f,0.56193f,0.54235f,0.51793f,0.48915f,0.46426f,0.44373f,0.42114f,0.51634f,0.00000f,0.91373f,0.91028f,0.89368f,0.87459f,0.86368f,0.84771f,0.81787f,0.80462f,0.78038f,0.76669f,0.73668f,0.71826f,0.70249f,0.68840f,0.67865f,0.67054f,0.65976f,0.64756f,0.63685f,0.62615f,0.60859f,0.59434f,0.57011f,0.54845f,0.52492f,0.50172f,0.47269f,0.44927f,0.42249f,0.39856f,0.37453f,0.46709f,
 0.00000f,1.00000f,0.99912f,1.00000f,0.99117f,0.96698f,0.99038f,0.99489f,0.98399f,0.96672f,0.96443f,0.99880f,0.98239f,0.98851f,0.98555f,0.98484f,0.98229f,0.97275f,0.97103f,0.97632f,0.97346f,0.97230f,0.99152f,0.98755f,0.98577f,0.98691f,0.99146f,0.99083f,0.99321f,0.99503f,0.99384f,0.99624f,1.00000f,0.00000f,0.99165f,0.98722f,0.98399f,0.97953f,0.97696f,0.97684f,0.97616f,0.97554f,0.97457f,0.97324f,0.97525f,0.97426f,0.97358f,0.97436f,0.97531f,0.97570f,0.97467f,0.97688f,0.97839f,0.97855f,0.97944f,0.98212f,0.98298f,0.98425f,0.98568f,0.98774f,0.98962f,0.99146f,0.99349f,0.99524f,0.99751f,1.00000f,0.00000f,0.99049f,0.98611f,0.98326f,0.97948f,0.97842f,0.97664f,0.97508f,0.97471f,0.97464f,0.97425f,0.97313f,0.97370f,0.97301f,0.97391f,0.97504f,0.97494f,0.97612f,0.97679f,0.97801f,0.97878f,0.98017f,0.98121f,0.98283f,0.98469f,0.98592f,0.98784f,0.98942f,0.99130f,0.99343f,0.99545f,0.99776f,1.00000f,0.00000f,0.98392f,0.98737f,0.98192f,0.97938f,0.97703f,0.97619f,0.97496f,0.97502f,0.97363f,0.97439f,0.97328f,0.97335f,0.97376f,0.97373f,0.97454f,0.97494f,0.97619f,0.97652f,0.97778f,0.97864f,0.97987f,0.98162f,0.98287f,0.98428f,0.98611f,0.98761f,0.98930f,0.99127f,0.99339f,0.99549f,0.99780f,1.00000f,0.00000f,0.94141f,0.98068f,0.98041f,0.97929f,0.97863f,0.97653f,0.97504f,0.97343f,0.97421f,0.97284f,0.97288f,0.97211f,0.97335f,0.97396f,0.97467f,0.97530f,0.97541f,0.97666f,0.97726f,0.97857f,0.98012f,0.98134f,0.98280f,0.98428f,0.98602f,0.98745f,0.98924f,0.99133f,0.99332f,0.99546f,0.99778f,1.00000f,0.00000f,0.90731f,0.95634f,0.97627f,0.97971f,0.97655f,0.97613f,0.97474f,0.97421f,0.97286f,0.97261f,0.97270f,0.97295f,0.97361f,0.97399f,0.97431f,0.97524f,0.97541f,0.97624f,0.97723f,0.97811f,0.97986f,0.98107f,0.98182f,0.98415f,0.98554f,0.98749f,0.98921f,0.99116f,0.99328f,0.99541f,0.99771f,1.00000f,0.00000f,0.89459f,0.92368f,0.95561f,0.97065f,0.97487f,0.97582f,0.97410f,0.97267f,0.97306f,0.97327f,0.97210f,0.97156f,0.97353f,0.97362f,0.97473f,0.97493f,0.97533f,0.97660f,0.97705f,0.97837f,0.97984f,0.98112f,0.98219f,0.98386f,0.98615f,0.98738f,0.98924f,0.99137f,0.99300f,0.99543f,0.99760f,1.00000f,0.00000f,0.90211f,0.90399f,0.92409f,0.95004f,0.96384f,0.97095f,0.97203f,0.97347f,0.97153f,0.97156f,0.97295f,0.97245f,0.97298f,0.97276f,0.97321f,0.97469f,0.97493f,0.97611f,0.97752f,0.97824f,0.97915f,0.98056f,0.98192f,0.98361f,0.98588f,0.98726f,0.98930f,0.99112f,0.99238f,0.99533f,0.99749f,1.00000f,0.00000f,0.90926f,0.89626f,0.90773f,0.92562f,0.94317f,0.95760f,0.96522f,0.96929f,0.97071f,0.97180f,0.97244f,0.97216f,0.97314f,0.97312f,0.97406f,0.97444f,0.97559f,0.97550f,0.97715f,0.97822f,0.97941f,0.98112f,0.98180f,0.98402f,0.98574f,0.98748f,0.98936f,0.99070f,0.99212f,0.99522f,0.99766f,1.00000f,0.00000f,0.91925f,0.89877f,0.89604f,0.90883f,0.92099f,0.93752f,0.94930f,0.95846f,0.96405f,0.96691f,0.97186f,0.97019f,0.97147f,0.97217f,0.97374f,0.97504f,0.97438f,0.97578f,0.97594f,0.97772f,0.97939f,0.98139f,0.98258f,0.98353f,0.98501f,0.98722f,0.98914f,0.98999f,0.99169f,0.99518f,0.99759f,1.00000f,0.00000f,0.92208f,0.89999f,0.89083f,0.89600f,0.90237f,0.91670f,0.93230f,0.94286f,0.94879f,0.95845f,0.96323f,0.96778f,0.96984f,0.97198f,0.97306f,0.97381f,0.97514f,0.97547f,0.97724f,0.97800f,0.97866f,0.97978f,0.98160f,0.98385f,0.98548f,0.98675f,0.98830f,0.98860f,0.99114f,0.99503f,0.99739f,1.00000f,0.00000f,0.92833f,0.90719f,0.89544f,0.89077f,0.89730f,0.90003f,0.91192f,0.92112f,0.93245f,0.94166f,0.95042f,0.95479f,0.96308f,0.96537f,0.96861f,0.97117f,0.97383f,0.97608f,0.97600f,0.97755f,0.97947f,0.97970f,0.98082f,0.98345f,0.98595f,0.98605f,0.98675f,0.98749f,0.99065f,0.99504f,0.99749f,1.00000f,0.00000f,0.93605f,0.90844f,0.89579f,0.88950f,0.89010f,0.89455f,0.89982f,0.90448f,0.91830f,0.92703f,0.93443f,0.94204f,0.94961f,0.95464f,0.96032f,0.96580f,0.96961f,0.97049f,0.97329f,0.97709f,0.97851f,0.97940f,0.98086f,0.98291f,0.98461f,0.98481f,0.98530f,0.98689f,0.99073f,0.99459f,0.99734f,1.00000f,0.00000f,0.93979f,0.91275f,0.89874f,0.89181f,0.88725f,0.88902f,0.88983f,0.89804f,0.90086f,0.90824f,0.91969f,0.92542f,0.93305f,0.94173f,0.94767f,0.95609f,0.96155f,0.96535f,0.96892f,0.97233f,0.97580f,0.97738f,0.98112f,0.98291f,0.98216f,0.98345f,0.98393f,0.98600f,0.99031f,0.99430f,0.99713f,1.00000f,0.00000f,0.94452f,0.91594f,0.90800f,0.89418f,0.88976f,0.88596f,0.88695f,0.89186f,0.89157f,0.89794f,0.90327f,0.90953f,0.91788f,0.92702f,0.93220f,0.93975f,0.94792f,0.95380f,0.96058f,0.96450f,0.96878f,0.97417f,0.97529f,0.97851f,0.97976f,0.98000f,0.98254f,0.98598f,0.99056f,0.99450f,0.99715f,1.00000f,0.00000f,0.94173f,0.92748f,0.90989f,0.90469f,0.88929f,0.88178f,0.88712f,0.88361f,0.88732f,0.88821f,0.89288f,0.89331f,0.90287f,0.90806f,0.91946f,0.92573f,0.93175f,0.93826f,0.94533f,0.95162f,0.95771f,0.96345f,0.96727f,0.97017f,0.97375f,0.97595f,0.97910f,0.98455f,0.99032f,0.99413f,0.99699f,1.00000f,0.00000f,0.94009f,0.93049f,0.92020f,0.90482f,0.89676f,0.88643f,0.88606f,0.88093f,0.88192f,0.88105f,0.88300f,0.88636f,0.88859f,0.89512f,0.90334f,0.90721f,0.91559f,0.92260f,0.92992f,0.93715f,0.94277f,0.94831f,0.95414f,0.95875f,0.96267f,0.96808f,0.97494f,0.98131f,0.98832f,0.99317f,0.99713f,1.00000f,0.00000f,0.93972f,0.93497f,0.92189f,0.90515f,0.89990f,0.89277f,0.88452f,0.87945f,0.87821f,0.87799f,0.87550f,0.88119f,0.87958f,0.88447f,0.88611f,0.89588f,0.90204f,0.90586f,0.91272f,0.91794f,0.92507f,0.93073f,0.93526f,0.94117f,0.94807f,0.95590f,0.96450f,0.97366f,0.98260f,0.99029f,0.99574f,1.00000f,0.00000f,0.94425f,0.94003f,0.91997f,0.91701f,0.90895f,0.88656f,0.88928f,0.87760f,0.87315f,0.87749f,0.87149f,0.87649f,0.87508f,0.87908f,0.87625f,0.88269f,0.88867f,0.88970f,0.89400f,0.90047f,0.90443f,0.90736f,0.91645f,0.92329f,0.93136f,0.94062f,0.94950f,0.96159f,0.97344f,0.98109f,0.99162f,0.99999f,0.00000f,0.95466f,0.93325f,0.92360f,0.91417f,0.90588f,0.89459f,0.88884f,0.88443f,0.87709f,0.87595f,0.86909f,0.87018f,0.86909f,0.87002f,0.86898f,0.87067f,0.87328f,0.87219f,0.87503f,0.87957f,0.88382f,0.88892f,0.89421f,0.90155f,0.91024f,0.92051f,0.93227f,0.94388f,0.95467f,0.96765f,0.98048f,0.99949f,0.00000f,0.94423f,0.93268f,0.93018f,0.92257f,0.90638f,0.89879f,0.89445f,0.88703f,0.87934f,0.87202f,0.86822f,0.86838f,0.86408f,0.86254f,0.86161f,0.86065f,0.85959f,0.85881f,0.85756f,0.85884f,0.85997f,0.86490f,0.87362f,0.88037f,0.88949f,0.89730f,0.90783f,0.92352f,0.93629f,0.94814f,0.96299f,0.99695f,0.00000f,0.94753f,0.94046f,0.92397f,0.92003f,0.90701f,0.89873f,0.89555f,0.88748f,0.87704f,0.87134f,0.87173f,0.86442f,0.86031f,0.85728f,0.85438f,0.85169f,0.84857f,0.84133f,0.83902f,0.84064f,0.83976f,0.84440f,0.84668f,0.85564f,0.86652f,0.87553f,0.88418f,0.89774f,0.91184f,0.92049f,0.93403f,0.98881f,0.00000f,0.94396f,0.93848f,0.93567f,0.92538f,0.91064f,0.90087f,0.90173f,0.88587f,0.87918f,0.87267f,0.87035f,0.86409f,0.85585f,0.85059f,0.84934f,0.84018f,0.83248f,0.82615f,0.82596f,0.82403f,0.82408f,0.82272f,0.82667f,0.83181f,0.83929f,0.85044f,0.85786f,0.86648f,0.87687f,0.88697f,0.89922f,0.97010f,0.00000f,0.95013f,0.93858f,0.93025f,0.91921f,0.91607f,0.90515f,0.89695f,0.88819f,0.87601f,0.86808f,0.86517f,0.85802f,0.85153f,0.84150f,0.83411f,0.82676f,0.82039f,0.81179f,0.80856f,0.80769f,0.80602f,0.80346f,0.80975f,0.80793f,0.81319f,0.82370f,0.82610f,0.83468f,0.84004f,0.84717f,0.85410f,0.94058f,0.00000f,0.94045f,0.93288f,0.92850f,0.92342f,0.91894f,0.90950f,0.90437f,0.88745f,0.88251f,0.88237f,0.86376f,0.84991f,0.84426f,0.82982f,0.82335f,0.81347f,0.80441f,0.79773f,0.79411f,0.78986f,0.78659f,0.78767f,0.78681f,0.78939f,0.79099f,0.79091f,0.79599f,0.79687f,0.80024f,0.80238f,0.80286f,0.90090f,0.00000f,0.93574f,0.94110f,0.93425f,0.92334f,0.90986f,0.90263f,0.89404f,0.88538f,0.87092f,0.87051f,0.86122f,0.84263f,0.83560f,0.81921f,0.80940f,0.79966f,0.79266f,0.78422f,0.78052f,0.77411f,0.77005f,0.76706f,0.76402f,0.76649f,0.76240f,0.75947f,0.76177f,0.75758f,0.75526f,0.75009f,0.74705f,0.85493f,0.00000f,0.93439f,0.93372f,0.92851f,0.92556f,0.91596f,0.90547f,0.89357f,0.88481f,0.87581f,0.86085f,0.84826f,0.83448f,0.82211f,0.80367f,0.79786f,0.78432f,0.77271f,0.76889f,0.76089f,0.75691f,0.75388f,0.74531f,0.74493f,0.73785f,0.73253f,0.72938f,0.72492f,0.71629f,0.70867f,0.69909f,0.68887f,0.80000f,0.00000f,0.93326f,0.93216f,0.93191f,0.91400f,0.90262f,0.90168f,0.89004f,0.87504f,0.86320f,0.85341f,0.83622f,0.81888f,0.80368f,0.78988f,0.78071f,0.77030f,0.75807f,0.75254f,0.74364f,0.73597f,0.73240f,0.72103f,0.72378f,0.71640f,0.70667f,0.69297f,0.68195f,0.67074f,0.65633f,0.64599f,0.62869f,0.74107f,0.00000f,0.93022f,0.92865f,0.92195f,0.91543f,0.90550f,0.89467f,0.88154f,0.87383f,0.85561f,0.84484f,0.82961f,0.80141f,0.79284f,0.77963f,0.76300f,0.74993f,0.74233f,0.73189f,0.72464f,0.72313f,0.71300f,0.70426f,0.69608f,0.68791f,0.66934f,0.65760f,0.64167f,0.62308f,0.60849f,0.58969f,0.57491f,0.68536f,0.00000f,0.92980f,0.91970f,0.92424f,0.91287f,0.89751f,0.88993f,0.87935f,0.85787f,0.84450f,0.82236f,0.80877f,0.79187f,0.77370f,0.75873f,0.74394f,0.73344f,0.72574f,0.71770f,0.70691f,0.69684f,0.68839f,0.67872f,0.67040f,0.65476f,0.63734f,0.62066f,0.59653f,0.57883f,0.55729f,0.53977f,0.51946f,0.62447f,0.00000f,0.92255f,0.91387f,0.91000f,0.90516f,0.88794f,0.88570f,0.86389f,0.84610f,0.83374f,0.80589f,0.79126f,0.76806f,0.75415f,0.73773f,0.72571f,0.70752f,0.70675f,0.69634f,0.68397f,0.67301f,0.66609f,0.65287f,0.63661f,0.61782f,0.59835f,0.58246f,0.55490f,0.53358f,0.50770f,0.48635f,0.46650f,0.57226f,0.00000f,0.91304f,0.91168f,0.90729f,0.88983f,0.87717f,0.87126f,0.85166f,0.82877f,0.81372f,0.78623f,0.76744f,0.74366f,0.73176f,0.71987f,0.70619f,0.69084f,0.68618f,0.67319f,0.66052f,0.64966f,0.64048f,0.62389f,0.60480f,0.58675f,0.56205f,0.54054f,0.51377f,0.48918f,0.46651f,0.43940f,0.42076f,0.51509f,0.00000f,0.91173f,0.91110f,0.89963f,0.88652f,0.87281f,0.84994f,0.83338f,0.80377f,0.78715f,0.76384f,0.74630f,0.72910f,0.71709f,0.69518f,0.67937f,0.66895f,0.65796f,0.65065f,0.63731f,0.62785f,0.61284f,0.59466f,0.57285f,0.55040f,0.52410f,0.49860f,0.47360f,0.44800f,0.42589f,0.40056f,0.37498f,0.46626f,
 0.00000f,1.00000f,0.99781f,0.99962f,0.99066f,0.96895f,0.99028f,0.99358f,0.98359f,0.96603f,0.96441f,0.99874f,0.98157f,0.98925f,0.98568f,0.98418f,0.98318f,0.97255f,0.97090f,0.97617f,0.97383f,0.97183f,0.99132f,0.98709f,0.98574f,0.98687f,0.99125f,0.99098f,0.99311f,0.99503f,0.99508f,0.99625f,1.00000f,0.00000f,0.99226f,0.98653f,0.98217f,0.97950f,0.97824f,0.97764f,0.97563f,0.97515f,0.97259f,0.97409f,0.97540f,0.97369f,0.97471f,0.97365f,0.97586f,0.97553f,0.97526f,0.97613f,0.97821f,0.97868f,0.97908f,0.98198f,0.98281f,0.98450f,0.98632f,0.98769f,0.98959f,0.99166f,0.99355f,0.99530f,0.99749f,1.00000f,0.00000f,0.98989f,0.98666f,0.98242f,0.97978f,0.97683f,0.97729f,0.97511f,0.97491f,0.97466f,0.97395f,0.97383f,0.97324f,0.97258f,0.97346f,0.97485f,0.97487f,0.97625f,0.97674f,0.97801f,0.97869f,0.97986f,0.98149f,0.98276f,0.98459f,0.98612f,0.98764f,0.98959f,0.99160f,0.99342f,0.99545f,0.99772f,1.00000f,0.00000f,0.98211f,0.98542f,0.98235f,0.97942f,0.97799f,0.97576f,0.97464f,0.97377f,0.97470f,0.97342f,0.97367f,0.97294f,0.97362f,0.97435f,0.97481f,0.97496f,0.97571f,0.97739f,0.97737f,0.97881f,0.97942f,0.98106f,0.98302f,0.98431f,0.98585f,0.98778f,0.98963f,0.99147f,0.99335f,0.99545f,0.99772f,1.00000f,0.00000f,0.93827f,0.98268f,0.98186f,0.97860f,0.97839f,0.97628f,0.97554f,0.97508f,0.97263f,0.97317f,0.97318f,0.97338f,0.97428f,0.97431f,0.97469f,0.97505f,0.97564f,0.97637f,0.97781f,0.97808f,0.97996f,0.98093f,0.98259f,0.98420f,0.98581f,0.98749f,0.98927f,0.99146f,0.99316f,0.99541f,0.99776f,1.00000f,0.00000f,0.90710f,0.95627f,0.97709f,0.97752f,0.97728f,0.97518f,0.97472f,0.97402f,0.97292f,0.97375f,0.97265f,0.97333f,0.97386f,0.97388f,0.97411f,0.97456f,0.97577f,0.97620f,0.97760f,0.97857f,0.97982f,0.98085f,0.98260f,0.98406f,0.98593f,0.98754f,0.98921f,0.99112f,0.99325f,0.99538f,0.99769f,1.00000f,0.00000f,0.90231f,0.92366f,0.95642f,0.97087f,0.97714f,0.97560f,0.97517f,0.97344f,0.97408f,0.97174f,0.97320f,0.97251f,0.97234f,0.97356f,0.97425f,0.97407f,0.97507f,0.97609f,0.97705f,0.97874f,0.97963f,0.98086f,0.98206f,0.98400f,0.98550f,0.98757f,0.98935f,0.99116f,0.99324f,0.99542f,0.99757f,1.00000f,0.00000f,0.90326f,0.90354f,0.92628f,0.94996f,0.96390f,0.96941f,0.97219f,0.97275f,0.97328f,0.97223f,0.97298f,0.97221f,0.97273f,0.97414f,0.97408f,0.97496f,0.97546f,0.97605f,0.97751f,0.97845f,0.97932f,0.98143f,0.98257f,0.98395f,0.98576f,0.98720f,0.98924f,0.99114f,0.99267f,0.99541f,0.99756f,1.00000f,0.00000f,0.90723f,0.89934f,0.90545f,0.92300f,0.94526f,0.95871f,0.96587f,0.97029f,0.97150f,0.97295f,0.97059f,0.97156f,0.97177f,0.97339f,0.97310f,0.97479f,0.97447f,0.97645f,0.97681f,0.97779f,0.97874f,0.98099f,0.98229f,0.98340f,0.98574f,0.98721f,0.98915f,0.99097f,0.99164f,0.99528f,0.99766f,1.00000f,0.00000f,0.91928f,0.89228f,0.89750f,0.90656f,0.92253f,0.93707f,0.94876f,0.95893f,0.96374f,0.96689f,0.97013f,0.97025f,0.97243f,0.97133f,0.97308f,0.97390f,0.97475f,0.97553f,0.97695f,0.97826f,0.97878f,0.98052f,0.98221f,0.98381f,0.98547f,0.98728f,0.98915f,0.99034f,0.99171f,0.99512f,0.99755f,1.00000f,0.00000f,0.92406f,0.90061f,0.89104f,0.89683f,0.90357f,0.91819f,0.93028f,0.94119f,0.94973f,0.95749f,0.96196f,0.96702f,0.97106f,0.97136f,0.97247f,0.97359f,0.97460f,0.97582f,0.97672f,0.97797f,0.97933f,0.98051f,0.98144f,0.98323f,0.98506f,0.98699f,0.98821f,0.98934f,0.99123f,0.99488f,0.99733f,1.00000f,0.00000f,0.93225f,0.90269f,0.88963f,0.89258f,0.89740f,0.90217f,0.91113f,0.92217f,0.93289f,0.94019f,0.95009f,0.95734f,0.96114f,0.96545f,0.96773f,0.97151f,0.97431f,0.97518f,0.97629f,0.97742f,0.97948f,0.98038f,0.98205f,0.98341f,0.98551f,0.98653f,0.98719f,0.98811f,0.99070f,0.99478f,0.99741f,1.00000f,0.00000f,0.94046f,0.90805f,0.89750f,0.88910f,0.89036f,0.89345f,0.89697f,0.90774f,0.91438f,0.92484f,0.93397f,0.94153f,0.94934f,0.95560f,0.95988f,0.96589f,0.96941f,0.97206f,0.97396f,0.97662f,0.97819f,0.97995f,0.98110f,0.98215f,0.98427f,0.98562f,0.98515f,0.98652f,0.99086f,0.99455f,0.99724f,1.00000f,0.00000f,0.94329f,0.91626f,0.90026f,0.89123f,0.88865f,0.88501f,0.89166f,0.89766f,0.89874f,0.90890f,0.91755f,0.92425f,0.93317f,0.94209f,0.95015f,0.95532f,0.95995f,0.96526f,0.96867f,0.97315f,0.97553f,0.97786f,0.98031f,0.98200f,0.98214f,0.98385f,0.98353f,0.98598f,0.99059f,0.99444f,0.99723f,1.00000f,0.00000f,0.94911f,0.92391f,0.90899f,0.89520f,0.88137f,0.88597f,0.88488f,0.88982f,0.89224f,0.89341f,0.90179f,0.91103f,0.91614f,0.92329f,0.93290f,0.94089f,0.94927f,0.95402f,0.96075f,0.96292f,0.97004f,0.97330f,0.97795f,0.97846f,0.97970f,0.98073f,0.98244f,0.98532f,0.98976f,0.99414f,0.99694f,1.00000f,0.00000f,0.94911f,0.92829f,0.90893f,0.89935f,0.89019f,0.88767f,0.88353f,0.88412f,0.88831f,0.88948f,0.89226f,0.89791f,0.90282f,0.90769f,0.91938f,0.92414f,0.93255f,0.93867f,0.94755f,0.95238f,0.96100f,0.96434f,0.96816f,0.97106f,0.97410f,0.97581f,0.98037f,0.98440f,0.98936f,0.99390f,0.99694f,1.00000f,0.00000f,0.95161f,0.93706f,0.91764f,0.90568f,0.89701f,0.88966f,0.88108f,0.88039f,0.88055f,0.88305f,0.88211f,0.88671f,0.88766f,0.89788f,0.90328f,0.90874f,0.91685f,0.92452f,0.93225f,0.93811f,0.94277f,0.94758f,0.95490f,0.95942f,0.96389f,0.96884f,0.97448f,0.98093f,0.98812f,0.99265f,0.99675f,1.00000f,0.00000f,0.93905f,0.93405f,0.92533f,0.91031f,0.89979f,0.88807f,0.88682f,0.88175f,0.88349f,0.87586f,0.87811f,0.87898f,0.88210f,0.88185f,0.88747f,0.89661f,0.90189f,0.90683f,0.91474f,0.92110f,0.92409f,0.92972f,0.93688f,0.93980f,0.94960f,0.95761f,0.96498f,0.97416f,0.98193f,0.98962f,0.99557f,1.00000f,0.00000f,0.94617f,0.93821f,0.92551f,0.90757f,0.90025f,0.89196f,0.89024f,0.87814f,0.87865f,0.87198f,0.87141f,0.87446f,0.87523f,0.87784f,0.87812f,0.88300f,0.88585f,0.88942f,0.89401f,0.89919f,0.90512f,0.91007f,0.91573f,0.92356f,0.93126f,0.94026f,0.95194f,0.96295f,0.97190f,0.98244f,0.99115f,0.99999f,0.00000f,0.94613f,0.93368f,0.92495f,0.91648f,0.90608f,0.89264f,0.88856f,0.87995f,0.88136f,0.87306f,0.87014f,0.87247f,0.86769f,0.86608f,0.87184f,0.87246f,0.87063f,0.87112f,0.87664f,0.87838f,0.88010f,0.88822f,0.89279f,0.90118f,0.91077f,0.92015f,0.93292f,0.94396f,0.95664f,0.96843f,0.98074f,0.99948f,0.00000f,0.94284f,0.93144f,0.93129f,0.91172f,0.90960f,0.90116f,0.89483f,0.88570f,0.88105f,0.87576f,0.86929f,0.86607f,0.86590f,0.86474f,0.86193f,0.86151f,0.85956f,0.85524f,0.85808f,0.85823f,0.86291f,0.86629f,0.87210f,0.87691f,0.88794f,0.89730f,0.91161f,0.92355f,0.93509f,0.94862f,0.96317f,0.99689f,0.00000f,0.94529f,0.93936f,0.92274f,0.92437f,0.90711f,0.90091f,0.89658f,0.88824f,0.88237f,0.87503f,0.86717f,0.86499f,0.85899f,0.85576f,0.85494f,0.84975f,0.84920f,0.84351f,0.84149f,0.84144f,0.84202f,0.84453f,0.84698f,0.85704f,0.86589f,0.87443f,0.88530f,0.89876f,0.90869f,0.92206f,0.93592f,0.98861f,0.00000f,0.94128f,0.93986f,0.93452f,0.91948f,0.91577f,0.90108f,0.89410f,0.88986f,0.87932f,0.87272f,0.86672f,0.85940f,0.85551f,0.85552f,0.84709f,0.84510f,0.83323f,0.83203f,0.82165f,0.82261f,0.82593f,0.82478f,0.82808f,0.83388f,0.83946f,0.85096f,0.85857f,0.86578f,0.87892f,0.88900f,0.89973f,0.97056f,0.00000f,0.94135f,0.93602f,0.93086f,0.91661f,0.91782f,0.90723f,0.90007f,0.88731f,0.88028f,0.86497f,0.86241f,0.85579f,0.85710f,0.84271f,0.83172f,0.82798f,0.82415f,0.81388f,0.80866f,0.80231f,0.80207f,0.80526f,0.80514f,0.81030f,0.81506f,0.82118f,0.82683f,0.83626f,0.84219f,0.84568f,0.85672f,0.94091f,0.00000f,0.94449f,0.93699f,0.93131f,0.92191f,0.91579f,0.90659f,0.89360f,0.88335f,0.88019f,0.86871f,0.86163f,0.85613f,0.84734f,0.83447f,0.82796f,0.81231f,0.80885f,0.79999f,0.79228f,0.79126f,0.78617f,0.79030f,0.78371f,0.79119f,0.79106f,0.79135f,0.79468f,0.79682f,0.80116f,0.80070f,0.80138f,0.90197f,0.00000f,0.93694f,0.93044f,0.93683f,0.92318f,0.91319f,0.90479f,0.89811f,0.88917f,0.87718f,0.86920f,0.85673f,0.84284f,0.83367f,0.82494f,0.80861f,0.79642f,0.79062f,0.78588f,0.77645f,0.77431f,0.76967f,0.76821f,0.76216f,0.76786f,0.75908f,0.76089f,0.75986f,0.75592f,0.75440f,0.74910f,0.74668f,0.85365f,0.00000f,0.93618f,0.93063f,0.92778f,0.91693f,0.90879f,0.90476f,0.89665f,0.88798f,0.87357f,0.86229f,0.85022f,0.83127f,0.82102f,0.81124f,0.79899f,0.78445f,0.77603f,0.77366f,0.76399f,0.75866f,0.75321f,0.74638f,0.74572f,0.73811f,0.73607f,0.73046f,0.72157f,0.71453f,0.70615f,0.69795f,0.68699f,0.79980f,0.00000f,0.93054f,0.93294f,0.92403f,0.91983f,0.91354f,0.90475f,0.89051f,0.87709f,0.86833f,0.85722f,0.84005f,0.82410f,0.80672f,0.79380f,0.78090f,0.76937f,0.76276f,0.75377f,0.74730f,0.74041f,0.73695f,0.73343f,0.72404f,0.71210f,0.70232f,0.69545f,0.68175f,0.66931f,0.65680f,0.64582f,0.62868f,0.74133f,0.00000f,0.92650f,0.93109f,0.92758f,0.91405f,0.91130f,0.90089f,0.88751f,0.87008f,0.86426f,0.84351f,0.83057f,0.80824f,0.79477f,0.78154f,0.76456f,0.75137f,0.74631f,0.73270f,0.73194f,0.71773f,0.71221f,0.70427f,0.69304f,0.68169f,0.67313f,0.65565f,0.63964f,0.62401f,0.60872f,0.58896f,0.57110f,0.68498f,0.00000f,0.92773f,0.91902f,0.91680f,0.91401f,0.89984f,0.89003f,0.87610f,0.86001f,0.84891f,0.83065f,0.80962f,0.79024f,0.77662f,0.76436f,0.74705f,0.73455f,0.72514f,0.71726f,0.71107f,0.69957f,0.69084f,0.67943f,0.66901f,0.65370f,0.63858f,0.61751f,0.59529f,0.57874f,0.55886f,0.53758f,0.51846f,0.62440f,0.00000f,0.92385f,0.91326f,0.90884f,0.90032f,0.89392f,0.88496f,0.86379f,0.84764f,0.83099f,0.81595f,0.78656f,0.77160f,0.75374f,0.74513f,0.72523f,0.71943f,0.70906f,0.70061f,0.68898f,0.67917f,0.66784f,0.65222f,0.63762f,0.61705f,0.59846f,0.57633f,0.55417f,0.53244f,0.51011f,0.49013f,0.46483f,0.56691f,0.00000f,0.92333f,0.90800f,0.90873f,0.89383f,0.88486f,0.87804f,0.84942f,0.83320f,0.81239f,0.79070f,0.77614f,0.75606f,0.73409f,0.72220f,0.70385f,0.69453f,0.68435f,0.67576f,0.66622f,0.65148f,0.63879f,0.62443f,0.60240f,0.58139f,0.56281f,0.53860f,0.51239f,0.48924f,0.46410f,0.44135f,0.42107f,0.51480f,0.00000f,0.90618f,0.89814f,0.89184f,0.88619f,0.86798f,0.85485f,0.83470f,0.81333f,0.79448f,0.76471f,0.74932f,0.73104f,0.71487f,0.69762f,0.68504f,0.67417f,0.66227f,0.65105f,0.63891f,0.62580f,0.61096f,0.58772f,0.57074f,0.54877f,0.52564f,0.49605f,0.47274f,0.44776f,0.42444f,0.39861f,0.37631f,0.46500f,
 0.00000f,1.00000f,0.99714f,1.00000f,0.99151f,0.96860f,0.99066f,0.99360f,0.98374f,0.96638f,0.96531f,0.99903f,0.98157f,0.98825f,0.98584f,0.98470f,0.98228f,0.97190f,0.97115f,0.97685f,0.97283f,0.97222f,0.99136f,0.98741f,0.98594f,0.98691f,0.99151f,0.99095f,0.99334f,0.99521f,0.99631f,0.99639f,1.00000f,0.00000f,0.99249f,0.98597f,0.98336f,0.98177f,0.97694f,0.97661f,0.97613f,0.97419f,0.97334f,0.97281f,0.97394f,0.97307f,0.97469f,0.97499f,0.97541f,0.97550f,0.97554f,0.97618f,0.97836f,0.97821f,0.97943f,0.98208f,0.98279f,0.98467f,0.98581f,0.98786f,0.98945f,0.99152f,0.99341f,0.99543f,0.99748f,1.00000f,0.00000f,0.99056f,0.98566f,0.98251f,0.98041f,0.97961f,0.97590f,0.97642f,0.97338f,0.97381f,0.97319f,0.97295f,0.97303f,0.97396f,0.97382f,0.97471f,0.97419f,0.97662f,0.97607f,0.97773f,0.97862f,0.97968f,0.98134f,0.98278f,0.98458f,0.98609f,0.98789f,0.98950f,0.99140f,0.99322f,0.99535f,0.99777f,1.00000f,0.00000f,0.98509f,0.98456f,0.98202f,0.97939f,0.97777f,0.97562f,0.97584f,0.97537f,0.97401f,0.97396f,0.97375f,0.97316f,0.97385f,0.97325f,0.97404f,0.97478f,0.97597f,0.97696f,0.97792f,0.97915f,0.98022f,0.98170f,0.98275f,0.98434f,0.98594f,0.98744f,0.98957f,0.99142f,0.99335f,0.99536f,0.99781f,1.00000f,0.00000f,0.93924f,0.98111f,0.98103f,0.97993f,0.97953f,0.97526f,0.97495f,0.97390f,0.97327f,0.97380f,0.97340f,0.97304f,0.97292f,0.97379f,0.97393f,0.97469f,0.97561f,0.97668f,0.97685f,0.97862f,0.97956f,0.98092f,0.98251f,0.98421f,0.98575f,0.98769f,0.98921f,0.99125f,0.99336f,0.99537f,0.99780f,1.00000f,0.00000f,0.90832f,0.95351f,0.97607f,0.97928f,0.97677f,0.97624f,0.97515f,0.97478f,0.97385f,0.97275f,0.97339f,0.97242f,0.97382f,0.97325f,0.97442f,0.97534f,0.97517f,0.97643f,0.97754f,0.97875f,0.97998f,0.98122f,0.98283f,0.98424f,0.98552f,0.98758f,0.98931f,0.99132f,0.99319f,0.99541f,0.99774f,1.00000f,0.00000f,0.90051f,0.92108f,0.95424f,0.97034f,0.97500f,0.97452f,0.97349f,0.97270f,0.97320f,0.97197f,0.97295f,0.97218f,0.97216f,0.97365f,0.97490f,0.97435f,0.97551f,0.97654f,0.97759f,0.97851f,0.98003f,0.98098f,0.98251f,0.98420f,0.98555f,0.98741f,0.98938f,0.99136f,0.99335f,0.99525f,0.99766f,1.00000f,0.00000f,0.89829f,0.90056f,0.92609f,0.95157f,0.96364f,0.97188f,0.97385f,0.97287f,0.97270f,0.97217f,0.97279f,0.97243f,0.97231f,0.97357f,0.97404f,0.97432f,0.97499f,0.97598f,0.97712f,0.97809f,0.97916f,0.98108f,0.98258f,0.98395f,0.98554f,0.98759f,0.98927f,0.99122f,0.99264f,0.99524f,0.99758f,1.00000f,0.00000f,0.90802f,0.89722f,0.90941f,0.92675f,0.94351f,0.95628f,0.96339f,0.97010f,0.97141f,0.97068f,0.97226f,0.97243f,0.97269f,0.97300f,0.97307f,0.97395f,0.97460f,0.97581f,0.97663f,0.97858f,0.97955f,0.98049f,0.98210f,0.98405f,0.98530f,0.98717f,0.98894f,0.99084f,0.99225f,0.99504f,0.99755f,1.00000f,0.00000f,0.91757f,0.89327f,0.89745f,0.90835f,0.92304f,0.93591f,0.94850f,0.95823f,0.96371f,0.96841f,0.97018f,0.97138f,0.97168f,0.97210f,0.97376f,0.97393f,0.97550f,0.97582f,0.97672f,0.97797f,0.97856f,0.98041f,0.98263f,0.98365f,0.98515f,0.98730f,0.98903f,0.99049f,0.99156f,0.99514f,0.99750f,1.00000f,0.00000f,0.93084f,0.90543f,0.89281f,0.89622f,0.90367f,0.91952f,0.92773f,0.94343f,0.94903f,0.95863f,0.96458f,0.96667f,0.96984f,0.97052f,0.97160f,0.97346f,0.97449f,0.97527f,0.97563f,0.97772f,0.97859f,0.98043f,0.98212f,0.98286f,0.98526f,0.98768f,0.98811f,0.98884f,0.99080f,0.99496f,0.99751f,1.00000f,0.00000f,0.92599f,0.90441f,0.89325f,0.89227f,0.89750f,0.90249f,0.91260f,0.92294f,0.93339f,0.94281f,0.95037f,0.95586f,0.96290f,0.96712f,0.96914f,0.97187f,0.97413f,0.97506f,0.97570f,0.97805f,0.97921f,0.98005f,0.98126f,0.98358f,0.98515f,0.98658f,0.98712f,0.98806f,0.99059f,0.99457f,0.99735f,1.00000f,0.00000f,0.92788f,0.91549f,0.89520f,0.88514f,0.89094f,0.89262f,0.89815f,0.90942f,0.91512f,0.92514f,0.93599f,0.94073f,0.94995f,0.95505f,0.96126f,0.96572f,0.97057f,0.97205f,0.97477f,0.97641f,0.97807f,0.97993f,0.98097f,0.98315f,0.98488f,0.98534f,0.98539f,0.98737f,0.99022f,0.99451f,0.99736f,1.00000f,0.00000f,0.93759f,0.91297f,0.90344f,0.88995f,0.89003f,0.88700f,0.89060f,0.89503f,0.90228f,0.90786f,0.91589f,0.92461f,0.93423f,0.94246f,0.94949f,0.95453f,0.96004f,0.96576f,0.96928f,0.97334f,0.97492f,0.97856f,0.98045f,0.98295f,0.98377f,0.98392f,0.98415f,0.98625f,0.98985f,0.99436f,0.99730f,1.00000f,0.00000f,0.94785f,0.92513f,0.90132f,0.89640f,0.88869f,0.88256f,0.88801f,0.88894f,0.89340f,0.89648f,0.90308f,0.90881f,0.91805f,0.92515f,0.93490f,0.93886f,0.94755f,0.95501f,0.95917f,0.96565f,0.96982f,0.97359f,0.97625f,0.97907f,0.98025f,0.98132f,0.98286f,0.98508f,0.99012f,0.99431f,0.99706f,1.00000f,0.00000f,0.93868f,0.92402f,0.91692f,0.90163f,0.88786f,0.88506f,0.87749f,0.88465f,0.88759f,0.88927f,0.89135f,0.89536f,0.90461f,0.91099f,0.91659f,0.92476f,0.93183f,0.93923f,0.94746f,0.95212f,0.95927f,0.96390f,0.96815f,0.97023f,0.97242f,0.97664f,0.98020f,0.98472f,0.98977f,0.99403f,0.99679f,1.00000f,0.00000f,0.94805f,0.93201f,0.91918f,0.90675f,0.89986f,0.88575f,0.88194f,0.88270f,0.88302f,0.88013f,0.88250f,0.88710f,0.89146f,0.89603f,0.90309f,0.90929f,0.91564f,0.92364f,0.93125f,0.93583f,0.94289f,0.94946f,0.95495f,0.95917f,0.96228f,0.96849f,0.97428f,0.98096f,0.98804f,0.99292f,0.99665f,1.00000f,0.00000f,0.93849f,0.93527f,0.92284f,0.91101f,0.89690f,0.89313f,0.88899f,0.88009f,0.87924f,0.87755f,0.87835f,0.88446f,0.88283f,0.88476f,0.88617f,0.89491f,0.90108f,0.90685f,0.91307f,0.91803f,0.92726f,0.93179f,0.93813f,0.94302f,0.94961f,0.95700f,0.96506f,0.97340f,0.98140f,0.98983f,0.99608f,1.00000f,0.00000f,0.95249f,0.93112f,0.92391f,0.90846f,0.90510f,0.89637f,0.88730f,0.88069f,0.87560f,0.87628f,0.87230f,0.87911f,0.87271f,0.87411f,0.87862f,0.88445f,0.88712f,0.89306f,0.89570f,0.90086f,0.90455f,0.90959f,0.91643f,0.92376f,0.93223f,0.94114f,0.95039f,0.96278f,0.97219f,0.98193f,0.99158f,0.99999f,0.00000f,0.95315f,0.93703f,0.92611f,0.91207f,0.90748f,0.89587f,0.88848f,0.88232f,0.87727f,0.87481f,0.87353f,0.87048f,0.87121f,0.86790f,0.86860f,0.87154f,0.87387f,0.87524f,0.87568f,0.88080f,0.88146f,0.88877f,0.89457f,0.90351f,0.91150f,0.92118f,0.93179f,0.94401f,0.95700f,0.96936f,0.98123f,0.99947f,0.00000f,0.94958f,0.93635f,0.92821f,0.91483f,0.90657f,0.90321f,0.88451f,0.88333f,0.87518f,0.87026f,0.86472f,0.86890f,0.86705f,0.86640f,0.85953f,0.86387f,0.85713f,0.85440f,0.85843f,0.85986f,0.86284f,0.86876f,0.87220f,0.87985f,0.88626f,0.89870f,0.91015f,0.92274f,0.93713f,0.94770f,0.96248f,0.99685f,0.00000f,0.94421f,0.94434f,0.93061f,0.92113f,0.91587f,0.90229f,0.89032f,0.88298f,0.88064f,0.87132f,0.86986f,0.86441f,0.85902f,0.85839f,0.85537f,0.85177f,0.84519f,0.84453f,0.84204f,0.84189f,0.84441f,0.84419f,0.85167f,0.85732f,0.86519f,0.87469f,0.88373f,0.89762f,0.90817f,0.92280f,0.93647f,0.98869f,0.00000f,0.94878f,0.93804f,0.92943f,0.92748f,0.90633f,0.90884f,0.89495f,0.89041f,0.87695f,0.87398f,0.86759f,0.86529f,0.85540f,0.85447f,0.84712f,0.83881f,0.83534f,0.82882f,0.82933f,0.82347f,0.82495f,0.82458f,0.83042f,0.83470f,0.84338f,0.84769f,0.85705f,0.86849f,0.87529f,0.88883f,0.89882f,0.96970f,0.00000f,0.94053f,0.93661f,0.93317f,0.91506f,0.91199f,0.90462f,0.89521f,0.89218f,0.88563f,0.87455f,0.86575f,0.85630f,0.85363f,0.84332f,0.83793f,0.82938f,0.81989f,0.81169f,0.81058f,0.81027f,0.80591f,0.80049f,0.80800f,0.81121f,0.81478f,0.82621f,0.82875f,0.83391f,0.84024f,0.84814f,0.85378f,0.93973f,0.00000f,0.94493f,0.93459f,0.93248f,0.91989f,0.91999f,0.90775f,0.90553f,0.88846f,0.88271f,0.87471f,0.86117f,0.85555f,0.84137f,0.83075f,0.82354f,0.81742f,0.81527f,0.80180f,0.79443f,0.79314f,0.78829f,0.79047f,0.79035f,0.78986f,0.79070f,0.79459f,0.79483f,0.79589f,0.80052f,0.80025f,0.80107f,0.90142f,0.00000f,0.93417f,0.93143f,0.92668f,0.92406f,0.90706f,0.91456f,0.89349f,0.89129f,0.87905f,0.86755f,0.86170f,0.84317f,0.83167f,0.81890f,0.81342f,0.80252f,0.79806f,0.79012f,0.78511f,0.77623f,0.77133f,0.77007f,0.76781f,0.76424f,0.76205f,0.76256f,0.75968f,0.75535f,0.75610f,0.75001f,0.74606f,0.85370f,0.00000f,0.93163f,0.93426f,0.93317f,0.91888f,0.91606f,0.91038f,0.89328f,0.89083f,0.87684f,0.86229f,0.84375f,0.83403f,0.82304f,0.81183f,0.80157f,0.79102f,0.77701f,0.76979f,0.76778f,0.75822f,0.75378f,0.75050f,0.74791f,0.73996f,0.73451f,0.72342f,0.72220f,0.71374f,0.70847f,0.69824f,0.68866f,0.79999f,0.00000f,0.93216f,0.93560f,0.92753f,0.91970f,0.90779f,0.90228f,0.88114f,0.88490f,0.86686f,0.85664f,0.84066f,0.82620f,0.81736f,0.79730f,0.78686f,0.77438f,0.76437f,0.75389f,0.74297f,0.73760f,0.73695f,0.72752f,0.72444f,0.71296f,0.70630f,0.69422f,0.68050f,0.67016f,0.65909f,0.64358f,0.62789f,0.74464f,0.00000f,0.93048f,0.92637f,0.92366f,0.91654f,0.90972f,0.89555f,0.88650f,0.87361f,0.85662f,0.83720f,0.83271f,0.81493f,0.79852f,0.77890f,0.76673f,0.75320f,0.74386f,0.73312f,0.72710f,0.71855f,0.71451f,0.70938f,0.69508f,0.68141f,0.67432f,0.65610f,0.64120f,0.62212f,0.60579f,0.59133f,0.57269f,0.68337f,0.00000f,0.92857f,0.92082f,0.91480f,0.91262f,0.89955f,0.89473f,0.87264f,0.86053f,0.85023f,0.82686f,0.81556f,0.79349f,0.78235f,0.76343f,0.75122f,0.73788f,0.72990f,0.71822f,0.70769f,0.70092f,0.68851f,0.67999f,0.66837f,0.65208f,0.63576f,0.61126f,0.59565f,0.57886f,0.55887f,0.53847f,0.51603f,0.62446f,0.00000f,0.91973f,0.91994f,0.90810f,0.90667f,0.89002f,0.88255f,0.87258f,0.84741f,0.83739f,0.81431f,0.79466f,0.77852f,0.75871f,0.74809f,0.73155f,0.72374f,0.70911f,0.70140f,0.68807f,0.67510f,0.67036f,0.65602f,0.64084f,0.61618f,0.59704f,0.57600f,0.55485f,0.53233f,0.51241f,0.48780f,0.46810f,0.56973f,0.00000f,0.90855f,0.91086f,0.90491f,0.89770f,0.88104f,0.86912f,0.85842f,0.83533f,0.81703f,0.79400f,0.77284f,0.75646f,0.73542f,0.72470f,0.70645f,0.69829f,0.68443f,0.67917f,0.66443f,0.65416f,0.63937f,0.62085f,0.60726f,0.58502f,0.56238f,0.53813f,0.51465f,0.48834f,0.46353f,0.43931f,0.41720f,0.51516f,0.00000f,0.91288f,0.90382f,0.89447f,0.88303f,0.87221f,0.85066f,0.84069f,0.81534f,0.79917f,0.77494f,0.75267f,0.73558f,0.71695f,0.70239f,0.68646f,0.67901f,0.66370f,0.65400f,0.64148f,0.62300f,0.61329f,0.59071f,0.56817f,0.55231f,0.52362f,0.49922f,0.47444f,0.44764f,0.42337f,0.39943f,0.37646f,0.46670f,
 0.00000f,1.00000f,0.99831f,1.00000f,0.98999f,0.96955f,0.99022f,0.99477f,0.98422f,0.96616f,0.96545f,0.99918f,0.98206f,0.98952f,0.98571f,0.98467f,0.98264f,0.97237f,0.97063f,0.97587f,0.97333f,0.97166f,0.99142f,0.98732f,0.98553f,0.98684f,0.99140f,0.99077f,0.99332f,0.99504f,0.99620f,0.99652f,1.00000f,0.00000f,0.99360f,0.98774f,0.98399f,0.98134f,0.97719f,0.97667f,0.97573f,0.97491f,0.97295f,0.97415f,0.97421f,0.97399f,0.97461f,0.97426f,0.97539f,0.97584f,0.97457f,0.97634f,0.97814f,0.97841f,0.97929f,0.98166f,0.98272f,0.98427f,0.98592f,0.98775f,0.98956f,0.99149f,0.99349f,0.99558f,0.99754f,1.00000f,0.00000f,0.99152f,0.98601f,0.98302f,0.98016f,0.97785f,0.97648f,0.97538f,0.97551f,0.97481f,0.97375f,0.97239f,0.97412f,0.97361f,0.97399f,0.97458f,0.97532f,0.97627f,0.97705f,0.97772f,0.97872f,0.97984f,0.98188f,0.98292f,0.98417f,0.98603f,0.98763f,0.98931f,0.99147f,0.99340f,0.99548f,0.99770f,1.00000f,0.00000f,0.98338f,0.98577f,0.98168f,0.98002f,0.97914f,0.97702f,0.97457f,0.97463f,0.97360f,0.97432f,0.97382f,0.97338f,0.97355f,0.97456f,0.97423f,0.97479f,0.97539f,0.97633f,0.97819f,0.97829f,0.97994f,0.98130f,0.98276f,0.98409f,0.98569f,0.98789f,0.98934f,0.99133f,0.99338f,0.99544f,0.99778f,1.00000f,0.00000f,0.93779f,0.98261f,0.98295f,0.97911f,0.97692f,0.97596f,0.97493f,0.97395f,0.97373f,0.97326f,0.97381f,0.97319f,0.97413f,0.97401f,0.97442f,0.97493f,0.97565f,0.97654f,0.97782f,0.97925f,0.97985f,0.98130f,0.98234f,0.98410f,0.98582f,0.98738f,0.98964f,0.99144f,0.99342f,0.99530f,0.99772f,1.00000f,0.00000f,0.90926f,0.95411f,0.97495f,0.97676f,0.97724f,0.97539f,0.97450f,0.97433f,0.97363f,0.97292f,0.97164f,0.97274f,0.97437f,0.97335f,0.97479f,0.97532f,0.97554f,0.97611f,0.97702f,0.97807f,0.97987f,0.98107f,0.98257f,0.98415f,0.98568f,0.98756f,0.98927f,0.99122f,0.99345f,0.99530f,0.99771f,1.00000f,0.00000f,0.90145f,0.91995f,0.95482f,0.97105f,0.97536f,0.97437f,0.97381f,0.97317f,0.97308f,0.97254f,0.97115f,0.97376f,0.97286f,0.97310f,0.97382f,0.97473f,0.97551f,0.97654f,0.97774f,0.97797f,0.97967f,0.98116f,0.98290f,0.98408f,0.98554f,0.98761f,0.98937f,0.99122f,0.99320f,0.99526f,0.99765f,1.00000f,0.00000f,0.90255f,0.90249f,0.92600f,0.94948f,0.96653f,0.97232f,0.97132f,0.97331f,0.97233f,0.97148f,0.97336f,0.97317f,0.97294f,0.97352f,0.97441f,0.97383f,0.97548f,0.97681f,0.97748f,0.97806f,0.97951f,0.98079f,0.98243f,0.98389f,0.98584f,0.98699f,0.98932f,0.99125f,0.99312f,0.99509f,0.99755f,1.00000f,0.00000f,0.91261f,0.89595f,0.90804f,0.92659f,0.94518f,0.95693f,0.96516f,0.97002f,0.97137f,0.97138f,0.97141f,0.97252f,0.97284f,0.97247f,0.97445f,0.97447f,0.97530f,0.97575f,0.97676f,0.97822f,0.97969f,0.98063f,0.98227f,0.98331f,0.98583f,0.98722f,0.98900f,0.99099f,0.99226f,0.99509f,0.99760f,1.00000f,0.00000f,0.91590f,0.89833f,0.89555f,0.90873f,0.92143f,0.93543f,0.95135f,0.95873f,0.96363f,0.96873f,0.97066f,0.97197f,0.97178f,0.97218f,0.97370f,0.97426f,0.97379f,0.97621f,0.97626f,0.97818f,0.97905f,0.98082f,0.98182f,0.98354f,0.98535f,0.98701f,0.98882f,0.99048f,0.99125f,0.99503f,0.99744f,1.00000f,0.00000f,0.92971f,0.90219f,0.89022f,0.89451f,0.90615f,0.91808f,0.92771f,0.94075f,0.94925f,0.95863f,0.96388f,0.96559f,0.97015f,0.97204f,0.97302f,0.97375f,0.97453f,0.97490f,0.97689f,0.97755f,0.97931f,0.98055f,0.98192f,0.98274f,0.98505f,0.98671f,0.98840f,0.98969f,0.99086f,0.99500f,0.99756f,1.00000f,0.00000f,0.93704f,0.90475f,0.89188f,0.89078f,0.89469f,0.90596f,0.91180f,0.92265f,0.93304f,0.94333f,0.95021f,0.95682f,0.96299f,0.96720f,0.96973f,0.97170f,0.97270f,0.97555f,0.97704f,0.97666f,0.97817f,0.98080f,0.98172f,0.98299f,0.98416f,0.98700f,0.98735f,0.98840f,0.99074f,0.99463f,0.99737f,1.00000f,0.00000f,0.93595f,0.90809f,0.89814f,0.88928f,0.89153f,0.89221f,0.89867f,0.90426f,0.91734f,0.92785f,0.93433f,0.94280f,0.95165f,0.95614f,0.96073f,0.96579f,0.96799f,0.97214f,0.97487f,0.97631f,0.97864f,0.98080f,0.98146f,0.98374f,0.98486f,0.98523f,0.98666f,0.98670f,0.99023f,0.99446f,0.99725f,1.00000f,0.00000f,0.94058f,0.91041f,0.90229f,0.88887f,0.89222f,0.88643f,0.89236f,0.89499f,0.90166f,0.90702f,0.91570f,0.92686f,0.93294f,0.94303f,0.94808f,0.95549f,0.95933f,0.96523f,0.96929f,0.97173f,0.97549f,0.97777f,0.98091f,0.98298f,0.98345f,0.98393f,0.98479f,0.98646f,0.98973f,0.99381f,0.99733f,1.00000f,0.00000f,0.93549f,0.92388f,0.90782f,0.89699f,0.88879f,0.88792f,0.88608f,0.89221f,0.89037f,0.89833f,0.90434f,0.91253f,0.91834f,0.92753f,0.93247f,0.93987f,0.94944f,0.95445f,0.95977f,0.96550f,0.96959f,0.97367f,0.97735f,0.97908f,0.98085f,0.98173f,0.98299f,0.98570f,0.98914f,0.99402f,0.99726f,1.00000f,0.00000f,0.94237f,0.92618f,0.91557f,0.90011f,0.88522f,0.88633f,0.88590f,0.88359f,0.88335f,0.88582f,0.89004f,0.89737f,0.90092f,0.90851f,0.91696f,0.92319f,0.93196f,0.94031f,0.94546f,0.95378f,0.95782f,0.96327f,0.96893f,0.97140f,0.97470f,0.97670f,0.98037f,0.98430f,0.98896f,0.99400f,0.99694f,1.00000f,0.00000f,0.94351f,0.93374f,0.91567f,0.90487f,0.89090f,0.88889f,0.88401f,0.88438f,0.88020f,0.88339f,0.88272f,0.88535f,0.89408f,0.89894f,0.90228f,0.91083f,0.91492f,0.92171f,0.92893f,0.93811f,0.94437f,0.95060f,0.95600f,0.95982f,0.96456f,0.96804f,0.97393f,0.98155f,0.98783f,0.99301f,0.99702f,1.00000f,0.00000f,0.94657f,0.93695f,0.91972f,0.90987f,0.89571f,0.89195f,0.88997f,0.87914f,0.87659f,0.87828f,0.87765f,0.87998f,0.87945f,0.88138f,0.89044f,0.89207f,0.90102f,0.90648f,0.91398f,0.91780f,0.92496f,0.93163f,0.93793f,0.94389f,0.95101f,0.95682f,0.96490f,0.97438f,0.98260f,0.98984f,0.99601f,1.00000f,0.00000f,0.94716f,0.93423f,0.91953f,0.91407f,0.89727f,0.89293f,0.88705f,0.88077f,0.87702f,0.87502f,0.87455f,0.87439f,0.87613f,0.88010f,0.87690f,0.88246f,0.88780f,0.88963f,0.89544f,0.90165f,0.90552f,0.91204f,0.91540f,0.92519f,0.93127f,0.94060f,0.94982f,0.96136f,0.97176f,0.98174f,0.99106f,0.99999f,0.00000f,0.93903f,0.93656f,0.92700f,0.92156f,0.91229f,0.89644f,0.88636f,0.88416f,0.88040f,0.87417f,0.87032f,0.87257f,0.87137f,0.86853f,0.87015f,0.86888f,0.87166f,0.87309f,0.87347f,0.87681f,0.88504f,0.88889f,0.89843f,0.90196f,0.91009f,0.92002f,0.93159f,0.94450f,0.95853f,0.96841f,0.98062f,0.99949f,0.00000f,0.95528f,0.93748f,0.92825f,0.91435f,0.90599f,0.90169f,0.89174f,0.88203f,0.87781f,0.87159f,0.86849f,0.86935f,0.86420f,0.86174f,0.85877f,0.86405f,0.86090f,0.86066f,0.86123f,0.86156f,0.86242f,0.86875f,0.87555f,0.88121f,0.88861f,0.89744f,0.91038f,0.92266f,0.93635f,0.94844f,0.96298f,0.99682f,0.00000f,0.94543f,0.93760f,0.93502f,0.91634f,0.90901f,0.90146f,0.89694f,0.88344f,0.87770f,0.87492f,0.86715f,0.86765f,0.86420f,0.85651f,0.85429f,0.85292f,0.84786f,0.84397f,0.84499f,0.84163f,0.84376f,0.84462f,0.85058f,0.85982f,0.86823f,0.87502f,0.88462f,0.89785f,0.90816f,0.92181f,0.93543f,0.98886f,0.00000f,0.94317f,0.93710f,0.93449f,0.91867f,0.91121f,0.89944f,0.89555f,0.89455f,0.87421f,0.87198f,0.86711f,0.86141f,0.85264f,0.85185f,0.84688f,0.84207f,0.83454f,0.82993f,0.82719f,0.82400f,0.82421f,0.82578f,0.83057f,0.83444f,0.84073f,0.84833f,0.85908f,0.86763f,0.87761f,0.88790f,0.89879f,0.97043f,0.00000f,0.94276f,0.93639f,0.93167f,0.91593f,0.91261f,0.90762f,0.89311f,0.88962f,0.88125f,0.87290f,0.86556f,0.85436f,0.85230f,0.84302f,0.84099f,0.83019f,0.81928f,0.81564f,0.81087f,0.80896f,0.80627f,0.80808f,0.80983f,0.81233f,0.81569f,0.82127f,0.82721f,0.83349f,0.84074f,0.84742f,0.85438f,0.94107f,0.00000f,0.94405f,0.93782f,0.92776f,0.92126f,0.90802f,0.90592f,0.90497f,0.88316f,0.88196f,0.87803f,0.86752f,0.85588f,0.84799f,0.83576f,0.82555f,0.81760f,0.81088f,0.80369f,0.80269f,0.79488f,0.78888f,0.78778f,0.78990f,0.78835f,0.79072f,0.79232f,0.79190f,0.79952f,0.79863f,0.80130f,0.80390f,0.90139f,0.00000f,0.93633f,0.93635f,0.92891f,0.92102f,0.90935f,0.90296f,0.89984f,0.88769f,0.87315f,0.87130f,0.86219f,0.84973f,0.83511f,0.82471f,0.81117f,0.80134f,0.79485f,0.79411f,0.78262f,0.77788f,0.77355f,0.77172f,0.76438f,0.76904f,0.76029f,0.75991f,0.75894f,0.75700f,0.75331f,0.74977f,0.74656f,0.85280f,0.00000f,0.93717f,0.93129f,0.92399f,0.91796f,0.91463f,0.90398f,0.89548f,0.88822f,0.87710f,0.86805f,0.85155f,0.83221f,0.83196f,0.81427f,0.80081f,0.78819f,0.78178f,0.77357f,0.76290f,0.76521f,0.75575f,0.75091f,0.74641f,0.73898f,0.73186f,0.72609f,0.72257f,0.71428f,0.70726f,0.69806f,0.68763f,0.79780f,0.00000f,0.93706f,0.93721f,0.92654f,0.91957f,0.91298f,0.90171f,0.89063f,0.88042f,0.86944f,0.85447f,0.84470f,0.82886f,0.81590f,0.79687f,0.78900f,0.77427f,0.76920f,0.75415f,0.74873f,0.74489f,0.73187f,0.72960f,0.72001f,0.71027f,0.70612f,0.69442f,0.68348f,0.66914f,0.65651f,0.64453f,0.62799f,0.74398f,0.00000f,0.92840f,0.92836f,0.92876f,0.91210f,0.91080f,0.90114f,0.88652f,0.88254f,0.86203f,0.84727f,0.83019f,0.81168f,0.79531f,0.78675f,0.77241f,0.76108f,0.74658f,0.73821f,0.73019f,0.72549f,0.71308f,0.70716f,0.69474f,0.68678f,0.67447f,0.65500f,0.64218f,0.62099f,0.60896f,0.59012f,0.57226f,0.68470f,0.00000f,0.92765f,0.92622f,0.91832f,0.91573f,0.90871f,0.89441f,0.88029f,0.86340f,0.84773f,0.83144f,0.80762f,0.79508f,0.78580f,0.76666f,0.75560f,0.74252f,0.72714f,0.72034f,0.71531f,0.70163f,0.69237f,0.68400f,0.66993f,0.65225f,0.63499f,0.61630f,0.59876f,0.57635f,0.55618f,0.53755f,0.51836f,0.62549f,0.00000f,0.92393f,0.91514f,0.90929f,0.90474f,0.89064f,0.88373f,0.86857f,0.85255f,0.83276f,0.82073f,0.80432f,0.77828f,0.76070f,0.75117f,0.73256f,0.72600f,0.70941f,0.69856f,0.69011f,0.67541f,0.66553f,0.65616f,0.63801f,0.61721f,0.59780f,0.58010f,0.55562f,0.53082f,0.51435f,0.48819f,0.46499f,0.56971f,0.00000f,0.90551f,0.91313f,0.90012f,0.89761f,0.88212f,0.87348f,0.85997f,0.84285f,0.81498f,0.80298f,0.78554f,0.75664f,0.73991f,0.72584f,0.71040f,0.70147f,0.69041f,0.67798f,0.66953f,0.65225f,0.63985f,0.62561f,0.60393f,0.58416f,0.56114f,0.53918f,0.51348f,0.48709f,0.46403f,0.44155f,0.41951f,0.51661f,0.00000f,0.90387f,0.90480f,0.89639f,0.88401f,0.86950f,0.85768f,0.83901f,0.82353f,0.80240f,0.77630f,0.75568f,0.73438f,0.71830f,0.70543f,0.69054f,0.68093f,0.66458f,0.65303f,0.64260f,0.62676f,0.61053f,0.59203f,0.56910f,0.54587f,0.51969f,0.49589f,0.47220f,0.44857f,0.42153f,0.39896f,0.37558f,0.46510f,
 0.00000f,1.00000f,0.99909f,1.00000f,0.99043f,0.96844f,0.99106f,0.99478f,0.98361f,0.96605f,0.96525f,0.99813f,0.98126f,0.98830f,0.98619f,0.98481f,0.98170f,0.97233f,0.97092f,0.97635f,0.97358f,0.97230f,0.99096f,0.98723f,0.98575f,0.98651f,0.99159f,0.99103f,0.99334f,0.99497f,0.99627f,0.99675f,1.00000f,0.00000f,0.99185f,0.98670f,0.98363f,0.98130f,0.97769f,0.97717f,0.97651f,0.97459f,0.97204f,0.97286f,0.97447f,0.97382f,0.97436f,0.97485f,0.97546f,0.97524f,0.97454f,0.97683f,0.97793f,0.97833f,0.97861f,0.98192f,0.98306f,0.98407f,0.98594f,0.98796f,0.98956f,0.99159f,0.99340f,0.99552f,0.99759f,1.00000f,0.00000f,0.99319f,0.98575f,0.98296f,0.98069f,0.97872f,0.97753f,0.97520f,0.97496f,0.97314f,0.97450f,0.97421f,0.97323f,0.97371f,0.97349f,0.97431f,0.97529f,0.97611f,0.97720f,0.97810f,0.97920f,0.97985f,0.98138f,0.98278f,0.98437f,0.98602f,0.98767f,0.98943f,0.99153f,0.99343f,0.99553f,0.99776f,1.00000f,0.00000f,0.98360f,0.98613f,0.98282f,0.97864f,0.97776f,0.97571f,0.97646f,0.97556f,0.97371f,0.97324f,0.97287f,0.97323f,0.97306f,0.97451f,0.97437f,0.97518f,0.97577f,0.97720f,0.97784f,0.97900f,0.97976f,0.98190f,0.98286f,0.98390f,0.98595f,0.98769f,0.98965f,0.99128f,0.99328f,0.99548f,0.99774f,1.00000f,0.00000f,0.93862f,0.98026f,0.98253f,0.97962f,0.97768f,0.97639f,0.97462f,0.97509f,0.97392f,0.97264f,0.97278f,0.97280f,0.97310f,0.97465f,0.97365f,0.97467f,0.97553f,0.97671f,0.97760f,0.97895f,0.97970f,0.98138f,0.98277f,0.98383f,0.98612f,0.98774f,0.98949f,0.99121f,0.99336f,0.99528f,0.99777f,1.00000f,0.00000f,0.90619f,0.95644f,0.97669f,0.97877f,0.97710f,0.97427f,0.97523f,0.97458f,0.97291f,0.97252f,0.97294f,0.97276f,0.97302f,0.97289f,0.97401f,0.97536f,0.97522f,0.97631f,0.97795f,0.97858f,0.97973f,0.98116f,0.98250f,0.98422f,0.98595f,0.98745f,0.98911f,0.99118f,0.99322f,0.99522f,0.99773f,1.00000f,0.00000f,0.89770f,0.91862f,0.95464f,0.97189f,0.97540f,0.97570f,0.97486f,0.97384f,0.97271f,0.97308f,0.97356f,0.97232f,0.97241f,0.97389f,0.97456f,0.97472f,0.97575f,0.97609f,0.97747f,0.97841f,0.97960f,0.98140f,0.98262f,0.98400f,0.98578f,0.98782f,0.98910f,0.99111f,0.99331f,0.99524f,0.99761f,1.00000f,0.00000f,0.90005f,0.90349f,0.92632f,0.94898f,0.96239f,0.97042f,0.97219f,0.97285f,0.97250f,0.97272f,0.97228f,0.97262f,0.97284f,0.97287f,0.97373f,0.97304f,0.97597f,0.97657f,0.97717f,0.97805f,0.98000f,0.98085f,0.98197f,0.98415f,0.98589f,0.98730f,0.98938f,0.99116f,0.99293f,0.99505f,0.99755f,1.00000f,0.00000f,0.90654f,0.89554f,0.90768f,0.92552f,0.94441f,0.95761f,0.96433f,0.96993f,0.97115f,0.97249f,0.97236f,0.97122f,0.97212f,0.97307f,0.97338f,0.97433f,0.97511f,0.97598f,0.97739f,0.97864f,0.97970f,0.98121f,0.98171f,0.98378f,0.98527f,0.98727f,0.98911f,0.99106f,0.99223f,0.99487f,0.99751f,1.00000f,0.00000f,0.91286f,0.89775f,0.89756f,0.90682f,0.92274f,0.94040f,0.94877f,0.95725f,0.96504f,0.96687f,0.96819f,0.96999f,0.97137f,0.97229f,0.97238f,0.97417f,0.97515f,0.97648f,0.97723f,0.97792f,0.97866f,0.98077f,0.98246f,0.98353f,0.98570f,0.98712f,0.98867f,0.99063f,0.99143f,0.99488f,0.99744f,1.00000f,0.00000f,0.92488f,0.89840f,0.89986f,0.90123f,0.90628f,0.91632f,0.93135f,0.93981f,0.95067f,0.95877f,0.96186f,0.96580f,0.97112f,0.97101f,0.97262f,0.97387f,0.97440f,0.97598f,0.97674f,0.97771f,0.97852f,0.98010f,0.98205f,0.98356f,0.98481f,0.98690f,0.98856f,0.98973f,0.99095f,0.99460f,0.99742f,1.00000f,0.00000f,0.93024f,0.90776f,0.89418f,0.88923f,0.89568f,0.90371f,0.91221f,0.92225f,0.93348f,0.94028f,0.95096f,0.95529f,0.96239f,0.96729f,0.96907f,0.97192f,0.97307f,0.97583f,0.97626f,0.97841f,0.97845f,0.98054f,0.98190f,0.98271f,0.98484f,0.98659f,0.98771f,0.98887f,0.99051f,0.99442f,0.99737f,1.00000f,0.00000f,0.93528f,0.91211f,0.89807f,0.89009f,0.89126f,0.89218f,0.89919f,0.90528f,0.91205f,0.92474f,0.93216f,0.94068f,0.94947f,0.95729f,0.96139f,0.96576f,0.96827f,0.97152f,0.97432f,0.97589f,0.97791f,0.98087f,0.98186f,0.98296f,0.98535f,0.98558f,0.98577f,0.98722f,0.98963f,0.99453f,0.99737f,1.00000f,0.00000f,0.93699f,0.91520f,0.89928f,0.89510f,0.88831f,0.88751f,0.89177f,0.89602f,0.90015f,0.90955f,0.91804f,0.92759f,0.93372f,0.94042f,0.94693f,0.95421f,0.95992f,0.96500f,0.96826f,0.97193f,0.97487f,0.97801f,0.98067f,0.98239f,0.98405f,0.98417f,0.98488f,0.98629f,0.98934f,0.99427f,0.99737f,1.00000f,0.00000f,0.94558f,0.91959f,0.90558f,0.89772f,0.88589f,0.88664f,0.89017f,0.88919f,0.89057f,0.89611f,0.90270f,0.90991f,0.91838f,0.92429f,0.93385f,0.94006f,0.94891f,0.95599f,0.96013f,0.96424f,0.97085f,0.97224f,0.97659f,0.97962f,0.98022f,0.98252f,0.98340f,0.98574f,0.98945f,0.99422f,0.99725f,1.00000f,0.00000f,0.93638f,0.92415f,0.90956f,0.89672f,0.88583f,0.88487f,0.88466f,0.88382f,0.88421f,0.88753f,0.89074f,0.89577f,0.90090f,0.90865f,0.91884f,0.92458f,0.93170f,0.93977f,0.94615f,0.95306f,0.95888f,0.96418f,0.96847f,0.97258f,0.97516f,0.97758f,0.98065f,0.98398f,0.98919f,0.99392f,0.99699f,1.00000f,0.00000f,0.93861f,0.92869f,0.91555f,0.90055f,0.89269f,0.88639f,0.88289f,0.88326f,0.88092f,0.88200f,0.88446f,0.88797f,0.88956f,0.89652f,0.90486f,0.90988f,0.91559f,0.92334f,0.93142f,0.93744f,0.94300f,0.95154f,0.95512f,0.96022f,0.96451f,0.96951f,0.97497f,0.98050f,0.98808f,0.99260f,0.99663f,1.00000f,0.00000f,0.94262f,0.93513f,0.92018f,0.90716f,0.89832f,0.88790f,0.88410f,0.87894f,0.87866f,0.87605f,0.87949f,0.87796f,0.88709f,0.88660f,0.88995f,0.89480f,0.89974f,0.90649f,0.91265f,0.92131f,0.92691f,0.93274f,0.93957f,0.94338f,0.95001f,0.95751f,0.96540f,0.97426f,0.98237f,0.98947f,0.99612f,1.00000f,0.00000f,0.94168f,0.93355f,0.92319f,0.91018f,0.90439f,0.89341f,0.88678f,0.88697f,0.87820f,0.87538f,0.87094f,0.87294f,0.87759f,0.87455f,0.87710f,0.88343f,0.88636f,0.88888f,0.89322f,0.90023f,0.90444f,0.91229f,0.91725f,0.92391f,0.93361f,0.94046f,0.95151f,0.96107f,0.97226f,0.98191f,0.99086f,0.99999f,0.00000f,0.94236f,0.93533f,0.92770f,0.91651f,0.90530f,0.89762f,0.89207f,0.88518f,0.87408f,0.87802f,0.87629f,0.87028f,0.86854f,0.87203f,0.87209f,0.87140f,0.87238f,0.87437f,0.87854f,0.87976f,0.88403f,0.89217f,0.89761f,0.90392f,0.91197f,0.92264f,0.93121f,0.94546f,0.95552f,0.96932f,0.98059f,0.99949f,0.00000f,0.94483f,0.93811f,0.93858f,0.92263f,0.90608f,0.89453f,0.88562f,0.88642f,0.87940f,0.87746f,0.87171f,0.86746f,0.86176f,0.85779f,0.86349f,0.86316f,0.85841f,0.85639f,0.86043f,0.86278f,0.86441f,0.86861f,0.87432f,0.88094f,0.88760f,0.90195f,0.91095f,0.92218f,0.93546f,0.94876f,0.96224f,0.99680f,0.00000f,0.95002f,0.93979f,0.92981f,0.91966f,0.91349f,0.89940f,0.89662f,0.89040f,0.87899f,0.87327f,0.86842f,0.86158f,0.86024f,0.85807f,0.85433f,0.85599f,0.84733f,0.84934f,0.84196f,0.84343f,0.84427f,0.84702f,0.85077f,0.85778f,0.86469f,0.87625f,0.88431f,0.89632f,0.90791f,0.92006f,0.93561f,0.98882f,0.00000f,0.94501f,0.94030f,0.93571f,0.92100f,0.90673f,0.90330f,0.89114f,0.89187f,0.88193f,0.87128f,0.87027f,0.86590f,0.85646f,0.85024f,0.84743f,0.84677f,0.83778f,0.83035f,0.82946f,0.82402f,0.82249f,0.82600f,0.83216f,0.83416f,0.84046f,0.84948f,0.85751f,0.86583f,0.87719f,0.88930f,0.89969f,0.97010f,0.00000f,0.94567f,0.93831f,0.92658f,0.92008f,0.91019f,0.90181f,0.89440f,0.88881f,0.87872f,0.87150f,0.86352f,0.86330f,0.85424f,0.84350f,0.83968f,0.83489f,0.82514f,0.82155f,0.81346f,0.80482f,0.80908f,0.80639f,0.81152f,0.81274f,0.81431f,0.82233f,0.82975f,0.83504f,0.84291f,0.84726f,0.85261f,0.94052f,0.00000f,0.93824f,0.93505f,0.93420f,0.92798f,0.91883f,0.90836f,0.89232f,0.89110f,0.88059f,0.86946f,0.85924f,0.85648f,0.84560f,0.83509f,0.82975f,0.82305f,0.81056f,0.80268f,0.79961f,0.79226f,0.79058f,0.78817f,0.79091f,0.79211f,0.79264f,0.79330f,0.79376f,0.79755f,0.79866f,0.80194f,0.80091f,0.90048f,0.00000f,0.93924f,0.93269f,0.93432f,0.92511f,0.91484f,0.90682f,0.89557f,0.88798f,0.87903f,0.86749f,0.85534f,0.85004f,0.83890f,0.82974f,0.81174f,0.81067f,0.80066f,0.79367f,0.78440f,0.77919f,0.77712f,0.76827f,0.76476f,0.76526f,0.76361f,0.75975f,0.76110f,0.75762f,0.75532f,0.74968f,0.74627f,0.85362f,0.00000f,0.93769f,0.94401f,0.93040f,0.91805f,0.91172f,0.90654f,0.89785f,0.88524f,0.88316f,0.86879f,0.85430f,0.83845f,0.82731f,0.81408f,0.80271f,0.79157f,0.78375f,0.77473f,0.77100f,0.76482f,0.75375f,0.75263f,0.74727f,0.74454f,0.73417f,0.72823f,0.72159f,0.71538f,0.70766f,0.69796f,0.68721f,0.79968f,0.00000f,0.93782f,0.93304f,0.92812f,0.92123f,0.91137f,0.89792f,0.89298f,0.88557f,0.86736f,0.85824f,0.83791f,0.82722f,0.81515f,0.80082f,0.78715f,0.77692f,0.76791f,0.75882f,0.75009f,0.74114f,0.73331f,0.73250f,0.72160f,0.71509f,0.70094f,0.69283f,0.67975f,0.67108f,0.65575f,0.64496f,0.63134f,0.74151f,0.00000f,0.92521f,0.92902f,0.92268f,0.91738f,0.91049f,0.89995f,0.88970f,0.87538f,0.86561f,0.85215f,0.84071f,0.81976f,0.80257f,0.78170f,0.77249f,0.76024f,0.74988f,0.74170f,0.73205f,0.72473f,0.71323f,0.70818f,0.69651f,0.68495f,0.67283f,0.65562f,0.63985f,0.62143f,0.60425f,0.59005f,0.57461f,0.68468f,0.00000f,0.92589f,0.91806f,0.91338f,0.91235f,0.90725f,0.88778f,0.88037f,0.86529f,0.85680f,0.83615f,0.81244f,0.80116f,0.78578f,0.76805f,0.75720f,0.73943f,0.73010f,0.72151f,0.70931f,0.70249f,0.69230f,0.68306f,0.66853f,0.65377f,0.64065f,0.61454f,0.59818f,0.57808f,0.55621f,0.53657f,0.52090f,0.62730f,0.00000f,0.92208f,0.91294f,0.91172f,0.90880f,0.89625f,0.88289f,0.86666f,0.85330f,0.84081f,0.82541f,0.79915f,0.77914f,0.76572f,0.74776f,0.73851f,0.72649f,0.71289f,0.70140f,0.69219f,0.67619f,0.66680f,0.65591f,0.63724f,0.62287f,0.59760f,0.57672f,0.55198f,0.53012f,0.50926f,0.48824f,0.46724f,0.56759f,0.00000f,0.91123f,0.91071f,0.90359f,0.89360f,0.88455f,0.86931f,0.85952f,0.83746f,0.81673f,0.80415f,0.78002f,0.76385f,0.74313f,0.72549f,0.71578f,0.70379f,0.68868f,0.68091f,0.66984f,0.65251f,0.63965f,0.62406f,0.60626f,0.58361f,0.56180f,0.53591f,0.51157f,0.48904f,0.46645f,0.44277f,0.42062f,0.51598f,0.00000f,0.90727f,0.90161f,0.90181f,0.88944f,0.87600f,0.85861f,0.83709f,0.82896f,0.80060f,0.77885f,0.76136f,0.74101f,0.72227f,0.70802f,0.69315f,0.67854f,0.66768f,0.65527f,0.64059f,0.62752f,0.61084f,0.59472f,0.56878f,0.54491f,0.52132f,0.49451f,0.47019f,0.44525f,0.42389f,0.39946f,0.37728f,0.46519f,
};

// 33x33 average albedo table calculated using moonshine_devtools/compute_albedo for the following ranges:
// F*Reflection + (1-F)*Transmission
// EtaT/EtaI [1, 1/3]
// Roughness [0,1]
const float
TransmissionAlbedo::mOneMinusEavgExit[COMP*COMP] =
{
 0.02804f,0.01759f,0.01569f,0.01559f,0.01576f,0.01622f,0.01674f,0.01763f,0.01898f,0.02111f,0.02416f,0.02802f,0.03321f,0.03921f,0.04604f,0.05311f,0.06013f,0.06747f,0.07428f,0.08099f,0.08799f,0.09424f,0.10080f,0.10620f,0.11381f,0.12030f,0.12725f,0.13438f,0.14140f,0.14883f,0.15683f,0.16362f,0.17126f,
 0.02211f,0.01695f,0.01565f,0.01563f,0.01581f,0.01613f,0.01677f,0.01762f,0.01894f,0.02091f,0.02363f,0.02708f,0.03213f,0.03841f,0.04626f,0.05576f,0.06597f,0.07739f,0.08936f,0.10150f,0.11329f,0.12490f,0.13671f,0.14749f,0.15819f,0.16869f,0.17846f,0.18995f,0.19914f,0.20937f,0.21898f,0.22869f,0.23874f,
 0.02070f,0.01676f,0.01561f,0.01560f,0.01572f,0.01615f,0.01666f,0.01751f,0.01870f,0.02062f,0.02321f,0.02664f,0.03120f,0.03728f,0.04489f,0.05436f,0.06553f,0.07848f,0.09230f,0.10697f,0.12216f,0.13793f,0.15270f,0.16768f,0.18276f,0.19716f,0.21101f,0.22413f,0.23733f,0.25010f,0.26320f,0.27626f,0.28944f,
 0.01941f,0.01637f,0.01563f,0.01564f,0.01583f,0.01619f,0.01662f,0.01737f,0.01864f,0.02040f,0.02284f,0.02619f,0.03072f,0.03651f,0.04364f,0.05322f,0.06422f,0.07680f,0.09150f,0.10776f,0.12440f,0.14183f,0.15994f,0.17772f,0.19507f,0.21253f,0.22945f,0.24653f,0.26312f,0.27920f,0.29558f,0.31113f,0.32632f,
 0.01881f,0.01614f,0.01558f,0.01562f,0.01584f,0.01618f,0.01672f,0.01738f,0.01858f,0.02023f,0.02281f,0.02580f,0.03022f,0.03590f,0.04295f,0.05197f,0.06269f,0.07546f,0.08999f,0.10688f,0.12436f,0.14368f,0.16269f,0.18137f,0.20162f,0.22101f,0.24154f,0.26070f,0.28088f,0.29991f,0.31824f,0.33611f,0.35516f,
 0.01779f,0.01609f,0.01555f,0.01563f,0.01576f,0.01605f,0.01664f,0.01733f,0.01853f,0.02031f,0.02261f,0.02567f,0.02997f,0.03564f,0.04221f,0.05079f,0.06148f,0.07373f,0.08805f,0.10458f,0.12238f,0.14167f,0.16206f,0.18304f,0.20516f,0.22687f,0.24875f,0.27091f,0.29303f,0.31464f,0.33658f,0.35567f,0.37660f,
 0.01675f,0.01598f,0.01562f,0.01565f,0.01579f,0.01620f,0.01655f,0.01749f,0.01845f,0.02015f,0.02244f,0.02570f,0.02970f,0.03506f,0.04157f,0.04995f,0.05977f,0.07235f,0.08631f,0.10290f,0.12052f,0.14041f,0.16065f,0.18330f,0.20622f,0.23026f,0.25292f,0.27686f,0.30142f,0.32492f,0.34780f,0.37179f,0.39291f,
 0.01605f,0.01582f,0.01547f,0.01568f,0.01573f,0.01606f,0.01668f,0.01735f,0.01841f,0.01987f,0.02236f,0.02546f,0.02946f,0.03461f,0.04111f,0.04936f,0.05951f,0.07079f,0.08454f,0.10090f,0.11883f,0.13798f,0.15947f,0.18220f,0.20636f,0.23084f,0.25623f,0.28143f,0.30745f,0.33282f,0.35696f,0.38112f,0.40462f,
 0.01522f,0.01576f,0.01556f,0.01560f,0.01584f,0.01611f,0.01662f,0.01734f,0.01854f,0.01995f,0.02229f,0.02521f,0.02906f,0.03446f,0.04069f,0.04898f,0.05850f,0.06991f,0.08372f,0.09930f,0.11713f,0.13652f,0.15803f,0.18089f,0.20539f,0.23105f,0.25670f,0.28300f,0.31022f,0.33626f,0.36253f,0.38798f,0.41350f,
 0.01497f,0.01569f,0.01559f,0.01568f,0.01584f,0.01601f,0.01657f,0.01734f,0.01833f,0.02000f,0.02225f,0.02532f,0.02925f,0.03401f,0.04056f,0.04819f,0.05776f,0.06898f,0.08278f,0.09766f,0.11572f,0.13486f,0.15584f,0.17942f,0.20341f,0.22963f,0.25620f,0.28413f,0.31212f,0.33956f,0.36664f,0.39315f,0.41983f,
 0.01475f,0.01571f,0.01561f,0.01561f,0.01582f,0.01612f,0.01659f,0.01734f,0.01849f,0.01992f,0.02214f,0.02524f,0.02895f,0.03401f,0.04002f,0.04786f,0.05732f,0.06844f,0.08138f,0.09664f,0.11386f,0.13318f,0.15445f,0.17734f,0.20246f,0.22870f,0.25582f,0.28347f,0.31195f,0.34105f,0.36930f,0.39621f,0.42430f,
 0.01423f,0.01559f,0.01557f,0.01565f,0.01578f,0.01616f,0.01662f,0.01739f,0.01848f,0.01982f,0.02202f,0.02484f,0.02883f,0.03350f,0.03988f,0.04754f,0.05652f,0.06733f,0.08062f,0.09564f,0.11210f,0.13135f,0.15283f,0.17505f,0.20159f,0.22723f,0.25520f,0.28285f,0.31251f,0.34108f,0.37016f,0.39982f,0.42747f,
 0.01429f,0.01557f,0.01546f,0.01559f,0.01576f,0.01600f,0.01654f,0.01733f,0.01837f,0.01994f,0.02211f,0.02484f,0.02864f,0.03360f,0.03969f,0.04708f,0.05658f,0.06701f,0.08004f,0.09434f,0.11099f,0.13039f,0.15140f,0.17376f,0.19953f,0.22583f,0.25339f,0.28305f,0.31117f,0.34142f,0.37090f,0.40037f,0.42894f,
 0.01378f,0.01556f,0.01555f,0.01565f,0.01579f,0.01616f,0.01649f,0.01738f,0.01833f,0.01995f,0.02201f,0.02507f,0.02841f,0.03315f,0.03938f,0.04672f,0.05549f,0.06685f,0.07900f,0.09381f,0.11025f,0.12868f,0.14986f,0.17251f,0.19708f,0.22449f,0.25224f,0.28122f,0.30919f,0.34117f,0.37131f,0.40068f,0.42989f,
 0.01373f,0.01546f,0.01555f,0.01554f,0.01571f,0.01605f,0.01664f,0.01728f,0.01839f,0.01989f,0.02211f,0.02495f,0.02864f,0.03323f,0.03950f,0.04681f,0.05518f,0.06614f,0.07847f,0.09287f,0.10952f,0.12747f,0.14779f,0.17097f,0.19566f,0.22254f,0.25057f,0.28034f,0.30969f,0.33987f,0.37040f,0.40058f,0.43112f,
 0.01393f,0.01558f,0.01563f,0.01555f,0.01580f,0.01597f,0.01668f,0.01735f,0.01837f,0.01987f,0.02204f,0.02486f,0.02846f,0.03291f,0.03906f,0.04618f,0.05502f,0.06527f,0.07805f,0.09206f,0.10830f,0.12664f,0.14730f,0.16984f,0.19411f,0.22119f,0.24965f,0.27822f,0.30846f,0.33908f,0.37039f,0.40106f,0.43118f,
 0.01365f,0.01552f,0.01549f,0.01563f,0.01583f,0.01607f,0.01656f,0.01741f,0.01836f,0.01995f,0.02193f,0.02463f,0.02825f,0.03309f,0.03895f,0.04610f,0.05474f,0.06510f,0.07790f,0.09127f,0.10797f,0.12613f,0.14544f,0.16842f,0.19302f,0.21971f,0.24737f,0.27720f,0.30746f,0.33860f,0.36961f,0.40067f,0.43244f,
 0.01345f,0.01560f,0.01553f,0.01564f,0.01584f,0.01616f,0.01646f,0.01734f,0.01840f,0.01988f,0.02198f,0.02475f,0.02812f,0.03283f,0.03867f,0.04613f,0.05434f,0.06468f,0.07727f,0.09066f,0.10672f,0.12442f,0.14473f,0.16750f,0.19234f,0.21820f,0.24703f,0.27533f,0.30578f,0.33716f,0.36755f,0.39984f,0.43098f,
 0.01350f,0.01557f,0.01547f,0.01556f,0.01575f,0.01611f,0.01655f,0.01727f,0.01840f,0.01971f,0.02194f,0.02462f,0.02824f,0.03278f,0.03864f,0.04556f,0.05416f,0.06497f,0.07685f,0.09017f,0.10611f,0.12406f,0.14384f,0.16603f,0.19026f,0.21691f,0.24498f,0.27466f,0.30453f,0.33634f,0.36746f,0.39945f,0.43036f,
 0.01353f,0.01550f,0.01554f,0.01558f,0.01586f,0.01607f,0.01658f,0.01743f,0.01844f,0.01979f,0.02204f,0.02460f,0.02822f,0.03275f,0.03833f,0.04554f,0.05411f,0.06411f,0.07636f,0.08970f,0.10541f,0.12271f,0.14310f,0.16467f,0.18939f,0.21562f,0.24401f,0.27307f,0.30434f,0.33522f,0.36730f,0.39831f,0.42965f,
 0.01335f,0.01549f,0.01546f,0.01566f,0.01581f,0.01613f,0.01655f,0.01726f,0.01845f,0.01978f,0.02176f,0.02472f,0.02819f,0.03262f,0.03782f,0.04548f,0.05380f,0.06412f,0.07581f,0.08939f,0.10497f,0.12239f,0.14177f,0.16335f,0.18845f,0.21506f,0.24215f,0.27216f,0.30209f,0.33409f,0.36567f,0.39817f,0.42972f,
 0.01336f,0.01550f,0.01548f,0.01558f,0.01586f,0.01607f,0.01660f,0.01733f,0.01828f,0.01984f,0.02182f,0.02457f,0.02804f,0.03272f,0.03816f,0.04511f,0.05348f,0.06332f,0.07570f,0.08908f,0.10457f,0.12180f,0.14087f,0.16306f,0.18703f,0.21338f,0.24171f,0.27064f,0.30134f,0.33324f,0.36500f,0.39715f,0.42849f,
 0.01324f,0.01548f,0.01557f,0.01558f,0.01577f,0.01608f,0.01656f,0.01735f,0.01844f,0.01985f,0.02179f,0.02451f,0.02808f,0.03235f,0.03805f,0.04518f,0.05368f,0.06360f,0.07535f,0.08897f,0.10405f,0.12108f,0.14056f,0.16237f,0.18654f,0.21210f,0.24078f,0.26969f,0.30070f,0.33110f,0.36276f,0.39571f,0.42775f,
 0.01314f,0.01551f,0.01549f,0.01561f,0.01588f,0.01597f,0.01663f,0.01739f,0.01827f,0.01979f,0.02180f,0.02447f,0.02786f,0.03256f,0.03798f,0.04510f,0.05350f,0.06324f,0.07513f,0.08825f,0.10345f,0.12093f,0.14011f,0.16084f,0.18583f,0.21147f,0.23939f,0.26886f,0.29916f,0.33136f,0.36329f,0.39500f,0.42704f,
 0.01319f,0.01549f,0.01548f,0.01562f,0.01584f,0.01609f,0.01657f,0.01731f,0.01836f,0.01983f,0.02197f,0.02453f,0.02813f,0.03235f,0.03781f,0.04493f,0.05287f,0.06308f,0.07468f,0.08815f,0.10296f,0.12003f,0.13883f,0.16100f,0.18482f,0.21029f,0.23785f,0.26819f,0.29749f,0.33035f,0.36146f,0.39459f,0.42700f,
 0.01338f,0.01544f,0.01554f,0.01560f,0.01593f,0.01610f,0.01657f,0.01725f,0.01829f,0.01968f,0.02169f,0.02436f,0.02794f,0.03245f,0.03784f,0.04496f,0.05297f,0.06282f,0.07458f,0.08784f,0.10259f,0.12014f,0.13843f,0.16045f,0.18410f,0.20966f,0.23772f,0.26738f,0.29724f,0.32935f,0.36120f,0.39299f,0.42590f,
 0.01324f,0.01547f,0.01555f,0.01558f,0.01588f,0.01600f,0.01654f,0.01725f,0.01838f,0.01979f,0.02186f,0.02441f,0.02794f,0.03214f,0.03803f,0.04456f,0.05330f,0.06240f,0.07361f,0.08745f,0.10233f,0.11867f,0.13902f,0.15951f,0.18298f,0.20914f,0.23706f,0.26625f,0.29594f,0.32735f,0.35992f,0.39249f,0.42504f,
 0.01305f,0.01551f,0.01558f,0.01561f,0.01576f,0.01603f,0.01651f,0.01736f,0.01820f,0.01963f,0.02171f,0.02454f,0.02769f,0.03224f,0.03771f,0.04473f,0.05286f,0.06279f,0.07406f,0.08740f,0.10221f,0.11879f,0.13747f,0.15913f,0.18227f,0.20864f,0.23611f,0.26499f,0.29561f,0.32702f,0.35944f,0.39116f,0.42517f,
 0.01309f,0.01545f,0.01550f,0.01560f,0.01583f,0.01618f,0.01657f,0.01741f,0.01820f,0.01969f,0.02166f,0.02446f,0.02792f,0.03223f,0.03766f,0.04469f,0.05279f,0.06263f,0.07366f,0.08702f,0.10158f,0.11846f,0.13715f,0.15877f,0.18147f,0.20742f,0.23486f,0.26482f,0.29437f,0.32605f,0.35896f,0.39118f,0.42314f,
 0.01310f,0.01544f,0.01553f,0.01562f,0.01580f,0.01612f,0.01659f,0.01720f,0.01833f,0.01978f,0.02171f,0.02437f,0.02799f,0.03233f,0.03787f,0.04422f,0.05244f,0.06219f,0.07362f,0.08717f,0.10150f,0.11800f,0.13651f,0.15859f,0.18149f,0.20791f,0.23461f,0.26327f,0.29419f,0.32555f,0.35806f,0.39050f,0.42319f,
 0.01296f,0.01549f,0.01561f,0.01557f,0.01591f,0.01605f,0.01658f,0.01728f,0.01836f,0.01972f,0.02177f,0.02426f,0.02778f,0.03217f,0.03761f,0.04449f,0.05257f,0.06191f,0.07319f,0.08639f,0.10171f,0.11781f,0.13641f,0.15789f,0.18057f,0.20642f,0.23389f,0.26276f,0.29395f,0.32562f,0.35653f,0.39005f,0.42161f,
 0.01298f,0.01549f,0.01550f,0.01563f,0.01577f,0.01616f,0.01659f,0.01720f,0.01824f,0.01974f,0.02173f,0.02425f,0.02762f,0.03225f,0.03726f,0.04463f,0.05225f,0.06212f,0.07337f,0.08652f,0.10092f,0.11737f,0.13662f,0.15766f,0.18009f,0.20635f,0.23350f,0.26230f,0.29224f,0.32406f,0.35616f,0.38859f,0.42187f,
 0.01300f,0.01553f,0.01546f,0.01558f,0.01583f,0.01617f,0.01651f,0.01732f,0.01827f,0.01978f,0.02165f,0.02428f,0.02792f,0.03229f,0.03742f,0.04445f,0.05226f,0.06168f,0.07340f,0.08568f,0.10090f,0.11747f,0.13603f,0.15684f,0.18009f,0.20527f,0.23208f,0.26216f,0.29170f,0.32337f,0.35584f,0.38814f,0.42102f,
};

//----------------------------------------------------------------------------

} // namespace shading
} // namespace moonray

